#include <ply/ply.h>
#include <utils/PMeshLite_Ply.H>

inline int NBytes( const int in_iType )
{
    int iSizeElem = 0;
    switch ( in_iType ) {
        case Uint8 :
        case Int8 : iSizeElem = 1; break;
        case Uint16 :
        case Int16 : iSizeElem = 2; break;
        case Uint32 :
        case Float32 :
        case Int32 : iSizeElem = 4; break;
        case Float64 : iSizeElem = 8; break;
        default : ASSERT(FALSE);
    }
    return iSizeElem;
}

inline WINbool IsInt( const int in_iType )
{
    int iSizeElem = 0;
    switch ( in_iType ) {
        case Uint8 :
        case Int8 : 
        case Uint16 :
        case Int16 : 
        case Uint32 :
        case Int32 : return TRUE;
        case Float32 :
        case Float64 : return FALSE;
        default : ASSERT(FALSE);
    }
    return FALSE;
}

inline int GetInt( const int in_iType, const int in_iStorage, FILE *file )
{
    const int iSize = NBytes( in_iType );
    char acData[50];
    int iRet = 0;

    if ( in_iStorage == PLY_BINARY_BE ) {
        const size_t iRead = fread( acData, iSize, 1, file );
        ASSERT( iRead == 1 );
        for ( int i = 0; i < iSize / 2; i++ ) {
            char cSave = acData[i];
            acData[i] = acData[iSize - i - 1];
            acData[iSize - i - 1] = cSave;
        }
    } else if ( in_iStorage == PLY_BINARY_LE ) {
        const size_t iRead = fread( acData, iSize, 1, file );
        ASSERT( iRead == 1 );
    } else if ( in_iStorage == PLY_ASCII ) {
        fscanf( file, "%d", &iRet );
        return iRet;
    } else {
        ASSERT(FALSE);
    }

    switch ( in_iType ) {
        case Uint8   : iRet = *( (unsigned char *) acData ); break;
        case Int8    : iRet = *( (char *) acData ); break;
        case Uint16  : iRet = *( (unsigned short *) acData ); break;
        case Int16   : iRet = *( (short *) acData ); break;
        case Uint32  : iRet = *( (unsigned int *) acData ); break;
        case Int32   : iRet = *( (int *) acData ); break;
        case Float64 : 
        case Float32 :
        default : ASSERT(FALSE);
    }
    return iRet;
}

inline double GetDouble( const int in_iType, const int in_iStorage, FILE *file )
{
    const int iSize = NBytes( in_iType );
    char acData[16];
    double dRet = 0;

    if ( in_iStorage == PLY_BINARY_BE ) {
        const size_t iRead = fread( acData, iSize, 1, file );
        ASSERT( iRead == 1 );
        for ( int i = 0; i < iSize / 2; i++ ) {
            char cSave = acData[i];
            acData[i] = acData[iSize - i - 1];
            acData[iSize - i - 1] = cSave;
        }
    } else if ( in_iStorage == PLY_BINARY_LE ) {
        const size_t iRead = fread( acData, iSize, 1, file );
        ASSERT( iRead == 1 );
    } else if ( in_iStorage == PLY_ASCII ) {
        fscanf( file, "%lf", &dRet );
        return dRet;
    } else {
        ASSERT(FALSE);
    }

    switch ( in_iType ) {
        case Float64 : dRet = *( (double *) acData ); break;
        case Float32 : dRet = *( (float *) acData ); break;
        case Uint8   : 
        case Int8    : 
        case Uint16  : 
        case Int16   :
        case Uint32  : 
        case Int32   : 
        default : ASSERT(FALSE);
    }
    return dRet;
}


void PLYReadPlyToPMeshLite( const char * in_strFName, PMeshLite &out_mesh, const WINbool in_bReverse )
{
    FILE *file = fopen( in_strFName, "rb");

    if ( file == NULL ) {
        cerr << "Bad file name " << in_strFName << "\n";
        return;
    }

    out_mesh.Clear();

    TRACE("Read ply %s reverse %s\n", in_strFName, in_bReverse ? "Yes" : "No");

    Array<Vertex> av, avF(3), avFTri(3);
    Array< Array<Vertex> > aavFFailed;
    R3Pt pt;
    int iVal = 0;
    double dVal = 0;

    PlyFile *ply = read_ply( file );

    int iFailed = 0;
    for (int i = 0; i < ply->num_elem_types; i++) {
        const int iNElems = ply->elems[i]->num;
        if ( ply->elems[i]->props[0]->is_list == 1 ) {
            const WINbool bIsFace = !strcmp( ply->elems[i]->name, "face" );
            for ( int iElem = 0; iElem < iNElems; iElem++ ) {
                const int iNGet = GetInt( Uint8, ply->file_type, file );
                if ( bIsFace ) avF.need( iNGet );
                for ( int j = 0; j < iNGet; j++ ) {
                    if ( IsInt( ply->elems[i]->props[0]->external_type ) ) {
                        iVal = GetInt( ply->elems[i]->props[0]->external_type, ply->file_type, file );
                    } else {
                        dVal = GetDouble( ply->elems[i]->props[0]->external_type, ply->file_type, file );
                    }
                    if ( bIsFace ) 
                        avF[j] = av[ iVal ];
                }
                ASSERT( !ferror( file ) );
                if ( bIsFace ) {
                    if ( avF.num() == 3 ) {
                        if ( in_bReverse ) avF.reverse();

                        if ( out_mesh.IsOkToAdd( avF ) ) {
                            out_mesh.AddFace( avF );
                        } else {
                            aavFFailed += avF;
                            for ( int iV = 0; iV < 3; iV++ ) {
                                TRACE("%d ", avF[iV]->IndexId());
                                Edge e = out_mesh.EdgeFromVertices( avF[iV], avF.wrap(iV+1) );
                                if ( e != NULL ) {
                                    Face f1 = out_mesh.EdgeFace1(e);
                                    Face f2 = out_mesh.EdgeFace2(e);
                                    if ( f1 )
                                        TRACE("\n%d  %d %d %d\n", f1->IndexId(), out_mesh.FaceVertices(f1)[0]->IndexId(),out_mesh.FaceVertices(f1)[1]->IndexId(),out_mesh.FaceVertices(f1)[2]->IndexId() );
                                    if ( f2 )
                                        TRACE("\n%d  %d %d %d\n", f2->IndexId(), out_mesh.FaceVertices(f2)[0]->IndexId(),out_mesh.FaceVertices(f2)[1]->IndexId(),out_mesh.FaceVertices(f2)[2]->IndexId() );

                                }
                            }

                            TRACE("\n");
                            iFailed++;
                        }

                    } else if ( avF.num() == 4 ) {
                        if ( in_bReverse ) avF.reverse();

                        avFTri[0] = avF[0];
                        avFTri[1] = avF[1];
                        avFTri[2] = avF[2];

                        if ( out_mesh.IsOkToAdd( avFTri ) ) {
                            out_mesh.AddFace( avFTri );
                        } else {
                            aavFFailed += avFTri;
                            iFailed++;
                        }

                        avFTri[0] = avF[2];
                        avFTri[1] = avF[3];
                        avFTri[2] = avF[0];
                        if ( out_mesh.IsOkToAdd( avFTri ) ) {
                            out_mesh.AddFace( avFTri );
                        } else {
                            aavFFailed += avFTri;
                            iFailed++;
                        }

                    } else {
                        if ( out_mesh.IsOkToAdd( avF ) ) {
                            out_mesh.AddFace( avF );
                        } else {
                            aavFFailed += avF;
                            iFailed++;
                        }
                    }
                }
            }
        } else {
            const WINbool bIsVertex = !strcmp( ply->elems[i]->name, "vertex" );
            if ( bIsVertex ) {
                TRACE("Number vertices: %d\n", iNElems );
                av.need( iNElems );
            }
            for ( int iElem = 0; iElem < iNElems; iElem++ ) {
                pt = R3Pt(1e30, 1e30, 1e30);
                for ( int j = 0; j < ply->elems[i]->nprops; j++ ) {
                    if ( IsInt( ply->elems[i]->props[j]->external_type ) ) {
                        iVal = GetInt( ply->elems[i]->props[j]->external_type, ply->file_type, file );
                    } else {
                        dVal = GetDouble( ply->elems[i]->props[j]->external_type, ply->file_type, file );
                    }
                    if ( bIsVertex ) {
                        if ( !strcmp( ply->elems[i]->props[j]->name, "x") || !strcmp( ply->elems[i]->props[j]->name, "X") )
                            pt[0] = dVal;
                        else if ( !strcmp( ply->elems[i]->props[j]->name, "y") || !strcmp( ply->elems[i]->props[j]->name, "Y") )
                            pt[1] = dVal;
                        else if ( !strcmp( ply->elems[i]->props[j]->name, "z") || !strcmp( ply->elems[i]->props[j]->name, "Z") )
                            pt[2] = dVal;
                    }
                }

                if ( bIsVertex ) {
                    if ( pt[0] == 1e30 || pt[1] == 1e30 || pt[2] == 1e30 )
                        TRACE("Bad points\n");

                    av[iElem] = out_mesh.AddVertex(pt );
                }
            }
        }
    }

    close_ply( ply );

    int iAddedReverse = 0;
    for ( int iF = 0; iF < aavFFailed.num(); iF++ ) {
        aavFFailed[iF].reverse();
        if ( out_mesh.IsOkToAdd( aavFFailed[iF] ) ) {
            out_mesh.AddFace( aavFFailed[iF] );
            iAddedReverse++;
        }
    }

    TRACE("Added %d faces, failed %d Euler %d\n", out_mesh.NumFaces(), iFailed, out_mesh.NumVertices() - out_mesh.NumEdges() + out_mesh.NumFaces() - 2 );
    if ( iAddedReverse != 0 )
        TRACE("WARNING: Added %d reverse faces\n", iAddedReverse);

    out_mesh.Compute_normals();
}


void PLYWritePMeshLiteToPly( const PMeshLite &in_mesh, const char * in_strFName )
{
    ofstream out;
    WINWriteBinary( in_strFName, out );

    Array<float> afPoints( in_mesh.NumVertices() * 3 );

    out << "ply\n";
    out << "format binary_little_endian 1.0\n";
    out << "element vertex " << in_mesh.NumVertices() << "\n";
    out << "property float x\n";
    out << "property float y\n";
    out << "property float z\n";
    out << "element face " << in_mesh.NumFaces() << "\n";
    out << "property list uchar int vertex_indices\n";
    out << "end_header\n";

    ForMeshVertex( in_mesh.Mesh(), v ) {
        afPoints[3 * v->IndexId() + 0 ] = (float) v->GetLoc()[0];
        afPoints[3 * v->IndexId() + 1 ] = (float) v->GetLoc()[1];
        afPoints[3 * v->IndexId() + 2 ] = (float) v->GetLoc()[2];
    } EndFor;
    out.write( (char *) &afPoints[0], sizeof(float) * afPoints.num() );

    ForMeshFace( in_mesh.Mesh(), f ) {
        unsigned char cNPer = in_mesh.FaceVertices(f).num();
        out.write( (char *) &cNPer, sizeof( unsigned char ) );

        ForFaceVertex( in_mesh.Mesh(), f, v ) {
            int iId = v->IndexId();
            out.write( (char *) &iId, sizeof(int) );
        }EndFor;
    }EndFor;
    out.close();
}

