#ifndef __MCALLMETRICS_DEFS_H
#define __MCALLMETRICS_DEFS_H

#include <WINSystemDefines.H>
#include <MC/MC_Point.H>
#include <MC/MC_Data.H>
#include <utils/Pm_MeshLite.H>

/**  \class MCAllMetrics MC_AllMetrics.H MC/MC_AllMetrics.H 
\brief Calculates curvature values on a mesh for various methods

If you only want to calculate the curvature using one method, use MCData instead
*/
class MCAllMetrics {
protected:
    PMeshLite m_Mesh;
    char *m_pMeshDescription;
    SArray< MCData > m_aData;

public:
    /**@name Constructors */
    //@{
    MCAllMetrics();
    ///
    virtual ~MCAllMetrics();
    //@}

    /**@name Initialization */
    //@{
    /// Read a mesh given the file name (do first)
    void ReadMesh( const char *filename );
    /// (Optional) If you have exact curvatures for the mesh, read them in
    int ReadExactCurvature( const char *filename );

    /// Calculate all curvature methods for all vertices
    int CalculateCurvature();
    /// Calculate just the given method for all vertices
    int CalculateCurvature(const MCData::MC_Method &method);
    /** \brief Calculate all of the methods for the given vertex. 
     * @param in_vertex the vertex to get the values for
     * @param out_data Curvature data for each of the methods. Methods are in order of MCData::MC_Method enumerated type
     * @returns -1 if no method
     */
    int CalculateCurvature(const Vertex &in_vertex, Array<MCPoint> &out_data);
    ///  Calculate just the given method for the given vertex. Returns the result in out_cpt
    int CalculateCurvature(const Vertex &in_vertex, 
                           const MCData::MC_Method &method,
                           MCPoint &out_cpt);
    //@}

    /**@name Accessors */
    //@{
    ///
    PMeshLite& Mesh() { return m_Mesh; }
    /** \brief Get the curvature using a specific method for this vertex. 
      * This is only valid after CalculateCurvature() has been called.
      *
      * @param method is which method (discrete curvature, fitting, etc)
      * @param in_vertex is the vertex
      * @param out_point is the data
      */
    int getPoint(const MCData::MC_Method method, const Vertex &in_vertex, 
                 MCPoint &out_point);
    /** \brief Get the curvature for all vertices calculated using the given method
      * This is only valid after CalculateCurvature() has been called.
      *
      * @param method ensures the method (discrete curvature, fitting, etc) has been calculated
      * @returns MCData, which has all of the data for all of vertices
      */
    MCData& getData(const MCData::MC_Method method);
    //@}

    /**@name Output */
    //@{ 
    ///
    void dump(FILE *out);
    ///
    void dumpRanges(FILE *out);
    //@}

};

#endif
