#ifndef __MCDATA_DEFS_H
#define __MCDATA_DEFS_H

#include <WINSystemDefines.H>
#include <MC/MC_Fit.H>
#include <MC/MC_Point.H>
#include <utils/Pm_MeshLite.H>

/**  \class MCData MC_Data.H MC/MC_Data.H 
\brief Stores curvature values for all vertices for one method

Also keeps the range of values over the mesh
*/
class MCData {
public:
    typedef enum { NONE=-1, EXACT=0, DCO, MDCO, ANCE, ANC, IEV, 
                   FIT1P, FIT2P, FIT3P, FIT1N, FIT2N, FIT3N,
                   RBF1P, RBF2P, RBF3P, RBF1N, RBF2N, RBF3N,
                   CON1, CON2, CON3, HST, DGM, NUM_METHODS } MC_Method;

protected:
    const PMeshLite *m_pMesh;
    char *m_pMeshDescription;
    MCData::MC_Method m_method;
    MCFit *m_pFit;
    Array< MCPoint > m_aPoints;
    R2Pt m_K1Range;
    R2Pt m_K2Range;
    R2Pt m_KgRange;
    R2Pt m_KbarRange;

protected:
    /**@name Protected constructors */
    //@{
    MCData( const MCData &in_data );
    ///
    MCData &operator=(const MCData &in_data);
    ///
    int setPoint(const Vertex &in_vertex, const MCPoint &in_point);
    //@}

	/// Sets k1/k2/gaussian/mean ranges
    void SetRange();

public:
    /**@name Constructors */
    //@{
    MCData();
    ///
    virtual ~MCData();
    //@}

    /**@name Accessors */
    //@{
    ///
    static const char *Description( const MC_Method in_method );
    ///
    const PMeshLite* Mesh() const { return m_pMesh; }
    ///
    const MCFit& Fit() { ASSERT(m_pFit != NULL);
                         return *m_pFit; }
    //@}

    /**@name Create the data */
    //@{
    ///
    void Clear();
    /// Set the mesh and the method (does not do calculation - call CalculateCurvature() )
    void Init(const PMeshLite &mesh, const MCData::MC_Method method);
    /// Set the mesh and the method then copy the data from in_data
    void Init(const PMeshLite &mesh, const MC_Method method,
              const MCData &in_data, const Array<Vertex> &in_aopVs);
	/// Actually calculate the curvature for a single vertex. Must call Init(mesh, method) first
    int CalculateCurvature( const Vertex &in_vertex, MCPoint &in_point ) const;
    /// Actually calculate the curvature for all vertices. Must call Init(mesh, method) first
    int CalculateCurvature();
    /// Read the data from a file
    int ReadMeshCurvature( const char *filename );
    /// Read just the normals from a file (for use with methods that need exact normals)
    int ReadMeshNormals( const char *filename );
    //@}

    /**@name Adjust the data */
    //@{
    ///
    void FixCurvatureSign(const int in_iFlip = -1);
    ///
    void FlipMeshNormals();
    ///
	void FlipCurvatureSign();
    //@}

    /**@name Access the data */
    //@{
    ///
	int NumPoints(void) const { return m_aPoints.num(); }
	///
    int getPoint(const Vertex &in_vertex, MCPoint &out_point) const;
    ///
    int getPoint(const int index, MCPoint &out_point) const;
    ///
    const R2Pt& getK1Range() const { return m_K1Range; }
    ///
    const R2Pt& getK2Range() const { return m_K2Range; }
    ///
    const R2Pt& getKgRange() const { return m_KgRange; }
    ///
    const R2Pt& getKbarRange() const { return m_KbarRange; }
    ///
    MCData::MC_Method GetMethod() const { return m_method; }
    /** \brief interpolate values across face
     * E.g. sum b_i kg(v_i)
     * @param i1, i2, i3 are vertex indices
     * @param b1, b2, b3 are barycentric coordinates for the vertices
     * @param out_kg averaged gaussian curvature
     * @param out_kbar averaged men curvature
     * @param out_k1 averaged k1 curvature
     * @param out_k2 averaged k2 curvature
     */
	int InterpolateBary(const int i1, const double b1, 
		                const int i2, const double b2, 
		                const int i3, const double b3, 
		                double &out_kg, double &out_kbar,
						double &out_k1, double &out_k2);
    //@}


    /**@name Dump the data to a file */
    //@{
    ///
    void dump(FILE *out);
    ///
    void dumpRanges(FILE *out);
    //@}

};

#endif
