#ifndef _FITTING_SVD_DEFS
#define _FITTING_SVD_DEFS

#include <mkl/mkl_solver.h>
#include <mkl/mkl_lapack.h>

/** \class FITToolsSVDLapack FITTools_SVD.H fitting/FITTools_SVD.H
    \ingroup FittingMatrix
    \brief Lapack Singular value decomposition (Also gives Eigen vectors and Eigen values)

  Factor a matrix into U D V^t, where U (Eigen vectors) and V are orthogonal, and D is a diagonal 
  matrix (eigen values).

  This does singular value decomposition, returning the eigen values/eigen vectors.

  If you don't need, say V, then it is faster not to compute it.

  Note that the A matrix data is overwritten when Solve is called. So calling Print after calling
  Solve will result in garbage.

<br>   Files: 
- include/fitting/FITTools_SVD.H (templated) */
template<class T>
class FITToolsSVDLapack {
private:
    int m_iNRows;    ///< m
    int m_iNCols;    ///< n

    /**@name Matrix storage. A is primary */
    //@{
    Array<T> m_adA;
    Array<T> m_adU, m_adVt, m_adD;
    //@}

    /**@name Scratch - store here */
    //@{
    Array<T> m_adE, m_adTauQ, m_adTauP, m_adWork;
    int iWork, iInfo;
    //@}

    WINbool m_bCalcU, ///< Calculate the U matrix, y/n
            m_bCalcVt;///< Calculate the V matrix, y/n

public:
    /**@name Accessors */
    //@{
    /// Number of rows in A
    int NRows()   const { return m_iNRows; }
    /// Number of columns in A
    int NCols() const { return m_iNCols; }
    /// 
    const T &A( const int in_iR, const int in_iC ) const;
    ///
    const T &U( const int in_iR, const int in_iC ) const;
    ///
    const T &Vt( const int in_iR, const int in_iC ) const;
    /// Diagonal
    const T &Sigma( const int in_i ) const;
    //@}

    /**@name Set up */
    //@{
    ///
    T &A( const int in_iR, const int in_iC );
    /// By default, don't calculate. Call if you want U.
    void CalcU() { m_bCalcU = TRUE; }
    /// By default, don't calculate. Call if you want V.
    void CalcVt() { m_bCalcVt = TRUE; }
    /// Delete all the data
    void Clear();
    /// Resize if necessary
    void Resize( const int in_iNRows, const int in_iNCols );
    //@}

    /**@name Solve */
    //@{
    /// Note: A is overwritten by solve
    WINbool Solve();
    //@}

    ///
    FITToolsSVDLapack( const int in_iNRows, const int in_iNCols, const WINbool in_bCalcU = FALSE, const WINbool in_bCalcVt = FALSE )
        : m_bCalcU( in_bCalcU ), m_bCalcVt( in_bCalcVt ) { Resize( in_iNRows, in_iNCols ); }
    ///
    ~FITToolsSVDLapack() { }

    /**@name Debuging */
    //@{
    void CheckOrthoU() const;
    ///
    void CheckOrthoVt() const;
    // Need to reset A for this call to work (A is overwritten by solve)
    void CheckAisUSVt() const;
    // Need to reset A for this call to work (A is overwritten by solve)
    void CheckAVisSU() const;
    //@}
};

template <class T>
inline const T &FITToolsSVDLapack<T>::A(const int in_iRow, const int in_iCol) const
{
   ASSERT(in_iRow >= 0 && in_iRow < m_iNRows);
   ASSERT(in_iCol >= 0 && in_iCol < m_iNCols);
   return m_adA[in_iCol * m_iNRows + in_iRow];
}

template <class T>
inline const T &FITToolsSVDLapack<T>::U(const int in_iRow, const int in_iCol) const
{
   ASSERT(in_iRow >= 0 && in_iRow < m_iNRows);
   ASSERT(in_iCol >= 0 && in_iCol < m_iNCols);
   return m_adU[in_iCol * m_iNRows + in_iRow];
}

template <class T>
inline const T &FITToolsSVDLapack<T>::Vt(const int in_iRow, const int in_iCol) const
{
   ASSERT(in_iRow >= 0 && in_iRow < m_iNCols);
   ASSERT(in_iCol >= 0 && in_iCol < m_iNCols);
   return m_adVt[in_iCol * m_iNRows + in_iRow];
}

template <class T>
inline const T &FITToolsSVDLapack<T>::Sigma(const int in_i) const
{
   ASSERT(in_i >= 0 && in_i < m_iNCols);
   return m_adD[in_i];
}

template <class T>
inline T &FITToolsSVDLapack<T>::A(const int in_iRow, const int in_iCol) 
{
   ASSERT(in_iRow >= 0 && in_iRow < m_iNRows);
   ASSERT(in_iCol >= 0 && in_iCol < m_iNCols);
   return m_adA[in_iCol * m_iNRows + in_iRow];
}

template <class T>
inline void FITToolsSVDLapack<T>::Clear()
{
    if ( m_adA.num() ) m_adA.memfill(0);
    if ( m_adU.num() ) m_adU.memfill(0);
    if ( m_adVt.num() ) m_adVt.memfill(0);
    if ( m_adD.num() ) m_adD.memfill(0);
}

template <class T>
inline void FITToolsSVDLapack<T>::Resize( const int in_iR, const int in_iC )
{
    ASSERT( in_iR > 0 && in_iC > 0 );
    m_iNRows = in_iR;
    m_iNCols = in_iC;
    m_adA.need( in_iR * in_iC );
    m_adA.memfill(0);
}

template <class T>
inline WINbool FITToolsSVDLapack<T>::Solve()
{
    m_adD.need( WINmax( m_iNRows, m_iNCols ) + 1);
    m_adE.need( WINmax( m_iNRows, m_iNCols ) + 1 );
    m_adTauP.need( WINmax( m_iNRows, m_iNCols ) + 1 );
    m_adTauQ.need( WINmax( m_iNRows, m_iNCols ) + 1 );
    m_adWork.need( (m_iNRows + m_iNCols) * 16 + 1 );
    iWork = (m_iNRows + m_iNCols) * 16 + 1;
    iInfo = 0;

    m_adD.last() = 0.0;
    m_adE.last() = 0.0;
    m_adTauQ.last() = 0.0;
    m_adTauP.last() = 0.0;
    m_adWork.last() = 0.0;

    // This is ugly, I know
    if ( sizeof(T) == sizeof(float) ) {
        sgebrd( &m_iNRows, &m_iNCols, (float *) &m_adA[0], &m_iNRows, (float *) &m_adD[0], (float *) &m_adE[0], (float *) &m_adTauQ[0], (float *) &m_adTauP[0], (float *) &m_adWork[0], &iWork, &iInfo );
    } else if ( sizeof(T) == sizeof(double) ) {
        dgebrd( &m_iNRows, &m_iNCols, (double *) &m_adA[0], &m_iNRows, (double *) &m_adD[0], (double *) &m_adE[0], (double *) &m_adTauQ[0], (double *) &m_adTauP[0], (double *) &m_adWork[0], &iWork, &iInfo );
    } else {
        ASSERT(FALSE);
    }
    ASSERT( m_adD.last() == 0.0 );
    ASSERT( m_adE.last() == 0.0 );
    ASSERT( m_adTauQ.last() == 0.0 );
    ASSERT( m_adTauP.last() == 0.0 );
    ASSERT( m_adWork.last() == 0.0 );
    if( iInfo != 0 ) {
        TRACE("Error calc'ing QBPt %d\n", iInfo);
        return FALSE;
    }

    float afU[1];
    float *aopfU = afU;
    double adU[1];
    double *aopdU = adU;
    if ( m_bCalcU == TRUE ) {
        m_adU.need( m_iNRows * m_iNCols + 1);
        m_adU.last() = 0.0;
        memcpy( m_adU, m_adA, m_iNRows * m_iNCols * sizeof(T) );
        char cWhichVec = 'Q';
        if ( sizeof(T) == sizeof(float) ) {
            sorgbr( &cWhichVec, &m_iNRows, &m_iNCols, &m_iNCols, (float *) &m_adU[0], &m_iNRows, (float *) &m_adTauQ[0], (float *) &m_adWork[0], &iWork, &iInfo );
            aopfU = (float *) &m_adU[0];
        } else {
            dorgbr( &cWhichVec, &m_iNRows, &m_iNCols, &m_iNCols, (double *) &m_adU[0], &m_iNRows, (double *) &m_adTauQ[0], (double *) &m_adWork[0], &iWork, &iInfo );
            aopdU = (double *) &m_adU[0];
        }

        ASSERT( m_adWork.last() == 0.0 );
        ASSERT( m_adU.last() == 0.0 );
        if( iInfo != 0 ) {
            TRACE("Error calc'ing U %d\n", iInfo);
            return FALSE;
        }
    }

    float afVt[1];
    float *aopfVt = afVt;
    double adVt[1];
    double *aopdVt = adVt;
    if ( m_bCalcVt == TRUE ) {
        m_adVt.need( m_iNRows * m_iNCols + 1);
        m_adVt.last() = 0.0;
        char cWhichVec = 'P';
        memcpy( m_adVt, m_adA, m_iNRows * m_iNCols * sizeof(T) );
        if ( sizeof(T) == sizeof(float) ) {
            sorgbr( &cWhichVec, &m_iNCols, &m_iNCols, &m_iNRows, (float *) &m_adVt[0], &m_iNRows, (float *) &m_adTauP[0], (float *) &m_adWork[0], &iWork, &iInfo );
            aopfVt = (float * )&m_adVt[0];
        } else {
            dorgbr( &cWhichVec, &m_iNCols, &m_iNCols, &m_iNRows, (double *) &m_adVt[0], &m_iNRows, (double *) &m_adTauP[0], (double *) &m_adWork[0], &iWork, &iInfo );
            aopdVt = (double *) &m_adVt[0];
        }

        ASSERT( m_adWork.last() == 0.0 );
        ASSERT( m_adVt.last() == 0.0 );
        if( iInfo != 0 ) {
            TRACE("Error calc'ing U %d\n", iInfo);
            return FALSE;
        }
    }

    int iNRowsU = m_bCalcU ? m_iNRows : 0;
    int iNColsVt = m_bCalcVt ? m_iNCols : 0;
    int iNo = 0, iOne = 1;
    float afC[1];
    double adC[1];
    afC[0] = 0.0f;
    adC[0] = 0.0;
    char cUpperOrLower = 'U';
    if ( sizeof(T) == sizeof(float) )
        sbdsqr( &cUpperOrLower, &m_iNCols, &iNColsVt, &iNRowsU, &iNo, (float *) &m_adD[0], (float *) &m_adE[0], aopfVt, &m_iNRows, aopfU, &m_iNRows, afC, &iOne, (float *) &m_adWork[0], &iInfo );
    else
        dbdsqr( &cUpperOrLower, &m_iNCols, &iNColsVt, &iNRowsU, &iNo, (double *) &m_adD[0], (double *) &m_adE[0], aopdVt, &m_iNRows, aopdU, &m_iNRows, adC, &iOne, (double *) &m_adWork[0], &iInfo );

    ASSERT( afC[0] == 0.0f );
    ASSERT( adC[0] == 0.0 );

    if( iInfo != 0 ) {
        TRACE("Error calc'ing SVD of B %d\n", iInfo);
        return FALSE;
    }
    return TRUE;
}

template <class T>
inline void FITToolsSVDLapack<T>::CheckOrthoU() const
{
    ASSERT( m_bCalcU == TRUE );
    for ( int iC = 0; iC < m_iNCols; iC++ ) {
        for ( int iC2 = iC+1; iC2 < m_iNCols; iC2++ ) {
            T dSum = 0.0;
            for ( int i = 0; i < m_iNRows; i++ ) 
                dSum += U( i, iC ) * U( i, iC2 );
            ASSERT( RNIsZero( dSum, (T) 1e-6 ) );
        }
    }
}

template <class T>
inline void FITToolsSVDLapack<T>::CheckOrthoVt() const
{
    ASSERT( m_bCalcVt == TRUE );
    for ( int iC = 0; iC < m_iNCols; iC++ ) {
        for ( int iC2 = iC+1; iC2 < m_iNCols; iC2++ ) {
            T dSum = 0.0;
            T dSum2 = 0.0;
            for ( int i = 0; i < 3; i++ ) {
                dSum += Vt( i, iC ) * Vt( i, iC2 );
                dSum2 += Vt( iC, i ) * Vt( iC2, i );
            }
            ASSERT( RNIsZero( dSum, (T) 1e-6 ) );
            ASSERT( RNIsZero( dSum2, (T) 1e-6 ) );
        }
    }
}

template <class T>
inline void FITToolsSVDLapack<T>::CheckAisUSVt() const
{
    ASSERT( m_bCalcU == TRUE && m_bCalcVt == TRUE );
    for ( int iR = 0; iR < m_iNRows; iR++ ) {
        for ( int iC = 0; iC < m_iNCols; iC++ ) {
            T dSum = 0;
            for ( int iS = 0; iS < 3; iS++ ) {
                dSum += U(iR, iS) * m_adD[iS] * Vt(iS,iC);
            }
            const T dCompare = A(iR, iC);
            ASSERT( RNApproxEqual( dSum, dCompare, (T) 1e-6 ) );
        }
    }
}

template <class T>
inline void FITToolsSVDLapack<T>::CheckAVisSU() const
{
    ASSERT( m_bCalcU == TRUE && m_bCalcVt == TRUE );
    for ( int iVec = 0; iVec < m_iNCols; iVec++ ) {
        for ( int iR = 0; iR < m_iNRows; iR++ ) {
            T dSum = 0.0;
            for ( int iC = 0; iC < 3; iC++ ) {
                dSum += A(iR, iC) * Vt( iVec, iC );
            }
            const T dU = m_adD[iVec] * U(iR, iVec);
            ASSERT( RNApproxEqual( dSum, dU, (T) 1e-6 ) );
        }
    }
}

/** \class FITToolsSVD FITTools_SVD.H fitting/FITTools_SVD.H
  \ingroup FittingMatrix
  \brief SVD routine from numerical recipies. 

  There really isn't much of a reason to use this, unless you can't get
  Lapack in any shape or form. Use FITToolsSVDLapack for singular value
  decomposition, or FITTools_LS for Ax=b.

<br>   Files: 
- include/fitting/FITTools_SVD.H
- src/fitting/FITTools/FITToolsSVD.cpp */
class FITToolsSVD {
public:
    static void householder_zero_col(double *a,double *u,int i,int j,int m,int n,double *hv);
    static void householder_zero_row(double *a,double *v,int i,int j,int m,int n,double *hv);
    static void rotate_cols(int i,int j,double cos,double sin, double *a, int start, int n,int mm,int nn);
    static void rotate_rows(int i, int j, double cos, double sin, double *a, int start,int n, int mm, int nn);
    static WINbool bidiagonal_svd( double *b,int m,int n,double *u,double *v);
    static void bidiagonalize( const double *a,int m,int n,double *u,double *b,double *v);
    static void golub_kahn_svd_rot(double * b,int p,int q,double * v,int m,int n,int min);
    static void clr_top_subdiag_elt(double * b,int p,int z,double * u,double * v,int m,int n,int min);
    static void clr_bot_supdiag_elt(double * b,int p,int z,double * v,int m,int n,int min);
    static void clr_top_supdiag_elt(double * b,int p,int z,double * u,int m,int n,int min);

    /**SVD.
       Singular-Value Decomposition. <br><br>
       
       Had to write this by hand, since the version in Numerical
       Recipes has a bug, carried over from a bug in Golub's
       original algorithm: <br> <br>
       
       Gene H. Golub & Charles F. Van Loan <br>
       MATRIX COMPUTATIONS <br>
       Johns Hopkins University Press, 1983 <br>
       Fifth printing, 1987 <br>
       Page 293 <br>
       Algorithm 8.3-2: The SVD Algorithm <br> <br>
     
     ...Find the largest q and the smallest p such that if <br>

                 +-            -+               <br>
                 | A11   0   0  |          p    <br>
                 |  0   A22  0  |        n-p-q  <br>
           A  =  |  0    0  A33 |          q    <br>
                 |  0    0   0  |         m-n   <br>
                 +-            -+               <br>
                 <br>
        then A33 is diagonal and A22 has a nonzero superdiagonal.<br>
        If q = n then quit.<br>
        If any diagonal entry in A22 is zero, then zero the superdiagonal
        element in the same row and go to Repeat...<br>
        <br>
        The last sentence above is the cause of the trouble, in the following case:<br>
        <br>
                     +-   -+  <br>
                     |0 1 0|  <br>
                 A = |0 1 1|  <br>
                     |0 0 0|  <br>
                     +-   -+  <br>
                     <br>
       In this case, q is 0, A33 is null, and A22 is the same as A.  The instruction
       <br>
            "if any diagonal entry in A22 is zero, then
             zero the superdiagonal element in the same row"<br>
             <br>
      cannot be applied -- A22 has a diagonal entry that is zero, but there
      is no superdiagonal element in that same row.  The proper thing to
      do seems to be to special-case: If A22 has 0 in its lower-right
      diagonal element, zero the superdiagonal element above it.<br>
      <br>

      Since the Num. Rec. code was cribbed from EISPACK or LINPACK
      or some such, which in turn cribbed from other places, I would
      consider an SVD routine suspect unless it works on the following
      test data: <br>
      		a = 	0 1 0  <br>
           		0 1 1  <br>
           		0 0 0  <br>
      <br>
      Anyway, Al Barr & I went through the references to figure out
      how it all worked.  I reimplimented it from scratch, first in
      lisp, and now in C.  Martin L. Livesey used the code and reported
      various bugs, which have been fixed. <br> <br>
      
      
      	Ronen Barzel   July 1989 <br>
            (bug fixes: Feb 1990) <br>
      	 (ansi-C, backsubst: May 1993) <br>
      
    */
    static WINbool SVD(const double *a, int m, int n,
                       double *u, double *w, double *vt);
    /**Back substitute to get Ax = b.

     SVD_BACKSUBST <br>
     <br>
     Given the svd of a matrix A, and a vector B, solves A X = B for
     a vector X.  Takes a condition-threshold factor "eps", singular
     values less than "eps" times the largest value are considered to be
     zero. <br> <br>

     input: <br>
     -	    u (m x min)			column-orthonormal >
     -          w (min-element vector)	diagonal elements
     -          vt (min x n)		column-orthonormal
        where min=min(m,n),
        such that a = u w vt, as decomposed by num_svd() 
     <-	    b (m-element vector)	 <br>
     <br>
     output: <br>
     - 	    x (n-element vector)
      */
    static void Backsubst(int m, int n, 
                          const double *u, const double *w, 
                          const double  *vt,
                          const double  b[], 
                          double x[], double eps);
};

#endif

