// Fit_RBF.H Radial basis function fitting class
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_FITRBF_H__C622E9A3_1A9D_422D_BAF4_50E33C53C211__INCLUDED_)
#define AFX_FITRBF_H__C622E9A3_1A9D_422D_BAF4_50E33C53C211__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <utils/Rn_Defs.H>
#include <utils/Mesh_Array.H>
#include <fitting/FITTools_Least_squares.H>
#include <utils/Utils_GeomArray.H>

/** \class FITRbfTC Fit_RBF.H fitting/Fit_RBF.H 
     \ingroup FittingElems
     \brief Radial Basis function fitting

  Define a function/surface as f(s,t) sum w_i b_i(s,t). There is usually
  one b_i for each data points, and the b_i are (usually) placed/centered around 
  a data point. The w_i are solved for so that f() interpolates the data points.

  This is the base class for all of the other Rbf classes. Each sub class defines
  a different b() function.

  Only templated on output dimension; could also template in input dimension, i.e., f(s)
  */
template< class PtType, class VecType >
class FITRbfTC
{
protected:
    Array< R2Pt >    m_aptRad;     ///< Where the basis functions are centered
    Array< PtType >  m_aptLambda;  ///< one for each Pt dim; weights w_i
    VecType          m_aptPoly[3]; ///< polynomial coefs for each dim

public:
    /// Output dimension (1,2,3, etc)
    int Dim() const { return m_aptPoly[0].Dim(); }

    /** \brief Set the input locations and corresponding output points
      * Set all the points at once. f( in_aptParam[i] ) = in_aptData[i]
      @param in_aptParam The s,t points
      @param in_aptData The data points to interpolate */
    virtual void Set( const Array< R2Pt > &in_aptParam, const Array< PtType > &in_aptData ) = 0;

    /**@name Evaluation */
    //@{
    ///
    virtual PtType Eval( const R2Pt &in_pt ) const = 0;

    ///
    virtual VecType DerivS( const R2Pt &in_pt ) const = 0;
    ///
    virtual VecType DerivT( const R2Pt &in_pt ) const = 0;
    ///
    virtual VecType DerivSS( const R2Pt &in_pt ) const = 0;
    ///
    virtual VecType DerivTT( const R2Pt &in_pt ) const = 0;
    ///
    virtual VecType DerivST( const R2Pt &in_pt ) const = 0;
    ///
    VecType SrfNormal( const R2Pt &in_pt ) const;
    ///
    virtual void AllEval( const R2Pt &in_pt,
                          PtType &out_pt,
                          VecType &out_vecDs, VecType &out_vecDt,
                          VecType &out_vecDss, VecType &out_vecDtt,
                          VecType &out_vecDst ) const = 0;
    ///
    virtual void AllEvalSub( const R2Pt &in_pt,
                             PtType &out_pt,
                             VecType &out_vecDs, VecType &out_vecDt ) const = 0;

    /// point area calculation
    double Area( const R2Pt &in_pt ) const;

    /// curvature calculation
	void Curvature( const R2Pt &in_pt, double &out_dCrv1, double &out_dCrv2, 
		            R2Vec &out_vec1, R2Vec &out_vec2, 
		            R3Vec &out_vecMin, R3Vec &out_vecMax, 
		            R3Vec &out_vecNorm ) const;
    //@}

    ///
    WINbool operator==( const FITRbfTC<PtType, VecType> &in_func ) const;
    ///
    FITRbfTC<PtType, VecType> &operator=( const FITRbfTC<PtType, VecType> &in_func );

    ///
    FITRbfTC( const FITRbfTC<PtType, VecType> &in_func ) { *this = in_func; }
    ///
    FITRbfTC();
    ///
    virtual ~FITRbfTC();

    /**@name Read/write */
    //@{
    ///
    virtual void Write( ofstream &out ) const;
    ///
    virtual void Read( ifstream &in );

    ///
    virtual void WriteBinary( ofstream &out ) const;
    ///
    virtual void ReadBinary( ifstream &in );
    //@}

    void Test() const;
};

extern int FITRbfTC_SortFc( const void *in_op1, const void *in_op2 );


#include "Fit_RBF_i.H"

typedef FITRbfTC<R3Pt,R3Vec> FITRbf;

#endif // !defined(AFX_FITRbf_H__C622E9A3_1A9D_422D_BAF4_50E33C53C211__INCLUDED_)

