// d(x,y - s,t) = sqrt( (x0 - s)^2 + (y0 - t)^2 )

// a0 + a1 s + a2 t + sum wi e^{d(x,y - s,t)^2  / sigmai^2}
template<class PtType, class VecType>
inline
PtType FITRbfGaussTC<PtType,VecType>::Eval( const R2Pt &in_pt ) const
{
    PtType ptRes;

    // polynomial
    for ( int i = 0; i < Dim(); i++ ) {
        ptRes[i] = m_aptPoly[0][i] + m_aptPoly[1][i] * in_pt[0] + m_aptPoly[2][i] * in_pt[1];
    }

    for ( FORINT i = 0; i < m_aptRad.num(); i++ ) {
        // d(x,y - s,t)^2
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt );

        // Kernel e^{d^2/sigma^2}
        const double dPhi = exp( -dDistSq / m_adSigmaSq[i] );

        // multiply by weight
        const PtType &ptLambda = m_aptLambda[i];

        for ( int j = 0; j < Dim(); j++ )
            ptRes[j] += ptLambda[j] * dPhi;
    }

    return ptRes;
}

// D/du e^x = dx/du e^x
template<class PtType, class VecType>
inline
VecType FITRbfGaussTC<PtType,VecType>::DerivS( const R2Pt &in_pt ) const
{
    VecType vecRes;

    // polynomial derivative
    for ( int i = 0; i < Dim(); i++ ) {
        vecRes[i] = m_aptPoly[1][i];
    }

    for ( FORINT i = 0; i < m_aptRad.num(); i++ ) {
        // d(x,y - s,t)^2
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt );
        // d/ds (s-x)^2 + (t-y)^2 = 2 (s - x)
        const double dDistSqDeriv = 2.0 * (m_aptRad[i][0] - in_pt[0]) / m_adSigmaSq[i];

        // Kernel e^{d^2/sigma^2}
        const double dPhi = exp( -dDistSq / m_adSigmaSq[i] );

        //
        const double dPhiDs = dDistSqDeriv * dPhi;

        // multiply by weight
        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ )
            vecRes[j] += ptLambda[j] * dPhiDs;
    }

    return vecRes;
}

template<class PtType, class VecType>
inline
VecType FITRbfGaussTC<PtType,VecType>::DerivT( const R2Pt &in_pt ) const
{
    VecType vecRes;

    for ( int i = 0; i < Dim(); i++ ) {
        vecRes[i] = m_aptPoly[2][i];
    }

    for ( FORINT i = 0; i < m_aptRad.num(); i++ ) {
        // d(x,y - s,t)^2
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt );
        // d/ds (s-x)^2 + (t-y)^2 = 2 (s - x)
        const double dDistSqDeriv = 2.0 * (m_aptRad[i][1] - in_pt[1]) / m_adSigmaSq[i];

        // Kernel e^{d^2/sigma^2}
        const double dPhi = exp( -dDistSq / m_adSigmaSq[i] );

        //
        const double dPhiDt = dDistSqDeriv * dPhi;

        // multiply by weight
        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ )
            vecRes[j] += ptLambda[j] * dPhiDt;
    }
    return vecRes;
}

// d/ds  dx e^x = dx^2/ds e^x + dx dx e^x
template<class PtType, class VecType>
inline
VecType FITRbfGaussTC<PtType,VecType>::DerivSS( const R2Pt &in_pt ) const
{
    VecType vecRes;

    RNZero( vecRes );

    for ( int i = 0; i < m_aptRad.num(); i++ ) {
        // d(x,y - s,t)^2
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt );
        // d/ds (s-x)^2 + (t-y)^2 = 2 (s - x)
        const double dDistSqDeriv = 2.0 * (m_aptRad[i][0] - in_pt[0]) / m_adSigmaSq[i];

        // d/ds  2 (s - x) = 2.0
        const double dDistSqDerivDeriv = -2.0 / m_adSigmaSq[i];

        // Kernel e^{d^2/sigma^2}
        const double dPhi = exp( -dDistSq / m_adSigmaSq[i] );

        //
        const double dPhiDss = dDistSqDerivDeriv * dPhi + dDistSqDeriv * dDistSqDeriv * dPhi;

        // multiply by weight
        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ )
            vecRes[j] += ptLambda[j] * dPhiDss;
    }
    return vecRes;
}

template<class PtType, class VecType>
inline
VecType FITRbfGaussTC<PtType,VecType>::DerivTT( const R2Pt &in_pt ) const
{
    VecType vecRes;

    RNZero( vecRes );

    for ( int i = 0; i < m_aptRad.num(); i++ ) {
        // d(x,y - s,t)^2
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt );
        // d/ds (s-x)^2 + (t-y)^2 = 2 (s - x)
        const double dDistSqDeriv = 2.0 * (m_aptRad[i][1] - in_pt[1]) / m_adSigmaSq[i];

        // d/ds  2 (s - x) = 2.0
        const double dDistSqDerivDeriv = -2.0 / m_adSigmaSq[i];

        // Kernel e^{d^2/sigma^2}
        const double dPhi = exp( -dDistSq / m_adSigmaSq[i] );

        //
        const double dPhiDtt = dDistSqDerivDeriv * dPhi + dDistSqDeriv * dDistSqDeriv * dPhi;

        // multiply by weight
        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ )
            vecRes[j] += ptLambda[j] * dPhiDtt;
    }
    return vecRes;
}


// d/ dt  dx e^x = 0 e^x + dx dy e^x
template<class PtType, class VecType>
inline
VecType FITRbfGaussTC<PtType,VecType>::DerivST( const R2Pt &in_pt ) const
{
    VecType vecRes;

    RNZero( vecRes );

    for ( int i = 0; i < m_aptRad.num(); i++ ) {
        // d(x,y - s,t)^2
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt );
        // d/ds (s-x)^2 + (t-y)^2 = 2 (s - x)
        const double dDistSqDerivU = 2.0 * (m_aptRad[i][0] - in_pt[0]) / m_adSigmaSq[i];
        const double dDistSqDerivV = 2.0 * (m_aptRad[i][1] - in_pt[1]) / m_adSigmaSq[i];

        // Kernel e^{d^2/sigma^2}
        const double dPhi = exp( -dDistSq / m_adSigmaSq[i] );

        //
        const double dPhiDst = dDistSqDerivU * dDistSqDerivV * dPhi;

        // multiply by weight
        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ )
            vecRes[j] += ptLambda[j] * dPhiDst;
    }
    return vecRes;
}

template<class PtType, class VecType>
inline
void FITRbfGaussTC<PtType,VecType>::AllEval( const R2Pt &in_pt,
                            PtType &out_pt,
                            VecType &out_vecDs, VecType &out_vecDt,
                            VecType &out_vecDss, VecType &out_vecDtt,
                            VecType &out_vecDst ) const
{

    for ( int i = 0; i < Dim(); i++ ) {
        out_pt[i] = m_aptPoly[0][i] + m_aptPoly[1][i] * in_pt[0] + m_aptPoly[2][i] * in_pt[1];
        out_vecDs[i] = m_aptPoly[1][i];
        out_vecDt[i] = m_aptPoly[2][i];
        out_vecDss[i] = 0.0;
        out_vecDst[i] = 0.0;
        out_vecDtt[i] = 0.0;
    }

    for ( FORINT i = 0; i < m_aptRad.num(); i++ ) {
        // d(x,y - s,t)^2
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt );
        // d/ds (s-x)^2 + (t-y)^2 = 2 (s - x)
        const double dDistSqDerivU = 2.0 * (m_aptRad[i][0] - in_pt[0]) / m_adSigmaSq[i];
        const double dDistSqDerivV = 2.0 * (m_aptRad[i][1] - in_pt[1]) / m_adSigmaSq[i];

        // d/ds  2 (s - x) = 2.0
        const double dDistSqDerivDeriv = -2.0 / m_adSigmaSq[i];

        // Kernel e^{d^2/sigma^2}
        const double dPhi = exp( -dDistSq / m_adSigmaSq[i] );

        //
        const double dPhiDs  = dDistSqDerivU * dPhi;
        const double dPhiDt  = dDistSqDerivV * dPhi;
        const double dPhiDss = dDistSqDerivDeriv * dPhi + dDistSqDerivU * dDistSqDerivU * dPhi;
        const double dPhiDtt = dDistSqDerivDeriv * dPhi + dDistSqDerivV * dDistSqDerivV * dPhi;
        const double dPhiDst = dDistSqDerivU * dDistSqDerivV * dPhi;

        // multiply by weight
        const PtType &ptLambda = m_aptLambda[i];
        for ( int iD = 0; iD < Dim(); iD++ ) {
            out_pt[iD]     += ptLambda[iD] * dPhi;
            out_vecDs[iD]  += ptLambda[iD] * dPhiDs;
            out_vecDt[iD]  += ptLambda[iD] * dPhiDt;
            out_vecDss[iD] += ptLambda[iD] * dPhiDss;
            out_vecDst[iD] += ptLambda[iD] * dPhiDst;
            out_vecDtt[iD] += ptLambda[iD] * dPhiDtt;
        }
    }
}

template<class PtType, class VecType>
inline
void FITRbfGaussTC<PtType,VecType>::AllEvalSub( const R2Pt &in_pt,
                              PtType &out_pt,
                              VecType &out_vecDs, VecType &out_vecDt ) const
{
    for ( int i = 0; i < Dim(); i++ ) {
        out_pt[i] = m_aptPoly[0][i] + m_aptPoly[1][i] * in_pt[0] + m_aptPoly[2][i] * in_pt[1];
        out_vecDs[i] = m_aptPoly[1][i];
        out_vecDt[i] = m_aptPoly[2][i];
    }

    for ( FORINT i = 0; i < m_aptRad.num(); i++ ) {
        // d(x,y - s,t)^2
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt );
        // d/ds (s-x)^2 + (t-y)^2 = 2 (s - x)
        const double dDistSqDerivU = 2.0 * (m_aptRad[i][0] - in_pt[0]) / m_adSigmaSq[i];
        const double dDistSqDerivV = 2.0 * (m_aptRad[i][1] - in_pt[1]) / m_adSigmaSq[i];

        // Kernel e^{d^2/sigma^2}
        const double dPhi = exp( -dDistSq / m_adSigmaSq[i] );

        //
        const double dPhiDs  = dDistSqDerivU * dPhi;
        const double dPhiDt  = dDistSqDerivV * dPhi;

        // multiply by weight
        const PtType &ptLambda = m_aptLambda[i];
        for ( int iD = 0; iD < Dim(); iD++ ) {
            out_pt[iD]     += ptLambda[iD] * dPhi;
            out_vecDs[iD]  += ptLambda[iD] * dPhiDs;
            out_vecDt[iD]  += ptLambda[iD] * dPhiDt;
        }
    }
}


template<class PtType, class VecType>
inline
void FITRbfGaussTC<PtType,VecType>::Set( const Array< R2Pt >   & in_acpt, 
                                         const Array< PtType > & in_apt )
{
    Array<double> adDist( in_apt.num() * in_apt.num() );
    for ( int i = 0; i < in_apt.num(); i++ ) {
        for ( int j = i; j < in_apt.num(); j++ ) {
            adDist[ i * in_apt.num() + j ] = LengthSq( in_acpt[i] - in_acpt[j] );
        }
    }

    m_adSigmaSq.need( in_acpt.num() );
    Array<double> adSort( in_acpt.num() );
    for ( FORINT i = 0; i < in_acpt.num(); i++ ) {
        for ( int j = 0; j < i; j++ )
            adSort[j] = adDist[ j * in_acpt.num() + i ];
        for ( FORINT j = i; j < in_acpt.num(); j++ )
            adSort[j] = adDist[ i * in_acpt.num() + j ];

        qsort( &adSort[0], adSort.num(), sizeof(double), FITRbfTC_SortFc );

        const int iN = WINmin( 10, in_acpt.num() );
        m_adSigmaSq[i] = 0.0;
        for ( FORINT j = 0; j < iN; j++ ) {
            m_adSigmaSq[i] += adSort[j];
        }
        m_adSigmaSq[i] /= (double) iN;
        if ( RNIsZero( m_adSigmaSq[i] ) )
            m_adSigmaSq[i] = 1.0;
    }


    FITTools_LS oLS( in_apt.num() + 3, in_apt.num() + 3, Dim() );

    for ( FORINT i = 0; i < in_apt.num(); i++ ) {
        for ( int j = i; j < in_apt.num(); j++ ) {

            const double dDistSq = adDist[i * in_apt.num() + j];

            oLS.Desired_clapack(i,j) = exp( -dDistSq / m_adSigmaSq[j] );
            oLS.Desired_clapack(j,i) = exp( -dDistSq / m_adSigmaSq[i] );
        }
        oLS.Desired_clapack( i, in_apt.num() + 0 ) = 1.0;
        oLS.Desired_clapack( i, in_apt.num() + 1 ) = in_acpt[i][0];
        oLS.Desired_clapack( i, in_apt.num() + 2 ) = in_acpt[i][1];

        oLS.Desired_clapack( in_apt.num() + 0, i ) = 1.0;
        oLS.Desired_clapack( in_apt.num() + 1, i ) = in_acpt[i][0];
        oLS.Desired_clapack( in_apt.num() + 2, i ) = in_acpt[i][1];

        for ( FORINT j = 0; j < Dim(); j++ ) {
            oLS.Rhs_clapack(i, j) = in_apt[i][j];

        }
    }
    VERIFY( oLS.Solve_clapack() );

    m_aptLambda.need( in_apt.num() );
    m_aptRad = in_acpt;
    for ( FORINT i = 0; i < in_apt.num(); i++ ) {
        for ( int j = 0; j < Dim(); j++ ) {
            m_aptLambda[i][j] = oLS.Solution_clapack( i, j );
        }
    }
    for ( FORINT i = 0; i < 3; i++ ) {
        for ( int j = 0; j < Dim(); j++ ) {
            m_aptPoly[i][j] = oLS.Solution_clapack( in_apt.num() + i, j );
        }
    }

    for ( FORINT i = 0; i < in_apt.num(); i++ ) {
        const PtType ptEval = Eval( in_acpt[i] );
        const double dLen = Length( ptEval - in_apt[i] );
        ASSERT( RNIsZero( dLen, 1e-4 ) );
    }
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

template<class PtType, class VecType>
inline
FITRbfGaussTC<PtType,VecType>::FITRbfGaussTC()
{
}

template<class PtType, class VecType>
inline
FITRbfGaussTC<PtType,VecType>::~FITRbfGaussTC()
{

}

///
template<class PtType, class VecType>
inline
FITRbfGaussTC<PtType,VecType> &FITRbfGaussTC<PtType,VecType>::operator=( const FITRbfGaussTC<PtType,VecType> &in_func )
{
    FITRbfTC<PtType,VecType>::operator=( in_func );

    m_adSigmaSq = in_func.m_adSigmaSq;

    return *this;
}

template<class PtType, class VecType>
inline
void FITRbfGaussTC<PtType,VecType>::Read( ifstream &in )
{
    FITRbfTC<PtType,VecType>::Read(in);
    m_adSigmaSq.need( m_aptRad.num() );

    for ( int i = 0; i < m_adSigmaSq.num(); i++ )
        in >> m_adSigmaSq[i];
}

template<class PtType, class VecType>
inline
void FITRbfGaussTC<PtType,VecType>::Write( ofstream &out ) const
{
    FITRbfTC<PtType,VecType>::Write(out);

    for ( int i = 0; i < m_adSigmaSq.num(); i++ )
        out << m_adSigmaSq[i] << " ";

    out << "\n";
}
