template<class PtType, class VecType>
inline
PtType FITRbfSplineTC<PtType,VecType>::Eval( const R2Pt &in_pt ) const
{
    PtType ptRes;

    for ( int i = 0; i < 3; i++ ) {
        ptRes[i] = m_aptPoly[0][i] + m_aptPoly[1][i] * in_pt[0] + m_aptPoly[2][i] * in_pt[1];
    }

    for ( FORINT i = 0; i < m_aptRad.num(); i++ ) {
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt ) * m_adSigmaSq[i];

        const double dPhi = m_crv( dDistSq );

        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ )
            ptRes[j] += ptLambda[j] * dPhi;
    }

    return ptRes;
}


template<class PtType, class VecType>
inline
VecType FITRbfSplineTC<PtType,VecType>::DerivS( const R2Pt &in_pt ) const
{
    VecType vecRes;

    for ( int i = 0; i < 3; i++ ) {
        vecRes[i] = m_aptPoly[1][i];
    }

    for ( FORINT i = 0; i < m_aptRad.num(); i++ ) {
        const double dXDist = 2.0 * (in_pt[0] - m_aptRad[i][0]) * m_adSigmaSq[i];
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt ) * m_adSigmaSq[i];

        const double dPhiDs = dXDist * m_crv.Derivative( dDistSq );

        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ )
            vecRes[j] += ptLambda[j] * dPhiDs;
    }

    return vecRes;
}

template<class PtType, class VecType>
inline
VecType FITRbfSplineTC<PtType,VecType>::DerivT( const R2Pt &in_pt ) const
{
    VecType vecRes;

    for ( int i = 0; i < 3; i++ ) {
        vecRes[i] = m_aptPoly[2][i];
    }

    for ( FORINT i = 0; i < m_aptRad.num(); i++ ) {
        const double dYDist = 2.0 * (in_pt[1] - m_aptRad[i][1]) * m_adSigmaSq[i];
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt ) * m_adSigmaSq[i];

        const double dPhiDt = dYDist * m_crv.Derivative( dDistSq );

        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ )
            vecRes[j] += ptLambda[j] * dPhiDt;
    }

    return vecRes;
}

template<class PtType, class VecType>
inline
VecType FITRbfSplineTC<PtType,VecType>::DerivSS( const R2Pt &in_pt ) const
{
    VecType vecRes;

    RNZero( vecRes );

    for ( int i = 0; i < m_aptRad.num(); i++ ) {
        const double dDistSq = LengthSq( in_pt - m_aptRad[i] ) * m_adSigmaSq[i];
        const double dDerivU = 2.0 * (in_pt[0] - m_aptRad[i][0]) * m_adSigmaSq[i];
        const double dDerivBasis = m_crv.Derivative( dDistSq );
        const double dPhiDss = (2.0 * dDerivBasis * m_adSigmaSq[i] + dDerivU * dDerivU * m_crv.SecondDerivative( dDistSq ));

        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ )
            vecRes[j] += ptLambda[j] * dPhiDss;
    }
    return vecRes;
}

template<class PtType, class VecType>
inline
VecType FITRbfSplineTC<PtType,VecType>::DerivTT( const R2Pt &in_pt ) const
{
    VecType vecRes;

    RNZero( vecRes );

    for ( int i = 0; i < m_aptRad.num(); i++ ) {
        const double dDistSq = LengthSq( in_pt - m_aptRad[i] ) * m_adSigmaSq[i];
        const double dDerivV = 2.0 * (in_pt[1] - m_aptRad[i][1]) * m_adSigmaSq[i];
        const double dDerivBasis = m_crv.Derivative( dDistSq );
        const double dPhiDtt = (2.0 * dDerivBasis * m_adSigmaSq[i] + dDerivV * dDerivV * m_crv.SecondDerivative( dDistSq ));

        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ )
            vecRes[j] += ptLambda[j] * dPhiDtt;
    }
    return vecRes;
}


template<class PtType, class VecType>
inline
VecType FITRbfSplineTC<PtType,VecType>::DerivST( const R2Pt &in_pt ) const
{
    VecType vecRes;

    RNZero( vecRes );
    for ( int i = 0; i < m_aptRad.num(); i++ ) {
        const double dXDist = 2.0 * (in_pt[0] - m_aptRad[i][0]) * m_adSigmaSq[i];
        const double dYDist = 2.0 * (in_pt[1] - m_aptRad[i][1]) * m_adSigmaSq[i];
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt ) * m_adSigmaSq[i];

        const double dPhiDst = dXDist * dYDist * m_crv.SecondDerivative( dDistSq );

        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ )
            vecRes[j] += ptLambda[j] * dPhiDst;
    }
    return vecRes;
}

template<class PtType, class VecType>
inline
void FITRbfSplineTC<PtType,VecType>::AllEval( const R2Pt &in_pt,
                                                PtType &out_pt,
                                                VecType &out_vecDs, VecType &out_vecDt,
                                                VecType &out_vecDss, VecType &out_vecDtt,
                                                VecType &out_vecDst ) const
{
    for ( int i = 0; i < Dim(); i++ ) {
        out_pt[i] = m_aptPoly[0][i] + m_aptPoly[1][i] * in_pt[0] + m_aptPoly[2][i] * in_pt[1];
        out_vecDs[i] = m_aptPoly[1][i];
        out_vecDt[i] = m_aptPoly[2][i];
        out_vecDss[i] = 0.0;
        out_vecDtt[i] = 0.0;
        out_vecDst[i] = 0.0;
    }

    for ( FORINT i = 0; i < m_aptRad.num(); i++ ) {
        const double dXDist = 2.0 * (in_pt[0] - m_aptRad[i][0]) * m_adSigmaSq[i];
        const double dYDist = 2.0 * (in_pt[1] - m_aptRad[i][1]) * m_adSigmaSq[i];
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt ) * m_adSigmaSq[i];

        const double dPhi = m_crv( dDistSq );

        const double dPhiDs = dXDist * m_crv.Derivative( dDistSq );
        const double dPhiDt = dYDist * m_crv.Derivative( dDistSq );
        const double dDerivBasis = m_crv.Derivative( dDistSq );
        const double dDerivBasis2 = m_crv.SecondDerivative( dDistSq );
        const double dPhiDss = (2.0 * dDerivBasis * m_adSigmaSq[i] + dXDist * dXDist * dDerivBasis2);
        const double dPhiDtt = (2.0 * dDerivBasis * m_adSigmaSq[i] + dYDist * dYDist * dDerivBasis2);
        const double dPhiDst = dXDist * dYDist * dDerivBasis2;
        
        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ ) {
            out_pt[j] += ptLambda[j] * dPhi;
            out_vecDs[j] += ptLambda[j] * dPhiDs;
            out_vecDt[j] += ptLambda[j] * dPhiDt;
            out_vecDss[j] += ptLambda[j] * dPhiDss;
            out_vecDtt[j] += ptLambda[j] * dPhiDtt;
            out_vecDst[j] += ptLambda[j] * dPhiDst;
        }
    }

}

template<class PtType, class VecType>
inline
void FITRbfSplineTC<PtType,VecType>::AllEvalSub( const R2Pt &in_pt,
                               PtType &out_pt,
                               VecType &out_vecDs, VecType &out_vecDt ) const
{
    for ( int i = 0; i < Dim(); i++ ) {
        out_pt[i] = m_aptPoly[0][i] + m_aptPoly[1][i] * in_pt[0] + m_aptPoly[2][i] * in_pt[1];
        out_vecDs[i] = m_aptPoly[1][i];
        out_vecDt[i] = m_aptPoly[2][i];
    }

    for ( FORINT i = 0; i < m_aptRad.num(); i++ ) {
        const double dXDist = 2.0 * (in_pt[0] - m_aptRad[i][0]) * m_adSigmaSq[i];
        const double dYDist = 2.0 * (in_pt[1] - m_aptRad[i][1]) * m_adSigmaSq[i];
        const double dDistSq = LengthSq( m_aptRad[i] - in_pt ) * m_adSigmaSq[i];

        const double dPhi = m_crv( dDistSq );

        const double dPhiDs = dXDist * m_crv.Derivative( dDistSq );
        const double dPhiDt = dYDist * m_crv.Derivative( dDistSq );
        
        const PtType &ptLambda = m_aptLambda[i];
        for ( int j = 0; j < Dim(); j++ ) {
            out_pt[j] += ptLambda[j] * dPhi;
            out_vecDs[j] += ptLambda[j] * dPhiDs;
            out_vecDt[j] += ptLambda[j] * dPhiDt;
        }
    }

}

template<class PtType, class VecType>
inline
void FITRbfSplineTC<PtType,VecType>::Set( const Array< R2Pt > &in_acpt, 
                        const Array< PtType > &in_apt )
{
    Array<double> adDist( in_apt.num() * in_apt.num() );
    for ( int i = 0; i < in_apt.num(); i++ ) {
        for ( int j = i; j < in_apt.num(); j++ ) {
            adDist[ i * in_apt.num() + j ] = LengthSq( in_acpt[i] - in_acpt[j] );
        }
    }

    m_adSigmaSq.need( in_acpt.num() );
    Array<double> adSort( in_acpt.num() );
    for ( FORINT i = 0; i < in_acpt.num(); i++ ) {
        for ( int j = 0; j < i; j++ )
            adSort[j] = adDist[ j * in_acpt.num() + i ];
        for ( FORINT j = i; j < in_acpt.num(); j++ )
            adSort[j] = adDist[ i * in_acpt.num() + j ];

        qsort( &adSort[0], adSort.num(), sizeof(double), FITRbfTC_SortFc );

        const int iN = WINmin( 10, in_acpt.num() );
        m_adSigmaSq[i] = 0.0;
        for ( FORINT j = 0; j < iN; j++ ) {
            m_adSigmaSq[i] += adSort[j];
        }
        m_adSigmaSq[i] /= (double) iN;
    }

    FITTools_LS oLS( in_apt.num() + 3, in_apt.num() + 3, Dim() );

    for ( FORINT i = 0; i < in_apt.num(); i++ ) {
        for ( int j = i; j < in_apt.num(); j++ ) {

            const double dDistSq = LengthSq( in_acpt[i] - in_acpt[j] );

            oLS.Desired_clapack(i,j) = m_crv( dDistSq * m_adSigmaSq[j] );
            oLS.Desired_clapack(j,i) = m_crv( dDistSq * m_adSigmaSq[i] );
        }
        oLS.Desired_clapack( i, in_apt.num() + 0 ) = 1.0;
        oLS.Desired_clapack( i, in_apt.num() + 1 ) = in_acpt[i][0];
        oLS.Desired_clapack( i, in_apt.num() + 2 ) = in_acpt[i][1];

        oLS.Desired_clapack( in_apt.num() + 0, i ) = 1.0;
        oLS.Desired_clapack( in_apt.num() + 1, i ) = in_acpt[i][0];
        oLS.Desired_clapack( in_apt.num() + 2, i ) = in_acpt[i][1];

        for ( FORINT j = 0; j < Dim(); j++ ) {
            oLS.Rhs_clapack(i, j) = in_apt[i][j];

        }
    }
    VERIFY( oLS.Solve_clapack() );

    m_aptLambda.need( in_apt.num() );
    m_aptRad = in_acpt;
    for ( FORINT i = 0; i < in_apt.num(); i++ ) {
        for ( int j = 0; j < Dim(); j++ ) {
            m_aptLambda[i][j] = oLS.Solution_clapack( i, j );
        }
    }
    for ( FORINT i = 0; i < 3; i++ ) {
        for ( int j = 0; j < Dim(); j++ ) {
            m_aptPoly[i][j] = oLS.Solution_clapack( in_apt.num() + i, j );
        }
    }

    for ( FORINT i = 0; i < in_apt.num(); i++ ) {
        const PtType ptEval = Eval( in_acpt[i] );
        const double dLen = Length( ptEval - in_apt[i] );
        ASSERT( RNIsZero( dLen, 1e-4 ) );
    }
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


template<class PtType, class VecType>
inline
void FITRbfSplineTC<PtType,VecType>::Write( ofstream &out ) const
{
    FITRbfTC<PtType,VecType>::Write(out);
    m_crv.Write(out);

    for ( int i = 0; i < m_adSigmaSq.num(); i++ )
        out << m_adSigmaSq[i] << " ";

    out << "\n";
}

template<class PtType, class VecType>
inline
void FITRbfSplineTC<PtType,VecType>::Read( ifstream &in ) 
{
    FITRbfTC<PtType,VecType>::Read(in);
    m_crv.Read(in);

    m_adSigmaSq.need( m_aptRad.num() );

    for ( int i = 0; i < m_adSigmaSq.num(); i++ )
        in >> m_adSigmaSq[i];
}

template<class PtType, class VecType>
inline
WINbool FITRbfSplineTC<PtType,VecType>::operator==( const FITRbfSplineTC<PtType,VecType> &in_func ) const
{
    if ( !( m_crv.Order() == in_func.m_crv.Order() ) )
        return FALSE;

    return FITRbfTC<PtType,VecType>::operator==( in_func );
}

///
template<class PtType, class VecType>
inline
FITRbfSplineTC<PtType,VecType> &FITRbfSplineTC<PtType,VecType>::operator=( const FITRbfSplineTC<PtType,VecType> &in_func )
{
    FITRbfTC<PtType,VecType>::operator=( in_func );

    m_crv = in_func.m_crv;
    m_adSigmaSq = in_func.m_adSigmaSq;

    return *this;
}

template<class PtType, class VecType>
inline
void FITRbfSplineTC<PtType,VecType>::SetContinuity( const int in_iC )
{
    m_crv = CRVBasis( WINmax(in_iC, -1), -0.05, 0.05 );
}



