/* Copyright 1995, Brown Computer Graphics Group.  All Rights Reserved. */

#ifndef __CRV_ALPHA_DEFS_H_
#define __CRV_ALPHA_DEFS_H_

/** \defgroup SplinesBSplinesRefinement Spline curve refinement
  \ingroup SplinesBSplines
  \brief Refining the curve by introducing new knots

   Refinement adds new control points by introducing new knots. The old
   control points plus the new ones are placed so that the curve is
   exactly the same as it was before refinement. These classes calculate
   the new locations as a weighted combination of the old ones.
   */
//@{
//@}

/** \class CRVAlpha Crv_Alpha.H utils/Crv_Alpha.H
  \ingroup SplinesBSplinesRefinement
  \brief Alpha function used to do refinement (see the 3B's book)
 
   Calculating the alpha functions for refinement
   One alpha function per each old index/new index pair

  Constant on knot intervals (one of these for each old control point).
  Each object has coefficients for all new control points.

  <br> Files:
- include/utils/Crv_Alpha.H
- src/utils/splines/CRVAlpha.cpp
  */
class CRVAlpha : public CRVSpace {
private:
    const int             m_iIndex;
    const Array<double> & m_adOldVec, & m_adNewVec;
    CRVAlpha            * m_alphaIth, * m_alphaIthPlus1;

public:
    double operator()(int in_iNewIndex) const;
    
    CRVAlpha( const CRVSpace &,
              const Array<double> &in_adOldVec, 
              const Array<double> &in_adNewVec,
              int in_iIndex);
    virtual ~CRVAlpha() {
        if (m_alphaIth) { 
            delete m_alphaIth; 
            delete m_alphaIthPlus1; 
        } 
    }
};

/** \class CRVNurb_alpha Crv_Alpha.H utils/Crv_Alpha.H
  \ingroup SplinesBSplinesRefinement
  \brief The functions used to do refinement (see the 3B's book)

  Collects one alpha function for each old control point.

  Given the old knot vector and the new knot vector, calculate the weight combinations
  of each old control point that are needed for the new control points, i.e.,
  each new control point is a linear combination of the old ones. 

  <br> Files:
- include/utils/Crv_Alpha.H
- src/utils/splines/CRVAlpha.cpp
*/
class CRVNurb_alpha : public CRVSpace {
private:
    CRVNurb_alpha(CRVNurb_alpha &);
    CRVNurb_alpha &operator=(CRVNurb_alpha &);

private:
    
    Array<double>      m_adOldKnot, m_adNewKnot;
    
    // One of these for each of the old control points
    SArray<CRVAlpha *> m_aAlphas;

public:
    double operator()(int in_iOldIndex, int in_iNewIndex);
    
    // Make a set of coefficients for changing control points out of the old
    // and new knot vecs
    CRVNurb_alpha( const CRVSpace &,
                   const Array<double> &in_adOldKnot, 
                   const Array<double> &in_adNewKnot);
    virtual ~CRVNurb_alpha();
    
    void Print() const;
};


#endif
