#ifndef _CRV_HERMITE_H_DEFS_
#define _CRV_HERMITE_H_DEFS_

/** \class CRVHermiteTC Crv_HermiteTC.H utils/Crv_HermiteTC.H
   \ingroup HermiteCrvs

  \brief Hermite curve, defined by end points and vectors.

This is a cubic curve built from segments, each of which is
 defined by two end points and two end vectors. 

  The curve is always evaluated between 0 and 1

Files:
 - include/utils/Crv_HermiteTC.H
 - include/utils/Crv_HermiteTC_i.H
*/
template<class Pt, class Vec>
class CRVHermiteTC {
protected:
    Array<Pt>  m_apt;
    Array<Vec> m_avec;

public:
    //*@name Access */
    //@{
    ///
    int NumPts() const { return m_apt.num(); }
    ///
    const Pt &GetPt( const int in_i ) const { return m_apt[in_i]; }
    ///
    const Vec &GetVec( const int in_i ) const { return m_avec[in_i]; }
    //@}

    //*@name Setting */
    //@{
    ///
    void Add( const Pt &in_pt, const Vec &in_vec ) { m_apt += in_pt; m_avec += in_vec; }
    ///
    Pt &SetPt( const int in_i ) { return m_apt[in_i]; }
    ///
    Vec &SetVec( const int in_i ) { return m_avec[in_i]; }
    //@}

    //@*name Evaulation */
    //@{
    /// 0 <= t <= 1
    Pt operator()( const double in_dT );
    ///
    Vec Deriv( const double in_dT );
    ///
    Vec SecondDeriv( const double in_dT );
    //@}

    ///
    CRVHermiteTC( const int in_iN = 0 ) : m_apt(in_iN), m_avec(in_iN) {}
    ///
    virtual ~CRVHermiteTC() {}

    ///
    void Write( ofstream &out ) const;
    ///
    void Read( ifstream &in );
};

#include "Crv_HermiteTC_i.H"

/** \ingroup HermiteCrvs
 \brief 3D Hermite curve */
typedef CRVHermiteTC<R3Pt, R3Vec> CRVHermite3;
/** \ingroup HermiteCrvs
 \brief 2D Hermite`fl curve */
typedef CRVHermiteTC<R2Pt, R2Vec> CRVHermite2;
/** \ingroup HermiteCrvs
 \brief 1D Hermite curve */
typedef CRVHermiteTC<R1Pt, R1Vec> CRVHermite1;

#endif