/* Copyright 1995, Brown Computer Graphics Group.  All Rights Reserved. */

#ifndef CRV_KNOT_VEC_DEFS
#define CRV_KNOT_VEC_DEFS

#include <utils/Crv_Basis.H>

/* -------------------  Constants  ------------------------------- */

/* ---------------------  Classes  ------------------------------- */

/** \defgroup SplinesBSplinesKnot Knot vector and basis functions
   \ingroup SplinesBSplines

  \brief The part of the curve that is not geometric, i.e., b(t) from C(t) = sum_i b_i(t) g_i

  Consists of a knot vector and basis functions (b_i(t)) built over that knot vector.
*/
//@{
//@}


/** \class CRVKnot_vec Crv_Knot_vec.H utils/Crv_Knot_vec.H
   \ingroup SplinesBSplinesKnot

  \brief The knot vector for a non uniform b-spline

   Knot vectors are defined by the desired degree (-1 points, 0
   polyline, 1 quadratic, 2 cubic, etc.), the number of points, and a
   monotonically increasing list of numbers which are the knots. Basis
   functions are defined for each interval (degree -1 implies 1 gap,
   degree 0 implies 2 gap, etc.), essentially bumps over some number
   of knots. <br>

   What is implemented here is the the knot vector, which is a non-decreasing, ordered list 
   of numbers. Each basis function (CRVBasis) is defined over k+1 knots. The basis functions are
   constructed using the recursion formula in the BBB spline book. A
   bit slow, but elegent. <br>

   The knot vector is the knots plus the basis functions defined over those knots.

<br> Files:
- include/utils/Crv_Knot_vector.H
- include/utils/Crv_Knot_vector_i.H
- src/utils/splines/CRVKnot_vector.cpp
*/
class CRVKnot_vec : public CRVSpace {
protected:
    Array<double>   m_adKnots; // Knot vector
    Array<CRVBasis> m_aBasis;  // Basis functions on the knot vector
    
    // Add the points at the given index
    // Remove n_pts from the index s_i on
    WINbool Add_knots(int in_iN, int in_iStart);
    WINbool Remove_knots(int in_iN, int iStart);
    
    // Should be same as Num_pts()
    int Num_basis() const { return m_aBasis.num(); }
    CRVBasis &Basis(int in_i) { return m_aBasis[in_i]; }
    
public:
    /**@name Accessors */
    //@{
    /// Number of points
    inline int Num_pts() const;

    /// Determined by number of points requested.
    inline int    Num_knots() const;
    /// Number of knots - 1
    inline int    Num_segments() const;
    /// Where the ith fully supported segment starts
    inline double Segment_start_t(int i) const ;
    /// Where the ith fully supported segment ends
    inline double Segment_end_t(int i) const;
    /// Return the t value which is perc along the segment. 0 < in_dPerc < 1
    inline double Segment_perc_t(int i, double in_dPerc) const;
    /// Which knot is in the middle of segment i
    inline double BasisCenter( const int in_iWhich ) const;
    ///
    const CRVBasis &Basis(int in_i) const { return m_aBasis[in_i]; }
   
    /// The parameter point at which the curve becomes valid
    inline double Start_t()   const;
    /// The parameter point at which the curve becomes invalid
    inline double End_t()     const;
    /// Percentage along valid part of curve. 0 < in_dPerc < 1
    inline double Perc_t(double in_dPerc) const;
    /// NOT length of curve.
    inline double DomainLength() const { return End_t() - Start_t(); } 

    /// NOT start of curve, but start of support of first basis function.
    inline double First_knot()    const;
    ///
    inline double Knot(int which) const;
    /// NOT end of curve, but end of support of last basis function.
    inline double Last_knot()     const;

    /** Index of first knot above or at t value.
        @return 0 or Knots().num() if out of range */
    int    Ceil_knot(double in_dT) const;
    /** Index of first knot below or at t value.
        @return 0 or Knots().num() if out of range */
    int    Floor_knot(double in_dT) const;
    /** Index of first valid segment below or at t value.
        @return -1 or Knots().Dim() if out of range */
    int    Segment(double in_dT) const;
    /// returns false if point outside of range
    WINbool Span( const double in_dT, int &out_iStart, int &out_iEnd ) const;
    ///
    inline const Array<double> Knots() const;
    //@}

    /**@name Evaluation of basis functions */
    //@{
    /// Evaluate a specific basis function
    inline double Basis(double in_dT, int in_i) const;
    /// The derivative of a specific basis function
    inline double BasisDerivative(double in_dT, int in_i) const;
    /// The second derivative of a specific basis function
    inline double BasisSecondDerivative(double in_dT, int in_i) const;
    /// Return a curve which evaluates to the derivative of this curve.
    CRVKnot_vec  Derivative() const;

    /** Evaluate all basis functions.
        Only the non-zero basis functions are returned.
        @param in_dT t value to evaluate.
        @param out_aiIndx The indices of the non-zero basis functions (at most order +1)
        @param out_adValue The values of each basis function at in_dT.
        @return the total sum ( which is 1 if in_dT is within Start_t and End_t ).
        */
    double EvalAllBasis( double in_dT, 
                         Array<int> &out_aiIndx, 
                         Array<double> &out_adValue ) const;
    /** Evaluate all basis functions.
        Only the non-zero basis functions are returned.
        @param in_dT t value to evaluate.
        @param out_aiIndx The indices of the non-zero basis functions (at most order +1)
        @param out_adValue The values of each basis function at in_dT.
        @return the total sum ( which is 1 if in_dT is within Start_t and End_t ).
        */
    double EvalAllBasisDeriv( double in_dT, 
                              Array<int> &out_aiIndx, 
                              Array<double> &out_adValue ) const;
    //@}

    /**@name Operations on the knot vector. */
    //@{
    /** Set the knots _between_ in_iStart and in_iEnd.
        
        To set all the knots, in_adNewKnots must have the same number
        of knots as the knot vector, in_iStart = -1, and in_iEnd =
        number of knots.

        @param in_adNewKnots Monotonically increasing, in_adNewKnots[
        in_iStart +1] > Knot[in_iStart], and in_adNewKnots[ in_iEnd-1 ] <
        Knot[in_iEnd ]
        @param in_iStart Knot in_iStart +1 is set to first knot in in_adNewKnots
        @param in_iEnd Knot in_iEnd is set to last knot in in_adNewKnots
        @return TRUE if creates valid knot vector
    */
    WINbool Set_knots(const Array<double> &in_adNewKnots, 
                      int in_iStart, int in_iEnd);
    /** Evenly spaced knots.
        @param in_dStart Knot(in_iStart) will be in_dStart
        @param in_dEnd Knot(in_iEnd) will be in_dEnd
        @param in_iStart Knot to start change at
        @param in_iEnd Knot to end change at
    */
    void Set_knots(double in_dStart, double in_dEnd, int in_iStart, int in_iEnd);
    /** Evenly spaced knots.
        @param in_dStart Start_t() will be in_dStart (first knot will be to the left of in_dStart).
        @param in_dEnd End_t() will be in_dEnd (last knot will be to the right of in_dEnd).
    */
    inline void Set_knots(double in_dStart, double in_dEnd);
    /** Evenly spaced knots.
        @param in_dStart Start_t() will be in_dStart (first knot will be to the left of in_dStart).
        @param in_dEnd End_t() will be in_dEnd (last knot will be to the right of in_dEnd).
        @param in_dEndSpacing controls how the knots to the left and right of Start_t() and End_t() will be set
    */
    inline void Set_knots(double in_dStart, double in_dEnd, double in_dEndSpacing);
    //@}

    /**@name Constructors, assignments, and comparator */
    //@{
    ///
    CRVKnot_vec &operator=(const CRVKnot_vec &);
    ///
    WINbool operator==(const CRVKnot_vec &) const ;
    ///
    CRVKnot_vec(const CRVKnot_vec &);
    ///
    CRVKnot_vec(ifstream &in, WINbool bIn_binary);
    ///
    CRVKnot_vec(const CRVSpace &in_oSp = CRVSpace(2), int in_iNumPts = 6);
    ///
    virtual ~CRVKnot_vec() {}
    //@}

    /**@name Read, write, print.
       Can read and write in binary or text. Named with KnotVec because stupid sun compiler can't do override properly.
       */
    //@{
    ///
    virtual void    PrintKnotVec(WINbool do_basis) const;
    ///
    virtual WINbool ReadKnotVec(ifstream &in);
    ///
    virtual void    WriteKnotVec(ofstream &out) const;
    /// Read the knot vector in binary; does not work cross platform (big indian vs little)
    virtual WINbool Read_binary(ifstream &in);
    /// Write the knot vector in binary; does not work cross platform (big indian vs little)
    virtual void    Write_binary(ofstream &out) const;
    /// Check that knot vector is monotonically increasing
    virtual WINbool CheckKnotVec(WINbool in_bCheckSum = FALSE) const;
    //@}
};

#include "Crv_Knot_vec_i.H"
#endif

