/* Copyright 1995, Brown Computer Graphics Group.  All Rights Reserved. */



/* -------------------------------------------------------------------------
 *
 * A non uniform rational b-spline
 *
 * ------------------------------------------------------------------------- */

#ifndef _CRVNURBS_DEFS
#define _CRVNURBS_DEFS

#include <utils/Crv_Knot_vec.H>
#include <utils/Rn_Projective.H>
#include <utils/Utils_Io.H>
#include <utils/Crv_Alpha.H>



/* -------------------  Constants  ------------------------------- */

/* ---------------------  Classes  ------------------------------- */

/** \class CRVNurbTC Crv_NurbTC.H utils/Crv_NurbTC.H
   \ingroup SplinesBSplines

  \brief An entire NURBS curve.

    Whether this is a nurb or nub curve depends upon the type of
    InPoint and OutPoint. If InPoint is a homogenous point, and
    OutPoint is a regular point, then the homogenous conversion
    happens automatically (corresponding to dividing by the weights). <br>

  Helper classes:
  - CRVSpace order of curve
  - CRVKnot_VecTC Knot vector
  - CRVBasis A single basis function
  - CRVAlpha Alpha function for refinement
  - CRVNurb_Alpha For refining the entire curve
  - CRVSampleTC A linear set of samples that represents the curve (for closest point)
  - CRVNurb_sampledTC A hierarchy of linear samples

Files:
<br> Files in: include/utils <br>
  - Crv_Nurb_defs.H
  - Crv_Nurb_i.H
  - Crv_NurbTC.H

<br> Files in: utils/splines/ <br>
  - CRVNurbTC_test.cpp
  - CRVNurbTC.cpp

 */
template<class InPoint, class OutPoint, class OutVector>
class CRVNurbTC : public CRVKnot_vec {
protected:
    // The points to go with the knot vector
    Array<InPoint> m_apts;      
    
    inline CRVNurbTC<InPoint, OutPoint, OutVector> Clip_crv(double in_dStart, double in_dEnd) const;
    
    inline double Integrate_interval(double in_dStart, double in_dEnd) const;
    inline double Integrate_segment(int in_iSeg) const;
    
    
public:
    /**@name Accessors */
    //@{
    ///
    inline int InDim() const;
    ///
    inline int OutDim() const;

    /// The control points
    const InPoint &Pt(int in_i) const { return m_apts[in_i]; }
    ///
          InPoint &Pt(int in_i)       { return m_apts[in_i]; }
    //@}

    /**@name Evaluating the curve */
    //@{
    /// First point on the curve
    inline OutPoint Start_crv()    const { return (*this)(Start_t()); }
    /// Last point on the curve
    inline OutPoint End_crv()      const { return (*this)(End_t()); }
    ///
    inline OutPoint operator()(double in_dT) const ;
    /// 0 < in_dPerc < 1
    inline OutPoint Perc(double in_dPerc) const;
    /// 0 < in_dPerc < 1, evaluate a percentage along the given segment
    inline OutPoint PercSegment(int in_iSeg, double in_dPerc) const;
    /// Return the in_iNDivs evenly spaced parameter values
    inline Array<double> Even_spacing(int in_iNDivs) const;
    
    /// Approximate length of the curve from in_dStart to in_dEnd
    inline double Length(double in_dStart, double in_dEnd) const;
    /// Approximate length of entire curve
    inline double Length() const;

    ///
    inline OutVector Derivative(double in_dT) const ;
    ///
    inline OutVector SecondDerivative(double in_dT) const ;
    /// A curve which evaluates to the derivative of this curve
    inline CRVNurbTC<InPoint, OutPoint, OutVector> Derivative() const;
    /** A coordinate system.
        Only really valid for dimension 3 curves.
        @param in_dT The curve point at which to get the frame
        @param out_vecT The tangent (derivative) of the curve
        @param out_vecM The second derivative of the curve
        @param out_vecB out_vecT cross out_vecM
        */
    inline void Frame(double in_dT, 
                    OutVector &out_vecT, OutVector &out_vecM, OutVector &out_vecB) const;

    //@}

    /**@name Altering the curve.
      
      See CRVKnot_vector for changing knot values. */
    //@{
    /// Remove point point[in_iStart] on.
    inline WINbool Remove_points(int in_iNPts, int in_iStart);
    /// Point[in_iStart] gets first point (fills in reasonable values for new points)
    inline WINbool Add_points(int in_iNPts, int in_iStart);
    /** Refine the curve.
        Refine by adding a single knot.
        @return The refined curve
        @param in_bDupsAllowed If TRUE,  in_dNewKnot can be the same as an existing knot. Default FALSE.
        @param in_dNewKnot The new knot. */
    inline CRVNurbTC<InPoint, OutPoint, OutVector>
    Refine(double in_dNewKnot, WINbool in_bDupsAllowed = FALSE)  const ;
    /** Refine the curve.
        Refine by adding several knots.
        @return The refined curve
        @param in_bDupsAllowed If TRUE,  in_dNewKnot can be the same as an existing knot. Default FALSE.
        @param in_adNewKnots The monotonically increasing list of new knots. */
    inline CRVNurbTC<InPoint, OutPoint, OutVector>
            Refine(const Array<double> &in_adNewKnots, 
                            WINbool in_bDupsAllowed =FALSE) const;
    //@}
    
    /**@name Closest points */
    //@{
    /// Return the parameter value of the closest point.
    inline double Closest_t(const OutPoint &in_p, int in_iNDivs = 50) const ;
    /// Return the actual closest point.
    inline OutPoint Closest_pt(const OutPoint &in_p, int in_iNDivs = 50) const ;
    //@}

    /**@name Construction, assignement, comparison */
    //@{
    /// Return a subset of the curve
    inline CRVNurbTC<InPoint, OutPoint, OutVector> Sub_crv(double in_dStart_t, double in_dEnd_t) const ;
    ///
    inline CRVNurbTC<InPoint, OutPoint, OutVector> &operator=(const CRVNurbTC &);
    ///
    inline WINbool operator==(const CRVNurbTC<InPoint, OutPoint, OutVector> &) const;

    /// 
    inline CRVNurbTC(const CRVNurbTC<InPoint, OutPoint, OutVector> &);
    ///
    inline CRVNurbTC(ifstream &in, WINbool bIs_binary);
    /// Default curve is cubic with 10 points
    inline CRVNurbTC(const CRVSpace &crvSp, int in_iNPts);
    /// Default curve is cubic with 10 points
    inline CRVNurbTC();
    ///
    virtual ~CRVNurbTC() {}
    //@}

    /**@name Read, write, print
       Can read or write in binary. Not cross platform. */
    //@{
    ///
    inline void    Print(WINbool print_basis = FALSE) const;
    ///
    inline WINbool Read(ifstream &in);
    ///
    inline void    Write(ofstream &out) const;
    ///
    inline WINbool Read_binary(ifstream &in);
    ///
    inline void    Write_binary(ofstream &out) const;
    //@}

    /**@name Debug routines */
    //@{ 
    /// Check that the number of basis functions and points matches the number of knots.
    inline WINbool Check( WINbool in_bDoKnots = TRUE) const;
    ///
    inline WINbool Test();
    //@}

};

/** \ingroup SplinesBSplines 
   \brief 3D curve*/
typedef CRVNurbTC<R3Pt, R3Pt, R3Vec> CRVNub3TC;
/** \ingroup SplinesBSplines 
   \brief 2D curve*/
typedef CRVNurbTC<R2Pt, R2Pt, R2Vec> CRVNub2TC;
/** \ingroup SplinesBSplines 
   \brief 1D curve*/
typedef CRVNurbTC<R1Pt, R1Pt, R1Vec> CRVNub1TC;

//@}

#include "Crv_NurbTC_i.H"


#endif

