#ifndef CRV_NURB_SAMPLED_DEFS
#define CRV_NURB_SAMPLED_DEFS

#include <utils/Crv_NurbTC.H>

#include <utils/Rn_Line_seg.H>

/** \defgroup SplinesBSplinesSamples Sampling a NURBS curve

   \ingroup SplinesBSplines

    \brief Generate samples along a CRVNurbsTC curve for drawing or closest point.

  These classes are used to speed up closest point and drawing functions. They
  sample the curve in an arc-length fashion and keep those samples. If the
  curve changes, these need to be re-calculated. 
  */
//@{
//@}

/** \class CRVSamplesTC Crv_Nurb_sampledTC.H utils/Crv_Nurb_sampledTC.H
   \ingroup SplinesBSplinesSamples

  \brief  A linearly sampled NURBS curve.

    Samples are not necessarily evenly spaced in t, but are in length.

<br> Files in: include/utils <br>
  - Crv_SamplesTC_i.H
  - Crv_Nurb_sampledTC.H
  - Crv_Nurb_sampledTC_i.H

<br> Files in: utils/splines/ <br>
  - CRVNurb_sampledTC.cpp
  - CRVSamplesTC.cpp
 */
template<class Point, class Vector, class LineSeg>
class CRVSamplesTC {
private:
    int           m_iNDivs;
    Array<Point>  m_aptSamples;
    Array<double> m_adTs;
    
    // Current error of sample
    inline double Calc_err(const CRVNurbTC<Point, Point, Vector> &);

    // closest interval bounded by in_iStart and in_iEnd
    inline int   Closest_intervalMe(int in_iStart, int in_iEnd, const Point &in_pt) const;
    inline Point Closest(const Point &in_pt) const;

public:
    /** Maximum length of sample.
        Used to keep the maximum sample length below a given value.
        @param out_dTotLen Total sum of length of polyline sample. */
    inline double Max_len(double &out_dTotLen) const;
    /// The ith sample
    const Point &operator[](int i) const { return m_aptSamples[i]; }
    /// Number of samples
    int    N() const                     { return m_aptSamples.num(); }
    /// T value (parameter) of ith sample
    double T(int i) const                { return m_adTs[i]; }
    /// Take n_samples of the curve
    inline double Sample(const CRVNurbTC<Point, Point, Vector> &, int n_samples);
    /// Take divs samples between the previous layer's samples
    inline double Sample(const CRVNurbTC<Point, Point, Vector> &, const CRVSamplesTC<Point, Vector, LineSeg> &prev_smpls, int divs);
    
    /// Call private closest_interval w/approp. interval
    inline int Closest_interval(const Point &in_pt) const;
    /// Call private closest_interval w/approp. interval
    inline int Closest_intervalClipped(const Point &in_pt, const double in_dT1, const double in_dT2) const;

    /// Look in the samples corresponding to in_iPrev_interval for the interval containing the closest point
    inline int Closest_interval(const Point &in_pt, int in_iPrev_interval) const;

    /// Look in the samples corresponding to in_iPrev_interval for the closest point
    inline Point Closest(const Point &in_pt, int in_iPrev_interval) const;
    /** Look in the samples corresponding to in_iPrev_interval for the closest point.
        Also return T of sample in out_dT.
        */
    inline Point Closest(const Point &in_pt, int in_iPrev_interval, double &out_dT) const;

    ///
    inline CRVSamplesTC(const CRVSamplesTC<Point, Vector, LineSeg> &);
    ///
    inline CRVSamplesTC<Point, Vector, LineSeg> &operator=(const CRVSamplesTC<Point, Vector, LineSeg> &);
    
    ///
    CRVSamplesTC() {}
    ///
    virtual ~CRVSamplesTC() {}
};

/** \class CRVNurb_sampledTC Crv_Nurb_sampledTC.H utils/Crv_Nurb_sampledTC.H
   \ingroup SplinesBSplinesSamples

  \brief  A hierarchy of linearly sampled NURBS curve.

    A hierarchy of samples. Each level (CRVSamplesTC) is a subset of the previous
    one. The lowest level of samples form a polyline which is within
    epsilon of the curve, and with a minimum edge length. <br>

    This class is useful if you're doing drawing or closest point
    evaluations of a curve which is not changing. <br>

    Very fast closest point because the layers form an n-tree search space. <br>

    Sampling should probably be re-written to actually guarantee epsilon accuracy.

    Collect all the layers. To find the closest point, essentially do a binary search.
    Start with top layer, find appropriate segment, then look in samples for that interval in the
    next layer down. At the bottom level, do a linear interpolation of the closest samples.

  <br> Files in: include/utils <br>
  - Crv_SamplesTC_i.H
  - Crv_Nurb_sampledTC.H
  - Crv_Nurb_sampledTC_i.H

<br> Files in: utils/splines/ <br>
  - CRVNurb_sampledTC.cpp
  - CRVSamplesTC.cpp
 */
template<class Point, class Vector, class LineSeg>
class CRVNurb_sampledTC : public CRVNurbTC<Point, Point, Vector> {
private:
    Array<CRVSamplesTC<Point, Vector, LineSeg> > m_aSamples;

public:
    /// Get all the samples at the lowest level
    inline void Samples(Array<Point> &in_pt) const;
    /// Get all the samples between the two parameter values.
    inline void Samples(Array<Point> &out_ptSamples, double in_dSt, double in_dEt) const;
    
    /** Create new layers until the desired accuracy is achieved.
        @param in_dAccuracy accuracy of samples. 0.25 is a good value
        @param in_dMax_len Maximum length of samples */
    inline void Resample(double in_dAccuracy, double in_dMax_len);

    ///
    inline Point Closest(const Point &in_pt, double &out_dT) const;
    
    ///
    inline Point Closest(const Point &in_pt, double &out_dT, const double in_dT1, const double in_dT2) const;
    
    /// Assign new curve and resample
    inline CRVNurbTC<Point, Point, Vector> &operator=(const CRVNurbTC<Point, Point, Vector> &crv);

    /// Just copy self
    inline CRVNurb_sampledTC<Point, Vector, LineSeg> &operator=(const CRVNurb_sampledTC<Point, Vector, LineSeg> &);

    /// Create with given curve and sample
    inline CRVNurb_sampledTC( const CRVNurbTC<Point, Point, Vector> &in_crv, 
                              double in_dAcc, double in_dMax_len) ;

    /// Create default curve, no sampling
    inline CRVNurb_sampledTC(const CRVSpace &in_crvspace, int in_iNPts);

    ///
    inline CRVNurb_sampledTC(const CRVNurb_sampledTC<Point, Vector, LineSeg> &);

    
    ///
    virtual ~CRVNurb_sampledTC() {}
    
    /// Read a curve from a file then resample
    inline void ReadCrv(ifstream &in, double acc, double maxLen );

    ///
    inline WINbool test();
};

#include "Crv_SamplesTC_i.H"
#include "Crv_Nurb_sampledTC_i.H"


/** \ingroup SplinesBSplines 
 typedef for 3D sampled curves */
typedef CRVNurb_sampledTC<R3Pt, R3Vec, R3Line_seg > CRVNub_sampled3TC;
/** \ingroup SplinesBSplines 
 typedef for 2D sampled curves */
typedef CRVNurb_sampledTC<R2Pt, R2Vec, R2Line_seg > CRVNub_sampled2TC;

//@}

#endif
