#ifndef _PICK_MESH_DEFS_H
#define _PICK_MESH_DEFS_H

#include <utils/Rn_Defs.H>
#include <utils/R2_Contour.H>
#include <utils/Pm_TextureMap.H>
#include <utils/Pm_UVCoords.H>
#include <utils/Utils_Element.H>
#include <utils/Utils_Toggle.H>

#include <utils/Mesh_GMesh.H>

class PMMesh;
#include <utils/Pm_Vertex.H>
#include <utils/Pm_Edge.H>
#include <utils/Pm_Face.H>


/** A mesh.

    See PMTextureMap for making texture maps and PMUVCoords for texture coordinates. <br>

    All adjacency information is found in this class (e.g., an edge's
    two vertices), not on the vertex, edge, and face classes. <br>

    This class uses the GMesh underlying data structures. <br>


    If using GMesh, call Compute_normals() if the
    mesh has been changed to update the normals.
    

<br><br> Files:
<dd><li> include/utils/Pm_Vertex.H
<dd><li> include/utils/Pm_Edge.H
<dd><li> include/utils/Pm_Face.H
<dd><li> include/utils/Pm_Mesh.H
<dd><li> include/utils/Pm_Mesh_i.H
<dd><li> include/utils/Pm_Mesh_GMesh_i.H
<dd><li> utils/pickmesh/PMMesh.cpp
<dd><li> utils/pickmesh/PMVertex.cpp
<dd><li> utils/pickmesh/PMEdge.cpp
<dd><li> utils/pickmesh/PMFace.cpp
<dd><li> utils/pickmesh/PMTextureMap.cpp
<dd><li> utils/pickmesh/PMMeshCombineTris.cpp
<dd><li> utils/pickmesh/PMMeshIO.cpp
<dd><li> utils/pickmesh/OpenGL/PMMeshOpenGL.cpp
<dd><li> utils/pickmesh/PMMesh/PMMesh_GMesh.cpp */
class PMMesh 
{
public:

protected:
    
    // The actual mesh
    GMesh  m_mesh;

public:
    // Conversions back and forth between GMesh data
    // and PMMesh data
    friend const Face   & PMGetFace   (const PMFace &);
    friend const Edge   & PMGetEdge   (const PMEdge &);
    friend const Vertex & PMGetVertex (const PMVertex &);
    friend       Face   & PMGetFace   (PMFace &);
    friend       Edge   & PMGetEdge   (PMEdge &);
    friend       Vertex & PMGetVertex (PMVertex &);
    const PMVertex  &   FindVertex  ( const Vertex ) const ;
    const PMEdge    &   FindEdge    ( const Edge ) const ;
    const PMFace    &   FindFace    ( const Face ) const ;
          PMVertex  &   FindVertex  ( const Vertex );
          PMEdge    &   FindEdge    ( const Edge );
          PMFace    &   FindFace    ( const Face );

protected:
    void AssignPMsAfterRead();
    
    // To override PMVertex type:
    //   Override NewElement DeleteElement, ElementsRenumber
    //  Note: m_aoVerts, etc., is constructed as an adding element array
    //   so data is allocated by NewVertex, etc., and deleted by
    //   DeleteVertex, etc. To override the type, the deriving class
    //   defines it's own UTILSElementArrays, and New/Delete routines,
    //   putting the data in it's own arrays and then assigning pointers to m_aoVerts
    virtual PMVertex &NewVertex(  );
    virtual PMEdge   &NewEdge( );
    virtual PMFace   &NewFace( );
    virtual void      KillVertex(PMVertex *);
    virtual void      KillEdge  (PMEdge   *);
    virtual void      KillFace  (PMFace   *);
    void              WriteElements       ( ofstream &out ) const;
    void              WriteElementsBinary ( ofstream &out ) const;
    virtual void      ReadElements        ( ifstream &in ) ;
    virtual void      ReadElementsBinary  ( ifstream &in ) ;
    virtual void      ElementsRenumber( Array<int> &out_aiVMap, 
                                        Array<int> &out_aiEMap, 
                                        Array<int> &out_aiFMap );
    Array<PMVertex *>           m_aopDeadVs; // For storage if 
    Array<PMEdge *>             m_aopDeadEs; // we're using
    Array<PMFace *>             m_aopDeadFs; // m_aoVerts, etc.
    UTILSElementArray<PMVertex> m_aoVerts;
    UTILSElementArray<PMEdge>   m_aoEdges;
    UTILSElementArray<PMFace>   m_aoFaces;
    friend void PMSetVertMap( PMMesh &in_pm, int iIndx, int iId );
    friend void PMSetEdgeMap( PMMesh &in_pm, int iIndx1, int iIndx2, int iId );
    friend void PMSetFaceMap( PMMesh &in_pm, int iIndx, int iId );
    Array<int>                  m_aiMapVert, m_aiMapFace;
    Array< Array<int> >         m_aaiMapEdgeIds, m_aaiMapEdge;

    /// OpenGL variables
    static UTILSToggle m_sbFlatShade, m_sbEdges;
    WINbool            m_bHasChanged;
    
    /// Color and texture map
    Array<PMTextureMap *> m_aopTextureMaps;
    UTILSColor            m_colDefault;

    // Delete the specified items from the GMesh/BMesh
    inline void delete_vertex( PMVertex &io_v );
    inline void add_vertex   ( PMVertex &io_v );
    void delete_face  ( PMFace   &io_f );
    void add_face     ( const Array<const PMVertex *> &in_v, PMFace &io_f );

    // Renumber the specific mesh
    void mesh_renumber(  );

    // Compute faces and vertices
    void Compute_normals( Array<PMVertex *> &aopVs, Array<PMFace *> &aopFs );

public:
    /**@name Texture mapping */
    //@{
    ///
    inline int                 NTextures    ()               const { return m_aopTextureMaps.num(); }
    ///
    inline const PMTextureMap &TextureMap   ( int in_iIndx ) const ;
    ///
    inline PMTextureMap       &TextureMap   ( int in_iIndx ) ;
    ///
    inline PMTextureMap       &AddTextureMap();
    ///
    inline void                RemoveAllTextureMaps();
    ///
    inline void                RemoveTextureMap( const int in_iWhich );
    //@}

    /**@name Coloring */
    //@{
    ///
    void Color( const UTILSColor &in_col );
    /// Color the vertices by the texture map ( if there is one );
    void ColorByTextureMap();
    /// Color the point by the texture map ( if there is one );
    UTILSColor ColorByTextureMap( const PMFace &in_face, const Array<double> &in_adBary ) const;
    ///
    inline void SetDefaultColor( const UTILSColor &in_col ) { m_colDefault = in_col; }
    ///
    inline const UTILSColor & DefaultColor(  ) const { return m_colDefault; }
    //@}

    /**@name Updating normals */
    //@{
    /// If point locations have changed
    void Compute_normals( Array<PMVertex *> &aopVs );
    /// Do all
    void Compute_normals(  );
    //@}

    /**@name Geometry properties */
    //@{
    /// Centroid and scale of entire mesh
    R3Pt         GetCenter( R3Pt &out_ptScale ) const;

    /// Point in a face
    R3Pt         GetPointInFace( const PMFace        & in_oFace, 
                                 const Array<double> & in_adPerc ) const;

    /// Point along an edge
    R3Pt         GetPointOnEdge( const PMEdge &in_oEdge, double in_dPerc ) const;
    /// Center the mesh at the given point with the given raduis
    void Center( const R3Pt &in_ptCenter = R3Pt(0,0,0), const double in_dRadius = 0.5 );
	/// Make anything that seems like it should align, do so
	virtual void Regularize( double in_dPerc = 0.1 );
    //@}

    /**@name Changing the topology of the mesh */
    //@{
    ///
    inline PMVertex &AddVertex();
    ///
    virtual inline PMVertex &AddVertex(const R3Pt &in_pt);
    /// Orient  all the faces the opposite way (turn it inside out)
    virtual void             Flip();

    /// Make sure the incoming face is oriented correctly before adding
    WINbool IsOkToAdd( const Array<const PMVertex *> &in_aopVs );
    /** Add face.
        @param in_oFT Set to PM_NONE unless you're doing manifold texture mapping
        @param in_iTexture Which texture map to use
        @param in_iElemId Element in manifold the face came from */
    PMFace &AddFace( const Array<PMVertex *> &aiVertId, 
                     PMFaceType in_oFT = PM_NONE,
                     int in_iTexture = -1,
                     int in_iElemId = -1 );
    /** Add face.
        @param in_oFT Set to PM_NONE unless you're doing manifold texture mapping
        @param in_iTexture Which texture map to use
        @param in_iElemId Element in manifold the face came from */
    virtual PMFace &AddFace( const Array<const PMVertex *> &aiVertId, 
                             PMFaceType in_oFT = PM_NONE,
                             int in_iTexture = -1,
                             int in_iElemId = -1 );
    /// Also deletes any edges which no longer have any faces
    inline virtual void    RemoveFace          ( PMFace & );
    /// Also removes adjacent edges and faces
    inline virtual void    RemoveVertex        ( PMVertex & );
    /// Also removes adjacent faces
    inline virtual void    RemoveEdge          ( PMEdge & );
    /// Also deletes any edges which no longer have any faces
    virtual void           RemoveFaces          ( Array<PMFace *> & );
    /// Also removes adjacent edges and faces
    virtual void           RemoveVertices        ( Array<PMVertex *> & );
    /// Does not remove vertex
    void    RemoveAdjacentFaces ( PMVertex & );
    /// Blow it all away
    virtual void    Clear();

    /// Transform the mesh by the given affine transformation
    virtual void Transform( const R4Matrix &in_oFrame ) ;

    //@}

    /**@name Elements of the entire mesh */
    //@{
    ///
    inline int NumVertices()  const;
    ///
    inline int NumEdges()     const;
    ///
    inline int NumFaces()     const;
    ///
    inline int CAllMeshVertices ( Array<const PMVertex *> & ) const;
    ///
    inline int CAllMeshEdges    ( Array<const PMEdge *>   & ) const;
    ///
    inline int CAllMeshFaces    ( Array<const PMFace *>   & ) const;
    ///
    inline int AllMeshVertices ( Array<PMVertex *> & ) ;
    ///
    inline int AllMeshEdges    ( Array<PMEdge *>   & ) ;
    ///
    inline int AllMeshFaces    ( Array<PMFace *>   & ) ;
    //@}

    /**@name Grow regions of elements.
      Adds to existing list, does not delete. */
    //@{
    ///
    int Grow( const Array<PMVertex *> &in_aopVs, 
                    Array<PMEdge   *> &out_aopEs,
                    Array<PMFace   *> &out_aopFs );
    ///
    int Grow( const Array<PMFace *> &  in_aopFs, 
                    Array<PMVertex *> &out_aopVs,
                    Array<PMEdge   *> &out_aopEs );
    ///
    int Grow( const Array<PMEdge *> &  in_aopEs, 
                    Array<PMVertex *> &out_aopVs,
                    Array<PMFace   *> &out_aopFs );
    ///
    int CGrow( const Array<const PMVertex *> &in_aopVs, 
                     Array<const PMEdge   *> &out_aopEs,
                     Array<const PMFace   *> &out_aopFs ) const;
    ///
    int CGrow( const Array<const PMFace *> &  in_aopFs, 
                     Array<const PMVertex *> &out_aopVs,
                     Array<const PMEdge   *> &out_aopEs ) const;
    ///
    int CGrow( const Array<const PMEdge *> &  in_aopEs, 
                     Array<const PMVertex *> &out_aopVs,
                     Array<const PMFace   *> &out_aopFs ) const;
    //@}

    /**@name Element from UTILSId()
      VertexFromId( PMVertex::UTILSId ) is an identity function. */
    //@{
    ///
    WINbool Valid( const PMVertex * ) const ;
    ///
    WINbool Valid( const PMEdge * ) const ;
    ///
    WINbool Valid( const PMFace * ) const ;
    ///
    const PMVertex &VertexFromId( int in_iId ) const { return m_aoVerts[ in_iId ]; }
    ///
          PMVertex &VertexFromId( int in_iId )       { return m_aoVerts[ in_iId ]; }
    ///
    const PMEdge   &EdgeFromId  ( int in_iId ) const { return m_aoEdges[ in_iId ]; }
    ///
          PMEdge   &EdgeFromId  ( int in_iId )       { return m_aoEdges[ in_iId ]; }
    ///
    const PMFace   &FaceFromId  ( int in_iId ) const { return m_aoFaces[ in_iId ]; }
    ///
          PMFace   &FaceFromId  ( int in_iId )       { return m_aoFaces[ in_iId ]; }
    //@}

    /**@name Adjacency information */
    //@{
    /**@name Vertex */
    //@{
    /// Number of adjacent edges
    int                              Degree             ( const PMVertex &in_pmV ) const ;
    ///
    inline Array<const PMVertex *>   CVertexVertices    ( const PMVertex & ) const;
    ///
    inline Array<const PMEdge *>     CVertexEdges       ( const PMVertex & ) const;
    ///
    inline Array<const PMFace *>     CVertexFaces       ( const PMVertex & ) const;
    /// So they would lie flat if you laid them out. Requires nice vertices.
    inline Array<const PMVertex *>   CVertexVerticesOrdered( const PMVertex & ) const;
    /// So they would lie flat if you laid them out. Requires nice vertices.
    inline Array<const PMFace *>     CVertexFacesOrdered( const PMVertex & ) const;
  /** So they would lie flat if you laid them out (counter clockwise orientation).
      Requires nice vertices. */
    void CVertexEdgeFace( const PMVertex &,
                          Array<const PMEdge *> &,
                          Array<const PMFace *> & ) const;

    ///
    inline Array<PMVertex *>   VertexVertices     ( const PMVertex & ) ;
    ///
    inline Array<PMEdge *>     VertexEdges        ( const PMVertex & ) ;
    ///
    inline Array<PMFace *>     VertexFaces        ( const PMVertex & ) ;
    /// So they would lie flat if you laid them out.  Requires nice vertices.
    inline Array<PMVertex *>   VertexVerticesOrdered( const PMVertex & ) ;
    /// So they would lie flat if you laid them out. Requires nice vertices.
    inline Array<PMFace *>     VertexFacesOrdered ( const PMVertex & ) ;
    /// So they would lie flat if you laid them out (counter clockwise orientation).
    ///  Requires nice vertices.
    void VertexEdgeFace( const PMVertex &,
                         Array<PMEdge *> &,
                         Array<PMFace *> & ) ;
    ///
    inline PMEdge *            EdgeFromVertices   ( const PMVertex &, const PMVertex & ) ;
    ///
    inline const PMEdge *      CEdgeFromVertices  ( const PMVertex &, const PMVertex & ) const;
    //@}

    /**@name Edge */
    //@{
    ///
    inline const PMVertex &    CEdgeVertex1 ( const PMEdge &  ) const;
    ///
    inline const PMVertex &    CEdgeVertex2 ( const PMEdge &  ) const;
    ///
    inline const PMFace *      CEdgeFace1   ( const PMEdge &  ) const;
    ///
    inline const PMFace *      CEdgeFace2   ( const PMEdge &  ) const;

    ///
    inline PMVertex &          EdgeVertex1 ( const PMEdge &  ) ;
    ///
    inline PMVertex &          EdgeVertex2 ( const PMEdge &  ) ;
    ///
    inline PMFace *            EdgeFace1   ( const PMEdge &  ) ;
    ///
    inline PMFace *            EdgeFace2   ( const PMEdge &  ) ;
    ///
    int                        HardEdges   ( Array<PMEdge *> &, float in_dDotClip  ) ;
    //@}

    /**@name Face */
    //@{
    ///
    inline Array<const PMVertex *>   CFaceVertices    ( const PMFace & )   const;
    ///
    inline Array<const PMEdge *>     CFaceEdges       ( const PMFace & )   const;
    /// return v0-e0-v1 where e0 goes from v0 to v1
    void  CFaceVertexEdge ( const PMFace            &,
                            Array<const PMVertex *> &,
                            Array<const PMEdge *>   & )   const;

    ///
    inline Array<PMVertex *>   FaceVertices    ( const PMFace & )   ;
    ///
    inline Array<PMEdge *>     FaceEdges       ( const PMFace & )   ;
    /// return v0-e0-v1 where e0 goes from v0 to v1
    void  FaceVertexEdge ( const PMFace      &,
                           Array<PMVertex *> &,
                           Array<PMEdge *>   & )   ;
    int   NSides( const PMFace & ) const ;
    //@}
    //@}

    /**@name Boundary queries */
    //@{
    ///
    int     NumBoundaries   ( const PMVertex &in_pmV ) const;
    ///
    WINbool IsBoundary      ( const PMVertex &in_pmV ) const ;
    ///
    WINbool IsMultipleBoundary( const PMVertex &in_pmV ) const ;
    ///
    WINbool IsBoundary      ( const PMEdge &in_pmE )   const ;
    /// 
    void    FollowBoundary( const PMVertex &, Array<const PMVertex *> & ) const;
    //@}


    /**@name Renumber and compress.
       Returns mapping of ids of elements. Called automatically before writing mesh */
    //@{
    ///
    void Renumber( Array<int> &out_aiVMap, Array<int> &out_aiEMap, Array<int> &out_aiFMap );
    /// Merge vertices, if safe. 
    void CombineTriangles( PMMesh          & out_mesh,
                           Array<PMEdge *>   & out_aopDuplicate,
                           Array<int>        & out_aiFMap );

    //@}

    /**@name Read and write */
    //@{
    /// Cross-platform safe
    void WriteBinary( ofstream & ) ;
    /// Cross-platform safe
    void ReadBinary( ifstream & );
    ///
    void Write( ofstream & ) ;
    ///
    void Read( ifstream & );
    /// Just the underlying mesh representation (GMesh or BMesh)
    void WriteMeshOnly( ofstream & ) ;
    ///
    void ReadMeshOnly( ifstream & );
    ///
    void ReadInventor( ifstream & );
    /** Write the inventor format.
        @param io_strm Where to write to
        @param in_strHeader (optional) where to grab the header from */
    void WriteInventor( ofstream &io_strm,
                        const char *in_strHeader = NULL) const;
    //@}

    GMesh       &Mesh()        { return m_mesh; }
    const GMesh &CMesh() const { return m_mesh; }


    /** A set of contours along the given vector at the given spacing. */
    Array<R2Contour>  Contours( const R3Vec &in_vecPlaneNormal,
                                double       in_dStartZ,
                                double       in_dWidth ) const;
    //@}

    /**@name OpenGL stuff -- See utils/Ogl_Routines.H */
    //@{

    /// Call SetLights before calling draw. No texture, but colors if defined.
    void DrawOpenGL() ;
    /// Draws just the edges
    void DrawEdgesOpenGL() ;
    /// Call SetLights before calling draw. Uses mesh color.
    void DrawTransparent( const double in_dAlpha ) ;
    /// Call SetLights before calling draw. Texture map, ignore colors on vertices.
    void DrawTextureOpenGL() const;

    //@}

    /// Non-functional
    PMMesh & operator=( const PMMesh &in_oPMMesh );
    
    ///
    PMMesh() ;
    ///
    inline virtual ~PMMesh() ;
};

#include "Pm_Mesh_i.H"
#include "Pm_Mesh_GMesh_i.H"

#endif

