#ifndef _PMESHLITE_DEFS_H
#define _PMESHLITE_DEFS_H

#include <utils/Rn_Defs.H>
#include <utils/Utils_Color.H>
#include <boost/dynamic_bitset/dynamic_bitset.hpp>

#include <utils/Mesh_GMesh.H>
#include <utils/Mesh_Intersect.H>

/** \class PMeshLite Pm_MeshLite.H utils/Pm_MeshLite.H
  \ingroup Meshes
  \brief A mesh, wrapper around GMesh

    See OGLMeshes_Texture for making texture maps  <br>

    All adjacency information is found in this class (e.g., an edge's
    two vertices), not on the vertex, edge, and face classes. <br>

    This class uses the GMesh underlying data structures <br>


    If using GMesh, call Compute_normals() if the
    mesh has been changed to update the normals.
    

<br><br> Files:
- include/utils/Pm_MeshLite.H
- utils/PMeshLite/PMeshLite.cpp
- utils/PMeshLite/PMeshLite_IO.cpp
- utils/PMeshLite/PMeshLite_Intersect.cpp
- utils/PMeshLite/PMeshLite_Tools.cpp */
class PMeshLite : private GMesh
{
protected:
    WINbool            m_bHasChanged; 

    /// Color 
    UTILSColor         m_colDefault;

    /// Check that the face can be added to the boundary
    WINbool CheckVs( Face in_of, const Array<Vertex > &in_aopVs ) const;

    /// grow disk subroutine
    WINbool EatWedge( Array<Vertex> &io_avBdry, Array<Edge> &io_aeBdry, boost::dynamic_bitset<> &io_abInDisk, const boost::dynamic_bitset<> &in_abNoCross ) const;

public:
    /**@name Tools */
    //@{
    /// Find a path from vertex 1 to any vertex in the list
    /// No crossing of marked vertices
    /// Use edge weights
    Array<Vertex> FindPath( const Vertex &in_v1, const Array<Vertex> &in_aopBdryVs, const boost::dynamic_bitset<> &in_abEdgeNoCross, const Array< double > &in_adEdgeWeights ) const;
    /// Find a path from vertex 1 to any vertex in the list
    Array<Vertex> FindPath( const Vertex &in_v1, const Array<Vertex> &in_aopBdryVs ) const;
    /// Don't cross
    Array<Vertex> FindPath( const Vertex &in_v1, const Array<Vertex> &in_aopBdryVs, const boost::dynamic_bitset<> &in_abEdgeNoCross ) const;
    /// Use edge weights
    Array<Vertex> FindPath( const Vertex &in_v1, const Array<Vertex> &in_aopBdryVs, const Array< double > &in_adEdgeWeights ) const;
    /// No edge weights, all cross
    Array<Vertex> FindPath( const Vertex &in_v1, const Vertex &in_v2 ) const;
    /// Use edge weights
    Array<Vertex> FindPath( const Vertex &in_v1, const Vertex &in_v2, const boost::dynamic_bitset<> &in_abEdgeNoCross, const Array< double > &in_adEdgeWeights ) const;
    /// Use no cross
    Array<Vertex> FindPath( const Vertex &in_v1, const Vertex &in_v2, const boost::dynamic_bitset<> &in_abEdgeNoCross ) const;
    /// Use edge weights
    Array<Vertex> FindPath( const Vertex &in_v1, const Vertex &in_v2, const Array<double> &in_adEdgeWeights ) const;
    /// run dijkstra's
    void GrowShortestPathTree( const Vertex &in_v1, const boost::dynamic_bitset<> &in_abNoCross, const Array<double> &in_adEdgeWeights, Array<Vertex> &out_avFrom, Array<double> &out_adDist, boost::dynamic_bitset<> &out_abTreeEdges, const double in_dStopGrow = 1e30) const;
    ///
    void DualMST( const Face &in_f1, const boost::dynamic_bitset<> &in_abNoCross, const Array<double> &in_adEdgeWeights, boost::dynamic_bitset<> & out_abTreeEdges) const;
    /// Keep boundary
    WINbool GrowDisk( Array<Vertex> &io_avBdry, 
                      Array<Edge> &io_aeBdry, 
                      boost::dynamic_bitset<> &io_abInDisk, 
                      const boost::dynamic_bitset<> &in_abNoCross ) const;
    /// Add input face to disk and eat wedges
    int GrowDiskByFace(  const Face &in_face,
                                        boost::dynamic_bitset<> &io_abInDisk,
                                        Array<Vertex> & io_avBdry, 
                                        Array<Edge>   & io_aeBdry,
                                        Array<int>   & out_wedgeEaten) const;
    /// Grow a disk with input face as center until the disk has targeted number of faces
    bool GrowDiskByDijkstra(  const Face &in_face,
                          const int targetFsInDisk,
                          boost::dynamic_bitset<> &out_abFsInDisk,
                          Array<Vertex> & out_avBdry, 
                          Array<Edge>   & out_aeBdry ) const;

    WINbool AddFaceToDisk(  const Face &in_face,
                                  boost::dynamic_bitset<> &io_abFChart,
                                  Array<Vertex> & io_avBdry, 
                                  Array<Edge>   & io_aeBdry) const ;
    ///check if Vs and Es on the boundary of the disk-chart are in correct order
    void CheckBoundary( const Array<Vertex> &in_av, const Array<Edge> &in_ae) const;
    ///check if the chart will remain a disk after adding this face
    WINbool OkayToAdd( const Face &in_of, const Array<Vertex > &in_aopVs ) const;

    /// Find the shortest path between the two faces,
    /// Not crossing the marked edges
    WINbool FindPath( const Face in_oF1, const Face in_oF2,
                      const boost::dynamic_bitset<> &in_aiEdgeNoCross,
                      Array<Face >   &out_aopFs ) const;
    /// Get the connected components
    int GetComponents( Array< unsigned short > &out_ausLabels, Array<int> &out_aiLabels ) const;
    /// Find the biggest component in the mesh and return it
    void GetOneComponent( PMeshLite &out_mesh ) const;
                             
    pair <int, int> ExtremePoints( const boost::dynamic_bitset<> &in_abVs ) const ;

    //@}


    /**@name Coloring */
    //@{
    ///
    void Color( const UTILSColor &in_col );
    ///
    inline void SetDefaultColor( const UTILSColor &in_col ) { m_colDefault = in_col; }
    ///
    inline const UTILSColor & DefaultColor(  ) const { return m_colDefault; }
    ///
    inline const UTILSColor & GetColor( Vertex in_v ) const { return in_v->GetColor(); }
    ///
    inline void SetColor( Vertex in_v, const UTILSColor &in_col ) { in_v->SetColor( in_col ); }
    ///
    void ColorByDistanceToInput( PMeshLite in_mesh );
    //@}

    /**@name Updating normals */
    //@{
    /// If point locations have changed
    void Compute_normals( const Array<Vertex> &aopVs, const Array<Face> &aopFs );
    /// Do all faces
    void Compute_face_normals(  );
    /// Do all
    void Compute_normals(  );
    //@}

    /**@name Geometry properties */
    //@{
    /// Centroid and scale of entire mesh
    R3Pt         GetCenter( R3Pt &out_ptScale ) const;
    /// Point furthest from centroid
    Vertex       GetExtreme() const;

    /// Point in a face
    R3Pt         GetPointInFace(       Face            in_oFace, 
                                 const Array<double> & in_adPerc ) const;
    /// Point in a face
    R3Pt         GetFaceCentroid( Face in_oFace ) const;
    /// Area of face
    double       GetFaceArea( const Face in_oFace ) const;
    ///
    void         GetFacePolygon( Face in_oFace, R3Polygon & out_poly ) const { GMesh::polygon(in_oFace, out_poly); }

    /// Point along an edge
    R3Pt         GetPointOnEdge( Edge in_oEdge, double in_dPerc ) const;
    ///
    double       GetEdgeLength( const Edge in_oEdge ) const;
    /// Returns array size of edge
    const Array<double> EdgeLengths() const;
    /// Center the mesh at the given point with the given raduis
    void Center( const R3Pt &in_ptCenter = R3Pt(0,0,0), const double in_dRadius = 0.5 );
	/// Make anything that seems like it should align, do so
	virtual void Regularize( double in_dPerc = 0.1 );

   //migrated from Tim's code - duplicated with other methods in Tim's code
    int AlphaBeta( const Face &in_face, const Vertex &in_vertex,
                             double &out_theta, 
                             double &out_alpha, double &out_beta,
                             double &t, double &a, double &b ) const;

    int AlphaBeta( const Vertex &in_vertex, const Edge &in_edge,
                             double &out_theta1, double &out_theta2, 
                             double &out_alpha, double &out_beta,
                             double &out_delta, double &out_gamma ) const;

    double CotAlphaBeta( const Vertex &v1, const Vertex &v2 ) const;

    double CotGammaDelta( const Vertex &v1, const Vertex &v2 ) const;

    int FaceAngles(const Face &in_face, const Vertex &in_vertex,
                             int &out_ThetaId, double &out_Theta, double &out_t, 
                             int &out_AlphaId, double &out_Alpha, double &out_a, 
                             int &out_BetaId, double &out_Beta, double &out_b ) const;
    
    void FaceAngles( const Face &in_face, Array<double> &out_adAngles ) const;

    int CotAngles( const Face &in_face, const Vertex &in_vertex,
                             int &out_ThetaId, double &out_CotTheta, double &out_t, 
                             int &out_AlphaId, double &out_CotAlpha, double &out_a, 
                             int &out_BetaId, double &out_CotBeta, double &out_b ) const;

    //@}

    /**@name Changing the topology of the mesh */
    //@{
    ///
    inline Vertex AddVertex();
    ///
    inline Vertex AddVertex(const R3Pt &in_pt);
    ///
    inline Vertex AddVertex(const R3Pt &in_pt, const UTILSColor &);
    ///
    inline Vertex AddVertex(const R3Pt &in_pt, const R3Vec &in_vec, const UTILSColor &);
    /// Orient  all the faces the opposite way (turn it inside out)
    void           Flip();

    /// Make sure the incoming face is oriented correctly before adding
    WINbool IsOkToAdd( const Array<Vertex> &in_aopVs );
    /** Add face.
        @param aiVertId Index indices of vertices to add */
    Face AddFace( const Array<Vertex> &in_aopVs, const int id = -1 );
    ///
    void AddFacesFixOrientation( const Array< Array<Vertex> > &in_aavface );

    /// Also deletes any edges which no longer have any faces
    inline virtual void    RemoveFace          ( Face );
    /// Also removes adjacent edges and faces
    inline virtual void    RemoveVertex        ( Vertex  );
    /// Also removes adjacent faces
    inline virtual void    RemoveEdge          ( Edge );
    /// Also deletes any edges which no longer have any faces
    virtual void           RemoveFaces          ( Array<Face> & );
    /// Also removes adjacent edges and faces
    virtual void           RemoveVertices        ( Array<Vertex> & );
    /// Does not remove vertex
    void    RemoveAdjacentFaces ( Vertex  );
    /// Blow it all away
    virtual void    Clear();

    /// Transform the mesh by the given affine transformation
    virtual void Transform( const R4Matrix &in_oFrame ) ;

	/// Subdivide mesh N times, N > 0
	///		subdivided mesh created in meshOut
	void SubdivideLoop( const int in_iNTimes, 
                              PMeshLite &out_mesh,
                        const WINbool in_bExactPts = FALSE, const WINbool in_bExactNorms = FALSE ) const;

	/// Subdivide mesh N times, N > 0
	///		subdivided mesh created in meshOut
	void SubdivideCC( const int in_iNTimes, 
                            PMeshLite &out_mesh ) const;
    /// Assume this mesh has moved; adjust locations of next level of subdivision
    void AdjustLocationsCC(       PMeshLite     & out_mesh, 
                            const Array<Vertex> & in_avMe, 
                                  Array<Vertex> & out_avNextLevel ) const;
    /// The set of vertices in the next level subdiv that are influenced by the in_av
    void InfluencesCC( const PMeshLite     & out_mesh, 
                       const Array<Vertex> & in_avMe, 
                             Array<Vertex> & out_avNextLevel ) const;
    /// The set of vertices in the next level subdiv that are influenced by the vertex
    void InfluenceCC( const PMeshLite       & in_meshChild, 
                      const Vertex            in_vChild, 
                            Array<Vertex>   & out_avParent, 
                            Array<double>   & out_adInfluence ) const;
    /// Which faces were derived from thisone
    void SubFacesCC( const PMeshLite   & in_meshNextLevel, 
                     const Array<Face> & in_afMe, 
                           Array<Face> & out_afNextLevel ) const;

    /// Make sure we can swap the edge
    WINbool IsOkToSwap( const Edge in_e ) const { return legalEdgeSwap( in_e ) ? TRUE : FALSE; }
    /// Make sure we can collapse the edge
    WINbool IsOkToCollapse( const Edge in_e ) const { return niceEdgeCollapse( in_e ) ? TRUE : FALSE; }

    ///
    void EdgeSwap( const Edge in_e ) { swapEdge(in_e); }
	///
	int SwapEdges( const Array<double> in_adDist );
	///
	void EdgeSplit( const Edge in_e ) { splitEdge(in_e); }
    ///
    void EdgeCollapse( const Edge in_e ) { collapseEdge(in_e); }
    ///
    void RemoveDupVertices( double in_dPerc = 0.01);
    //@}


    /**@name Elements of the entire mesh */
    //@{
    ///
    inline int NumVertices()  const;
    ///
    inline int NumEdges()     const;
    ///
    inline int NumFaces()     const;
    ///
    inline int AllMeshVertices ( Array<Vertex > & ) const ;
    ///
    inline int AllMeshEdges    ( Array<Edge >   & ) const ;
    ///
    inline int AllMeshFaces    ( Array<Face >   & ) const ;
    //@}

    /**@name Grow regions of elements.
      Adds to existing list, does not delete. */
    //@{
    ///
    int Grow( const Array<Vertex > &in_aopVs, 
                    Array<Edge   > &out_aopEs,
                    Array<Face   > &out_aopFs ) const;
    ///
    int Grow( const Array<Face   > &in_aopFs, 
                    Array<Vertex > &out_aopVs,
                    Array<Edge   > &out_aopEs ) const;
    ///
    int Grow( const Array<Edge   > &in_aopEs, 
                    Array<Vertex > &out_aopVs,
                    Array<Face   > &out_aopFs ) const;
    ///
    int GrowRings( const Vertex &in_vertex, 
                   const int nRings,
                   Array<Vertex> &out_aopVs,
				   Array<Face>   &out_aopFs ) const;
    //Copy a part of the mesh and make a new mesh out of it
    //might be the same as ExtractCopy
    void  MakeSubMesh(  const boost::dynamic_bitset<> &in_abFChart,
					    PMeshLite &newMesh ) const;
    ///
    int ExtractCopy( const Array<Face> &in_aopFs,
					 PMeshLite &newMesh ) const;
    ///
    int ExtractCopy( const Array<Face> &in_aopFs,
					 PMeshLite &newMesh, Array<Vertex> &out_aopVs ) const;
    //@}

    /**@name Element from IndexId()
      VertexFromIndexId( IndexId ) is an identity function. */
    //@{
    ///
    int MaxVertexId() const { return vertexnum; }
    ///
    int MaxEdgeId() const { return edgenum; }
    ///
    int MaxFaceId() const { return facenum; }
    ///
    WINbool Valid( Vertex in_v ) const { return valid(in_v) ? TRUE : FALSE; }
    ///
    WINbool Valid( Edge in_e ) const { return valid(in_e) ? TRUE : FALSE; }
    ///
    WINbool Valid( Face in_f ) const { return valid(in_f) ? TRUE : FALSE; }
    ///
    Vertex VertexFromId( int in_iId ) const { return idvertex( in_iId ); }
    ///
    Vertex VertexFromIndexId( int in_iId ) const { return idvertex( in_iId + 1 ); }
    ///
    Edge   EdgeFromId  ( int in_iId ) const { return idedge( in_iId ); }
    ///
    Edge   EdgeFromIndexId  ( int in_iId ) const { return idedge( in_iId + 1 ); }
    ///
    Face   FaceFromId  ( int in_iId ) const { return idface( in_iId ); }
    ///
    Face   FaceFromIndexId  ( int in_iId ) const { return idface( in_iId + 1 ); }
    ///
    Face   FaceFromVertices( const Vertex in_v1, const Vertex in_v2, const Vertex in_v3 ) const;
    ///
    Face   FaceFromVertices( const Vertex in_v1, const Vertex in_v2 ) const;
    //@}

    /**@name Adjacency information */
    //@{
    /**@name Vertex */
    //@{
    /// Number of adjacent edges
    int                    Degree            ( Vertex in_v ) const { return degree(in_v); }
    /// The vertices are NOT sorted. See VertexVerticesOrdered.
    inline Array< Vertex > VertexVertices    ( Vertex in_v ) const;
    /// The Edges are NOT sorted. See VertexEdgeFace.
    inline Array< Edge >   VertexEdges       ( Vertex in_v ) const;
    /// 
    inline Array< Face >   VertexFaces       ( Vertex in_v ) const;
    /// So they would lie flat if you laid them out. Requires nice vertices.
    inline Array< Vertex > VertexVerticesOrdered( Vertex in_v ) const;
    /// So they would lie flat if you laid them out. Requires nice vertices.
    /// Face i lies between edge i and edge i+1
    inline Array< Face >   VertexFacesOrdered( Vertex in_v ) const;
    /// So they would lie flat if you laid them out. Requires nice vertices.
    inline Array< Edge >   VertexEdgesOrdered( Vertex in_v ) const;
    /** So they would lie flat if you laid them out (counter clockwise orientation).
      Requires nice vertices. */
    void VertexEdgeFace(  Vertex in_v ,
                          Array< Edge > &,
                          Array< Face > & ) const;

    /// returns if there is an edge from v to w, or w to v
    inline Edge  EdgeFromVertices ( Vertex in_v,  Vertex in_w ) const { return queryEdge(in_v, in_w); }
    /// only checks for half edge v to w
    inline Edge  EdgeFromVerticesOrdered ( Vertex in_v,  Vertex in_w ) const { return queryEdgeOrdered(in_v, in_w); }
    ///
    Edge  EdgeFromFaces( Face in_f,  Face in_g ) const; 
    //@}

    /**@name Edge */
    //@{
    ///
    inline Vertex  EdgeVertex1 ( Edge in_e ) const { return vertex1( in_e ); }
    ///
    inline Vertex  EdgeOppositeVertex ( Edge in_e, Vertex in_v ) const { return oppVertex( in_v, in_e ); }
    ///
    inline Vertex  EdgeVertex2 ( Edge in_e ) const { return vertex2( in_e ); }
    /// Assumes triangular mesh; may return NULL if boundary
    inline Vertex  SideVertex1 ( Edge in_e ) const { return sideVertex1( in_e ); }
    /// Assumes triangular mesh; may return NULL if boundary
    inline Vertex  SideVertex2 ( Edge in_e ) const { return sideVertex2( in_e ); }
    ///
    inline Face    EdgeFace1   ( Edge in_e ) const { return face1( in_e ); }
    ///
    inline Face    EdgeFace2   ( Edge in_e ) const { return face2( in_e ); }
    ///
    inline Face    EdgeOtherFace   ( Edge in_e, Face in_f ) const;

    //@}

    /**@name Face */
    //@{
    /// Ordered
    inline Array< Vertex >  FaceVertices  ( Face )   const;
    /// Ordered 
    inline Array<Edge >     FaceEdges     ( Face )   const;
    /// return v0-e0-v1 where e0 goes from v0 to v1
    inline Array<Face >     FaceFaces     ( Face )   const;
    /// return f0-f1-f2 where e0 goes from v0 to v1 and f0 shares e0 with in_face
    void  FaceVertexEdge ( Face,
                           Array< Vertex > &,
                           Array< Edge >   & )   const;

    int   NSides( Face in_f ) const { return numVertices(in_f); }
    //@}
    //@}

    /**@name Boundary queries */
    //@{
    ///
    int     NumBoundaries   ( Vertex v ) const { return numBoundaries( v ); }
    ///
    WINbool IsBoundary      ( Vertex v ) const { return isBoundary( v ) ? TRUE : FALSE; }
    ///
    WINbool IsMultipleBoundary( Vertex v ) const { return isNice( v ) ? TRUE : FALSE; }
    ///
    WINbool IsBoundary      ( Edge e )   const { return isBoundary( e ) ? TRUE : FALSE; }
    /// 
    void    FollowBoundary( Vertex v, Array<Vertex > & ) const;
    //@}

    /**@name Silhouette queries */
    //@{
    ///
    WINbool IsSilho(const Edge & in_e, const R3Vec & in_lookVec) const;
    ///
    void SilhouetteEdgesVertices(const R3Vec &in_lookVec, Array< Array<Vertex> > &o_aVertices, Array< Array<Edge> > &o_aEdges, Array < Array<int> > &o_adDist ) const;
    //@}

    /**@name Filtering/smoothing */
    //@{
    ///
    void FilterLocations( Array<R3Pt> &out_apt, const double dPercMove ) const;
    ///
    void Filter( const double in_dPercMove );
    //@}

    /**@name Renumber and compress.
       Returns mapping of ids of elements. Called automatically before writing mesh */
    //@{
    ///
    void Renumber( );
    /// Merge vertices, if safe. 
    void CombineTriangles( PMeshLite         & out_mesh );
    ///
    void MakeTriangular();
    //@}

    /**@name Read and write */
    //@{
    ///
    void Write( const char *in_str ) const;
    ///
    void Read( const char *in_str, const WINbool in_bReverse = FALSE ) ;
    ///
    void Write( ofstream & ) const;
    ///
    void Read( ifstream &, const WINbool in_bReverse = FALSE );
    ///
    void ReadOFF( ifstream &, const WINbool in_bReverse = FALSE );
    /// A limited version of that file format (single mesh)
    void ReadObj( ifstream &, const WINbool in_bReverse = FALSE );
    ///
    void WriteBinary( ofstream & ) const;
    ///
    void ReadBinary( ifstream &, const WINbool in_bReverse = FALSE );
    ///
    void ReadInventor( ifstream & );
    /** Write the inventor format.
        @param io_strm Where to write to
        @param in_strHeader (optional) where to grab the header from */
    void WriteInventor( ofstream &io_strm,
                        const char *in_strHeader = NULL) const;
    /** Write the inventor format.
        @param in_strFName Where to write to
        @param in_strHeader (optional) where to grab the header from */
    void WriteInventor( const char *in_strFName,
                        const char *in_strHeader = NULL) const;
    //@}

    /// 
    const GMesh &Mesh() const { return *this; }
    /// 
    GMesh &Mesh()  { return *this; }

    /**@name Contours
    /** A set of contours along the given vector at the given spacing. */
    void  Contours( const R3Vec &in_vecPlaneNormal,
                    const double       in_dStartZ,
                    const double       in_dWidth,
                    Array<R3Pt>  &out_apt,
                    Array<R3Vec> &out_avec) const;
    //@}

    /**@name OpenGL stuff -- See utils/Ogl_Routines.H */
    //@{

    /// Call SetLights before calling draw. No texture, but colors if defined.
    void DrawOpenGL() ;
    /// Draws just the edges
    void DrawEdgesOpenGL() ;
    /// Call SetLights before calling draw. Uses color given.
    void DrawTransparent( const UTILSColor &, const double in_dAlpha, const WINbool in_bDoEdges ) ;
    /// Call SetLights before calling draw. Uses mesh color.
    void DrawTransparent( const float in_fAlpha ) ;

    /// 
    PMeshLite & operator=( const PMeshLite &in_oPMMesh );
    
    ///
    PMeshLite() ;
    ///
    PMeshLite( const PMeshLite &in_mesh ) { *this = in_mesh; }
    ///
    inline virtual ~PMeshLite() ;

    static void TestSubdivCC();
};

#include "Pm_MeshLite_i.H"

#endif

