#ifndef _SUBDIV_PICK_MESH_DEFS_H
#define _SUBDIV_PICK_MESH_DEFS_H

#include <utils/Pm_Mesh.H>
#include <utils/Pm_SubdivPt.H>
#include <utils/Mesh_Intersect.H>

/**@name A subdivision mesh.

    Produce the next n level of subdivision. In addition to creating
    the new vertices and faces, save the linear combination of the
    original vertices at each vertex for quick re-evaluation. <br>

    The mesh generates, keeps, and updates the n subdivision levels. 
    

<br><br> Files:
<dd><li> include/utils/Pm_SubdivMesh.H
<dd><li> src/utils/pickmesh/PMSubdivMesh.cpp */
//@{

class PMSubdivFace;
class PMSubdivEdge;
class PMSubdivVertex;
class PMSubdivMesh;


/** Vertex in subdivision mesh.

    Additional data is a subdivision point and a vertex in the
    submesh, if any. Also store the linear combination of PMMesh
    Vertices which makes up the subdivision point.

<br>Files:
<dd><li> include/utils/Pm_SubdivMesh.H
<dd><li> src/utils/pickMesh/PMSubdivMesh.cpp */
class PMSubdivVertex : public PMVertex, public PMSubdivPt {
protected:

    PMSubdivPt  m_oSelfSubdivPt;

public:
    /**@name Subdivision point */
    //@{
    /// Calculate the blend values
    void CalcBlend( const Array<const PMSubdivFace *> &, 
                    const Array<const PMSubdivEdge *> &,
                    const WINbool in_bDoLocal );
    //@}

    ///
    PMSubdivPt  &SetSelfSubdiv() { return m_oSelfSubdivPt; }
    ///
    PMSubdivPt  &SelfSubdiv() { return m_oSelfSubdivPt; }
    ///
    const PMSubdivPt  &SelfSubdiv() const { return m_oSelfSubdivPt; }

    ///
    PMSubdivVertex() ;
    ///
    virtual ~PMSubdivVertex() {}

    /**@name Read/write */
    //@{
    ///
    virtual void Write(ofstream &out) const ;
    ///
    virtual void Read(ifstream &in);
    ///
    virtual void ReadBinary(ifstream &in);
    ///
    virtual void WriteBinary(ofstream &out) const;
    //@}
};

/** Edge in subdivision mesh.

    Additional data is a subdivision point and a vertex in the
    constructed next-level mesh (if any). Also store the linear combination of
    PMMesh Vertices which makes up the subdivision point.

<br>Files:
<dd><li> include/utils/Pm_SubdivMesh.H
<dd><li> src/utils/pickMesh/PMSubdivMesh.cpp */
class PMSubdivEdge : public PMEdge, public PMSubdivPt {
protected:

public:
    /**@name Subdivision point */
    //@{
    /// Calculate the blend values
    void CalcBlend( const PMSubdivFace * in_opF1, 
                    const PMSubdivFace * in_opF2, 
                    const PMSubdivVertex & in_opV1, 
                    const PMSubdivVertex & in_opV2,
                    const WINbool in_bDoLocal );

    //@}

    ///
    PMSubdivEdge()  {}
    ///
    virtual ~PMSubdivEdge() {}

    /**@name Read/write */
    //@{
    ///
    virtual void Write(ofstream &out) const ;
    ///
    virtual void Read(ifstream &in);
    ///
    virtual void ReadBinary(ifstream &in);
    ///
    virtual void WriteBinary(ofstream &out) const;
    //@}
};

/** Face in mesh generator.

    Additional data is a subdivision point, a vertex in the constructed
    next-level mesh (if any), and one face in the constructed next-level mesh for each
    vertex in this face.  Also store the linear combination of
    PMMesh Vertices which makes up the subdivision point. If doing exact
    evaluation, keep the grid of points used to evaluate the patch (see Jos Stam's
    paper in Siggraph '98).

<br>Files:
<dd><li> include/utils/Pm_SubdivMesh.H
<dd><li> src/utils/pickMesh/PMSubdivMesh.cpp */
class PMSubdivFace : public PMFace, public PMSubdivPt {
protected:
    Array< PMSubdivFace * >  m_aopSubFaces;

    /// For exact evaluation
    Array< const PMSubdivVertex * > m_aopGrid;
    Array< R3Pt >                   m_aptGridEval;
    
    R2Pt MapUV( const R2Pt &in_pt, const PMSubdivMesh & in_oMesh ) const;

public:
    /**@name Subdivision point */
    //@{
    /// Calculate the blend values
    void CalcBlend(const Array<const PMSubdivVertex *> &,
                   const WINbool in_bDoLocal);
    //@}

    /**@name Exact evaluation data */
    //@{
    ///
    Array< const PMSubdivVertex * > &GridVs() { return m_aopGrid; }
    ///
    Array< R3Pt> & GridPts() { return m_aptGridEval; }
    ///
    WINbool ExactEvalValid() const { return (m_aopGrid.num() > 0) ? TRUE : FALSE; }
    ///
    R3Pt ExactEvalPt( const R2Pt &, const PMSubdivMesh & in_oMesh) const ;
    ///
    PMSubdivPt ExactEvalSubdiv( const R2Pt &, const PMSubdivMesh & in_oMesh) const ;
    //@}

    /**@name The constructed Faces in the next-level mesh */
    //@{
    /// Number of constructed faces (same as number of vertices)
    int            NFaces() const { return m_aopSubFaces.num(); }
    /// PMFace i
    const PMSubdivFace * SubFace(int in_i) const { return m_aopSubFaces[in_i]; }
    /// PMFace i
          PMSubdivFace * SubFace(int in_i)       { return m_aopSubFaces[in_i]; }
    /// Constructing mesh
    void CreateFaces( const PMSubdivMesh & in_mesh,
                            PMSubdivMesh & io_meshNextLevel );
    ///
    void ClearFaces() { m_aopSubFaces.need(0); }
    //@}

    ///
    PMSubdivFace() {  }
    ///
    virtual ~PMSubdivFace() {}

    /**@name Read/write */
    //@{
    ///
    virtual void Write(ofstream &out) const ;
    ///
    virtual void Read(ifstream &in);
    ///
    virtual void ReadBinary(ifstream &in);
    ///
    virtual void WriteBinary(ofstream &out) const;
    //@}
};

/** The Subdivision mesh constructor.

    Each vertex, edge, and face in this mesh produces a vertex in the
    next-level mesh. Each n-sided face in this mesh produces n faces in
    the next-level mesh. <br>


<br>Files:
<dd><li> include/utils/Pm_SubdivMesh.H
<dd><li> include/utils/Pm_SubdivMesh_i.H
<dd><li> src/utils/pickMesh/PMSubdivMeshOps.cpp
<dd><li> src/utils/pickMesh/PMSubdivMesh.cpp */
class PMSubdivMesh : public PMMesh {
protected:
    // This mesh's arrays with pointers to the derived classes
    // Mirror's PMMesh's arrays
    UTILSElementArray<PMSubdivVertex> m_aoSubVerts;
    UTILSElementArray<PMSubdivEdge>   m_aoSubEdges;
    UTILSElementArray<PMSubdivFace>   m_aoSubFaces;

    GMeshIntersect  m_meshIntersectData;

    // Top level mesh. Possibly self.
    PMMesh         *m_meshpTopLevel;
    // The next level down. Possibly null.
    PMSubdivMesh   *m_meshpNextLevel;

    // Whether or not to do exact evaluation
    WINbool         m_bDoExactEval;

    // Whether or not to do local or top evaluation
    WINbool         m_bDoLocal;

    // What needs to be updated (topology changed) or re-computed (geometry moved)
    UTILSBits          m_oUpdateV, m_oUpdateE, m_oUpdateF;
    UTILSBits          m_oMoveV, m_oMoveE, m_oMoveF;
    void               ExpandUpdate();
    void               ExpandMove();

    /**@name Update routines called by PMMesh when creating/deleting items */
    //@{
    ///
    void AfterNew( PMSubdivVertex & );
    ///
    void AfterNew( PMSubdivEdge & );
    ///
    void AfterNew( PMSubdivFace & );

    ///
    void BeforeKill( PMSubdivVertex & );
    ///
    void BeforeKill( PMSubdivEdge & );
    ///
    void BeforeKill( PMSubdivFace & );

    ///
    PMVertex &NewVertex(  );
    ///
    PMEdge   &NewEdge( );
    ///
    PMFace   &NewFace( );
    ///
    void      KillVertex(PMVertex *);
    ///
    void      KillEdge  (PMEdge   *);
    ///
    void      KillFace  (PMFace   *);
    ///
    void      ElementsRenumber( Array<int> &out_aiVMap, 
                                Array<int> &out_aiEMap, 
                                Array<int> &out_aiFMap );
    /// Make elements and read in any data
    virtual void ReadElements        ( ifstream &in ) ;
    ///
    virtual void ReadElementsBinary  ( ifstream &in ) ;
    //@}


    /**@name Update internal structures */
    //@{
    /// Make new vertices and faces
    void CalcVs();
    ///  Make new vertices and faces
    void CalcEs();
    ///  Make new vertices and faces
    void CalcFs();
    ///  Recalculate subdivision linear combinations
    void CalcBlends(  );
    ///  Recompute subdivision points using subdivision linear combinations
    void CalcPts(  );
    ///  Clear out arrays indicating items to be updated
    void Done();

    /// Which vertices to add
    void AddVertices( PMSubdivMesh &in_oMesh );
    /// Which faces to add
    void AddFaces( PMSubdivMesh &in_oMesh );
    //@}

    ///
    PMSubdivMesh( WINbool in_bDoAlloc );

    /** Fill in the array of control points/vertices as
       specified by Figure 3 in Jos Stam's '98 paper.
     Bails if two non-regular corners
       @returns degree of corner. */
    int FillArray( const PMFace &in_f ) ;

    /// Where the point would eval to
    R3Pt EvalPt ( const PMFace & in_f, const R2Pt &in_ptUV ) const;

    /// Where the point would eval to
    PMSubdivPt EvalSubdiv ( const PMFace & in_f, const R2Pt &in_ptUV ) const;
    
    /// Where the point would eval to
    PMFace *QuadPt ( const PMFace & in_f, const R2Pt &in_ptUV ) const;
    
    // Map array to point
    R2Pt FacePt ( const PMFace & in_f, const Array<double> &in_adPerc ) const;

    // Map face quad point to point in next level face
    const PMFace * MapQuad ( const PMFace & in_opF, 
                             const R2Pt   & in_pt,
                             R2Pt & out_ptMap ) const;

    // Map to the next level
    const PMFace * MapNext ( const PMFace & in_opF, 
                             const R2Pt   & in_pt,
                             R2Pt & out_ptMap ) const;
    
    /// Make meshes for each face in the bottom level
    void TestCorners( const char *in_strName, int in_iDepth ) const;
    void TestSubExact( const char *in_strName, int in_iDepth ) const;
    void TestSubEval( const char *in_strName, int in_iDepth ) const;

    ///
    virtual PMFace &AddFaceNoUpdate( const Array<const PMVertex *> &aiVertId, 
                                     PMFaceType in_oFT = PM_NONE,
                                     int in_iTexture = -1,
                                     int in_iElemId = -1 );

    virtual void RemoveFaceNoUpdate( PMFace &in_oF );
    virtual void RemoveVertexNoUpdate( PMVertex &in_oF );

public:
    /**@name Changing mesh.
       Use PMMesh routines to change topology. Call GenerateMesh() to make next-level
       mesh. */
    //@{
    ///
    virtual PMFace &AddFace( const Array<const PMVertex *> &aiVertId, 
                             PMFaceType in_oFT = PM_NONE,
                             int in_iTexture = -1,
                             int in_iElemId = -1 );
    /// If arrays are empty, use defaults
    virtual Array<PMFace *> AddFacesWithType( const Array< Array<const PMVertex *> > & , 
                                              const Array< PMFaceType >              & ,
                                              const Array< int >                     & ,
                                              const Array< int >                     &  );
    /// Use defaults
    virtual Array<PMFace *> AddFaces( const Array< Array<const PMVertex *> > & );
 
    /// Remove any vertices not attatched to a face ( called upon Write )
    virtual void RemoveDanglingVertices();
    /// Also deletes any edges which no longer have any faces
    virtual void RemoveFace ( PMFace & );
    /// Also removes adjacent edges and faces
    virtual void RemoveVertex        ( PMVertex & );
    virtual void RemoveFaces         ( Array<PMFace *> & );
    /// Also removes adjacent edges and faces
    virtual void RemoveVertices      ( Array<PMVertex *> & );

    /// Move a single vertex. Change will be propogated at next update self.
    virtual void MoveVertex  ( PMVertex &, const R3Vec &in_vec );
    /// Move a single vertex. Change will be propogated at next update self.
    virtual void ChangeVertex  ( PMVertex &, const R3Pt &in_pt );

    /// Transform the mesh by the given affine transformation
    virtual void Transform( const R4Matrix &in_oFrame ) ;
	/// Make anything that seems like it should align, do so
	virtual void Regularize( double in_dPerc = 0.1 );

    //@}

    /**@name Setting top-level mesh. */
    //@{
    /// 
    void     TopLevelRenumbered( const Array<int> &in_aiVMap );
    ///
    PMMesh *&TopLevelMesh() { return m_meshpTopLevel; }
    /** Change the location of the vertices so we're interpolating
      at the lowest level. Found in fit/FITMesh */
    static void Interpolate(PMSubdivMesh &in_oMesh);
    /** Approximate the given points. Found in fit/FITMesh */
    static void Fit(PMSubdivMesh &in_oMesh, const Array<R3Pt> &, const double in_dSmooth = 0.0);
    /** Approximate the given points. Found in fit/FITMesh. */
    static void FitPartial(PMSubdivMesh &in_oMesh, const int in_iLevel, 
                           const Array<R3Pt> &, const Array<R3Vec> &);
    /** Approximate the given points. Found in fit/FITMesh. */
    static void FitPartial(PMSubdivMesh &in_oMesh, const int in_iLevel, 
                           const Array<R3Pt> &, const Array<R3Pt> &, const Array<PMSubdivPt> &);
    //@}

    /**@name Changing next-level mesh. */
    //@{
    /** Generate mesh.
      @param in_iDepth levels (0 == no levels).
      @param in_bDoExact if TRUE, sets bottom level vertices to exact values
      @returns bottom level mesh
      */
    const PMSubdivMesh &GenerateSubdiv( int in_iDepth, WINbool in_bDoExact );
    /** Move the next-level mesh to new positions.
        Call after moving the vertices or changing topology. */
    virtual void UpdateSelf(  );
    ///
    void PropogateColor();
    ///
    void DoLocalPropogate();
    ///
    void DoGlobalPropogate();
    /// return mesh at depth, or bottom level if not that deep
    const PMSubdivMesh &Subdiv( int in_iDepth ) const;
    /// return mesh at depth, or bottom level if not that deep
     PMSubdivMesh &Subdiv( int in_iDepth ) ;
    /// Current depth of mesh
    int Depth() const;
    //@}

    /**@name Sub type access */
    //@{
    ///
    PMSubdivVertex &SubVertex( const PMVertex & );
    ///
    PMSubdivEdge   &SubEdge( const PMEdge & );
    ///
    PMSubdivFace   &SubFace( const PMFace & );
    ///
    const PMSubdivVertex &CSubVertex( const PMVertex & ) const;
    ///
    const PMSubdivEdge   &CSubEdge( const PMEdge & ) const;
    ///
    const PMSubdivFace   &CSubFace( const PMFace & ) const;

    ///
    const R3Pt       &Subdiv_pt( const PMVertex & ) const;
    ///
    const PMSubdivPt &Subdiv   ( const PMVertex & v, int in_iDepth = 0 ) const;
    ///
    const R3Pt       &Subdiv_pt( const PMEdge & ) const;
    ///
    const PMSubdivPt &Subdiv   ( const PMEdge &, int in_iDepth = 0 ) const;
    ///
    const R3Pt       &Subdiv_pt( const PMFace & ) const;
    ///
    const PMSubdivPt &Subdiv   ( const PMFace &, int in_iDepth = 0 ) const;
    //@}

    /**@name Linear combinations */
    //@{
    /** Linear combination of PMMesh vertices which determines subdiv vertex location.
        @param in_v Vertex in the current mesh
        */
    const PMSubdivPt &VertexBasis ( const PMVertex & in_v ) const;
    //@}

    /**@name Exact evaluation */
    //@{
    /// Where the point would eval to
    R3Pt ExactEvalPt ( const PMFace & in_f, const Array<double> &in_adPerc ) const;

    /// Where the point would eval to
    PMSubdivPt ExactEvalSubdiv ( const PMFace & in_f, const Array<double> &in_adPerc ) const;

    /** Where the point would eval to.
        The input point is an R2Pt in the unit polygon. **/
    PMSubdivPt ExactEvalSubdiv ( const PMFace & in_f, const R2Pt &in_ptUnitPolygon ) const;

    /** Where the point would eval to.
        The input point is an R2Pt in the unit polygon. **/
    R3Pt ExactEvalPt ( const PMFace & in_f, const R2Pt &in_ptUnitPolygon ) const;

    /// Where the point would eval to
    R3Pt ExactEvalPt ( const PMVertex & in_v ) const;

    /// Return the vertices influence by this one at that depth
    PMSubdivPt Influenced( const PMVertex &in_v, const int in_iDepth );
    /// Where the point would eval to
    PMSubdivPt ExactEvalSubdiv ( const PMVertex & in_v ) const;

    /// Project point and find closest
    R3Pt Closest( const R3Pt &in_pt, PMSubdivPt &out_ptSubdiv ) const;

    /// Project point and find closest
    R3Pt ClosestLocal( const R3Pt &in_pt, const PMFace &, PMSubdivPt &out_ptSubdiv ) ;

    /// Project point and find closest
    R3Pt ClosestLocal( const R3Pt &in_pt, PMSubdivPt &out_ptSubdiv ) ;

    /// Intersect ray 
    Array<R3Pt> Intersect( const R3Pt &in_pt, const R3Vec &, Array<PMSubdivPt> &out_aptSubdiv ) const;
    //@}

    /**@name Drawing */
    //@{
    ///
    void DrawEdgesOpenGL() { PMMesh::DrawEdgesOpenGL(); }
    ///
    void DrawOpenGL() { PMMesh::DrawOpenGL(); }
    ///
    void DrawTextureOpenGL() { PMMesh::DrawTextureOpenGL(); }
    ///
    void DrawTransparent( const double in_dAlpha ) { PMMesh::DrawTransparent( in_dAlpha ); }
    //@}

    ///
    PMSubdivMesh &operator=( const PMMesh &in_oMesh ) ;
    ///
    virtual void    Clear();
    ///
    PMSubdivMesh(  ) ;
    ///
    ~PMSubdivMesh() ;

    /// Make meshes for each face in the bottom level
    void Test( const char *in_strName ) const;
};

//@}

#include "Pm_SubdivMesh_i.H"
#endif

