#ifndef _PICK_MESH_TEXTUREMAP_DEFS_H
#define _PICK_MESH_TEXTUREMAP_DEFS_H

#include <utils/Rn_Defs.H>
#include <utils/Pm_UVCoords.H>
#include <utils/Utils_Color.H>

/**@name Convert from UTILSColor to PMPixel */
//@{
///
typedef R3VectorTC<unsigned char> PMPixel;

/// Clips UTILScolor before assignment.
inline PMPixel Convert( UTILSColor in_pix ) 
{
    in_pix.Clip();

    return PMPixel( (unsigned char) (in_pix[0] * 255.0),
                    (unsigned char) (in_pix[1] * 255.0),
                    (unsigned char) (in_pix[2] * 255.0) );
}

///
inline UTILSColor Convert( PMPixel in_pix ) 
{
    return UTILSColor( in_pix[0] / 255.0,
                       in_pix[1] / 255.0,
                       in_pix[2] / 255.0 );
}
//@}

/** A texture map.

    The texture map can be divided up into rectangles of equal
    sizes. This is used in the manifold code to allocate space for
    some number of vertex, edge, or face charts. To get this to behave
    like a normal texture map, call SetNumber(1,1) and use in_iIndex = 2. <br>

    Each rectangle can be assigned a unique id.

<br> Files:
<dd><li> include/utils/Pm_TextureMap.H
<dd><li> src/utils/pickmesh/PMTextureMap.cpp
    */
class PMTextureMap {
protected:
    int                  m_iId;
    Array<unsigned char> m_acImage;

    int m_iHeight, m_iWidth;
    int m_iNAcross, m_iNDown;
    int m_iNPixHeight, m_iNPixWidth;
    
    Array<int> m_aoElementIds;
    Array<int> m_aiXStart, m_aiYStart;

    int PixStart( int in_iX, int in_iY ) const { return 3 * ( in_iY * m_iWidth + in_iX ); }

public:
    /**@name Dividing up the texture map into regions */
    //@{
    /// Have we used up all the rectangles?
    WINbool Full() const;
    /// Have we used any of the rectangles? If not, safe to delete. 
    WINbool Empty() const;
    /// Grab the next available rectangle of in_iElemId.
    int     SetNextEmpty( const int in_iElemId );
    /// Set this guy
    WINbool SetId( const int in_iGridSq, const int in_iElemId );
    /// Release hold of rectangle, setting it's id to unclaimed.
    void    Remove( const int in_iElemId );
    /// Release all the rectangles.
    void ClearIndices();
    //@}

    /**@name Size of entire texture map */
    //@{
    ///
    int  Height()        const { return m_iHeight; }
    ///
    int  Width()         const { return m_iWidth; }
    ///
    void SetSize( const int in_iWidth, const int in_iHeight );
    /// Id of this texture map
    int  Id()                  const { return m_iId; }
    //@}

    /**@name Integer coords to uv coords*/
    //@{
    ///
    R2Pt ToIndex( const PMUVCoord &in_uv ) const { return R2Pt( in_uv[0] * m_iWidth, in_uv[1] * m_iHeight ); }
    //@}

    /**@name Rectangle information */
    //@{
    /// Size of rectangles
    int  NPixelsWidth()  const { return m_iNPixWidth; }
    ///
    int  NPixelsHeight() const { return m_iNPixHeight; }
    /// Number of rectangles across
    int  NAcross()       const { return m_iNAcross; }
    /// Number of rectangles down
    int  NDown()         const { return m_iNDown; }
    /// Total number of rectangles
    int  NTotal()        const { return m_iNAcross * m_iNDown; }
    /// Id stored in rectangle 
    int  ElementId( const int in_i ) const { return m_aoElementIds[ in_i ]; }
    /** Set number of rectangles across and down.
        Size of rectangle is the biggest possible that will fit. Removes all rectangle owners.*/
    void SetNumber( const int in_iNAcross, const int in_iNDown );
    /// Make as many rectangles of the given size as possible
    void SetNPixels( const int in_iNPixAcross, const int in_iPixDown );
    //@}

    /**@name Setting pixels in a rectangle. 
       @param in_iIndex is the rectangle id */
    //@{
    /** Get a (u,v) coordinate for the point in the rectangle.
        @param dXRatio 0 < dXRatio < 1
        @param dYRatio 0 < dYRatio < 1 */
    PMUVCoord Map   ( const int in_iIndex, const double dXRatio, const double dYRatio ) const;
    /** Get a (NPixelsWidth,NPixelsHeight) coordinate for the point in the rectangle.
        @param dXRatio 0 < dXRatio < NPixelsWidth
        @param dYRatio 0 < dYRatio < NPixelsHeight */
    PMUVCoord Map   ( const int in_iIndex, const R2Pt &in_pt ) const;
    /** Get a (u,v) coordinate for the point in the rectangle.
        @param dXRatio 0 < dXRatio < 1
        @param dYRatio 0 < dYRatio < 1 */
    PMUVCoord MapInt( const int in_iIndex, const double dXLoc, const double dYLoc ) const;
    /** Get an int x,y location in widthXheight.
        @param in_iIndex square
        @param in_ipt x,y in square */
    R2Pt_i MapInt( const int in_iIndex, const R2Pt_i &in_ipt ) const;
    /** Get which square the (x,y) in (width,height) belongs to
        @param in_iX x in 0, width
        @param in_iY y in 0, height */
    int MapToSquare( R2Pt_i &out_ipt, const int in_iX, const int in_iY ) const;
    /** Set the pixel percentage across rectangle.
        @param dXRatio 0 < dXRatio < 1
        @param dYRatio 0 < dYRatio < 1 */
    void      SetPixel( const int in_iIndex, const double dXRatio, const double dYRatio,
                        const PMPixel &);
    /** Set the pixel percentage across rectangle.
        @param dXRatio 0 < dXRatio < 1
        @param dYRatio 0 < dYRatio < 1 */
    void      SetPixelInt( const int in_iX, const int in_iY, const PMPixel &);
    /** Set the pixel percentage across rectangle.
        @param dXRatio 0 < NPixelsWidth < 1
        @param dYRatio 0 < NPixelsHeight < 1 */
    void      SetPixel( const int in_iIndex, const R2Pt &in_pt,
                        const PMPixel &);
    /** Set the pixel (x,y).
        @param in_iX 0 < in_iX < NPixelsWidth()
        @param in_iY 0 < in_iY < NPixelsHeight() */
    void      SetPixelInt( const int in_iIndex, const int in_iX, const int in_iY,
                           const PMPixel &);
    /// Set the rectangle to be the given color
    void SetSolid( const int in_iIndx, const PMPixel &oPixCol );
    //@}

    /**@name Setting pixels in entire texture map */
    //@{
    /// Set the pixel at the given coordinate to the given pixel.
    void      SetPixel( const PMUVCoord &, const PMPixel & );
    /// Set all the rectangles to be different colors
    void SetSolid();
    //@}

    /**@name Getting pixels */
    //@{
    ///
    unsigned char Red  ( const int in_iX, const int in_iY ) const;
    ///
    unsigned char Green( const int in_iX, const int in_iY ) const;
    ///
    unsigned char Blue ( const int in_iX, const int in_iY ) const;
    /** Get the pixel (x,y).
        @param in_iX 0 < in_iX < NPixelsWidth()
        @param in_iY 0 < in_iY < NPixelsHeight() */
    PMPixel GetPixelInt( const int in_iIndex, const int in_iX, const int in_iY) const;
    ///
    PMPixel GetPixel( const double in_dX, const double in_dY) const;
    ///
    PMPixel GetPixelInt( const int in_iX, const int in_iY) const;
    //@}

    /// Call to make this texture be the active one
    void OpenGLSet() const ;

    /// In general, make a new texture map by calling PMMesh::AddTextureMap
    PMTextureMap( const int in_iId );
    /// 
    virtual ~PMTextureMap() {}

    /**@name Read/write */
    //@{
    /// Cross-platform
    void Read(ifstream &);
    /// Cross-platform
    void Write(ofstream &) const;
    /// Cross-platform
    void ReadBinary(ifstream &);
    /// Cross-platform
    void WriteBinary(ofstream &) const;
    /// Write out a .ppm file
    void WriteImage( char *in_strFname) const;
    //@}
};


#endif
