#ifndef _PMUVCOORDS_DEFS_H
#define _PMUVCOORDS_DEFS_H

#include <utils/Rn_Defs.H>
#include <utils/Utils_Io.H>

/**@name Texture map coordinates.

   Manifolds have several texture maps, one set for the vertex charts,
   one for the edges, and one for the faces. This means that a given
   vertex might have texture coordinates in up to three texture maps,
   one each of a vertex, edge, and face. PMFaceType is used by the
   texture mapping routine to figure out which of the UVCoordinates to
   grab from the vertices of the face. <br>

   To use texture maps without all that nonsense, don't set the
   PMFaceType of the faces (it defaults to PM_NONE) and use the
   PMUVCoords::SetUV without the PMFaceType.

<br><br> Files:
<dd><li> include/utils/Pm_UVCoords.H
<dd><li> src/utils/PickMesh/PMUVCoords.cpp
   */

//@{

/**@name PMFaceType */
//@{
typedef enum {
    ///
    PM_DELETED = 0,
    ///
    PM_NONE,
    ///
    PM_INTERIOR,
    ///
    PM_EDGE,
    ///
    PM_EDGE_LEFT,
    ///
    PM_EDGE_RIGHT,
    ///
    PM_CORNER,

    ///

    PM_MULTIPLE
} PMFaceType;
//@}

/// Format of a coordiate
typedef R2Pt PMUVCoord;

/** UV coordinates for a vertex. */
class PMUVCoords {
private:
   Array<PMUVCoord>   m_afCoords;
   Array<int>         m_aiElemId;

public:
    /**@name Access */

    //@{

    ///

    int              NCoords(  ) const { return m_afCoords.num(); }

    ///

    int              Elem( const int in_i ) const { return m_aiElemId[in_i]; }

    ///

    const PMUVCoord &Coord( const int in_i ) const { return m_afCoords[in_i]; }

    //@}



    /**@name Generic coordinates */
    //@{
    ///
    double            U(  ) const;
    ///
    double            V(  ) const;
    ///
    const PMUVCoord  &UV(  ) const;
    ///
    void              SetUV( const PMUVCoord &in_dUV );

    //@}

    /**@name Manifold elaborate coordinates */
    //@{
    ///
    double            U( const PMFaceType in_oType, const int in_oElemId ) const;
    ///
    double            V( const PMFaceType in_oType, const int in_oElemId ) const;
    ///
    const PMUVCoord  &UV( const PMFaceType in_oType, const int in_oElemId ) const;
    ///
    void              SetUV( const PMUVCoord &in_dUV, const PMFaceType in_oType, const int in_iElemId );
    ///

    void              Clear(); 

    //@}

    /**@name Read/write */
    //@{
    ///
    void              Read(ifstream &in);
    ///
    void              Write(ofstream &out) const;
    ///
    void              ReadBinary(ifstream &in);
    ///
    void              WriteBinary(ofstream &out) const;
    //@}

    ///
    PMUVCoords(  ) { }
    ///
    PMUVCoords( const PMUVCoords & );
    ///
    PMUVCoords &operator=( const PMUVCoords & );
};

//@}

inline     PMUVCoords &
PMUVCoords::operator=( const PMUVCoords &in_uv )
{
    m_afCoords = in_uv.m_afCoords;
    m_aiElemId = in_uv.m_aiElemId;

    return *this;
}

inline PMUVCoords::PMUVCoords( const PMUVCoords &in_uv )
{
    (*this) = in_uv;
}

inline const PMUVCoord &PMUVCoords::UV( PMFaceType in_oType, int in_iElemId )  const
{

    int i;
    switch (in_oType) {
        case PM_INTERIOR    : 
            ASSERT(m_aiElemId[0] == in_iElemId);
            return m_afCoords[0];
        case PM_EDGE        :
            if ( m_aiElemId[1] == in_iElemId )
                return m_afCoords[1];
            else if ( m_aiElemId[2] == in_iElemId )
                return m_afCoords[2];
            ASSERT(FALSE);
            break;
        case PM_EDGE_LEFT   : 
            ASSERT(m_aiElemId[1] == in_iElemId);
            return m_afCoords[1];
        case PM_EDGE_RIGHT  : 
            ASSERT(m_aiElemId[2] == in_iElemId);
            return m_afCoords[2];
        case PM_CORNER      : 
            ASSERT(m_aiElemId[3] == in_iElemId);
            return m_afCoords[3];

        case PM_MULTIPLE :

            for (i = 0; i < m_aiElemId.num(); i++)

                if ( m_aiElemId[i] == in_iElemId )

                    return m_afCoords[i];

            ASSERT(FALSE);

            break;
        case PM_NONE      : 
            ASSERT(m_aiElemId.num() == 1);
            ASSERT(m_aiElemId[0] == -1);
            ASSERT(m_afCoords.num() == 1);
            return m_afCoords[0];
    }
    ASSERT(FALSE);
    return m_afCoords[0];
}

inline const PMUVCoord &PMUVCoords::UV(  )  const
{
    ASSERT(m_aiElemId.num() == 1);
    ASSERT(m_aiElemId[0] == -1);
    ASSERT(m_afCoords.num() == 1);
    return m_afCoords[0];
}

inline double PMUVCoords::U( const PMFaceType in_oType, const int in_iElemId ) const
{
    return UV( in_oType, in_iElemId )[0];
}

inline double PMUVCoords::U(  ) const
{
    return m_afCoords[0][0];
}


inline double PMUVCoords::V( const PMFaceType in_oType, const int in_iElemId ) const
{
    return UV( in_oType, in_iElemId )[1];
}

inline double PMUVCoords::V(  ) const
{
    return m_afCoords[0][1];
}


inline void 
PMUVCoords::SetUV( const PMUVCoord &in_oCoord, const PMFaceType in_oType, const int in_iElemId ) 
{
    int iNeed = 0, iIndx = -1;
    switch (in_oType) {
    case PM_INTERIOR : iNeed = 1;
        break;
    case PM_EDGE_LEFT     : 
    case PM_EDGE          : iNeed = 2;
        break;
    case PM_EDGE_RIGHT     : iNeed = 3; 
        break;
    case PM_CORNER   : iNeed = 4; 
        break;

    case PM_MULTIPLE :

        iIndx = m_aiElemId.index( in_iElemId );

        if ( iIndx == -1 ) {

            m_afCoords += in_oCoord;

            m_aiElemId += in_iElemId;

        } else {

            m_afCoords[iIndx] = in_oCoord;

        }

        return;
    default : ASSERT(FALSE);
    }
    for (int i = m_afCoords.num(); i < iNeed; i++) {
        m_afCoords += in_oCoord;
        m_aiElemId += 0;
    }

    switch (in_oType) {
    case PM_INTERIOR : 
        m_afCoords[0] = in_oCoord;
        m_aiElemId[0] = in_iElemId;
        break;
    case PM_EDGE_LEFT     : 
    case PM_EDGE          : 
        m_afCoords[1] = in_oCoord;
        m_aiElemId[1] = in_iElemId;
        break;
    case PM_EDGE_RIGHT     : 
        m_afCoords[2] = in_oCoord;
        m_aiElemId[2] = in_iElemId;
        break;
    case PM_CORNER   : 
        m_afCoords[3] = in_oCoord;
        m_aiElemId[3] = in_iElemId;
        break;
    default : ASSERT(FALSE);
    }

}



inline void PMUVCoords::Clear()

{

    m_aiElemId.need(0);

    m_afCoords.need(0);

}

inline void 
PMUVCoords::SetUV( const PMUVCoord &in_oCoord ) 
{
    m_aiElemId.need(1);
    m_aiElemId[0] = -1;
    m_afCoords.need(1);
    m_afCoords[0] = in_oCoord;
}

inline void PMUVCoords::Read( ifstream &in ) 
{
    int iN = 0;
    in >> iN;
    m_afCoords.need( iN );
    m_aiElemId.need( iN );
    for (int i = 0; i < iN; i++)
        in >> m_afCoords[i][0] >> m_afCoords[i][1] >> m_aiElemId[i];
}

inline void PMUVCoords::Write( ofstream &out ) const 
{
    out << m_afCoords.num() << " ";
    for (int i = 0; i < m_afCoords.num(); i++)
        out << " " << m_afCoords[i][0] << " " << m_afCoords[i][1] << " "
            << m_aiElemId[i] << "\n";

}

inline void PMUVCoords::ReadBinary( ifstream &in ) 
{
    float fDummy = 0;
    UTILSIoReadBinaryPtVec( m_afCoords, in, fDummy );
    UTILSIo::ReadBinary( m_aiElemId, in );

    ASSERT( m_afCoords.num() == m_aiElemId.num() );
}

inline void PMUVCoords::WriteBinary( ofstream &out ) const 
{
    float fDummy = 0;
    UTILSIoWriteBinaryPtVec( m_afCoords, out, fDummy );
    UTILSIo::WriteBinary( m_aiElemId, out );
}

#endif

