#ifndef __PM_VERTEX_DEFS_H
#define __PM_VERTEX_DEFS_H

#include <utils/Utils_Color.H>

/** A Vertex.

    This class contains a Vertex pointer. <br>

    Note: To get adjacent edges, faces, and vertices, use the routines in PMMesh. <br>

    Data in this class: <br>

    A note on the normal and color data: These will be computed when
    PMMesh::compute_normals_and_colors is called (in the GMesh case,
    the color is written as a string on the vertex, and must be
    converted to an actual color). Normals are computed by taking the
    average of the adjacent face normals. The above function is called
    after a Read or before a Draw. If you've been mucking with the
    mesh, call after you're done to reset the normals. <br>

<dd><li> R3Pt the location of the vertex.
<dd><li> R3Vec_f the normal at the vertex (only valid if PMMesh::compute_normals_and_colors has been called)
<dd><li> R3Pt_f the color at the vertex (only valid if PMMesh::compute_normals_and_colors has been called). Defaults to a pinkish color.
<dd><li> PMUVCoords UV coordinates for texture mapping. Default is none.
<dd><li> VertId The id of the GMesh or Bvert vertex.
<dd><li> OwnerId User defined. For manifolds, the owning vertex chart.  

<br><br> Files:
<dd><li> include/utils/Pm_Vertex.H
<dd><li> utils/pickmesh/PMVertex.cpp
*/
class PMVertex : public UTILSElement {
protected:
    Vertex  m_vVert;
    GMesh  *m_opMesh;

    int            m_iVertId;
    int            m_iOwnerId;
    PMUVCoords     m_uvCoords;

    friend void          PMSetMesh  (       PMVertex &pmV, GMesh *in_opm ) { pmV.m_opMesh = in_opm; }
    friend const Vertex &PMGetVertex( const PMVertex &pmV ) { return pmV.m_vVert; }
    friend       Vertex &PMGetVertex(       PMVertex &pmV ) { return pmV.m_vVert; }
    friend void PMSetVertMap( PMMesh &in_pm, int iIndx, int iId );

public:
          void        SetIdFromVert( PMMesh &in_pm ) ;
          void        SetVertFromId( PMMesh &in_pm ) ;

    /**@name Accessors */
    //@{
    ///
          R3Pt         GetLoc()  const;
    ///
          void         SetLoc   ( const R3Pt & ) ;
    ///
          void         SetColor ( const UTILSColor & ) ;
    ///
          void         SetColor ( const double in_dR, const double in_dG, const double in_dB ) ;

    ///
    const UTILSColor & Color()    const { return m_vVert->GetColor(); }
    /// Normal of vertex. Not guaranteed to be valid.
    const R3Vec      & Norm()    const { return m_vVert->GetNorm(); }
    ///
          void         SetNorm(const R3Vec &in_v) { m_vVert->SetNorm(in_v); }
    ///
    const PMUVCoords & UV()      const { return m_uvCoords; }
    ///
          PMUVCoords & UV()            { return m_uvCoords; }
    ///
          int          OwnerId() const { return m_iOwnerId; }
    ///
          int        & OwnerId()       { return m_iOwnerId; }
    ///
          int          VertId()  const { return m_iVertId; }
    //@}

    /**@name Queries */
    //@{
    /** Does this vertex lie on the boundary?
        At least one adjacent edge has only one face */
    WINbool            IsBoundary() const ;
    ///
    WINbool operator==( const PMVertex & ) const;
    //@}

    /**@name Constructors, assignments. In general, don't call directly. */
    //@{
    /// I don't recomend calling this.
    PMVertex &operator=(const PMVertex &);
    /// Don't call. Make vertices using PMMesh::AddVertex
    PMVertex( const PMVertex & );
    /// Don't call. Make vertices using PMMesh::AddVertex
    PMVertex( );
    /// Don't delete vertices; use PMMesh::RemoveVertex
    virtual ~PMVertex( );
    //@}

    /**@name Read and write. 

       When deriving, overload in derived class, then call this
       Read/Write to make sure that PMVertex's data gets read/written
       ok. */
    //@{
    ///
    virtual void Write(ofstream &out) const ;
    ///
    virtual void Read(ifstream &in);
    ///
    virtual void ReadBinary(ifstream &in) ;
    ///
    virtual void WriteBinary(ofstream &out) const ;
    //@}
};

#endif

