#ifndef _CIRCLE_ANGLE_DEFS_H__
#define _CIRCLE_ANGLE_DEFS_H__

/** \class S1CircleAngleTC S1_CircleAngleTC.H utils/S1_CircleAngleTC.H
 \ingroup ImplicitGeometry
 \brief A circle manifold.

    This class is for adding/subtracting/sorting circle angles. It keeps the angle in the range [0,2pi),
    then checks if 2 pi should be added/subtracted from one of the two angles before performing
    on operation on them
<br> Files: 
 - include/utils/S1_CircleAngleTC.H <br>
*/
template <class T>
class S1CircleAngleTC {
private:
    T m_dAng;
public:
    /**@name Accessors and data */
    //@{
    /// The actual angle in the range [0, 2pi)
    double Angle() const { return m_dAng; }
    /// Given an angle, return the (clockwise) tangent at that point

    R2VectorTC<T> Tangent( const S1CircleAngleTC<T> &in_dAngDir ) const ;
    /// Given a circle center and a radius, return the point at m_dAng on that circle
    R2PointTC<T> Circle( const R2PointTC<T> &in_pt, const T &in_dRad ) const { return R2PointTC<T>( in_pt[0] + in_dRad * cos( m_dAng ), in_pt[1] + in_dRad * sin( m_dAng ) ); }
    /// Return the point on the unit circle given by the angle
    R2PointTC<T> Circle( ) const { return R2PointTC<T>( cos( m_dAng ), sin( m_dAng ) ); }
    //@}

    /**@name Comparisons */
    //@{
    /// Sort the input angle into the range [0, 2pi)
    static T Sort( const T in_dAng );
    /// Is in_ang Clockwise from this one?
    WINbool Clockwise( const S1CircleAngleTC<T> & in_ang ) const ;
    /** \brief Is this angle between the two input angles?
      The two input angles partition the circle into two pieces. Inside is considered to be inside the smaller of the two pieces
      @param in_ang1  Angle one
      @param in_ang2  Angle two
      @returns If the class's angle is between the other two
      */
    WINbool Inside( const S1CircleAngleTC<T> & in_ang1, const S1CircleAngleTC<T> & in_ang2 ) const ;
    /** \brief Is this angle between the two input angles?
      Returns percentage
      @param in_ang1  Angle one
      @param in_ang2  Angle two
      @returns If the class's angle is between the other two then [0,1], else outside
      */
    double InsidePerc( const S1CircleAngleTC<T> & in_ang1, const S1CircleAngleTC<T> & in_ang2 ) const ;
    /** \brief Is this angle between the two input angles? Checks clockwise
      Returns percentage
      @param in_ang1  Angle one
      @param in_ang2  Angle two
      @returns If the class's angle is between the other two then [0,1], else outside
      */
    double ClockwiseInsidePerc( const S1CircleAngleTC<T> & in_ang1, const S1CircleAngleTC<T> & in_ang2 ) const ;
    ///
    T Difference( const S1CircleAngleTC<T> & in_ang ) const ;
    /// Exact equality
    WINbool operator==( const S1CircleAngleTC<T> & in_ang ) const { return m_dAng == in_ang.m_dAng; }
    /// Exact inequality
    WINbool operator!=( const S1CircleAngleTC<T> & in_ang ) const { return m_dAng != in_ang.m_dAng; }
    //@}

    /// Initialize with angle 0
    S1CircleAngleTC( ) { m_dAng = 0.0; }
    /// Sort the input angle before storing it
    S1CircleAngleTC( const T in_dAng ) { m_dAng = Sort( in_dAng ); }
    /// Take atan2( in_vec.y, in_vec.x ) to get angle, then Sort
    S1CircleAngleTC( const R2VectorTC<T> &in_vec ) ;
    /// Take atan2( in_dY, in_dX) to get angle, then Sort
    S1CircleAngleTC( const T in_dX, const T in_dY ) ;
    ///
    ~S1CircleAngleTC() {}

    static void Test();
};

template <class T> inline
R2VectorTC<T> S1CircleAngleTC<T>::Tangent( const S1CircleAngleTC<T> &in_dAngDir ) const 
{ 
    if ( fabs( in_dAngDir.m_dAng - m_dAng ) > M_PI ) {
        if ( in_dAngDir.m_dAng > m_dAng ) 
            return R2VectorTC<T>( sin( m_dAng ), -cos( m_dAng ) );

        return R2VectorTC<T>( -sin( m_dAng ), cos( m_dAng) );
    }

    if ( m_dAng < in_dAngDir.m_dAng ) {
        return R2VectorTC<T>( -sin( m_dAng ), cos( m_dAng ) );
    }
    return R2VectorTC<T>( sin( m_dAng ), -cos( m_dAng ) );
}

template <class T> inline
T S1CircleAngleTC<T>::Sort( const T in_dAng )
{
    if ( in_dAng >= 0.0 && in_dAng < 2.0 * M_PI )
        return in_dAng;

    if ( in_dAng >= 2.0 * M_PI ) {
        const int iSubtract = (int) (in_dAng / ( 2.0 * M_PI ));
        return in_dAng - iSubtract * 2.0 * M_PI;
    }

    ASSERT( in_dAng < 0.0 );

    const int iAdd = (int) (-in_dAng / ( 2.0 * M_PI ));
    return in_dAng + (1 + iAdd) * 2.0 * M_PI;
}

template <class T> inline
WINbool S1CircleAngleTC<T>::Clockwise( const S1CircleAngleTC<T> & in_ang ) const 
{
    if ( fabs( in_ang.m_dAng - m_dAng ) > M_PI ) {
        if ( in_ang.m_dAng > m_dAng ) 
            return TRUE;
        return FALSE;
    }
    if ( in_ang.m_dAng > m_dAng ) 
        return FALSE;
    return TRUE;
}

template <class T> inline
WINbool S1CircleAngleTC<T>::Inside( const S1CircleAngleTC<T> & in_ang1, const S1CircleAngleTC<T> & in_ang2 ) const 
{
    if ( fabs( in_ang1.m_dAng - in_ang2.m_dAng ) > M_PI ) {
        if ( in_ang1.m_dAng > in_ang2.m_dAng ) {
            if ( m_dAng >= in_ang1.m_dAng || m_dAng <= in_ang2.m_dAng )
                return TRUE;
            else
                return FALSE;
        } else {
            if ( m_dAng >= in_ang2.m_dAng || m_dAng <= in_ang1.m_dAng )
                return TRUE;
            else
                return FALSE;
        }
    } 

    if ( in_ang1.m_dAng > in_ang2.m_dAng ) {
        if ( m_dAng >= in_ang2.m_dAng && m_dAng <= in_ang1.m_dAng )
            return TRUE;
        else 
            return FALSE;
    }
    if ( m_dAng >= in_ang1.m_dAng && m_dAng <= in_ang2.m_dAng )
        return TRUE;
    else 
        return FALSE;
}

template <class T> inline
double S1CircleAngleTC<T>::InsidePerc( const S1CircleAngleTC<T> & in_ang1, const S1CircleAngleTC<T> & in_ang2 ) const 
{
    if ( ApproxEqual( in_ang1, in_ang2 ) ) {
        if ( ApproxEqual( *this, in_ang1 ) ) {
            return 0.5;
        } else {
            return -1.0;
        }
    }
    if ( fabs( in_ang1.m_dAng - in_ang2.m_dAng ) > M_PI ) {
        if ( in_ang1.m_dAng > in_ang2.m_dAng ) {
            if ( m_dAng >= in_ang1.m_dAng )
                return (m_dAng - in_ang1.m_dAng) / ( in_ang2.m_dAng + 2.0 * M_PI - in_ang1.m_dAng );
            else if ( m_dAng <= in_ang2.m_dAng )
                return (m_dAng - (in_ang1.m_dAng - 2.0 * M_PI) ) / ( in_ang2.m_dAng + 2.0 * M_PI - in_ang1.m_dAng );
            else
                return -1.0;
        } else {
            if ( m_dAng >= in_ang2.m_dAng )
                return (m_dAng - in_ang2.m_dAng) / ( in_ang1.m_dAng + 2.0 * M_PI - in_ang2.m_dAng );
            else if ( m_dAng <= in_ang1.m_dAng )
                return (m_dAng - (in_ang2.m_dAng - 2.0 * M_PI) ) / ( in_ang1.m_dAng + 2.0 * M_PI - in_ang2.m_dAng );
            else
                return -1.0;
        }
    } 

    if ( in_ang1.m_dAng > in_ang2.m_dAng ) {
        if ( m_dAng >= in_ang2.m_dAng && m_dAng <= in_ang1.m_dAng )
            return (m_dAng - in_ang2.m_dAng) / ( in_ang1.m_dAng - in_ang2.m_dAng );
        else 
            return -1.0;
    }
    if ( m_dAng >= in_ang1.m_dAng && m_dAng <= in_ang2.m_dAng )
        return (m_dAng - in_ang1.m_dAng) / ( in_ang2.m_dAng - in_ang1.m_dAng );
    else 
        return -1.0;
}

template <class T> inline
double S1CircleAngleTC<T>::ClockwiseInsidePerc( const S1CircleAngleTC<T> & in_ang1, const S1CircleAngleTC<T> & in_ang2 ) const 
{
    if ( ApproxEqual( in_ang1, in_ang2 ) ) {
        if ( ApproxEqual( *this, in_ang1 ) ) {
            return 0.5;
        } else {
            return -1.0;
        }
    }
    if ( in_ang1.m_dAng > in_ang2.m_dAng ) {
        if ( m_dAng >= in_ang1.m_dAng )
            return (m_dAng - in_ang1.m_dAng) / ( in_ang2.m_dAng + 2.0 * M_PI - in_ang1.m_dAng );
        else if ( m_dAng <= in_ang2.m_dAng )
            return (m_dAng - (in_ang1.m_dAng - 2.0 * M_PI) ) / ( in_ang2.m_dAng + 2.0 * M_PI - in_ang1.m_dAng );
        else
            return -1.0;
    } else {
        if ( m_dAng >= in_ang2.m_dAng )
            return (m_dAng - in_ang2.m_dAng) / ( in_ang1.m_dAng + 2.0 * M_PI - in_ang2.m_dAng );
        else if ( m_dAng <= in_ang1.m_dAng )
            return (m_dAng - (in_ang2.m_dAng - 2.0 * M_PI) ) / ( in_ang1.m_dAng + 2.0 * M_PI - in_ang2.m_dAng );
        else
            return -1.0;
    }
    return -1.0;
}

template <class T> inline
T S1CircleAngleTC<T>::Difference( const S1CircleAngleTC<T> & in_ang ) const 
{ 
    const T dAng = fabs( m_dAng - in_ang.Angle() );
    if ( dAng < M_PI )
        return dAng;

    const T dAng2 = fabs( m_dAng + 2.0 * M_PI - in_ang.Angle() );
    if ( dAng2 < M_PI )
        return dAng2;

    return fabs( m_dAng - 2.0 * M_PI - in_ang.Angle() );
}

template <class T> inline
S1CircleAngleTC<T>::S1CircleAngleTC( const T in_dX, const T in_dY ) 
{ 
    ASSERT( !( RNIsZero( in_dX ) && RNIsZero( in_dY ) ) );
    m_dAng = Sort( atan2( in_dY, in_dX ) ); 
}

template <class T> inline
S1CircleAngleTC<T>::S1CircleAngleTC( const R2VectorTC<T> &in_vec ) 
{ 
    ASSERT( !( RNIsZero( in_vec[0] ) && RNIsZero( in_vec[1] ) ) );
    m_dAng = Sort( atan2( in_vec[1], in_vec[0] ) ); 
}


/** \brief linearly interpolate between two angles
 
   Sorts the angles so that second angle is after first and checks for 2pi boundaries
   @param in_ang1 First angle
   @param in_ang2 Second angle
   @param in_dPerc Percent along
   @returns Linearly interpolated angle.
   */
template <class T> inline
S1CircleAngleTC<T> Lerp( const S1CircleAngleTC<T> &in_ang1, const S1CircleAngleTC<T> &in_ang2, const T in_dPerc )
{
    if ( fabs( in_ang1.Angle() - in_ang2.Angle() ) > M_PI ) {
        if ( in_ang1.Angle() > in_ang2.Angle() ) 
            return S1CircleAngleTC<T>( (1.0 - in_dPerc) * in_ang1.Angle() + ( in_ang2.Angle() + 2.0 * M_PI ) * in_dPerc );

        return S1CircleAngleTC<T>( (1.0 - in_dPerc) * in_ang1.Angle() + ( in_ang2.Angle() - 2.0 * M_PI ) * in_dPerc );
    } 

    return S1CircleAngleTC<T>( (1.0 - in_dPerc) * in_ang1.Angle() + in_ang2.Angle() * in_dPerc );
}

/// Checks for 2pi wrap, angles within in_dEps of each other
template <class T> inline
WINbool ApproxEqual( const S1CircleAngleTC<T> &in_ang1, const S1CircleAngleTC<T> &in_ang2, const T in_dEps )
{
    if ( fabs( in_ang1.Angle() - in_ang2.Angle() ) > M_PI ) {
        if ( in_ang1.Angle() > in_ang2.Angle() ) 
            return RNApproxEqual( in_ang1.Angle(), in_ang2.Angle() + 2.0 * M_PI, in_dEps );

        return RNApproxEqual( in_ang1.Angle() + 2.0 * M_PI, in_ang2.Angle(), in_dEps );
    } 

    return RNApproxEqual( in_ang1.Angle(), in_ang2.Angle(), in_dEps );
}

/// Checks for 2pi wrap, angles within RNEpsilon_d of each other
template <class T> inline
WINbool ApproxEqual( const S1CircleAngleTC<T> &in_ang1, const S1CircleAngleTC<T> &in_ang2 )
{
    return ApproxEqual( in_ang1, in_ang2, RNEpsilon_d );
}

///
template<class T>
inline istream& operator>>( istream &in, S1CircleAngleTC<T> &out_dAng )
{
    T dAng;
    in >> dAng;
    out_dAng = S1CircleAngleTC<T>( dAng );
    return in;
}

///
template<class T>
inline ostream& operator<<( ostream &out, const S1CircleAngleTC<T> &in_dAng )
{
    out << in_dAng.Angle() << " ";
    return out;
}

typedef S1CircleAngleTC<double> S1CircleAngle;
#endif

