#ifndef SRF_DRAW_DEF
#define SRF_DRAW_DEF

#include <utils/Rn_Defs.H>
#include <utils/Rn_Polygon.H>
#include <utils/Mesh_GMesh.H>
#include <utils/Crv_Nurb_sampledTC.H>
#include <utils/Srf_NurbTC.H>

/** \class SRFDraw Srf_Draw.H utils/Srf_Draw.H
   \ingroup SplinesSrfs
    \brief Turn a trimmed spline surface into a mesh.

    Begin with gridding the spline surface. Keep dividing sub grids
    until the sub grids are within some epsilon fit of the
    surface. <br>

First level grid. <br>
    |-------|   <br>
    |   |   |   <br>
    |---+---|   <br>
    |   |   |   <br>
    |-------|   <br>

<br> Each quadrant is then gridded again. <br>

<br> After these grids are set, some grids may be gridded further if their neighbors have a much finer grid. <br> <br>

Optionally, 4 trim curves can be passed in to trim the 4 edges. The
initial gridding checks that the gridding where the trim curves is
sufficiently fine. After the gridding is done, the triangles along the
boundaries are ripped out and replaced with fans of triangles which
better approximate the edge. There's a bug in this code somewhere, but
it's mostly well-behaved. <br>

I don't think this compiles under the suns. <br>

Mesh is stored in m_mesh. The vertices are tagged with the u,v point
they came from and a unique number. <br>

Only works on $R^3$ surfaces, although this could be fixed by templating the class (I think). <br>

Files:
- include/utils/Srf_Draw.H
- utils/splines/SRFDraw.cpp 
- utils/splines/SRFDrawClip.cpp */
class SRFDraw {
private:
    class SRFuvs {
    public:
        double m_dU, m_dV;
        short m_iSrfName;
    };
    Array<SRFuvs> m_aSrfUvs;
    
    static short m_iSrfNum;
    short m_iSrfName;
    
    void Set_vert(GMesh &io_mesh, Vertex io_vert, const R2Pt &in_pt );
    Vertex New_vert( const R2Pt &in_pt);
    
    const SRFNub3TC &m_srf;
    CRVNub2TC m_crvTrim;
    
    GMesh m_mesh;
    
    void Write_mesh();
    
    double m_dEpsFit;
    // Pad the boundaries of the clipped mesh with faces
    void Fill_boundary(const Array<double> &in_adTs, 
                       const Array<Vertex> &in_avertBdry, 
                       const Array<double> &in_adCorners, 
                       const CRVNub_sampled2TC &in_aCrv);
    
    typedef Array<int> DepthRow;
    typedef Array <DepthRow> DepthGrid;
    typedef Array<bool> BoolRow;
    typedef Array <BoolRow> BoolGrid;
    typedef Array<Vertex> VertRow;
    typedef Array <VertRow> VertGrid;
    
    int m_iDoubled_u, m_iDoubled_v;
    VertGrid m_VertGrid;
    double Fit_edge(const R2Pt &in_pt1, const R2Pt &in_pt2);
    void Checker_fit_depth(DepthGrid &io_depthGrid,
                           double su, double eu,
                           double sv, double ev, int depth);
    bool Checker_adjust_depth(const DepthGrid &in_depthGrid, 
                              DepthGrid &io_depthGrid, 
                              int su, int eu, int sv, int ev, int depth);
    void Checker_sample(const DepthGrid &in_depthGrid,
                        BoolGrid &io_boolGrid,
                        int su, int eu, int sv, int ev, int depth);
    
    void Fill(Array<Vertex> &out_aVs);
    void Fill_square(int in_iSi, int in_iSj, int in_iSeperation);
    void Fill_faces(const BoolGrid &in_boolGrid);
    void Start_check_sample();
    
    void Make_draw_crv(CRVNub_sampled2TC &in_crvSampled, CRVNub3TC &out_crvDraw);
    void Fix_bad();
    int Fix_boundary(Array<Vertex> &bdry, 
                     const CRVNub_sampled2TC &in_crvSampled, bool bAll);
    void Sort_boundary(Array<double> &io_adTs,
                       Array<Vertex> &io_avertBdry, 
                       const CRVNub_sampled2TC &in_crvSampled);
    
    Array<double> m_adCorners;
    void Find_outside(Array<Face> &out_afaceOutside, 
                      const BoolGrid &in_boolGridOutMask);
    void Flood_fill(const BoolGrid &bdry_grid, BoolGrid &out_boolGridInsideOut);
    void Check_inside(const BoolGrid &bdry_grid, BoolGrid &g1, BoolGrid &g2);
    void Make_trim(const CRVNub2TC &e0, const CRVNub2TC &e1, 
                   const CRVNub2TC &e2, const CRVNub2TC &e3);
    void Make_corners();
   void Make_bdry_inside_grid(BoolGrid &out_boolGridInside, 
                              BoolGrid &out_boolGridBdry, 
                              const R2Polygon &in_polyBdry);
    void Clip(CRVNub3TC &out_crvDraw);
    
public:
    /// Vertices in the mesh will be tagged with this index
    static int m_iSrfIndex;
    
    /// The u,v coordinate of the vertex.
    double SrfU ( const Vertex in_v ) const;
    /// The u,v coordinate of the vertex.
    double SrfV ( const Vertex in_v ) const;
    /// The unique id of the vertex. Should be m_iSrfIndex.
    short  SrfID( const Vertex in_v ) const;
    
    /// The constructed mesh.
    const GMesh &Mesh() const { return m_mesh; }
    /// The surface the mesh was constructed from
    const SRFNub3TC &Srf() const { return m_srf; }
    
    /// 
    void Set_epsilon_fit(double in_dEps) { m_dEpsFit = in_dEps; }
    
    /// Construct with surface and no trim curves.
    SRFDraw(const SRFNub3TC &in_srf, double in_dEps);
    /// Construct with surface and the given trim curves
    SRFDraw(const SRFNub3TC &in_srf, double in_dEps,
            const CRVNub2TC &e0, const CRVNub2TC &e1, 
            const CRVNub2TC &e2, const CRVNub2TC &e3, CRVNub3TC &out_crvDraw);
};

inline double SRFDraw::SrfU(const Vertex in_v) const 
{
    return m_aSrfUvs[m_mesh.vertexid(in_v)].m_dU;
}

inline double SRFDraw::SrfV(const Vertex in_v) const 
{
    return m_aSrfUvs[m_mesh.vertexid(in_v)].m_dV;
}

inline short SRFDraw::SrfID(const Vertex in_v) const 
{
    return m_aSrfUvs[m_mesh.vertexid(in_v)].m_iSrfName;
}

#endif
