/* -------------------------------------------------------------------------
 *
 * A non uniform rational b-spline surface
 *
 * ------------------------------------------------------------------------- */

#ifndef SRFNurbTCs_DEFS
#define SRFNurbTCs_DEFS

#include <utils/Crv_Knot_vec.H>


/* -------------------  Constants  ------------------------------- */

/* ---------------------  Classes  ------------------------------- */

/** \class SRFNurbTC Srf_NurbTC.H utils/Srf_NurbTC.H
   \ingroup SplinesSrfs
  \brief A Tensor product spline surface.

    Sum_i_j B_i(s) B_j(t). Templated on input and output point
    type. To get a NURBS surface, the input point type should be a
    homogenous point (HPoint3TC). The weights are then the last
    coordinate of the homgenous point. The output point should be
    Point3TC; the division by the weight happens when the HPoint is
    cast to a Point. <br>

    SRFDraw supposedly has code to create a mesh from a patch; I really,
    really don't know what state that class is in

    I'm not entirely sure what state this class is in. I'm pretty sure
    all the functionality described here exists. <br>

    I don't think it will compile under the sun compiler version 4.2 <br>

    There are, of course, two knot vectors u and v. <br>

    Files: 
 - include/utils/Srf_NurbTC.H
 - include/utils/Srf_NurbTC.H
 - utils/splines/SRFNurbTC.cpp */
template<class InPoint, class OutPoint, class OutVector>
class SRFNurbTC  {
protected:
    CRVKnot_vec m_kvecU, m_kvecV;
    
    // The points to go with the knot vectors
    Array<InPoint> m_apts;      // Pts have dim+1 values (last value is weight if homogenous)
    
public:
    /**@name Accessors */
    //@{
    ///
    inline int InDim() const;
    ///
    inline int OutDim() const;
    
    /// The knot vectors
    const CRVKnot_vec &U_kvec() const { return m_kvecU; }
    ///
    const CRVKnot_vec &V_kvec() const { return m_kvecV; }
    /// Start and stop parameters of the surface
    double Start_u() const { return m_kvecU.Start_t(); }
    ///
    double Start_v() const { return m_kvecV.Start_t(); }
    ///
    double End_u() const { return m_kvecU.End_t(); }
    ///
    double End_v() const { return m_kvecV.End_t(); }
    /// 0 < in_dT < 1, find the parameter value in_dT along the knot vector
    double PercU(double in_dT) const { return m_kvecU.Perc_t(in_dT); }
    /// 0 < in_dT < 1, find the parameter value in_dT along the knot vector
    double PercV(double in_dT) const { return m_kvecV.Perc_t(in_dT); }

    /// NUpts * NVpts
    int    Num_pts() const { return m_apts.num(); }
    ///
    int NUpts() const { return m_kvecU.Num_pts(); }
    ///
    int NVpts() const { return m_kvecV.Num_pts(); }
    ///
    inline int PtIndx( int in_iU, int in_iV ) const;
    ///
    inline int PtXIndx( const int in_iPtIndx) const;
    ///
    inline int PtYIndx( const int in_iPtIndx) const;
    ///
    inline       InPoint &Pt(int in_iU, int in_iV);
    ///
    inline const InPoint &Pt(int in_iU, int in_iV) const;
    ///
    inline       InPoint &Pt(const int in_iIndx ) { return m_apts[in_iIndx]; }
    ///
    inline const InPoint &Pt(const int in_iIndx ) const { return m_apts[in_iIndx]; }
    ///
    inline const Array<InPoint> &Pts( ) const { return m_apts; }
    //@}

    /**@name Evaluate the curve */
    //@{
    /** Return the indices and values of the non-zero basis functions for $(u, v)$.
        The out_* arrays are all the same length.
        @param in_pt (u,v) point
        @param out_adValues B_u(in_dU) * B_v(in_dV)
        @param out_aiIndex
        */
    void Basis( const R2Pt &in_pt, 
                Array<double> & out_adValues, 
                Array<int>    & out_aiIndex ) const;
    /** Return the indices and values of the non-zero second derivative basis functions for $(u, v)$.
        The out_* arrays are all the same length.
        @param in_iWhich Which basis, given by index i,j
        @param in_pt (u,v) point
        @param out_adValues B_u(in_dU) * B_v(in_dV)
        @param out_aiIndx u * npts v + v
        */
    void BasisDerivative( const int in_iWhich, const R2Pt &in_pt, 
                          Array<double> & out_adValues, 
                          Array<int>    & out_aiIndx) const;
    /** Return the indices and values of the non-zero derivative basis functions for $(u, v)$.
        The out_* arrays are all the same length.
        @param in_iWhich Which basis, given by index i,j
        @param in_pt (u,v) point
        @param out_adValues B_u(in_dU) * B_v(in_dV)
        @param out_aiIndx u * npts v + v
        */
    void BasisSecondDerivative( const R2Pt_i &in_iWhich, const R2Pt &in_pt, 
                                Array<double> & out_adValues, 
                                Array<int>    & out_aiIndx) const;
    /** Return the indices and values of the non-zero derivative basis functions for $(u, v)$.
        The out_* arrays are all the same length.
        @param in_pt Input (u,v) point
        @param out_adDerivDu Derivative wrt u
        @param out_adDerivDv Derivative wrt v
        @param out_adDerivDuu Second derivative wrt u
        @param out_adDerivDvv Second derivative wrt u
        @param out_adDerivDuv Mixed derivative wrt u,v
        @param out_adValues B_u(in_dU) * B_v(in_dV)
        @param out_aiIndx u * npts v + v
        */
    void BasisAll( const R2Pt &in_pt,
                   Array<double> & out_adValues, 
                   Array<double> & out_adDerivDu,   
                   Array<double> & out_adDerivDv,  
                   Array<double> & out_adDerivDuu,  
                   Array<double> & out_adDerivDvv, 
                   Array<double> & out_adDerivDuv, 
                   Array<int>    & out_aiIndx) const;
    ///
    inline OutPoint operator()( const R2Pt &in_pt ) const ;
    /// Evaluate at a percentage along the surface. 0 < in_dU < 1 and 0 < in_dV < 1
    inline OutPoint Perc( const R2Pt &in_pt ) const ;
    ///
    inline OutVector Derivative( const int in_iWhich, const R2Pt & in_ptUV ) const;
    ///
    inline OutVector SecondDerivative( const R2Pt_i &in_iptWhich, const R2Pt & in_ptUV ) const;
    ///
    inline void AllEval( const R2Pt &in_pt,
                         OutPoint &out_pt,
                         OutVector &out_vecDs, OutVector &out_vecDt,
                         OutVector &out_vecDss, OutVector &out_vecDtt,
                         OutVector &out_vecDst ) const;
    ///
    inline void AllEval( const R2Pt &in_pt,
                         OutPoint &out_pt,
                         OutVector &out_vecDs, OutVector &out_vecDt ) const;
    ///
    inline OutVector Normal( const R2Pt &in_p ) const;
    //@}

    /**@name Operations */
    //@{
    /// Clip io_dU and io_dV to be between ( (Start_u, End_u), (Start_v, End_v) )
    inline void Clip_uv( R2Pt &io_uv ) const; 
    /** Find closest point on surface (very slow).
        in_iNDivs is the number of sample points per segment to try. */
    inline OutPoint Closest_pt( const InPoint &in_pt, 
                                R2Pt &out_pt,
                                const int in_iNDivs = 50) const ;
    //@}

    /**@name Constructors, assignments, comparators */
    //@{
    ///
    inline WINbool operator==(const SRFNurbTC &) const;
    ///
    inline SRFNurbTC &operator=(const SRFNurbTC &)  ;
    /// Uses default order in CRVknot_vector
    inline SRFNurbTC(const SRFNurbTC &);
    ///
    inline SRFNurbTC( const R2Pt_i &in_iOrder, const R2Pt_i &in_iNPts );
    ///
    inline SRFNurbTC();
    ///
    void SetKnots( const R2Pt &in_ptStart, const R2Pt &in_ptEnd, const R2Vec &in_vecSpacing );
    ///
    void SetKnots( const R2Pt &in_ptStart, const R2Pt &in_ptEnd );
    ///
    inline void SetKnots( const Array<double> &in_adUKnots, const Array<double> &in_adVKnots );
    ///
    inline SRFNurbTC(ifstream &in, WINbool bIs_binary = FALSE);
    ///
    virtual ~SRFNurbTC() {}
    //@}

    /**@name Read, write, print */
    //@{
    ///
    inline void    Print(const WINbool in_bPrintBasis = FALSE) const;
    ///
    inline WINbool Read(ifstream &in);
    ///
    inline void    Write(ofstream &out) const;
    /// Not cross-platform safe
    inline WINbool Read_binary(ifstream &in);
    /// Not cross-platform safe
    inline void    Write_binary(ofstream &out) const;
    //@}

    ///
    static WINbool Test();
};

#include "Srf_NurbTC_i.H"

/** \ingroup SplinesSrfs
 \brief 3D NUBS patch */
typedef SRFNurbTC<R3PointTC<double>, R3PointTC<double>, R3VectorTC<double> > SRFNub3TC;
/** \ingroup SplinesSrfs
 \brief 2D NUBS patch */
typedef SRFNurbTC<R2PointTC<double>, R2PointTC<double>, R2VectorTC<double> > SRFNub2TC;
/** \ingroup SplinesSrfs
 \brief 1D NUBS patch */
typedef SRFNurbTC<R1PointTC<double>, R1PointTC<double>, R1VectorTC<double> > SRFNub1TC;

/** \ingroup SplinesSrfs
 \brief 3D NURBS patch */
typedef SRFNurbTC<R3HPointTC<double>, R3PointTC<double>, R3VectorTC<double> > SRFNurb3TC;
/** \ingroup SplinesSrfs
 \brief 2D NURBS patch */
typedef SRFNurbTC<R2HPointTC<double>, R2PointTC<double>, R2VectorTC<double> > SRFNurb2TC;

/** \ingroup SplinesSrfs
 \brief Evaluate the curvature of a patch at a given point. 
  
 @param in_srf The surface
 @param in_pt The point to evaluate at
 @param out_vecP1 Principle curvature direction, in the domain
 @param out_vecP2 Principle curvature direction, in the domain
 @param out_vecNorm Surface normal
 @param out_vecS1 Principle curvature direction, on the surface
 @param out_vecS2 Principle curvature direction, on the surface
 @param out_dCrv1 Curvature 1
 @param out_dCrv2 Curvature 2
 @param out_dArea infitesmal area
 */
extern void SRFNurb3TC_CrvEval( const SRFNub3TC &in_srf,
                                const R2Pt  &in_pt,
                                R2Vec &out_vecP1, R2Vec &out_vecP2,
                                R3Vec &out_vecNorm,
                                R3Vec &out_vecS1, R3Vec &out_vecS2,
                                double &out_dCrv1, double &out_dCrv2,
                                double &out_dArea);
#endif

