template<class InPoint, class OutPoint, class OutVector>
inline
void SRFNurbTC<InPoint, OutPoint, OutVector>::SetKnots( const R2Pt &in_ptStart, const R2Pt &in_ptEnd, const R2Vec &in_vecSpacing )
{
    m_kvecU.Set_knots( in_ptStart[0], in_ptEnd[0], in_vecSpacing[0] );
    m_kvecV.Set_knots( in_ptStart[1], in_ptEnd[1], in_vecSpacing[1] );
}

template<class InPoint, class OutPoint, class OutVector>
inline
void SRFNurbTC<InPoint, OutPoint, OutVector>::SetKnots( const R2Pt &in_ptStart, const R2Pt &in_ptEnd )
{
    m_kvecU.Set_knots( in_ptStart[0], in_ptEnd[0] );
    m_kvecV.Set_knots( in_ptStart[1], in_ptEnd[1] );
}

template<class InPoint, class OutPoint, class OutVector>
inline void SRFNurbTC<InPoint, OutPoint, OutVector>::SetKnots( const Array<double> &in_adUKnots, const Array<double> &in_adVKnots )
{
    m_kvecU.Set_knots( in_adUKnots, -1, in_adUKnots.num() );
    m_kvecV.Set_knots( in_adVKnots, -1, in_adVKnots.num() );
}


template<class InPoint, class OutPoint, class OutVector>
inline
OutPoint SRFNurbTC<InPoint, OutPoint, OutVector>::Perc( const R2Pt &in_pt ) const 
{
   return (*this)( R2Pt( m_kvecU.Perc_t( in_pt[0] ), m_kvecV.Perc_t( in_pt[1] ) ) );
}

template<class InPoint, class OutPoint, class OutVector>
inline
int SRFNurbTC<InPoint, OutPoint, OutVector>::OutDim() const
{
   OutPoint outpt;
   return outpt.Dim();
}

template<class InPoint, class OutPoint, class OutVector>
inline
int SRFNurbTC<InPoint, OutPoint, OutVector>::InDim() const
{
   InPoint inpt;
   return inpt.Dim();
}

template<class InPoint, class OutPoint, class OutVector>
inline
int SRFNurbTC<InPoint, OutPoint, OutVector>::PtIndx(int in_iU, int in_iV) const 
{ 
   return in_iV * m_kvecU.Num_pts() + in_iU; 
}

template<class InPoint, class OutPoint, class OutVector>
inline int SRFNurbTC<InPoint, OutPoint, OutVector>::PtXIndx( const int in_iPtIndx) const
{
    return in_iPtIndx % m_kvecU.Num_pts();
}
///
template<class InPoint, class OutPoint, class OutVector>
inline int SRFNurbTC<InPoint, OutPoint, OutVector>::PtYIndx( const int in_iPtIndx) const
{
    return in_iPtIndx / m_kvecV.Num_pts();
}

template<class InPoint, class OutPoint, class OutVector>
inline InPoint &
SRFNurbTC<InPoint, OutPoint, OutVector>::Pt(int in_iU, int in_iV) 
{ 
   return m_apts[PtIndx(in_iU,in_iV)]; 
}

template<class InPoint, class OutPoint, class OutVector>
inline const InPoint &
SRFNurbTC<InPoint, OutPoint, OutVector>::Pt(int in_iU, int in_iV) const 
{ 
   return m_apts[PtIndx(in_iU,in_iV)]; 
}

template<class InPoint, class OutPoint, class OutVector>
inline
OutPoint
SRFNurbTC<InPoint, OutPoint, OutVector>::operator()(const R2Pt &in_pt)
const
{
   InPoint out_pt;
   
   RNZero( out_pt );

   static Array<double> adValues;
   static Array<int> aiIndex;

   Basis(in_pt, adValues, aiIndex);

   for (int i = 0; i < adValues.num(); i++) {
      const InPoint &pt = Pt(aiIndex[i]);
      for (int j = 0; j < InDim(); j++)
        out_pt[j] += adValues[i] * pt[j];
   }
   
   return out_pt;
}

template<class InPoint, class OutPoint, class OutVector>
inline
OutVector
SRFNurbTC<InPoint, OutPoint, OutVector>::Derivative(const int in_iWhich, const R2Pt &in_pt)
const
{
   OutVector out_vec;
   
   RNZero( out_vec );

   static Array<double> adValues;
   static Array<int> aiIndex;

   BasisDerivative(in_iWhich, in_pt, adValues, aiIndex);

   for (int i = 0; i < adValues.num(); i++) {
      const InPoint &pt = Pt(aiIndex[i]);
      for (int j = 0; j < InDim(); j++)
        out_vec[j] += adValues[i] * pt[j];
   }
   
   return out_vec;
}

template<class InPoint, class OutPoint, class OutVector>
inline
OutVector
SRFNurbTC<InPoint, OutPoint, OutVector>::SecondDerivative(const R2Pt_i &in_ipt, const R2Pt &in_pt)
const
{
   OutVector out_vec;
   
   RNZero( out_vec );

   static Array<double> adValues;
   static Array<int> aiIndex;

   BasisSecondDerivative(in_ipt, in_pt, adValues, aiIndex);

   for (int i = 0; i < adValues.num(); i++) {
      const InPoint &pt = Pt(aiIndex[i]);
      for (int j = 0; j < InDim(); j++)
        out_vec[j] += adValues[i] * pt[j];
   }
   
   return out_vec;
}


template<class InPoint, class OutPoint, class OutVector>
inline
OutVector
SRFNurbTC<InPoint, OutPoint, OutVector>::Normal( const R2Pt &in_pt )
const
{
    OutPoint ptOut;
    OutVector vecDu, vecDv;
    AllEval( in_pt, ptOut, vecDu, vecDv );

    return Cross( vecDu, vecDv );
}

/* ------------------------------------------------------------
 * DESCR   :	Copy the curve
 * ------------------------------------------------------------ */
template<class InPoint, class OutPoint, class OutVector>
inline
SRFNurbTC<InPoint, OutPoint, OutVector> &
SRFNurbTC<InPoint, OutPoint, OutVector>::operator=(const SRFNurbTC<InPoint, OutPoint, OutVector> &in_srf)
{
   m_kvecU = in_srf.m_kvecU;
   m_kvecV = in_srf.m_kvecV;

   m_apts = in_srf.m_apts;

   return *this;
}

/* ------------------------------------------------------------
 * DESCR   :	Compare the surface
 * ------------------------------------------------------------ */
template<class InPoint, class OutPoint, class OutVector>
inline
WINbool
SRFNurbTC<InPoint, OutPoint, OutVector>::operator==(const SRFNurbTC<InPoint, OutPoint, OutVector> &in_srf)
const
{
   if ((m_kvecU == in_srf.m_kvecU) == FALSE) return FALSE;
   if ((m_kvecV == in_srf.m_kvecV) == FALSE) return FALSE;

   for (int i = 0; i < m_apts.num(); i++) {
      if (ApproxEqual(m_apts[i], in_srf.m_apts[i]) == FALSE) return FALSE;
   }

   return TRUE;
}

/* ------------------------------------------------------------
 * DESCR   :	Make a copy of the given curve
 * ------------------------------------------------------------ */
template<class InPoint, class OutPoint, class OutVector>
inline
SRFNurbTC<InPoint, OutPoint, OutVector>::SRFNurbTC(const SRFNurbTC<InPoint, OutPoint, OutVector> &in_srf) 
: m_kvecU(in_srf.m_kvecU), m_kvecV(in_srf.m_kvecV)
{
   m_apts = in_srf.m_apts;
}

template<class InPoint, class OutPoint, class OutVector>
inline
SRFNurbTC<InPoint, OutPoint, OutVector>::SRFNurbTC( ) 
: m_kvecU(CRVSpace(1), 5), m_kvecV(CRVSpace(1), 5)
{
   m_apts.need(m_kvecU.Num_pts() * m_kvecV.Num_pts());
   memset( &m_apts[0], 0, sizeof( InPoint ) * m_apts.num() );
}

template<class InPoint, class OutPoint, class OutVector>
inline
SRFNurbTC<InPoint, OutPoint, OutVector>::SRFNurbTC(const R2Pt_i &in_iOrder, const R2Pt_i &in_iNPts)
: m_kvecU(CRVSpace(in_iOrder[0]), in_iNPts[0]), 
  m_kvecV(CRVSpace(in_iOrder[1]), in_iNPts[1])
{
    m_apts.need(m_kvecU.Num_pts() * m_kvecV.Num_pts());
    memset( &m_apts[0], 0, sizeof( InPoint ) * m_apts.num() );
}

template<class InPoint, class OutPoint, class OutVector>
inline
SRFNurbTC<InPoint, OutPoint, OutVector>::SRFNurbTC(ifstream &in, WINbool bIs_binary) 
: m_kvecU(in, bIs_binary), m_kvecV(in, bIs_binary)
{
   m_apts.need(m_kvecU.Num_pts() * m_kvecV.Num_pts());

   if (bIs_binary) {
      int iNData = (InDim()) * m_kvecU.Num_pts() * m_kvecV.Num_pts();
      double *adData = new double [iNData];
      in.read((char *) adData, sizeof(double) * iNData);
      ASSERT (in.gcount() == sizeof(double) * iNData);
      
      int indx = 0;
      for (int i = 0; i < m_apts.num(); i++) 
        for (int j = 0; j < InDim(); j++, indx++)
          m_apts[i][j] = adData[indx];
      
      delete [] adData;
   } else {
      for (int i = 0; i < m_apts.num(); i++) 
          ASSERT(m_apts[i].Read( in ));
   }
}

template<class InPoint, class OutPoint, class OutVector>
inline
void SRFNurbTC<InPoint, OutPoint, OutVector>::Clip_uv( R2Pt &in_pt ) const
{
   in_pt[0] = WINminmax(in_pt[0], U_kvec().Start_t(), U_kvec().End_t());
   in_pt[1] = WINminmax(in_pt[1], V_kvec().Start_t(), V_kvec().End_t());
}


/* ------------------------------------------------------------
 * DESCR   :	Evaluate a point
 * DETAILS :    Multiply basis functions by control points, 
 *              divide by sum of weights 
 * ------------------------------------------------------------ */
template<class InPoint, class OutPoint, class OutVector>
inline
void
SRFNurbTC<InPoint, OutPoint, OutVector>::Basis( const R2Pt &in_pt, 
                                                Array<double> &out_adValues,
                                                Array<int> &out_aiIndex )
const
{
   out_adValues.need(0);
   out_aiIndex.need(0);

   // Evaluate basis function/control point pairs

   int s_u, e_u, s_v, e_v;
   if ( m_kvecU.Span( in_pt[0], s_u, e_u ) == FALSE )
       return;

   if ( m_kvecV.Span( in_pt[1], s_v, e_v ) == FALSE )
       return;

   for (int ui = s_u; ui < e_u; ui++) {
      const double bu = m_kvecU.Basis(in_pt[0], ui);
      if ( bu > 0 ) {
         for (int vi = s_v; vi < e_v; vi++) {
            const double bv = m_kvecV.Basis(in_pt[1], vi);
            if ( bv > 0 ) {
               out_adValues += bu * bv;
               out_aiIndex += PtIndx( ui, vi );
            }
         }
      }
   }
}

/* ------------------------------------------------------------
 * DESCR   :	Evaluate a point
 * DETAILS :    Multiply basis functions by control points, 
 *              divide by sum of weights 
 * ------------------------------------------------------------ */
template<class InPoint, class OutPoint, class OutVector>
inline
void
SRFNurbTC<InPoint, OutPoint, OutVector>::BasisDerivative( const int in_iWhich, const R2Pt &in_pt,
                                                          Array<double> &out_adValues,
                                                          Array<int> &out_aiIndex)
const
{
    out_adValues.need(0);
    out_aiIndex.need(0);

    // Evaluate basis function/control point pairs

    int s_u, e_u, s_v, e_v;
    if ( m_kvecU.Span( in_pt[0], s_u, e_u ) == FALSE )
        return;

    if ( m_kvecV.Span( in_pt[1], s_v, e_v ) == FALSE )
        return;

    double bu, bv;
    for (int ui = s_u; ui < e_u; ui++) {
        if ( in_iWhich == 0 ) {
            bu = m_kvecU.BasisDerivative( in_pt[0], ui );
        } else {
            bu = m_kvecU.Basis(in_pt[0], ui);
        }
        if ( RNIsZero(bu) ) continue;

        for (int vi = s_v; vi < e_v; vi++) {
            if ( in_iWhich == 1 ) {
                bv = m_kvecV.BasisDerivative( in_pt[1], vi );
            } else {
                bv = m_kvecV.Basis(in_pt[1], vi);
            }
            if ( RNIsZero(bv) ) continue;
            out_adValues += bu * bv;
            out_aiIndex += PtIndx( ui, vi );
        }
    }
}

/* ------------------------------------------------------------
 * DESCR   :	Evaluate a point
 * DETAILS :    Multiply basis functions by control points, 
 *              divide by sum of weights 
 * ------------------------------------------------------------ */
template<class InPoint, class OutPoint, class OutVector>
inline
void
SRFNurbTC<InPoint, OutPoint, OutVector>::BasisSecondDerivative( const R2Pt_i &in_iptWhich, const R2Pt &in_pt,
                                                                Array<double> &out_adValues,
                                                                Array<int> &out_aiIndex)
const
{
    out_adValues.need(0);
    out_aiIndex.need(0);

    // Evaluate basis function/control point pairs

    int s_u, e_u, s_v, e_v;
    if ( m_kvecU.Span( in_pt[0], s_u, e_u ) == FALSE )
        return;

    if ( m_kvecV.Span( in_pt[1], s_v, e_v ) == FALSE )
        return;

    double bu, bv;
    int iU = 0, iV = 0;
    if ( in_iptWhich[0] == 0 ) iU++;
    else                       iV++;
    if ( in_iptWhich[1] == 0 ) iU++;
    else                       iV++;

    for (int ui = s_u; ui < e_u; ui++) {
        if ( iU == 0 ) {
            bu = m_kvecU.Basis(in_pt[0], ui);
        } else if ( iU == 1 ) {
            bu = m_kvecU.BasisDerivative( in_pt[0], ui );
        } else if ( iU == 2 ) {
            bu = m_kvecU.BasisSecondDerivative( in_pt[0], ui );
        }
        if ( RNIsZero(bu) ) continue;

        for (int vi = s_v; vi < e_v; vi++) {
            if ( iV == 0 ) {
                bv = m_kvecV.Basis(in_pt[1], vi);
            } else if ( iV == 1 ) {
                bv = m_kvecV.BasisDerivative( in_pt[1], vi );
            } else {
                bv = m_kvecV.BasisSecondDerivative( in_pt[1], vi );
            }
            if ( RNIsZero(bv) ) continue;
            out_adValues += bu * bv;
            out_aiIndex += PtIndx( ui, vi );
        }
    }
}

/* ------------------------------------------------------------
 * DESCR   :	Evaluate a point
 * DETAILS :    Multiply basis functions by control points, 
 *              divide by sum of weights 
 * ------------------------------------------------------------ */
template<class InPoint, class OutPoint, class OutVector>
inline
void
SRFNurbTC<InPoint, OutPoint, OutVector>::BasisAll( const R2Pt &in_pt,
                                                   Array<double> &out_adValues,
                                                   Array<double> &out_adDerivDu,
                                                   Array<double> &out_adDerivDv,
                                                   Array<double> &out_adDerivDuu,
                                                   Array<double> &out_adDerivDvv,
                                                   Array<double> &out_adDerivDuv,
                                                   Array<int> &out_aiIndex)
const
{
    out_adValues.need(0);
    out_adDerivDu.need(0);
    out_adDerivDv.need(0);
    out_adDerivDuu.need(0);
    out_adDerivDvv.need(0);
    out_adDerivDuv.need(0);
    out_aiIndex.need(0);

    // Evaluate basis function/control point pairs

    int s_u, e_u, s_v, e_v;
    if ( m_kvecU.Span( in_pt[0], s_u, e_u ) == FALSE )
        return;

    if ( m_kvecV.Span( in_pt[1], s_v, e_v ) == FALSE )
        return;

    for (int ui = s_u; ui < e_u; ui++) {
        const double bu = m_kvecU.Basis(in_pt[0], ui);
        const double bDu = m_kvecU.BasisDerivative( in_pt[0], ui );
        const double bDuu = m_kvecU.BasisSecondDerivative( in_pt[0], ui );

        if ( RNIsZero(bu) && RNIsZero(bDu) && RNIsZero(bDuu) ) continue;

        for (int vi = s_v; vi < e_v; vi++) {
            const double bv = m_kvecV.Basis(in_pt[1], vi);
            const double bDv = m_kvecV.BasisDerivative( in_pt[1], vi );
            const double bDvv = m_kvecV.BasisSecondDerivative( in_pt[1], vi );

            if ( RNIsZero(bv) && RNIsZero(bDv) && RNIsZero(bDvv) ) continue;

            out_adValues += bu * bv;
            out_adDerivDu += bDu   * bv;
            out_adDerivDv += bu    * bDv;
            out_adDerivDuu += bDuu * bv;
            out_adDerivDvv += bu   * bDvv;
            out_adDerivDuv += bDu  * bDv;
            out_aiIndex += PtIndx( ui, vi );
        }
    }
}


template<class InPoint, class OutPoint, class OutVector>
inline 
void 
SRFNurbTC<InPoint, OutPoint, OutVector>::AllEval( const R2Pt &in_pt,
                                                  OutPoint &out_pt,
                                                  OutVector &out_vecDs, OutVector &out_vecDt,
                                                  OutVector &out_vecDss, OutVector &out_vecDtt,
                                                  OutVector &out_vecDst ) const
{
    RNZero( out_pt );
    RNZero( out_vecDs );
    RNZero( out_vecDt );
    RNZero( out_vecDss );
    RNZero( out_vecDtt );
    RNZero( out_vecDst );

    int s_u, e_u, s_v, e_v;
    if ( m_kvecU.Span( in_pt[0], s_u, e_u ) == FALSE )
        return;

    if ( m_kvecV.Span( in_pt[1], s_v, e_v ) == FALSE )
        return;

    const int iNU = e_u - s_u;
    const int iNV = e_v - s_v;
    Array<double> adU(iNU), adUD(iNU), adUDD(iNU), adV(iNV), adVD(iNV), adVDD(iNV);

    for ( int iU = s_u; iU < e_u; iU++ ) {
        m_kvecU.Basis( iU ).AllEval( in_pt[0], adU[iU - s_u], adUD[iU - s_u], adUDD[iU - s_u] );
    }
    for ( int iV = s_v; iV < e_v; iV++ ) {
        m_kvecV.Basis( iV ).AllEval( in_pt[1], adV[iV - s_v], adVD[iV - s_v], adVDD[iV - s_v] );
    }

    InPoint ptOut;
    RNZero( ptOut );
    for ( FORINT iU = 0; iU < iNU; iU++ ) {
        for ( int iV = 0; iV < iNV; iV++ ) {
            const InPoint &pt = Pt( s_u + iU, s_v + iV );
            for ( int iD = 0; iD < InDim(); iD++ ) {
                ptOut[iD] += pt[iD] * adU[iU] * adV[iV];
                out_vecDs[iD] += pt[iD] * adUD[iU] * adV[iV];
                out_vecDt[iD] += pt[iD] * adU[iU] * adVD[iV];
                out_vecDss[iD] += pt[iD] * adUDD[iU] * adV[iV];
                out_vecDtt[iD] += pt[iD] * adU[iU] * adVDD[iV];
                out_vecDst[iD] += pt[iD] * adUD[iU] * adVD[iV];
            }
        }
    }
    out_pt = ptOut;
}


template<class InPoint, class OutPoint, class OutVector>
inline 
void 
SRFNurbTC<InPoint, OutPoint, OutVector>::AllEval( const R2Pt &in_pt,
                                                  OutPoint &out_pt,
                                                  OutVector &out_vecDs, OutVector &out_vecDt ) const
{
    RNZero( out_pt );
    RNZero( out_vecDs );
    RNZero( out_vecDt );

    int s_u, e_u, s_v, e_v;
    if ( m_kvecU.Span( in_pt[0], s_u, e_u ) == FALSE )
        return;

    if ( m_kvecV.Span( in_pt[1], s_v, e_v ) == FALSE )
        return;

    const int iNU = e_u - s_u;
    const int iNV = e_v - s_v;
    Array<double> adU(iNU), adUD(iNU), adV(iNV), adVD(iNV);

    for ( int iU = s_u; iU < e_u; iU++ ) {
        m_kvecU.Basis( iU ).AllEval( in_pt[0], adU[iU - s_u], adUD[iU - s_u] );
    }
    for ( int iV = s_v; iV < e_v; iV++ ) {
        m_kvecV.Basis( iV ).AllEval( in_pt[1], adV[iV - s_v], adVD[iV - s_v] );
    }

    InPoint ptOut;
    RNZero( ptOut );
    for ( FORINT iU = 0; iU < iNU; iU++ ) {
        for ( int iV = 0; iV < iNV; iV++ ) {
            const InPoint &pt = Pt( s_u + iU, s_v + iV );
            for ( int iD = 0; iD < InDim(); iD++ ) {
                ptOut[iD] += pt[iD] * adU[iU] * adV[iV];
                out_vecDs[iD] += pt[iD] * adUD[iU] * adV[iV];
                out_vecDt[iD] += pt[iD] * adU[iU] * adVD[iV];
            }
        }
    }
    out_pt = ptOut;
}

/* ------------------------------------------------------------
 * DESCR   :	Find the closest point
 * ------------------------------------------------------------ */
template<class InPoint, class OutPoint, class OutVector>
inline
OutPoint
SRFNurbTC<InPoint, OutPoint, OutVector>::Closest_pt(const InPoint &in_pt, 
                                                    R2Pt &out_pt,
                                                    const int in_iNDivs) const 
{
    double min_d = 1e30;
   
    const int iNDivs = WINmax(2, in_iNDivs);
    const double div = 1.0 / (iNDivs - 1);
    for (int u = 0; u < m_kvecU.Num_segments(); u++) {
        for (int i = 0; i < iNDivs; i++) {
            const double du = m_kvecU.Segment_perc_t(u, div * i);

            for (int v = 0; v < m_kvecV.Num_segments(); v++) {
                for (int j = 0; j < iNDivs; j++) {
                    const double dv = m_kvecV.Segment_perc_t(v, div * j);
                    const double d = Length((*this)(R2Pt( du, dv ) ) - in_pt);

                    if (d < min_d) {
                        min_d = d;
                        out_pt[0] = du;
                        out_pt[1] = dv;
                    }
                }
            }
        }
    }
    return (*this)( out_pt );
}


/* ------------------------------------------------------------
 * DESCR   :	Print the knot vector and the points
 * ------------------------------------------------------------ */
template<class InPoint, class OutPoint, class OutVector>
inline
void
SRFNurbTC<InPoint, OutPoint, OutVector>::Print(const WINbool do_basis)
     const 
{
   cout << "Dim " << InDim() << " " << OutDim() << "\n";
   
   for (int u = 0; u < m_kvecU.Num_pts(); u++) {
      for (int v = 0; v < m_kvecV.Num_pts(); v++) {
          cout << Pt(u,v) << " ";
      }
      cout << "\n";
   }
   m_kvecU.PrintKnotVec(do_basis);
   m_kvecV.PrintKnotVec(do_basis);
}

/* ------------------------------------------------------------
 * DESCR   :	Write out and read in
 * DETAILS
 *    order
 *    dimension (of points+1)
 *    num of points
 *    num of data allocd
 *    knot dimension
 *    knot (RNvec::write)
 *    pts 0..n (Point::write)
 * ------------------------------------------------------------ */
template<class InPoint, class OutPoint, class OutVector>
inline
WINbool
SRFNurbTC<InPoint, OutPoint, OutVector>::Read(ifstream &in)
{
   WINbool out_bRes = TRUE;
   out_bRes = (out_bRes && m_kvecU.ReadKnotVec(in)) ? TRUE : FALSE;
   out_bRes = (out_bRes && m_kvecV.ReadKnotVec(in)) ? TRUE : FALSE;

   m_apts.need(m_kvecU.Num_pts() * m_kvecV.Num_pts());
   
   for (int i = 0; i < m_apts.num(); i++) 
       if ( m_apts[i].Read( in ) == FALSE)
           out_bRes = FALSE;
   
   return out_bRes;
}

template<class InPoint, class OutPoint, class OutVector>
inline
void
SRFNurbTC<InPoint, OutPoint, OutVector>::Write(ofstream &out)
const 
{
   m_kvecU.WriteKnotVec(out);
   m_kvecV.WriteKnotVec(out);
   
   for (int i = 0; i < m_apts.num(); i++)
       out << m_apts[i] << " ";
}


template<class InPoint, class OutPoint, class OutVector>
inline
WINbool SRFNurbTC<InPoint, OutPoint, OutVector>::Read_binary(ifstream &in)
{
   WINbool bRes = TRUE;

   if (m_kvecU.Read_binary(in) == FALSE) bRes = FALSE;
   if (m_kvecV.Read_binary(in) == FALSE) bRes = FALSE;

   m_apts.need(m_kvecU.Num_pts() * m_kvecV.Num_pts());
   
   for (int i = 0; i < m_apts.num(); i++) {
       in.read((char *) &m_apts[i], sizeof(double) * InDim() );
       if (in.gcount() != sizeof(double) * InDim() ) bRes = FALSE;
   }

   return bRes;
}

template<class InPoint, class OutPoint, class OutVector>
inline
void SRFNurbTC<InPoint, OutPoint, OutVector>::Write_binary(ofstream &out) const
{
   m_kvecU.Write_binary(out);
   m_kvecV.Write_binary(out);

   for (int i = 0; i < m_apts.num(); i++) 
     out.write( (const char *) &m_apts[i][0], InDim() * sizeof(double) );
}

