


template <class TData>
inline
void STATKMeans<TData>::BuildTree()
{
    for ( int i = 0; i < m_aadCenters.num(); i++ )
        for ( int j = 0; j < m_aadCenters[i].num(); j++ )
            m_aptsData[i][j] = m_aadCenters[i][j];

    if ( m_KDTree != NULL )
        delete m_KDTree;

    m_KDTree = new ANNkd_tree(			// build search structure
		    m_aptsData,			// the data points
		    m_aadCenters.num(),			// number of points
		    m_aadCenters[0].num());			// dimension of space

}

template <class TData>
inline
int  STATKMeans<TData>::ClosestInTree( const int in_i )
{
    const Array<double> &ad = GetData(in_i).m_adColorToCompare;

    for ( int j = 0; j < ad.num(); j++ )
        m_ptQuery[j] = ad[j];

	m_KDTree->annkSearch(			// search
		m_ptQuery,			// query point
		m_iNeighSize,				// number of near neighbors
		m_aiNeigh,				// nearest neighbors (returned)
     	m_adNeighDist,				// distance (returned)
		m_dAvgDiff);				// error bound

    int iClosest = m_aiNeigh[0];
    double dClosest = Distance( GetData(in_i), m_aiNeigh[0] );

    for (int iC = 1; iC < m_iNeighSize; iC++) {
        const double dDist = Distance( GetData(in_i), m_aiNeigh[iC] );
        if ( dDist < dClosest ) {
            dClosest = dDist;
            iClosest = m_aiNeigh[iC];
        }
   }

   return iClosest;
}


template <class TData>
inline
double STATKMeans<TData>::Distance( const Data &in_oData, const int in_iCenter ) const
{
    return in_oData.Distance( m_aadCenters[ in_iCenter ] );
}

template <class TData>
inline
void  STATKMeans<TData>::CalculateMean( const Array< const Data *> &in_aopData, const int in_iCenter ) 
{
    Array<double> &adCenter = m_aadCenters[in_iCenter];

    memset( &adCenter[0], 0, adCenter.num() * sizeof( double ) );

    if ( in_aopData.num() == 0 )
        return;

    for ( int i = 0; i < in_aopData.num(); i++ )
        in_aopData[i]->Add( adCenter );

    const double dDiv = 1.0 / in_aopData.num();

    for ( FORINT i = 0; i < adCenter.num(); i++ )
        adCenter[i] *= dDiv;
}

template <class TData>
inline
double  STATKMeans<TData>::AvgDist( const Array< const Data *> &in_aopData, const int in_iCenter ) const
{
    double dDiff = 0;

    if ( in_aopData.num() == 0 )
        return 0;

    const Array<double> &adCenter = m_aadCenters[in_iCenter];
    for ( int i = 0; i < in_aopData.num(); i++ )
        dDiff += in_aopData[i]->AvgDiff( adCenter );

    return dDiff / (double) in_aopData.num();
}

template<class TData>
inline
void STATKMeans<TData>::AssignRandom()
{
    ASSERT ( m_aadCenters.num() != 0 );

    for ( int iC = 0; iC < m_aadCenters.num(); iC++ ) {
        m_aaopData[iC].need(0);
    }

    m_aiDataCenter.need( NData() );

    const int iMod = m_aadCenters.num();
    for ( int i = 0; i < NData(); i++ ) {
        const int iRand = rand() % iMod;
        m_aiDataCenter[i] = iRand;

        m_aaopData[ m_aiDataCenter[i] ] += GetDataPtr( i );
    }
}

template<class TData>
inline
int STATKMeans<TData>::ReClassify()
{
    if ( m_aadCenters.num() == 0 )
        return 0;

    int out_iMoved = 0;

    for ( int iC = 0; iC < m_aadCenters.num(); iC++ ) {
        m_aaopData[iC].need(0);
        m_abChanged[iC] = FALSE;
    }

    for ( int iD = 0; iD < NData(); iD++ ) {
        int iClosest = ClosestInTree(iD);

        if ( m_aiDataCenter[iD] != iClosest ) {
            m_abChanged[ m_aiDataCenter[iD] ] = TRUE;
            m_abChanged[ iClosest ] = TRUE;
            out_iMoved++;
        }

        m_aiDataCenter[iD] = iClosest;

        m_aaopData[iClosest] += GetDataPtr(iD);
    }

    int iChanged = 0;
    for ( FORINT iC = 0; iC < m_abChanged.num(); iC++ )
        if ( m_abChanged[iC] == TRUE )
            iChanged++;

    PRINT("Changed %d/%d ", iChanged, m_abChanged.num() );
    return out_iMoved;
}

template<class TData>
int STATKMeans<TData>::CheckDiff()
{
    Array<int>    aiFree, aiSplit;
    Array<double> adWidth;
    for ( int iC = 0; iC < m_aadCenters.num(); iC++ ) {
        if ( m_aaopData[iC].num() == 0 ) {
            m_adCenterWidth[iC] = 0.0;
            aiFree += iC;
        } else {
            m_adCenterWidth[iC] = AvgDist( m_aaopData[iC], iC );
            if ( m_adCenterWidth[iC] > m_dAvgDiff ) {
                aiSplit += iC;
                adWidth += 1.0 - m_adCenterWidth[iC];
            }
        }
    }

    aiSplit.sort( adWidth );

    PRINT("Free %d Split %d ", aiFree.num(), aiSplit.num() );

    const int iDiff = aiSplit.num() - aiFree.num();
    const int iAdd = ( iDiff <= 0 ) ? 0 : WINmax(1, (int) (iDiff * 0.85));
    const int iRemove = ( iDiff >= 0 || aiSplit.num() > 0) ? 0 : iDiff * -0.5;

    if ( iAdd > 0 ) {
        for ( int iA = 0; iA < iAdd; iA++ )
            aiFree += m_aadCenters.num() + iA;
        IncreaseSize( iAdd );
    }
    if ( iRemove > 0 ) {
        DecreaseSize( iRemove );
    }


    PRINT("Add %d Remove %d  ", iAdd, iRemove );
    const int iLoop = WINmin( aiFree.num(), aiSplit.num() );

    if ( adWidth.num() != 0 ) PRINT("%f ", adWidth[0]);

    for ( int i = 0; i < iLoop; i++ ) {

        Array< const SYNTHData * > &aopData = m_aaopData[ aiSplit[i] ];

        int iPick0 = 0, iPick1 = 1;
        double dDiff = aopData[iPick0]->AvgDiff( aopData[iPick1]->m_adColorToCompare );

        for ( int iTry0 = 0; iTry0 < aopData.num(); iTry0++ ) {
            for (int iTry1 = iTry0 + 1; iTry1 < aopData.num(); iTry1++ ) {
                double dDiffTry = aopData[iTry0]->AvgDiff( aopData[iTry1]->m_adColorToCompare );
                if ( dDiffTry > dDiff ) {
                    dDiff = dDiffTry;
                    iPick0 = iTry0;
                    iPick1 = iTry1;
                }
            }
        }

        m_aadCenters[ aiSplit[i] ] = aopData[iPick0]->m_adColorToCompare;
        m_aadCenters[ aiFree[i] ] = aopData[iPick1]->m_adColorToCompare;
        m_abChanged[ aiSplit[i] ] = TRUE;
        m_abChanged[ aiFree[i] ] = TRUE;
    }
    return iLoop;
}

template<class TData>
inline
WINbool STATKMeans<TData>::Iterate( const int in_iNTimes )
{
    if ( NData() == 0 )
        return TRUE;

    const int iStop = (int) (m_dMoveThreshold * 0.2 * NData());

    int iNMoved = NData(), iCheckDiff = 0;

    AssignRandom();
    for ( int i = 0; i < in_iNTimes; i++ ) {
        for ( int iC = 0; iC < m_aadCenters.num(); iC++ ) {
            if ( m_abChanged[iC] == TRUE ) {
                CalculateMean( m_aaopData[iC], iC );
            }
        }

        // We have the current centers with the current assignment
        if ( iNMoved < 0.1 * NData() ) {
            iNMoved = CheckDiff();
            iCheckDiff = iNMoved;
        } else {
            iNMoved = 0;
        }

        BuildTree();

        PRINT("Checked %d ", iNMoved);

        iNMoved += ReClassify();

        PRINT("Iteration %d moved %d\n", i, iNMoved);
        if ( iCheckDiff == 0 && iNMoved < iStop )
            return TRUE;

    }

    const double dMoved = (double) iNMoved / (double) NData();

    PRINT("Bailing at %f\n", dMoved);
    if ( dMoved < m_dMoveThreshold )
        return TRUE;

    return FALSE;
}

template<class TData>
inline
int STATKMeans<TData>::AssignedCenter( const int in_iDataId ) const
{
    return m_aiDataCenter[in_iDataId];
}

///
template<class TData>
inline
const Array<double> &STATKMeans<TData>::CenterValue( const int in_iCenter ) const
{
    return m_aadCenters[ in_iCenter ];
}

template<class TData>
inline
void STATKMeans<TData>::GetDataForCenter( Array<const TData *> &out_aopData, const int in_iCenter ) const
{
    out_aopData.need(0);

    for ( int i = 0; i < NData(); i++ ) {
        if ( m_aiDataCenter[i] == in_iCenter )
            out_aopData += GetDataPtr(i);
    }
}

template<class TData>
inline
void STATKMeans<TData>::SetNCenters( const int in_i, const int in_iSize )
{
    m_aadCenters.need(in_i);
    m_aaopData.need( in_i );
    m_adCenterWidth.need( in_i );
    m_abChanged.need( in_i );
    m_aptsData = annAllocPts( in_i, in_iSize );
    m_ptQuery = annAllocPt( in_iSize );
    m_aiNeigh = new ANNidx[m_iNeighSize];			// allocate near neigh indices
    m_adNeighDist = new ANNdist[m_iNeighSize];			// allocate near neighbor dists

    for ( int i = 0; i < m_aadCenters.num(); i++ ) {
        m_aadCenters[i].need( in_iSize );
        memset( &m_aadCenters[i][0], 0, in_iSize * sizeof(double) );
        m_abChanged[i] = TRUE;
    }
}

template<class TData>
inline
void STATKMeans<TData>::IncreaseSize( const int in_iN )
{
    const int iN = m_aadCenters.num();

    m_aadCenters.add(in_iN);
    m_aaopData.add( in_iN );
    m_adCenterWidth.add( in_iN );
    m_abChanged.add( in_iN );
    
    annDeallocPts( m_aptsData );
    m_aptsData = annAllocPts( m_aadCenters.num(), m_aadCenters[0].num() );

    for ( int i = iN; i < m_aadCenters.num(); i++ ) {
        m_aadCenters[i].need( m_aadCenters[0].num() );
        memset( &m_aadCenters[i][0], 0, m_aadCenters[0].num() * sizeof(double) );
        m_adCenterWidth[ i ] = 0.0;
        m_abChanged[i] = TRUE;
   }
}

template<class TData>
inline
void STATKMeans<TData>::DecreaseSize( const int in_iN )
{
    Array<int> aiFree, aiNotFree;
    for ( int i = 0; i < m_aadCenters.num(); i++ ) {
        if ( m_aaopData[i].num() == 0 ) {
            aiFree += i;
        } else {
            aiNotFree += i;
        }
    }

    const int iLast = m_aadCenters.num() - in_iN;
    int iFree = 0;
    for ( FORINT i = 0; i < aiNotFree.num(); i++ ) {
        if ( aiNotFree[i] >= iLast ) {
            const int iMove = aiNotFree[i];
            const int iTo = aiFree[ iFree++ ];

            ASSERT( iTo < iLast );

            const Array<double> &adMove = m_aadCenters[iMove];
            Array<double> &adTo = m_aadCenters[iTo];
            for ( int j = 0; j < adMove.num(); j++ )
                adTo[j] = adMove[j];

            m_adCenterWidth[iTo] = m_adCenterWidth[iMove];        
            m_abChanged[iTo] = m_abChanged[iMove];


            const Array< const TData * > &aopMove = m_aaopData[iMove];
            Array< const TData * > &aopTo = m_aaopData[iTo];
            for ( FORINT j = 0; j < aopMove.num(); j++ )
                aopTo += aopMove[j];

            for ( FORINT j = 0; j < NData(); j++ ) {
                if ( m_aiDataCenter[j] == iMove )
                    m_aiDataCenter[j] = iTo;
            }
        }
    }

    m_aadCenters.sub(in_iN);
    m_aaopData.sub( in_iN );
    m_adCenterWidth.sub( in_iN );    

    annDeallocPts( m_aptsData );
    m_aptsData = annAllocPts( m_aadCenters.num(), m_aadCenters[0].num() );
}

template<class TData>
inline
WINbool STATKMeans<TData>::Solve( const int in_iTry )
{
    for ( int i = 0; i < in_iTry; i++ )
        if ( Iterate(30) == TRUE )
            return TRUE;

    return FALSE;
}

template<class TData>
inline
STATKMeans<TData>::STATKMeans() : m_dMoveThreshold( 0.05 ), m_dAvgDiff( 0.05 ), m_iNeighSize(3), m_KDTree(NULL)
{
}

///
template<class TData>
inline
STATKMeans<TData>::~STATKMeans()
{
}


