#include "StdAfx.H"
#include <stdio.h>
#include <MC/MC_Conic.H>
#include <MC/MC_Point.H>
#include <MC/MC_Param.H>
#include <MC/MC_Fit.H>
#include <MC/MC_MeshSubset.H>
#include <MC/MC_Transform.H>
#include <utils/Pm_MeshLite.H>
#include <fitting/FITTools_Least_squares.H>

MCConic::MCConic(const PMeshLite &in_mesh,
                 const int in_nRings )
: MCFit(in_mesh,CONIC,MCParam::NONE,in_nRings)
{

}

MCConic::~MCConic()
{
}

double MCConic::FunctionValue(const R3Pt &in_point)
{
    double value;
    R3Pt P;

    P = in_point;
    m_Transform.toLocal(P);

    // Ax^2 + By^2 + Cz^2 + Dxy + Exz + Fyz + Gx + Hy + Iz + J = 0
    // By^2 + (Dx+Fz+H)y + (Ax^2+Cz^2+Exz+Gx+Iz+J) = 0

    value = A[0]*P[0]*P[0] + A[1]*P[1]*P[1] + A[2]*P[2]*P[2]
          + A[3]*P[0]*P[1] + A[4]*P[0]*P[2] + A[5]*P[1]*P[2]
          + A[6]*P[0]      + A[7]*P[1]      + A[8]*P[2]   + A[9];

    return value;
}

int 
MCConic::Curvature(const Vertex &in_vertex, MCPoint &cpt)
{
    int ier;
    double k1, k2, vsign;
    R3Pt P;
    R3Vec NN, Xu, Xv, Xuu, Xuv, Xvv, V;
    double gMag, e1, e2;
    R2Vec s, t;
    R3Vec b1, b2, N, k1vec, k2vec;
    R2Matrix M;
    FILE *out = NULL;

    //if( !valid || in_vertex != m_Vertex ) {
        ier = CalculateFit(in_vertex);
		if( ier != 0 ) return ier;
    //}

    // Evaluate the function to get point, normal, and derivatives 
    
    P = in_vertex->GetLoc();

    // Convert to transformed coordinates
    m_Transform.toLocal(P);
    
    if( A.num() < 10 ) return -1;
    // Ax^2 + By^2 + Cz^2 + Dxy + Exz + Fyz + Gx + Hy + Iz + J = 0

    // Set up the Hessian matrix
    const R3Matrix H = R3Matrix( R3CoVec( 2.0*A[0],     A[3],     A[4] ),
                                 R3CoVec( A[3], 2.0*A[1],     A[5] ),
                                 R3CoVec( A[4],     A[5], 2.0*A[2] ) );
    
    // Find a set of basis functions in the local normal plane
    N = getGradient(P);
    N.Normalize();
    b1 = getBasis1(N);
    b2 = Cross(N,b1);
    b2.Normalize();
    gMag = getGradientMag(P);
    if( gMag == 0.0 ) return -1;

    // Set up the matrix Aij = [bi H bj]
    const R3CoVec preMult( b1[0], b1[1], b1[2] );
    const R3CoVec preMultTwo( b2[0], b2[1], b2[2] );
    R3CoVec v = preMult * H;
    M(0,0) = Dot(v,preMult);
    M(0,1) = Dot(v,preMultTwo);

    v = preMultTwo * H;
    M(1,0) = Dot(v,preMult);
    M(1,1) = Dot(v,preMultTwo);

    // Curvatures are the eigenvalues of A
    // Get the eigenvectors of A
    M.EigenReal( e1, e2, s, t );
    k1 = e1 / gMag;
    k2 = e2 / gMag;

    k1vec = s[0]*b1 + s[1]*b2;
    k2vec = t[0]*b1 + t[1]*b2;

    
    // Transform back to physical coordinates
    m_Transform.fromLocal(k1vec);
    m_Transform.fromLocal(k2vec);
    m_Transform.fromLocal(N);

    /*
    N = getNref();
    if( Dot(out_vecNorm,N) < 0 ) {
        out_dCrv1 = -out_dCrv1;
        out_dCrv2 = -out_dCrv2;
    }
    */
	V = Nref();
	vsign = V[0]*N[0] + V[1]*N[1] + V[2]*N[2];
	if( vsign < 0.0 ) {
		k1 = -k1;
		k2 = -k2;
	}

    cpt.setId(in_vertex->IndexId());
	if( k1 < k2 ) {
		cpt.setMaxMin(k2,k1);
		cpt.setMaxMinVec(k2vec,k1vec);
	} else {
		cpt.setMaxMin(k1,k2);
		cpt.setMaxMinVec(k1vec,k2vec);
	}

	if( vsign < 0.0 ) {
		cpt.setNormal(R3Vec(-N[0],-N[1],-N[2]));
	} else {
		cpt.setNormal(N);
	}

    return 0;
}
    
int 
MCConic::CalculateFit(const Vertex &in_vertex)
{
    int i, nv;
    double d;
    R3Pt P, Q;
    R3Vec N;

    // Create the mesh subset
    m_Param.Create(in_vertex,m_iRings);

    // Compute the transformation
    P = in_vertex->GetLoc();
    m_Param.CalculateNormal(in_vertex, MCMeshSubset::ANGLE, N);
    m_Transform.setTransform(P,N);

    // Setup least square problem
    nv = m_Param.Vertices().num();
    m_dDia = 0.0;
    
    FITTools_LS prob_ls( nv+1, 10, 1 );

    // Set up the matrix to solve 
    for( i=0; i<nv; i++ ) {
        Q = m_Param.Vertices()[i]->GetLoc();
        d = m_Param.Distance(P,Q);
        if( d > m_dDia ) m_dDia = d;

        m_Transform.toLocal(Q);

        prob_ls.Desired_clapack( i, 0 ) = Q[0]*Q[0];
        prob_ls.Desired_clapack( i, 1 ) = Q[1]*Q[1];
        prob_ls.Desired_clapack( i, 2 ) = Q[2]*Q[2];
        prob_ls.Desired_clapack( i, 3 ) = Q[0]*Q[1];
        prob_ls.Desired_clapack( i, 4 ) = Q[0]*Q[2];
        prob_ls.Desired_clapack( i, 5 ) = Q[1]*Q[2];
        prob_ls.Desired_clapack( i, 6 ) = Q[0];
        prob_ls.Desired_clapack( i, 7 ) = Q[1];
        prob_ls.Desired_clapack( i, 8 ) = Q[2];
        prob_ls.Desired_clapack( i, 9 ) = 1.0;
        prob_ls.Rhs_clapack( i, 0 ) = 0.0;
    }

    // Add a non-zero constraint 
    for( i=0; i<10; i++ ) {
        prob_ls.Desired_clapack( nv, i ) = 0.0;  
    }
    prob_ls.Desired_clapack( nv, 0 ) = 1.0;
    prob_ls.Rhs_clapack( nv, 0 ) = 1.0;

    // Do the least squares fit 
    if( prob_ls.Solve_clapack() == FALSE ) {
        printf("ERROR - Solving Least Square problem\n" );
        prob_ls.Print_clapack( cout, FALSE );
    }
    
    // Store the polynomial coefficients
    A.need(0);
    for( i=0; i<10; i++ ) {
        A += prob_ls.Solution_clapack( i, 0 );
    }

    return 0;
}

/*
void MCConic::Analyze(TDGMesh *mesh)
{
    int i;
    float Dmin, Dmax, Dmean, Dsd;
    double x, y, z;
    R3Pt p;
    R3Vec N;
    Array< const PMVertex * > aopVs;
    Array< float > data;

    mesh->CAllMeshVertices(aopVs);
    for( i=0; i<mesh->NumVertices(); i++ ) {
        p = aopVs[i]->GetLoc();
        m_Transform.toLocal(p[0],p[1],p[2],"point");
        data += Function((float)p[0],(float)p[1],(float)p[2]);
    }

    N = getNref();
    Normalize(N);
    N *= 0.1;
    m_Transform.toLocal(N[0],N[1],N[2],"vector");

    getCenter(x,y,z);
    m_Transform.toLocal(x,y,z,"point");

    TRACE("Center = %f, 0.1N = %f\n", Function((float)x, (float)y, (float)z),
        Function((float)(x+N[0]), (float)(y+N[1]), (float)(z+N[2])));
    ArrayStatistics(data, Dmin, Dmax, Dmean, Dsd );
    TRACE("Conic Fit (%d points): Min=%f, Max=%f, Mean=%f, StdDev=%f\n",
        mesh->NumVertices(), Dmin, Dmax, Dmean, Dsd );
}
*/
    
int MCConic::ProjectLocalX(R3Pt &point) const
{
    // Transformed coordinates
    double a, b, c, r, x, y, z, x1, x2, magX1, magX2;

    x = point[0];
    y = point[1];
    z = point[2];

    if( A.num() < 10 ) return -1;
    // Ax^2 + By^2 + Cz^2 + Dxy + Exz + Fyz + Gx + Hy + Iz + J = 0
    // Ax^2 + (Dy+Ez+G)x + (By^2+Cz^2+Fyz+Hy+Iz+J) = 0

    a = A[0];
    b = A[3]*y + A[4]*z + A[6];
    c = A[1]*y*y + A[2]*z*z + A[5]*y*z + A[7]*y + A[8]*z + A[9];

    if( a == 0.0 && b == 0.0 ) {
        return -2;
    }
    if( a == 0.0 ) {
        point[0] = -c/b;
        return 0;
    }
    r = b*b - 4.0*a*c;
    if( r < 0.0 ) {
        r = 0.0;
        point[0] = -b / (2.0*a);
        return -2;
    }
    x1 = (-b + sqrt(r)) / (2.0*a);
    x2 = (-b - sqrt(r)) / (2.0*a);
    magX1 = x1<x ? x-x1 : x1-x;
    magX2 = x2<x ? x-x2 : x2-x;
    point[0] = (magX2 < magX1) ? x2 : x1;
    
    return 0;
}

int MCConic::ProjectLocalY(R3Pt &point) const
{
    // Transformed coordinates
    double a, b, c, r, x, y, z, y1, y2, magY1, magY2;

    x = point[0];
    y = point[1];
    z = point[2];
    
    //y = 0.0;
    if( A.num() < 10 ) return -1;
    // Ax^2 + By^2 + Cz^2 + Dxy + Exz + Fyz + Gx + Hy + Iz + J = 0
    // By^2 + (Dx+Fz+H)y + (Ax^2+Cz^2+Exz+Gx+Iz+J) = 0

    a = A[1];
    b = A[3]*x + A[5]*z + A[7];
    c = A[0]*x*x + A[2]*z*z + A[4]*x*z + A[6]*x + A[8]*z + A[9];

    if( a == 0.0 && b == 0.0 ) {
        return -2;
    }
    if( a == 0.0 ) {
        point[1] = -c/b;
        return 0;
    }
    r = b*b - 4.0*a*c;
    if( r < 0.0 ) {
        r = 0.0;
        point[1] = -b / (2.0*a);
        return -2;
    }
    y1 = (-b + sqrt(r)) / (2.0*a);
    y2 = (-b - sqrt(r)) / (2.0*a);
    magY1 = y1<y ? y-y1 : y1-y;
    magY2 = y2<y ? y-y2 : y2-y;
    point[1] = (magY2 < magY1) ? y2 : y1;
    
    return 0;
}

int MCConic::ProjectLocalZ(R3Pt &point) const
{
    // Transformed coordinates
    double a, b, c, r, x, y, z, z1, z2, magZ1, magZ2;

    x = point[0];
    y = point[1];
    z = point[2];
    
    //z = 0.0;
    if( A.num() < 10 ) return -1;
    // Ax^2 + By^2 + Cz^2 + Dxy + Exz + Fyz + Gx + Hy + Iz + J = 0
    // Cz^2 + (Ex+Fy+I)z + (Ax^2+By^2+Dxy+Gx+Hy+J) = 0

    a = A[2];
    b = A[4]*x + A[5]*y + A[8];
    c = A[0]*x*x + A[1]*y*y + A[3]*x*y + A[6]*x + A[7]*y + A[9];

    if( a == 0.0 && b == 0.0 ) {
        return -2;
    }
    if( a == 0.0 ) {
        point[2] = -c/b;
        return 0;
    }
    r = b*b - 4.0*a*c;
    if( r < 0.0 ) {
        r = 0.0;
        point[2] = -b / (2.0*a);
        return -2;
    }
    z1 = (-b + sqrt(r)) / (2.0*a);
    z2 = (-b - sqrt(r)) / (2.0*a);
    magZ1 = z1<z ? z-z1 : z1-z;
    magZ2 = z2<z ? z-z2 : z2-z;
    point[2] = (magZ2 < magZ1) ? z2 : z1;
    
    return 0;
}

R3Vec MCConic::getGradient(const R3Pt &in_point) const
{
    // Gradient of a point in transformed (local) coordinates
    double x, y, z, Nx, Ny, Nz, Nmag;

    // Ax^2 + By^2 + Cz^2 + Dxy + Exz + Fyz + Gx + Hy + Iz + J = 0
    
    x = in_point[0];
    y = in_point[1];
    z = in_point[2];
    
    Nx = 2.0*A[0]*x + A[3]*y + A[4]*z + A[6];
    Ny = 2.0*A[1]*y + A[3]*x + A[5]*z + A[7];
    Nz = 2.0*A[2]*z + A[4]*x + A[5]*y + A[8];

    Nmag = Nx*Nx + Ny*Ny + Nz*Nz;
    if( Nmag == 0.0 ) return R3Vec(0.0,0.0,0.0);
    Nmag = sqrt(Nmag);

    return R3Vec(Nx/Nmag,Ny/Nmag,Nz/Nmag);
}

double MCConic::getGradientMag(const R3Pt &in_point) const
{
    // Transformed coordinates
    double x, y, z, Nx, Ny, Nz;
    
    x = in_point[0];
    y = in_point[1];
    z = in_point[2];
    
    // Ax^2 + By^2 + Cz^2 + Dxy + Exz + Fyz + Gx + Hy + Iz + J = 0
    
    Nx = 2.0*A[0]*x + A[3]*y + A[4]*z + A[6];
    Ny = 2.0*A[1]*y + A[3]*x + A[5]*z + A[7];
    Nz = 2.0*A[2]*z + A[4]*x + A[5]*y + A[8];

    return sqrt(Nx*Nx + Ny*Ny + Nz*Nz);
}

R3Vec MCConic::getBasis1(const R3Vec &N) const
{
    // Physical coordinates
    int mode = 0;
    double a, b, c;
    R3Vec V;
    
    a = N[0]*N[0];
    b = N[1]*N[1];
    c = N[2]*N[2];
    if( b <= a && b <= c ) mode = 1;
    if( c <= a || c <= b ) mode = 2;
    if( a == 0.0 && b == 0.0 && c == 0.0 ) return R3Vec( 0.0, 0.0, 1.0 );

    switch(mode) {
        case 0:
            V = Cross(N,R3Vec(1.0,0.0,0.0));
            break;
        case 1:
            V = Cross(N,R3Vec(0.0,1.0,0.0));
            break;
        case 2:
            V = Cross(N,R3Vec(0.0,0.0,1.0));
            break;
    }
    V.Normalize();
    return V;
}

/*
int MCConic::Evaluate(const R3Pt X, const int &axis, R3Pt &Xnew ) const
{
    // Physical coordinates
    int ier;
    
    if( A.num() < 10 ) return -1;
    if( axis < 0 || axis > 3 ) return -2;

    Xnew = X;
    
    m_Transform.toLocal(Xnew);

    switch(axis) {
        case 1:
            ier =  ProjectLocalX(Xnew);
            break;
        case 2:
            ier =  ProjectLocalY(Xnew);
            break;
        case 3:
            ier =  ProjectLocalZ(Xnew);
            break;
        default:
            ier =  EvaluateN(X,Xnew);
            break;
    }
    m_Transform.fromLocal(Xnew);
    return ier;
}

int MCConic::EvaluateN(const R3Pt X, R3Pt &Xnew) const
{
    // Transformed coordinates
    double a, b, c, r, s, s1, s2, u, v, w;
    R3Vec N;
    FILE *out;
    int verbose = 1;

    if( A.num() < 10 ) return -1;
    
    Xnew = X;
    N = getGradient(X);
    u = N[0];
    v = N[1];
    w = N[2];

    x = X[0];
    y = X[1];
    z = X[2];

    a = A[0]*u*u + A[1]*v*v + A[2]*w*w + 
        A[3]*u*v + A[4]*u*w + A[5]*v*w;
    b = 2.0*(A[0]*u*x + A[1]*v*y + A[2]*w*z ) +
        A[3]*(u*y+v*x) + A[4]*(u*z + w*x) + A[5]*(v*z + w*y) +
        A[6]*u + A[7]*v + A[8]*w;
    c = A[0]*x*x + A[1]*y*y + A[2]*z*z + 
        A[3]*x*y + A[4]*x*z + A[5]*y*z + A[6]*x + A[7]*y + A[8]*z + A[9];

    r = b*b - 4.0*a*c;
    if( r < 0.0 ) r = 0.0;

    s1 = (-b + sqrt(r)) / (2.0*a);
    s2 = (-b - sqrt(r)) / (2.0*a);

    if( s1*s1 < s2*s2 ) { 
        s = s1;
    } else {
        s = s2;
    }

    Xnew[0] = x + s*u;
    Xnew[1] = y + s*v;
    Xnew[2] = z + s*w;

    return 0;
}

int MCConic::Project(const R3Pt &X, const R3Vec &NN, R3Pt &Xnew ) const
{
    // Physical coordinates
    double a, b, c, r, s, s1, s2, u, v, w, x, y, z;
    R3Vec N;
    FILE *out;
    int verbose = 1;

    if( A.num() < 10 ) return -1;
    
    Xnew = X;
    m_Transform.toLocal(Xnew);

    N = NN;
    m_Transform.toLocal(N);

    u = N[0];
    v = N[1];
    w = N[2];

    a = A[0]*u*u + A[1]*v*v + A[2]*w*w + 
        A[3]*u*v + A[4]*u*w + A[5]*v*w;
    b = 2.0*(A[0]*u*x + A[1]*v*y + A[2]*w*z ) +
        A[3]*(u*y+v*x) + A[4]*(u*z + w*x) + A[5]*(v*z + w*y) +
        A[6]*u + A[7]*v + A[8]*w;
    c = A[0]*x*x + A[1]*y*y + A[2]*z*z + 
        A[3]*x*y + A[4]*x*z + A[5]*y*z + A[6]*x + A[7]*y + A[8]*z + A[9];

    r = b*b - 4.0*a*c;
    if( r < 0.0 ) r = 0.0;

    s1 = (-b + sqrt(r)) / (2.0*a);
    s2 = (-b - sqrt(r)) / (2.0*a);

    if( s1*s1 < s2*s2 ) { 
        s = s1;
    } else {
        s = s2;
    }

    Xnew[0] = x + s*u;
    Xnew[1] = y + s*v;
    Xnew[2] = z + s*w;
    m_Transform.fromLocal(Xnew);

    return 0;
}
*/

void
MCConic::MakeMesh(const int nu, const int nv,
                  const double scalefactor, const int flip,
                  PMeshLite &out_mesh) const
{
    // Physical coordinates
    int i, j, n, i1, i2, i3, i4;
    double s, t;
    R3Pt Q;
    R3Vec N, b1, b2;
    Array<Vertex> aovn;
    Array<Vertex> aopVs;
    Array<Face> aopFs;

    n = A.num();
    if( n<10 ) return;

    // Find a set of basis functions in the normal plane
    b1 = R3Vec(1.0, 0.0, 0.0);
    b2 = R3Vec(0.0, 1.0, 0.0);

    out_mesh.Clear();

    /* Generate the vertices */
    for( j=0; j<nv; j++ ) {
        t = scalefactor * m_dDia * (double) (2*j+1-nv) / (double) (nv-1);
        for( i=0; i<nu; i++ ) {
            s = scalefactor * m_dDia * (double) (2*i+1-nu) / (double) (nu-1);
            Q[0] = s*b1[0] + t*b2[0];
            Q[1] = s*b1[1] + t*b2[1];
            Q[2] = s*b1[2] + t*b2[2];
            ProjectLocalZ( Q );
            m_Transform.fromLocal(Q);
            aopVs += out_mesh.AddVertex( Q );
        }
    }

    /* Make the faces */
    for( j=1; j<nv; j++ ) {
        for( i=1; i<nu; i++ ) {
            i1 = i + j*nu;
            i2 = i1 - 1;
            i3 = i2 - nu;
            i4 = i3 + 1;

            if( flip ) {
                aovn += aopVs[i1];
                aovn += aopVs[i3];
                aovn += aopVs[i2];
            } else {
                aovn += aopVs[i1];
                aovn += aopVs[i2];
                aovn += aopVs[i3];
            }
            if( out_mesh.IsOkToAdd( aovn ) == 0 )
                printf("Error adding face\n");
            aopFs += out_mesh.AddFace( aovn );
            aovn.need(0);

            if( flip ) {
                aovn += aopVs[i1];
                aovn += aopVs[i4];
                aovn += aopVs[i3];
            } else {
                aovn += aopVs[i1];
                aovn += aopVs[i3];
                aovn += aopVs[i4];
            }
            if( out_mesh.IsOkToAdd( aovn ) == 0 )
                printf("Error adding face\n");
            aopFs += out_mesh.AddFace( aovn );
            aovn.need(0);
        }
    }
}

void 
MCConic::dump(FILE *out)
{
    int i;

    if( out != NULL ) {
        fprintf(out,"Conic: %d rings\n", m_iRings );
        for( i=0; i<A.num(); i++ ) {
            fprintf(out," A[%d]: %lf\n", i, A[i] );
        }
    }
}
