#include "StdAfx.H"
#include <MC/MC_CurvatureMap.H>
#include <MC/MC_MeshUtils.H>

MCCurvatureMap::MCCurvatureMap()
{
    m_iNumVertices = 0;
    m_iNumData = 0;
    m_iMaxRings = 0;
    m_sMeshDescription = NULL;
}

MCCurvatureMap::~MCCurvatureMap()
{
    if( m_sMeshDescription != NULL ) delete [] m_sMeshDescription;
}
    

void MCCurvatureMap::set_description(const char *desc)
{
    if( m_sMeshDescription != NULL ) delete [] m_sMeshDescription;
    if( desc == NULL ) {
        m_sMeshDescription = NULL;
        return;
    }
    m_sMeshDescription = new char[strlen(desc)+1];
    strcpy(m_sMeshDescription, desc);
}

/*
void MCCurvatureMap::set_dataDescription(const int index, 
                                         const char *desc)
{
    if( index < 0 || index >= m_iNumData ) return;

    if( m_dataDescriptions[index] != NULL ) delete [] m_dataDescriptions[index];
    if( desc == NULL ) {
        m_dataDescriptions[index] = NULL;
        return;
    }
    m_dataDescriptions[index] = new char[strlen(desc)+1];
    strcpy(m_dataDescriptions[index], desc);
}
*/

void MCCurvatureMap::Clear()
{
}

int MCCurvatureMap::Init(const int in_nv, const int in_nd, 
                         const int in_maxRings)
{
    if( m_iNumVertices != 0 || m_iNumData != 0 ) return -1; // Only init once

    m_iNumVertices = in_nv;
    m_iNumData = in_nd;
    m_iMaxRings = in_maxRings;

    m_aMap.init(in_nv);
    m_aRingCount.init(in_nv);
    //m_dataDescriptions.init(in_nd);

    return 0;
}

int MCCurvatureMap::Copy(MCCurvatureMap &cmap)
{
	int ier, i, nv;

	nv = cmap.m_iNumVertices;

	ier = Init(nv, cmap.m_iNumData, cmap.m_iMaxRings);
	if( ier != 0 ) return ier;

	set_description(cmap.Description());

    // Copy distributions
	for( i=0; i<nv; i++ ) {
		ier = m_aMap[i].Copy(cmap.Distribution(i));
		if( ier != 0 ) return ier;

		m_aRingCount[i] = cmap.NumRings(i);
	}
	return 0;
}

int MCCurvatureMap::Extend(MCCurvatureMap &cmap, const int size)
{
	int ier, i, nv;
	MCDistribution flatDist;

	nv = cmap.m_iNumVertices;
	if( size < nv ) return Truncate(cmap, size);
	if( size == nv ) return Copy(cmap);

	ier = Init(size, cmap.m_iNumData, cmap.m_iMaxRings);
	if( ier != 0 ) return ier;

	set_description(cmap.Description());

    // Copy original distributions
	for( i=0; i<nv; i++ ) {
		ier = m_aMap[i].Copy(cmap.Distribution(i));
		if( ier != 0 ) return ier;

		m_aRingCount[i] = cmap.NumRings(i);
	}

	// Create additional (flat) distributions
	flatDist.Copy(cmap.Distribution(0));
	flatDist *= 0.0;

	for( i=nv; i<size; i++ ) {
		ier = m_aMap[i].Copy(flatDist);
		if( ier != 0 ) return ier;

		m_aRingCount[i] = cmap.NumRings(0);
	}

	return 0;
}

int MCCurvatureMap::Truncate(MCCurvatureMap &cmap, const int size)
{
	int ier, i, nv;

	nv = cmap.m_iNumVertices;
	if( size > nv ) return Extend(cmap, size);
	if( size == nv ) return Copy(cmap);

	ier = Init(size, cmap.m_iNumData, cmap.m_iMaxRings);
	if( ier != 0 ) return ier;

	set_description(cmap.Description());

    // Copy original distributions
	for( i=0; i<size; i++ ) {
		ier = m_aMap[i].Copy(cmap.Distribution(i));
		if( ier != 0 ) return ier;

		m_aRingCount[i] = cmap.NumRings(i);
	}

	return 0;
}

int MCCurvatureMap::ExtractSubset(MCCurvatureMap &cmap, Array<int> indices)
{
	int ier, i, j, nv, count;

	int *vflag = NULL;

	nv = cmap.m_iNumVertices;
	if( nv < 3 ) return -1;

	vflag = new int[nv];
	for( i=0; i<nv; i++ ) {
		vflag[i] = -1;
	}

	count = 0;
	for( j=0; j<indices.num(); j++ ) {
		i = indices[j];
		if( i < 0 || i >= nv ) continue;
		if( vflag[i] == -1 ) {
			vflag[i] = count++;
		}
	}

	if( count < 3 ) {
		return -1;
	}

	ier = Init(count, cmap.m_iNumData, cmap.m_iMaxRings);
	if( ier != 0 ) return ier;

	set_description(cmap.Description());

    // Copy distributions
	for( i=0; i<nv; i++ ) {
		j = vflag[i];
		if( j == -1 ) continue;

		ier = m_aMap[j].Copy(cmap.Distribution(i));
		if( ier != 0 ) return ier;

		m_aRingCount[j] = cmap.NumRings(i);
	}

	delete [] vflag;

	return 0;
}

int MCCurvatureMap::ExtractFeature(Array<int> &features, const int index,
		                           MCCurvatureMap &cmap)
{
	int ier, i, j;
	Array<int> indices;

	indices.need(0);
	for( i=0; i<NumVertices(); i++ ) {
		j = features[i];
		if( j == index || j == -index ) indices += i;
	}
	if( indices.num() < 3 ) return -1;

	ier = cmap.ExtractSubset(*this, indices);
	if( ier != 0 ) return ier;

	return 0;
}

int MCCurvatureMap::NumRings(const int index) const
{
    if( index < 0 || index >= m_aRingCount.num() ) return -1;

    return m_aRingCount[index];
}

int MCCurvatureMap::Create(const char *meshfile, 
						   MCData &data, 
                           const int maxRings)
{
    int ier, i, nv, ndata;  
    int nRings = 1;
    //double pi;

    R3Pt P;
    R3Vec N;
    Vertex v;
    Array<double> Areas;
    Array<R3Vec> Normals;
    Array<Vertex> aopVs;

    MCMeshSubset::MC_NormalType normalType = MCMeshSubset::COTANGENT;
    MCMeshSubset::MC_AreaType areaType = MCMeshSubset::MIXED;

    nv = data.Mesh()->NumVertices();
    ndata = 10;

    ier = Init(nv, ndata, maxRings);
    if( ier != 0 ) return -1;

    // Calculate area and normal
    printf("Calculating areas and normals\n");
    data.Mesh()->AllMeshVertices(aopVs);
    MCMeshSubset ring(*(data.Mesh()), nRings);
    
    for( i=0; i<nv; i++ ) {
        v = aopVs[i];
        ring.Create(v,nRings);
        Areas += ring.VertexArea(v, areaType);
        ier = ring.CalculateNormal(v, normalType, N);
        Normals += N;
    }

    // Set description
    set_description(meshfile);
    for( i=0; i<nv; i++ ) {
        m_aMap[i].set_description(meshfile);
    }

	ArrayMesh amesh;
	amesh.Init(*(data.Mesh()));

	int *queue = NULL;
	int *counts = NULL;
	int *flags = NULL;
	int *list = NULL;
	int *kp = NULL;
	int *ki = NULL;
	int *in = NULL;
	int *ij = NULL;
	int *ji = NULL;
	int *jk = NULL;
	float *xc = NULL;

    timer.Reset();
    timer.Start();
    printf("Creating curvature maps\n");
    for( i=0; i<nv; i++ ) {
	//	i = 2136;
        //ier = m_aMap[i].Create(data, amesh, Areas, i, maxRings);
        ier = m_aMap[i].Create(data, amesh, Areas, i, maxRings,
			                   queue, counts, flags, list,
						       kp, ki, in, ij, ji, jk, xc);
        if( ier != 0 ) {
            printf("ERROR Initializing curvature map %d: %d\n", i, ier);
            return ier;
        }
        m_aRingCount[i] = m_aMap[i].Size();

        if( i>0 && (i+1)%100 == 0 ) printf("Completed %d of %d\n", i+1, nv);
    }
    timer.Stop();
    printf("Creating map: Elapsed time = %lf\n", timer.GetElapsedTime());
    
	delete [] queue;
	delete [] counts;
	delete [] flags;
	delete [] list;
	delete [] kp;
	delete [] ki;
	delete [] in;
	delete [] ij;
	delete [] ji;
	delete [] jk;
	delete [] xc;

    return 0;
}

int MCCurvatureMap::Create2(const char *meshfile, 
						   MCData &data, 
                           const int maxRings)
{
    int ier, i, j, k, k1, k2, k3, n, n1, n2, nv, nf, ndata;
	int first, iLast, iNext, nextFace, nRings;
    double a, b, c, aa, bb, cc, d, s, pi, Acum;
	double area, angA, angB, angC, cotA, cotB, cotC, Reff, Rlast, Rcur;
	double KgSum, KgSumSq, KmSum, KmSumSq, AreaSum;
	double avgKg, avgKgSD, avgKm, avgKmSD;
	double kgmin, kgmax, kmmin, kmmax;
    tdgTimer timer;

    R3Pt P;
    R3Vec N;
    MCPoint cpt;
	Array<int> aof;
	Array<int> subsetVertex;
    Array<int> subsetFace;
    Array<int> activeVertex;
    Array<int> activeFace;
    Array<double> aodKg;
    Array<double> aodKm;
    Array<double> aodArea;
	ArrayMesh am;

    printf("Setup and Initializing\n");
    timer.Reset();
    timer.Start();
    
	pi = 4.0*atan(1.0);
	ndata = 10;

	am.Init(*(data.Mesh()));

    nv = am.get_nVertex();
    nf = am.get_nFace();

    ier = Init(nv, ndata, maxRings);
    if( ier != 0 ) return -1;
    set_description(meshfile);

	subsetVertex.need(nv);
    activeVertex.need(nv);
    aodKg.need(nv);
    aodKm.need(nv);
    aodArea.need(nv);
    subsetFace.need(nf);
    activeFace.need(nf);
    
	for( i=0; i<nv; i++ ) {
		ier = data.getPoint(i, cpt);
		if( ier != 0 ) return ier;

		aodKg[i] = cpt.Kg();
		aodKm[i] = cpt.Kbar();
		aodArea[i] = 0.0;
		activeVertex[i] = 0;

		m_aMap[i].set_description(meshfile);
    }

	for( i=0; i<nf; i++ ) {
		activeFace[i] = 0;

		// Calculate area
		ier = am.getFace(i, k1, k2, k3);
		if( ier != 0 ) return ier;
		
		cc = am.DistanceSquared(k1, k2);
		aa = am.DistanceSquared(k2, k3);
		bb = am.DistanceSquared(k3, k1);
		if( aa <= 0.0 || bb <= 0.0 || cc <= 0.0 ) continue;
		a = sqrt(aa);
		b = sqrt(bb);
		c = sqrt(cc);

		if( aa > bb + cc ) {
			s = 0.5*(a+b+c);
			area = sqrt(s*(s-a)*(s-b)*(s-c));
			aodArea[k1] += 0.5*area;
			aodArea[k2] += 0.25*area;
			aodArea[k3] += 0.25*area;
			continue;
		}

		if( bb > aa + cc ) {
			s = 0.5*(a+b+c);
			area = sqrt(s*(s-a)*(s-b)*(s-c));
			aodArea[k1] += 0.25*area;
			aodArea[k2] += 0.5*area;
			aodArea[k3] += 0.25*area;
			continue;
		}

		if( cc > aa + bb ) {
			s = 0.5*(a+b+c);
			area = sqrt(s*(s-a)*(s-b)*(s-c));
			aodArea[k1] += 0.25*area;
			aodArea[k2] += 0.25*area;
			aodArea[k3] += 0.5*area;
			continue;
		}

		d = (cc + bb - aa)/(2.0*c*b);
		angA = (d <= -1.0) ? pi : (d >= 1.0) ? 0.0 : acos(d);

		d = (aa + cc - bb)/(2.0*a*c);
		angB = (d <= -1.0) ? pi : (d >= 1.0) ? 0.0 : acos(d);

		d = (aa + bb - cc)/(2.0*a*b);
		angC = (d <= -1.0) ? pi : (d >= 1.0) ? 0.0 : acos(d);

		d = tan(angA);
		cotA = (d == 0.0) ? 1.0 : 1.0/d;

		d = tan(angB);
		cotB = (d == 0.0) ? 1.0 : 1.0/d;

		d = tan(angC);
		cotC = (d == 0.0) ? 1.0 : 1.0/d;

		aodArea[k1] += 0.125*(bb*cotB + cc*cotC);
		aodArea[k2] += 0.125*(cc*cotC + aa*cotA);
		aodArea[k3] += 0.125*(aa*cotA + bb*cotB);
	}

    timer.Stop();
    printf("Initialized: Elapsed time = %lf\n", timer.GetElapsedTime());

    printf("Creating curvature maps\n");
    timer.Reset();
    timer.Start();

    int pct;
    int lastPct = 0;
	int iDebug = 317;

    m_iMaxRings = 0;
    for( i=0; i<nv; i++ ) {
		if( i == iDebug ) {
			ier++;
		}
		iLast = -1;
		iNext = 0;
		nextFace = 0;
		nRings = 0;
		Acum = aodArea[i];
		Rcur = sqrt(Acum/pi);
		Reff = 0.0;
		subsetVertex[++iLast] = i;
		activeVertex[i] = 1;
		ier = m_aMap[i].AddPoint(aodKg[i], aodKm[i], 0.0, 0.0, aodArea[i], Reff, 
			                     aodKg[i], aodKg[i], aodKm[i], aodKm[i]);
		if( ier != 0 ) return ier;

		// Process a ring at a time
		while( iNext <= iLast && nRings < maxRings )
		{
			nRings++;
			KgSum = 0.0;
			KgSumSq = 0.0;
			KmSum = 0.0;
			KmSumSq = 0.0;
			AreaSum = 0.0;

			n1 = iNext;
			n2 = iLast;
			first = 1;
			for( j=n1; j<=n2; j++ ) {
				ier = am.FindAdjacentFace(subsetVertex[j], aof);
				if( ier != 0 ) continue;

				for( k=0; k<aof.num(); k++ ) {
					if( activeFace[aof[k]] ) continue;
					subsetFace[nextFace++] = aof[k];
					activeFace[aof[k]] = 1;
					ier = am.getFace(aof[k], k1, k2, k3);
					if( ier != 0 ) return ier;

					if( activeVertex[k1] == 0 ) {
						subsetVertex[++iLast] = k1;
						activeVertex[k1] = 1;
						KgSum += aodKg[k1];
						KgSumSq += aodKg[k1]*aodKg[k1];
						KmSum += aodKm[k1];
						KmSumSq += aodKm[k1]*aodKm[k1];
						AreaSum += aodArea[k1];
						if( first ) {
							kgmin = aodKg[k1];
							kgmax = aodKg[k1];
							kmmin = aodKm[k1];
							kmmax = aodKm[k1];
							first = 0;
						}
						if( aodKg[k1] < kgmin ) kgmin = aodKg[k1];
						if( aodKg[k1] > kgmax ) kgmax = aodKg[k1];
						if( aodKm[k1] < kmmin ) kmmin = aodKm[k1];
						if( aodKm[k1] > kmmax ) kmmax = aodKm[k1];
					}
					if( activeVertex[k2] == 0 ) {
						subsetVertex[++iLast] = k2;
						activeVertex[k2] = 1;
						KgSum += aodKg[k2];
						KgSumSq += aodKg[k2]*aodKg[k2];
						KmSum += aodKm[k2];
						KmSumSq += aodKm[k2]*aodKm[k2];
						AreaSum += aodArea[k2];
						if( first ) {
							kgmin = aodKg[k2];
							kgmax = aodKg[k2];
							kmmin = aodKm[k2];
							kmmax = aodKm[k2];
							first = 0;
						}
						if( aodKg[k2] < kgmin ) kgmin = aodKg[k2];
						if( aodKg[k2] > kgmax ) kgmax = aodKg[k2];
						if( aodKm[k2] < kmmin ) kmmin = aodKm[k2];
						if( aodKm[k2] > kmmax ) kmmax = aodKm[k2];
					}
					if( activeVertex[k3] == 0 ) {
						subsetVertex[++iLast] = k3;
						activeVertex[k3] = 1;
						KgSum += aodKg[k3];
						KgSumSq += aodKg[k3]*aodKg[k3];
						KmSum += aodKm[k3];
						KmSumSq += aodKm[k3]*aodKm[k3];
						AreaSum += aodArea[k3];
						if( first ) {
							kgmin = aodKg[k3];
							kgmax = aodKg[k3];
							kmmin = aodKm[k3];
							kmmax = aodKm[k3];
							first = 0;
						}
						if( aodKg[k3] < kgmin ) kgmin = aodKg[k3];
						if( aodKg[k3] > kgmax ) kgmax = aodKg[k3];
						if( aodKm[k3] < kmmin ) kmmin = aodKm[k3];
						if( aodKm[k3] > kmmax ) kmmax = aodKm[k3];
					}
				}
			}
			n = iLast - n2;
			iNext = n2+1;
			avgKg = KgSum / (double) n;
			avgKgSD = KgSumSq / (double) n - avgKg*avgKg;
			avgKgSD = (avgKgSD<0.0) ? 0.0 : avgKgSD;
			avgKm = KmSum / (double) n;
			avgKmSD = KmSumSq / (double) n - avgKm*avgKm;
			avgKmSD = (avgKmSD<0.0) ? 0.0 : avgKmSD;
			Acum += AreaSum;
			Rlast = Rcur;
			Rcur = sqrt(Acum/pi);
			Reff = 0.5*(Rlast + Rcur);
			ier = m_aMap[i].AddPoint(avgKg, avgKm, sqrt(avgKgSD), sqrt(avgKmSD), AreaSum, 
				                     Reff, kgmax, kgmin, kmmax, kmmin);
			if( ier != 0 ) return ier;
		}
		m_aMap[i].SetPointIndex(i);
        m_aRingCount[i] = m_aMap[i].Size();
        if( m_aRingCount[i] > m_iMaxRings ) m_iMaxRings = m_aRingCount[i];

		// Clear the active lists
		for( j=0; j<=iLast; j++ ) {
			activeVertex[subsetVertex[j]] = 0;
		}
		for( j=0; j<nextFace; j++ ) {
			activeFace[subsetFace[j]] = 0;
		}

        pct = (int) ((double) (100*(i+1)) / (double) nv);
        if( pct > lastPct && ( pct <= 5 || (pct%10) == 0 ) ) {
            printf("%d pct complete\n", pct);
        }
        lastPct = pct;
    }
    timer.Stop();
    printf("Creating map: Elapsed time = %lf\n", timer.GetElapsedTime());
    
    return 0;
}

int MCCurvatureMap::Create(const char *meshfile, 
                           MCData &data, 
                           const double maxRadius)
{
    int ier, i, nv, ndata;  
    int nRings = 1;
    int maxRings = 1;
    //double pi;
    tdgTimer timer;

    R3Pt P;
    R3Vec N;
    Vertex v;
    Array<double> Areas;
    Array<R3Vec> Normals;
    Array<Vertex> aopVs;

    MCMeshSubset::MC_NormalType normalType = MCMeshSubset::COTANGENT;
    MCMeshSubset::MC_AreaType areaType = MCMeshSubset::MIXED;

    nv = data.Mesh()->NumVertices();
    ndata = 10;

    ier = Init(nv, ndata, maxRings);
    if( ier != 0 ) return -1;

    timer.Reset();
    timer.Start();
    // Calculate area and normal
    printf("Calculating areas and normals\n");
    data.Mesh()->AllMeshVertices(aopVs);
    MCMeshSubset ring(*(data.Mesh()), nRings);
    
    for( i=0; i<nv; i++ ) {
        v = aopVs[i];
        ring.Create(v,nRings);
        Areas += ring.VertexArea(v, areaType);
        ier = ring.CalculateNormal(v, normalType, N);
        Normals += N;
    }
    timer.Stop();
    printf("Areas and normals: Elapsed time = %lf\n", timer.GetElapsedTime());

    // Set description
    set_description(meshfile);
    for( i=0; i<nv; i++ ) {
        m_aMap[i].set_description(meshfile);
    }

    ArrayMesh amesh;
    amesh.Init(*(data.Mesh()));

    int *queue = NULL;
    int *counts = NULL;
    int *flags = NULL;
    int *list = NULL;
    int *kp = NULL;
    int *ki = NULL;
    int *in = NULL;
    int *ij = NULL;
    int *ji = NULL;
    int *jk = NULL;
    float *xc = NULL;

    timer.Reset();
    timer.Start();
    printf("Creating curvature maps\n");
    m_iMaxRings = 0;
    for( i=0; i<nv; i++ ) {
    //  i = 2136;
        //ier = m_aMap[i].Create(data, amesh, Areas, i, maxRings);
        //ier = m_aMap[i].Create(data, amesh, Areas, i, maxRings,
        //                     queue, counts, flags, list,
        //                     kp, ki, in, ij, ji, jk, xc);
        ier = m_aMap[i].Create(data, amesh, Areas, i, maxRadius,
                               queue, counts, flags, list,
                               kp, ki, in, ij, ji, jk, xc);
        if( ier != 0 ) {
            printf("ERROR Initializing curvature map %d: %d\n", i, ier);
            return ier;
        }
        m_aRingCount[i] = m_aMap[i].Size();
        if( m_aRingCount[i] > m_iMaxRings ) m_iMaxRings = m_aRingCount[i];

        if( i>0 && (i+1)%100 == 0 ) printf("Completed %d of %d\n", i+1, nv);
    }
    timer.Stop();
    printf("Creating map: Elapsed time = %lf\n", timer.GetElapsedTime());
    
    delete [] queue;
    delete [] counts;
    delete [] flags;
    delete [] list;
    delete [] kp;
    delete [] ki;
    delete [] in;
    delete [] ij;
    delete [] ji;
    delete [] jk;
    delete [] xc;

    return 0;
}

int MCCurvatureMap::Create2(const char *meshfile, 
                            MCData &data, 
                            const double maxRadius)
{
    int ier, i, j, k, k1, k2, k3, n, n1, n2, nv, nf, ndata;
	int first, iLast, iNext, nextFace, nRings;
    double a, b, c, aa, bb, cc, d, s, pi, Acum;
	double area, angA, angB, angC, cotA, cotB, cotC, Reff, Rlast, Rcur;
	double KgSum, KgSumSq, KmSum, KmSumSq, AreaSum;
	double avgKg, avgKgSD, avgKm, avgKmSD;
	double kgmin, kgmax, kmmin, kmmax;
    tdgTimer timer;

    R3Pt P;
    R3Vec N;
    MCPoint cpt;
	Array<int> aof;
	Array<int> subsetVertex;
    Array<int> subsetFace;
    Array<int> activeVertex;
    Array<int> activeFace;
    Array<double> aodKg;
    Array<double> aodKm;
    Array<double> aodArea;
	ArrayMesh am;

    printf("Setup and Initializing\n");
    timer.Reset();
    timer.Start();
    
	pi = 4.0*atan(1.0);
	ndata = 10;

    int maxRings = 1;

	am.Init(*(data.Mesh()));

    nv = am.get_nVertex();
    nf = am.get_nFace();

    ier = Init(nv, ndata, maxRings);
    if( ier != 0 ) return -1;
    set_description(meshfile);

	subsetVertex.need(nv);
    activeVertex.need(nv);
    aodKg.need(nv);
    aodKm.need(nv);
    aodArea.need(nv);
    subsetFace.need(nf);
    activeFace.need(nf);
    
	for( i=0; i<nv; i++ ) {
		ier = data.getPoint(i, cpt);
		if( ier != 0 ) return ier;

		aodKg[i] = cpt.Kg();
		aodKm[i] = cpt.Kbar();
		aodArea[i] = 0.0;
		activeVertex[i] = 0;

		m_aMap[i].set_description(meshfile);
    }

	for( i=0; i<nf; i++ ) {
		activeFace[i] = 0;

		// Calculate area
		ier = am.getFace(i, k1, k2, k3);
		if( ier != 0 ) return ier;
		
		cc = am.DistanceSquared(k1, k2);
		aa = am.DistanceSquared(k2, k3);
		bb = am.DistanceSquared(k3, k1);
		if( aa <= 0.0 || bb <= 0.0 || cc <= 0.0 ) continue;
		a = sqrt(aa);
		b = sqrt(bb);
		c = sqrt(cc);

		if( aa > bb + cc ) {
			s = 0.5*(a+b+c);
			area = sqrt(s*(s-a)*(s-b)*(s-c));
			aodArea[k1] += 0.5*area;
			aodArea[k2] += 0.25*area;
			aodArea[k3] += 0.25*area;
			continue;
		}

		if( bb > aa + cc ) {
			s = 0.5*(a+b+c);
			area = sqrt(s*(s-a)*(s-b)*(s-c));
			aodArea[k1] += 0.25*area;
			aodArea[k2] += 0.5*area;
			aodArea[k3] += 0.25*area;
			continue;
		}

		if( cc > aa + bb ) {
			s = 0.5*(a+b+c);
			area = sqrt(s*(s-a)*(s-b)*(s-c));
			aodArea[k1] += 0.25*area;
			aodArea[k2] += 0.25*area;
			aodArea[k3] += 0.5*area;
			continue;
		}

		d = (cc + bb - aa)/(2.0*c*b);
		angA = (d <= -1.0) ? pi : (d >= 1.0) ? 0.0 : acos(d);

		d = (aa + cc - bb)/(2.0*a*c);
		angB = (d <= -1.0) ? pi : (d >= 1.0) ? 0.0 : acos(d);

		d = (aa + bb - cc)/(2.0*a*b);
		angC = (d <= -1.0) ? pi : (d >= 1.0) ? 0.0 : acos(d);

		d = tan(angA);
		cotA = (d == 0.0) ? 1.0 : 1.0/d;

		d = tan(angB);
		cotB = (d == 0.0) ? 1.0 : 1.0/d;

		d = tan(angC);
		cotC = (d == 0.0) ? 1.0 : 1.0/d;

		aodArea[k1] += 0.125*(bb*cotB + cc*cotC);
		aodArea[k2] += 0.125*(cc*cotC + aa*cotA);
		aodArea[k3] += 0.125*(aa*cotA + bb*cotB);
	}

    timer.Stop();
    printf("Initialized: Elapsed time = %lf\n", timer.GetElapsedTime());

    printf("Creating curvature maps\n");
    timer.Reset();
    timer.Start();

    int pct;
    int lastPct = 0;
	int iDebug = 317;

    m_iMaxRings = 0;
    for( i=0; i<nv; i++ ) {
		if( i == iDebug ) {
			ier++;
		}
		iLast = -1;
		iNext = 0;
		nextFace = 0;
		nRings = 0;
		Acum = aodArea[i];
		Rcur = sqrt(Acum/pi);
		Reff = 0.0;
		subsetVertex[++iLast] = i;
		activeVertex[i] = 1;
		ier = m_aMap[i].AddPoint(aodKg[i], aodKm[i], 0.0, 0.0, aodArea[i], Reff, 
			                     aodKg[i], aodKg[i], aodKm[i], aodKm[i]);
		if( ier != 0 ) return ier;

		// Process a ring at a time
		while( iNext <= iLast && Reff < maxRadius )
		{
			nRings++;
			KgSum = 0.0;
			KgSumSq = 0.0;
			KmSum = 0.0;
			KmSumSq = 0.0;
			AreaSum = 0.0;

			n1 = iNext;
			n2 = iLast;
			first = 1;
			for( j=n1; j<=n2; j++ ) {
				ier = am.FindAdjacentFace(subsetVertex[j], aof);
				if( ier != 0 ) continue;

				for( k=0; k<aof.num(); k++ ) {
					if( activeFace[aof[k]] ) continue;
					subsetFace[nextFace++] = aof[k];
					activeFace[aof[k]] = 1;
					ier = am.getFace(aof[k], k1, k2, k3);
					if( ier != 0 ) return ier;

					if( activeVertex[k1] == 0 ) {
						subsetVertex[++iLast] = k1;
						activeVertex[k1] = 1;
						KgSum += aodKg[k1];
						KgSumSq += aodKg[k1]*aodKg[k1];
						KmSum += aodKm[k1];
						KmSumSq += aodKm[k1]*aodKm[k1];
						AreaSum += aodArea[k1];
						if( first ) {
							kgmin = aodKg[k1];
							kgmax = aodKg[k1];
							kmmin = aodKm[k1];
							kmmax = aodKm[k1];
							first = 0;
						}
						if( aodKg[k1] < kgmin ) kgmin = aodKg[k1];
						if( aodKg[k1] > kgmax ) kgmax = aodKg[k1];
						if( aodKm[k1] < kmmin ) kmmin = aodKm[k1];
						if( aodKm[k1] > kmmax ) kmmax = aodKm[k1];
					}
					if( activeVertex[k2] == 0 ) {
						subsetVertex[++iLast] = k2;
						activeVertex[k2] = 1;
						KgSum += aodKg[k2];
						KgSumSq += aodKg[k2]*aodKg[k2];
						KmSum += aodKm[k2];
						KmSumSq += aodKm[k2]*aodKm[k2];
						AreaSum += aodArea[k2];
						if( first ) {
							kgmin = aodKg[k2];
							kgmax = aodKg[k2];
							kmmin = aodKm[k2];
							kmmax = aodKm[k2];
							first = 0;
						}
						if( aodKg[k2] < kgmin ) kgmin = aodKg[k2];
						if( aodKg[k2] > kgmax ) kgmax = aodKg[k2];
						if( aodKm[k2] < kmmin ) kmmin = aodKm[k2];
						if( aodKm[k2] > kmmax ) kmmax = aodKm[k2];
					}
					if( activeVertex[k3] == 0 ) {
						subsetVertex[++iLast] = k3;
						activeVertex[k3] = 1;
						KgSum += aodKg[k3];
						KgSumSq += aodKg[k3]*aodKg[k3];
						KmSum += aodKm[k3];
						KmSumSq += aodKm[k3]*aodKm[k3];
						AreaSum += aodArea[k3];
						if( first ) {
							kgmin = aodKg[k3];
							kgmax = aodKg[k3];
							kmmin = aodKm[k3];
							kmmax = aodKm[k3];
							first = 0;
						}
						if( aodKg[k3] < kgmin ) kgmin = aodKg[k3];
						if( aodKg[k3] > kgmax ) kgmax = aodKg[k3];
						if( aodKm[k3] < kmmin ) kmmin = aodKm[k3];
						if( aodKm[k3] > kmmax ) kmmax = aodKm[k3];
					}
				}
			}
			n = iLast - n2;
			iNext = n2+1;
			avgKg = KgSum / (double) n;
			avgKgSD = KgSumSq / (double) n - avgKg*avgKg;
			avgKgSD = (avgKgSD<0.0) ? 0.0 : avgKgSD;
			avgKm = KmSum / (double) n;
			avgKmSD = KmSumSq / (double) n - avgKm*avgKm;
			avgKmSD = (avgKmSD<0.0) ? 0.0 : avgKmSD;
			Acum += AreaSum;
			Rlast = Rcur;
			Rcur = sqrt(Acum/pi);
			Reff = 0.5*(Rlast + Rcur);
			ier = m_aMap[i].AddPoint(avgKg, avgKm, sqrt(avgKgSD), sqrt(avgKmSD), AreaSum, 
				                     Reff, kgmax, kgmin, kmmax, kmmin);
			if( ier != 0 ) return ier;
		}
		m_aMap[i].SetPointIndex(i);
        m_aRingCount[i] = m_aMap[i].Size();
        if( m_aRingCount[i] > m_iMaxRings ) m_iMaxRings = m_aRingCount[i];

		// Clear the active lists
		for( j=0; j<=iLast; j++ ) {
			activeVertex[subsetVertex[j]] = 0;
		}
		for( j=0; j<nextFace; j++ ) {
			activeFace[subsetFace[j]] = 0;
		}

        pct = (int) ((double) (100*(i+1)) / (double) nv);
        if( pct > lastPct && ( pct <= 5 || (pct%10) == 0 ) ) {
            printf("%d pct complete\n", pct);
        }
        lastPct = pct;
    }
    timer.Stop();
    printf("Creating map: Elapsed time = %lf\n", timer.GetElapsedTime());
    
    return 0;
}

int MCCurvatureMap::Create(const char *meshfile, 
                           const char *fanDataFile, 
                           const int maxRings, 
						   const int isBinary, const int verbose)
{
    int ier, i, nv, ndata, printInterval;  
    int nRings = 1;
	//printInterval = 100;
	printInterval = 1000;

	int test = 0;
    /*
	R3Pt P;
    R3Vec N;
    Vertex v;
    Array<double> Areas;
    Array<R3Vec> Normals;
    Array<Vertex> aopVs;
	*/
	ArrayMesh mesh;
	Array<int> spokeSize;
	Array<R3Pt> xyz;
	Array<double> K1;
	Array<double> K2;
	Array<int> valid;

    MCMeshSubset::MC_NormalType normalType = MCMeshSubset::COTANGENT;
    MCMeshSubset::MC_AreaType areaType = MCMeshSubset::MIXED;

	// Read mesh (just to get number of vertices)
	mesh.ReadMesh(meshfile);
	nv = mesh.get_nVertex();

    //nv = data.Mesh()->NumVertices();
    ndata = 10;

    ier = Init(nv, ndata, maxRings);
    if( ier != 0 ) return -1;

	/*
    // Calculate area and normal
    printf("Calculating areas and normals\n");
    data.Mesh()->AllMeshVertices(aopVs);
    MCMeshSubset ring(*(data.Mesh()), nRings);
    
    for( i=0; i<nv; i++ ) {
        v = aopVs[i];
        ring.Create(v,nRings);
        Areas += ring.VertexArea(v, areaType);
        ier = ring.CalculateNormal(v, normalType, N);
        Normals += N;
    }
	*/

    // Set description
    set_description(meshfile);

	// Read the fan data file

	FILE *in = NULL;

	if( isBinary ) {
		in = fopen(fanDataFile,"rb");
	} else {
		in = fopen(fanDataFile,"r");
	}
	if( in == NULL ) return -1;

	char junkFile[132];
	strcpy(junkFile,fanDataFile);
	strcat(junkFile,".copy");

	FILE *out = NULL;
	out = fopen(junkFile,"w");

	int nd = 5;

	printf("Creating curvature maps\n");
    for( i=0; i<nv; i++ ) {
		//if( i == 6199 ) {
		//	test = 1;
		//}
		if( isBinary ) {
			ier = MCReadFanDataBinary(in, spokeSize, xyz, K1, K2, valid);
		} else {
			ier = MCReadFanData(in, spokeSize, xyz, K1, K2, valid);
		}
		if( ier != 0 ) return ier;

		if( out != NULL ) {
			ier = MCWriteFan(out, spokeSize, xyz, K1, K2, valid);
		}

        ier = m_aMap[i].CreateFromFan(spokeSize, xyz, K1, K2, valid,
                                      i, maxRings);
        if( ier != 0 ) {
            printf("ERROR creating fan %d: %d\n", i, ier);
            return ier;
        }

        m_aRingCount[i] = m_aMap[i].Size();
		m_aMap[i].set_description(meshfile);

        if( i>0 && (i+1)%printInterval == 0 ) printf("Completed %d of %d\n", i+1, nv);
    }
    
    return 0;
}

int MCCurvatureMap::WriteBinary(const char *outfile) const
{
    int ier, i, j, printInterval;
    double dbuf[20];
    int *count = NULL;
    char names[80];

	printInterval = 1000;

    // Check for valid curvature map
    if( m_iNumVertices == 0 || m_iNumData == 0 || m_iMaxRings == 0 ) return -2;
    if( m_aMap.num() != m_iNumVertices ) return -2;
    if( m_aRingCount.num() != m_iNumVertices ) return -2;

    // Open the file
    FILE *out = NULL;
    out = fopen(outfile,"wb");
    if( out == NULL ) return -1;

    // Write the map header info
    //dbuf[0] = (double) m_iNumVertices;
    //dbuf[1] = (double) m_iNumData;
    //dbuf[2] = (double) m_iMaxRings;
    //ier = fwrite(&dbuf[0], sizeof(double), 3, out);
    ier = fwrite(&m_iNumVertices, sizeof(int), 1, out);
    ier = fwrite(&m_iNumData, sizeof(int), 1, out);
    ier = fwrite(&m_iMaxRings, sizeof(int), 1, out);

    // Write the ring counts
    count = new int[m_aRingCount.num()];
    for( i=0; i<m_aRingCount.num(); i++ ) {
        count[i] = m_aRingCount[i];
    }
    ier = fwrite(count, sizeof(int), m_aRingCount.num(), out);
    delete [] count;

    // Write the curvature data
    for( i=0; i<m_iNumVertices; i++ ) {
        for( j=0; j<m_aRingCount[i]; j++ ) {
            dbuf[0] = (double) i;
            dbuf[1] = (double) j;
            ier = Value(MCDistribution::AREA   , i, j, dbuf[ 2]);
            ier = Value(MCDistribution::RADIUS , i, j, dbuf[ 3]);
            ier = Value(MCDistribution::KG     , i, j, dbuf[ 4]);
            ier = Value(MCDistribution::KBAR   , i, j, dbuf[ 5]);
            ier = Value(MCDistribution::KGSD   , i, j, dbuf[ 6]);
            ier = Value(MCDistribution::KBARSD , i, j, dbuf[ 7]);
            ier = Value(MCDistribution::KGMIN  , i, j, dbuf[ 8]);
            ier = Value(MCDistribution::KGMAX  , i, j, dbuf[ 9]);
            ier = Value(MCDistribution::KBARMIN, i, j, dbuf[10]);
            ier = Value(MCDistribution::KBARMAX, i, j, dbuf[11]);
            ier = fwrite(&dbuf[0], sizeof(double), m_iNumData+2, out);
        }
        if( i>0 && (i+1)%printInterval == 0 ) {
            printf("Completed %d of %d\n", i+1, m_iNumVertices);
        }
    }

    // Write the description
	if( m_sMeshDescription != NULL ) {
		strncpy(names,m_sMeshDescription,79);
	} else {
		strncpy(names,"Unknown",79);
	}
    ier = fwrite(&names[0],sizeof(char), 80, out);

    fclose(out);

    return 0;
}

int MCCurvatureMap::Write(const char *outfile) const
{
    int ier, i, printInterval;

	printInterval = 1000;

    // Check for valid curvature map
    if( m_iNumVertices == 0 || m_iNumData == 0 || m_iMaxRings == 0 ) return -2;
    if( m_aMap.num() != m_iNumVertices ) return -2;
    if( m_aRingCount.num() != m_iNumVertices ) return -2;

    // Open the file
    FILE *out = NULL;
    out = fopen(outfile,"w");
    if( out == NULL ) return -1;

    // Write the curvature data
    for( i=0; i<m_iNumVertices; i++ ) {
        ier = m_aMap[i].WriteMap(out);
        if( ier != 0 ) return ier;

        if( i>0 && (i+1)%printInterval == 0 ) {
            printf("Completed %d of %d\n", i+1, m_iNumVertices);
        }
    }

    fclose(out);

    return 0;
}

int MCCurvatureMap::ReadBinary(const char *infile)
{
    int ier, i, j, nv, nd, mr;
    double dbuf[20];
    int *count = NULL;
    char names[80];

    // Open the file
    FILE *in = NULL;
    in = fopen(infile,"rb");
    if( in == NULL ) return -1;

    // Read the map header info
    ier = fread(&nv, sizeof(int), 1, in);
    ier = fread(&nd, sizeof(int), 1, in);
    ier = fread(&mr, sizeof(int), 1, in);
    //ier = fread(&dbuf[0], sizeof(double), 3, in);
    //nv = (int) dbuf[0];
    //nd = (int) dbuf[1];
    //mr = (int) dbuf[2];

    // Initialize the curvature map
    ier = Init(nv, nd, mr);
    if( ier != 0 ) return -1;

    // Read the ring counts
    count = new int[m_iNumVertices];
    ier = fread(count, sizeof(int), m_iNumVertices, in);
    for( i=0; i<m_iNumVertices; i++ ) {
        m_aRingCount[i] = count[i];
    }
    delete [] count;

    // Read the curvature data
    for( i=0; i<m_iNumVertices; i++ ) {
        m_aMap[i].Clear();
        for( j=0; j<m_aRingCount[i]; j++ ) {
            ier = fread(&dbuf[0], sizeof(double), m_iNumData+2, in);
            ier = m_aMap[i].AddPoint(i, j, &dbuf[2], m_iNumData);
            /*
            dbuf[0] = (double) i;
            dbuf[1] = (double) j;
            ier = Value(MCDistribution::MC_AREA   , index, j, dbuf[ 2]);
            ier = Value(MCDistribution::MC_RADIUS , index, j, dbuf[ 3]);
            ier = Value(MCDistribution::MC_KG     , index, j, dbuf[ 4]);
            ier = Value(MCDistribution::MC_KBAR   , index, j, dbuf[ 5]);
            ier = Value(MCDistribution::MC_KGSD   , index, j, dbuf[ 6]);
            ier = Value(MCDistribution::MC_KBARSD , index, j, dbuf[ 7]);
            ier = Value(MCDistribution::MC_KGMIN  , index, j, dbuf[ 8]);
            ier = Value(MCDistribution::MC_KGMAX  , index, j, dbuf[ 9]);
            ier = Value(MCDistribution::MC_KBARMIN, index, j, dbuf[10]);
            ier = Value(MCDistribution::MC_KBARMAX, index, j, dbuf[11]);
            */
        }
    }

    // Read the description
    ier = fread(&names[0],sizeof(char), 80, in);
    delete [] m_sMeshDescription;
    m_sMeshDescription = new char[strlen(names)+1];
    strcpy(m_sMeshDescription,names);

    fclose(in);

    return 0;
}

int MCCurvatureMap::Read(const char *infile)
{
    return -1;

    int ier, i, j, nv, nd, mr;
    double dbuf[20];
    int *count = NULL;
    char names[80];

    // Open the file
    FILE *in = NULL;
    in = fopen(infile,"r");
    if( in == NULL ) return -1;

    // Read the map header info
    fread(&dbuf[0], sizeof(double), 3, in);
    nv = (int) dbuf[0];
    nd = (int) dbuf[1];
    mr = (int) dbuf[2];

    // Initialize the curvature map
    ier = Init(nv, nd, mr);
    if( ier != 0 ) return -1;

    // Read the ring counts
    count = new int[m_iNumVertices];
    fread(count, sizeof(int), m_iNumVertices, in);
    for( i=0; i<m_iNumVertices; i++ ) {
        m_aRingCount[i] = count[i];
    }
    delete [] count;

    // Read the curvature data
    for( i=0; i<m_iNumVertices; i++ ) {
        m_aMap[i].Clear();
        for( j=0; j<m_aRingCount[i]; j++ ) {
            fread(&dbuf[0], sizeof(double), m_iNumData+2, in);
            m_aMap[i].AddPoint(i, j, &dbuf[2], m_iNumData);
            /*
            dbuf[0] = (double) i;
            dbuf[1] = (double) j;
            ier = Value(MCDistribution::MC_AREA   , index, j, dbuf[ 2]);
            ier = Value(MCDistribution::MC_RADIUS , index, j, dbuf[ 3]);
            ier = Value(MCDistribution::MC_KG     , index, j, dbuf[ 4]);
            ier = Value(MCDistribution::MC_KBAR   , index, j, dbuf[ 5]);
            ier = Value(MCDistribution::MC_KGSD   , index, j, dbuf[ 6]);
            ier = Value(MCDistribution::MC_KBARSD , index, j, dbuf[ 7]);
            ier = Value(MCDistribution::MC_KGMIN  , index, j, dbuf[ 8]);
            ier = Value(MCDistribution::MC_KGMAX  , index, j, dbuf[ 9]);
            ier = Value(MCDistribution::MC_KBARMIN, index, j, dbuf[10]);
            ier = Value(MCDistribution::MC_KBARMAX, index, j, dbuf[11]);
            */
        }
    }

    // Read the description
    fread(&names[0],sizeof(char), 80, in);
    delete [] m_sMeshDescription;
    m_sMeshDescription = new char[strlen(names)+1];
    strcpy(m_sMeshDescription,names);

    fclose(in);

    return 0;
}

const MCDistribution& MCCurvatureMap::Distribution(const int index) const
{
    ASSERT(index >= 0 && index < m_aMap.num());
    return m_aMap[index];
}

int MCCurvatureMap::Replace(const int index, 
							const MCDistribution &dist)
{
	int nring;

	if( index < 0 || index >= m_iNumVertices ) return -1;

	m_aMap[index].Copy(dist);
	m_aMap[index].SetPointIndex(index);

	nring = dist.Size();
	m_aRingCount[index] = nring;
	if( nring > m_iMaxRings ) m_iMaxRings = nring;

	return 0;
}

int MCCurvatureMap::Value(MCDistribution::MC_DataType dataType, const int index, 
                          const int iRing, double &value ) const
{
    int ier;

    if( index < 0 || index >= m_iNumVertices ) return -1;

    ier = m_aMap[index].Value(dataType, iRing, value);

    return ier;
}

int MCCurvatureMap::CreateCurve(const int iPoint,
								MCDistribution::MC_DataType indep, 
								MCDistribution::MC_DataType dep,
								const int maxPoints, 
								tdgCurve2D &curve ) const
{
    int ier, i, n;
	double x, y;

	if( iPoint < 0 || iPoint >= m_iNumVertices ) return -1;

	curve.Clear();

	n = m_aMap[iPoint].Size();
	if( n > maxPoints ) n = maxPoints;

    for( i=0; i<n; i++ ) {
		ier = m_aMap[iPoint].Value(indep, i, x);
		if( ier != 0 ) return ier;
		ier = m_aMap[iPoint].Value(dep, i, y);
		if( ier != 0 ) return ier;
		curve.Add(x,y);
	}
	curve.Label(Description());
	curve.Id(iPoint);
	curve.Index(dep);

    return 0;
}

int MCCurvatureMap::InterpolateBary(const int i1, const double b1, 
		                            const int i2, const double b2, 
		                            const int i3, const double b3, 
		                            double &kg, double &kbar)
{
	int ier, iRing;
	double value;

	if( i1 < 0 || i1 >= m_iNumVertices ||
		i2 < 0 || i2 >= m_iNumVertices ||
		i3 < 0 || i3 >= m_iNumVertices ) return -1;

	iRing = 0;
	kg = 0.0;
	kbar = 0.0;

	ier = m_aMap[i1].Value(MCDistribution::KG, iRing, value);
	if( ier != 0 ) return ier;
	kg += value * b1;

	ier = m_aMap[i2].Value(MCDistribution::KG, iRing, value);
	if( ier != 0 ) return ier;
	kg += value * b2;

	ier = m_aMap[i3].Value(MCDistribution::KG, iRing, value);
	if( ier != 0 ) return ier;
	kg += value * b3;

	ier = m_aMap[i1].Value(MCDistribution::KBAR, iRing, value);
	if( ier != 0 ) return ier;
	kbar += value * b1;

	ier = m_aMap[i2].Value(MCDistribution::KBAR, iRing, value);
	if( ier != 0 ) return ier;
	kbar += value * b2;

	ier = m_aMap[i3].Value(MCDistribution::KBAR, iRing, value);
	if( ier != 0 ) return ier;
	kbar += value * b3;

	return 0;
}

int MCCurvatureMap::Smooth(const int nTerm)
{
	int ier, i;

	for( i=0; i<m_iNumVertices; i++ ) {
		ier = m_aMap[i].Smooth(nTerm);
		if( ier != 0 ) return ier;
	}

	return 0;
}

int MCCurvatureMap::RadiusStats(double &Rmin, double &Rmax, 
								double &Rmean, double &Rstddev, 
								double &Rmedian)
{
	int ier, i, nRing;
	double r, sum, sum2;
	Array<double> aodR;
	Array<int> sortOrder;
	
	if( m_iNumVertices < 3 ) return -1;

	aodR.need(0);
	sum = 0.0;
	sum2 = 0.0;
	for( i=0; i<m_iNumVertices; i++ ) {
		nRing = m_aRingCount[i];
		ier = Value(MCDistribution::RADIUS, i, nRing-1, r);
		if( ier != 0 ) {
			return ier;
		}
		aodR += r;
		sum += r;
		sum2 += r*r;
	}
	sortOrder.sort(aodR);

	Rmin = aodR[m_iNumVertices-1];
	Rmax = aodR[0];
	Rmedian = aodR[m_iNumVertices/2];
	Rmean = sum / (double) m_iNumVertices;
	Rstddev = sqrt(sum2 / (double) m_iNumVertices - Rmean*Rmean);

	return 0;
}
int MCCurvatureMap::Filter(const double sigma)
{
	int ier, i;

	for( i=0; i<m_iNumVertices; i++ ) {
		ier = m_aMap[i].Filter(sigma);
		if( ier != 0 ) return ier;
	}

	return 0;
}

int MCCurvatureMap::Filter(const int npts)
{
	int ier, i, count;
	double sigma, sum;

	count = 0;
	sum = 0.0;
	for( i=0; i<m_iNumVertices; i++ ) {
		count += m_aMap[i].Size()-1;
		sum += m_aMap[i].Radius();
	}
	if( count < 1 ) return -1;
	sigma = npts * sum / count;

	for( i=0; i<m_iNumVertices; i++ ) {
		ier = m_aMap[i].Filter(sigma);
		if( ier != 0 ) return ier;
	}

	return 0;
}

int MCCurvatureMap::ComputeFeatureDistributions(const int type, 
		                                        Array<int> &features, 
								                Array<MCDistribution *> &signatures)
{
	int ier, i, j, k, n, nv;
	//Array<int> count;
	Array< Array<int> > points;
	//Array< Array<double> > values;
	//Array< Array<int> > indices;

	// Type: 0=All, 1=Max freq, 2=Max Kbar
	if( type < 0 || type > 2 ) return -1;

	// Compute number of features
	n = 0;
	nv = features.num();
	for( i=0; i<nv; i++ ) {
		j = features[i];
		if( j >  n ) n =  j;
		if( j < -n ) n = -j;
	}

	// Initialize work space
	points.need(n);
	//values.need(n);
	//indices.need(n);
	for( i=0; i<n; i++ ) {
		(points[i]).need(0);
		//(values[i]).need(0);
		//(indices[i]).need(0);
	}

	// Find active indices lists
	switch(type) {
		case 0:     // All
			for( i=0; i<nv; i++ ) {
				j = features[i];
				if( j == 0 ) continue;
				k = (j<0) ? -j-1 : j-1;
				if( k >= 0 && k < n ) {
					points[k] += i;
				}
			}
			break;
		case 1:     // Max frequency
			for( i=0; i<nv; i++ ) {
				j = features[i];
				if( j == 0 ) continue;
				if( j < 0 && j >= -n ) {
					points[-j-1] += i;
				}
			}
			break;
		//case 2:     // Max mean curvature
		//	break;
		default:
			return -1;
	}

	// Delete old signatures
	for( i=0; i<signatures.num(); i++ ) {
		if( signatures[i] != NULL ) delete signatures[i];
	}
	//count.need(n);
	signatures.need(n);
	for( i=0; i<n; i++ ) {
		//count[i] = 0;
		signatures[i] = NULL;
		signatures[i] = new MCDistribution();
	}

	// Compute feature signatures
	for( i=0; i<n; i++ ) {
		ier = (signatures[i])->MCDistribution::Average(*this, points[i]);
		if( ier != 0 ) return ier;
	}

	return 0;
}

int MCCurvatureMap::ComputeListOfDistributions(Array< Array<int> > &points, 
								               Array<MCDistribution *> &signatures)
{
	int ier, i, n;

	n = points.num();
	if( n <= 0 ) return -1;

	// Delete old signatures
	for( i=0; i<signatures.num(); i++ ) {
		if( signatures[i] != NULL ) delete signatures[i];
	}
	//count.need(n);
	signatures.need(n);
	for( i=0; i<n; i++ ) {
		//count[i] = 0;
		signatures[i] = NULL;
		signatures[i] = new MCDistribution();
	}

	// Compute feature signatures
	for( i=0; i<n; i++ ) {
		ier = (signatures[i])->MCDistribution::Average(*this, points[i]);
		if( ier != 0 ) return ier;
	}

	return 0;
}

int MCCurvatureMap::ComputeFeatureDeviations(const double radius,
		                                     Array<int> &features,
											 Array<MCDistribution *> &signatures,
								             Array<double> &deviation)
{
	int i, j, n, nv, type, areaWt;
	double d;
	MCDistribution *sig;
	MCDistribution::MC_MapMetric metric;

	Array<int> count;

	// Initialize work space
	n = signatures.num();
	nv = features.num();
	count.need(n);
	deviation.need(n);
	for( i=0; i<n; i++ ) {
		count[i] = 0;
		deviation[i] = 0.0;
	}

	// Type: 0=All, 1=Max freq
	type = 0;
	areaWt = 0;
	metric = MCDistribution::LOG_MOD;

	for( i=0; i<nv; i++ ) {
		j = features[i];
		if( j == 0 ) continue;
		if( j > 0 ) {
			if( type == 1 ) continue;
		} else {
			j = -j;
		}
		sig = signatures[j-1];
		d = sig->Compare(Distribution(i), metric, radius, areaWt);

		deviation[j-1] += d;
		count[j-1] += 1;
	}

	for( i=0; i<n; i++ ) {
		if( count[i] > 0 ) {
			deviation[i] /= (double) count[i];
		}
	}

	return 0;
}

int MCCurvatureMap::ComputeDeviations(const double radius,
		                              Array< Array<int> > &indexLists,
									  Array<MCDistribution *> &signatures,
								      Array<double> &deviation)
{
	int i, j, k, n, nv, areaWt;
	double d;
	MCDistribution *sig;
	MCDistribution::MC_MapMetric metric;
	Array<int> count;

	nv = NumVertices();
	n = signatures.num();
	if( indexLists.num() != n ) return -1;

	count.need(n);
	deviation.need(n);
	for( i=0; i<n; i++ ) {
		count[i] = 0;
		deviation[i] = 0.0;
	}

	areaWt = 0;
	metric = MCDistribution::LOG_MOD;

	for( i=0; i<n; i++ ) {
		sig = signatures[i];
		for( j=0; j<indexLists[i].num(); j++ ) {
			k = indexLists[i][j];
			if( k < 0 || k >= nv ) continue;

			d = sig->Compare(Distribution(k), metric, radius, areaWt);

			deviation[i] += d;
			count[i] += 1;
		}
	}

	for( i=0; i<n; i++ ) {
		if( count[i] > 0 ) {
			deviation[i] /= (double) count[i];
		}
	}

	return 0;
}

int MCCurvatureMap::ComputeScale(ArrayMesh &mesh, MCCurvatureMap &cmap, 
					             Array<int> &corrFace, Array<double> &corrB1, 
					             Array<double> &corrB2, Array<double> &corrB3, 
					             const double radius,
		                         MCDistribution::MC_MapMetric method,
					             const int areaWt, Array<double> &scale, 
					             const double scaleLimit, const int maxIter)
{
	int ier, i, pct, lastPct, iDebug;
	double sc;

	lastPct = 0;
	iDebug = -1;
	scale.need(m_iNumVertices);
	for( i=0; i<m_iNumVertices; i++ ) {
		if( i == iDebug ) {
			ier++;
		}
		ier = ComputeScale(mesh, cmap, corrFace, corrB1, corrB2, corrB3,
			               radius, i, method, areaWt, sc, scaleLimit, maxIter);
		if( ier != 0 ) return ier;
		scale[i] = sc;

        pct = (int) ((double) (100*(i+1)) / (double) m_iNumVertices);
        if( pct > lastPct && ( pct <= 5 || (pct%10) == 0 ) ) {
            printf("%d pct complete\n", pct);
        }
        lastPct = pct;
	}
	return 0;
}

int MCCurvatureMap::ComputeScale(ArrayMesh &mesh, MCCurvatureMap &cmap, 
					             Array<int> &corrFace, Array<double> &corrB1, 
					             Array<double> &corrB2, Array<double> &corrB3, 
					             const double radius, const int index,
		                         MCDistribution::MC_MapMetric method,
					             const int areaWt, double &scale, 
					             const double scaleLimit, const int maxIter)
{
	int ier, i, k1, k2, k3;
	MCDistribution dist;
	Array<int> pts;
	Array<double> wts;

	if( index < 0 || index >= m_iNumVertices ) return -1;
	if( index >= corrFace.num() || index >= corrB1.num() || 
		index >= corrB2.num() || index >= corrB3.num() ) return -1;

	if( corrFace[index] < 0 ) {
		scale = 1.0;
		return 0;
	}

	i = corrFace[index];
	ier = mesh.getFace(i, k1, k2, k3);
	if( ier != 0 ) return ier;

	pts.need(0);
	pts += k1;
	pts += k2;
	pts += k3;

	wts.need(0);
	wts += corrB1[index];
	wts += corrB2[index];
	wts += corrB3[index];

	ier = dist.WeightedAverage(cmap, pts, wts);
	if( ier != 0 ) return ier;

	ier = Distribution(index).ComputeScale(dist, method, radius, areaWt, 
		                                   scale, scaleLimit, maxIter);
	if( ier != 0 ) return ier;

	return 0;
}

void MCCurvatureMap::dump(FILE *out) const
{
    int ier, i, j;
    double A, R, Kg, Kbar, KgSD, KbarSD, 
           KgMin, KgMax, KbarMin, KbarMax;

    if( out != NULL ) {
        if( m_sMeshDescription != NULL ) {
            fprintf(out,"Mesh: \"%s\"\n",m_sMeshDescription);
        } else {
            fprintf(out,"Mesh: UNKNOWN\n");
        }
        fprintf(out,"Point Ring");
        fprintf(out,"      Area     Radius     ");
        fprintf(out,"       Kg       Kbar      ");
        fprintf(out,"   Std Dev Kg Std Dev Kbar");
        fprintf(out,"     KgMax     KgMin      ");
        fprintf(out,"    KbarMax   KbarMin     \n");

        for( i=0; i<m_iNumVertices; i++ ) {
            for( j=0; j<m_aRingCount[i]; j++ ) {
                ier = Value(MCDistribution::AREA   , i, j, A);
                ier = Value(MCDistribution::RADIUS , i, j, R);
                ier = Value(MCDistribution::KG     , i, j, Kg);
                ier = Value(MCDistribution::KBAR   , i, j, Kbar);
                ier = Value(MCDistribution::KGSD   , i, j, KgSD);
                ier = Value(MCDistribution::KBARSD , i, j, KbarSD);
                ier = Value(MCDistribution::KGMIN  , i, j, KgMin);
                ier = Value(MCDistribution::KGMAX  , i, j, KgMax);
                ier = Value(MCDistribution::KBARMIN, i, j, KbarMin);
                ier = Value(MCDistribution::KBARMAX, i, j, KbarMax);
                fprintf(out," %d %d %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf\n",
                            i, j, A, R, Kg, Kbar, KgSD, KbarSD, 
                                  KgMin, KgMax, KbarMin, KbarMax);
            }
            fprintf(out,"\n\n");
        }
    }
}
