#include "StdAfx.H"
#include <MC/MC_Data.H>
#include <MC/MC_Point.H>
#include <MC/MC_Fit.H>
#include <MC/MC_Polynomial.H>
#include <MC/MC_DiscreteOperator.H>
#include <MC/MC_RadialBasisFunction.H>
#include <MC/MC_Conic.H>
#include <MC/MC_MeshUtils.H>
#include <utils/Pm_MeshLite.H>


static char s_astrs[24][150] = 
{ "None", "Exact (if provided)", 
  "Desbrun", "Modified Desbrun",
  "Goldfeather exact normals", "Goldfeather calc normals",
  "Taubin",
  "Fit polynomial planar 1 ring", "Fit polynomial planar 2 ring", "Fit polynomial planar 3 ring", 
  "Fit polynomial natural param 1 ring", "Fit polynomial natural param 2 ring", "Fit polynomial natural param 3 ring", 
  "Fit RBF planar 1 ring", "Fit RBF planar 2 ring", "Fit RBF planar 3 ring", 
  "Fit RBF natural param 1 ring", "Fit RBF natural param 2 ring", "Fit RBF natural param 3 ring", 
  "Fit Conic 1 ring", "Fit Conic 2 ring", "Fit Conic 3 ring", 
  "Taubin modified",
  "Angle deficit" };

const char *MCData::Description( const MC_Method in_method )
{
    return s_astrs[ (int) in_method + 1 ];
}

MCData::MCData()
{
    m_pMeshDescription = NULL;
    m_pMesh = NULL;
    m_method = NONE;
    m_pFit = NULL;
}

MCData::MCData( const MCData &in_data ) 
{
    (*this) = in_data;
}
    
MCData::~MCData()
{
    if( m_pMeshDescription != NULL ) delete [] m_pMeshDescription;
    if( m_pFit != NULL ) delete m_pFit;
}
    
MCData& MCData::operator=(const MCData &in_data)
{
    m_pMesh = in_data.m_pMesh;
    m_aPoints = in_data.m_aPoints;
    m_pMeshDescription = new char[strlen(in_data.m_pMeshDescription)+1];
    strcpy(m_pMeshDescription,in_data.m_pMeshDescription);

    m_pFit = in_data.m_pFit; // Bad, need to create new fit

    return *this;
}

void MCData::Clear()
{
    m_aPoints.need(0);
}

void MCData::Init(const PMeshLite &mesh, const MC_Method method)
{
    m_pMesh = &mesh;
    m_method = method;
    switch(method) {
        case EXACT:
            break;
        case DCO:
            m_pFit = new MCDiscreteOperator(*m_pMesh, MCDiscreteOperator::DCO);
            break;
        case MDCO:
            m_pFit = new MCDiscreteOperator(*m_pMesh, MCDiscreteOperator::DCO);
            break;
        case ANCE:
            m_pFit = new MCDiscreteOperator(*m_pMesh, MCDiscreteOperator::ANCe);
            break;
        case ANC:
            m_pFit = new MCDiscreteOperator(*m_pMesh, MCDiscreteOperator::ANC);
            break;
        case IEV:
            m_pFit = new MCDiscreteOperator(*m_pMesh, MCDiscreteOperator::IEV);
            break;
        case FIT1P:
            m_pFit = new MCPolynomial(*m_pMesh, 1, MCParam::PLANAR);
            break;
        case FIT2P:
            m_pFit = new MCPolynomial(*m_pMesh, 2, MCParam::PLANAR);
            break;
        case FIT3P:
            m_pFit = new MCPolynomial(*m_pMesh, 3, MCParam::PLANAR);
            break;
        case FIT1N:
            m_pFit = new MCPolynomial(*m_pMesh, 1, MCParam::NATURAL);
            break;
        case FIT2N:
            m_pFit = new MCPolynomial(*m_pMesh, 2, MCParam::NATURAL);
            break;
        case FIT3N:
            m_pFit = new MCPolynomial(*m_pMesh, 3, MCParam::NATURAL);
            break;
        case RBF1P:
            m_pFit = new MCRadialBasisFunction(*m_pMesh, 1, MCParam::PLANAR);
            break;
        case RBF2P:
            m_pFit = new MCRadialBasisFunction(*m_pMesh, 2, MCParam::PLANAR);
            break;
        case RBF3P:
            m_pFit = new MCRadialBasisFunction(*m_pMesh, 3, MCParam::PLANAR);
            break;
        case RBF1N:
            m_pFit = new MCRadialBasisFunction(*m_pMesh, 1, MCParam::NATURAL);
            break;
        case RBF2N:
            m_pFit = new MCRadialBasisFunction(*m_pMesh, 2, MCParam::NATURAL);
            break;
        case RBF3N:
            m_pFit = new MCRadialBasisFunction(*m_pMesh, 3, MCParam::NATURAL);
            break;
        case CON1:
            m_pFit = new MCConic(*m_pMesh, 1);
            break;
        case CON2:
            m_pFit = new MCConic(*m_pMesh, 2);
            break;
        case CON3:
            m_pFit = new MCConic(*m_pMesh, 3);
            break;
        case HST:
            m_pFit = new MCDiscreteOperator(*m_pMesh, MCDiscreteOperator::HST);
            break;
        case DGM:
            m_pFit = new MCDiscreteOperator(*m_pMesh, MCDiscreteOperator::DGM);
            break;
        default:
            ASSERT(false);
            break;
    }
}

void MCData::Init(const PMeshLite &in_mesh, MC_Method method,
                  const MCData &in_data, const Array<Vertex> &in_aopVs)
{
    int i;
    MCPoint cpt;
    Array<Vertex> aopVs;

    FILE *out = NULL;
    out = fopen(WINDataHome("extract.dat"),"a");

    in_mesh.AllMeshVertices(aopVs);

    Init(in_mesh, method);
    
    if( out != NULL ) {
        fprintf(out,"\nExtracting data: %d vertices from %d vertices, %d faces\n",
                    in_aopVs.num(), in_data.Mesh()->NumVertices(),
                    in_data.Mesh()->NumFaces() );
    }

    if( in_mesh.NumVertices() == in_aopVs.num() ) {
        for( i=0; i<in_aopVs.num(); i++ ) {
            in_data.getPoint(in_aopVs[i], cpt);
            m_aPoints += cpt;
            if( out != NULL ) {
                R3Pt P = aopVs[i]->GetLoc();
                fprintf(out,"i=%d, IndexId=%d (%lf %lf %lf) %d %lf %lf\n", 
                            i, in_aopVs[i]->IndexId(), P[0], P[1], P[2],
                            cpt.Id(), cpt.K1(), cpt.K2() );
            }
        }
    }
    if( out != NULL ) fclose(out);

    SetRange();
}

int MCData::getPoint(const Vertex &in_vertex, MCPoint &out_point) const
{
    int id;

    id = in_vertex->IndexId();
    if( id < 0 || id >= m_aPoints.num() ) return -1;

    out_point = m_aPoints[id];
    return 0;
}

int MCData::getPoint(const int index, MCPoint &out_point) const
{
    if( index < 0 || index >= m_aPoints.num() ) return -1;

    out_point = m_aPoints[index];
    return 0;
}

int 
MCData::setPoint(const Vertex &in_vertex, const MCPoint &in_point)
{
    

    if( m_aPoints.num() != m_pMesh->NumVertices() ) {
        m_aPoints.need( m_pMesh->NumVertices() );
    }

    int i = in_vertex->IndexId();
    if( i < 0 || i >= m_aPoints.num() ) return -1;

    m_aPoints[i] = in_point;

    return 0;
}

int
MCData::ReadMeshCurvature( const char *filename )
{
    // This routine reads curvature for a mesh in Dr. Grimm's format

    int i, n, nv, nread;
    double area;
    double k1, k2;
    R3Pt pt;
    R3Vec N, V1, V2;

    MCPoint cpt;
    
    FILE *in;
    char str[255];
    char *ptr;
    WINbool flip;
    double mag;

    flip = TRUE;

    Clear();  // Clear out any old data

    // Open input file

    in = fopen( filename, "r" );
	if( in == NULL ) return -1;

    // Read vertices

    ptr = fgets(str, 254, in);  // Read the header line
    
    nread = fscanf( in, "%d", &nv );
    if( m_pMesh == NULL ) return -1;
    if( m_pMesh->NumVertices() != nv ) return -2;

    for( i=0; i<nv; i++ ) {
        nread = fscanf( in, "%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
                  &n, &pt[0], &pt[1], &pt[2], &area, &N[0], &N[1], &N[2], 
                  &k1, &k2, &V1[0], &V1[1], &V1[2], &V2[0], &V2[1], &V2[2] );
        ASSERT(nread >= 16 );

        // Normalize the normal
        mag = N[0]*N[0] + N[1]*N[1] + N[2]*N[2];
        if( mag != 0.0 ) {
            mag = 1.0/sqrt(mag);
            N *= -mag;
        }
        
        if( flip ) {
            k1 = -k1;
            k2 = -k2;
        }
        if( k2*k2 > k1*k1 ) {
            cpt.setId(n);
            cpt.setMaxMin(k2,k1);
            cpt.setNormal(N);
            cpt.setMaxMinVec(V2,V1);
        } else {
            cpt.setId(n);
            cpt.setMaxMin(k1,k2);
            cpt.setNormal(N);
            cpt.setMaxMinVec(V1,V2);
        }
        m_aPoints += cpt;
    }
    fclose( in );

	FixCurvatureSign(2);

    SetRange();

    return 0;
}

void 
MCData::FixCurvatureSign(const int flip)
{
    int i;
	double dot;
    //double k1A, k2A, k1B, k2B;
    double k1, k2;
    R3Pt P;
	R3Vec N, NN;
    //Vertex V;
    MCPoint cpt;
    Array<Vertex> aopVs;

    switch(flip) {
        case 0:      // Do not flip
            break;
        case 1:      // Flip
            for( i=0; i<m_aPoints.num(); i++ ) {
                k1 = m_aPoints[i].K1();
                k2 = m_aPoints[i].K2();
                m_aPoints[i].setMaxMin(-k1,-k2);
            }
            SetRange();
            break;
        case 2:      // Flip if the surface normal is opposite curvature normal
			N = m_pMesh->VertexFromIndexId(0)->GetNorm();
			NN = m_aPoints[0].Normal();
			dot = N[0]*NN[0] + N[1]*NN[1] + N[2]*NN[2];
            if( dot < 0.0 ) {
                for( i=0; i<m_aPoints.num(); i++ ) {
                    k1 = m_aPoints[i].K1();
                    k2 = m_aPoints[i].K2();
                    m_aPoints[i].setMaxMin(-k1,-k2);

					N = m_aPoints[1].Normal();
					N *= -1.0;
					m_aPoints[i].setNormal(N);
                }
                SetRange();
            }

            
        default:
            break;
    }
}
    
int
MCData::ReadMeshNormals( const char *filename )
{
    // This routine reads normals for a mesh in Dr. Grimm's format

    int i, n, nv, nread;
    double area;
    double k1, k2, dot;
    R3Pt pt;
    R3Vec N, NN, V1, V2;

    MCPoint cpt;
    
    FILE *in;
    char str[255];
    char *ptr;
    //int flip = 0;
    double mag;

    // Open input file

    in = fopen( filename, "r" );

    // Read vertices

    ptr = fgets(str, 254, in);  // Read the header line
    
    nread = fscanf( in, "%d", &nv );
    if( m_pMesh == NULL ) return -1;
    if( m_pMesh->NumVertices() != nv ) return -2;

    // Make sure data array is initialized
    if( m_aPoints.num() != nv ) {
        Clear();
        for( i=0; i<nv; i++ ) {
            m_aPoints += cpt;
        }
    }

    for( i=0; i<nv; i++ ) {
        nread = fscanf( in, "%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
                  &n, &pt[0], &pt[1], &pt[2], &area, &N[0], &N[1], &N[2], 
                  &k1, &k2, &V1[0], &V1[1], &V1[2], &V2[0], &V2[1], &V2[2] );
        ASSERT(nread >= 16 );

		// Compare to the mesh normal to see if the normal should be flipped
		// Nref = m_pMesh->GetNorm();
		// if( N[0]*Nref[0] + N[0]*Nref[0] + N[0]*Nref[0] > 0.0 ) {
		//     flip++;
		// } else {
		//     flip--;
	    // }


        // Normalize the normal
        mag = N[0]*N[0] + N[1]*N[1] + N[2]*N[2];
        if( mag != 0.0 ) {
            mag = 1.0/sqrt(mag);
            N *= -mag;
        }
        m_aPoints[i].setNormal(N);
    }
    fclose( in );

	N = m_pMesh->VertexFromIndexId(0)->GetNorm();
	NN = m_aPoints[0].Normal();
	dot = N[0]*NN[0] + N[1]*NN[1] + N[2]*NN[2];
    if( dot < 0.0 ) {
		FlipMeshNormals();
	}

    return 0;
}

void
MCData::FlipMeshNormals()
{
    // This routine flips the normal direction for a mesh

    int i;
    R3Vec N;

    for( i=0; i<m_aPoints.num(); i++ ) {
        N = m_aPoints[i].Normal();
        N *= -1.0;
        m_aPoints[i].setNormal(N);
    }
}
   
/* 
void
MCData::FlipCurvatureSign()
{
    // This routine flips the sign of the curvature for a mesh

    int i;
    R3Vec N;

    for( i=0; i<m_aPoints.num(); i++ ) {

        N = m_aPoints[i].Normal();
        //MCData::FixCurvatureSign();
        N *= -1.0;
        m_aPoints[i].setNormal(N);
    }
}
*/
    
int 
MCData::CalculateCurvature( const Vertex &in_vertex, MCPoint &point ) const
{
    int ier = 0;

    ier = m_pFit->Curvature(in_vertex, point);

    return ier;
}

int 
MCData::CalculateCurvature()
{
    int i, ier;
    Array< Vertex > aopVs;
    MCPoint cpt;

    m_pMesh->AllMeshVertices(aopVs);
    for( i=0; i<aopVs.num(); i++ ) {
        ier = CalculateCurvature(aopVs[i],cpt);
        if( ier < 0 ) return ier;
        setPoint(aopVs[i],cpt);
    }
    SetRange();

    return 0;
}
    
int MCData::InterpolateBary(const int i1, const double b1, 
		                    const int i2, const double b2, 
		                    const int i3, const double b3, 
		                    double &kg, double &kbar,
							double &k1, double &k2)
{
	double f, rad;

	if( i1 < 0 || i1 >= m_aPoints.num() ||
		i2 < 0 || i2 >= m_aPoints.num() ||
		i3 < 0 || i3 >= m_aPoints.num() ) return -1;

	kg = 0.0;
	kg += b1 * m_aPoints[i1].Kg();
	kg += b2 * m_aPoints[i2].Kg();
	kg += b3 * m_aPoints[i3].Kg();

	kbar = 0.0;
	kbar += b1 * m_aPoints[i1].Kbar();
	kbar += b2 * m_aPoints[i2].Kbar();
	kbar += b3 * m_aPoints[i3].Kbar();

	f = b1 + b2 + b3;
	if( f > 0.0 && f != 1.0) {
		f = 1.0/f;
		kg *= f;
		kbar *= f;
	}

	rad = kbar*kbar - kg;
	if( rad < 0.0 ) rad = 0.0;
	if( kbar < 0.0 ) {
		k1 = kbar - sqrt(rad);
		k2 = kbar + sqrt(rad);
	} else {
		k1 = kbar + sqrt(rad);
		k2 = kbar - sqrt(rad);
	}

	return 0;
}

void 
MCData::SetRange()
{
    int i;
    double k1min, k1max, k2min, k2max, kgmin, kgmax, kbarmin, kbarmax;
    Array< Vertex > aopVs;
    MCPoint cpt;

    for( i=0; i<m_aPoints.num(); i++ ) {
        cpt = m_aPoints[i];
        if( i == 0 ) {
            k1min = cpt.K1();
            k1max = cpt.K1();
            k2min = cpt.K2();
            k2max = cpt.K2();
            kgmin = cpt.Kg();
            kgmax = cpt.Kg();
            kbarmin = cpt.Kbar();
            kbarmax = cpt.Kbar();
        } else {
            if( cpt.K1() < k1min ) k1min = cpt.K1();
            if( cpt.K1() > k1max ) k1max = cpt.K1();
            if( cpt.K2() < k2min ) k2min = cpt.K2();
            if( cpt.K2() > k2max ) k2max = cpt.K2();
            if( cpt.Kg() < kgmin ) kgmin = cpt.Kg();
            if( cpt.Kg() > kgmax ) kgmax = cpt.Kg();
            if( cpt.Kbar() < kbarmin ) kbarmin = cpt.Kbar();
            if( cpt.Kbar() > kbarmax ) kbarmax = cpt.Kbar();
        }
    }
    m_K1Range = R2Pt(k1min,k1max);
    m_K2Range = R2Pt(k2min,k2max);
    m_KgRange = R2Pt(kgmin,kgmax);
    m_KbarRange = R2Pt(kbarmin,kbarmax);
}
    
void MCData::dump(FILE *out)
{
    int i;
    MCPoint pt;

    if( out != NULL ) {
        if( m_pMeshDescription != NULL ) {
            fprintf(out,"Mesh: %s\n",m_pMeshDescription);
        }
        if( m_pMesh != NULL ) {
            fprintf(out,"  %d vertices, %d faces\n", 
                        m_pMesh->NumVertices(), m_pMesh->NumFaces() );
        }
        fprintf(out,"  Method: %d\n", (int)m_method );
        if( m_aPoints.num() > 0 ) {
            fprintf(out," K1 Range: %lf to %lf\n", m_K1Range[0], m_K1Range[1]);
            fprintf(out," K2 Range: %lf to %lf\n", m_K2Range[0], m_K2Range[1]);
            for( i=0; i<m_aPoints.num(); i++ ) {
                pt = m_aPoints[i];
                fprintf(out," Point %d: k1= %lf, k2= %lf\n",
                            i, pt.K1(), pt.K2() );
            }
            if(m_pFit != NULL ) m_pFit->dump(out);
        }
    }
}

void MCData::dumpRanges(FILE *out)
{
    MCPoint pt;

    if( out != NULL ) {
        if( m_pMeshDescription != NULL ) {
            fprintf(out,"Mesh: %s\n",m_pMeshDescription);
        }
        if( m_pMesh != NULL ) {
            fprintf(out,"  %d vertices, %d faces\n", 
                        m_pMesh->NumVertices(), m_pMesh->NumFaces() );
        }
        fprintf(out,"  Method: %d\n", (int)m_method );
        if( m_aPoints.num() > 0 ) {
            fprintf(out," K1 Range: %lf to %lf\n", m_K1Range[0], m_K1Range[1]);
            fprintf(out," K2 Range: %lf to %lf\n", m_K2Range[0], m_K2Range[1]);
        }
    }
}
