#include "StdAfx.H"
#include <MC/MC_DistanceGrid.H>
#include <MC/MC_MeshUtils.H>

MCDistanceGrid::MCDistanceGrid()
{
    m_sMeshTitle = NULL;
    m_sMetricTitle = NULL;
    m_pCmap = NULL;
    m_pFanData = NULL;
    points.need(0);
    m_iRings = 0;
	m_iInteg = 0;
	m_iAreaWt = 0;
	m_GridType = UNKNOWN;
}

MCDistanceGrid::MCDistanceGrid( const MCDistanceGrid &in_dist ) 
{
    (*this) = in_dist;
}
    
MCDistanceGrid::~MCDistanceGrid()
{
    delete [] m_sMeshTitle;
    delete [] m_sMetricTitle;
}
    
MCDistanceGrid& MCDistanceGrid::operator=(const MCDistanceGrid &in_dist)
{
    m_sMeshTitle = new char[strlen(in_dist.m_sMeshTitle)+1];
    strcpy(m_sMeshTitle,in_dist.m_sMeshTitle);

    m_sMetricTitle = new char[strlen(in_dist.m_sMetricTitle)+1];
    strcpy(m_sMetricTitle,in_dist.m_sMetricTitle);

    return *this;
}

void MCDistanceGrid::MeshTitle(const char *title)
{
    delete [] m_sMeshTitle;
    m_sMeshTitle = new char[strlen(title)+1];
    strcpy(m_sMeshTitle, title);
}

void MCDistanceGrid::MetricTitle(const char *title)
{
    delete [] m_sMetricTitle;
    m_sMetricTitle = new char[strlen(title)+1];
    strcpy(m_sMetricTitle, title);
}

int MCDistanceGrid::GetPointIndex(const int index) const
{
	if( index < 0 || index >= points.num() ) return -1;

	return points[index];
}

int MCDistanceGrid::Create(MCCurvatureMap &cmap, 
                           const Array<int> &pts,
                           const MCDistribution::MC_MapMetric metric,
                           const int nRing, const int integ,
						   const int areaWt)
{
    int i, j, ii, jj, n, nn;
	double delta;

    n = pts.num();
	if( grid.num() == 0 ) {
        grid.init(n);
        for( i=0; i<n; i++ ) {
            grid[i].init(n);
        }
    }
    if( n != grid.num() ) return -1;

    points.need(0);
    for( i=0; i<n; i++ ) {
        points += pts[i];
    }

    nn = cmap.NumVertices();
    for( j=0; j<n; j++ ) {
        grid[j][j] = 0.0;
        for( i=0; i<j; i++ ) {
			ii = points[i];
            jj = points[j];
			if( ii < 0 || ii >= nn || jj < 0 || jj >= nn ) {
				grid[i][j] = -987.654;
				grid[j][i] = -987.654;
				continue;
			}
			if( integ == 0 ) {
				delta = (cmap.Distribution(ii)).Compare(
						 cmap.Distribution(jj), metric, nRing, areaWt);
			} else {
				delta = (cmap.Distribution(ii)).CompareIntegral(
						 cmap.Distribution(jj), metric, nRing, areaWt);
			}
			grid[i][j] = delta;
			grid[j][i] = delta;
        }
    }
    m_pCmap = &cmap;
    m_iRings = nRing;
	m_iInteg = integ;
	m_iAreaWt = areaWt;
    MetricTitle(MCDistribution::Metric(metric));
	m_GridType = CMAP;

    return 0;
}

int MCDistanceGrid::Create(SArray<tdgFanData> &fdat, 
                           const Array<int> &pts,
                           const MCDistribution::MC_MapMetric metric,
                           const int nRing, const int integ,
						   const int areaWt)
{
    int ier, i, j, ii, jj, n, nn;
	double delta, ang;
    tdgFanData *refFan;
    tdgFanData *curFan;

    n = pts.num();
	if( grid.num() == 0 ) {
        grid.init(n);
        for( i=0; i<n; i++ ) {
            grid[i].init(n);
        }
    }
    if( n != grid.num() ) return -1;

    points.need(0);
    for( i=0; i<n; i++ ) {
        points += pts[i];
    }

	nn = fdat.num();
    for( j=0; j<n; j++ ) {
        grid[j][j] = 0.0;
        jj = points[j];
        curFan = &((fdat)[jj]);
        for( i=0; i<j; i++ ) {
			ii = points[i];
			refFan = &((fdat)[ii]);
			if( ii < 0 || ii >= nn || jj < 0 || jj >= nn ) {
				grid[i][j] = -987.654;
				grid[j][i] = -987.654;
				continue;
			}
            ier = curFan->Compare(*refFan, metric, delta, ang, nRing);
            if( ier != 0 ) {
                return ier;
            }
            grid[i][j] = delta;
			grid[j][i] = delta;
        }
    }
    m_pFanData = &fdat;
    m_iRings = nRing;
	m_iInteg = integ;
	m_iAreaWt = areaWt;
    MetricTitle(MCDistribution::Metric(metric));
	m_GridType = FANDATA;

    return 0;
}

int MCDistanceGrid::Update(const MCDistribution::MC_MapMetric metric,
                           const int nRing, const int integ,
						   const int areaWt)
{
    int ier, i, j, ii, jj, n, nn;
	double delta, ang;
    tdgFanData *refFan;
    tdgFanData *curFan;

    if( m_GridType == CMAP && m_pCmap == NULL ) return -1;
    if( m_GridType == FANDATA && m_pFanData == NULL ) return -1;

    n = grid.num();
    if( n <= 0 ) return -1;

	switch(m_GridType) {
	case CMAP:
		nn = m_pCmap->NumVertices();
    	break;
	case FANDATA:
		nn = m_pFanData->num();
    	break;
	default:
		return -1;
	}
    for( j=0; j<n; j++ ) {
        grid[j][j] = 0.0;
        jj = points[j];
        for( i=0; i<j; i++ ) {
            ii = points[i];
            if( ii < 0 || ii >= nn || jj < 0 || jj >= nn ) {
				grid[i][j] = -987.654;
				grid[j][i] = -987.654;
				continue;
			}
			switch(m_GridType) {
			case CMAP:
				if( integ == 0 ) {
					delta = (m_pCmap->Distribution(ii)).Compare(
							 m_pCmap->Distribution(jj), metric, nRing, areaWt);
				} else {
					delta = (m_pCmap->Distribution(ii)).CompareIntegral(
                             m_pCmap->Distribution(jj), metric, nRing, areaWt);
				}
				break;
			case FANDATA:
				curFan = &((*m_pFanData)[jj]);
				refFan = &((*m_pFanData)[ii]);
				ier = curFan->Compare(*refFan, metric, delta, ang, nRing);
				if( ier != 0 ) {
				    return ier;
				}
				break;
			default:
				return -1;
			}
			grid[i][j] = delta;
			grid[j][i] = delta;
        }
    }
    m_iRings = nRing;
    m_iInteg = integ;
	m_iAreaWt = areaWt;
    MetricTitle(MCDistribution::Metric(metric));

    return 0;
}


double MCDistanceGrid::Value(const int i, const int j) const
{
    if( i < 0 || i >= points.num() ||
        j < 0 || j >= points.num() ) {
        return -987.654;
    }

    return grid[i][j];
}

double MCDistanceGrid::Max() const
{
    int i, j, n;
    double value;

    n = points.num();
    value = grid[0][0];
    for( j=0; j<n; j++ ) {
        for( i=0; i<n; i++ ) {
            if( grid[i][j] > value ) {
                value = grid[i][j];
            }
        }
    }
    return value;
}

double MCDistanceGrid::Min() const
{
    int i, j, n;
    double value;

    n = points.num();
    value = grid[0][0];
    for( j=0; j<n; j++ ) {
        for( i=0; i<n; i++ ) {
            if( grid[i][j] < value ) {
                value = grid[i][j];
            }
        }
    }
    return value;
}
	
int MCDistanceGrid::MeanVariance(const int i0, const int i1,
		                         const int j0, const int j1,
					             double &mean, double &variance,
								 int &count)
{
	int i, j, n, ni, nj;
	double sum, sumsq;
	FILE *out = NULL;

	//out = fopen(WINDataHome("stat_work.dat"),"a");

	n = Size();
	if( i0 < 0 || i0 >= n || i1 < 0 || i1 >= n ||
		j0 < 0 || j0 >= n || j1 < 0 || j1 >= n ||
		i1 < i0 || j1 < j0 ) return -1;

	ni = 1+i1-i0;
	nj = 1+j1-j0;
	if( ni < 1 || nj < 1 ) return -1;

	count = 0;
	sum = 0.0;
	sumsq = 0.0;
	for( j=j0; j<=j1; j++ ) {
		for( i=i0; i<=i1; i++ ) {
			if( i == j ) continue;
			if( i >= j0 && i <= j1 && 
				j >= i0 && j <= i1 && i > j ) continue;
			sum += grid[i][j];
			sumsq += grid[i][j] * grid[i][j];
			count++;
			if( out != NULL ) {
				fprintf(out,"grid[%d][%d]=%lf, sum=%lf, sq=%lf, %d\n",
					        i, j, grid[i][j], sum, sumsq, count);
			}
		}
	}
	if( count < 1 ) return -1;
	mean = sum / (double) count;

	if( count < 2 ) {
		variance = 0.0;
		return 0;
	}

	variance = sumsq - (double) count * mean * mean;
	variance /= (double) (count-1);

	if( out != NULL ) {
		fprintf(out,"mean=%lf, var=%lf\n", mean, variance);
		fclose(out);
	}

	return 0;
}

int MCDistanceGrid::tTest(const char *filename, const int groupSize)
{
	int ier, i, j, i0, i1, j0, j1, n, ng, ni, nj, Na, Nb, count;
	double M, Ma, Mb, V, Va, Vb, Vd, t;
	SArray< SArray<double> > aoMean;
	SArray< SArray<double> > aoVar;
    SArray< SArray<int> > aoNum;
    SArray< SArray<double> > aoT;
    SArray< SArray<int> > aoDof;

	int nThresh1 = 0;
	int nThresh2 = 0;
	double thresh1 =  6.0;
	double thresh2 = 10.0;

    int numperline = 10;
	FILE *out = NULL;

	// Compute number of groups
	n = Size();
	ng = n/groupSize;
	if( groupSize*ng < n ) ng++;

	//out = fopen(WINDataHome("stat_work.dat"),"w");
	if( out != NULL ) {
        if( m_sMeshTitle != NULL ) {
            fprintf(out,"Mesh: \"%s\"\n",m_sMeshTitle);
        } else {
            fprintf(out,"Mesh: UNKNOWN\n");
        }
        if( m_sMetricTitle != NULL ) {
            fprintf(out,"Metric: \"%s\"\n",m_sMetricTitle);
        } else {
            fprintf(out,"Metric: UNKNOWN\n");
        }
		fprintf(out,"  %d rings, %d points, %d groups\n", 
			        m_iRings, n, ng);
		fprintf(out,"  Integ = %d, Area Wt = %d\n", m_iInteg, m_iAreaWt);
		fprintf(out,"\n");
		fclose(out);
	}

	// Initialize array size
	aoMean.init(ng);
	aoVar.init(ng);
	aoNum.init(ng);
	aoT.init(ng);
	aoDof.init(ng);
	for( i=0; i<ng; i++ ) {
		aoMean[i].init(ng);
		aoVar[i].init(ng);
		aoNum[i].init(ng);
		aoT[i].init(ng);
		aoDof[i].init(ng);
	}

	// Compute mean and variance of distances 
	// within and between groups
	nj = 0;
	for( j0=0; j0<n; j0+=groupSize ) {
		j1 = j0 + groupSize - 1;
		if( j1 >= n ) j1 = n-1;
		ni = 0;
		for( i0=0; i0<=j0; i0+=groupSize ) {
			i1 = i0 + groupSize - 1;
			if( i1 >= n ) i1 = n-1;
			ier = MeanVariance(i0, i1, j0, j1, M, V, count);
			aoMean[nj][ni] = M;
			aoVar[nj][ni] = V;
			aoNum[nj][ni] = count;
			aoMean[ni][nj] = aoMean[nj][ni];
			aoVar[ni][nj] = aoVar[nj][ni];
			aoNum[ni][nj] = aoNum[nj][ni];
			ni++;
		}
		nj++;
	}

	//out = fopen(WINDataHome("stat_work.dat"),"a");
	if( out != NULL ) fprintf(out,"\n");

	// Compute t-Test between each pair of groups
	for( j=0; j<ng; j++ ) {
		aoT[j][j] = 0.0;
		for( i=0; i<=j; i++ ) {
			Ma = aoMean[i][i];
			Mb = aoMean[j][i];
			Va = aoVar[i][i];
			Vb = aoVar[j][i];
			Na = aoNum[i][i];
			Nb = aoNum[j][i];
			//Vd = Va/(double)Na + Vb/(double)Nb;
			Vd = (double) (Na-1) * Va + (double) (Nb-1) * Vb;
			Vd *= (double) (Na + Nb)/(double) (Na*Nb*(Na+Nb-2));
			t = (Mb - Ma) / sqrt(Vd);
			//t = (Mb - Ma) / Vd;
			aoT[j][i] = (t<0.0) ? -t : t;
			aoDof[j][i] = Na + Nb - 2;
			if( out != NULL ) {
				fprintf(out,"(%d,%d) Ma=%lf, Mb=%lf, Va=%lf, Vb = %lf\n",
					        j, i, Ma, Mb, Va, Vb);
				fprintf(out,"      Na=%d, Nb=%d, Vd=%lf, t = %lf\n",
					        Na, Nb, Vd, t);
			}
			if( i == j ) continue;

			Ma = aoMean[j][j];
			Mb = aoMean[j][i];
			Va = aoVar[j][j];
			Vb = aoVar[j][i];
			Na = aoNum[j][j];
			Nb = aoNum[j][i];
			//Vd = Va/(double)Na + Vb/(double)Nb;
			Vd = (double) (Na-1) * Va + (double) (Nb-1) * Vb;
			Vd *= (double) (Na + Nb)/(double) (Na*Nb*(Na+Nb-2));
			t = (Mb - Ma) / sqrt(Vd);
			//t = (Mb - Ma) / Vd;
			aoT[i][j] = (t<0.0) ? -t : t;
			aoDof[i][j] = Na + Nb - 2;
			if( out != NULL ) {
				fprintf(out,"(%d,%d) Ma=%lf, Mb=%lf, Va=%lf, Vb = %lf\n",
					        i, j, Ma, Mb, Va, Vb);
				fprintf(out,"      Na=%d, Nb=%d, Vd=%lf, t = %lf\n",
					        Na, Nb, Vd, t);
			}
		}
	}
	if( out != NULL ) {
		fprintf(out,"\n");
		fclose(out);
		out = NULL;
	}

	// Write out the t-Test data
	out = fopen(filename,"a");
	if( out != NULL ) {
        if( m_sMeshTitle != NULL ) {
            fprintf(out,"Mesh: \"%s\"\n",m_sMeshTitle);
        } else {
            fprintf(out,"Mesh: UNKNOWN\n");
        }
        if( m_sMetricTitle != NULL ) {
            fprintf(out,"Metric: \"%s\"\n",m_sMetricTitle);
        } else {
            fprintf(out,"Metric: UNKNOWN\n");
        }
		fprintf(out,"  %d rings, %d points, %d groups\n", 
			        m_iRings, n, ng);
		fprintf(out,"  Integ = %d, Area Wt = %d\n", m_iInteg, m_iAreaWt);
        fprintf(out,"  DoF (within group) = %d, DoF (between groups) = %d\n",
			        aoDof[0][0], aoDof[0][1]);
        for( j=0; j<ng; j++ ) {
            fprintf(out,"Row %d:\n",j);
            for( i=0; i<ng; i++ ) {
                fprintf(out,"%6.2lf ", aoT[i][j]);
				if( aoT[i][j] >= thresh1 ) nThresh1++;
                if( aoT[i][j] >= thresh2 ) nThresh2++;
                if( (i+1)%numperline == 0 ) fprintf(out,"\n");
            }
            if( (i)%numperline != 0 ) fprintf(out,"\n");
        }
		fprintf(out,"%d above %lf,  %d above %lf\n",
			        nThresh1, thresh1, nThresh2, thresh2);
		fprintf(out,"\n");
		fclose(out);
    }

	return 0;
}

void MCDistanceGrid::dump(FILE *out) const
{
    int i, j, n;
    int numperline = 10;

    if( out != NULL ) {
        if( m_sMeshTitle != NULL ) {
            fprintf(out,"Mesh: \"%s\"\n",m_sMeshTitle);
        } else {
            fprintf(out,"Mesh: UNKNOWN\n");
        }
        if( m_sMetricTitle != NULL ) {
            fprintf(out,"Metric: \"%s\"\n",m_sMetricTitle);
        } else {
            fprintf(out,"Metric: UNKNOWN\n");
        }
		fprintf(out,"  %d rings\n", m_iRings);
        fprintf(out,"  Integ = %d, Area Wt = %d\n", m_iInteg, m_iAreaWt);
        n = points.num();
        if( n > 0 ) {
            for( j=0; j<n; j++ ) {
                fprintf(out,"Row %d:\n",j);
                for( i=0; i<n; i++ ) {
                    fprintf(out,"%lf ", grid[i][j]);
                    if( (i+1)%numperline == 0 ) fprintf(out,"\n");
                }
                if( (i)%numperline != 0 ) fprintf(out,"\n");
            }
        }
    }
}
