#include "StdAfx.H"
#include <MC/MC_Point.H>
#include <MC/MC_Fit.H>
#include <MC/MC_MeshSubset.H>
#include <MC/MC_Param.H>
#include <utils/Pm_MeshLite.H>

MCFit::MCFit(const PMeshLite &in_mesh, 
             const MC_FitType &in_fitType,
             const MCParam::MC_ParamType &in_paramType,
             const int in_nRings )
: m_pMesh(in_mesh)
, m_FitType(in_fitType)
, m_ParamType(in_paramType)
, m_iRings(in_nRings)
, m_Param(in_mesh,in_nRings,in_paramType)
{
    valid = FALSE;
}
    
MCFit::~MCFit()
{
}

void 
MCFit::Curvature(const R3Vec &Ds, const R3Vec &Dt, 
                 const R3Vec &Dss, const R3Vec &Dst, const R3Vec &Dtt,
                 double &out_dArea, MCPoint &out_data ) const
{
	Curvature2(Ds, Dt, Dss, Dst, Dtt, out_dArea, out_data);
	return;

    R3Vec v1, v2;
    double k1, k2;
        
    const double dE = Dot( Ds, Ds );
    const double dF = Dot( Ds, Dt );
    const double dG = Dot( Dt, Dt );

    v1 = Cross( Ds, Dt );
    v1.Normalize();
    R3Vec refNormal;
    refNormal = m_Param.Nref();

    double factor = 1.0;
    if( refNormal[0] != 0.0 ||  refNormal[1] != 0.0 ||  refNormal[2] != 0.0 ) {
        if( Dot(v1,refNormal) > 0.0 ) factor = -1.0;
    }

	if( factor > 0.0 ) {
	    out_data.setNormal(R3Vec(-v1[0], -v1[1], -v1[2]));
	} else {
	    out_data.setNormal(v1);
	}

    const double de = Dot( Dss, v1 );
    const double df = Dot( Dst, v1 );
    const double dg = Dot( Dtt, v1 );

    out_dArea = sqrt( dE * dG - dF * dF );

    const double dA = (dF * dg - dG * df);
    const double dB = (dE * dg - dG * de);
    const double dC = (dE * df - dF * de);

    double dR = dB * dB - 4.0 * dA * dC;
	if( RNApproxEqual(dR,0.0,1.E-6) && dR < 0.0 ) dR = 0.0;

	const double dRoot = dR;
    if ( dRoot < 0 ) {
        ASSERT(FALSE);
    } 

    const double dL1 = ( -dB + sqrt(dRoot) ) / (2.0 * dA);
    const double dL2 = ( -dB - sqrt(dRoot) ) / (2.0 * dA);

    const double dAng1 = atan( dL1 );
    const double dAng2 = atan( dL2 );

    k1 = factor * (de + df * dL1) / (dE + dF * dL1);
    k2 = factor * (de + df * dL2) / (dE + dF * dL2);
    
    if( k1*k1 < k2*k2 ) {
        out_data.setMaxMin(k2,k1);

        out_data.setMaxMinVecPar( R2Vec(cos(dAng2), sin(dAng2) ),
                                  R2Vec(cos(dAng1), sin(dAng1) ) );

        out_data.setMaxMinVec( Ds*cos(dAng2) + Dt*sin(dAng2),
                               Ds*cos(dAng1) + Dt*sin(dAng1) );
    } else {
        out_data.setMaxMin(k1,k2);

        out_data.setMaxMinVecPar( R2Vec( cos(dAng1), sin(dAng1) ),
                                  R2Vec( cos(dAng2), sin(dAng2) ) );

        out_data.setMaxMinVec( Ds*cos(dAng1) + Dt*sin(dAng1),
                               Ds*cos(dAng2) + Dt*sin(dAng2) );
    }
}

void 
MCFit::Curvature2(const R3Vec &Ds, const R3Vec &Dt, 
                  const R3Vec &Dss, const R3Vec &Dst, const R3Vec &Dtt,
                  double &out_dArea, MCPoint &out_data ) const
{
    R3Vec n, v1, v2;
    double k1, k2;
	double E, F, G, L, M, N, fC0, fC1, fC2, fA0, fA1; 
	double fTemp, fMult, fLength, epsilon, factor; 

	epsilon = 1.0e-6;
        
    E = Dot( Ds, Ds );
    F = Dot( Ds, Dt );
    G = Dot( Dt, Dt );

    n = Cross( Ds, Dt );
    n.Normalize();
    R3Vec refNormal;
    refNormal = m_Param.Nref();

    factor = 1.0;
    if( refNormal[0] != 0.0 ||  refNormal[1] != 0.0 ||  refNormal[2] != 0.0 ) {
        if( Dot(n,refNormal) < 0.0 ) factor = -1.0;
    }

	if( factor > 0.0 ) {
	    out_data.setNormal(n);
	} else {
	    out_data.setNormal(R3Vec(-n[0], -n[1], -n[2]));
	}

    L = -Dot( Dss, n );
    M = -Dot( Dst, n );
    N = -Dot( Dtt, n );

    out_dArea = sqrt( E*G - F*F );

	fC0 = L*N - M*M;
	fC1 = 2.0*F*M - (E*N + G*L);
	fC2 = E*G - F*F;

	fTemp = fC1*fC1 - 4.0*fC0*fC2;
	fTemp = (fTemp < 0.0) ? sqrt(-fTemp) : sqrt(fTemp);

	fMult = (fC2 == 0.0) ? 1.0 : 0.5/fC2;

	k1 = -fMult*( fC1+fTemp);
	k2 =  fMult*(-fC1+fTemp);

	fA0 = M - k1*F;
	fA1 = k1*E - L;
	fLength = sqrt(fA0*fA0 + fA1*fA1);

	if( fLength > epsilon ) {
		v1 = fA0*Ds + fA1*Dt;
	} else {
		fA0 = N - k1*G;
		fA1 = k1*F - M;
		fLength = sqrt(fA0*fA0 + fA1*fA1);

		if( fLength > epsilon ) {
			v1 = fA0*Ds + fA1*Dt;
		} else {
			fA0 = 1.0;
			fA1 = 0.0;
			v1 = Ds;
		}
	}

	v1.Normalize();
	v2 = Cross(v1, n);
	v2.Normalize();

	if( factor < 0.0 ) {
		k1 = -k1;
		k2 = -k2;
	}
    
    //if( k1*k1 < k2*k2 ) {
    if( k1 < k2 ) {
        out_data.setMaxMin(k2,k1);

        out_data.setMaxMinVecPar( R2Vec(fA1,  fA0),
                                  R2Vec(fA0, -fA1) );

        out_data.setMaxMinVec( v2, v1 );
    } else {
        out_data.setMaxMin(k1,k2);

        out_data.setMaxMinVecPar( R2Vec(fA0,  fA1),
                                  R2Vec(fA1, -fA0) );

        out_data.setMaxMinVec( v1, v2 );
    }
}
