#include "StdAfx.H"
#include <MC/MC_Point.H>
#include <MC/MC_Data.H>
#include <MC/MC_Param.H>
#include <MC/MC_MeshSubset.H>
#include <MC/MC_FlattenedSurface.H>
#include <MC/MC_MeshUtils.h>
#include <utils/Pm_MeshLite.H>

MCFlattenedSurface::MCFlattenedSurface(MCData &in_data)
: m_pCurvature(in_data)
{
    MCParam param( *m_pCurvature.Mesh(), 999,MCParam::NATURAL );
    param.CreateFull();
    double lambda = 1.0;
    double mu = 0.0;
    param.CalculateUVnatural(lambda, mu);
    param.MakeMesh(m_FlatMesh);
}

MCFlattenedSurface::MCFlattenedSurface(MCData &in_data, MCParam::MC_ParamType in_type,
									   const double in_lambda, const double in_mu)
: m_pCurvature(in_data)
{
    MCParam param( *m_pCurvature.Mesh(), 999,in_type );
    param.CreateFull();
    switch( in_type ) {
	case MCParam::PLANAR:
		param.CalculateUVplanar();
		break;
	case MCParam::CIRCULAR:
		param.CalculateUVcircular(in_lambda, in_mu);
		break;
	case MCParam::NATURAL:
		param.CalculateUVnatural(in_lambda, in_mu);
		break;
	default:
		param.CalculateUVnatural(in_lambda, in_mu);
		break;
	}
    param.MakeMesh(m_FlatMesh);
}

MCFlattenedSurface::~MCFlattenedSurface()
{
}

void 
MCFlattenedSurface::Init(const int method)
{
}

void 
MCFlattenedSurface::Init(PMeshLite &in_flatMesh, MCData &in_data)
{
}

MCData& 
MCFlattenedSurface::Data()
{
    return m_pCurvature;
}
 
/*
PMeshLite* 
MCFlattenedSurface::Mesh3D()
{
    return m_pCurvature.Mesh();
}
*/

int 
MCFlattenedSurface::AreaCorrect(int method, double BBfactor, double SZfactor)
{
    const PMeshLite *orig;
    Array <Vertex> aoVorig;
    Array <Vertex> aoVflat;
    Array <Face> aoForig;
    Array <Face> aoFflat;
    Array <Edge> aoEorig;
    Array <Edge> aoEflat;
    Array <Vertex> aov;
    Array <Edge> aoe;
    Edge e1, e2;
    Vertex v1, v2, v3, v4;
    R3Pt P, Pmin, Pmax;
    R3Vec V;
    Array<double> edgeLength;
    Array<double> edgeRatio;
    Array<double> temp;
    Array<double> nodeSize;
    Array<int> nodeEdges;
    int i, j, nv, ne, nf, id1, id2, ipt, iCell, jCell, ier, num_error;
    double d1, d2, r, medianLength, size, epsilon;
    double s, t, err, max_error;
    double u, v, umin, umax, vmin, vmax;
    double x, y;
    int nBdryEdge;
    double bdryEdgeRatio;

    orig = m_pCurvature.Mesh();
    orig->AllMeshVertices(aoVorig);
    orig->AllMeshFaces(aoForig);
    orig->AllMeshEdges(aoEorig);

    m_FlatMesh.AllMeshVertices(aoVflat);
    m_FlatMesh.AllMeshFaces(aoFflat);
    m_FlatMesh.AllMeshEdges(aoEflat);

    nv = orig->NumVertices();
    nf = orig->NumFaces();
    ne = orig->NumEdges();

    edgeLength.need(ne);
    edgeRatio.need(ne);
    nodeSize.need(nv);
    nodeEdges.need(nv);

    // Calculate edge data
    nBdryEdge = 0;
    bdryEdgeRatio = 0.0;
    for( i=0; i<ne; i++ ) {
        e1 = aoEorig[i];
        e2 = aoEflat[i];
        id1 = aoEorig[i]->IndexId();
        id2 = aoEflat[i]->IndexId();
        v1 = orig->EdgeVertex1(e1);
        v2 = orig->EdgeVertex2(e1);
        v3 = m_FlatMesh.EdgeVertex1(e2);
        v4 = m_FlatMesh.EdgeVertex2(e2);
        V = v2->GetLoc() - v1->GetLoc();
        d1 = sqrt(V[0]*V[0] + V[1]*V[1] + V[2]*V[2]);
        V = v4->GetLoc() - v3->GetLoc();
        d2 = sqrt(V[0]*V[0] + V[1]*V[1] + V[2]*V[2]);
        //edgeLength.set(id2,d2);
        edgeLength[id2] = d2;
        r = (d1 == 0.0) ? 1.0 : d2/d1;
        //edgeRatio.set(id2,r);
        edgeRatio[id2] = r;
        if( orig->IsBoundary(e1) ) {
            nBdryEdge++;
            bdryEdgeRatio += r;
        }
    }
    if( nBdryEdge > 0 ) bdryEdgeRatio /= nBdryEdge;
    bdryEdgeRatio = (bdryEdgeRatio == 0.0) ? 1.0 : 1.0/bdryEdgeRatio;

    // Find the median edge length
    for( i=0; i<ne; i++ ) { temp += edgeLength[i]; }
    edgeLength.sort( temp );
    medianLength = edgeLength[ne/2];

    TRACE("No. of boundary edges: %d, Avg. boundary ratio: %lf\n", 
        nBdryEdge, bdryEdgeRatio );
    TRACE("Median edge length: %lf\n", medianLength);

    // Compute node sizing function
    for( i=0; i<nv; i++ ) {
        size = 0.0;
        aoe = m_FlatMesh.VertexEdges(aoVflat[i]);
        for( j=0; j<aoe.num(); j++ ) {
            size += edgeRatio[ aoe[j]->IndexId() ];
        }
        if( aoe.num() > 0 ) size /= (double) (aoe.num());
        //nodeSize.set(i,size);
        nodeSize[i] = size*bdryEdgeRatio;

        // Also get the bounding box
        BoundingBox(aoVflat, Pmin, Pmax);
    }

    // Expand the bounding box by a factor
    umin = Pmin[0] - 0.5*(BBfactor-1.0)*(Pmax[0]-Pmin[0]);
    umax = Pmax[0] + 0.5*(BBfactor-1.0)*(Pmax[0]-Pmin[0]);
    vmin = Pmin[1] - 0.5*(BBfactor-1.0)*(Pmax[1]-Pmin[1]);
    vmax = Pmax[1] + 0.5*(BBfactor-1.0)*(Pmax[1]-Pmin[1]);

    // Make a grid
    Array<R2Pt> grid;
    Array<double> gridNodeSize;
    Array<double> error;

    int ni, nj;
    medianLength *= SZfactor;

    ni = (int) ((umax - umin)/medianLength);
    nj = (int) ((vmax - vmin)/medianLength);
    if( ni < 11 ) ni = 11;
    if( nj < 11 ) nj = 11;
    //ni = 21;
    //nj = 21;

    for( j=0; j<nj; j++ ) {
        v = (vmin*(nj-(j+1)) + vmax*(j))/(nj-1);
        for( i=0; i<ni; i++ ) {
            u = (umin*(ni-(i+1)) + umax*(i))/(ni-1);
            grid += R2Pt(u,v);
            //gridNodeSize += 1.0;
        }
    }
    gridNodeSize.needSet(ni*nj, 1.0);

    // Transfer mesh sizes to grid
    epsilon = 0.01 * medianLength;

    TRACE("Interpolating desired size onto grid nodes\n");

    InterpolateToGrid( grid, gridNodeSize, ni, nj, 
                       umin, umax, vmin, vmax,
                       m_FlatMesh, aoFflat, nodeSize, epsilon );

    // Set the size function for edges of the grid

    TRACE("Computing desired edge size\n");

    // method: 0=transform, 1=inverse transform
    if( method == 1 ) {
        for( j=0; j<nj; j++ ) {
            for( i=0; i<ni; i++ ) {
                ipt = i + j*ni;
                if( gridNodeSize[ipt] != 0.0 ) {
                    gridNodeSize[ipt] = 1.0/gridNodeSize[ipt];
                }
            }
        }
        // Could check and set to max if size == 0.0
    }

    Array<R2Pt> gridEdgeSize;
    gridEdgeSize.needSet(ni*nj,R2Pt(1.0,1.0));

    for( j=0; j<nj-1; j++ ) {
        for( i=0; i<ni-1; i++ ) {
            ipt = i + j*ni;
            gridEdgeSize[ipt][0] = 0.5 *( gridNodeSize[ipt] +
                                          gridNodeSize[ipt+1] );
            gridEdgeSize[ipt][1] = 0.5 *( gridNodeSize[ipt] +
                                          gridNodeSize[ipt+ni] );
        }
    }
        
    //TRACE("Writing original grid (%dx%d)\n", ni, nj);

    // Write out the original grid
    //MCWriteGridData(WINDataHome("grid-orig.txt"), ni, nj, 
    //                            grid, gridNodeSize);  // Original grid

    //TRACE("Done Writing original grid as a mesh\n");

    // Relax the grid

    int max_it;
    double residual, tol;

    TRACE("Smoothing grid\n");

    max_it = 1000;
    tol = 0.000001*medianLength*bdryEdgeRatio;
    residual = LaplacianSmoothing(grid, gridEdgeSize, ni, nj, max_it, tol );

    //TRACE("Writing relaxed grid (%dx%d)\n", ni, nj);

    // Write out the modified grid
    //MCWriteGridData(WINDataHome("grid-mod.txt"), ni, nj, 
    //                grid, gridNodeSize);  // Relaxed grid

    //TRACE("Writing relaxed grid as a mesh\n");

    TRACE("Searching for new mesh locations\n");

    // Compute new locations for nodes in the flat mesh
    max_error = 0.0;
    num_error = 0;
    if( method == 1 ) {
        for( i=0; i<aoVflat.num(); i++ ) {
            P = aoVflat[i]->GetLoc();
            FindCell( P[0], P[1], ni, nj, umin, umax, vmin, vmax,
                      iCell, jCell, s, t );

            ier = GridInterp( grid, ni, nj, iCell, jCell, s,  t, x, y );
            if( ier == 0 ) {
                aoVflat[i]->SetLoc(R3Pt(x, y, 0.0));
            } else {
                num_error++;
            }
        }
    } else {
        for( i=0; i<aoVflat.num(); i++ ) {
            P = aoVflat[i]->GetLoc();
            FindCell(grid, ni, nj, P[0], P[1], iCell, jCell, s, t, err );
            if( err > max_error ) max_error = err;
    
            ier = GridInterp( iCell, jCell, s,  t, 
                              umin, umax, vmin, vmax, ni, nj, x, y );
            if( ier == 0 ) {
                aoVflat[i]->SetLoc(R3Pt(x, y, 0.0));
            } else {
                num_error++;
            }
        }
    }

    TRACE("Max FindCell error = %lf", max_error);
    TRACE(", No. of interpolation error = %d\n", num_error);
    TRACE("*** Done ***\n");

    // Restrict flattened mesh coordinates to the unit square 
    NormalizeUVCoords(aoVflat);

    // Dump data
    FILE *out = NULL;
    //out = fopen(WINDataHome("flat.dat"),"w");
    if( out != NULL ) {
        fprintf(out,"Median edge length = %lf\n", medianLength);
        fprintf(out,"Bounds: min(u,v) = (%lf,%lf)\n", Pmin[0], Pmin[1] );
        fprintf(out,"      : max(u,v) = (%lf,%lf)\n", Pmax[0], Pmax[1] );
        fprintf(out,"Expand: min(u,v) = (%lf,%lf)\n", umin, vmin );
        fprintf(out,"      : max(u,v) = (%lf,%lf)\n", umax, vmax );
        for( i=0; i<ne; i++ ) {
            fprintf(out,"Edge %5d: Len= %15.5lf, Ratio= %15.5lf\n",
                i, edgeLength[i], edgeRatio[i] );
        }
        for( i=0; i<nv; i++ ) {
            fprintf(out,"Node %5d: Size function = %15.5lf\n",
                i, nodeSize[i] );
        }
        for( j=0; j<nj; j++ ) {
            for( i=0; i<ni; i++ ) {
                ipt = i + j*ni;
                fprintf(out,"%4d %4d: Xsize= %15.5lf, Ysize= %15.5lf\n",
                            i, j, gridEdgeSize[ipt][0], gridEdgeSize[ipt][1] );
            }
        }
        fclose(out);
    }
    return 0;
}

R3Pt 
MCFlattenedSurface::ComputeCentroid(void)
{
    int i;
    double weight, curvature, x, y, z;
    R3Pt P, Q;
    MCPoint cpt;
    Array <Vertex> aoVs;

    m_FlatMesh.AllMeshVertices(aoVs);

    weight = 0.0;
    x = 0.0;
    y = 0.0;
    z = 0.0;
    for( i=0; i<aoVs.num(); i++ ) {
        m_pCurvature.getPoint(aoVs[i],cpt);
        curvature = (cpt.Kg()<0.0) ? -cpt.Kg() : cpt.Kg();
        weight += curvature;

        Q = aoVs[i]->GetLoc();
        x += Q[0] * curvature;
        y += Q[1] * curvature;
        z += Q[2] * curvature;
    }
    if( weight == 0.0 ) return R3Pt(0.0,0.0,0.0);
    
    return R3Pt(x/weight, y/weight, z/weight);
}

void 
MCFlattenedSurface::AngleBins( int method, const R3Pt &centroid, int nBins, 
                               Array<double> &Ksum )
{
    int i, j, k, ilast, inext, count, iBin;
    double curvature, angle, pi;
    R3Pt P, Q;
    MCPoint cpt;
    Array <Vertex> aoVs;
    Array <int> Nsum;

    FILE *out = NULL;
    //out = fopen(WINDataHome("bin.dat"),"a");

    pi = 4.0 * atan(1.0);

    m_FlatMesh.AllMeshVertices(aoVs);

    Ksum.need(nBins);
    Nsum.need(nBins);
    for( i=0; i<nBins; i++ ) {
        Ksum[i] = 0.0;
        Nsum[i] = 0;
    }
    
    for( i=0; i<aoVs.num(); i++ ) {
        m_pCurvature.getPoint(aoVs[i],cpt);
        switch(method) {
            case 0: // Average of Gaussian curvature
            case 3: // Maximum of Gaussian curvature
                curvature = cpt.Kg();
                break;
            case 1: // Average of Mean curvature
            case 4: // Maximum of Mean curvature
                curvature = cpt.Kbar();
                break;
            case 2: // Average of maximum curvature
            case 5: // Maximum of maximum curvature
                curvature = cpt.K1();
                break;
            default:
                return;
        }
        
        Q = aoVs[i]->GetLoc();
        angle = atan2(Q[0]-centroid[0],Q[1]-centroid[1]);
        while( angle < 0.0 ) angle += (2.0*pi);
        while( angle >= 2.0*pi ) angle -= (2.0*pi);

        iBin = (int) (angle*nBins/(2.0*pi));
        while(iBin<0) iBin += nBins;
        while(iBin>=nBins) iBin -= nBins;

        if( out != NULL ) {
            fprintf(out,"%3d: angle = %12.5lf, bin = %3d, curvature = %lf\n", 
                        i, angle*180.0/pi, iBin, curvature);
        }

        if( method >= 3 && method <= 5 ) {
            if( curvature*curvature > Ksum[iBin]*Ksum[iBin] ) {
               Ksum[iBin] = curvature;
            }
        } else {
            Ksum[iBin] += curvature;
        }
        Nsum[iBin]++;
    }

    if( method >= 0 && method <= 2 ) {
        for( i=0; i<nBins; i++ ) {
            if( Nsum[i] != 0 ) Ksum[i] /= Nsum[i];
        }
    }

    // Average neighboring bins for bins without data
    for( i=0; i<nBins; i++ ) {
        if( Nsum[i] == 0 ) {

            // Find previous bin with data
            ilast = i-1;
            if( i == 0 ) {
                for( j=nBins-1; j>0; j-- ) {
                    if( Nsum[j] != 0 ) {
                        ilast = j;
                        break;
                    }
                }
            }
            if( ilast < 0 ) break;   // No data found

            // Find next bin with data
            inext = 0;
            for( j=i+1; j<nBins; j++ ) {
                if( Nsum[j] != 0 ) {
                    inext = j;
                    break;
                }
            }

            // Count number of empty bins
            count = inext - (ilast+1);
            if( ilast > inext ) count += nBins;

            // Interpolate data
            for( j=0; j<count; j++ ) {
                k = i+j;
                if( k >= nBins ) k -= nBins;
                Ksum[k] = ((count-j)*Ksum[ilast] + 
                               (j+1)*Ksum[inext])/(count+1);
                Nsum[k] = -1;
            }
        }
    }

    if( out != NULL ) {
        for( i=0; i<nBins; i++ ) {
            fprintf(out,"Bin %3d: %15.7lf, %d\n", i, Ksum[i], Nsum[i]);
        }
        fclose(out);
    }
}

void 
MCFlattenedSurface::BinPoints( int method, int nBins, 
                               int iBin1, int iBin2, int iBin3, int iBin4,
                               R3Pt &P1, R3Pt &P2, R3Pt &P3, R3Pt &P4 )
{
    int i, iBin;
    double curvature, angle, pi, k1, k2, k3, k4;
    R3Pt P, Q;
    MCPoint cpt;
    Array <Vertex> aoVf;
    Array <Vertex> aoVs;

    m_FlatMesh.AllMeshVertices(aoVf);
    m_pCurvature.Mesh()->AllMeshVertices(aoVs);

    pi = 4.0*atan(1.0);
    k1 = 0.0;
    k2 = 0.0;
    k3 = 0.0;
    k4 = 0.0;
    P = ComputeCentroid();
    
    for( i=0; i<aoVs.num(); i++ ) {
        m_pCurvature.getPoint(aoVf[i],cpt);
        switch(method) {
            case 0: // Average of Gaussian curvature
            case 3: // Maximum of Gaussian curvature
                curvature = cpt.Kg();
                break;
            case 1: // Average of Mean curvature
            case 4: // Maximum of Mean curvature
                curvature = cpt.Kbar();
                break;
            case 2: // Average of maximum curvature
            case 5: // Maximum of maximum curvature
                curvature = cpt.K1();
                break;
            default:
                return;
        }
        
        Q = aoVf[i]->GetLoc();
        angle = atan2(Q[0]-P[0],Q[1]-P[1]);
        while( angle < 0.0 ) angle += (2.0*pi);
        while( angle >= 2.0*pi ) angle -= (2.0*pi);

        iBin = (int) (angle*nBins/(2.0*pi));
        while(iBin<0) iBin += nBins;
        while(iBin>=nBins) iBin -= nBins;

        if( iBin == iBin1 ) {
            if( curvature*curvature > k1*k1 ) {
                k1 = curvature;
                P1 = aoVs[i]->GetLoc();
            }
            continue;
        }

        if( iBin == iBin2 ) {
            if( curvature*curvature > k2*k2 ) {
                k2 = curvature;
                P2 = aoVs[i]->GetLoc();
            }
            continue;
        }

        if( iBin == iBin3 ) {
            if( curvature*curvature > k3*k3 ) {
                k3 = curvature;
                P3 = aoVs[i]->GetLoc();
            }
            continue;
        }

        if( iBin == iBin4 ) {
            if( curvature*curvature > k4*k4 ) {
                k4 = curvature;
                P4 = aoVs[i]->GetLoc();
            }
        }   
    }
}

void 
MCFlattenedSurface::AlignUV(const R3Pt &center, const R3Pt &newCenter, 
                            const double angle, const int flip)
{
    int i, nv;
    double x, y, z, x1, y1, z1, x2, y2;
    Array <Vertex> aoVflat;
    R3Pt P;

    m_FlatMesh.AllMeshVertices(aoVflat);
    nv = aoVflat.num();
    
    for( i=0; i<nv; i++ ) {
       P = aoVflat[i]->GetLoc();

       // Translate center to zero
       x1 = P[0] - center[0];
       y1 = P[1] - center[1];
       z1 = P[2] - center[2];
       
       // Flip
       if( flip ) x1 = -x1;

       // Rotate
       x2 = x1*cos(angle) - y1*sin(angle);
       y2 = x1*sin(angle) + y1*cos(angle);
       
       // Translate to new center
       x = x2 + newCenter[0];
       y = y2 + newCenter[1];
       z = z1 + newCenter[2];

       aoVflat[i]->SetLoc(R3Pt(x, y, z));
    }
    if( flip ) m_FlatMesh.Flip();
}

int 
MCFlattenedSurface::OrientUV( MCFlattenedSurface &ref, int metric, 
                              int nBins, int nSep, FILE *out )
{
    int flip;
    double angle, pi;
    R3Pt C_A, C_B;
    Array<double> kBinA;
    Array<double> kBinB;
    //Array<double> matrix;
    //ofstream os;
    //PMeshLite mesh;

    //int iBin1, iBin2, iBin3, iBin4;
    //R3Pt A1, A2, A3, A4, B1, B2, B3, B4;
    //R3Pt P1, P2, P3, P4;
    //R3Pt T1, T2;

    //int flex = 2;
    
    pi = 4.0 * atan(1.0);

    if( nBins < 4 ) {
        TRACE("Error: Too few bins\n");
        return -1;
    }

    if( metric < 0 || metric > 5 ) {
        TRACE("Error: metric out of range(0-5)\n");
        return -2;
    }
    //metric = 0; // Average of Gaussian curvature
    //metric = 1; // Average of Mean curvature
    //metric = 2; // Average of maximum curvature
    //metric = 3; // Maximum of Gaussian curvature
    //metric = 4; // Maximum of Mean curvature
    //metric = 5; // Maximum of maximum curvature
            
    C_A = ref.ComputeCentroid();
    C_B = this->ComputeCentroid();

    ref.AngleBins(metric, C_A, nBins, kBinA );
    this->AngleBins(metric, C_B, nBins, kBinB );

    AlignBins(kBinA, kBinB, angle, flip);

    if( out != NULL ) {
        fprintf(out,"metric %d\n", metric);
        //for( int i=0; i<nBins; i++ ) {
        //    fprintf(out,"Bin %3d: %15.7lf, %15.7lf\n", i, kBinA[i], kBinB[i]);
        //}
        fprintf(out,"Centroid A (%12.5lf,%12.5lf,%12.5lf)\n", 
                    C_A[0], C_A[1], C_A[2] );
        fprintf(out,"Centroid B (%12.5lf,%12.5lf,%12.5lf)\n", 
                    C_B[0], C_B[1], C_B[2] );
        fprintf(out,"--- Angle = %lf (%lf), flip = %d\n", 
                    angle, angle*180.0/pi, flip );
    }

    this->AlignUV(C_B, C_A, angle, flip);

    return 0;
}

void 
MCFlattenedSurface::CurvatureAlign( MCFlattenedSurface &ref, 
                                    int metric, int nPts, int nSep,
                                    int flip, int flex )
{
    // Aligns the current 3D object to a specified 3D object
    int i, ier;
    double curvature;
    R3Pt P, Q, P1, P2, P3, P4;
    MCPoint cpt;
    Array <Vertex> aoVf;
    Array <Vertex> aoVr;
    Array <Vertex> aoVs;

    Array<double> values;
    Array<R3Pt> PA, PB;
    Array<double> VA, VB;
    Array<int> IA, IB, CA, CB;

    PMeshLite mesh;

    // Find the highest curvature points for this mesh
    m_FlatMesh.AllMeshVertices(aoVf);
    values.need(0);
    for( i=0; i<aoVf.num(); i++ ) {
        m_pCurvature.getPoint(aoVf[i],cpt);
        switch(metric) {
            case 0: // Average of Gaussian curvature
            case 3: // Maximum of Gaussian curvature
                curvature = cpt.Kg();
                break;
            case 1: // Average of Mean curvature
            case 4: // Maximum of Mean curvature
                curvature = cpt.Kbar();
                break;
            case 2: // Average of maximum curvature
            case 5: // Maximum of maximum curvature
                curvature = cpt.K1();
                break;
            default:
                return;
        }
        values += curvature;
    }
    ier = FindMaxPoints(m_FlatMesh, values, nPts, nSep, PA, VA, IA);
        
    // Find the highest curvature points for the reference mesh
    (ref.Mesh()).AllMeshVertices(aoVr);
    values.need(0);
    for( i=0; i<aoVr.num(); i++ ) {
        m_pCurvature.getPoint(aoVr[i],cpt);
        switch(metric) {
            case 0: // Average of Gaussian curvature
            case 3: // Maximum of Gaussian curvature
                curvature = cpt.Kg();
                break;
            case 1: // Average of Mean curvature
            case 4: // Maximum of Mean curvature
                curvature = cpt.Kbar();
                break;
            case 2: // Average of maximum curvature
            case 5: // Maximum of maximum curvature
                curvature = cpt.K1();
                break;
            default:
                return;
        }
        values += curvature;
    }
    ier = FindMaxPoints(ref.Mesh(), values, nPts, nSep, PB, VB, IB);

    // Find the best correspondence between the high curvature points
    ier = BestCorrespondence(PA, VA, IA, PB, VB, IB, CA, CB);

    // Now get the corresponding 3D points
    PA.need(0);
    PB.need(0);

    m_pCurvature.Mesh()->AllMeshVertices(aoVs);
    PA += aoVs[IA[CA[0]]]->GetLoc();
    PA += aoVs[IA[CA[1]]]->GetLoc();
    PA += aoVs[IA[CA[2]]]->GetLoc();
    PA += aoVs[IA[CA[3]]]->GetLoc();

    mesh = *(ref.Data().Mesh());
    
    mesh.AllMeshVertices(aoVr);
    PB += aoVr[IB[CB[0]]]->GetLoc();
    PB += aoVr[IB[CB[1]]]->GetLoc();
    PB += aoVr[IB[CB[2]]]->GetLoc();
    PB += aoVr[IB[CB[3]]]->GetLoc();

    // Compute the transformation
    R3Pt T1, T2;
    Array<double> matrix;

    TransMatrix(PA[0], PA[1], PA[2], PA[3],
                PB[0], PB[1], PB[2], PB[3], 
                flex, T1, T2, matrix);

    DumpMatrix(WINDataHome("matrices.dat"), "Max Transform", T1, T2, matrix);

    Transform(mesh, T1, T2, matrix);

    MCWrite4PointMesh(WINDataHome("pointsAorig.txt"), 
                                  PA[0], PA[1], PA[2], PA[3] );
    
    MCWrite4PointMesh(WINDataHome("pointsBorig.txt"), 
                                  PB[0], PB[1], PB[2], PB[3] );
   
    P1 = Transform(PB[0], T1, T2, matrix);
    P2 = Transform(PB[1], T1, T2, matrix);
    P3 = Transform(PB[2], T1, T2, matrix);
    P4 = Transform(PB[3], T1, T2, matrix);

    MCWrite4PointMesh(WINDataHome("pointsBmod.txt"), P1, P2, P3, P4 );
    
    
}

void 
MCFlattenedSurface::WriteMeshData( char *filename )
{
    MCWriteMeshData(filename, m_FlatMesh, m_pCurvature);
}

void 
MCFlattenedSurface::dump(FILE *out)
{
    m_pCurvature.dump(out);
}

void
MCFlattenedSurface::test(void)
{
    int i, nd, nf, nv, id1, id2, count;

    FILE *out = NULL;
    out = fopen(WINDataHome("testflat.dat"),"a");

    nf = m_FlatMesh.NumVertices();
    nv = m_pCurvature.Mesh()->NumVertices();
    nd = m_pCurvature.NumPoints();
    
    if( out != NULL ) {
        fprintf(out,"Mesh vertices: %d\n", nv);
        fprintf(out,"FlatMesh vertices: %d\n", nf);
        fprintf(out,"Data points: %d\n", nd);
    }

    Array<Vertex> aoVs;
    Array<Vertex> aoVf;
        
    if( nf == nv ) {
        m_pCurvature.Mesh()->AllMeshVertices(aoVs);
        m_FlatMesh.AllMeshVertices(aoVf);

        count = 0;
        for( i=0; i<nv; i++ ) {
            id1 = aoVs[i]->IndexId();
            id2 = aoVf[i]->IndexId();
            if( id1 != i || id2 != i ) {
                if( out != NULL ) {
                    count++;
                    fprintf(out,"%d: Id1: %d, Id2: %d\n", i, id1, id2);
                }
            }
        }
        if( out != NULL ) {
            if( count == 0 ) fprintf(out,"All indices match\n");
        }
    } else {
        if( out != NULL ) {
            fprintf(out,"WARNING: NV = %d, NF = %d, ND = %d\n", nv, nf, nd);
        }
    }

    if( out != NULL ) {

        fclose(out);
        out = NULL;
    }
}
