#include "StdAfx.H"
#include <MC/MC_MeshSubset.H>
#include <utils/Pm_MeshLite.H>

MCMeshSubset::MCMeshSubset(const PMeshLite &in_mesh, 
                           const int in_nRings )
: m_pMesh(in_mesh)
, m_iRings(in_nRings)
, valid(FALSE)
{
}
    
MCMeshSubset::~MCMeshSubset()
{
}

void
MCMeshSubset::Create(const Vertex &in_vertex, const int in_nRings)
{
    int ier;

    if( !valid || in_vertex != m_Vertex || in_nRings != m_iRings ) {
        // Extract the N-ring mesh
        ier = m_pMesh.GrowRings(in_vertex, in_nRings, m_aopVs, m_aopFs);
        m_Vertex = in_vertex;
        m_iRings = in_nRings;
        m_Nref = m_Vertex->GetNorm();
        valid = TRUE;
        m_BoundaryEdges.need(0);
        m_BoundaryVertexIndices.need(0);
    }
}

void
MCMeshSubset::CreateFull()
{
    if( !valid ) {
        // Extract the entire mesh
        m_pMesh.AllMeshVertices(m_aopVs);
        m_pMesh.AllMeshFaces(m_aopFs);
        m_Vertex = m_aopVs[0];
        m_iRings = 999;
        m_Nref = m_Vertex->GetNorm();
        valid = TRUE;
        m_BoundaryEdges.need(0);
        m_BoundaryVertexIndices.need(0);
    }
}

int 
MCMeshSubset::ExtractMesh(PMeshLite &new_mesh)
{
    return m_pMesh.ExtractCopy( m_aopFs, new_mesh );
}

void
MCMeshSubset::SetBoundary() 
{
    int i, j, i1, i2, index, index1, index2;
    Array<Edge> aoe;
    Face f1;
    Face f2;

    m_BoundaryEdges.need(0);
    for( i=0; i<m_aopFs.num(); i++ ) {
        aoe = m_pMesh.FaceEdges(m_aopFs[i]);
        for( j=0; j<aoe.num(); j++ ) {
            f1 = m_pMesh.EdgeFace1(aoe[j]);
            f2 = m_pMesh.EdgeFace2(aoe[j]);
            if( f1 != NULL ) {
                index1 = m_aopFs.index(f1);
            } else {
                index1 = -1;
            }
            if( f2 != NULL ) {
                index2 = m_aopFs.index(f2);
            } else {
                index2 = -1;
            }
            if( index1 == -1 || index2 == -1 ) {
                if( m_BoundaryEdges.index(aoe[j]) == -1 ) {
                    m_BoundaryEdges += aoe[j];
                }
            }
        }
    }

    m_BoundaryVertexIndices.need(0);
    for( i=0; i<m_BoundaryEdges.num(); i++ ) {
        i1 = m_pMesh.EdgeVertex1(m_BoundaryEdges[i])->IndexId();
        index = m_BoundaryVertexIndices.index(i1);
        if( index == -1 ) m_BoundaryVertexIndices += i1;
        i2 = m_pMesh.EdgeVertex2(m_BoundaryEdges[i])->IndexId();
        index = m_BoundaryVertexIndices.index(i2);
        if( index == -1 ) m_BoundaryVertexIndices += i2;
    }
}

int
MCMeshSubset::ParameterizeBoundary(Array<int> &out_Order, 
                                   Array<R2Pt> &out_aopUVs )
{
    int i, n, first, last, nbv, bv, index;
    double total, s, theta, pi;
    Array< double > d;
    SArray< int > EdgeIndex1;
    SArray< int > EdgeIndex2;
    Vertex vertex;
    Edge edge;
    Vertex lastVertex;
    Edge lastEdge;

    FILE *out = NULL;
    //out = fopen(WINDataHome("edges.dat"),"a");

    out_Order.need(0);
    out_aopUVs.need(0);

    pi = 4.0 * atan(1.0);

    if(m_BoundaryEdges.num() == 0 ) {
        SetBoundary();
        if(m_BoundaryEdges.num() == 0) return -1;
    }
    nbv = m_BoundaryVertexIndices.num();
    EdgeIndex1.init(nbv);
    EdgeIndex2.init(nbv);
    for( i=0; i<nbv; i++ ) {
        EdgeIndex1[i] = -1;
        EdgeIndex2[i] = -1;
    }

    for( i=0; i<nbv; i++ ) {
        edge = m_BoundaryEdges[i];
        vertex = m_pMesh.EdgeVertex1(edge);
        index = m_BoundaryVertexIndices.index(vertex->IndexId());
        if( EdgeIndex1[index] == -1 ) {
            EdgeIndex1[index] = i;
        } else {
            EdgeIndex2[index] = i;
        }
        vertex = m_pMesh.EdgeVertex2(edge);
        index = m_BoundaryVertexIndices.index(vertex->IndexId());
        if( EdgeIndex1[index] == -1 ) {
            EdgeIndex1[index] = i;
        } else {
            EdgeIndex2[index] = i;
        }
    }

    if(out != NULL) {
        for( i=0; i<nbv; i++ ) {
            fprintf(out,"m_BoundaryVertexIndices[%d] = %d\n", 
                        i, m_BoundaryVertexIndices[i] );
            fprintf(out,"Index %d: EdgeIndex1= %d, EdgeIndex2= %d\n", 
                        i, EdgeIndex1[i], EdgeIndex2[i] );
        }
    }

    // Find the first boundary vertex
    edge = m_BoundaryEdges[0];
    vertex = m_pMesh.EdgeVertex1(edge);
    first = vertex->IndexId();
    last = first;
    lastVertex = vertex;
    lastEdge = edge;
    bv = first;

    int VertexIndex, Edge1Index, Edge2Index;

    // March around the boundary 
    n = 0;
    do {
        // Store the current vertex 
        n++;
        out_Order += bv;
        if( n>1 ) d += Distance( vertex, lastVertex );
        last = bv;
        lastVertex = vertex;

        // Find the next edge
        VertexIndex = m_BoundaryVertexIndices.index(bv);
        Edge1Index = EdgeIndex1[VertexIndex];
        Edge2Index = EdgeIndex2[VertexIndex];
        if( m_BoundaryEdges[Edge1Index] == edge ) {
            edge = m_BoundaryEdges[Edge2Index];
        } else {
            edge = m_BoundaryEdges[Edge1Index];
        }

        // Find the next vertex
        vertex = m_pMesh.EdgeVertex1(edge);
        if( vertex == lastVertex ) {
            vertex = m_pMesh.EdgeVertex2(edge);
        } else {
            vertex = m_pMesh.EdgeVertex1(edge);
        }
        bv = vertex->IndexId();
    } 
    while( bv != first );
    d += Distance( vertex, lastVertex );

    // Get relative distance between points 
    total = 0.0;
    for( i=0; i<d.num(); i++ ) {
        total += d[i];
    }

    s = 0.0;
    for( i=0; i<d.num(); i++ ) {
        s += d[i];
        theta = 2.0*pi*s/total;
        out_aopUVs += R2Pt( 0.5*(1.0+sin(theta)), 0.5*(1.0+cos(theta)) );
    }

    if(out != NULL) {
        for( i=0; i<nbv; i++ ) {
            fprintf(out,"Point %d: Order= %d, u= %lf, v=%lf\n", 
                        i, out_Order[i], out_aopUVs[i][0], out_aopUVs[i][1] );
        }
    }

    if(out != NULL) {
        fclose(out);
        out = NULL;
    }
	return n;
}

int
MCMeshSubset::BoundaryExtremePoints(int &index1, int &index2 )
{
    int i, j, nv;
    double d, dmax;
    Vertex vertex1, vertex2;

    if(m_BoundaryEdges.num() == 0 ) {
        SetBoundary();
        if(m_BoundaryEdges.num() == 0) return -1;
    }
    
    nv = m_BoundaryVertexIndices.num();
    for( i=0; i<nv-1;i++ ) {
        vertex1 = m_pMesh.VertexFromIndexId(m_BoundaryVertexIndices[i]);
        for( j=i+1; j<nv; j++ ) {
            vertex2 = m_pMesh.VertexFromIndexId(m_BoundaryVertexIndices[j]);
        
            d = Distance(vertex1, vertex2);
            if( i==0 && j==1 ) {
                index1 = m_BoundaryVertexIndices.index(vertex1->IndexId());
                index2 = m_BoundaryVertexIndices.index(vertex2->IndexId());
                dmax = d;
            } else {
                if( d > dmax ) {
                    index1 = m_BoundaryVertexIndices.index(vertex1->IndexId());
                    index2 = m_BoundaryVertexIndices.index(vertex2->IndexId());
                    dmax = d;
                }
            }
        }
    }
    return 0;
}

int
MCMeshSubset::ExtremePoints(int &index1, int &index2 )
{
    int i, j, nv;
    double d, dmax;
    Vertex vertex1, vertex2;

    nv = m_aopVs.num();
    if(nv == 0 ) return -1;
    
    for( i=0; i<nv-1;i++ ) {
        vertex1 = m_aopVs[i];
        for( j=i+1; j<nv; j++ ) {
            vertex2 = m_aopVs[j];
        
            d = Distance(vertex1, vertex2);
            if( i==0 && j==1 ) {
                index1 = i;
                index2 = j;
                dmax = d;
            } else {
                if( d > dmax ) {
                    index1 = i;
                    index2 = j;
                    dmax = d;
                }
            }
        }
    }
    return 0;
}

WINbool
MCMeshSubset::IsBoundary(Vertex &v) 
{
    if( m_BoundaryVertexIndices.num() == 0 ) SetBoundary();

    if( m_BoundaryVertexIndices.index(v->IndexId()) == -1 ) return FALSE;
    return TRUE;
}

Array<Face>&
MCMeshSubset::VertexFaces(const Vertex &v)
{
    Array< Vertex > aov;

    m_aof.need(0);
    for( int i=0; i<m_aopFs.num(); i++ ) {
        aov = m_pMesh.FaceVertices(m_aopFs[i]);
        for( int j=0; j<aov.num(); j++ ) {
            if( v->IndexId() == aov[j]->IndexId() ) {
                m_aof += m_aopFs[i];
                break;
            }
        }
    }
    return m_aof;
}

double
MCMeshSubset::VertexArea(const Vertex &v, MC_AreaType in_AreaType)
{
    int i, ier;
    double Area, theta, alpha, beta, t, a, b, s;
    Array<Face> aof;

    aof = VertexFaces(v);
    Area = 0.0;
    for( i=0; i<aof.num(); i++ ) {
        ier = AlphaBeta(aof[i],v,theta,alpha,beta,t,a,b);
        s = 0.5*(t+a+b);
        switch(in_AreaType) {
            case VORONOI:
                Area += 0.125*(a*a/tan(alpha) + b*b/tan(beta));
                break;
            case MIXED:
                if( t*t > a*a+b*b ) {
                    Area += sqrt(s*(s-t)*(s-a)*(s-b))/2.0;
                    break;
                }
                if( a*a > b*b+t*t || b*b > t*t+a*a ) {
                    Area += sqrt(s*(s-t)*(s-a)*(s-b))/4.0;
                } else {
                    Area += 0.125*(a*a/tan(alpha) + b*b/tan(beta));
                }
                break;
            case MODIFIED:
                if( a*a > b*b+t*t ) {
                    Area += 0.125 * b*b * tan(theta);
                    break;
                }
                if( b*b > t*t+a*a ) {
                    Area += 0.125 * a*a * tan(theta);
                    break;
                }
                if(  t*t > a*a+b*b ) {
                    Area += sqrt(s*(s-t)*(s-a)*(s-b))
                          - 0.125 * (a*a*tan(beta) + b*b*tan(alpha));
                } else {
                    Area += 0.125*(a*a/tan(alpha) + b*b/tan(beta));
                }
                break;
            case BARYCENTRIC:
            default:
                Area += sqrt(s*(s-t)*(s-a)*(s-b))/3.0;
                break;
        }
    }
    return Area;
}

int 
MCMeshSubset::AlphaBeta( const Face &in_face, const Vertex &in_vertex,
                         double &out_theta, 
                         double &out_alpha, double &out_beta,
                         double &t, double &a, double &b ) const
{
    int id;
    Array< Vertex > aov;
    R3Pt T, A, B;

    aov = m_pMesh.FaceVertices(in_face);
    id = -1;
    if( in_vertex->IndexId() == aov[0]->IndexId() ) id = 0;
    if( in_vertex->IndexId() == aov[1]->IndexId() ) id = 1;
    if( in_vertex->IndexId() == aov[2]->IndexId() ) id = 2;
    if( id == -1 ) return id;

    T = aov[id]->GetLoc();
    A = aov[(id+1)%3]->GetLoc();
    B = aov[(id+2)%3]->GetLoc();

    t = Distance(A,B);
    a = Distance(B,T);
    b = Distance(T,A);

    out_alpha = acos( ( b*b + t*t - a*a ) / ( 2.0 * b * t ) );
    out_beta  = acos( ( t*t + a*a - b*b ) / ( 2.0 * t * a ) );
    out_theta = acos( ( a*a + b*b - t*t ) / ( 2.0 * a * b ) );
    
    return 0;
}

int 
MCMeshSubset::AlphaBeta( const Vertex &in_vertex, const Edge &in_edge,
                         double &out_theta1, double &out_theta2, 
                         double &out_alpha, double &out_beta,
                         double &out_delta, double &out_gamma ) const
{
    int i;
    Face f1, f2;
    Vertex opposite;
    Array< Vertex > aov;
    Array<double> angles;
    R3Pt P, Q, R;

    opposite = m_pMesh.EdgeVertex1(in_edge);
    if( in_vertex == opposite ) opposite = m_pMesh.EdgeVertex2(in_edge);

	out_theta1 = 0.0;
	out_theta2 = 0.0;
	out_alpha  = 0.0;
	out_beta   = 0.0;
	out_delta  = 0.0;
	out_gamma  = 0.0;

    f1 = m_pMesh.EdgeFace1(in_edge);
    f2 = m_pMesh.EdgeFace2(in_edge);

	if( f1 != NULL ) {
		angles.need(0);
		FaceAngles(f1,angles);
		ASSERT( angles.num() == 3 );

		aov = m_pMesh.FaceVertices(f1);
		for( i=0; i<aov.num(); i++ ) {
		    if(aov[i] == in_vertex ) {
			    out_theta1 = angles[i];
		    } else {
		        if( aov[i] == opposite ) {
		            out_delta = angles[i];
		        } else {
		            out_alpha = angles[i];
		        }
		    }
		}
	}

	if( f2 != NULL ) {
		angles.need(0);
		FaceAngles(f2,angles);
		ASSERT( angles.num() == 3 );

		aov = m_pMesh.FaceVertices(f2);
		for( i=0; i<aov.num(); i++ ) {
		    if(aov[i] == in_vertex ) {
		        out_theta2 = angles[i];
		    } else {
		        if( aov[i] == opposite ) {
		            out_gamma = angles[i];
		       } else {
		          out_beta = angles[i];
		      }
		   }
		}
	}
    return 0;
}

double 
MCMeshSubset::CotAlphaBeta( const Vertex &v1, const Vertex &v2 ) const
{
    double theta1, theta2, alpha, beta, delta, gamma, 
           tan_alpha, tan_beta, result;

    const Edge e = m_pMesh.EdgeFromVertices(v1,v2);
    if( e == NULL )  {
        ASSERT(FALSE);
        return 0.0;
    }
    result = 0.0;

    const int ier = AlphaBeta(v1, e, theta1, theta2, alpha, beta, delta, gamma);

	if( m_pMesh.EdgeFace1(e) != NULL ) {
		// Compute the alpha component 
		tan_alpha = tan(alpha);
        if ( ! RNIsZero( tan_alpha ) ) {
		    result += 1.0 / tan_alpha;
        }
	}
    
	if( m_pMesh.EdgeFace2(e) != NULL ) {
		// Compute the beta component 
		tan_beta = tan(beta);
        if ( !RNIsZero( tan_beta ) ) {
            result += 1.0 / tan_beta;
        }
	}

    return result;
}

double 
MCMeshSubset::CotGammaDelta( const Vertex &v1, const Vertex &v2 ) const
{
    int ier;
    double theta1, theta2, alpha, beta, delta, gamma, 
           tan_delta, tan_gamma, result;
    Edge e;

    e = m_pMesh.EdgeFromVertices(v1,v2);
    if( e == NULL ) return 0.0;
    result = 0.0;

    FILE *out = NULL;
    //out = fopen(WINDataHome("edges.dat"),"a");
    if( out != NULL ) {
        fprintf(out,"Result: %d to %d\n", 
                    m_pMesh.EdgeVertex1(e)->IndexId(),
                    m_pMesh.EdgeVertex2(e)->IndexId() );
        fclose(out);
        out = NULL;
    }

    ier = AlphaBeta(v1, e, theta1, theta2, alpha, beta, delta, gamma);

	if( m_pMesh.EdgeFace2(e) != NULL ) {
		// Compute the gamma component 
		tan_gamma = tan(gamma);
		if( tan_gamma != 0.0 ) result += 1.0 / tan_gamma;
	}
    
	if( m_pMesh.EdgeFace2(e) != NULL ) {
		// Compute the delta component 
		tan_delta = tan(delta);
		if( tan_delta != 0.0 ) result += 1.0 / tan_delta;
	}

    return result;
}

int 
MCMeshSubset::FaceAngles(const Face &in_face, const Vertex &in_vertex,
                         int &out_ThetaId, double &out_Theta, double &out_t, 
                         int &out_AlphaId, double &out_Alpha, double &out_a, 
                         int &out_BetaId, double &out_Beta, double &out_b ) const
{
    int i, n, id;
    Array< Vertex > aov;
    R3Pt P, Q, R;

	if( in_face == NULL ) return -1;

    aov = m_pMesh.FaceVertices(in_face);

    n = aov.num();
    id = -1;
    for( i=0; i<n; i++ ) {
        if( aov[i] == in_vertex ) {
            id = i;
            break;
        }
    }
    if( id == -1 ) return -1;

    out_ThetaId = aov[i]->IndexId();
    P = aov[id]->GetLoc();
    Q = aov[(id+1)%n]->GetLoc();
    R = aov[(id+2)%n]->GetLoc();

    out_t = Distance(Q,R);
    out_a = Distance(R,P);
    out_b = Distance(P,Q);

    out_Theta = acos( ( out_a*out_a + out_b*out_b - out_t*out_t ) / 
                      ( 2.0 * out_a * out_b ) );
    out_Alpha = acos( ( out_t*out_t + out_b*out_b - out_a*out_a ) / 
                      ( 2.0 * out_t * out_b ) );
    out_Beta  = acos( ( out_a*out_a + out_t*out_t - out_b*out_b ) / 
                      ( 2.0 * out_a * out_t ) );

    out_ThetaId = aov[id]->IndexId();
    out_AlphaId = aov[(id+1)%n]->IndexId();
    out_BetaId  = aov[(id+2)%n]->IndexId();

    return 0;
}
    
void
MCMeshSubset::FaceAngles( const Face &in_face, Array<double> &out_adAngles ) const
{
    const Array< Vertex > aov = m_pMesh.FaceVertices(in_face);
    out_adAngles.need( aov.num() );

    // In case something goes wrong
    const double dAng = M_PI / (double) out_adAngles.num();
    for ( int i = 0; i < out_adAngles.num(); i++ )
        out_adAngles[i] = dAng;

    Array< double > adDistances( aov.num() );
    WINbool bIsZero = FALSE;
    for ( FORINT i = 0; i < aov.num(); i++ ) {
        adDistances[i] = Distance( aov[i], aov.wrap(i+1) );
        if ( RNIsZero( adDistances[i] ) )
            bIsZero = TRUE;
    }
    if ( bIsZero == TRUE ) {
        ASSERT( aov.num() == 3 ); // This is only the right thing to do if this is a triangle
        return; 
    }

    double dSumAllAngles = 0.0;
    for ( FORINT i = 0; i < aov.num(); i++ ) {
        const double dDenom = 2.0 * adDistances[i] * adDistances.wrap(i-1);  // Two adjacent sides
        const double dSumAngles = ( adDistances[i] * adDistances[i] + adDistances.wrap(i-1) * adDistances.wrap(i-1) - adDistances.wrap(i+1)*adDistances.wrap(i+1) ) / dDenom;
        out_adAngles[i] = acos( WINminmax( dSumAngles, -1.0, 1.0 ) );

        dSumAllAngles += out_adAngles[i];
    }

    if ( aov.num() == 3 )
        ASSERT( RNApproxEqual( dSumAllAngles, M_PI, 1e-6 ) );
}

int
MCMeshSubset::CotAngles( const Face &in_face, const Vertex &in_vertex,
                         int &out_ThetaId, double &out_CotTheta, double &out_t, 
                         int &out_AlphaId, double &out_CotAlpha, double &out_a, 
                         int &out_BetaId, double &out_CotBeta, double &out_b ) const
{
    int ier;
    double theta, alpha, beta, TanTheta, TanAlpha, TanBeta;

    ier = FaceAngles(in_face, in_vertex, 
                     out_ThetaId, theta, out_t, 
                     out_AlphaId, alpha, out_a,
                     out_BetaId, beta, out_b);
    if( ier != 0 ) return ier;

    out_CotTheta = 0.0;
    TanTheta = tan(theta);
    if( TanTheta != 0.0 ) out_CotTheta = 1.0/TanTheta;

    out_CotAlpha = 0.0;
    TanAlpha = tan(alpha);
    if( TanAlpha != 0.0 ) out_CotAlpha = 1.0/TanAlpha;

    out_CotBeta = 0.0;
    TanBeta = tan(beta);
    if( TanBeta != 0.0 ) out_CotBeta = 1.0/TanBeta;

    return 0;
}

int
MCMeshSubset::Centroid(const Face &in_face, const Vertex &in_vertex,
                       R3Pt &out_Centroid, double &out_CentroidDist) const
{
    int i, n;
    R3Pt P, Q;
	Array< Vertex > aov;

	if( in_face == NULL ) return -1;
	if( in_vertex == NULL ) return -1;

	out_Centroid = R3Pt(0.0, 0.0, 0.0);

    aov = m_pMesh.FaceVertices(in_face);
    n = aov.num();

    for( i=0; i<n; i++ ) {
		P = aov[i]->GetLoc();
		out_Centroid[0] += P[0];
        out_Centroid[1] += P[1];
        out_Centroid[2] += P[2];
    }
	out_Centroid[0] /= 3.0;
	out_Centroid[1] /= 3.0;
	out_Centroid[2] /= 3.0;
	Q = in_vertex->GetLoc();

    out_CentroidDist = Distance(out_Centroid,Q);

    return 0;
}

int
MCMeshSubset::CentroidSum(const Vertex &in_vertex, double &out_Sum) const
{
    int ier, i, n;
	double d;
    R3Pt P;
	Array<Face> aof;

	if( in_vertex == NULL ) return -1;

    aof = m_pMesh.VertexFaces( in_vertex );
    n = aof.num();

	out_Sum = 0.0;
    for( i=0; i<n; i++ ) {
		ier = Centroid(aof[i], in_vertex, P, d);
		if( ier != 0 ) return ier;

		out_Sum += d;
    }

	return 0;
}

int 
MCMeshSubset::MaxWeight(const Face &in_face, const Vertex &in_vertex,
                        double &out_Weight) const
{
    int i, n, id;
	double area, a, b, denom;
    Array< Vertex > aov;
    R3Pt P, Q, R;

	if( in_face == NULL ) return -1;

    aov = m_pMesh.FaceVertices(in_face);

    n = aov.num();
    id = -1;
    for( i=0; i<n; i++ ) {
        if( aov[i] == in_vertex ) {
            id = i;
            break;
        }
    }
    if( id == -1 ) return -1;

    P = aov[id]->GetLoc();
    Q = aov[(id+1)%n]->GetLoc();
    R = aov[(id+2)%n]->GetLoc();

    a = Distance(R,P);
    b = Distance(P,Q);

	area = FaceArea(in_face);

	denom = a*a + b*b;
	if( denom == 0.0 ) {
		out_Weight = 0.0;
	} else {
		out_Weight = area / denom;
	}

    return 0;
}
    

int 
MCMeshSubset::CalculateNormal(const Vertex &in_vertex,
                              const MC_NormalType &in_type,
                              R3Vec &out_normal ) const
{
    int i, ier, ThetaId, AlphaId, BetaId;
    double Theta, Alpha, Beta, CotTheta, CotAlpha, CotBeta, t, a, b, area, wt;
    Array<Face> aof;
    R3Vec vector;
	R3Pt P;

	int verbose = 0;

    out_normal[0] = 0.0;
    out_normal[1] = 0.0;
    out_normal[2] = 0.0;

    aof = m_pMesh.VertexFaces( in_vertex );
    for( i=0; i<aof.num(); i++ ) {
        FaceNormal(aof[i],vector);
        switch( in_type ) {
            case AREA:
                area = FaceArea(aof[i]);
                vector *= area;
                out_normal += vector;
                break;
            case ANGLE:
                ier = FaceAngles(aof[i], in_vertex,
                                 ThetaId, Theta, t, 
                                 AlphaId, Alpha, a, 
                                 BetaId, Beta, b );
                if( ier != 0 ) return ier;
                vector *= Theta;
                out_normal += vector;
                break;

            case COTANGENT:
                ier = CotAngles( aof[i], in_vertex,
                                 ThetaId, CotTheta, t, 
                                 AlphaId, CotAlpha, a, 
                                 BetaId, CotBeta, b );
                if( ier != 0 ) return ier;
                vector *= (CotAlpha + CotBeta);
                out_normal += vector;
                break;

            case CENTROID:
                ier = Centroid(aof[i], in_vertex, P, wt );
                if( ier != 0 ) return ier;
				if( verbose == 1 ) {
					printf("Vector=(%lf,%lf,%lf), wt=%lf\n", 
						   vector[0], vector[1], vector[2], wt);
				}
                vector *= wt;
                out_normal += vector;
                break;

            case MAX99:
                ier = MaxWeight(aof[i], in_vertex, wt);
                if( ier != 0 ) return ier;
                vector *= wt;
                out_normal += vector;
                break;
            case AVERAGE:
            default:
                out_normal += vector;
                break;
        }
    }
    out_normal.Normalize();

	if( verbose == 1 ) {
		if( in_type == CENTROID ) {
			printf("Vector=(%lf,%lf,%lf)\n", 
				   out_normal[0], out_normal[1], out_normal[2]);
		}
	}

    return 0;
}

void MCMeshSubset::FaceNormal(const Face &in_face, R3Vec &out_normal) const
{
    R3Vec va, vb;
    Array<Vertex> aov;

    aov = m_pMesh.FaceVertices(in_face);
    ASSERT( aov.num() == 3);

    va = aov[2]->GetLoc() - aov[1]->GetLoc();
    vb = aov[1]->GetLoc() - aov[0]->GetLoc();
    out_normal[0] = vb[1] * va[2] - vb[2] * va[1];
    out_normal[1] = vb[2] * va[0] - vb[0] * va[2];
    out_normal[2] = vb[0] * va[1] - vb[1] * va[0];
    out_normal.Normalize();
}

double 
MCMeshSubset::FaceArea(const Face &in_face) const
{
    double a, b, c, s, area;
    Array<Vertex> aov;

    area = 0;
    aov = m_pMesh.FaceVertices(in_face);
    ASSERT( aov.num() == 3);

    a = Distance( aov[0], aov[1] );
    b = Distance( aov[1], aov[2] );
    c = Distance( aov[2], aov[0] );
    s = 0.5 * ( a + b + c );
    area = sqrt( s * (s-a) * (s-b) * (s-c) );

    return area;
}
    
void 
MCMeshSubset::dump(FILE *out)
{
    int i, j;
    R3Pt P;
    Array< Vertex > aov;
    
    for( i=0; i<m_aopVs.num(); i++ ) {
        P = m_aopVs[i]->GetLoc();
        fprintf(out,"Vertex %d: IndexId %d (%lf, %lf, %lf)\n", 
                    i, m_aopVs[i]->IndexId(), P[0], P[1], P[2]);
    }
    for( i=0; i<m_aopFs.num(); i++ ) {
        aov = m_pMesh.FaceVertices(m_aopFs[i]);
        fprintf(out,"Face %d:", i );
        for( j=0; j<aov.num(); j++ ) {
            fprintf(out," %d (%d)", aov[j]->Id(), aov[j]->IndexId() );
        }
        fprintf(out,"\n");
    }
    SetBoundary();
    for( i=0; i<m_BoundaryVertexIndices.num(); i++ ) {
        if( i%5 == 0 ) fprintf(out,"\nBoundary(%d):", i );
        fprintf(out,"%d ", m_BoundaryVertexIndices[i] );
    }
    fprintf(out,"\n");
}
