// MCMeshUtils.cpp : Defines mesh utility routines.
//

#include "stdafx.h"
#include <MC/MC_MeshUtils.h>
#include <MC/MC_Data.H>
#include <utils/Pm_MeshLite.H>
#include <fitting/FITTools_Least_squares.H>
#include <fitting/FITTools_SVD.H>

void 
MCMeshRange(const PMeshLite &in_mesh, R3Pt &Xmin, R3Pt &Xmax )
{
    int i;
    R3Pt P;
    Array<Vertex> aopVs;

    in_mesh.AllMeshVertices(aopVs);
    for( i=0; i<aopVs.num(); i++ ) {
        P = aopVs[i]->GetLoc();
        if( i==0 ) {
            Xmin = P;
            Xmax = P;
        } else {
            if( P[0] < Xmin[0] ) Xmin[0] = P[0];
            if( P[1] < Xmin[1] ) Xmin[1] = P[1];
            if( P[2] < Xmin[2] ) Xmin[2] = P[2];
            if( P[0] > Xmax[0] ) Xmax[0] = P[0];
            if( P[1] > Xmax[1] ) Xmax[1] = P[1];
            if( P[2] > Xmax[2] ) Xmax[2] = P[2];
        }
    }
}

void 
MCFixMeshNormals(PMeshLite &in_mesh)
{
    int Isign, Bsign, iConsist;
    //int ier, jConsist;
    Array<int> dummy;

	int verbose = 1;

    MCFindNormalSign(in_mesh, Isign, Bsign);
	if( verbose ) {
		printf("Isign = %d, Bsign = %d\n", Isign, Bsign);
	}

    if( Isign > 1 || (Isign + Bsign) > 5) {
	}

    if( Isign < -1 || (Isign + Bsign) < -5) {

		if( verbose ) {
			printf("Flipping normals\n");
		}
		MCFlipMeshNormals(in_mesh);
	}

	iConsist = MCCheckNormalConsistency(in_mesh);
	if( verbose ) {
		printf("iConsist = %d\n", iConsist);
	}

    if( !iConsist ) {
			
		if( verbose ) {
			printf("Flipping mesh\n");
		}
		in_mesh.Flip();
	}

	//MCFindNormalSign(in_mesh, Isign, Bsign);
	//iConsist = MCCheckNormalConsistency(in_mesh);

    //ier = MCWriteMeshWithScalar(WINDataHome("Mesh_fix.txt"),
    //                            in_mesh, dummy, dummy);
}

int
MCCheckNormalConsistency(PMeshLite &in_mesh)
{
    int i, j, consistent;
    double NdotN, vmag;
    R3Pt P, p0, p1, p2;
    R3Vec N, NN, v, v1, v2;
    Array<Vertex> aopVs;
    Array<Face> aopFs;
    Array<Vertex> aov;
    Array<Face> aof;

    in_mesh.AllMeshVertices(aopVs);
    in_mesh.AllMeshFaces(aopFs);

    consistent = 0;
    for( i=0; i<aopVs.num(); i++ ) {
        N = aopVs[i]->GetNorm();
        P = aopVs[i]->GetLoc();

        NN[0] = 0.0;
        NN[1] = 0.0;
        NN[2] = 0.0;
        aof = in_mesh.VertexFaces(aopVs[i]);
        for( j=0; j<aof.num(); j++ ) {
            aov = in_mesh.FaceVertices(aof[j]);
            p0 = aov[0]->GetLoc();
            p1 = aov[1]->GetLoc();
            p2 = aov[2]->GetLoc();
            v1 = p1 - p0;
            v2 = p2 - p1;
            v[0] = v2[1]*v1[2] - v2[2]*v1[1];
            v[1] = v2[2]*v1[0] - v2[0]*v1[2];
            v[2] = v2[0]*v1[1] - v2[1]*v1[0];
            vmag = sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
            v *= -1.0;
            if( vmag != 0.0 ) {
                v /= vmag;
            }
            NN[0] += v[0];
            NN[1] += v[1];
            NN[2] += v[2];
        }

        vmag = sqrt(NN[0]*NN[0] + NN[1]*NN[1] + NN[2]*NN[2]);
        if( vmag != 0.0 ) {
            NN /= vmag;
        }

        NdotN = N[0]*NN[0] + N[1]*NN[1] + N[2]*NN[2];

        if( NdotN < 0.0 ) {
            consistent--;
        } else {
            consistent++;
        }
    }
    if( consistent > 0 ) return 1;
    return 0;
}

void
MCFlipMeshNormals(PMeshLite &in_mesh)
{
    int i;
    R3Vec N;
    Array<Vertex> aopVs;

    in_mesh.AllMeshVertices(aopVs);
    for( i=0; i<aopVs.num(); i++ ) {
        N[0] = -1.0 * aopVs[i]->GetNorm()[0];
        N[1] = -1.0 * aopVs[i]->GetNorm()[1];
        N[2] = -1.0 * aopVs[i]->GetNorm()[2];
        aopVs[i]->SetNorm(N);
    }
}

void 
MCFindNormalSign(const PMeshLite &in_mesh, 
				 int &IPos, int &INeg, 
				 int &BPos, int &BNeg)
{
    // IPos - count of interior extreme points with outward pointing normal
    // INeg - count of interior extreme points with inward pointing normal
    // BPos - count of boundary extreme points with outward pointing normal
    // BNeg - count of boundary extreme points with inward pointing normal

    int i, ix0, ix1, iy0, iy1, iz0, iz1;
    R3Pt P, Xmin, Xmax;
    R3Vec N;
    Array<Vertex> aopVs;
    Vertex v;

    IPos = 0;
    INeg = 0;
    BPos = 0;
    BNeg = 0;

    in_mesh.AllMeshVertices(aopVs);
    for( i=0; i<aopVs.num(); i++ ) {
        P = aopVs[i]->GetLoc();
        if( i==0 ) {
            Xmin = P;
            Xmax = P;
            ix0 = iy0 = iz0 = 1;
            ix1 = iy1 = iz1 = 1;
        } else {
            if( P[0] < Xmin[0] ) {
                Xmin[0] = P[0];
                ix0 = i;
            }
            if( P[1] < Xmin[1] ) {
                Xmin[1] = P[1];
                iy0 = i;
            }
            if( P[2] < Xmin[2] ) {
                Xmin[2] = P[2];
                iz0 = i;
            }
            if( P[0] > Xmax[0] ) {
                Xmax[0] = P[0];
                ix1 = i;
            }
            if( P[1] > Xmax[1] ) {
                Xmax[1] = P[1];
                iy1 = i;
            }
            if( P[2] > Xmax[2] ) {
                Xmax[2] = P[2];
                iz1 = i;
            }
        }
    }

    // Check the Xmin point
    v = aopVs[ix0];
    N = v->GetNorm();
    if( in_mesh.IsBoundary(v) ) {
        if( N[0] < 0.0 ) {
            BPos++;
        } else {
            BNeg++;
        }
    } else {
        if( N[0] < 0.0 ) {
            IPos++;
        } else {
            INeg++;
        }
    }

    // Check the Xmax point
    v = aopVs[ix1];
    N = v->GetNorm();
    if( in_mesh.IsBoundary(v) ) {
        if( N[0] > 0.0 ) {
            BPos++;
        } else {
            BNeg++;
        }
    } else {
        if( N[0] > 0.0 ) {
            IPos++;
        } else {
            INeg++;
        }
    }

    // Check the Ymin point
    v = aopVs[iy0];
    N = v->GetNorm();
    if( in_mesh.IsBoundary(v) ) {
        if( N[1] < 0.0 ) {
            BPos++;
        } else {
            BNeg++;
        }
    } else {
        if( N[1] < 0.0 ) {
            IPos++;
        } else {
            INeg++;
        }
    }

    // Check the Ymax point
    v = aopVs[iy1];
    N = v->GetNorm();
    if( in_mesh.IsBoundary(v) ) {
        if( N[1] > 0.0 ) {
            BPos++;
        } else {
            BNeg++;
        }
    } else {
        if( N[1] > 0.0 ) {
            IPos++;
        } else {
            INeg++;
        }
    }

    // Check the Zmin point
    v = aopVs[iz0];
    N = v->GetNorm();
    if( in_mesh.IsBoundary(v) ) {
        if( N[2] < 0.0 ) {
            BPos++;
        } else {
            BNeg++;
        }
    } else {
        if( N[2] < 0.0 ) {
            IPos++;
        } else {
            INeg++;
        }
    }

    // Check the Zmax point
    v = aopVs[iz1];
    N = v->GetNorm();
    if( in_mesh.IsBoundary(v) ) {
        if( N[2] > 0.0 ) {
            BPos++;
        } else {
            BNeg++;
        }
    } else {
        if( N[2] > 0.0 ) {
            IPos++;
        } else {
            INeg++;
        }
    }
}

void 
MCFindNormalSign(const PMeshLite &in_mesh, int &Isign, int &Bsign )
{
    // Isign > 0 indicates outward pointing normal
    // Isign < 0 indicates inward pointing normal
    // Isign uses extreme points not on a boundary
    // Bsign uses extreme points on a boundary

	int IPos, INeg, BPos, BNeg;

	MCFindNormalSign(in_mesh, IPos, INeg, BPos, BNeg);

	Isign = IPos - INeg;
	Bsign = BPos - BNeg;
}

int MCAnalyzeMesh(const PMeshLite &in_mesh, 
				  int &out_iFaceOrient, 
				  int &out_iNormOrient)
{
    int i, j, k, ndx[6], count;
	double val[6];
    R3Pt P;
	R3Vec N;
	Array<Vertex> aov;

	// Find indices for extreme points
	for( j=0; j<6; j++ ) {
		k = j%3;
		ndx[j] = 0;
		val[j] = in_mesh.VertexFromIndexId(0)->GetLoc()[k];
	}
	ForMeshVertex( in_mesh.Mesh(), v ) {
		i = v->IndexId();
		P = v->GetLoc();
		for( k=0; k<3; k++ ) {
			j = k+3;
			if( P[k] > val[k] ) {
				ndx[k] = i;
				val[k] = P[k];
			}
			if( P[k] < val[j] ) {
				ndx[j] = i;
				val[j] = P[k];
			}
		}
    }EndFor;

	// Flag extreme points on boundary
	for( j=0; j<6; j++ ) {
		if( in_mesh.IsBoundary(in_mesh.VertexFromIndexId(ndx[j])) ) ndx[j] = -1;
	}

	// Check the face vertex ordering
	out_iFaceOrient = 0;
	for( j=0; j<6; j++ ) {
		if( ndx[j] < 0 ) continue;
		k = j%3;
		count = 0;
		Vertex v = in_mesh.VertexFromIndexId(ndx[j]);
        ForVertexFace( in_mesh.Mesh(), v, f ) {
			aov = in_mesh.FaceVertices(f);
			R3Vec N = UnitSafe( Cross( aov[1]->GetLoc() - aov[0]->GetLoc(), aov[2]->GetLoc() - aov[0]->GetLoc()) );
			if( k==j ) {
				if( N[k] > 0.0 ) count++;
				if( N[k] < 0.0 ) count--;
			} else {
				if( N[k] > 0.0 ) count--;
				if( N[k] < 0.0 ) count++;
			}
        } EndFor;
		if( count > 0 ) out_iFaceOrient++;
		if( count < 0 ) out_iFaceOrient--;
	}

	// Check for normal orientation
	out_iNormOrient = 0;
	for( j=0; j<6; j++ ) {
		if( ndx[j] < 0 ) continue;
		k = j%3;
		Vertex v = in_mesh.VertexFromIndexId(ndx[j]);
        R3Vec N = v->GetNorm();
		R3Pt Q = v->GetLoc();
		if( k == j ) {
			if( N[k] > 0.0 ) out_iNormOrient++;
			if( N[k] < 0.0 ) out_iNormOrient--;
		} else {
			if( N[k] > 0.0 ) out_iNormOrient--;
			if( N[k] < 0.0 ) out_iNormOrient++;
		}
	}

	return 0;
}

int 
MCFindPointId(const PMeshLite &in_mesh, const int dir, const double tol,
              int &out_indexId )
{
    int i, i1, i2, axis, axis_sign;
    double x1, x2, d, dmin, dh, d1, d2;
    R3Pt P, Q, Xmin, Xmax;
    Array<Vertex> aopVs;
    WINbool found = false;

    axis_sign = (dir>0) ? 1 : -1;
    axis = dir * axis_sign - 1;
    if( axis < 0 || axis > 2 ) return -1;
    i1 = (axis+1)%3;
    i2 = (axis+2)%3;

    MCMeshRange(in_mesh, Xmin, Xmax);
    x1 = 0.5 * (Xmax[i1] + Xmin[i1]);
    x2 = 0.5 * (Xmax[i2] + Xmin[i2]);

    in_mesh.AllMeshVertices(aopVs);
    for( i=0; i<aopVs.num(); i++ ) {
        Q = aopVs[i]->GetLoc();
        d1 = Q[i1]-x1;
        d2 = Q[i2]-x2;
        d = sqrt( d1*d1 + d2*d2 );
        dh = d - Q[axis]*axis_sign;
        if( !found ) {
            if( d < tol ) {
                found = true;
                P = Q;
                out_indexId = i;
            } else {
                if( i == 0 ) dmin = dh+1;
                if( dh < dmin ) {
                    P = Q;
                    out_indexId = i;
                    dmin = dh;
                }
            }
        } else {
            if( d < tol && Q[axis]*axis_sign > P[axis]*axis_sign ) {
                P = Q;
                out_indexId = i;
            }
        }
    }
    return 0;
}

int MCFlagBoundedRegion(const PMeshLite &in_mesh, const Vertex &in_vertex, 
                        const int dir, const double depth, 
                        Array<Face> &out_aopFs )
{
    int i, j, k, jndx, kndx, nleft, nv, axis, axis_sign;
    double d;
    R3Pt P, Q;
    Array<Vertex> aopVs;
    Array<Vertex> aov;
    Array<Face> aof;
    Array<int> vertexFlag;
    WINbool accept;

    axis_sign = (dir > 0) ? 1 : -1;
    axis = dir*axis_sign - 1;
    if( axis < 0 || axis > 2 ) return -1;

    in_mesh.AllMeshVertices(aopVs);
    nv = aopVs.num();
    P = in_vertex->GetLoc();

    out_aopFs.need(0);
    vertexFlag.need(0);
    for( i=0; i<nv; i++ ) {
        vertexFlag += -1;
    }

    vertexFlag[aopVs.index(in_vertex)] = 0;
    nleft = 1;

    while(nleft > 0) {
        nleft = 0;
        for( i=0; i<nv; i++ ) {
            if(vertexFlag[i] == 0) {
                aof = in_mesh.VertexFaces(aopVs[i]);
                for( j=0; j<aof.num(); j++ ) {
                    accept = true;
                    aov = in_mesh.FaceVertices(aof[j]);
                    for( k=0; k<aov.num(); k++ ) {
                        kndx = aov[k]->IndexId();
                        if( vertexFlag[kndx] == -1 ) {
                            Q = aov[k]->GetLoc();
                            d = (P[axis]-Q[axis])*axis_sign;
                            if( d > depth ) {
                                vertexFlag[kndx] = -2;
                                accept = false;
                            } else {
                                vertexFlag[kndx] = 0;
                                nleft++;
                            }
                        } else {
                            if( vertexFlag[kndx] == -2 ) accept = false;
                        }
                    }
                    if( accept ) {
                        jndx = out_aopFs.index(aof[j]);
                        if( jndx == -1 ) out_aopFs += aof[j];
                    }
                }
                vertexFlag[i] = 1;
            }
        }
    }
    return 0;
}

int MCTrimMesh(const PMeshLite &in_mesh, const int trimType,
               const double trimFraction, Array<Face> &out_aopFs )
{
    int i, j, axis, axis_sign, count;
    double value;
    R3Pt Xmin, Xmax, P;
    Array<Vertex> aov;
    Face f;
    Array<int> vertexFlag;

    switch(trimType) {
        case 1:
            axis = 0;
            axis_sign = -1;
            break;
        case 2:
            axis = 0;
            axis_sign = 1;
            break;
        case 3:
            axis = 1;
            axis_sign = -1;
            break;
        case 4:
            axis = 1;
            axis_sign = 1;
            break;
        case 5:
            axis = 2;
            axis_sign = -1;
            break;
        case 6:
            axis = 2;
            axis_sign = 1;
            break;
        default:
            return -1;
    }

    MCMeshRange(in_mesh, Xmin, Xmax);
    value = (1.0-trimFraction)*Xmin[axis] + trimFraction*Xmax[axis];
    value *= axis_sign;

    printf("Xmin = (%lf, %lf, %lf)\n", Xmin[0], Xmin[1], Xmin[2]);
    printf("Xmax = (%lf, %lf, %lf)\n", Xmax[0], Xmax[1], Xmax[2]);
    printf("Trimming at X[%d] = %lf\n", axis, value*axis_sign);

    out_aopFs.need(0);
    vertexFlag.need(0);
    for( i=0; i<in_mesh.NumVertices(); i++ ) {
        P = in_mesh.VertexFromIndexId(i)->GetLoc();
        if( P[axis]*axis_sign < value ) {
            vertexFlag += 1;
        } else {
            vertexFlag += 0;
        }
    }

    for( i=0; i<in_mesh.NumFaces(); i++ ) {
        f = in_mesh.FaceFromIndexId(i);
        aov = in_mesh.FaceVertices(f);
        count = 0;
        for( j=0; j<aov.num(); j++ ) {
            count += vertexFlag[aov[j]->IndexId()];
        }
        if( count == aov.num() ) {
            out_aopFs += f;
        }
    }
    return 0;
}

void 
InterpolateToGrid( Array <R2Pt> &grid, Array <double> &data,
                   int ni, int nj, 
                   double umin, double umax,
                   double vmin, double vmax,
                   PMeshLite &mesh,
                   Array <Face> &faces,
                   Array <double> &faceData,
                   double max_error )
{
    int i, j, n, nf, ipt, i0, i1, j0, j1;
    double u, v, w, err, err_old, s, t;
    Array <double> error;
    Array <Vertex> aov;
    R3Pt P, Pmin, Pmax;
    R2Pt P1, P2, P3;

    error.needSet(ni*nj, max_error);
    //for( i=0; i<ni*nj; i++ ) error.set(i,max_error);

    nf = faces.num();
    for( n=0; n<nf; n++ ) {

        // Get the bounding box for this face
        aov = mesh.FaceVertices(faces[n]);
        BoundingBox(aov, Pmin, Pmax);
        
        // Find the range of grid node indices that can be in this face
        FindCell(Pmin[0], Pmin[1], ni, nj, umin, umax, vmin, vmax,
                 i0, j0, s, t);
        FindCell(Pmax[0], Pmax[1], ni, nj, umin, umax, vmin, vmax,
                 i1, j1, s, t);
        i1++;
        j1++;

        // Interpolate the points
        for( j=j0; j<=j1; j++ ) {
            for( i=i0; i<=i1; i++ ) {
                ipt = i + j*ni;
                err_old = error[ipt];
                P1 = R2Pt(aov[0]->GetLoc()[0], aov[0]->GetLoc()[1]);
                P2 = R2Pt(aov[1]->GetLoc()[0], aov[1]->GetLoc()[1]);
                P3 = R2Pt(aov[2]->GetLoc()[0], aov[2]->GetLoc()[1]);
                Barycentric(P1, P2, P3, grid[ipt], u, v, w, err );
                if( err < err_old ) {
                    error[ipt] = err;
                    data[ipt] = u * faceData[aov[0]->IndexId()] +
                                v * faceData[aov[1]->IndexId()] +
                                w * faceData[aov[2]->IndexId()];
                }
            }
        }
    }
}

double 
LaplacianSmoothing( Array <R2Pt> &uv, const Array <R2Pt> &s,
                    int nu, int nv, int max_it, double tol )
{
    int i, j, n, ipt;
    double residual, u, v, d, dx, dy, denom, L, L1, L2, L3, L4;
    Array <R2Vec> newUV;
    R2Pt P, Q;

    newUV.need(nu*nv);

    for( n=0; n<max_it; n++ ) {
        residual = 0.0;
        for( j=1; j<nv-1; j++ ) {
            for( i=1; i<nu-1; i++ ) {
                ipt = i + j*nu;
                P = uv[ipt];
                u = 0.0;
                v = 0.0;
                denom = 0.0;

                Q = uv[ipt-1];
                L = s[ipt-1][0];
                u += Q[0]/L;
                v += Q[1]/L;
                denom += 1.0/L;
                L1 = L;

                Q = uv[ipt+1];
                L = s[ipt][0];
                u += Q[0]/L;
                v += Q[1]/L;
                denom += 1.0/L;
                L2 = L;

                Q = uv[ipt-nu];
                L = s[ipt-nu][1];
                u += Q[0]/L;
                v += Q[1]/L;
                denom += 1.0/L;
                L3 = L;

                Q = uv[ipt+nu];
                L = s[ipt][1];
                u += Q[0]/L;
                v += Q[1]/L;
                denom += 1.0/L;
                L4 = L;

                L1 /= denom;
                L2 /= denom;
                L3 /= denom;
                L4 /= denom;

                newUV[ipt][0] = u/denom;
                newUV[ipt][1] = v/denom;
                dx = newUV[ipt][0] - P[0];
                dy = newUV[ipt][1] - P[1];
                d  = dx*dx + dy*dy;
                if( d > residual ) residual = d;
            }
        }
        for( j=1; j<nv-1; j++ ) {
            for( i=1; i<nu-1; i++ ) {
                ipt = i + j*nu;
                uv[ipt][0] = newUV[ipt][0];
                uv[ipt][1] = newUV[ipt][1];
            }
        }
        if( residual < tol ) break;
        if( (n%10) == 0 ) printf("Iteration %d: Residual = %lf\n", n, residual);
    }
    printf("Laplacian smoothing: %d iterations, Residual = %lf\n", n, residual);

    return residual;
}

void 
BoundingBox( const Array <Vertex> &aov, 
             R3Pt &BBmin, R3Pt &BBmax )
{
    int i;
    R3Pt P;

    BBmin = aov[0]->GetLoc();
    BBmax = aov[0]->GetLoc();

    for( i=1; i<aov.num(); i++ ) {
        P = aov[i]->GetLoc();
        if( P[0] < BBmin[0] ) BBmin[0] = P[0];
        if( P[0] > BBmax[0] ) BBmax[0] = P[0];
        if( P[1] < BBmin[1] ) BBmin[1] = P[1];
        if( P[1] > BBmax[1] ) BBmax[1] = P[1];
        if( P[2] < BBmin[2] ) BBmin[2] = P[2];
        if( P[2] > BBmax[2] ) BBmax[2] = P[2];
    }
}

double 
Area( const R2Pt &p1, const R2Pt &p2, const R2Pt &p3 )
{
    return Area(p1[0], p1[1], p2[0], p2[1], p3[0], p3[1]);
}

double 
Area( double x1, double y1,
      double x2, double y2, 
      double x3, double y3 )
{
    double a;

    a = 0.5*(x1*y2 + x2*y3 + x3*y1 - (x1*y3 +x2*y1 + x3*y2));
    return a;
}

void
Barycentric( const R2Pt &p1, const R2Pt &p2, 
             const R2Pt &p3, const R2Pt &p,
             double &u, double &v, double &w,
             double &error )
{
    double a, a1, a2, a3, aa, aa1, aa2, aa3;

    a  = Area(p1, p2, p3);
    a1 = Area(p2, p3, p);
    a2 = Area(p3, p1, p);
    a3 = Area(p1, p2, p);
    u  = a1/a;
    v  = a2/a;
    w  = a3/a;

    // See if the point may be outside the triangle
    error = 0.0;
    aa  = (a>=0)  ? a  : -a;
    aa1 = (a1>=0) ? a1 : -a1;
    aa2 = (a2>=0) ? a2 : -a2;
    aa3 = (a3>=0) ? a3 : -a3;
    error = (aa1 + aa2 + aa3) / aa - 1.0;
    error = (error<0.0) ? -error : error;
}

void 
FindCell( Array <R2Pt> &uv, int ni, int nj,
          double x,  double y, 
          int &iCell, int &jCell, double &s, double &t, double &error )
{
    // Find the cell and position within a cell for a point (x,y) in a 
    // deformed Cartesian grid

    int i, j, ipt1, ipt2, ipt3, ipt4;
    double u, v, w, err;
    R2Pt P;

    P = R2Pt(x,y);
    Barycentric(uv[0], uv[1], uv[ni], P, u, s, t, error);
    iCell = 0;
    jCell = 0;
        
    for( j=0; j<nj-1; j++ ) {
       for( i=0; i<ni-1; i++ ) {
           ipt1 = i + j*ni;
           ipt2 = ipt1 + 1;
           ipt3 = ipt1 + ni;
           ipt4 = ipt3 + 1;

           Barycentric(uv[ipt1], uv[ipt2], uv[ipt3], P, u, v, w, err);
           if( err < error ) {
               iCell = i;
               jCell = j;
               s = v;
               t = w;
               error = err;
           }

           Barycentric(uv[ipt2], uv[ipt4], uv[ipt3], P, u, v, w, err);
           if( err < error ) {
               iCell = i;
               jCell = j;
               s = u+v;
               t = v+w;
               error = err;
           }
       }
    }
}

void 
FindCell( double x,  double y,  int ni, int nj,
          double umin, double umax, double vmin, double vmax,
          int &iCell, int &jCell, double &s, double &t )
{
    // Find the cell and position within a cell for a point (x,y) in a 
    // regular Cartesian grid

    double u, v;

    u = (ni-1) * (x - umin) / (umax - umin);
    v = (nj-1) * (y - vmin) / (vmax - vmin);

    iCell = (int) u;
    if( iCell < 0   ) iCell = 0;
    if( iCell >= ni ) iCell = ni-1;
    s = u - iCell;

    jCell = (int) v;
    if( jCell < 0   ) jCell = 0;
    if( jCell >= nj ) jCell = nj-1;
    t = v - jCell;
}

int  
GridInterp( int iCell, int jCell, double s,  double t, 
            double umin, double umax, double vmin, double vmax,
            int ni, int nj, double &x, double &y )
{
    // Interpolate to get a physical point in a Cartesian grid 
    // from cell parameters
    double u, v;

    if( iCell < 0 || iCell >= ni ) return -1;
    if( jCell < 0 || jCell >= nj ) return -2;

    u = iCell + s;
    v = jCell + t;

    x = (umin*(ni-u) + umax*u) / (double) ni;
    y = (vmin*(nj-v) + vmax*v) / (double) nj;
    
    return 0;
}

int  
GridInterp( const Array <R2Pt> &grid, int ni, int nj,
            int iCell, int jCell, double s,  double t, 
            double &x, double &y )
{
    // Interpolate to get a physical point in a deformed grid 
    // from cell parameters
    int ipt;
    R2Pt P0, P1, P2, P3;

    if( iCell < 0 || iCell >= ni ) return -1;
    if( jCell < 0 || jCell >= nj ) return -2;

    ipt = iCell + jCell*ni;

    P0 = grid[ipt];
    P1 = grid[ipt+1];
    P2 = grid[ipt+ni];
    P3 = grid[ipt+ni+1];

    x = (1.0-s)*(1.0-t)*P0[0] + s*(1.0-t)*P1[0] + (1.0-s)*t*P2[0] + s*t*P3[0];
    y = (1.0-s)*(1.0-t)*P0[1] + s*(1.0-t)*P1[1] + (1.0-s)*t*P2[1] + s*t*P3[1];

    return 0;
}

void NormalizeUVCoords(PMeshLite &mesh, int nDir)
{
    Array <Vertex> aoVs;

    mesh.AllMeshVertices(aoVs);
    NormalizeUVCoords(aoVs,nDir);
}

void NormalizeUVCoords(Array <Vertex> &aoVs, int nDir)
{
    int i, nv;
    double uscale, vscale, scale, u, v;
    R3Pt P, Pmin, Pmax;
    R3Vec Range;

    // Get the current bounding box size
    BoundingBox(aoVs, Pmin, Pmax);

    // Compute the scale factors in each direction
    Range = Pmax - Pmin;
    uscale = (Range[0]>0.0) ? 1.0/Range[0] : 1.0;
    vscale = (Range[1]>0.0) ? 1.0/Range[1] : 1.0;
    if( nDir == 1 ) {
        scale = (uscale>vscale) ? vscale : uscale;
        uscale = scale;
        vscale = scale;
    }

    nv = aoVs.num();
    for(i=0; i<nv; i++) {
        P = aoVs[i]->GetLoc();
        u = (P[0]-Pmin[0])*uscale;
        v = (P[1]-Pmin[1])*vscale;
        aoVs[i]->SetLoc(R3Pt(u,v,0.0));
    }
}

int 
MCWriteMeshData(const char *filename, const PMeshLite &in_mesh,
                const MCData &in_data, Array<R2Pt> *in_uv)
{
    int i, j, k, ier;
    double area, k1, k2;
    R3Pt P;
    R2Pt Q;
    R3Vec N, k1vec, k2vec;
    Array<Vertex> aopVs;
    Array<Vertex> aovs;
    Array<Vertex> aov;
    Array<Face> aopFs;
    MCPoint cpt;
    const PMeshLite *dataMesh;

    FILE *out = NULL;
    int haveUV = 0;

    out = fopen(filename, "w");
    if( out == NULL ) return -1;

    area = 0.0;

    in_mesh.AllMeshVertices(aopVs);
    in_mesh.AllMeshFaces(aopFs);
    dataMesh = in_data.Mesh();
    dataMesh->AllMeshVertices(aovs);

    if( aovs.num() != aopVs.num() ) return -1;
    if( in_uv != NULL ) {
        if( in_uv->num() == aopVs.num() ) haveUV = 1;
    }

    fprintf(out,"vertID Pt Area, Normal, PrincipCrv1, PrincipCrv2, ");
    fprintf(out,"PrincipVec1, PrincipVec2\n");
    fprintf(out,"%d\n", aopVs.num() );

    if( haveUV ) {
        for( i=0; i<aopVs.num(); i++ ) {
            P = aopVs[i]->GetLoc();
            ier = in_data.getPoint(aovs[i],cpt);
            N = cpt.Normal();
            k1 = cpt.K1();
            k2 = cpt.K2();
            k1vec = cpt.KmaxVec();
            k2vec = cpt.KminVec();

            Q = (*in_uv)[i];
            fprintf(out,"%d %lf %lf %lf %lf ", i, P[0], P[1], P[2], area);
            fprintf(out,"%lf %lf %lf %lf %lf ", N[0], N[1], N[2], k1, k2);
            fprintf(out,"%lf %lf %lf ", k1vec[0], k1vec[1], k1vec[2]);
            fprintf(out,"%lf %lf %lf ", k2vec[0], k2vec[1], k2vec[2]);
            fprintf(out,"%lf %lf\n", Q[0], Q[1]);
        }
    } else {
        for( i=0; i<aopVs.num(); i++ ) {
            P = aopVs[i]->GetLoc();
            ier = in_data.getPoint(aovs[i],cpt);
            N = cpt.Normal();
            k1 = cpt.K1();
            k2 = cpt.K2();
            k1vec = cpt.KmaxVec();
            k2vec = cpt.KminVec();
            fprintf(out,"%d %lf %lf %lf %lf ", i, P[0], P[1], P[2], area);
            fprintf(out,"%lf %lf %lf %lf %lf ", N[0], N[1], N[2], k1, k2);
            fprintf(out,"%lf %lf %lf ", k1vec[0], k1vec[1], k1vec[2]);
            fprintf(out,"%lf %lf %lf\n", k2vec[0], k2vec[1], k2vec[2]);
        }
    }
    
    fprintf(out,"NFaces %d\n", aopFs.num() );

    for( i=0; i<aopFs.num(); i++ ) {
        aov = in_mesh.FaceVertices(aopFs[i]);
        for( j=0; j<aov.num(); j++ ) {
            k = aov[j]->IndexId();
            fprintf(out,"%d ", k);
        }
        fprintf(out,"\n");
    }
    fclose(out);

    return 0;
}

int 
MCWriteMeshData(const char *filename, const PMeshLite &in_mesh,
                const MCData &in_data, const PMeshLite &in_UVmesh)
{
    Array<R2Pt> uv;
    Array<Vertex> aov;
    R3Pt P;
    int i;

    in_UVmesh.AllMeshVertices(aov);

    uv.need(aov.num());
    for( i=0; i<aov.num(); i++ ) {
        P = aov[i]->GetLoc();
        uv[i][0] = P[0];
        uv[i][1] = P[1];
    }

    return MCWriteMeshData(filename, in_mesh, in_data, &uv);
}

int MCWriteMeshWithScalar(const char *filename, const PMeshLite &mesh,
                          const Array<double> &scalar1,
                          const Array<double> &scalar2 )
{
    int i, j, nv, nf, ns1, ns2;
    double s1, s2;
    double Z = 0.0;
    R3Pt P;
	R3Vec N;

    nv = mesh.NumVertices();
    nf = mesh.NumFaces();
    ns1 = scalar1.num();
    ns2 = scalar2.num();
    Array<Vertex> aopVs;
    Array<Face> aopFs;
    Array<Vertex> aov;

    mesh.AllMeshVertices(aopVs);
    mesh.AllMeshFaces(aopFs);

    FILE *out = NULL;
    out = fopen(filename,"w");
    if( out != NULL ) {
        fprintf(out,"vertID Pt Area, Normal, PrincipCrv1, PrincipCrv2, ");
        fprintf(out,"PrincipVec1, PrincipVec2\n");
        fprintf(out,"%d\n",nv);
        for( i=0; i<nv; i++ ) {
            P = aopVs[i]->GetLoc();
            N = aopVs[i]->GetNorm();
            s1 = (i<ns1) ? scalar1[i] : 0.0;
            s2 = (i<ns2) ? scalar2[i] : 0.0;
            fprintf(out,"%d %lf %lf %lf %lf ", i, P[0], P[1], P[2], Z);
            fprintf(out,"%lf %lf %lf %lf %lf ", N[0], N[1], N[2], s1, s2);
            fprintf(out,"%lf %lf %lf %lf %lf %lf\n", Z, Z, Z, Z, Z, Z);
        }
        fprintf(out,"NFaces %d\n", nf );
        for( i=0; i<nf; i++ ) {
            aov = mesh.FaceVertices(aopFs[i]);
            for( j=0; j<aov.num(); j++ ) {
                fprintf(out,"%d ",aov[j]->IndexId());
            }
            fprintf(out,"\n");
        }
        fclose(out);
    }

    return 0;
}

int MCWriteMeshWithScalar(const char *filename, const PMeshLite &mesh,
                          const Array<int> &scalar1,
                          const Array<int> &scalar2 )
{
    int i, j, k, nv, nf, ns1, ns2;
    double s1, s2;
    double Z = 0.0;
    R3Pt P;
	R3Vec N;

    nv = mesh.NumVertices();
    nf = mesh.NumFaces();
    ns1 = scalar1.num();
    ns2 = scalar2.num();
    Array<Vertex> aopVs;
    Array<Face> aopFs;
    Array<Vertex> aov;

    mesh.AllMeshVertices(aopVs);
    mesh.AllMeshFaces(aopFs);

    FILE *out = NULL;
    out = fopen(filename,"w");
    if( out != NULL ) {
        fprintf(out,"vertID Pt Area, Normal, PrincipCrv1, PrincipCrv2, ");
        fprintf(out,"PrincipVec1, PrincipVec2\n");
        fprintf(out,"%d\n",nv);
        for( i=0; i<nv; i++ ) {
            P = aopVs[i]->GetLoc();
            N = aopVs[i]->GetNorm();
            s1 = (i<ns1) ? (double) scalar1[i] : 0.0;
            s2 = (i<ns2) ? (double) scalar2[i] : 0.0;
            fprintf(out,"%d %lf %lf %lf %lf ", i, P[0], P[1], P[2], Z);
            fprintf(out,"%lf %lf %lf %lf %lf ", N[0], N[1], N[2], s1, s2);
            fprintf(out,"%lf %lf %lf %lf %lf %lf\n", Z, Z, Z, Z, Z, Z);
        }
        fprintf(out,"NFaces %d\n", nf );
        for( i=0; i<nf; i++ ) {
            aov = mesh.FaceVertices(aopFs[i]);
            for( j=0; j<aov.num(); j++ ) {
                k = aov[j]->IndexId();
                fprintf(out,"%d ", k);
            }
            fprintf(out,"\n");
        }
        fclose(out);
    }

    return 0;
}

int 
MCWriteGarlandMesh(const char *filename, const PMeshLite &in_mesh)
{
    int i, j, k;
    R3Pt P;
    Array<Vertex> aopVs;
    Array<Vertex> aov;
    Array<Face> aopFs;

    FILE *out = NULL;

    out = fopen(filename, "w");
    if( out == NULL ) return -1;

    in_mesh.AllMeshVertices(aopVs);
    in_mesh.AllMeshFaces(aopFs);

    fprintf(out,"#$SMF 1.0\n");
    fprintf(out,"# created by curvature program, from surface with ??? floaters\n");

    for( i=0; i<aopVs.num(); i++ ) {
        P = aopVs[i]->GetLoc();
        fprintf(out,"v %lf %lf %lf\n", P[0], P[1], P[2]);
    }
    
    fprintf(out,"# %d vertices, %d triangles\n", aopVs.num(), aopFs.num() );

    for( i=0; i<aopFs.num(); i++ ) {
        aov = in_mesh.FaceVertices(aopFs[i]);
        for( j=0; j<aov.num(); j++ ) {
            k = aov[j]->IndexId();
            fprintf(out,"%d ", k+1);
        }
        fprintf(out,"\n");
    }
    fclose(out);

    return 0;
}


int MCScalarStats(const Array<int> &scalar, int &nValues, 
				  Array<int> &Values, Array<int> &groupSizes)
{
	int i, j, k;
	int val;

	nValues = 0;
	Values.need(0);
	groupSizes.need(0);

	// Find the unique values
	for( i=0; i<scalar.num(); i++ ) {
		val = scalar[i];
		j = -1;
		for( k=0; k<Values.num(); k++ ) {
			if( val == Values[k] ) {
				j = k;
				break;
			}
		}
		if( j == -1 ) {
			nValues++;
			Values += val;
		}
	}

	// Count the number of points for each value
	for( k=0; k<Values.num(); k++ ) {
		j = 0;
		for( i=0; i<scalar.num(); i++ ) {
			if( scalar[i] == Values[k] ) j++;
		}
		groupSizes += j;
	}
	nValues = Values.num();

	return 0;
}

void tdgStretchMetric(const R3Pt &A1, const R3Pt &B1, const R3Pt &C1, 
                      const R3Pt &A2, const R3Pt &B2, const R3Pt &C2, 
                      double &area1, double &area2, 
					  double &gamma1, double &gamma2, const double f = 1.0)
{
    int index;
    double a, b, c, d, s, aa, bb, hh;
    double c1, h1, x1, c2, h2, x2, d1[3], d2[3], t;
    R3Vec va, vb, vc;

    // Find the longest side of the triangle 1
    va = C1 - B1;
    vb = A1 - C1;
    vc = B1 - A1;
    d1[0] = sqrt(va[0]*va[0] + va[1]*va[1] + va[2]*va[2]);
    d1[1] = sqrt(vb[0]*vb[0] + vb[1]*vb[1] + vb[2]*vb[2]);
    d1[2] = sqrt(vc[0]*vc[0] + vc[1]*vc[1] + vc[2]*vc[2]);
    index = 2;
    if( d1[0] >= d1[1] && d1[0] >= d1[2] ) index = 0;
    if( d1[1] >= d1[0] && d1[1] >= d1[2] ) index = 1;

    // Compute the perimeter and area
    s = 0.5*(d1[0]+d1[1]+d1[2]);
    area1 = sqrt(s*(s-d1[0])*(s-d1[1])*(s-d1[2]));

    // Compute the scaled planar reference coords for triangle 1
    hh = 2.0*area1/d1[index];
    bb = d1[(index+2)%3];
    c1 = f*d1[index];
    h1 = f*hh;
    x1 = f*sqrt(bb*bb - hh*hh);

    // For triangle 2
    va = C2 - B2;
    vb = A2 - C2;
    vc = B2 - A2;
    d2[0] = sqrt(va[0]*va[0] + va[1]*va[1] + va[2]*va[2]);
    d2[1] = sqrt(vb[0]*vb[0] + vb[1]*vb[1] + vb[2]*vb[2]);
    d2[2] = sqrt(vc[0]*vc[0] + vc[1]*vc[1] + vc[2]*vc[2]);

    // Compute the perimeter and area
    s = 0.5*(d2[0]+d2[1]+d2[2]);
    area2 = sqrt(s*(s-d2[0])*(s-d2[1])*(s-d2[2]));

    // Compute the planar reference coords for triangle 2
    aa = d2[(index+1)%3];
    bb = d2[(index+2)%3];
    c2 = d2[index];
    h2 = 2.0*area2/c2;
    if( aa >= c2 && aa >= bb ) {
        x2 = c2 - sqrt(aa*aa - h2*h2);
    } else {
        x2 = sqrt(bb*bb - h2*h2);
    }

    // Compute the map
    a = c2/c1;
    b = (c1*x2 - c2*x1)/(c1*h1);
    c = 0.0;
    d = h2/h1;

    // Compute the eigenvalues
    t = 4.0*b*c + (a-d)*(a-d);
    if( t < 0.0 ) t = 0.0;
    gamma1 = 0.5*((a+d) + sqrt(t));
    gamma2 = 0.5*((a+d) - sqrt(t));
}


double tdgStretchMetric(const R3Pt &Am, const R3Pt &Bm, const R3Pt &Cm, 
                        const R3Pt &As, const R3Pt &Bs, const R3Pt &Cs, 
                        double &areaM, double &areaS, const double f = 1.0)
{
    int m, n;
    double a, b, c, aa, bb, cc, s, ss, rs, g1, g2;
    double hm, cm1, cm2, d, da, db, dc, stretch, area1, area2;
    double ar[6], u[9], w[9], vt[9];
    R3Vec vma, vmb, vmc, vsa, vsb, vsc;
    R3Pt Ps, P1, P2, P3, Q1, Q2, Q3;
    WINbool good;

    m = 3;
    n = 2;

    // For the mesh M
    vma = Cm - Bm;
    vmb = Am - Cm;
    vmc = Bm - Am;
    da = sqrt(vma[0]*vma[0] + vma[1]*vma[1] + vma[2]*vma[2]);
    db = sqrt(vmb[0]*vmb[0] + vmb[1]*vmb[1] + vmb[2]*vmb[2]);
    dc = sqrt(vmc[0]*vmc[0] + vmc[1]*vmc[1] + vmc[2]*vmc[2]);
    if( da >= db && da >= dc ) {
        P1 = Bm;
        P2 = Cm;
        P3 = Am;
        Q1 = Bs;
        Q2 = Cs;
        Q3 = As;
        a = db;
        b = dc;
        c = da;
    } else {
        if( db >= dc ) {
            P1 = Cm;
            P2 = Am;
            P3 = Bm;
            Q1 = Cs;
            Q2 = As;
            Q3 = Bs;
            a = dc;
            b = da;
            c = db;
        } else {
            P1 = Am;
            P2 = Bm;
            P3 = Cm;
            Q1 = As;
            Q2 = Bs;
            Q3 = Cs;
            a = da;
            b = db;
            c = dc;
        }
    }

    s = 0.5*(a+b+c);
    rs = sqrt(s*(s-a)*(s-b)*(s-c));

    hm  = f*2.0*rs/c;
    cm1 = f*sqrt(b*b - hm*hm);
    cm2 = f*sqrt(a*a - hm*hm);
    areaM = rs;

    // For the sphere S

    // Find the split point on the spherical triangle
    Ps[0] = (Q2[0]*cm1 + Q1[0]*cm2)/(cm1+cm2);
    Ps[1] = (Q2[1]*cm1 + Q1[1]*cm2)/(cm1+cm2);
    Ps[2] = (Q2[2]*cm1 + Q1[2]*cm2)/(cm1+cm2);

    // Put the split point on the sphere
    d = Ps[0]*Ps[0] + Ps[1]*Ps[1] + Ps[2]*Ps[2];
    if( d != 0.0 ) {
        d = 1.0/sqrt(d);
        Ps[0] = Ps[0] * d;
        Ps[1] = Ps[1] * d;
        Ps[2] = Ps[2] * d;
    }

    // For the first sub triangle
    vsa = Ps - Q3;
    vsb = Ps - Q1;
    vsc = Q3 - Q1;
    aa = sqrt(vsa[0]*vsa[0] + vsa[1]*vsa[1] + vsa[2]*vsa[2]);
    bb = sqrt(vsb[0]*vsb[0] + vsb[1]*vsb[1] + vsb[2]*vsb[2]);
    cc = sqrt(vsc[0]*vsc[0] + vsc[1]*vsc[1] + vsc[2]*vsc[2]);

    ss = 0.5*(aa+bb+cc);
    area1 = sqrt(ss*(ss-aa)*(ss-bb)*(ss-cc));

    // Compute the derivatives
    ar[0] = vsa[0]/hm;  // dxds 
    ar[1] = vsa[1]/hm;  // dyds 
    ar[2] = vsa[2]/hm;  // dzds 
    ar[3] = vsb[0]/cm1; // dxdt 
    ar[4] = vsb[1]/cm1; // dydt 
    ar[5] = vsb[2]/cm1; // dzdt 

    // Compute the gammas
    g1 = 0.0;
    g2 = 0.0;
    good = FITToolsSVD::SVD(ar, m, n, u, w, vt);
    if( !good ) return -1;

    if( w[0] != 0.0 ) g1 = 1.0/(w[0]*w[0]);
    if( w[1] != 0.0 ) g2 = 1.0/(w[1]*w[1]);
    //stretch = (g1 + g2) * area1;
    stretch = (w[0]*w[0] + w[1]*w[1]) * area1;

    // For the second sub triangle
    //vsa = Ps - Q3;
    vsb = Ps - Q2;
    vsc = Q3 - Q2;
    //aa = sqrt(vsa[0]*vsa[0] + vsa[1]*vsa[1] + vsa[2]*vsa[2]);
    bb = sqrt(vsb[0]*vsb[0] + vsb[1]*vsb[1] + vsb[2]*vsb[2]);
    cc = sqrt(vsc[0]*vsc[0] + vsc[1]*vsc[1] + vsc[2]*vsc[2]);

    ss = 0.5*(aa+bb+cc);
    area2 = sqrt(ss*(ss-aa)*(ss-bb)*(ss-cc));

    areaS = area1 + area2;

    // Compute the derivatives
    ar[0] = vsa[0]/hm;  // dxds 
    ar[1] = vsa[1]/hm;  // dyds 
    ar[2] = vsa[2]/hm;  // dzds 
    ar[3] = vsb[0]/cm2; // dxdt 
    ar[4] = vsb[1]/cm2; // dydt 
    ar[5] = vsb[2]/cm2; // dzdt 

    // Compute the gammas
    g1 = 0.0;
    g2 = 0.0;
    good = FITToolsSVD::SVD(ar, m, n, u, w, vt);
    if( !good ) return -1;

    if( w[0] != 0.0 ) g1 = 1.0/(w[0]*w[0]);
    if( w[1] != 0.0 ) g2 = 1.0/(w[1]*w[1]);
    //stretch += (g1 + g2) * area2;
    stretch += (w[0]*w[0] + w[1]*w[1]) * area2;

    return stretch;
}


int MCStretchStats(const PMeshLite &in_mesh1, const PMeshLite &in_mesh2,
                   double &out_area1, double &out_area2,
                   double &out_scaleFactor, double &out_stretchEfficiency)
{
    // Determine the stretch efficiency between two meshes
    int i;
    double sumD, sf, area1, area2;
	double sumD2, sumD3, LT, g1, g2;
    Array<Vertex> aov1(3), aov2(3);

    int verbose = 1;

    // Check for valid mesh size
    if( in_mesh1.NumVertices() != in_mesh2.NumVertices() ) return -1;
    if( in_mesh1.NumFaces() != in_mesh2.NumFaces() ) return -1;

    // First pass to compute areas and check connectivity
    out_area1 = 0.0;
    out_area2 = 0.0;
    WINbool connect = TRUE;
    R3Polygon poly(3);
    ForMeshFace( in_mesh1.Mesh(), f1 ) {
        i = f1->IndexId();
        Face f2 = in_mesh2.FaceFromIndexId(i);

        aov1 = in_mesh1.FaceVertices( f1 );
        aov2 = in_mesh2.FaceVertices( f2 );

        // Check for corresponding connectivity
        if( (aov1[0]->IndexId() != aov2[0]->IndexId()) ||
            (aov1[1]->IndexId() != aov2[1]->IndexId()) ||
            (aov1[2]->IndexId() != aov2[2]->IndexId()) ) {
            connect = FALSE;
        }
        
        for ( int j = 0; j < 3; j++ )
            poly[j] = aov1[j]->GetLoc();

        out_area1 += poly.Area();

        for ( FORINT j = 0; j < 3; j++ )
            poly[j] = aov2[j]->GetLoc();

        out_area2 += poly.Area();

    }EndFor;
    if( out_area1 <= 0.0 ) {
        out_scaleFactor = -1.0;
        out_stretchEfficiency = 0.0;
        if( verbose ) printf("WARNING: Mesh 1 has zero area\n");
        return -1;
    }
    if( out_area2 <= 0.0 ) {
        out_scaleFactor = 0.0;
        out_stretchEfficiency = 0.0;
        if( verbose ) printf("WARNING: Mesh 2 has zero area\n");
        return -1;
    }
    sf = sqrt(out_area2 / out_area1);
    out_scaleFactor = sf; 

    // Second pass to compute stretch efficiency
    //bool first = true;
    sumD = 0.0;
    sumD2 = 0.0;
    sumD3 = 0.0;
    ForMeshFace( in_mesh1.Mesh(), f1 ) {
        i = f1->IndexId();
        Face f2 = in_mesh2.FaceFromIndexId(i);

        aov1 = in_mesh1.FaceVertices( f1 );
        aov2 = in_mesh2.FaceVertices( f2 );
        
		tdgStretchMetric(aov1[0]->GetLoc(), 
                         aov1[1]->GetLoc(),
                         aov1[2]->GetLoc(),
                         aov2[0]->GetLoc(), 
                         aov2[1]->GetLoc(),
                         aov2[2]->GetLoc(),
                         area1, area2, g1, g2, sf);

		LT = 0.5*(g1*g1 + g2*g2);
		sumD += LT*area1;

		tdgStretchMetric(aov1[0]->GetLoc(), 
                         aov1[1]->GetLoc(),
                         aov1[2]->GetLoc(),
                         aov2[0]->GetLoc(), 
                         aov2[1]->GetLoc(),
                         aov2[2]->GetLoc(),
                         area1, area2, g1, g2);

		LT = (1.0/(g1*g1) + 1.0/(g2*g2))*area1*out_area2/(out_area1*out_area1)
		   + (    (g1*g1) +     (g2*g2))*area2*out_area1/(out_area2*out_area2);
		sumD2 += LT;

		LT = (    (g1*g1) +     (g2*g2))*area1*out_area2/(out_area1*out_area1)
		   + (1.0/(g1*g1) + 1.0/(g2*g2))*area2*out_area1/(out_area2*out_area2);
		sumD3 += LT;
    }EndFor;
    if( sumD == 0.0 ) {
        out_stretchEfficiency = 0.0;
        if( verbose ) printf("WARNING: Integrated stretch metric denominator is zero \n");
        return -1;
    }

	double s0 = sumD/out_area1;
	double s1 = 2.0/sumD2;
	double s2 = 2.0/sumD3;
    out_stretchEfficiency = s1;

	printf("...Area 1 = %lf, Area 2 = %lf, Scale Factor = %lf\n", 
		    out_area1, out_area2, out_scaleFactor);
	printf("...Stretch Efficiency: Method 1 = %lf\n", out_stretchEfficiency);
	//printf("...Stretch Efficiency: Method 1 = %lf, Method 2 = %lf\n", s1, s2);

	/*
	double s1 = sumD2/(double) (in_mesh1.NumVertices());
	double s2 = sumD3/out_area1;
	double s3 = sumD3/(out_area1+out_area2);
	printf("...sumD2=%lf, s1=%lf\n", sumD2, s1);
    printf("...sumD3=%lf, s2=%lf, s3=%lf\n", sumD3, s2, s3);
	*/

    if(!connect) printf("WARNING: Connectivity inconsistent\n");
    if( verbose ) {
        printf("Stretch Am = %lf, As = %lf, factor = %lf\n", 
                out_area1, out_area2, out_scaleFactor);
        printf("Stretch Efficiency = %lf\n", out_stretchEfficiency);
    }

    if( !connect ) return 1;
    return 0;
}

int MCReadMeshWithScalar(const char *filename, PMeshLite &mesh,
                         Array<double> &scalar1,
                         Array<double> &scalar2 )
{
    int i, nv, nf, v1, v2, v3;
    double s1, s2;
    double Z = 0.0;
    R3Pt P;
	R3Vec N;
	char buffer[256];
	char word[80];

    Array<Vertex> aoVs;
    Array<Face> aoFs;
    Array<Vertex> aov;

    FILE *in = NULL;
    in = fopen(filename,"r");
    if( in == NULL ) return -1;

	mesh.Clear();
    scalar1.need(0);
	scalar2.need(0);

	// Read vertices with scalar data
	fgets(buffer, 256, in);
    fgets(buffer, 256, in);
    sscanf(buffer,"%d", &nv);
    for( i=0; i<nv; i++ ) {
        fgets(buffer, 256, in);
        sscanf(buffer,
			   "%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
	           &i, &P[0], &P[1], &P[2], &Z, &N[0], &N[1], &N[2], 
	           &s1, &s2, &Z, &Z, &Z, &Z, &Z, &Z);
        scalar1 += s1;
		scalar2 += s2;

		// Add vertices to the mesh
        aoVs += mesh.AddVertex(P);
        aoVs.last()->SetNorm(N);
    }

	// Read face data
    fgets(buffer, 256, in);
    sscanf(buffer,"%s %d\n", &word[0], &nf );

    /* Generate faces */
    for( i=0; i<nf; i++ ) {
		fgets(buffer, 256, in);
	    sscanf(buffer,"%d %d %d", &v1, &v2, &v3); 
        aov.need(0);
        aov += aoVs[v1];
        aov += aoVs[v2];
        aov += aoVs[v3];
        aoFs += mesh.AddFace(aov);
    }
    fclose(in);

    return 0;
}

int MCReadMeshWithScalar(const char *filename, PMeshLite &mesh,
                         Array<int> &scalar1,
                         Array<int> &scalar2 )
{
    int i, nv, nf, v1, v2, v3;
    double s1, s2;
    double Z = 0.0;
    R3Pt P;
	R3Vec N;
	char buffer[256];
	char word[80];

    Array<Vertex> aoVs;
    Array<Face> aoFs;
    Array<Vertex> aov;

    FILE *in = NULL;
    in = fopen(filename,"r");
    if( in == NULL ) return -1;

	mesh.Clear();
    scalar1.need(0);
	scalar2.need(0);

	// Read vertices with scalar data
	fgets(buffer, 256, in);
    fgets(buffer, 256, in);
    sscanf(buffer,"%d", &nv);
    for( i=0; i<nv; i++ ) {
        fgets(buffer, 256, in);
        sscanf(buffer,
			   "%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
	           &i, &P[0], &P[1], &P[2], &Z, &N[0], &N[1], &N[2], 
	           &s1, &s2, &Z, &Z, &Z, &Z, &Z, &Z);
        scalar1 += (int) s1;
		scalar2 += (int) s2;

		// Add vertices to the mesh
        aoVs += mesh.AddVertex(P);
        aoVs.last()->SetNorm(N);
    }

	// Read face data
    fgets(buffer, 256, in);
    sscanf(buffer,"%s %d\n", &word[0], &nf );

    /* Generate faces */
    for( i=0; i<nf; i++ ) {
		fgets(buffer, 256, in);
	    sscanf(buffer,"%d %d %d", &v1, &v2, &v3); 
        aov.need(0);
        aov += aoVs[v1];
        aov += aoVs[v2];
        aov += aoVs[v3];
        aoFs += mesh.AddFace(aov);
    }
    fclose(in);

    return 0;
}

int
AlignBins( Array<double> &Kbin1, Array<double> &Kbin2, 
           double &angle, int &flip )
{
    int i, j, k, kf, nBins;
    double err, errf, pi;
    Array<double> error;
    Array<double> error2;

    pi = 4.0 * atan(1.0);
    nBins = Kbin1.num();
    if( Kbin2.num() != nBins ) return -1;

    error.need(nBins);
    error2.need(nBins);

    for( i=0; i<nBins; i++ ) {
        error[i] = 0.0;
        error2[i] = 0.0;
        for( j=0; j<nBins; j++ ) {
            k = i+j;
            kf = nBins - (1+k);
            while( k >= nBins ) {
                k -= nBins;
            }
            while( kf < 0 ) {
                kf += nBins;
            }

            err = Kbin2[k] - Kbin1[j];
            error[i] += err*err;
        
            err = Kbin2[kf] - Kbin1[j];
            error2[i] += err*err;
        }
    }

    k=0;
    kf=0;
    err = error[0];
    errf = error2[0];
    for( i=1; i<nBins; i++ ) {
        if( error[i] < err ) {
            k = i;
            err = error[i];
        }
        if( error2[i] < errf ) {
            kf = i;
            errf = error2[i];
        }
    }
    if( errf < err ) {
        flip = 1;
        angle = 2.0*pi*kf/(double) nBins;
    } else {
        flip = 0;
        angle = 2.0*pi*k/(double) nBins;
    }

    FILE *out = NULL;
    //out = fopen(WINDataHome("error.dat"),"a");
    if( out != NULL ) {
        for( i=0; i<nBins; i++ ) {
            fprintf(out,"Angle index %d: %20.8lf, %20.8lf\n", 
                        i, error[i], error2[i]);
        }
        fclose(out);
        out = NULL;
    }

    return 0;
}

int
SelectBins( Array<double> &Kbin1, Array<double> &Kbin2, 
            int &iBin1, int &iBin2, int &iBin3, int &iBin4 )
{
    int i, nBins, nClose;
    double d, sumsq;
    Array<double> error;
    Array<int> iError;
    FILE *out = NULL;

    // Method: 0 = best fit
    //         1 = max total curvature
    int method = 1;

    out = fopen(WINDataHome("select.dat"),"w");

    nBins = Kbin1.num();
    if( Kbin2.num() != nBins ) return -1;
    nClose = nBins/8;
    //nClose = nBins/16;
    //nClose = 1;

    error.need(nBins);

    switch(method) {
        case 0:
            for( i=0; i<nBins; i++ ) {
                d = Kbin2[i] - Kbin1[i];
                sumsq = Kbin1[i]*Kbin1[i] + Kbin2[i]*Kbin2[i];
                if( sumsq != 0.0 ) {
                    error[i] = sqrt(d*d/sumsq);
                } else {
                    error[i] = sqrt(d*d);
                }
                iError += i;
            }
            break;
        case 1:
            sumsq = 0.0;
            for( i=0; i<nBins; i++ ) {
                d = Kbin2[i] + Kbin1[i];
                error[i] = d*d;
                if( error[i] > sumsq ) sumsq = error[i];
                iError += i;
            }
            for( i=0; i<nBins; i++ ) {
                error[i] = sumsq - error[i];
            }
            break;
        default:
            return -1;
    }

    if( out != NULL ) {
        fprintf(out,"Pre-sort\n");
        for( i=0; i<nBins; i++ ) {
            fprintf(out,"%2d: Error = %lf, Index = %d\n", 
                        i, error[i], iError[i] );
        }
    }

    iError.sort(error);

    if( out != NULL ) {
        fprintf(out,"Post-sort\n");
        for( i=0; i<nBins; i++ ) {
            fprintf(out,"%2d: Error = %lf, Index = %d\n", 
                        i, error[i], iError[i] );
        }
    }

    //iBin1 = iError[nBins-1];
    iBin1 = iError[0];
    for( i=0; i<nBins; i++ ) {
        if( iBin1-nClose <= iError[i] && iError[i] <= iBin1+nClose ) {
            error[i] = -1.0;
        }
        if( iBin1<nClose && iError[i] >= nBins-(nClose-iBin1) ) error[i] = -1.0;
        if( iBin1>=nBins-nClose && iError[i] <= nClose-iBin1 ) error[i] = -1.0;
    }

    if( out != NULL ) {
        fprintf(out,"iBin1 = %d\n", iBin1);
        for( i=0; i<nBins; i++ ) {
            fprintf(out,"%2d: Error = %lf, Index = %d\n", 
                        i, error[i], iError[i] );
        }
    }

    //for( i=nBins-1; i>=0; i-- ) {
    for( i=0; i<nBins; i++ ) {
        if( error[i] != -1.0 ) {
            iBin2 = iError[i];
            break;
        }
    }

    for( i=0; i<nBins; i++ ) {
        if( iBin2-nClose <= iError[i] && iError[i] <= iBin2+nClose ) {
            error[i] = -1.0;
        }
        if( iBin2<nClose && iError[i] >= nBins-(nClose-iBin2) ) error[i] = -1.0;
        if( iBin2>=nBins-nClose && iError[i] <= nClose-iBin2 ) error[i] = -1.0;
    }

    if( out != NULL ) {
        fprintf(out,"iBin2 = %d\n", iBin2);
        for( i=0; i<nBins; i++ ) {
            fprintf(out,"%2d: Error = %lf, Index = %d\n", 
                        i, error[i], iError[i] );
        }
    }

    //for( i=nBins-1; i>=0; i-- ) {
    for( i=0; i<nBins; i++ ) {
        if( error[i] != -1.0 ) {
            iBin3 = iError[i];
            break;
        }
    }

    for( i=0; i<nBins; i++ ) {
        if( iBin3-nClose <= iError[i] && iError[i] <= iBin3+nClose ) {
            error[i] = -1.0;
        }
        if( iBin3<nClose && iError[i] >= nBins-(nClose-iBin3) ) error[i] = -1.0;
        if( iBin3>=nBins-nClose && iError[i] <= nClose-iBin3 ) error[i] = -1.0;
    }

    if( out != NULL ) {
        fprintf(out,"iBin3 = %d\n", iBin3);
        for( i=0; i<nBins; i++ ) {
            fprintf(out,"%2d: Error = %lf, Index = %d\n", 
                        i, error[i], iError[i] );
        }
    }

    //for( i=nBins-1; i>=0; i-- ) {
    for( i=0; i<nBins; i++ ) {
        if( error[i] != -1.0 ) {
            iBin4 = iError[i];
            break;
        }
    }

    if( out != NULL ) {
        fprintf(out,"iBin4 = %d\n", iBin4);
        fclose(out);
    }

    return 0;
}

void
TransMatrix(R3Pt &A1, R3Pt &A2, R3Pt &A3, R3Pt &A4, 
            R3Pt &B1, R3Pt &B2, R3Pt &B3, R3Pt &B4,
            int method, R3Pt &t1, R3Pt &t2, Array<double> &matrix)
{
    int i, j;
    double s, wx, wy, wz;
    double PA[4][3];
    double PB[4][3];
    
    // Setup least square problem
    FITTools_LS prob_ls0( 3, 3, 3 );
    FITTools_LS prob_ls1( 12, 3, 1 );
    FITTools_LS prob_ls2( 12, 4, 1 );

    switch(method) {
        case 0: // General
            t1[0] = A1[0];
            t1[1] = A1[1];
            t1[2] = A1[2];
            t2[0] = B1[0];
            t2[1] = B1[1];
            t2[2] = B1[2];
            break;
        default:
            t1[0] = (A1[0] + A2[0] + A3[0] + A4[0])*0.25;
            t1[1] = (A1[1] + A2[1] + A3[1] + A4[1])*0.25;
            t1[2] = (A1[2] + A2[2] + A3[2] + A4[2])*0.25;
            t2[0] = (B1[0] + B2[0] + B3[0] + B4[0])*0.25;
            t2[1] = (B1[1] + B2[1] + B3[1] + B4[1])*0.25;
            t2[2] = (B1[2] + B2[2] + B3[2] + B4[2])*0.25;
            break;
    }
    
    PA[0][0] = A1[0] - t1[0];
    PA[0][1] = A1[1] - t1[1];
    PA[0][2] = A1[2] - t1[2];
    PA[1][0] = A2[0] - t1[0];
    PA[1][1] = A2[1] - t1[1];
    PA[1][2] = A2[2] - t1[2];
    PA[2][0] = A3[0] - t1[0];
    PA[2][1] = A3[1] - t1[1];
    PA[2][2] = A3[2] - t1[2];
    PA[3][0] = A4[0] - t1[0];
    PA[3][1] = A4[1] - t1[1];
    PA[3][2] = A4[2] - t1[2];

    PB[0][0] = B1[0] - t2[0];
    PB[0][1] = B1[1] - t2[1];
    PB[0][2] = B1[2] - t2[2];
    PB[1][0] = B2[0] - t2[0];
    PB[1][1] = B2[1] - t2[1];
    PB[1][2] = B2[2] - t2[2];
    PB[2][0] = B3[0] - t2[0];
    PB[2][1] = B3[1] - t2[1];
    PB[2][2] = B3[2] - t2[2];
    PB[3][0] = B4[0] - t2[0];
    PB[3][1] = B4[1] - t2[1];
    PB[3][2] = B4[2] - t2[2];

    switch(method) {
        case 0: // General
            // Set up the matrix to solve
            for( i=0; i<3; i++ ) {
                for( j=0; j<3; j++ ) {
                    prob_ls0.Desired_clapack( i, j ) = PA[i+1][j];
                    prob_ls0.Rhs_clapack( i, j ) = PB[i+1][j];
                }
            }
            
            // Do the least squares fit 
            if( prob_ls0.Solve_clapack() == FALSE ) {
                printf("ERROR - Solving Least Square problem\n" );
                prob_ls0.Print_clapack( cout, FALSE );
            }
    
            // Create the transformation matrix
            matrix.need(0);
            for( i=0; i<3; i++ ) {
                for( j=0; j<3; j++ ) {
                    matrix += prob_ls0.Solution_clapack( j, i );
                }
            }
            break;
        case 1:  // Rigid
            // Set up the matrix to solve 
            for( i=0; i<4; i++ ) {
                prob_ls1.Desired_clapack( 3*i, 0 ) =  0.0;
                prob_ls1.Desired_clapack( 3*i, 1 ) =  PA[i][2];
                prob_ls1.Desired_clapack( 3*i, 2 ) = -PA[i][1];
                prob_ls1.Rhs_clapack( 3*i, 0 )     =  PB[i][0] - PA[i][0];
                
                prob_ls1.Desired_clapack( 3*i+1, 0 ) = -PA[i][2];
                prob_ls1.Desired_clapack( 3*i+1, 1 ) =  0.0;
                prob_ls1.Desired_clapack( 3*i+1, 2 ) =  PA[i][0];
                prob_ls1.Rhs_clapack( 3*i+1, 0 )     =  PB[i][1] - PA[i][1];
                
                prob_ls1.Desired_clapack( 3*i+2, 0 ) =  PA[i][1];
                prob_ls1.Desired_clapack( 3*i+2, 1 ) = -PA[i][0];
                prob_ls1.Desired_clapack( 3*i+2, 2 ) =  0.0;
                prob_ls1.Rhs_clapack( 3*i+2, 0 )     =  PB[i][2] - PA[i][2];
            }
                
            // Do the least squares fit 
            if( prob_ls1.Solve_clapack() == FALSE ) {
                printf("ERROR - Solving Least Square problem\n" );
                prob_ls1.Print_clapack( cout, FALSE );
            }
    
            // Get the matrix values
            wx = prob_ls1.Solution_clapack( 0, 0 );
            wy = prob_ls1.Solution_clapack( 1, 0 );
            wz = prob_ls1.Solution_clapack( 2, 0 );
            s  = 1.0;
            
            // Create the transformation matrix
            matrix.need(0);
            matrix +=  1.0;
            matrix += -prob_ls1.Solution_clapack( 2, 0 );
            matrix +=  prob_ls1.Solution_clapack( 1, 0 );

            matrix +=  prob_ls1.Solution_clapack( 2, 0 );
            matrix +=  1.0;
            matrix += -prob_ls1.Solution_clapack( 0, 0 );
            
            matrix += -prob_ls1.Solution_clapack( 1, 0 );
            matrix +=  prob_ls1.Solution_clapack( 0, 0 );
            matrix +=  1.0;
            break;
        case 2:  // Rigid + scaling
            // Set up the matrix to solve 
            for( i=0; i<4; i++ ) {
                prob_ls2.Desired_clapack( 3*i, 0 ) =  0.0;
                prob_ls2.Desired_clapack( 3*i, 1 ) =  PA[i][2];
                prob_ls2.Desired_clapack( 3*i, 2 ) = -PA[i][1];
                prob_ls2.Desired_clapack( 3*i, 3 ) =  PA[i][0];
                prob_ls2.Rhs_clapack( 3*i, 0 )     =  PB[i][0];
                
                prob_ls2.Desired_clapack( 3*i+1, 0 ) = -PA[i][2];
                prob_ls2.Desired_clapack( 3*i+1, 1 ) =  0.0;
                prob_ls2.Desired_clapack( 3*i+1, 2 ) =  PA[i][0];
                prob_ls2.Desired_clapack( 3*i+1, 3 ) =  PA[i][1];
                prob_ls2.Rhs_clapack( 3*i+1, 0 )     =  PB[i][1];
                
                prob_ls2.Desired_clapack( 3*i+2, 0 ) =  PA[i][1];
                prob_ls2.Desired_clapack( 3*i+2, 1 ) = -PA[i][0];
                prob_ls2.Desired_clapack( 3*i+2, 2 ) =  0.0;
                prob_ls2.Desired_clapack( 3*i+2, 3 ) =  PA[i][2];
                prob_ls2.Rhs_clapack( 3*i+2, 0 )     =  PB[i][2];
            }
                
            // Do the least squares fit 
            if( prob_ls2.Solve_clapack() == FALSE ) {
                printf("ERROR - Solving Least Square problem\n" );
                prob_ls2.Print_clapack( cout, FALSE );
            }
    
            // Get the matrix values
            wx = prob_ls2.Solution_clapack( 0, 0 );
            wy = prob_ls2.Solution_clapack( 1, 0 );
            wz = prob_ls2.Solution_clapack( 2, 0 );
            s  = prob_ls2.Solution_clapack( 3, 0 );
            if( s != 0.0 ) {
                wx /= s;
                wy /= s;
                wz /= s;
            }
            
            // Create the transformation matrix
            matrix.need(0);
            matrix +=  prob_ls2.Solution_clapack( 3, 0 );
            matrix += -prob_ls2.Solution_clapack( 2, 0 );
            matrix +=  prob_ls2.Solution_clapack( 1, 0 );

            matrix +=  prob_ls2.Solution_clapack( 2, 0 );
            matrix +=  prob_ls2.Solution_clapack( 3, 0 );
            matrix += -prob_ls2.Solution_clapack( 0, 0 );
            
            matrix += -prob_ls2.Solution_clapack( 1, 0 );
            matrix +=  prob_ls2.Solution_clapack( 0, 0 );
            matrix +=  prob_ls2.Solution_clapack( 3, 0 );
            break;
    }
}

void
Transform(PMeshLite &mesh, R3Pt &T1, R3Pt &T2, Array<double> &matrix)
{
    int i;
    Array<Vertex> aoVs;
    R3Pt P, Q;

    mesh.AllMeshVertices(aoVs);

    for( i=0; i<aoVs.num(); i++ ) {
        P = aoVs[i]->GetLoc();
        Q[0] = P[0] - T1[0];
        Q[1] = P[1] - T1[1];
        Q[2] = P[2] - T1[2];

        // Apply the transformation matrix
        P[0] = matrix[0]*Q[0] + matrix[1]*Q[1] + matrix[2]*Q[2];
        P[1] = matrix[3]*Q[0] + matrix[4]*Q[1] + matrix[5]*Q[2];
        P[2] = matrix[6]*Q[0] + matrix[7]*Q[1] + matrix[8]*Q[2];

        Q[0] = P[0] + T2[0];
        Q[1] = P[1] + T2[1];
        Q[2] = P[2] + T2[2];

        aoVs[i]->SetLoc(Q);
    }
}

R3Pt
Transform(R3Pt &in_P, R3Pt &T1, R3Pt &T2, Array<double> &matrix)
{
    Array<Vertex> aoVs;
    R3Pt P, Q;

    P = in_P;
    Q[0] = P[0] - T1[0];
    Q[1] = P[1] - T1[1];
    Q[2] = P[2] - T1[2];

        // Apply the transformation matrix
    P[0] = matrix[0]*Q[0] + matrix[1]*Q[1] + matrix[2]*Q[2];
    P[1] = matrix[3]*Q[0] + matrix[4]*Q[1] + matrix[5]*Q[2];
    P[2] = matrix[6]*Q[0] + matrix[7]*Q[1] + matrix[8]*Q[2];

    Q[0] = P[0] + T2[0];
    Q[1] = P[1] + T2[1];
    Q[2] = P[2] + T2[2];

    return Q;
}

int 
FindMaxPoints( PMeshLite &mesh, Array<double> &func, int nPts, int nSep,
               Array<R3Pt> &pts, Array<double> &val, Array<int> &ndx )
{
    int i, j, k, nv;
    Array <Vertex> aoVs;
    Array <Vertex> aov;
    Array <Face> aof;
    Array <double> value;
    Array <int> index;

    FILE *out = NULL;
    out = fopen(WINDataHome("sort.dat"),"a");

    nv = mesh.NumVertices();
    if( func.num() != nv ) return -1;

    mesh.AllMeshVertices(aoVs);
    pts.need(0);
    val.need(0);
    ndx.need(0);

    for( i=0; i<nv; i++ ) {
        value += func[i]*func[i];
        index += i;
    }

    if( out != NULL ) {
        fprintf(out,"\nPre-sort\n");
        for( k=0; k<nv; k++ ) {
            fprintf(out,"%2d: Value = %lf (%lf), Index = %d\n", 
                        k, value[k], func[k], index[k] );
        }
    }

    index.sort(value);

    if( out != NULL ) {
        fprintf(out,"\nInitial sort\n");
        for( k=0; k<nv; k++ ) {
            fprintf(out,"%2d: Value = %lf, Index = %d\n", 
                        k, value[k], index[k] );
        }
    }

    for( i=0; i<nPts; i++ ) {
        for( j=nv-1; j>=0; j-- ) {
            if( value[j] != -1.0 ) {
                pts += aoVs[index[j]]->GetLoc();
                val += func[index[j]];
                ndx += index[j];
                break;
            }
        }
        if( pts.num() <= i ) return -(1000+i);
        mesh.GrowRings(aoVs[j], nSep, aov, aof);
        for( j=0; j<aov.num(); j++ ) {
            k = aov[j]->IndexId();
            value[k] = -1.0;
        }

        if( out != NULL ) {
            fprintf(out,"\nStep %d: Vertex %d\n", i, ndx[i], val[i]);
            for( k=0; k<nv; k++ ) {
                fprintf(out,"%2d: Value = %lf, Index = %d\n", 
                            k, value[k], index[k] );
            }
        }
    }

    if( out != NULL ) fclose(out);

    return 0;       
}

R3Pt 
Centroid( Array<R3Pt> &pts )
{
    int i, n;
    double x, y, z;
    R3Pt P;

    n = 0;
    x = 0.0;
    y = 0.0;
    z = 0.0;
    for( i=0; i<pts.num(); i++ ) {
        P = pts[i];
        x += P[0];
        y += P[1];
        z += P[2];
        n++;
    }
    if( n > 0 ) {
        x /= (double) n;
        y /= (double) n;
        z /= (double) n;
    }
    return R3Pt(x,y,z);
}

int BestCorrespondence( Array<R3Pt> &PA, Array<double> &VA, Array<int> &IA,
                        Array<R3Pt> &PB, Array<double> &VB, Array<int> &IB,
                        Array<int> &CA, Array<int> &CB )
{
    int i, j, k, l, ii, jj, kk, ll, n;
    double dd, df, d1, d2, d3, d4, f1, f2, f3, f4;
    double wd, wf, dRef, fRef, err, error;
    R3Vec v1, v2, v3, v4;

    wd = 1.0;
    wf = 1.0;

    n = PA.num();
    if( PB.num() != n || VA.num() != n || VB.num() != n ||
                         IA.num() != n || IB.num() != n ) return -1;

    if( n < 4 ) return -2;

    dRef = 0.0;
    fRef = 0.0;
    for( i=0; i<4; i++ ) {
        for( j=0; j<4; j++ ) {
            df = VB[j] - VA[i];
            if( df*df > fRef ) fRef = df*df;
            v1 = PB[j] - PA[i];
            dd = v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2];
            if( dd > dRef ) dRef = dd;
        }
    }
    dRef = sqrt(dRef);
    fRef = sqrt(fRef);
    if( dRef == 0.0 ) return -3;
    wd /= dRef;
    if( fRef != 0.0 ) wf /= fRef;

    CA.need(0);
    CB.need(0);

    ii = -1;
    error = 0.0;

    for( i=0; i<4; i++ ) {
        v1 = PB[i] - PA[0];
        for( j=0; j<4; j++ ) {
            if( j == i ) continue;
            v2 = PB[j] - PA[1];
            for( k=0; k<4; k++ ) {
                if( k == i || k == j ) continue;
                v3 = PB[k] - PA[2];
                for( l=0; l<4; l++ ) {
                    if( l == i || l == j || l == k ) continue;
                    v4 = PB[l] - PA[3];
                    d1 = sqrt(v1[0]*v1[0]+v1[1]*v1[1]+v1[2]*v1[2]);
                    d2 = sqrt(v2[0]*v2[0]+v2[1]*v2[1]+v2[2]*v2[2]);
                    d3 = sqrt(v3[0]*v3[0]+v3[1]*v3[1]+v3[2]*v3[2]);
                    d4 = sqrt(v4[0]*v4[0]+v4[1]*v4[1]+v4[2]*v4[2]);
                    f1 = sqrt( (VB[i]-VA[0]) * (VB[i]-VA[0]) );
                    f2 = sqrt( (VB[j]-VA[1]) * (VB[j]-VA[1]) );
                    f3 = sqrt( (VB[k]-VA[2]) * (VB[k]-VA[2]) );
                    f4 = sqrt( (VB[l]-VA[3]) * (VB[l]-VA[3]) );
                    err = wd*(d1+d2+d3+d4) + wf*(f1+f1+f3+f4);
                    if( ii == -1 ) {
                        ii = i;
                        jj = j;
                        kk = k;
                        ll = l;
                        error = err;
                    }
                    if( err < error ) {
                        ii = i;
                        jj = j;
                        kk = k;
                        ll = l;
                        error = err;
                    }
                }
            }
        }
    }
    CA += 0;
    CA += 1;
    CA += 2;
    CA += 3;
    CB += ii;
    CB += jj;
    CB += kk;
    CB += ll;

    return 0;
}

int 
MCWriteGridData(const char *filename, int ni, int nj, 
                Array<R2Pt> grid, Array<double> data)
{
    int i, j, i0, i1, i2, i3;
    double area, k1, k2, x, y, z, d;
    R2Pt P;
    R3Vec N, k1vec, k2vec;
    //const PMeshLite *dataMesh;

    FILE *out = NULL;

    out = fopen(filename, "w");
    if( out == NULL ) return -1;

    area = 0.0;
    N = R3Vec(0.0, 0.0, 0.0);
    k1 = 0.0;
    k2 = 0.0;
    k1vec = R3Vec(1.0, 0.0, 0.0);
    k2vec = R3Vec(0.0, 1.0, 0.0);

    fprintf(out,"vertID Pt Area, Normal, PrincipCrv1, PrincipCrv2, ");
    fprintf(out,"PrincipVec1, PrincipVec2\n");
    fprintf(out,"%d\n", ni*nj );

    for( i=0; i<ni*nj; i++ ) {
        x = grid[i][0];
        y = grid[i][1];
        z = 0.0;
        d = data[i];
        fprintf(out,"%d %lf %lf %lf %lf ", i, x, y, z, area);
        fprintf(out,"%lf %lf %lf %lf %lf ", N[0], N[1], N[2], d, k2);
        fprintf(out,"%lf %lf %lf ", k1vec[0], k1vec[1], k1vec[2]);
        fprintf(out,"%lf %lf %lf\n", k2vec[0], k2vec[1], k2vec[2]);
    }
    
    fprintf(out,"NFaces %d\n", 2*(ni-1)*(nj-1) );

    for( j=0; j<nj-1; j++ ) {
        for( i=0; i<ni-1; i++ ) {
            i0 = i + j*ni;
            i1 = i0+1;
            i2 = i0+ni;
            i3 = i2+1;
            fprintf(out,"%d %d %d\n", i0, i1, i3 );
            fprintf(out,"%d %d %d\n", i0, i3, i2 );
        }
    }
    fclose(out);

    return 0;
}

int 
MCWrite4PointMesh(const char *filename, const R3Pt &P1, const R3Pt &P2, 
                  const R3Pt &P3, const R3Pt &P4 )
{
    int i;
    double area, k1, k2;
    R3Pt P;
    R2Pt Q;
    R3Vec N, k1vec, k2vec;

    FILE *out = NULL;
    int haveUV = 0;

    out = fopen(filename, "w");
    if( out == NULL ) return -1;

    area = 0.0;
    k1 = 0.0;
    k2 = 0.0;
    for( i=0; i<3; i++ ) {
        N[i] = 0.0;
        k1vec[i] = 0.0;
        k2vec[i] = 0.0;
    }

    fprintf(out,"vertID Pt Area, Normal, PrincipCrv1, PrincipCrv2, ");
    fprintf(out,"PrincipVec1, PrincipVec2\n");
    fprintf(out,"%d\n", 4 );

    fprintf(out,"%d %lf %lf %lf %lf ", 0, P1[0], P1[1], P1[2], area);
    fprintf(out,"%lf %lf %lf %lf %lf ", N[0], N[1], N[2], k1, k2);
    fprintf(out,"%lf %lf %lf ", k1vec[0], k1vec[1], k1vec[2]);
    fprintf(out,"%lf %lf %lf\n", k2vec[0], k2vec[1], k2vec[2]);
    
    fprintf(out,"%d %lf %lf %lf %lf ", 1, P2[0], P2[1], P2[2], area);
    fprintf(out,"%lf %lf %lf %lf %lf ", N[0], N[1], N[2], k1, k2);
    fprintf(out,"%lf %lf %lf ", k1vec[0], k1vec[1], k1vec[2]);
    fprintf(out,"%lf %lf %lf\n", k2vec[0], k2vec[1], k2vec[2]);
    
    fprintf(out,"%d %lf %lf %lf %lf ", 2, P3[0], P3[1], P3[2], area);
    fprintf(out,"%lf %lf %lf %lf %lf ", N[0], N[1], N[2], k1, k2);
    fprintf(out,"%lf %lf %lf ", k1vec[0], k1vec[1], k1vec[2]);
    fprintf(out,"%lf %lf %lf\n", k2vec[0], k2vec[1], k2vec[2]);
    
    fprintf(out,"%d %lf %lf %lf %lf ", 3, P4[0], P4[1], P4[2], area);
    fprintf(out,"%lf %lf %lf %lf %lf ", N[0], N[1], N[2], k1, k2);
    fprintf(out,"%lf %lf %lf ", k1vec[0], k1vec[1], k1vec[2]);
    fprintf(out,"%lf %lf %lf\n", k2vec[0], k2vec[1], k2vec[2]);
    
    fprintf(out,"NFaces %d\n", 4 );
    fprintf(out,"%d %d %d\n", 0, 1, 2 );
    fprintf(out,"%d %d %d\n", 0, 3, 1 );
    fprintf(out,"%d %d %d\n", 1, 3, 2 );
    fprintf(out,"%d %d %d\n", 2, 3, 0 );

    fclose(out);

    return 0;
}

void
DumpMatrix(char *filename, char *desc, R3Pt &T1, R3Pt &T2, Array<double> &m)
{
    FILE *out = NULL;

    out = fopen(filename, "a");

    if( out != NULL ) {
        fprintf(out,"\n");
        fprintf(out,"%s\n", desc);
        fprintf(out,"T1: (%lf, %lf, %lf)\n", T1[0], T1[1], T1[2]);
        fprintf(out,"Matrix: %15.7lf, %15.7lf, %15.7lf\n", m[0], m[1], m[2]);
        fprintf(out,"      : %15.7lf, %15.7lf, %15.7lf\n", m[3], m[4], m[5]);
        fprintf(out,"      : %15.7lf, %15.7lf, %15.7lf\n", m[6], m[7], m[8]);
        fprintf(out,"T2: (%lf, %lf, %lf)\n", T2[0], T2[1], T2[2]);
    }
}

void RampColor(const double val,
               const double valmin, const double valmax,
               const int type, double &red,
               double &green, double &blue )
{
    /* type: 0 = decreasing blue to magenta */
    /*       1 = increasing blue to magenta */
    /*       2 = decreasing grayscale */
    /*       3 = increasing grayscale */
    int seg;
    double fract, v, vmin, vmax;

	vmin = 0.05;
	vmax = 0.95;

    if( valmax <= valmin ) {
        red   = 0.5;
        green = 0.5;
        blue  = 0.5;
        return;
    }

    v = (val - valmin) / (valmax - valmin);
    switch(type) {
    // Color
    case 0:
        v = 1.0 - v;
    case 1:
        seg = (int) ( 5.0 * v );
        seg = seg<0 ? 0 : ( seg>4 ? 4 : seg );
        fract = 5.0 * v - seg;
        fract = fract<0.0 ? 0.0 : ( fract>1.0 ? 1.0 : fract );
        switch( seg ) {
        case 0:
            red = 0.0;
            green = fract;
            blue = 1.0;
            break;
        case 1:
            red = 0.0;
            green = 1.0;
            blue = 1.0-fract;
            break;
        case 2:
            red = fract;
            green = 1.0;
            blue = 0.0;
            break;
        case 3:
            red = 1.0;
            green = 1.0-fract;
            blue = 0.0;
            break;
        case 4:
            red = 1.0;
            green = 0.0;
            blue = fract;
            break;
        default:
            red = 1.0;
            green = 1.0;
            blue = 1.0;
            break;
        }
        break;

    // Grayscale
    case 2:
        v = 1.0 - v;
    case 3:
		if( vmax > vmin ) {
			v = (v*vmax + (1.0-v)*vmin) / (vmax-vmin);
		}
        red   = v;
        green = red;
        blue  = red;
        break;
    }
}

int MCReadFanData(FILE *in, Array<int> &spokeSize,
			                Array<R3Pt> &xyz, 
			                Array<double> &K1, 
						    Array<double> &K2,
							Array<int> &valid)
{
	int i, j, k, nSpoke, nread, v;
	int size[10];
	double x, y, z, k1, k2;
	char line[132];
	R3Pt P;
	char *ptr;

	if( in == NULL ) return 0;

	// Initialize the arrays
	spokeSize.need(0);
	xyz.need(0);
	K1.need(0);
	K2.need(0);
	valid.need(0);

	// Read the number of spokes
	ptr = fgets(line, 132, in);
	nread = sscanf(line,"%d", &nSpoke );
	if( nread != 1 ) return -1;

	// Read the number of points in each spoke
	k = 0;
	do
	{
		ptr = fgets(line, 132, in);
		if( ptr == NULL ) return -2;

		nread = sscanf(line,"%d %d %d %d %d %d %d %d %d %d",
			           &size[0], &size[1], &size[2], &size[3], &size[4],
					   &size[5], &size[6], &size[7], &size[8], &size[9]);
		if( (nSpoke-k < 11 && nread != nSpoke-k) ||
			(nSpoke-k > 10 && nread != 10 ) ) return -2;
		for( i=0; i<nread; i++ ) {
			spokeSize += size[i];
			k++;
		}
	} while( k < nSpoke );

	// Read the center point
	ptr = fgets(line, 132, in);
	nread = sscanf(line,"%lf %lf %lf %lf %lf %d", &x, &y, &z, &k1, &k2, &v);
	if( nread < 3 ) return -3;
	xyz += R3Pt(x, y, z);
	if( nread >= 5 ) {
		K1 += k1;
		K2 += k2;
		valid += v;
	}

	// Read the spoke points
	for( i=0; i<nSpoke; i++ ) {
		for( j=0; j<spokeSize[i]; j++ ) {
			ptr = fgets(line, 132, in);
			nread = sscanf(line,"%lf %lf %lf %lf %lf %d", &x, &y, &z, &k1, &k2, &v);
			if( nread < 3 ) return -3;
			xyz += R3Pt(x, y, z);
			if( nread >= 5 ) {
				K1 += k1;
				K2 += k2;
				valid += v;
			}
		}
	}
	return 0;
}

int MCReadFanDataBinary(FILE *in, Array<int> &spokeSize,
			                      Array<R3Pt> &xyz, 
			                      Array<double> &K1, 
						          Array<double> &K2,
							      Array<int> &valid)
{
	int ier, i, j, nSpoke, v, n;
	//int k, nread;
	//int size[10];
	//double x, y, z;
	double k1, k2;
	//char line[132];
	R3Pt P;
	//char *ptr;

	if( in == NULL ) return 0;

	// Initialize the arrays
	spokeSize.need(0);
	xyz.need(0);
	K1.need(0);
	K2.need(0);
	valid.need(0);

	// Read the number of spokes
	ier = fread(&nSpoke, sizeof(int), 1, in);

	// Read the number of points in each spoke
	for( i=0; i<nSpoke; i++ ) {
		ier = fread(&n, sizeof(int), 1, in);
		spokeSize += n;
	}

	// Read the center point
	ier = fread( &P, sizeof(R3Pt), 1, in);
	xyz += R3Pt(P[0], P[1], P[2]);

	ier = fread( &k1, sizeof(double), 1, in);
	ier = fread( &k2, sizeof(double), 1, in);
	ier = fread( &v, sizeof(int), 1, in);
	K1 += k1;
	K2 += k2;
	valid += v;

	// Read the spoke points
	for( i=0; i<nSpoke; i++ ) {
		for( j=0; j<spokeSize[i]; j++ ) {
			ier = fread( &P, sizeof(R3Pt), 1, in);
			xyz += R3Pt(P[0], P[1], P[2]);

			ier = fread( &k1, sizeof(double), 1, in);
			ier = fread( &k2, sizeof(double), 1, in);
			ier = fread( &v, sizeof(int), 1, in);
			K1 += k1;
			K2 += k2;
			valid += v;
		}
	}
	return 0;
}

int MCWriteFan(FILE *out, Array<int> &spokeSize,
			              Array<R3Pt> &xyz, 
			              Array<double> &K1, 
						  Array<double> &K2,
						  Array<int> &valid)
{
	int i, j, k;
	R3Pt P;

	if( out == NULL ) return 0;

	fprintf(out,"%d\n", spokeSize.num() );
	for( i=0; i<spokeSize.num(); i++ ) {
		fprintf(out,"%d ", spokeSize[i]);
		if( (i+1)%10 == 0 ) fprintf(out,"\n");
	}
	if( i%10 != 0 ) fprintf(out,"\n");

	// Write out the center point
	k = 0;
	P = xyz[k];
	fprintf(out,"%lf %lf %lf", P[0], P[1], P[2]);
	if( K1.num() > 0 && K2.num() > 0 ) {
		fprintf(out," %lf %lf %d", K1[k], K2[k], valid[k]);
	}
	fprintf(out,"\n");
	k++;

	for( i=0; i<spokeSize.num(); i++ ) {
		for( j=0; j<spokeSize[i]; j++ ) {
			if( xyz.num() > k ) {
				P = xyz[k];
			} else {
				fprintf(out,"k out of range[1-%d]: %d\n", xyz.num(), k);
			}
			fprintf(out,"%lf %lf %lf", P[0], P[1], P[2]);
			if( K1.num() > k && K2.num() > k ) {
				fprintf(out," %lf %lf %d", K1[k], K2[k], valid[k]);
			}
			fprintf(out,"\n");
			k++;
		}
	}

	return 0;
}

int MCWriteFanBinary(FILE *out, Array<int> &spokeSize,
			                    Array<R3Pt> &xyz, 
			                    Array<double> &K1, 
						        Array<double> &K2,
						        Array<int> &valid)
{
	int ier, i, j, k, n;
	//double v;
	R3Pt P;

	if( out == NULL ) return 0;

	n = spokeSize.num();
	ier = fwrite(&n, sizeof(int), 1, out);
	for( i=0; i<spokeSize.num(); i++ ) {
		n = spokeSize[i];
		ier = fwrite(&n, sizeof(int), 1, out);
	}

	// Write out the center point
	k = 0;
	P = xyz[k];
	ier = fwrite(&P, sizeof(R3Pt), 1, out);
	if( K1.num() > 0 && K2.num() > 0 ) {
		ier = fwrite(&K1[k], sizeof(double), 1, out);
		ier = fwrite(&K2[k], sizeof(double), 1, out);
		ier = fwrite(&valid[k], sizeof(int), 1, out);
	}
	k++;

	for( i=0; i<spokeSize.num(); i++ ) {
		for( j=0; j<spokeSize[i]; j++ ) {
			if( xyz.num() > k ) {
				P = xyz[k];
			} else {
				printf("k out of range[1-%d]: %d\n", xyz.num(), k);
			}
			fwrite(&P, sizeof(R3Pt), 1, out);
			//fprintf(out,"%lf %lf %lf", P[0], P[1], P[2]);
			if( K1.num() > k && K2.num() > k ) {
				fwrite(&K1[k], sizeof(double), 1, out);
				fwrite(&K2[k], sizeof(double), 1, out);
				fwrite(&valid[k], sizeof(int), 1, out);
			}
			k++;
		}
	}

	return 0;
}

int MCWriteFanMesh(const char *filename,
				   Array<int> &spokeSize,
			       Array<R3Pt> &xyz, 
			       Array<double> &K1, 
				   Array<double> &K2,
				   Array<int> &valid)
{
    /*
	int ier;
	ArrayMesh mesh;

	ier = mesh.FromFan(spokeSize, xyz, valid);

	ier = mesh.WriteMeshWithScalar(filename, K1, K2);
*/
	return 0;
}

int MCMakeFanMesh(const char *inputFile,
	 			  const char *outputFile,
				  const int index)
{
	int ier, i;
	Array<int> spokeSize;
	Array<R3Pt> xyz;
	Array<double> K1;
	Array<double> K2;
	Array<int> valid;

	FILE *in = NULL;
	//in = fopen(inputFile,"r");
	in = fopen(inputFile,"rb");
	if( in == NULL ) return -1;

	for( i=0; i<=index; i++ ) {
		//ier = MCReadFanData(in, spokeSize, xyz, K1, K2, valid);
		ier = MCReadFanDataBinary(in, spokeSize, xyz, K1, K2, valid);
		if( ier != 0 ) return ier;
	}

	ier = MCWriteFanMesh(outputFile, spokeSize, xyz, K1, K2, valid);
	return ier;
}

double MCGetLength(const Array<Vertex> &aov)
{
	int i;
	double d;
	R3Vec v;

	d = 0;
	for( i=1; i<aov.num(); i++ ) {
		v = aov[i]->GetLoc() - aov[i-1]->GetLoc();
		d += sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
	}
	return d;
}