#include "StdAfx.H"
#include <MC/MC_Param.H>
#include <MC/MC_Point.H>
#include <MC/MC_Fit.H>
#include <utils/Pm_MeshLite.H>
#include <fitting/FITTools_Least_squares.H>

MCParam::MCParam(const PMeshLite &in_mesh, 
                 const int in_nRings,
                 const MC_ParamType &in_paramType)
: MCMeshSubset(in_mesh,in_nRings)
, m_ParamType(in_paramType)
{
    m_lambda = 1.0;
    m_mu = 0.0;
}
    
MCParam::~MCParam()
{
}

int
MCParam::CalculateUV(const Vertex &in_vertex, const int in_nRings)
{
    if( !valid || in_vertex != m_Vertex || in_nRings != m_iRings ) {
        Create(in_vertex, in_nRings);
    }

    switch(m_ParamType) {
        case PLANAR:
            return CalculateUVplanar();
        case CIRCULAR:
            return CalculateUVcircular(m_lambda, m_mu);
        case NATURAL:
            return CalculateUVnatural(m_lambda, m_mu);
        default:
            break;
    }

    return -1;
}

int 
MCParam::CalculateUVplanar()
{
    R3Pt p;
    R3Vec origin, normal, pvec, uvec, vvec;
    int i;
    double u, v;
    double umin, umax, vmin, vmax;

    normal = m_Vertex->GetNorm();
    normal.Normalize();
    //normal.unitSafe();

    m_PtXYZ = m_Vertex->GetLoc();
    origin = R3Vec(m_PtXYZ[0],m_PtXYZ[1],m_PtXYZ[2]);
    
    // Find the u direction 
    if( abs(Dot(normal,R3Vec(1.0,0.0,0.0))) >
        abs(Dot(normal,R3Vec(0.0,1.0,0.0))) ) {
        uvec = Cross( normal, R3Vec(0.0,1.0,0.0) );
    } else {
        uvec = Cross( normal, R3Vec(1.0,0.0,0.0) );
    }
    uvec.Normalize();
    
    vvec = Cross(normal,uvec);
    vvec.Normalize();

    // Compute u,v coordinates 
    m_aUVs.need(m_aopVs.num());
    for(i=0; i<m_aopVs.num(); i++) {
        p = m_aopVs[i]->GetLoc();
        pvec = R3Vec(p[0],p[1],p[2]) - origin;
        u = Dot(pvec,uvec);
        v = Dot(pvec,vvec);
        if(i==0) {
            umin = u;
            umax = u;
            vmin = v;
            vmax = v;
        } else {
            umin = u<umin ? u : umin;
            umax = u>umax ? u : umax;
            vmin = v<vmin ? v : vmin;
            vmax = v>vmax ? v : vmax;
        }
        m_aUVs[i] = R2Pt(u,v);
    }
    m_UVmin = R2Pt(umin,vmin);
    m_UVmax = R2Pt(umax,vmax);
    m_PtUV = m_aUVs[m_aopVs.index(m_Vertex)];

    return 0;
}

int 
MCParam::CalculateUVcircular(const double in_lambda, const double in_mu)
{
    Array< Vertex > aov;
    Array< int > IndexIds;
    Array< int > Order;
    Array<R2Pt> BoundaryUV;
    int i, j, jj, nInterior, nBoundary, nVertex;
    double d, mija, mija_sum, mijx, mijx_sum, u, v;

    FILE *out = NULL;
    //out = fopen(WINDataHome("test.out"),"a");
    if( out != NULL ) MCMeshSubset::dump(out);

    // Count the number of interior and boundary nodes 
    nInterior = 0;
    nBoundary = 0;
    nVertex = m_aopVs.num();
    for(i=0; i<nVertex; i++) {
        IndexIds += m_aopVs[i]->IndexId();
        if(IsBoundary(m_aopVs[i])) {
            nBoundary++;
        } else {
            nInterior++;
        }
    }

    ParameterizeBoundary(Order,BoundaryUV);
    if( BoundaryUV.num() != nBoundary ) {
        return -1;
    }

    if( out != NULL ) {
        for( i=0; i<Order.num(); i++ ) {
            fprintf(out,"Order %d = %d\n", i, Order[i] );
        }

        for( i=0; i<BoundaryUV.num(); i++ ) {
            fprintf(out,"BoundaryUV %d = (%lf, %lf)\n", 
                        i, BoundaryUV[i][0], BoundaryUV[i][1] );
        }
    }

    FITTools_LS prob_ls( nVertex, nVertex, 2 );

    // Initialize the matrix to zero 
    for( i=0; i<nVertex; i++ ) {
        prob_ls.Rhs_clapack( i, 0 ) = 0.0;
        prob_ls.Rhs_clapack( i, 1 ) = 0.0;
        for( j=0; j<nVertex; j++ ) {
            prob_ls.Desired_clapack( i, j ) = 0.0;
        }
    }

    for( i=0; i<nVertex; i++ ) {
        if( IsBoundary(m_aopVs[i]) ) {
            // Load the matrix for boundary points 
            prob_ls.Desired_clapack( i, i ) = 1.0;
            if( out != NULL ) {
                fprintf(out,"Vertex %d: Boundary (%d,%d) = 1.0\n",i,i,i);
            }

        } else {
            // Load the matrix for interior points 
            aov = m_pMesh.VertexVertices(m_aopVs[i]);
            mija_sum = 0.0;
            mijx_sum = 0.0;
            for( j=0; j<aov.num(); j++ ) {
                mija = CotAlphaBeta( m_aopVs[i], aov[j] );
                
                mija_sum += mija;
                d = Distance( m_aopVs[i], aov[j] );
                mijx = 0.0;
                if( d > 0.0 ) {
                    mijx = CotGammaDelta( m_aopVs[i], aov[j] )/(d*d);
                    mijx_sum += mijx;
                }
                jj = m_aopVs.index(aov[j]);
                prob_ls.Desired_clapack(i,jj) += m_lambda*mija + m_mu*mijx;
                if( out != NULL ) {
                    fprintf(out,"Vertex %d-%d: Interior (%d,%d) d = %lf\n",
                                i, j, i, jj, d);
                }
            }
            prob_ls.Desired_clapack(i,i) += -(m_lambda*mija_sum + m_mu*mijx_sum);
        }
    }

    // Load the constraint values 
    for( i=0; i<nBoundary; i++ ) {
        jj = IndexIds.index(Order[i]);
        prob_ls.Rhs_clapack( jj, 0 ) = BoundaryUV[i][0];
        prob_ls.Rhs_clapack( jj, 1 ) = BoundaryUV[i][1];
        if( out != NULL ) {
            fprintf(out,"Boundary Vertex %d: Index = %d (%lf, %lf)\n",
                        i, jj, BoundaryUV[i][0], BoundaryUV[i][1]);
        }
    }   

    // Solve for the UV coordinates 
    if( prob_ls.Solve_clapack() == FALSE ) {
        printf("ERROR - Solving Least Square problem\n" );
        prob_ls.Print_clapack( cout, FALSE );
    }
    
    m_aUVs.need(nVertex);
    for( i=0; i<nVertex; i++ ) {
        u = prob_ls.Solution_clapack( i, 0 );
        v = prob_ls.Solution_clapack( i, 1 );
        m_aUVs[i] = R2Pt(u,v);
        if( out != NULL ) {
            fprintf(out,"Vertex %d: UV=(%lf, %lf)\n", i, u, v );
        }
    }
    m_PtUV = m_aUVs[m_aopVs.index(m_Vertex)];

    //NormalizeUVCoords(2);

    if( out != NULL ) {
        fclose(out);
        out = NULL;
    }

    return 0;
}

int 
MCParam::CalculateUVnatural(const double in_lambda, const double in_mu)
{
    int i, i1, i2, j, jj, kk, ier, nv;
    double d, mija, mijx, u, v, umin, umax, vmin, vmax;
    int thetaId, alphaId, betaId;
    double cotTheta, t, cotAlpha, a, cotBeta, b;

    Array<int> Order;
    Array<int> IndexIds;
    Array<R2Pt> UV;
    PMeshLite ring;
    Array< Vertex > aov;
    Array< Face > aof;
    R3Pt P;

	/*
	int verbose = 0;

    FILE *out = NULL;
	if( verbose ) {
		out = fopen(WINDataHome("natural_new.out"),"w");
		printf("Writing file %s\n", WINDataHome("natural_new.out"));
	}
	*/

    nv = m_aopVs.num();
	if( nv < 4 ) return -1;  // Not a valid subset

    IndexIds.need(0);
    for( i=0; i<nv; i++ ) { IndexIds += m_aopVs[i]->IndexId(); }

    FITTools_LS prob_ls( 2*nv+4, 2*nv, 1 );

	/*
    if( out != NULL ) {
        fprintf(out,"Natural: %d points, lambda = %lf, mu = %lf\n", 
                    nv, in_lambda, in_mu);
    }
	if( out != NULL && verbose == 2 ) {
		for( i=0; i<nv; i++ ) {
			fprintf(out,"Pt %d: Index %d\n", i, IndexIds[i]);
		}
	}
	*/

    for( i=0; i<nv; i++ ) {
        /*
        P = m_aopVs[i]->GetLoc();
		if( out != NULL && verbose == 1 ) {
            fprintf(out,"%d(%d): (%lf, %lf, %lf) %d\n", i, m_aopVs[i]->IndexId(), 
                        P[0], P[1], P[2], IsBoundary(m_aopVs[i]) );
        }
		*/

        if( IsBoundary(m_aopVs[i]) ) {
            // Load the matrix for boundary points 
            aof = VertexFaces(m_aopVs[i]);
			/*
			if( out != NULL && verbose == 2 ) {
				for( j=0; j<aof.num(); j++ ) {
					aov = m_pMesh.FaceVertices(aof[j]);
					fprintf(out,"Face %d (%d): %d, %d, %d\n", 
						    j, aof[j]->IndexId(), aov[0]->IndexId(), 
							aov[1]->IndexId(), aov[2]->IndexId());
				}
			}
			*/
            for( j=0; j<aof.num(); j++ ) {
                ier = CotAngles(aof[j],m_aopVs[i],
                                thetaId,cotTheta,t,
                                alphaId,cotAlpha,a,
                                betaId,cotBeta,b);
				if( ier != 0 ) return ier;

                jj = IndexIds.index(alphaId);
                kk = IndexIds.index(betaId);
				/*
				if( out != NULL && verbose == 2 ) {
					fprintf(out,"Pt %d, Face %d, Index %d",
						        i, j, m_aopVs[i]->IndexId());
					fprintf(out,", thetaId %d, alphaId %d, betaId %d\n",
						        thetaId, alphaId, betaId); 
				}
				*/

				if( jj < 0 || kk < 0 ) {
					/*
					if( out != NULL ) fclose(out);
					*/
					ier = -2;
					return ier;
				}

                // U component 
                prob_ls.Desired_clapack( i, i ) += -(cotAlpha+cotBeta);
                prob_ls.Desired_clapack( i, jj ) += cotBeta;
                prob_ls.Desired_clapack( i, kk ) += cotAlpha;
                prob_ls.Desired_clapack( i, jj+nv ) += 1.0;
                prob_ls.Desired_clapack( i, kk+nv ) += -1.0;
                prob_ls.Rhs_clapack(i,0) = 0.0;

                // V component 
                prob_ls.Desired_clapack( i+nv, i+nv ) += -(cotAlpha+cotBeta);
                prob_ls.Desired_clapack( i+nv, jj+nv ) += cotBeta;
                prob_ls.Desired_clapack( i+nv, kk+nv ) += cotAlpha;
                prob_ls.Desired_clapack( i+nv, jj ) += -1.0;
                prob_ls.Desired_clapack( i+nv, kk ) += 1.0;
                prob_ls.Rhs_clapack(i+nv,0) = 0.0;

				/*
                if( out != NULL && verbose == 1 ) {
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i, i, -(cotAlpha+cotBeta) );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i, jj, cotBeta );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i, kk, cotAlpha );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i, jj+nv, 1.0 );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i, kk+nv, -1.0 );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i+nv, i+nv, -(cotAlpha+cotBeta) );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i+nv, jj+nv, cotBeta );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i+nv, kk+nv, cotAlpha );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i+nv, jj, -1.0 );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i+nv, kk, 1.0 );
                }
				*/
            }

        } else {
            // Load the matrix for interior points 
            aov = m_pMesh.VertexVertices(m_aopVs[i]);;
            for( j=0; j<aov.num(); j++ ) {
                mija = CotAlphaBeta( m_aopVs[i], aov[j] );
                
                d = Distance( m_aopVs[i], aov[j] );
                mijx = 0.0;
                if( d > 0.0 ) {
                    mijx = CotGammaDelta( m_aopVs[i], aov[j] )/(d*d);
                }
                //jj = aov[j]->IndexId();
                jj = m_aopVs.index(aov[j]);
				if( jj < 0 ) {
					ier = -3;
					return ier;
				}

                // U components 
                prob_ls.Desired_clapack(i,jj) += m_lambda*mija + m_mu*mijx;
                prob_ls.Desired_clapack(i,i) += -(m_lambda*mija + m_mu*mijx);
                prob_ls.Rhs_clapack(i,0) = 0.0;
                // V components 
                prob_ls.Desired_clapack(i+nv,jj+nv) += m_lambda*mija + m_mu*mijx;
                prob_ls.Desired_clapack(i+nv,i+nv) += -(m_lambda*mija + m_mu*mijx);
                prob_ls.Rhs_clapack(i+nv,0) = 0.0;
				/*
                if( out != NULL && verbose == 1 ) {
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i, jj, m_lambda*mija + m_mu*mijx );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i, i, -(m_lambda*mija + m_mu*mijx) );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i+nv, jj+nv, m_lambda*mija + m_mu*mijx );
                    fprintf(out,"...Adding to (%d,%d) %lf\n",
                                i+nv, i+nv, -(m_lambda*mija + m_mu*mijx) );
                }
				*/
            }
        }
    }

    // Find the extreme points 
    ExtremePoints(i1, i2);

	/*
    if( out != NULL && verbose == 1 ) {
        fprintf(out,"Extreme points: %d, %d\n", i1, i2);
    }
	*/

    prob_ls.Desired_clapack(2*nv,i1) += 1.0;
    prob_ls.Rhs_clapack(2*nv,0) += 0.0;
    prob_ls.Desired_clapack(2*nv+1,i1+nv) += 1.0;
    prob_ls.Rhs_clapack(2*nv+1,0) += 0.0;
    prob_ls.Desired_clapack(2*nv+2,i2) += 1.0;
    prob_ls.Rhs_clapack(2*nv+2,0) += 0.0;
    prob_ls.Desired_clapack(2*nv+3,i2+nv) += 1.0;
    prob_ls.Rhs_clapack(2*nv+3,0) += 1.0;
	/*
    if( out != NULL && verbose == 1 ) {
        fprintf(out,"...Adding to (%d,%d) %lf\n",
                    2*nv, i1, 1.0 );
        fprintf(out,"...Adding to (%d,%d) %lf\n",
                    2*nv+1, i1+nv, 1.0 );
        fprintf(out,"...Adding to (%d,%d) %lf\n",
                    2*nv+2, i2, 1.0 );
        fprintf(out,"...Adding to (%d,%d) %lf\n",
                    2*nv+3, 0, 1.0 );
    }
	*/

	/*
    // Print the matrix 
    if( out != NULL && verbose == 1 ) {
        fprintf(out,"  Var: ");
        for( j=0; j<nv; j++ ) fprintf(out,"%5du ",j);
        for( j=0; j<nv; j++ ) fprintf(out,"%5dv ",j);
        fprintf(out,"\n");

        for( j=0; j<2*nv; j++ ) {
            if( j<nv ) {
                fprintf(out,"%5du ",j);
            } else {
                fprintf(out,"%5dv ",j-nv);
            }
            for( i=0; i<2*nv; i++ ) {
                fprintf(out,"%6.2lf ",prob_ls.Desired_clapack(j,i));
            }
            fprintf(out,"= %6.2lf\n", prob_ls.Rhs_clapack(j,0));
        }
    }
	*/

    // Solve for the UV coordinates 
    if( prob_ls.Solve_clapack() == FALSE ) {
        printf("ERROR - Solving Least Square problem\n" );
        prob_ls.Print_clapack( cout, FALSE );
    }
    
	/*
    // Print the solution 
    if( out != NULL && verbose == 1 ) {
        fprintf(out,"    U: ");
        for( i=0; i<nv; i++ ) {
            fprintf(out,"%6.2lf ",prob_ls.Solution_clapack( i, 0 ));
        }
        fprintf(out,"\n");
        fprintf(out,"    V: ");
        for( i=0; i<nv; i++ ) {
            fprintf(out,"%6.2lf ",prob_ls.Solution_clapack( i+nv, 0 ));
        }
        fprintf(out,"\n");
    }
	*/

    // Compute u,v coordinates 
    m_aUVs.need(nv);
    for( i=0; i<nv; i++ ) {
        u = prob_ls.Solution_clapack( i, 0 );
        v = prob_ls.Solution_clapack( i+nv, 0 );
 	/*
       if( out != NULL && verbose == 1 ) {
            fprintf(out,"%d: (%lf, %lf)\n", i, u, v );
        }
	*/

        if(i==0) {
            umin = u;
            umax = u;
            vmin = v;
            vmax = v;
        } else {
            umin = u<umin ? u : umin;
            umax = u>umax ? u : umax;
            vmin = v<vmin ? v : vmin;
            vmax = v>vmax ? v : vmax;
        }
        m_aUVs[i] = R2Pt(u,v);
    }
    m_UVmin = R2Pt(umin,vmin);
    m_UVmax = R2Pt(umax,vmax);
    m_PtUV = m_aUVs[m_aopVs.index(m_Vertex)];

    NormalizeUVCoords(1);
    //NormalizeUVCoords(2);
	/*
    if( out != NULL && verbose == 1 ) {
		for( i=0; i<m_aUVs.num(); i++ ) {
			fprintf(out,"%d(%d): (%lf, %lf)\n", 
				        i, IndexIds[i], m_aUVs[i][0], m_aUVs[i][1] );
		}
    }

    if( out != NULL ) {
        fclose(out);
        out = NULL;
    }
	*/

    return 0;
}

void MCParam::SetUVRange()
{
    int i, nv;
    double u, v, umin, umax, vmin, vmax;

    // Get the range of the UV coordinates 
    nv = m_aUVs.num();
    umin = m_aUVs[0][0];
    umax = m_aUVs[0][0];
    vmin = m_aUVs[0][1];
    vmax = m_aUVs[0][1];
    for(i=1; i<nv; i++) {
        u = m_aUVs[i][0];
        v = m_aUVs[i][1];
        umin = (u<umin) ? u : umin;
        umax = (u>umax) ? u : umax;
        vmin = (v<vmin) ? v : vmin;
        vmax = (v>vmax) ? v : vmax;
    }
    m_UVmin = R2Pt(umin,vmin);
    m_UVmax = R2Pt(umax,vmax);
}

void MCParam::NormalizeUVCoords(const int nDir)
{
    int i, nv;
    double uscale, vscale, scale, u, v;
    R2Vec Range;

    // Get the range of the UV coordinates 
    SetUVRange();

    Range = m_UVmin - m_UVmax;
    
    uscale = (Range[0]>0.0) ? 1.0/Range[0] : 1.0;
    vscale = (Range[1]>0.0) ? 1.0/Range[1] : 1.0;
    if( nDir == 1 ) {
        scale = (uscale>vscale) ? vscale : uscale;
        uscale = scale;
        vscale = scale;
    }

    nv = m_aUVs.num();
    for(i=0; i<nv; i++) {
        u = (m_aUVs[i][0]-m_UVmin[0])*uscale;
        v = (m_aUVs[i][1]-m_UVmin[1])*vscale;
        m_aUVs[i] = R2Pt(u,v);
    }

    m_PtUV = R2Pt((m_PtUV[0]-m_UVmin[0])*uscale,
                  (m_PtUV[1]-m_UVmin[1])*vscale);
    m_UVmax = R2Pt((m_UVmax[0]-m_UVmin[0])*uscale,
                   (m_UVmax[1]-m_UVmin[1])*vscale);
    m_UVmin = R2Pt(0.0,0.0);
}

void MCParam::ScaleUVCoords(const double scale)
{
    int i, nv;
    double u, v;

    nv = m_aUVs.num();
    for(i=0; i<nv; i++) {
        u = m_aUVs[i][0]*scale;
        v = m_aUVs[i][1]*scale;
        m_aUVs[i] = R2Pt(u,v);
    }

    m_PtUV  = R2Pt(m_PtUV[0]*scale, m_PtUV[1]*scale);
    m_UVmax = R2Pt(m_UVmax[0]*scale, m_UVmax[1]*scale);
    m_UVmin = R2Pt(m_UVmin[0]*scale, m_UVmin[1]*scale);
}

void MCParam::MakeMesh(PMeshLite &out_mesh) const
{
    int id, id2, i, j;
    R3Pt p;
    Array<int> VertexId;
    Array<Vertex> aopVs;
    Array<Vertex> aov;
    Array<Vertex> aovn;
    Array<Face> aopFs;

    FILE *out = NULL;
    //out = fopen(WINDataHome("test.out"),"a");

    out_mesh.Clear();

    // Make the vertices 
    for( i=0; i<m_aUVs.num(); i++ ) {
        aopVs += out_mesh.AddVertex( R3Pt(m_aUVs[i][0],m_aUVs[i][1],0.0) );
        VertexId += m_aopVs[i]->IndexId();
        if( out != NULL ) {
            fprintf(out,"Mesh Vertex %d, IndexId %d\n",
                        i, m_aopVs[i]->IndexId() );
        }
    }

    // Copy the faces 
    for( i=0; i<m_aopFs.num(); i++ ) {
        aov = m_pMesh.FaceVertices( m_aopFs[i] );

        for( j=0; j<aov.num(); j++ ) {
            id = aov[j]->IndexId();
            id2 = VertexId.index(id);
            aovn += aopVs[id2];
            if( out != NULL ) {
                fprintf(out,"Mesh Face %d Vertex %d, Id %d, IndexId %d\n", 
                            i, j, id2, id );
            }
        }
        if( out_mesh.IsOkToAdd( aovn ) == 0 )
            printf("Error adding face\n");
        aopFs += out_mesh.AddFace( aovn );
        aovn.need(0);
    }
    if( out != NULL ) {
        fclose(out);
        out = NULL;
    }
}

void MCParam::dump(FILE *out) const
{
    int i;
    R3Pt pt;
    R2Pt uv;
    WINbool hasUVs = FALSE;

    uv = R2Pt(0.0,0.0);
    hasUVs = ( m_aUVs.num() >= m_aopVs.num() ) ? TRUE : FALSE;
    if( out != NULL ) { 
        fprintf(out,"ParamType: %d\n",m_ParamType);
        fprintf(out,"  lambda = %lf, mu = %lf\n", m_lambda, m_mu);
        for( i=0; i<m_aopVs.num(); i++ ) {
            pt = m_aopVs[i]->GetLoc();
            if( hasUVs ) uv = m_aUVs[i];
            fprintf(out," Point %d: x= %lf, y= %lf, z= %lf, u= %lf, v= %lf\n",
                        i, pt[0], pt[1], pt[2], uv[0], uv[1] );
        }
    }
}
