/* Copyright 1996, Microsoft.  All Rights Reserved. */


/* -------------------------------------------------------------------------
 *
 *               Solve Ax = b
 *
 * ------------------------------------------------------------------------- */
#include "StdAfx.H"

#include <fitting/FITTools_Least_squares.H>
#include <fitting/FITTools_SVD.H>
#include <ctime>
#include <mkl/mkl_solver.h>
#include <mkl/mkl_lapack.h>

/* -----------------------  Private Methods  ------------------------------- */

/* ----------------------- Protected Methods ------------------------------- */


/* -----------------------  Public Methods   ------------------------------- */

// solves overdetermined or underdetermined real linear systems
//  involving an M-by-N matrix A, or its transpose, using a QR or LQ
//  factorization of A.  It is assumed that A has full rank.
//   NOTE: X is column-row ordering (Fortran style)
static long int CallDgels( int in_lNRows,
                           int in_lNCols,
                           Array<double>  &in_adA,  // A input in column row format
                           int in_lNColsB,
                           Array<double>  &in_adB,  // rhs B input in column row format
                           Array<double>  &out_adX  // return the result (each column is a solution)
              )
{
    
    int lWork = WINmin( in_lNRows, in_lNCols ) + WINmax( WINmax((int)1, in_lNRows), 
                                                         WINmax(in_lNCols, in_lNColsB) );

    Array<double> adWork( lWork );
    char cTrans = 'N';
    
    int lInfo, lLdb = WINmax( in_lNRows, in_lNCols);

    double *adB = NULL;
    // B is too small, copy to X before doing dgels
    if ( in_lNRows < in_lNCols ) {
        ASSERT( out_adX.num() >= in_lNColsB * in_lNCols );
        ASSERT( in_adB.num() >= in_lNColsB * in_lNRows );

        for ( int i = 0; i < in_lNColsB; i++ ) {
            for ( int j = 0; j < in_lNRows; j++ ) {
                out_adX[ (int) (i * in_lNCols) + j ] = in_adB[ (int) (i * in_lNRows) + j ];
            }
            memcpy( &out_adX[ (int) (i * in_lNCols) ], 
                    &in_adB[ (int) (i * in_lNRows) ], 
                    (int) in_lNRows * sizeof(double) );
        }
        adB = &out_adX[0];
    } else {
        adB = &in_adB[0];
    }

    dgels( &cTrans, 
           &in_lNRows, &in_lNCols, &in_lNColsB, 
           in_adA, &in_lNRows, 
           adB, &lLdb, 
           adWork, &lWork, &lInfo );

    // Result was in B. Copy to X.
    if ( in_lNRows >= in_lNCols ) {
        for ( int i = 0; i < in_lNColsB; i++ ) {
            memcpy( &out_adX[(int) (i * in_lNCols)], 
                    &in_adB[(int) (i * in_lNRows)], 
                    in_lNCols * sizeof(double) );
        }
    }

    return lInfo;
}

/* -------------------------------------------------------------------------
 * DESCR   :	Pick the appropriate thing to do
 * ------------------------------------------------------------------------- */
WINbool
FITTools_LS::Solve_utils(  )
{
   double *u = new double [m_iNDesired * m_iNVariables];
   double *vt = new double [WINmin(m_iNDesired, m_iNVariables) * m_iNVariables];
   double *sigma = new double [WINmin(m_iNDesired, m_iNVariables)];
   
   double dTol = 0;
   WINbool bRes = TRUE;
   for (int i = 0; i < m_iNDimension; i++)  {
      
       WINbool bResSvd = 
           FITToolsSVD::SVD(m_adDesiredData, m_iNDesired, m_iNVariables,
                         u, sigma, vt);
       FITToolsSVD::Backsubst(m_iNDesired, m_iNVariables,
                           u, sigma, vt,
                           &m_adRhsData[i * m_iNDesired],
                           &m_adSolutionData[i * m_iNVariables], 
                           dTol);
       
       if ( bResSvd == FALSE )
           bRes = FALSE;

   }
   delete [] u;
   delete [] vt;
   delete [] sigma;
   
   return bRes;
}

WINbool
FITTools_LS::Solve_clapack()
{
    int iRes = 0;
    WINbool bRes = TRUE;
    iRes = CallDgels( m_iNDesired, m_iNVariables,
                     m_adDesiredData, m_iNDimension,
                     m_adRhsData,
                     m_adSolutionData);

    if ( iRes != 0 ) {
       bRes = FALSE;
       cerr << m_iNDesired << " " << m_iNVariables << " " << iRes << "\n";
    }

    for (int i = 0; i < m_iNDimension; i++) {
        for (int c = 0; c < m_iNVariables; c++) {
            //TRACE("%f %f\n", lsCopy.Solution_utils(c, i), Solution_clapack(c, i));
            if ( ! ( Solution_clapack(c, i) > -1e20 && Solution_clapack(c, i) < 1e20 ) )
                bRes = FALSE;
        }
    }

    return bRes;
}

extern "C" int omp_get_max_threads();

WINbool
FITTools_LS::Solve_sparse(const WINbool in_bPrintProgress)
{
    static WINbool s_bInit = FALSE;
    static void *s_aPt[64];
    static int s_aiParams[64];
    int iMaxFct = 1, iMNum = 1, iMType = 11, iPhase = 11, iMsgLvl = 1, iErr = 0, iDummy = 0;
    double dDummy = 0.0;

    if ( s_bInit == FALSE ) {
        for ( int i = 0; i < 64; i++ ) {
            s_aPt[i] = 0;
            s_aiParams[i] = 0;
        }
        s_bInit = TRUE;
        s_aiParams[0] = 0;
        s_aiParams[2] = omp_get_max_threads();
    }

    const clock_t start_t = clock( );
    // pardiso(pt, &maxfct, &mnum, &mtype, &phase, 
    //         &n, a, ia, ja, perm, &nrhs,
    //         iparm, &msglvl, b, x, &error);
    iPhase = 11;
    pardiso_( s_aPt, &iMaxFct, &iMNum,  &iMType, &iPhase,
              &m_iNDesired, &m_adDesiredData[0], &m_aiRowIndex[0], &m_aiColIndex[0], &iDummy, &m_iNDimension,
              s_aiParams, &iMsgLvl, &dDummy, &dDummy, &iErr );

    if ( iErr != 0 ) {
        TRACE("Error in Symbolic Factorization %d %d\n", iErr, s_aiParams[20]);
        return FALSE;
    }
    if ( in_bPrintProgress ) {
        TRACE("Reordering complete\n");
        TRACE("Number of non-zero factors %d\n", s_aiParams[17] );
        TRACE("Number of factorizations %d\n", s_aiParams[18] );
    }

    iPhase = 22;
    pardiso_( s_aPt, &iMaxFct, &iMNum,  &iMType, &iPhase,
              &m_iNDesired, &m_adDesiredData[0], &m_aiRowIndex[0], &m_aiColIndex[0], &iDummy, &m_iNDimension,
              s_aiParams, &iMsgLvl, &dDummy, &dDummy, &iErr );

    if ( iErr != 0 ) {
        TRACE("Error in Numerical Factorization %d\n", iErr);
        return FALSE;
    }

    if ( in_bPrintProgress ) {
        TRACE("Factorization complete\n");
    }

    iPhase = 33;
    pardiso_( s_aPt, &iMaxFct, &iMNum,  &iMType, &iPhase,
              &m_iNDesired, &m_adDesiredData[0], &m_aiRowIndex[0], &m_aiColIndex[0], NULL, &m_iNDimension,
              s_aiParams, &iMsgLvl, &m_adRhsData[0], &m_adSolutionData[0], &iErr );

    const clock_t end_t = clock( );
    const double dDiff = end_t - start_t;

    if ( in_bPrintProgress ) {
        if ( iErr == 0 ) {
            TRACE("Time %f Iterations %d\n", dDiff, s_aiParams[20]);
        }
    }

    /*
    iPhase = -1;
    int iErrMem = 0;
    pardiso_( s_aPt, &iMaxFct, &iMNum,  &iMType, &iPhase,
              &m_iNDesired, &m_adDesiredData[0], &m_aiRowIndex[0], &m_aiColIndex[0], &iDummy, &m_iNDimension,
              s_aiParams, &iMsgLvl, &dDummy, &dDummy, &iErrMem );

    if ( iErrMem != 0 )
        TRACE("Problems with memory release\n");
              */

    if ( iErr == 0 ) {
        return TRUE;
    }

    TRACE("Failed: time %f iterations %d  Error", dDiff, iErr );
    switch ( -iErr ) {
    case 1 : TRACE("Large fluctuations in residual\n"); break;
    case 2 : TRACE("Slow convergence\n"); break;
    case 3 : TRACE("Stopping criteria not reached\n"); break;
    case 4 : TRACE("Perturbed pivots caused iterative refinement\n"); break;
    case 5 : TRACE("Try again with iparam[3] = 0\n"); break;
    }

    return FALSE;
}

/* -------------------------------------------------------------------------
 * DESCR   :	Create the matrices that will Ae needed
 * ------------------------------------------------------------------------- */
FITTools_LS::FITTools_LS( int in_iNDesired, // number of rows/constraints
                          int in_iNVariables, // number of variables
                          int in_iNDimension, //  number of cols in solution
                          const WINbool in_bSparse ) 
{
    m_bSparse = in_bSparse;
    Resize( in_iNDesired, in_iNVariables, in_iNDimension, in_bSparse );
}

void FITTools_LS::Clear()
{
    m_adDesiredData.memfill(0);
    m_adSolutionData.memfill(0);
    m_adRhsData.memfill(0);
    m_aiRowIndex.memfill(0);

    if ( m_bSparse == TRUE ) {
        m_adDesiredData.need( 20 );
        m_aiColIndex.need(0);
        m_aiRowIndex[0] = 1;
    }
}

void FITTools_LS::Resize( const int in_iNDesired, const int in_iNVariables, const int in_iDimension, const WINbool in_bSparse )
{
    ASSERT( in_iNDesired > 0 );
    ASSERT( in_iNVariables > 0 );
    ASSERT( in_iDimension > 0 );

    m_iNDesired = in_iNDesired;
    m_iNVariables = in_iNVariables;
    m_iNDimension = in_iDimension;

    if ( in_bSparse == FALSE ) {
        m_adDesiredData.need( in_iNDesired * m_iNVariables );
        m_adSolutionData.need( in_iNVariables * in_iDimension );
        m_adRhsData.need( in_iNDesired * in_iDimension );
    } else {
        // Add extra data at the end
        m_adDesiredData.need( 20 );
        m_adSolutionData.need( in_iNVariables * in_iDimension );
        m_adRhsData.need( in_iNDesired * in_iDimension );
        m_aiRowIndex.need( in_iNDesired + 1 );
    }

    Clear();
}

/* -------------------------------------------------------------------------
 * DESCR   :	Zap everything
 * ------------------------------------------------------------------------- */
FITTools_LS::~FITTools_LS()
{
}

/*
 * Standard qsort comparision function 
 * Points store col, value
 */
static int sortOrder( const void *in_op1, const void * in_op2 )
{
    const pair<int,double> *pt1 = static_cast< const pair<int,double> * >( in_op1 );
    const pair<int,double> *pt2 = static_cast< const pair<int,double> * >( in_op2 );

    if ( pt1->first < pt2->first )
        return -1;
    else if ( pt1->first > pt2->first )
        return 1;

    ASSERT(FALSE);
    return 0;
}
/// A row of the matrix (sorts before adding)
void FITTools_LS::Desired_sparse   (const int in_iRow, Array< pair<int,double> > &io_apairColValue )
{
    qsort( &io_apairColValue[0], io_apairColValue.num(), sizeof( pair<int,double> ), sortOrder );

    const int iOldSize = m_aiColIndex.num();

    ASSERT( m_aiRowIndex[ in_iRow ] == m_aiColIndex.num() + 1 );

    m_adDesiredData.add( io_apairColValue.num() );
    m_aiColIndex.add( io_apairColValue.num() );
    m_aiRowIndex[ in_iRow+1 ] = m_aiColIndex.num() + 1;

    for ( int i = 0; i < io_apairColValue.num(); i++ ) {
        m_adDesiredData[ iOldSize + i ] = io_apairColValue[i].second;
        m_aiColIndex[ iOldSize + i ] = io_apairColValue[i].first + 1;
    }
}


void FITTools_LS::Print_utils(ostream &out, WINbool in_bDoTrace )
{
    Array<double> adSum( N_variables() );
    Array<double> adSumAbs( N_variables() );

    int iV = 0;
    for (iV = 0; iV < N_variables(); iV++) {
        adSum[iV] = 0;
        adSumAbs[iV] = 0;
    }

    if ( in_bDoTrace )
        TRACE("desired %d variables %d\n", N_desired(), N_variables());
    out << "Desired " << N_desired() << " " << N_variables() << "\n";
    for (int iD = 0; iD < N_desired(); iD++) {
        double dSum = 0;
        for (int iV = 0; iV < N_variables(); iV++) {
            dSum += Desired_utils(iD, iV);
            adSum[iV] += Desired_utils(iD, iV);
            adSumAbs[iV] += fabs( Desired_utils(iD, iV) );
            if (!RNIsZero( Desired_utils(iD, iV) ) ) {
                out << iV << " " << Desired_utils(iD, iV) << "  ";
                if ( in_bDoTrace )
                    TRACE("%d %f  ", iV, Desired_utils(iD, iV) );
            }
        }

        if ( in_bDoTrace )
            TRACE(" = ");
        out << " = ";
        for (int iS = 0; iS < N_dimension(); iS++) {
            if ( in_bDoTrace )
                TRACE("%f ", Rhs_utils(iD, iS));
            out << Rhs_utils(iD, iS) << " ";
        }
        out << "  Sum " << dSum << "\n";
        if ( in_bDoTrace )
            TRACE("  Sum %f\n", dSum);
    }
    for ( iV = 0; iV < N_variables(); iV++) {
        if ( in_bDoTrace )
            TRACE("%f ", adSum[iV]);
        out << adSum[iV] << " ";
    }
    out << "\n";
    for ( iV = 0; iV < N_variables(); iV++) {
        if ( in_bDoTrace )
            TRACE("%f ", adSumAbs[iV]);
        out << adSumAbs[iV] << " ";
    }
    out << "\n\n";
    if ( in_bDoTrace )
        TRACE("\n\n");
}

void FITTools_LS::Print_clapack(ostream &out, WINbool in_bDoTrace )
{
    Array<double> adSum( N_variables() );
    Array<double> adSumAbs( N_variables() );

    for (int iV = 0; iV < N_variables(); iV++) {
        adSum[iV] = 0;
        adSumAbs[iV] = 0;
    }

    if ( in_bDoTrace )
        TRACE("desired %d variables %d\n", N_desired(), N_variables());
    out << "Desired " << N_desired() << " " << N_variables() << "\n";
    for (int iD = 0; iD < N_desired(); iD++) {
        double dSum = 0;
        for (int iV = 0; iV < N_variables(); iV++) {
            dSum += Desired_clapack(iD, iV);
            adSum[iV] += Desired_clapack(iD, iV);
            adSumAbs[iV] += fabs( Desired_clapack(iD, iV) );
            if (!RNIsZero( Desired_clapack(iD, iV) ) ) {
                out << iV << " " << Desired_clapack(iD, iV) << "  ";
                if ( in_bDoTrace )
                    TRACE("%d %f  ", iV, Desired_clapack(iD, iV) );
            }
        }

        if ( in_bDoTrace )
            TRACE(" = ");
        out << " = ";
        for (int iS = 0; iS < N_dimension(); iS++) {
            if ( in_bDoTrace )
                TRACE("%f ", Rhs_clapack(iD, iS));
            out << Rhs_clapack(iD, iS) << " ";
        }
        out << "  Sum " << dSum << "\n";
        if ( in_bDoTrace )
            TRACE("  Sum %f\n", dSum);
    }
    if ( in_bDoTrace )
        TRACE("\n\n");
    for ( FORINT iV = 0; iV < N_variables(); iV++) {
        if ( in_bDoTrace )
            TRACE("%f ", adSum[iV]);
        out << adSum[iV] << " ";
    }
    out << "\n";
    for ( FORINT iV = 0; iV < N_variables(); iV++) {
        if ( in_bDoTrace )
            TRACE("%f ", adSumAbs[iV]);
        out << adSumAbs[iV] << " ";
    }
    out << "\n\n";
    if ( in_bDoTrace )
        TRACE("\n\n");
}

void   FITTools_LS::Print_sparse( ostream &out, WINbool in_bDoTrace  )
{
    for ( int iR = 0; iR < m_iNDesired; iR++ ) {
        out << iR << ": ";
        if ( in_bDoTrace ) TRACE("%d: ", iR);

        for ( int iC = m_aiRowIndex[iR]; iC < m_aiRowIndex[iR+1]; iC++ ) {
            if ( in_bDoTrace ) {
                TRACE("%d %0.5f", m_aiColIndex[iC-1] - 1, m_adDesiredData[iC-1]);
            }
            out << m_aiColIndex[iC-1] - 1 << " " << m_adDesiredData[iC-1] << " ";
        }
        if ( in_bDoTrace ) TRACE("  = ");
        out << "  = ";

        for ( int k = 0; k < m_iNDimension; k++ ) {
            if ( in_bDoTrace ) TRACE("%f ", Rhs_sparse( iR, k ) );
            out << Rhs_sparse(iR, k) << " ";
        }
        if ( in_bDoTrace ) TRACE("\n");
        out << "\n";
    }
}

void   FITTools_LS::Check_clapack( const double dMinA, const double dMaxA, const double dMinR, const double dMaxR )
{
    Array<double> adSum( N_variables() );
    Array<double> adSumAbs( N_variables() );

    for (int iV = 0; iV < N_variables(); iV++) {
        adSum[iV] = 0;
        adSumAbs[iV] = 0;
    }

    for (int iD = 0; iD < N_desired(); iD++) {
        double dSum = 0;
        for (int iV = 0; iV < N_variables(); iV++) {
            dSum += Desired_clapack(iD, iV);
            adSum[iV] += Desired_clapack(iD, iV);
            adSumAbs[iV] += fabs( Desired_clapack(iD, iV) );
            ASSERT ( Desired_clapack(iD, iV) >= dMinA && Desired_clapack(iD, iV ) <= dMaxA ) ;
        }

        for (int iS = 0; iS < N_dimension(); iS++) {
            ASSERT (  Rhs_clapack(iD, iS) >= dMinR && Rhs_clapack(iD, iS) <= dMaxR ) ;
        }
        if ( RNIsZero( dSum ) )
            ASSERT(FALSE);
    }
    for ( FORINT iV = 0; iV < N_variables(); iV++) {
        if ( RNIsZero( adSumAbs[iV]) )
            ASSERT(FALSE);
    }
}

Array<int>   FITTools_LS::Clip_variables_utils( double in_dPerc )
{
    ASSERT( in_dPerc <= 1.0 && in_dPerc > 0.0 );

    Array<int> aiRevMap(N_variables());
    for ( int iV = 0; iV < N_variables(); iV++ ) {
        aiRevMap[iV] = iV;
    }

    int iCount = WINminmax((int) (N_desired() * in_dPerc), 1, N_variables() );
    if ( iCount == N_variables() ) return aiRevMap;

    Array<double> adSum( N_variables() );

    for ( FORINT iV = 0; iV < N_variables(); iV++)
        adSum[iV] = 0;

    for (int iD = 0; iD < N_desired(); iD++) {
        for (int iV = 0; iV < N_variables(); iV++) {
            adSum[iV] += Desired_utils(iD, iV);
        }
    }

    double dAvg = 0;
    for ( FORINT iV = 0; iV < N_variables(); iV++ ) {
        dAvg += adSum[iV];
    }

    dAvg /= N_variables();

    Array<int> aiMap( N_variables() );
    for ( FORINT iV = 0; iV < N_variables(); iV++ ) {
        aiMap[iV] = iV;
    }
    aiMap.sort( adSum );
    aiMap.reverse();


    aiRevMap.need( iCount );
    for ( FORINT iV = 0; iV < iCount; iV++ ) {
        aiRevMap[iV] = aiMap[iV];
    }

    Array<double> adDesired( iCount * N_desired() );

    for ( FORINT iD = 0; iD < N_desired(); iD++) {
        for (int iV = 0; iV < iCount; iV++) {
            adDesired[iD * iCount + iV] = Desired_utils(iD, aiMap[iV]);
        }
    }

    m_adDesiredData = adDesired;
    m_iNVariables = iCount;

    return aiRevMap;
}

void   FITTools_LS::Duplicate_rows_utils( )
{
    Array<WINbool> abDupl(N_desired());

    for (int iD = 0; iD < N_desired(); iD++) {
        abDupl[iD] = FALSE;
    }

    for ( FORINT iD = 0; iD < N_desired(); iD++) {
        if ( abDupl[iD] == TRUE ) continue;

        for (int iD2 = iD + 1; iD2 < N_desired(); iD2++) {
            abDupl[iD2] = TRUE;
            double dScl = 0.0;
            for (int iV = 0; iV < N_variables(); iV++) {
                if (RNIsZero( Desired_utils(iD, iV) ) ) {
                    if ( RNIsZero( Desired_utils(iD2, iV) ) == FALSE ) {
                        abDupl[iD2] = FALSE;
                        break;
                    }
                } else {
                    dScl = Desired_utils(iD2, iV) / Desired_utils(iD, iV);
                    break;
                }
            }
            if ( abDupl[iD2] == FALSE ) continue;
            for ( FORINT iV = 0; iV < N_variables(); iV++) {
                double dVal = dScl * Desired_utils(iD, iV);
                if (RNApproxEqual( dVal, Desired_utils(iD2, iV) ) == FALSE ) {
                    abDupl[iD2] = FALSE;
                    break;
                }
            }
        }
    }
    for ( FORINT  iD = 0; iD < N_desired(); iD++ ) {
        if ( abDupl[iD] == TRUE )
            TRACE("%d ", iD);
    }
    TRACE("\n");
}


void   FITTools_LS::Variable_sums_utils( Array<double> &out_adSums ) 
{
    out_adSums.need( N_variables() );
    for (int iV = 0; iV < N_variables(); iV++) {
        out_adSums[iV] = 0;
    }

    for (int iD = 0; iD < N_desired(); iD++) {
        for (int iV = 0; iV < N_variables(); iV++) {
            out_adSums[iV] += fabs( Desired_utils(iD, iV) );
        }
    }
}

WINbool   FITTools_LS::Solve_avg(  ) 
{
    for (int iV = 0; iV < N_variables(); iV++) {
        Array<double> adVec( N_dimension() );
        for ( int i = 0; i < adVec.num(); i++ )
            adVec[i] = 0;

        double dSum = 0;
        for (int iD = 0; iD < N_desired(); iD++) {
            double dScl = Desired_clapack(iD, iV);
            dSum += dScl;
            for ( int i = 0; i < adVec.num(); i++ )
                adVec[i] += dScl * Rhs_clapack( iD, i );
        }

        if ( RNIsZero( dSum ) ) {
            for ( FORINT  i = 0; i < adVec.num(); i++ )
                m_adSolutionData[ iV * m_iNVariables + i ] = adVec[i] ;
        } else {
            for ( FORINT  i = 0; i < adVec.num(); i++ )
                m_adSolutionData[ iV * m_iNVariables + i ] = adVec[i] / dSum;
        }
    }
    return TRUE;
}


void FITTools_LS::Test()
{
    FITTools_LS oLS(2,2,1);

    oLS.Desired_clapack(0,0) = 1.0;
    oLS.Desired_clapack(0,1) = 0.0;
    oLS.Desired_clapack(1,0) = 0.0;
    oLS.Desired_clapack(1,1) = 1.0;

    oLS.Rhs_clapack(0, 0) = 3.0;
    oLS.Rhs_clapack(1, 0) = 4.0;

    VERIFY( oLS.Solve_clapack() == TRUE );

    cout << "(should be 3.0) " << oLS.Solution_clapack(0,0 ) << "\n";
    cout << "(should be 4.0) " << oLS.Solution_clapack(1,0 ) << "\n\n";

    ASSERT( RNApproxEqual(3.0, oLS.Solution_clapack(0,0 ), 1e-4 ) );
    ASSERT( RNApproxEqual(4.0, oLS.Solution_clapack(1,0 ), 1e-4 ) );

    FITTools_LS oLS2(3,2,1);

    oLS2.Desired_clapack(0,0) = 1.0;
    oLS2.Desired_clapack(0,1) = 0.0;
    oLS2.Desired_clapack(1,0) = 0.0;
    oLS2.Desired_clapack(1,1) = 1.0;
    oLS2.Desired_clapack(2,0) = 0.8;
    oLS2.Desired_clapack(2,1) = 0.2;

    oLS2.Rhs_clapack(0, 0) = 3.0;
    oLS2.Rhs_clapack(1, 0) = 4.0;
    oLS2.Rhs_clapack(2, 0) = 2.0;

    VERIFY( oLS2.Solve_clapack() == TRUE );

    cout << "(should be 2.42857) " << oLS2.Solution_clapack(0,0 ) << "\n";
    cout << "(should be 3.85714) " << oLS2.Solution_clapack(1,0 ) << "\n\n";

    ASSERT( RNApproxEqual(2.42857, oLS2.Solution_clapack(0,0 ), 1e-4 ) );
    ASSERT( RNApproxEqual(3.85714, oLS2.Solution_clapack(1,0 ), 1e-4 ) );

    FITTools_LS oLS3(3,2,2);

    oLS3.Desired_clapack(0,0) = 1.0;
    oLS3.Desired_clapack(0,1) = 0.0;
    oLS3.Desired_clapack(1,0) = 0.0;
    oLS3.Desired_clapack(1,1) = 1.0;
    oLS3.Desired_clapack(2,0) = 0.8;
    oLS3.Desired_clapack(2,1) = 0.2;

    oLS3.Rhs_clapack(0, 0) = 6.0;
    oLS3.Rhs_clapack(1, 0) = 8.0;
    oLS3.Rhs_clapack(2, 0) = 4.0;

    oLS3.Rhs_clapack(0, 1) = 3.0;
    oLS3.Rhs_clapack(1, 1) = 4.0;
    oLS3.Rhs_clapack(2, 1) = 2.0;

    VERIFY( oLS3.Solve_clapack() == TRUE );
    
    cout << "(should be 4.85714) " << oLS3.Solution_clapack(0,0 ) << "\n";
    cout << "(should be 7.71429) " << oLS3.Solution_clapack(1,0 ) << "\n";
    cout << "(should be 2.42857) " << oLS3.Solution_clapack(0,1 ) << "\n";
    cout << "(should be 3.85714) " << oLS3.Solution_clapack(1,1 ) << "\n\n";

    ASSERT( RNApproxEqual( 4.85714,  oLS3.Solution_clapack(0,0 ), 1e-4 ) );
    ASSERT( RNApproxEqual( 7.71429,  oLS3.Solution_clapack(1,0 ), 1e-4 ) );
    ASSERT( RNApproxEqual( 2.42857,  oLS3.Solution_clapack(0,1 ), 1e-4 ) );
    ASSERT( RNApproxEqual( 3.85714,  oLS3.Solution_clapack(1,1 ), 1e-4 ) );

    cerr << "Over constrained\n";
    FITTools_LS oLS4(2,3,2);

    oLS4.Desired_clapack(0,0) = 1.0;
    oLS4.Desired_clapack(0,1) = 0.0;
    oLS4.Desired_clapack(0,2) = 0.8;
    oLS4.Desired_clapack(1,0) = 0.0;
    oLS4.Desired_clapack(1,1) = 1.0;
    oLS4.Desired_clapack(1,2) = 0.2;

    oLS4.Rhs_clapack(0, 0) = 3.0;
    oLS4.Rhs_clapack(1, 0) = 4.0;
    oLS4.Rhs_clapack(0, 1) = 6.0;
    oLS4.Rhs_clapack(1, 1) = 8.0;

    VERIFY( oLS4.Solve_clapack() == TRUE );

    cout << "(should be 1.47619) " << oLS4.Solution_clapack(0,0 ) << "\n";
    cout << "(should be 3.61905) " << oLS4.Solution_clapack(1,0 ) << "\n";
    cout << "(should be 1.90476) " << oLS4.Solution_clapack(2,0 ) << "\n";
    cout << "(should be 2.95238) " << oLS4.Solution_clapack(0,1 ) << "\n";
    cout << "(should be 7.2381) " << oLS4.Solution_clapack(1,1 ) << "\n";
    cout << "(should be 3.80952) " << oLS4.Solution_clapack(2,1 ) << "\n";

    ASSERT( RNApproxEqual( 1.47619,  oLS4.Solution_clapack(0,0 ), 1e-4 ) );
    ASSERT( RNApproxEqual( 3.61905,  oLS4.Solution_clapack(1,0 ), 1e-4 ) );
    ASSERT( RNApproxEqual( 1.90476,  oLS4.Solution_clapack(2,0 ), 1e-4 ) );
    ASSERT( RNApproxEqual( 2.95238,  oLS4.Solution_clapack(0,1 ), 1e-4 ) );
    ASSERT( RNApproxEqual( 7.2381,  oLS4.Solution_clapack(1,1 ), 1e-4 ) );
    ASSERT( RNApproxEqual( 3.80952,  oLS4.Solution_clapack(2,1 ), 1e-4 ) );
}
