///////////////////////////////////////////////////////////////////////////////
// Chuck and Brian's SVD from \\atr\animation\atila\atilasrc\mathematics
// Revision History Abstract
// Programmer: C. Rose
// 21Jan1997 Move into Math namespace
// 21Jul1997 Fix array bound error in underdetermined case
// /////////////////////////////////////////////////////////////////////////////
#include "StdAfx.H"

#include <fitting/FITTools_SVD.H>

/*************************************************************
** Singular-Value Decomposition.
** 
** Had to write this by hand, since the version in Numerical
** Recipes has a bug, carried over from a bug in Golub's
** original algorithm:

        Gene H. Golub & Charles F. Van Loan
        MATRIX COMPUTATIONS
        Johns Hopkins University Press, 1983
        Fifth printing, 1987

        Page 293
        Algorithm 8.3-2: The SVD Algorithm

        ...Find the largest q and the smallest p such that if

                 +-            -+
                 | A11   0   0  |          p
                 |  0   A22  0  |        n-p-q
           A  =  |  0    0  A33 |          q
                 |  0    0   0  |         m-n
                 +-            -+

        then A33 is diagonal and A22 has a nonzero superdiagonal.
        If q = n then quit.
        If any diagonal entry in A22 is zero, then zero the superdiagonal
        element in the same row and go to Repeat...

The last sentence above is the cause of the trouble, in the following case:

                     +-   -+
                     |0 1 0|
                 A = |0 1 1|
                     |0 0 0|
                     +-   -+

In this case, q is 0, A33 is null, and A22 is the same as A.  The instruction

            "if any diagonal entry in A22 is zero, then
             zero the superdiagonal element in the same row"

cannot be applied -- A22 has a diagonal entry that is zero, but there
is no superdiagonal element in that same row.  The proper thing to
do seems to be to special-case: If A22 has 0 in its lower-right
diagonal element, zero the superdiagonal element above it.


**
** 
**
** Since the Num. Rec. code was cribbed from EISPACK or LINPACK
** or some such, which in turn cribbed from other places, I would
** consider an SVD routine suspect unless it works on the following
** test data:
**		a = 	0 1 0
**      		0 1 1
**      		0 0 0
**
** Anyway, Al Barr & I went through the references to figure out
** how it all worked.  I reimplimented it from scratch, first in
** lisp, and now in C.  Martin L. Livesey used the code and reported
** various bugs, which have been fixed.
**
**
** public routines:
**	num_svd()
**	num_svd_backsubst()
**
**	Ronen Barzel   July 1989
**       (bug fixes: Feb 1990)
**	 (ansi-C, backsubst: May 1993)
**
****************************************************************/

#include <stdio.h>
#include <memory.h>
#include <malloc.h>
#include <math.h>

#define MIN(A,B)	(((A)<(B)) ? (A):(B))
#define MAX(A,B)	(((A)>(B)) ? (A):(B))
#define FREE(p)			free((char*)(p))
static void CLEAR2D(double *a, int m, int n) 
{ 
    (void) memset((char*)(a),0,(m)*(n)*sizeof(double)); 
}

static void COPY2D(double *d, int m, int n, double const *s)	
{ 
    (void) memcpy((char*)(d),(char*)(s),(int)((m)*(n)*sizeof(double))); 
}

static const double &ref2D(const double *a, int m, int n, int i, int j) 
{	
    return a[(n)*(i)+(j)]; 
}

static double &assign2D(double *a, int m, int n, int i, int j) 
{	
    return a[(n)*(i)+(j)]; 
}

#define pythag	fhypot

static double *alloc2D(int m, int n, double *&v)	
{  
    return v = (double *) malloc((unsigned)(m)*(n)*sizeof(double)); 
}

double *ALLOC1D(int n, double *&v) 
{	
    return v = (double *) malloc((unsigned)(n)*sizeof(double)); 
}

static double SQRT(double t) { return (double) sqrt((double) t); }

static double FABS(double t) { return (double) fabs((double) t); }

static double fhypot (double a, double b)
{
    return (double) sqrt ((double) a*a + b*b);
}

/* householder transformations
**
** looks at the submatrix of A below and to the right of the [i,j]th element
** (inclusive) (where i and j go from 0 to n-1).  Performs a householder
** transformation on the submatrix, to zero out all but the first
** elements of the first column.
**
** Matrix u (a rotation matrix) is transformed by the inverse of the householder
** transformation, so that (old u)(old a) = (new u)(new a)
**
** a is m x n,  u is m x m
**
** If the submatrix is (X1, X2, ...) the householder transformation is
** P_X1,  X1 transforms to (|X1|,0,0,0,...), and the resulting matrix
** is ((|X1|,0,0,...),P_X1 X2, P_X1 X3, ...)
*/

void
FITToolsSVD::householder_zero_col(double *a,double *u,int i,int j,int m,int n,double *hv)
	/* hv is a work vector, length m */
{
    double	scale,	/* a scale factor for X1 */
		sigma,		/* the norm^2 of X1 */
		snorm,		/* +/- the norm of X1 */
		f, h, s, dot;
    int k,l;
    
    /* we will scale X1 by its l1-norm.  squawk! */
    for (scale=0, k=i; k<m; k++) scale += FABS(ref2D(a,m,n,k,j));

    /* if X1 is 0, no point doing anything else */
    if (!scale) return;

    /* divide out the l1-norm, and calculate sigma = |X|^2 */
    for (k=i; k<m; k++) hv[k] = ref2D(a,m,n,k,j) / scale;
    for (sigma=0, k=i; k<m; k++) sigma += hv[k]*hv[k];

    /* The householder vector is X1 +/- (|X1|,0,0,...).  We will
    ** contruct this vector in hv and use it to perform the householder
    ** transformation on the matrix.  The plus or minus on the norm
    ** is to reduce roundoff error.  */
    f = hv[i];
    snorm = (f > 0) ? -SQRT(sigma) : SQRT(sigma);
    h = f*snorm - sigma;	/* eqn 11.2.4 in Press et.al. */
    hv[i] = f - snorm;

    /* set the first column of a to be the (|X1|,0,...) -- this is
    ** what we would get by performing the transformation, but this
    ** way's faster */
    assign2D(a,m,n,i,j) = scale * snorm;
    for (k=i+1; k<m; k++) assign2D(a,m,n,k,j) = 0;

    /* Now perform the householder transformation on the rest of
    ** the columns.  If the householder vector is X, and -half its norm^2
    ** is h, the householder transform is P_ij : delta_ij + x_i x_j / h.
    ** We don't actually create the P matrix, we just do the calcuation.
    */
    for (k=j+1; k<n; k++) {
	for (dot=0, l=i; l<m; l++) dot += ref2D(a,m,n,l,k)*hv[l];
	s = dot/h;
	for (l=i; l<m; l++) assign2D(a,m,n,l,k) += s*hv[l];
    }

    /* Similarly, perform the householder transformation on (all)
    ** the rows of u.  Note that it's the rows of u rather than
    ** the columns, because we want U to invert what we've done to a.
    */
    for (k=0; k<m; k++) {
	for (dot=0, l=i; l<m; l++) dot += ref2D(u,m,m,k,l)*hv[l];
	s = dot/h;
	for (l=i; l<m; l++) assign2D(u,m,m,k,l) += s*hv[l];
    }
}

/* this is the same as householder_zero_col, but with rows
** and cols swapped.  
*/

 void
FITToolsSVD::householder_zero_row(double *a,double *v,int i,int j,int m,int n,double *hv)
{
    double scale, sigma,snorm, f, h, s, dot;
    int k,l;

    for (scale=0, k=j; k<n; k++) scale += FABS(ref2D(a,m,n,i,k));
    if (!scale) return;

    for (k=j; k<n; k++) hv[k] = ref2D(a,m,n,i,k) / scale;
    for (sigma=0, k=j; k<n; k++) sigma += hv[k]*hv[k];

    f = hv[j];
    snorm = (f > 0) ? -SQRT(sigma) : SQRT(sigma);
    h = f*snorm - sigma;
    hv[j] = f - snorm;

    assign2D(a,m,n,i,j) = scale * snorm;
    for (k=j+1; k<n; k++) assign2D(a,m,n,i,k) = 0;

    for (k=i+1; k<m; k++) {
	for (dot=0, l=j; l<n; l++) dot += ref2D(a,m,n,k,l)*hv[l];
	s = dot/h;
	for (l=j; l<n; l++) assign2D(a,m,n,k,l) += s*hv[l];
    }

    for (k=0; k<n; k++) {
	for (dot=0, l=j; l<n; l++) dot += ref2D(v,n,n,l,k)*hv[l];
	s = dot/h;
	for (l=j; l<n; l++) assign2D(v,n,n,l,k) += s*hv[l];
    }
}

/*
** performs a Givens rotation on the ith and jth columns of a, looking
** at the n elements starting at start.  a is mm x nn.
*/
 void 
FITToolsSVD::rotate_cols(int i,int j,double cos,double sin, double *a, int start, int n,int mm,int nn)
{
    int end = start+n, k;
    double x,y;
    for (k=start; k<end; k++) {
	x = ref2D(a,mm,nn,k,i);
	y = ref2D(a,mm,nn,k,j);
	assign2D(a,mm,nn,k,i) =  cos*x + sin*y;
	assign2D(a,mm,nn,k,j) = -sin*x + cos*y;
    }
}


/*
** performs a Givens rotation on the ith and jth rows of a, looking
** at the n elements starting at start.  a is mm x nn.
*/
 void FITToolsSVD::rotate_rows(int i, int j, double cos, double sin, double *a, int start,int n, int mm, int nn)	
{	
  int end = (start)+(n), _k;	
  double _x,_y;	
  for (_k=(start); _k<end; _k++) {	
	   _x = ref2D(a,mm,nn,i,_k);	
	   _y = ref2D(a,mm,nn,j,_k);	
	   assign2D(a,mm,nn,i,_k) =  cos*_x + sin*_y;	
	   assign2D(a,mm,nn,j,_k) = -sin*_x + cos*_y;	
  }	
}

/*
** This takes a submatrix of b (from p through z inclusive).  The submatrix
** must be bidiagonal, with a zero in the upper-left corner element.  Mucks
** with the sumatrix so that the top row is completely 0, accumulating
** the rotations into u.  b is m x n.  u is m x min.
**
** Suppose the matrix looks like
**   0 R 0 0 0...
**   0 X X 0 0...
**   0 0 X X 0...
**   ...
** Where R & X's are different & non-zero.  We can rotate the first
** and second rows, giving
**   0 0 R 0 0...
**   0 X X 0 0...
**   0 0 X X 0...
**   ...
** with new R's and X's.  We rotate the first and third rows, moving
** R over again, etc.  till R falls off the end.
*/
 void
FITToolsSVD::clr_top_supdiag_elt(double * b,int p,int z,double * u,int m,int n,int min)
//double *b, *u;
//int p,z,m,n,min;
{
    int i;
    double r, x, hypot, cos, sin;

    for (i=p+1; i<=z; i++) {
	r = ref2D(b,m,n,p,i);
	x = ref2D(b,m,n,i,i);
	if (r==0) break;
	hypot = SQRT (r*r + x*x); //hypot = pythag(r,x);
	cos = x/hypot;
	sin = r/hypot;
	/* update the diagonal and superdiagonal elements */
	assign2D(b,m,n,p,i) = 0;
	assign2D(b,m,n,i,i) = hypot;
	/* Rotate the remainder of rows p and i (only need to
	** rotate one more element, since the rest are zero) */
	if (i != z) rotate_rows(p,i,cos,sin,b,i+1,1,m,n);
	/* accumulate the rotation into u */
	rotate_cols(i,p,cos,sin,u,0,m,m,min);
    }
}

/*
** This takes a submatrix of b (from p through z inclusive).  The submatrix
** must be bidiagonal, with a zero in the lower-right corner element.  Mucks
** with the sumatrix so that the right column is completely 0, accumulating
** the rotations into v.  b is m x n.  v is min x n
**
** Suppose the matrix looks like
**   X X 0 0 
**   0 X X 0
**   0 0 X R
**   0 0 0 0
** Where R & X's are different & non-zero.  We can rotate the last
** and second-to-last columns, yielding
**   X X 0 0
**   0 X X R
**   0 0 X 0
**   0 0 0 0
** with new R's and X's.  We rotate the last and third-to-last cols, moving
** R over again, etc.  till R falls off the end.
*/
 void 
FITToolsSVD::clr_bot_supdiag_elt(double * b,int p,int z,double * v,int m,int n,int min)
//double *b, *v;
//int p,z,m,n,min;
{
    int i;
    double r, x, hypot, cos, sin;

    for (i=z-1; i>=p; i--) {
	r = ref2D(b,m,n,i,z);
	x = ref2D(b,m,n,i,i);
	if (r==0) break;
	hypot = SQRT (r*r + x*x); //hypot = pythag(r,x);
	cos = x/hypot;
	sin = r/hypot;
	/* update the diagonal and superdiagonal elements */
	assign2D(b,m,n,i,z) = 0;
	assign2D(b,m,n,i,i) = hypot;
	/* Rotate the remainder of cols z and i (only need to
	** rotate one more element, since the rest are zero) */
	if (i != p) rotate_cols(i,z,cos,sin,b,i-1,1,m,n);
	/* accumulate the rotation into v */
	rotate_rows(i,z,cos,sin,v,0,n,min,n);
    }
}

/*
** This takes a submatrix of b (from p through z inclusive).  The submatrix
** must be bidiagonal except that the topmost subdiagonal element is non-zero.
** Mucks with the submatrix to make it bidiagonal, accumulating the rotations
** into u and v.  b is m x n  u is m x min   v is min x n
**
** Suppose the matrix looks like
**   X X 0 0 0...
**   R X X 0 0...
**   0 0 X X 0...
**   ...
** Where R & X's are different & non-zero.  We can rotate the first and
** second rows, giving
**   X X R 0 0...
**   0 X X 0 0...
**   0 0 X X 0...
**   ...
** with new R &X's.  Now rotate the second and third columns, getting
**   X X 0 0 0...
**   0 X X 0 0...
**   0 R X X 0...
**   ...
** which is the same as the initial problem, but decreased in
** size by 1.  Eventually, we'll reach the situation where we have
**      ...
**   ... X X
**   ... R X
** and the row rotation will eliminate R.
*/
 void 
FITToolsSVD::clr_top_subdiag_elt(double * b,int p,int z,double * u,double * v,int m,int n,int min)
//double *b, *u, *v;
//int p,z,m,n,min;
{
    int i;
    double x, r, hypot, cos, sin;

    for (i=p; ; i++) {
	/* figure out row rotation to zero out the subdiagonal element */
	x = ref2D(b,m,n,i,i);
	r = ref2D(b,m,n,i+1,i);
	hypot = pythag(x,r);
	cos = x/hypot;
	sin = r/hypot;
	/* rotate the leading elements of the row */
	assign2D(b,m,n,i,i) = hypot;
	assign2D(b,m,n,i+1,i) = 0;
	/* rotate out the rest of the row */
	rotate_rows(i,i+1,cos,sin,b,i+1,(i+1==z)?1:2,m,n);
	/* accumulate transformation into columns of u */
	rotate_cols(i,i+1,cos,sin,u,0,m,m,min);

	/* end with a row rotation */
	if (i+1==z) break;

	/* figure out column rotation */
	x = ref2D(b,m,n,i,i+1);
	r = ref2D(b,m,n,i,i+2);
	hypot = pythag(x,r);
	cos = x/hypot;
	sin = r/hypot;
	/* rotate the leading elements of the column */
	assign2D(b,m,n,i,i+1) = hypot;
	assign2D(b,m,n,i,i+2) = 0;
	/* rotate the rest of the column */
	rotate_cols(i+1,i+2,cos,sin,b,i+1,2,m,n);
	/* accumulate transformation into columns of v */
	rotate_rows(i+1,i+2,cos,sin,v,0,n,min,n);
    }
}

/*
** This is the first part of an implicit-shift QR step.  We do some
** magic eigenvalue calculation, to calculate a Givens rotation for
** the top-left corner.
**
** This rotation is described as part of Golub's Algorithm 3.3-1
**
** This is also described as the implicit QL algorithm for symmetric
** tridiagonal matrices, in Numerical Recipes.  Here's what's going
** on, as far as I can tell:
**
** Hypothetically, one could do a series of QR decompositions of a symmetric
** tridiagonal matrix A.
**   - Ignoring the R's, one serially computes An+1 = Qn^t An Q
**   - Eventually, An will approach diagonal
**   - The rate of convergence goes like A[i,j] = u_i/u_j, where
**     u_i and u_j are the eigenvalues of A
**   - To make it converge faster, we shift the eigenvalues by some amount
**     ("uu" in the code) by decomposing the matrix A - uu I.
**   - uu is computed so as to attempt to minimize (u_i-uu)/(u_j-uu), which
**     is the convergence rate of the shifted matrix.  Ideally uu would
**     be equal to the smallest eigenvalue.  Since we don't know the
**     smallest eigenvalue, we look at the eigenvalues of the
**     trailing 2x2 submatrix.
**   - Rather than actually computing the matrix A - uu I, we just keep
**     track of the shift in the calculation of the coefficients for the
**     rotations that make up the Q's.  Hence the "implicit" in the
**     name of the algorithm.
**
** For SVD, we are looking at a bidiagonal matrix, rather than a tridiagonal
** one.  Thus we will do one more level of implicitness, by computing the
** coefficients we WOULD get were we to consider the tridiagonal matrix
** T = B^t B.
**
** This particular routine just performs the initial rotation on the
** bidiagonal matrix, based on the eigenvalue-shift stuff.  This
** makes the matrix no loger bidiagonal.  Other routines will have to
** clean up the non-bidiagonal terms.  The net rotation that is performed
** by this routine and the cleanup routines is the Q at one step of
** the above iteration.  We don't ever explicitly compute Q, we
** just keep updating the U and V matrices.
**
** b is m x n,   v is min x n
*/

 void
FITToolsSVD::golub_kahn_svd_rot( double * b,int p,int q,double * v,int m,int n,int min)
//double *b, *v;
//int p,q,m,n,min;
{
    double t11, t12, t22;
    double uu;
    double b11, b12;
    double dm,fm,dn,fn;
    double hypot,cos,sin;
    double d, s, y, z;

    /* grab the last diagonal and superdiagonal elements of b22 */
    fm = (q-2) < 0 ? 0 : ref2D(b,m,n,q-2,q-1);
    dm = ref2D(b,m,n,q-1,q-1);	fn = ref2D(b,m,n,q-1,q);
				dn = ref2D(b,m,n,q,q);
    
    /* create the trailing 2x2 submatrix of double = b22^t b22 */
    t11 = dm*dm + fm * fm;
    t12 = dm * fn;		t22 = dn * dn + fn * fn;

    /* find the eigenvalues of the T matrix.
    **
    ** The quadratic equation for the eigenvalues has coefficients:
    ** a = 1
    ** b = -(t11+t22)
    ** c = t11 t22 - t12 t12
    **
    ** b^2 - 4ac = (t11^2 + t22^2 + 2 t11 t12) - 4 t11 t22 + 4 t12 t12
    **           = (t11 - t22)^2 + 4 t12 t12
    **
    ** sqrt(b^2-4ac) = sqrt((t11 - t22)^2 + 4 t12 t12) -- use "pythag()"
    **
    ** using quadratic formula, we have the eigenvalues:
    ** (u1,u2) = .5 (t11 + t22 +/- pythag(...))
    **         = t22 + .5 (t11 - t22 +/- pythag(...))
    **
    ** We propogate the .5 into the pythag term, yielding golub's equation 8.2-2
    ** for the "wilkinson shift".
    ** [Note:  Golub says to use (t22 + d - signum(d) s) to find the eigenvalue
    ** closer to t22.  He's right.  ]
    **/
    d = .5f*(t11 - t22);
    s = pythag(d,t12);
    uu = t22 + d + ((d > 0) ? -s : s);

    /* grab the leading 2x1 of b */
    b11 = ref2D(b,m,n,p,p);	b12 = ref2D(b,m,n,p,p+1);

    /* make the leading 2x1 submatrix of T */
    t11 = b11 * b11;
    t12 = b11 * b12;

    /* calculate the rotation that would zero the off-diagonal term of the
    ** shifted matrix T - uu I
    */
    y = t11 - uu;
    z = t12;
    hypot = pythag(y,z);
    cos = y / hypot;
    sin = z / hypot;

    /* perform the rotation on B.  This sprays some glop into the upper-left
    ** subdiagonal element.
    */
    rotate_cols(p,p+1,cos,sin,b,p,2,m,n);
    /* accumulate the rotation into the rows of v */
    rotate_rows(p,p+1,cos,sin,v,0,n,min,n);
}


/* bidiagonalize
**
** Given a (m x n)
** computes u (m x min)	orthogonal cols
**	    b (m x n)	bidiagonal
**	    v (min x n)	orthogonal rows
** such that a = u b v  (looking at only the min x min leading part of b)
**
** Works by starting with B = A, then performing a series of Householder
** transformations to zero-out the off-diagonal elements, while
** accumulating the transformations into U and V.
**
** b may point to the same array as a, in which case the result
** overwrites the original data.
*/

 void
FITToolsSVD::bidiagonalize( const double *a,int m,int n,double *u,double *b,double *v)
{
    int i,j, min=MIN(m,n);
    double *usave = u, *vsave = v, *h = ALLOC1D(MAX(m,n), h);

    /* we need square matrices to accumulate the transformations */
    if (min != m) u = alloc2D(m,m, u);
    if (min != n) v = alloc2D(n,n, v);

    /* start off with u and v equal to the identity, and b equal to a */
    CLEAR2D(u,m,m);
    CLEAR2D(v,n,n);
    for (i=0; i<m; i++) assign2D(u,m,m,i,i) = 1;
    for (i=0; i<n; i++) assign2D(v,n,n,i,i) = 1;
    if (b != a) COPY2D(b,m,n,a);

    /* walk down the diagonal */
    for (i = 0; i<min; i++) {
	/* zero the entries below b[i,i] */
	householder_zero_col(b,u,i,i,m,n,h);
	/* zero the entries to the right of b[i,i+1] */
	householder_zero_row(b,v,i,i+1,m,n,h);
    }
    /* when m < n (matrix wider than tall), the above
       leaves a non-0 element in the [m-1,m]th spot.
       This is the bottom superdiagonal element of an (m+1)x(m+1);
       use the clear routine to get rid of it. */
    if (min != n)
	clr_bot_supdiag_elt(b,0,min,v,m,n,n);

    /* For non-square arrays, lop off the parts we don't need */
    if (min!=m) {
	for (i=0; i<m; i++) for (j=0; j<min; j++)
	    assign2D(usave,m,min,i,j) = ref2D(u,m,m,i,j);
	FREE(u);
    }
    if (min!=n) {
	for (i=0; i<n; i++) for (j=0; j<min; j++)
	    assign2D(vsave,min,n,j,i) = ref2D(v,n,n,j,i);
	FREE(v);
    }
    FREE(h);
}

/*
** Finds the SVD of a bidiagonal matrix.
**
** Zeroes the superdiagonal of a bidiagonal matrix, accumulating
** left- and right-hand transforms into u and v.  The matrix
** is modified in place.
**
** That is, given u, b, v  where b is bidiagonal and u & v are
** rotations, modify the matrices so that (old) u b v = (new) u b v,
** where the new b is diagonal, and u & v are still rotations.
**
** b is m x n,  u is m x min ,  and v is min x n
**
** This is Golub's SVD algorithm (Algorithm 8.3-2)
*/

 WINbool
FITToolsSVD::bidiagonal_svd(double *b,int m,int n,double *u,double *v)
{
    double anorm, t;
    int p, q, i, j, z, iter, min=MIN(m,n);

    /* use the max of the sum of each diagonal element and the
    ** superdiagonal element above it as a norm for the array.
    ** The use of this norm comes from the Numerical Recipes (Press et al)
    ** code.  squawk!
    */

    // CHUCKR: this is where the array access if fouling up
    //for (anorm=ref2D(b,m,n,0,0), i=1; i<n; i++) {
    for (anorm = ref2D(b,m,n,0,0), i=1 ; i < min ; i++) {
	    t = FABS(ref2D(b,m,n,i,i)) + FABS(ref2D(b,m,n,i-1,i));
	    if (t > anorm) anorm = t;
    }

    /* Each iteration of this loop will zero out the superdiagonal
    ** element above b[q][q] -- i.e. b[q][q] will be a singular value.
    */
    WINbool bRes = TRUE;
    for (q = min-1; q>=0; q--) {

	/* Each iteration will make the superdiagonal elements smaller,
	** till they drop below numerical noise.  It ought to converge
	** by 30 tries.  squawk!
	*/
	for (iter=0; iter<30; iter++) {
	    
	    /* Find a block that has a zero on the diagonal or superdiagonal.
	    ** That is, we are breaking up b into three submatrices,
	    ** b11, b22, b33, such that b33 is diagonal and b22 has no zeros 
	    ** on the superdiagonal.
	    ** b33 goes from q+1 to n-1 -- it's the part we already did
	    ** b22 goes from p through q -- it's a bidiagonal block
	    */

	    /* sweep back till we reach a numerically 0 superdiagonal element */
	    for (p = q; p>0; p--) if (ref2D(b,m,n,p-1,p) + anorm == anorm) {
		assign2D(b,m,n,p-1,p) = 0;
		break;
	    }

	    /* if b22 is 1x1, i.e. there is a 0 above b[q,q], then
	    ** b[q,q] is the singular value.  Go on to the next
	    ** singular value.  (But first, make sure it's
	    ** positive.  If it's negative, negate it and the
	    ** corresponding row of v)
	    */

	    if (p==q) {
		if (ref2D(b,m,n,q,q) < 0) {
		    assign2D(b,m,n,q,q) *= -1;
		    for (j=0; j<n; j++) assign2D(v,min,n,q,j) *= -1;
		}
		break;
	    }

	    /* check for zero on the diagonal */
	    for (z= -1, i=q; i>=p; i--) if (ref2D(b,m,n,i,i)+anorm==anorm) {
		z = i;
		break;
	    }
	    if (z >= 0) {
		/* get rid of zero on the diagonl.  where is it? */
		if (z == q)
		    /* lower-right corner.  clr element above it */
		    clr_bot_supdiag_elt(b,0,z,v,m,n,min);
		else
		    /* in the middle.  clr elt to its right */
		    clr_top_supdiag_elt(b,z,q,u,m,n,min);
	    }
	    else {
		/* b22 has no zeroes on the diagonal or superdiagonal.
		** Do magic rotation, leaving glop in the uppermost
		** subdiagonal element
		*/
		golub_kahn_svd_rot(b,p,q,v,m,n,min);
		/* get rid of the glop in the uppermost subdiagonal */
		clr_top_subdiag_elt(b,p,q,u,v,m,n,min);
	    }
	}
    if (iter>=30) {
        bRes = FALSE;
	    (void) fprintf(stderr,"svd: didn't converge after 30 iterations\n");
    }
    }
    return bRes;
}


/**************************************************************************
**
** SVD
**
** Given a (m x n)
** computes u (m x min)	column-orthonormal
**          w (min x min)	diagonal
**          vt (min x n)	column-orthonormal
** where min=min(m,n),
** such that a = u w vt
** 
** w is returned as a vector of length min
**
**                                                t
** NOTE:  the SVD is commonly represented as U W V
**        where U is m x min and V is n x min.  This routine
**        computes the min x n transpose of V, placing the result
**        in (the memory pointed to by) parameter "vt"
**
****************************************************************************/

WINbool
FITToolsSVD::SVD(const double *a, int m, int n,
                 double *u, double *w, double *vt)
{
   int i, j, k, min = MIN(m,n);
   double *g, *p, wi;
   
   g = alloc2D(m,n, g);
   bidiagonalize(a,m,n,u,g,vt);
   
   
   WINbool bRes = bidiagonal_svd(g,m,n,u,vt);
   for (i=0; i<min; i++) w[i] = ref2D(g,m,n,i,i);

    /* insertion sort singular values.  Note that since
    ** the svd algorithm produces values sorted or nearly sorted,
    ** an insertion sort is efficient.
    */
    for (i=1; i<min; i++) {
	if (w[i] > w[i-1]) {
	    /* save w[i], and col i of u and row i of vt.  use "g" as buffer */
	    wi = w[i];
	    p=g;
	    for (k=0; k<m; k++) *p++ = ref2D(u,m,min,k,i);
	    for (k=0; k<n; k++) *p++ = ref2D(vt,min,n,i,k);
	    /* slide columns over */
	    for (j=i; j>0 && wi > w[j-1]; j--) {
		w[j] = w[j-1];
		for (k=0; k<m; k++) assign2D(u,m,min,k,j)=ref2D(u,m,min,k,j-1);
		for (k=0; k<n; k++) assign2D(vt,min,n,j,k)=ref2D(vt,min,n,j-1,k);
	    }
	    /* put original i stuff where we ended up */
	    w[j] = wi;
	    p=g;
	    for (k=0; k<m; k++) assign2D(u,m,min,k,j) = *p++;
	    for (k=0; k<n; k++) assign2D(vt,min,n,j,k) = *p++;
	}
    }
	    
    FREE(g);

    return bRes;
}

/**************************************************************************
**
** SVD_BACKSUBST
**
** Given the svd of a matrix A, and a vector B, solves A X = B for
** a vector X.  Takes a condition-threshold factor "eps", singular
** values less than "eps" times the largest value are considered to be
** zero.
**
** input:
**	    u (m x min)			column-orthonormal
**          w (min-element vector)	diagonal elements
**          vt (min x n)		column-orthonormal
**    where min=min(m,n),
**    such that a = u w vt, as decomposed by num_svd()
**	    b (m-element vector)	
**
** output:
** 	    x (n-element vector)
**  
****************************************************************************/

void
FITToolsSVD::Backsubst(int m, int n, 
                       const double *u, const double  *w, const double  *vt,
                       const double  b[], double x[], double eps)
{
    const int min = MIN(m,n);

/* singular vals are sorted, w[0] is max */
    const double thresh = eps * w[0]; 
    double *tmp = ALLOC1D(min, tmp);
    int i,j,k;
    
    for (j=0; j<min; j++) {
        double s = 0;
        if (w[j] >= thresh) {
            for (i=0; i<m; i++) s += ref2D(u,m,min,i,j) * b[i];
            s /= w[j];
        }
        tmp[j] = s;
    }
    for (k=0; k<n; k++) {
        double s=0;
        for (j=0; j<min; j++) s += ref2D(vt,min,n,j,k) * tmp[j];
        x[k] = s;
    }
    FREE(tmp);
}

/********** Debugging routines below.  define BIDIAG_MAIN
*********** to test out the bidiagonalization.  define TESTMAIN
*********** to test out the full svd.  
*/
#if defined(BIDIAG_MAIN) || defined(TESTMAIN)

/* multiplies results of svd, to reconstruct the original matrix */
void
num_svd_mult(double *r,int m,int n,const double *u,const double *w,const double *v)
{
    int i,j,k,min=MIN(m,n);
    double s;
    for (i=0; i<m; i++) for (j=0; j<n; j++) {
        for (s=0,k=0; k<min; k++)
            s += ref2D(u,m,min,i,k) * ref2D(v,min,n,k,j) * w[k];
        ref2D(r,m,n,i,j) = s;
    }
}

pmat(s,a,m,n)
char *s;
double *a;
int m,n;
{
    int i,j;
    (void) printf("%s: (%dx%d)\n", s, m, n);
    for (i=0; i<m; i++) {
	for (j=0; j<n; j++) (void) printf("\t%e", ref2D(a,m,n,i,j));
	(void) printf("\n");
    }
}

pmult(s,a,b,m,n,p)
char *s;
double *a,*b;
int m,n,p;
{
    double r[BUFSIZ];
    int i,j,k;

    for (i=0; i<m; i++) for (j=0; j<p; j++)
	for (assign2D(r,m,p,i,j)=0, k=0; k<n; k++) 
	    assign2D(r,m,p,i,j) += ref2D(a,m,n,i,k) * ref2D(b,n,p,k,j);
    pmat(s,r,m,p);
}
pmultt(s,a,b,m,n,p)
char *s;
double *a,*b;
int m,n,p;
{
    double r[BUFSIZ];
    int i,j,k;

    for (i=0; i<m; i++) for (j=0; j<p; j++)
	for (assign2D(r,m,p,i,j)=0, k=0; k<n; k++) 
	    assign2D(r,m,p,i,j) += ref2D(a,m,n,i,k) * ref2D(b,p,n,j,k);
    pmat(s,r,m,p);
}

/* #define M 20 */
/* #define N 20 */
/* #define MN 20 */
/* static double xa[M][N] = { */
    /* {1, 3, 2, -2, -4, 6, -5, 1, 2, 7, -9, 2,11, 1, 6, 8, 2, 1, 8, 2}, */
    /* {3, 3, 2, -2, -4, 7, -5, 1, 2, 7, -9, 2, 2, 4, 1, 8, 2, 1, 8, 2}, */
    /* {6, 3, 9, -2, -4, 8, -2, 1, 2, 7, -3, 5, 2, 4, 6, 1, 2, 1, 8, 2}, */
    /* {9, 3, 2, -2, -4, 9, -5, 1, 2, 7, -9, 2, 2, 4, 6, 8, 1, 1, 8, 2}, */
    /* {2, 3, 8, -2, -4, 0, -5, 1, 2, 7, -9, 2, 2, 4, 6, 8, 2, 2, 8, 2}, */
    /* {0, 1, 2, -2, -4, 2, -5, 1, 2, 7, -9, 2, 2, 4, 6, 8, 2, 1, 3, 2}, */
    /* {4, 2, 0, -2, -4, 2, -5, 1, 2, 7, -9, 2, 2, 4, 6, 8, 2, 1, 8, 5}, */
    /* {9, 3, 2, -3, -4, 3, -5, 1, 7, 7, -9, 2, 2, 4, 6, 8, 2, 1, 8, 2}, */
    /* {1, 4, 7, -2, -4, 4, -5, 1, 2, 9, -9, 2, 2, 4, 6, 8, 2, 1, 8, 2}, */
    /* {2, 5, 2, -3, -4, 5, -5, 1, 2, 7, -9, 2, 2, 4, 6, 8, 2, 1, 8, 2}, */
    /* {0, 6, 6, -0, -4, 7, -5, 1, 3, 7, -9, 2, 2, 4, 6, 8, 2, 1, 8, 2}, */
    /* {6, 7, 2, -8, -4, 8, -5, 1, 2, 7, -9, 2, 2, 4, 6, 8, 2, 1, 8, 2}, */
    /* {7, 8, 5, -9, -4, 9, -5, 1, 2, 7, -9, 2, 2, 4, 6, 8, 2, 1, 8, 2}, */
    /* {9, 9, 2, -2, -4, 0, -5, 1, 2, 7, -9, 2, 2, 4, 6, 8, 2, 1, 8, 2}, */
    /* {2, 3, 4, -2, -4,-1, -5, 1, 2, 7, -9, 1, 2, 4, 6, 8, 2, 1, 8, 2}, */
    /* {1, 3, 2, -2, -4,-3, -5, 1, 2, 7, -9, 2, 2, 4, 6, 8, 2, 1, 8, 2}, */
    /* {1, 3, 3, -6, -4,-4, -3, 1, 2, 7, -4, 2, 8, 4, 6, 8, 2, 1, 8, 2}, */
    /* {1, 3, 1, -4, -4,-5, -5, 1, 2, 7, -9, 2, 2, 4, 6, 8, 2, 1, 8, 2}, */
    /* {1, 3, -1, 2, -4,-6, -5, 1, 2, 2, -9, 2, 2, 4, 6, 8, 2, 1, 8, 2}, */
    /* {1, 3, -2, 1, -4,-7, -5, 1, 2, 7, -9, 2, 2, 4, 6, 8, 2, 1, 8, 2}}; */
/* #define M 1 */
/* #define N 10 */
/* #define MN 1 */
/* static double xa[M][N] = { { */
/* 3.964648e-01,8.404854e-01,3.533361e-01,4.465834e-01,3.186928e-01,8.864284e-01,1.558285e-02,5.840902e-01,1.593686e-01,3.837159e-01 */
/* }}; */

#define M 3
#define N 4
#define MN 3
static double xa[M][N] = {{1, 3, 2, -2}, {-4, 6, -5, 1}, {2, 7, -9, 2}};

/* #define M 4 */
/* #define N 3 */
/* #define MN 3 */
/* static double xa[M][N] = {{1, 3, 2}, {-4, 6, 5}, {2, 7, -9}, {8, -4, -3}}; */


#endif


#ifdef BIDIAG_MAIN
double xu[M][MN], xb[M][N], xv[MN][N];
double xub[M][MN], xubv[M][N];
main()
{
    int i, j, k;
    double s;

    bidiagonalize((double*)xa,M,N,(double*)xu,(double*)xb,(double*)xv);
    pmat("a",(double*)xa,M,N);
    pmat("b",(double*)xb,M,N);
    pmat("u",(double*)xu,M,MN);
    pmat("vt",(double*)xv,MN,N);
    for (i=0; i<M; i++) for (j=0; j<MN; j++) {
	for (s=0,k=0;k<MN;k++) s += xu[i][k]*xb[k][j];
	xub[i][j] = s;
    }
    for (i=0; i<M; i++) for (j=0; j<N; j++) {
	for (s=0,k=0;k<MN;k++) s += xub[i][k]*xv[k][j];
	xubv[i][j] = s;
    }
    pmat("ubv",(double*)xubv,M,N);
    return(0);
}
#endif /* BIDIAG_MAIN */



#ifdef TESTMAIN
double xu[M][MN], xw[MN], xv[MN][N], xr[M][N];
main()
{
    int i, j, k;
    double s, rn;

    num_svd((double*)xa,M,N,(double*)xu,(double*)xw,(double*)xv);
    num_svd_mult((double*)xr,M,N,(double*)xu,(double*)xw,(double*)xv);
    for (rn=0,i=0; i<M; i++) for (j=0; j<N; j++) {
	xr[i][j] -= xa[i][j];
	rn += fabs(xr[i][j]);
    }
    pmat("a",(double*)xa,M,N);
    pmat("u",(double*)xu,M,MN);
    pmat("w",(double*)xw,1,MN);
    pmat("vt",(double*)xv,MN,N);
    pmat("r",(double*)xr,M,N);
    printf("%e\n",rn/(M*N));
    return(0);
}
#endif /* TESTMAIN */


