/* Copyright 1996, Microsoft.  All Rights Reserved. */


/* -------------------------------------------------------------------------
 *
 *               Solve Ax = b
 *
 * ------------------------------------------------------------------------- */
#include "StdAfx.H"

#include <fitting/FITTools_LSConstrained.H>
#include <fitting/FITTools_SVD.H>
#include <mkl/mkl_lapack.h>

/* -----------------------  Private Methods  ------------------------------- */

/* ----------------------- Protected Methods ------------------------------- */


/* -----------------------  Public Methods   ------------------------------- */

// solves overdetermined or underdetermined real linear systems
//  involving an M-by-N matrix A, or its transpose, using a QR or LQ
//  factorization of A.  It is assumed that A has full rank.
//   NOTE: X is column-row ordering (Fortran style)
static long int CallDgels( int in_lNRows,
                           int in_lNCols,
                           Array<double>  &in_adA,  // A input in column row format
                           int in_lNColsB,
                           Array<double>  &in_adB,  // rhs B input in column row format
                           Array<double>  &out_adX  // return the result (each column is a solution)
              )
{
    
    int lWork = WINmin( in_lNRows, in_lNCols ) + WINmax( WINmax(( int)1, in_lNRows), 
                                                         WINmax(in_lNCols, in_lNColsB) );

    Array<double> adWork( lWork );
    char cTrans = 'N';
    
    int lInfo, lLdb = WINmax( in_lNRows, in_lNCols);

    double *adB = NULL;
    // B is too small, copy to X before doing dgels
    if ( in_lNRows < in_lNCols ) {
        ASSERT( out_adX.num() >= in_lNColsB * in_lNCols );
        ASSERT( in_adB.num() >= in_lNColsB * in_lNRows );

        for ( int i = 0; i < in_lNColsB; i++ ) {
            for ( int j = 0; j < in_lNRows; j++ ) {
                out_adX[ (int) (i * in_lNCols) + j ] = in_adB[ (int) (i * in_lNRows) + j ];
            }
            memcpy( &out_adX[ (int) (i * in_lNCols) ], 
                    &in_adB[ (int) (i * in_lNRows) ], 
                    (int) in_lNRows * sizeof(double) );
        }
        adB = &out_adX[0];
    } else {
        adB = &in_adB[0];
    }

    dgels( &cTrans, 
           &in_lNRows, &in_lNCols, &in_lNColsB, 
           in_adA, &in_lNRows, 
           adB, &lLdb, 
           adWork, &lWork, &lInfo );

    // Result was in B. Copy to X.
    if ( in_lNRows >= in_lNCols ) {
        for ( int i = 0; i < in_lNColsB; i++ ) {
            memcpy( &out_adX[(int) (i * in_lNCols)], 
                    &in_adB[(int) (i * in_lNRows)], 
                    in_lNCols * sizeof(double) );
        }
    }

    return lInfo;
}


inline int Index( int in_iRow, int in_iCol, int in_iNRows, int in_iCols ) 
{
   ASSERT(in_iRow >= 0 && in_iRow < in_iNRows);
   ASSERT(in_iCol >= 0 && in_iCol < in_iCols);
   return in_iCol * in_iNRows + in_iRow;
}

// solves overdetermined or underdetermined real linear systems
//  involving an M-by-N matrix A, or its transpose, using a QR or LQ
//  factorization of A.  It is assumed that A has full rank.
//   NOTE: X is column-row ordering (Fortran style)
// out_VT will be of size in_lNCols X out_lNNull
static long int GetNullSpace( int in_lNRows,
                              int in_lNCols,
                              Array<double>  &in_adQ,  // A input in column row format
                              int &out_lNNull,
                              Array<double>  &out_VT // return the result (each column is a solution)
              )
{
    char jobu = 'N';
    char jobv = 'A';
    int lda = in_lNRows;
    int ldu = in_lNCols;
    int ldvt = in_lNCols;
    int lwork = 10 * WINmax(3*WINmin(in_lNRows,in_lNCols)+WINmax(in_lNRows,in_lNCols),5*WINmin(in_lNRows,in_lNCols));
    int linfo = 0;
    double *a = new double [ in_adQ.num() ];
    double *work = new double [lwork];
    double *s = new double [WINmax( in_lNRows, in_lNCols )];
    double *u = new double [in_lNRows * in_lNRows];
    double *v = new double [ in_lNCols * in_lNCols ];

    memcpy( a, &in_adQ[0], sizeof(double) * in_adQ.num() );
    dgesvd( &jobu, &jobv, &in_lNRows, &in_lNCols, a, &in_lNRows, s, u, &in_lNRows, v, &in_lNCols, work, &lwork, &linfo );

    out_lNNull = in_lNCols - in_lNRows;

    out_VT.need( out_lNNull * in_lNCols );

    int iCount = 0;
    for ( int i = in_lNRows; i < in_lNCols; i++ ) {

        for ( int j = 0; j < in_lNCols; j++ ) {
            out_VT[ Index( iCount, j, out_lNNull, in_lNCols) ] = v[ Index(i,j,in_lNCols, in_lNCols) ];
        }
        iCount++;
    }

    /*
    for ( int r = 0; r < in_lNCols; r++ ) {
        for ( int c = 0; c < in_lNCols; c++ ) {
            cout << v[ Index(r,c, in_lNCols, in_lNCols) ] << " ";
        }
        cout << "\n";
    }
    cout << "\n";
    */

    delete a;
    delete work;
    delete s;
    delete u;
    delete v;

    return linfo;
}

WINbool
FITTools_LSConstrained::Solve()
{
    Array<double> adVT;

    //Print(cout, FALSE);

    int iNNull;
    const long int linfo = GetNullSpace( m_iNConstrained, m_iNVariables, m_adConstrainedData, iNNull, adVT );
    VERIFY( linfo == 0 );

    for ( int r = 0; r < iNNull; r++ ) {
        double dSum = 0.0;
        for ( int i = 0; i < m_iNConstrained; i++ ) {
            for ( int j = 0; j < m_iNVariables; j++ ) {
                dSum += Constrained( i, j ) * adVT[ Index(r,j,iNNull,m_iNVariables) ];
            }
        }
        ASSERT( RNIsZero( dSum, 1e-6 ) );
    }

    /*
    cout << "Null space " << iNNull << " " << m_iNVariables << "\n";
    for ( FORINT r = 0; r < iNNull; r++ ) {
        for ( int c = 0; c < m_iNVariables; c++ )
            cout << adVT[ Index(r,c,iNNull, m_iNVariables) ] << " ";
        cout << "\n";
    }
    cout << "\n";
    */

    int iRes2 = 0;
    const double dR1 = Constrained(0,0);
    const double dR2 = Constrained(0,1);
    iRes2 = CallDgels( m_iNConstrained, m_iNVariables,
                       m_adConstrainedData, m_iNDimension,
                       m_adRhsConstrainedData,
                       m_adSolutionData);

    // A (x-x') = A x - A x' = b
    Array<double> adA( m_iNDesired * iNNull );
    //cout << "New A " << m_iNDesired << " " << iNNull << "\n";
    for ( int i = 0; i < m_iNDesired; i++ ) {
        for ( int j = 0; j < iNNull; j++ ) {
            double &d = adA[ Index(i,j, m_iNDesired, iNNull) ];
            d = 0.0;
            for ( int k = 0; k < m_iNVariables; k++ ) {
                d += Desired( i, k ) * adVT[ Index(j,k,iNNull, m_iNVariables) ];
            }
            //cout << d << " ";
        }
        //cout << "\n";
    }
    //cout << "\n\n";

    // b = b - A * x'
    for ( FORINT i = 0; i < m_iNDesired; i++ ) {
        for ( int j = 0; j < m_iNDimension; j++ ) {
            double &d = Rhs( i, j );
            for ( int k = 0; k < m_iNVariables; k++ )
                d -= Desired( i, k ) * Solution(k, j );
        }
    }

    int iRes = 0;
    WINbool bRes = TRUE;
    Array<double> adSoln( iNNull * m_iNDimension );
    iRes = CallDgels( m_iNDesired, iNNull,
                      adA, m_iNDimension,
                      m_adRhsData,
                      adSoln);

    if ( iRes != 0 || iRes2 != 0 ) {
        bRes = FALSE;
        cerr << m_iNDesired << " " << m_iNVariables << " " << iRes << "\n";
    }

    // Multiply the solution in adSoln by adVT
    Array<double> adXAdd( m_iNVariables * m_iNDimension );
    for ( FORINT i = 0; i < m_iNVariables; i++ ) {
        for ( int j = 0; j < m_iNDimension; j++ ) {
            double &d = adXAdd[ Index(i,j, m_iNVariables, m_iNDimension) ];
            d = 0.0;
            for ( int k = 0; k < iNNull; k++ ) {
                d += adVT[ Index(k,i,iNNull, m_iNVariables) ] * adSoln[ Index(k, j, iNNull, m_iNDimension) ];
            }
        }
    }

    // Q (x + x') = c
    //Add result into original solution
    for ( FORINT i = 0; i < m_iNVariables * m_iNDimension; i++ )
        m_adSolutionData[i] += adXAdd[i];

    for (FORINT i = 0; i < m_iNDimension; i++) {
       for (int c = 0; c < m_iNVariables; c++) {
           //TRACE("%f %f\n", lsCopy.Solution_utils(c, i), Solution_clapack(c, i));
           if ( ! ( Solution(c, i) > -1e20 && Solution(c, i) < 1e20 ) )
               bRes = FALSE;
       }
   }

   return bRes;
}

/* -------------------------------------------------------------------------
 * DESCR   :	Create the matrices that will Ae needed
 * ------------------------------------------------------------------------- */
FITTools_LSConstrained::FITTools_LSConstrained( const int in_iNDesired, // number of rows/constraints
                                                const int in_iNConstrained, // number of variables
                                                const int in_iNVariables, // number of variables
                                                const int in_iNDimension //  number of cols in solution
                                                ) 
: m_iNDesired(in_iNDesired), m_iNVariables(in_iNVariables),
  m_iNConstrained( in_iNConstrained ),
   m_iNDimension(in_iNDimension),
   m_adDesiredData( in_iNDesired * m_iNVariables ),
   m_adConstrainedData( in_iNConstrained * m_iNVariables ),
   m_adSolutionData( in_iNVariables * in_iNDimension ),
   m_adRhsConstrainedData( in_iNConstrained * in_iNDimension ),
   m_adRhsData( in_iNDesired * in_iNDimension )
{
    ASSERT( in_iNDesired > 0 );
    ASSERT( in_iNVariables > 0 );
    ASSERT( in_iNConstrained > 0 );
    ASSERT( in_iNDimension > 0 );
    ASSERT( in_iNConstrained <= in_iNVariables );

    Clear();
}

void FITTools_LSConstrained::Clear()
{
    m_adDesiredData.memfill(0);
    m_adConstrainedData.memfill(0);
    m_adSolutionData.memfill(0);
    m_adRhsData.memfill(0);
    m_adRhsConstrainedData.memfill(0);
}

/* -------------------------------------------------------------------------
 * DESCR   :	Zap everything
 * ------------------------------------------------------------------------- */
FITTools_LSConstrained::~FITTools_LSConstrained()
{
}



void FITTools_LSConstrained::Print(ostream &out, WINbool in_bDoTrace )
{
    Array<double> adSum( N_variables() );
    Array<double> adSumAbs( N_variables() );

    for (int iV = 0; iV < N_variables(); iV++) {
        adSum[iV] = 0;
        adSumAbs[iV] = 0;
    }

    if ( in_bDoTrace )
        TRACE("desired %d variables %d\n", N_desired(), N_variables());
    out << "Desired " << N_desired() << " " << N_variables() << "\n";
    for (int iD = 0; iD < N_desired(); iD++) {
        double dSum = 0;
        for (int iV = 0; iV < N_variables(); iV++) {
            dSum += Desired(iD, iV);
            adSum[iV] += Desired(iD, iV);
            adSumAbs[iV] += fabs( Desired(iD, iV) );
            if (!RNIsZero( Desired(iD, iV) ) ) {
                out << iV << " " << Desired(iD, iV) << "  ";
                if ( in_bDoTrace )
                    TRACE("%d %f  ", iV, Desired(iD, iV) );
            }
        }

        if ( in_bDoTrace )
            TRACE(" = ");
        out << " = ";
        for (int iS = 0; iS < N_dimension(); iS++) {
            if ( in_bDoTrace )
                TRACE("%f ", Rhs(iD, iS));
            out << Rhs(iD, iS) << " ";
        }
        out << "  Sum " << dSum << "\n";
        if ( in_bDoTrace )
            TRACE("  Sum %f\n", dSum);
    }
    if ( in_bDoTrace )
        TRACE("\n\n");
    for ( FORINT iV = 0; iV < N_variables(); iV++) {
        if ( in_bDoTrace )
            TRACE("%f ", adSum[iV]);
        out << adSum[iV] << " ";
    }
    out << "\n";
    for ( FORINT iV = 0; iV < N_variables(); iV++) {
        if ( in_bDoTrace )
            TRACE("%f ", adSumAbs[iV]);
        out << adSumAbs[iV] << " ";
    }
    out << "\n\n";
    if ( in_bDoTrace )
        TRACE("\n\n");


    for (FORINT iV = 0; iV < N_variables(); iV++) {
        adSum[iV] = 0;
        adSumAbs[iV] = 0;
    }

    if ( in_bDoTrace )
        TRACE("constrained %d variables %d\n", N_constrained(), N_variables());
    out << "constrained " << N_constrained() << " " << N_variables() << "\n";
    for (FORINT iD = 0; iD < N_constrained(); iD++) {
        double dSum = 0;
        for (int iV = 0; iV < N_variables(); iV++) {
            dSum += Constrained(iD, iV);
            adSum[iV] += Constrained(iD, iV);
            adSumAbs[iV] += fabs( Constrained(iD, iV) );
            if (!RNIsZero( Constrained(iD, iV) ) ) {
                out << iV << " " << Constrained(iD, iV) << "  ";
                if ( in_bDoTrace )
                    TRACE("%d %f  ", iV, Constrained(iD, iV) );
            }
        }

        if ( in_bDoTrace )
            TRACE(" = ");
        out << " = ";
        for (int iS = 0; iS < N_dimension(); iS++) {
            if ( in_bDoTrace )
                TRACE("%f ", RhsConstrained(iD, iS));
            out << RhsConstrained(iD, iS) << " ";
        }
        out << "  Sum " << dSum << "\n";
        if ( in_bDoTrace )
            TRACE("  Sum %f\n", dSum);
    }
    if ( in_bDoTrace )
        TRACE("\n\n");
    for ( FORINT iV = 0; iV < N_variables(); iV++) {
        if ( in_bDoTrace )
            TRACE("%f ", adSum[iV]);
        out << adSum[iV] << " ";
    }
    out << "\n";
    for ( FORINT iV = 0; iV < N_variables(); iV++) {
        if ( in_bDoTrace )
            TRACE("%f ", adSumAbs[iV]);
        out << adSumAbs[iV] << " ";
    }
    out << "\n\n";
    if ( in_bDoTrace )
        TRACE("\n\n");

}


static void Test2By2( const R2Vec &vec1, const R2Vec &vec2, const double dLH1, const double dLH2 )
{
    FITTools_LSConstrained oLS(1,1,2,1);

    oLS.Desired(0,0) = vec1[0];
    oLS.Desired(0,1) = vec1[1];
    oLS.Constrained(0,0) = vec2[0];
    oLS.Constrained(0,1) = vec2[1];

    oLS.Rhs(0, 0) = dLH1;
    oLS.RhsConstrained(0, 0) = dLH2;

    VERIFY( oLS.Solve() == TRUE );

    const double dRes1 =  vec1[0] * oLS.Solution(0,0) + vec1[1] * oLS.Solution(1,0) - dLH1;
    const double dRes2 =  vec2[0] * oLS.Solution(0,0) + vec2[1] * oLS.Solution(1,0) - dLH2;

    ASSERT( RNIsZero( dRes1, 1e-6 ) );
    ASSERT( RNIsZero( dRes2, 1e-6 ) );
}

static void Test3By3a( const R3Vec &vec1, const R3Vec &vec2, const R3Vec &vec3, const double dLH1, const double dLH2, const double dLH3 )
{
    FITTools_LSConstrained oLS(2,1,3,1);

    oLS.Desired(0,0) = vec1[0];
    oLS.Desired(0,1) = vec1[1];
    oLS.Desired(0,2) = vec1[2];
    oLS.Desired(1,0) = vec2[0];
    oLS.Desired(1,1) = vec2[1];
    oLS.Desired(1,2) = vec2[2];
    oLS.Constrained(0,0) = vec3[0];
    oLS.Constrained(0,1) = vec3[1];
    oLS.Constrained(0,2) = vec3[2];

    oLS.Rhs(0, 0) = dLH1;
    oLS.Rhs(1, 0) = dLH2;
    oLS.RhsConstrained(0, 0) = dLH3;

    VERIFY( oLS.Solve() == TRUE );

    const double dRes1 =  vec1[0] * oLS.Solution(0,0) + vec1[1] * oLS.Solution(1,0) + vec1[2] * oLS.Solution(2,0) - dLH1;
    const double dRes2 =  vec2[0] * oLS.Solution(0,0) + vec2[1] * oLS.Solution(1,0) + vec2[2] * oLS.Solution(2,0) - dLH2;
    const double dRes3 =  vec3[0] * oLS.Solution(0,0) + vec3[1] * oLS.Solution(1,0) + vec3[2] * oLS.Solution(2,0) - dLH3;

    ASSERT( RNIsZero( dRes1, 1e-6 ) );
    ASSERT( RNIsZero( dRes2, 1e-6 ) );
    ASSERT( RNIsZero( dRes3, 1e-6 ) );
}

static void Test3By3b( const R3Vec &vec1, const R3Vec &vec2, const R3Vec &vec3, const double dLH1, const double dLH2, const double dLH3 )
{
    FITTools_LSConstrained oLS(1,2,3,1);

    oLS.Constrained(0,0) = vec1[0];
    oLS.Constrained(0,1) = vec1[1];
    oLS.Constrained(0,2) = vec1[2];
    oLS.Constrained(1,0) = vec2[0];
    oLS.Constrained(1,1) = vec2[1];
    oLS.Constrained(1,2) = vec2[2];
    oLS.Desired(0,0) = vec3[0];
    oLS.Desired(0,1) = vec3[1];
    oLS.Desired(0,2) = vec3[2];

    oLS.RhsConstrained(0, 0) = dLH1;
    oLS.RhsConstrained(1, 0) = dLH2;
    oLS.Rhs(0, 0) = dLH3;

    VERIFY( oLS.Solve() == TRUE );

    const double dRes1 =  vec1[0] * oLS.Solution(0,0) + vec1[1] * oLS.Solution(1,0) + vec1[2] * oLS.Solution(2,0) - dLH1;
    const double dRes2 =  vec2[0] * oLS.Solution(0,0) + vec2[1] * oLS.Solution(1,0) + vec2[2] * oLS.Solution(2,0) - dLH2;
    const double dRes3 =  vec3[0] * oLS.Solution(0,0) + vec3[1] * oLS.Solution(1,0) + vec3[2] * oLS.Solution(2,0) - dLH3;

    ASSERT( RNIsZero( dRes1, 1e-6 ) );
    ASSERT( RNIsZero( dRes2, 1e-6 ) );
    ASSERT( RNIsZero( dRes3, 1e-6 ) );
}

void FITTools_LSConstrained::Test()
{
    Test2By2( R2Vec(0.75,1.5), R2Vec(-0.25,0.5), 4.0, 3.0 );
    Test2By2( R2Vec(1,0), R2Vec(0,1), 4.0, 3.0 );
    Test2By2( R2Vec(0.5,0.5), R2Vec(-0.5,0.5), 4.0, 3.0 );

    Test3By3a( R3Vec(1,0,0), R3Vec(0,1,0), R3Vec(0,0,1), 4.0, 3.0, 5.0 );
    Test3By3a( R3Vec(0.2,0.5,1.0), R3Vec(0.3,1.2,-0.1), R3Vec(-0.3,0.2,1.3), 4.0, 3.0, 5.0 );
    Test3By3b( R3Vec(1,0,0), R3Vec(0,1,0), R3Vec(0,0,1), 4.0, 3.0, 5.0 );
    Test3By3b( R3Vec(0.2,0.5,1.0), R3Vec(0.3,1.2,-0.1), R3Vec(-0.3,0.2,1.3), 4.0, 3.0, 5.0 );
}
