#include "stdafx.h"
#include <utils/Pm_MeshLite.H>
#include "PMeshLiteUtils.h"

/***************** MakeEllipsoidNRing ***********************************
  This routine generates a mesh for a N-ring neighborhood around a 
  point on an ellipsoid shaped surface
  *******************************************************************/
WINbool
MakeEllipsoidNRing(PMeshLite &new_mesh,
                   const double r, const double a, const double b,
                   const int n, const double phi,
                   const double dr_edge, const double dr_cent,
                   const double dphi_edge, const double dphi_cent,
                   const double dtheta, const int nRings )
{
    int i, j, npts, offset, i1, i2, i3, i4, i5, pointGrowthLimit;
    double x, y, z, PI, radical, Nmag, Nx, Ny, Nz, alpha, beta, theta, tmp;
    Array< Vertex > aoVs;
    Array< Vertex > aonv;
    Array< Face> aoFs;

    int verbose = 0;

    new_mesh.Clear();

    /* Eqn: (x/r)^2 + (y/a)^2 + (z/b)^2 = 1 */

    PI = 4.0 * atan( 1.0 );
    alpha = 0.25 * PI;
    pointGrowthLimit = 4;

    /* Print theoretical curvature */

    float k1, k2, kbar, kg;
    if( a*a < b*b ) {
        k1 = (float) (r / (a*a));
        k2 = (float) (r / (b*b));
    } else {
        k1 = (float) (r / (b*b));
        k2 = (float) (r / (a*a));
    }
    kbar = (float) (0.5 * ( k1 + k2 ));
    kg = k1 * k2;

    if( verbose > 0 ) {
        printf("\n%s: Target Curvature K1 = %f, K2 = %f, Kbar = %f, Kg = %f\n",
                "Ellipsoid", k1, k2, kbar, kg );
    }

    /* Generate the center point (which can include error) */

    x = r * cos( dphi_cent );
    y = r * sin( dphi_cent );
    z = 0.0;

    radical = 1.0 - ( y*y/(a*a) + z*z/(b*b) );
    if( radical >= 0.0 ) {
        x = r * sqrt( radical );
    } else {
        printf( "Radical out of range: %lf\n", radical );
        new_mesh.Clear();
        return FALSE;
    }

    Nmag = sqrt( x*x/(r*r*r*r) + y*y/(a*a*a*a) + z*z/(b*b*b*b) );
    Nx = x/(r*r*Nmag);
    Ny = y/(a*a*Nmag);
    Nz = z/(b*b*Nmag);

    x += Nx * dr_cent - r;
    y += Ny * dr_cent;
    z += Nz * dr_cent;

    tmp = y * cos(alpha) - z * sin(alpha);
    z   = y * sin(alpha) + z * cos(alpha);
    y   = tmp;
    tmp = Ny * cos(alpha) - Nz * sin(alpha);
    Nz  = Ny * sin(alpha) + Nz * cos(alpha);
    Ny  = tmp;

    aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
    aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));

    /* Generate the first 1-ring edge point (which can include error) */

    x = r * cos( phi + dphi_edge );
    y = r * sin( phi + dphi_edge ) * cos( dtheta );
    z = r * sin( phi + dphi_edge ) * sin( dtheta );

    radical = 1.0 - ( y*y/(a*a) + z*z/(b*b) );
    if( radical >= 0.0 ) {
        x = r * sqrt( radical );
    } else {
        printf( "Radical out of range: %lf\n", radical );
        new_mesh.Clear();
        return FALSE;
    }

    Nmag = sqrt( x*x/(r*r*r*r) + y*y/(a*a*a*a) + z*z/(b*b*b*b) );
    Nx = x/(r*r*Nmag);
    Ny = y/(a*a*Nmag);
    Nz = z/(b*b*Nmag);

    x += Nx * dr_edge - r;
    y += Ny * dr_edge;
    z += Nz * dr_edge;

    tmp = y * cos(alpha) - z * sin(alpha);
    z   = y * sin(alpha) + z * cos(alpha);
    y   = tmp;
    tmp = Ny * cos(alpha) - Nz * sin(alpha);
    Nz  = Ny * sin(alpha) + Nz * cos(alpha);
    Ny  = tmp;

    aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
    aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));

    /* Make the remainder of the first ring of points (no errors) */

    for( i=1; i<n; i++ ) {
        theta = i * 2.0 * PI / (double) n;
        x = r * cos( phi );
        y = r * sin( phi ) * cos( theta );
        z = r * sin( phi ) * sin( theta );

        radical = 1.0 - ( y*y/(a*a) + z*z/(b*b) );
        if( radical >= 0.0 ) {
            x = r * sqrt( radical );
        } else {
            printf( "Radical out of range: %lf\n", radical );
            new_mesh.Clear();
            return FALSE;
        }

        Nmag = sqrt( x*x/(r*r*r*r) + y*y/(a*a*a*a) + z*z/(b*b*b*b) );
        Nx = x/(r*r*Nmag);
        Ny = y/(a*a*Nmag);
        Nz = z/(b*b*Nmag);
        x -= r;

        tmp = y * cos(alpha) - z * sin(alpha);
        z   = y * sin(alpha) + z * cos(alpha);
        y   = tmp;
        tmp = Ny * cos(alpha) - Nz * sin(alpha);
        Nz  = Ny * sin(alpha) + Nz * cos(alpha);
        Ny  = tmp;

        aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
        aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));
    }

    /* Make additional rings of points (no errors) */

    npts = n;
    for( j=2; j<=nRings; j++ ) {
        beta = j*phi;
        if( j<pointGrowthLimit ) npts *= 2;
        for( i=0; i<npts; i++ ) {
            theta = i*2.0*PI/(double)npts;
            if( j>=pointGrowthLimit ) {
                theta -= (1+j-pointGrowthLimit)*PI/(double)npts;
            }
            x = r * cos( beta );
            y = r * sin( beta ) * cos( theta );
            z = r * sin( beta ) * sin( theta );
            radical = 1.0 - ( y*y/(a*a) + z*z/(b*b) );
            if( radical >= 0.0 ) {
                x = r * sqrt( radical );
            } else {
                printf( "Radical out of range: %lf\n", radical );
                new_mesh.Clear();
                return FALSE;
            }

            Nmag = sqrt( x*x/(r*r*r*r) + y*y/(a*a*a*a) + z*z/(b*b*b*b) );
            Nx = x/(r*r*Nmag);
            Ny = y/(a*a*Nmag);
            Nz = z/(b*b*Nmag);
            x -= r;

            tmp = y * cos(alpha) - z * sin(alpha);
            z   = y * sin(alpha) + z * cos(alpha);
            y   = tmp;
            tmp = Ny * cos(alpha) - Nz * sin(alpha);
            Nz  = Ny * sin(alpha) + Nz * cos(alpha);
            Ny  = tmp;

           aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
           aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));
        }
    }

    /* Generate faces */
    for( i=0; i<n; i++ ) {
        aonv.need(0);
        aonv += aoVs[0];
        aonv += aoVs[1+i];
        aonv += aoVs[1+((i+1)%n)];
        if( new_mesh.IsOkToAdd( aonv ) == 0 )
            printf("Error adding face\n");
        aoFs += new_mesh.AddFace(aonv);
    }

    offset = 1;
    npts = n;
    for( j=1; j<nRings; j++ ) {
        for( i=0; i<npts; i++ ) {
            if( j<pointGrowthLimit-1 ) {
                i1 = offset + i;
                i2 = offset + (i+1)%npts;
                i3 = offset + npts + 2*i;
                i4 = i3 + 1;
                i5 = offset + npts + (2*i+2)%(2*npts);

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i3];
                aonv += aoVs[i4];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i4];
                aonv += aoVs[i2];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);

                aonv.need(0);
                aonv += aoVs[i2];
                aonv += aoVs[i4];
                aonv += aoVs[i5];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);
            } else {
                i1 = offset + i;
                i2 = offset + (i+1)%npts;
                i3 = i1 + npts;
                i4 = i2 + npts;

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i3];
                aonv += aoVs[i4];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i4];
                aonv += aoVs[i2];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);
            }
        }
        offset += npts;
        if( j<pointGrowthLimit-1 ) npts *= 2;
    }

    return TRUE;
}

/***************** MakeSaddleNRing ***********************************
  This routine generates a mesh for a N-ring neighborhood around a 
  point on a saddle shaped surface
  *******************************************************************/
WINbool
MakeSaddleNRing(PMeshLite &new_mesh,
                const double r, const double a, const double b,
                const int n, const double phi,
                const double dr_edge, const double dr_cent,
                const double dphi_edge, const double dphi_cent,
                const double dtheta, const int nRings )
{
    int i, j, npts, offset, i1, i2, i3, i4, i5, pointGrowthLimit;
    double x, y, z, PI, Nmag, Nx, Ny, Nz;
    double ydenom, zdenom, alpha, beta, theta, tmp;
    Array< Vertex > aoVs;
    Array< Vertex > aonv;
    Array< Face > aoFs;

    int verbose = 0;

    new_mesh.Clear();

    /* Eqn: x - (A - sqrt( A^2 - y^2 )) + (B - sqrt( B^2 - z^2 )) = 0 */

    PI = 4.0 * atan( 1.0 );
    alpha = 0.25 * PI;
    pointGrowthLimit = 4;

    /* Print theoretical curvature */

    float k1, k2, kbar, kg;
    if( a*a < b*b ) {
        k1 = (float) (1.0 / a);
        k2 = (float) (-1.0 / b);
    } else {
        k1 = (float) (-1.0 / b);
        k2 = (float) (1.0 / a);
    }
    kbar = (float) (0.5 * ( k1 + k2 ));
    kg = k1 * k2;

    if( verbose > 0 ) {
        printf("\nSaddle: Target Curvature K1 = %f, K2 = %f, Kbar = %f, Kg = %f\n",
               k1, k2, kbar, kg );
    }

    /* Generate the center point (which can include error) */

    x = r * cos( dphi_cent );
    y = r * sin( dphi_cent );
    z = 0.0;

    if( y*y >= a*a ) {
        printf( "Error: y(%lf) > a(%lf)\n", y, a );
        return FALSE;
    }
    ydenom = sqrt( a*a - y*y );
    if( z*z >= b*b ) {
        printf( "Error: z(%lf) > b(%lf)\n", z, b );
        return FALSE;
    }
    zdenom = sqrt( b*b - z*z );
    x = a - b + zdenom - ydenom;

    Nmag = sqrt( 1.0 + y*y/(ydenom*ydenom) + z*z/(zdenom*zdenom) );
    Nx = 1.0/Nmag;
    Ny = -y/(ydenom*Nmag);
    Nz =  z/(zdenom*Nmag);

    x += Nx * dr_cent;
    y += Ny * dr_cent;
    z += Nz * dr_cent;

    tmp = y * cos(alpha) - z * sin(alpha);
    z   = y * sin(alpha) + z * cos(alpha);
    y   = tmp;
    tmp = Ny * cos(alpha) - Nz * sin(alpha);
    Nz  = Ny * sin(alpha) + Nz * cos(alpha);
    Ny  = tmp;

    aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
    aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));

    /* Generate the first 1-ring edge point (which can include error) */

    x = r * cos( phi + dphi_edge );
    y = r * sin( phi + dphi_edge ) * cos( dtheta );
    z = r * sin( phi + dphi_edge ) * sin( dtheta );

    if( y*y >= a*a ) {
        printf( "Error: y(%lf) > a(%lf)\n", y, a );
        return FALSE;
    }
    ydenom = sqrt( a*a - y*y );
    if( z*z >= b*b ) {
        printf( "Error: z(%lf) > b(%lf)\n", z, b );
        return FALSE;
    }
    zdenom = sqrt( b*b - z*z );
    x = a - b + zdenom - ydenom;

    Nmag = sqrt( 1.0 + y*y/(ydenom*ydenom) + z*z/(zdenom*zdenom) );
    Nx = 1.0/Nmag;
    Ny = -y/(ydenom*Nmag);
    Nz =  z/(zdenom*Nmag);

    x += Nx * dr_edge;
    y += Ny * dr_edge;
    z += Nz * dr_edge;

    tmp = y * cos(alpha) - z * sin(alpha);
    z   = y * sin(alpha) + z * cos(alpha);
    y   = tmp;
    tmp = Ny * cos(alpha) - Nz * sin(alpha);
    Nz  = Ny * sin(alpha) + Nz * cos(alpha);
    Ny  = tmp;

    aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
    aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));

    /* Make the remainder of the first ring of points (no errors) */

    for( i=1; i<n; i++ ) {
        theta = i * 2.0 * PI / (double) n;
        x = r * cos( phi );
        y = r * sin( phi ) * cos( theta );
        z = r * sin( phi ) * sin( theta );

        if( y*y >= a*a ) {
            printf( "Error: y(%lf) > a(%lf)\n", y, a );
            return FALSE;
        }
        ydenom = sqrt( a*a - y*y );
        if( z*z >= b*b ) {
            printf( "Error: z(%lf) > b(%lf)\n", z, b );
            return FALSE;
        }
        zdenom = sqrt( b*b - z*z );
        x = a - b + zdenom - ydenom;

        Nmag = sqrt( 1.0 + y*y/(ydenom*ydenom) + z*z/(zdenom*zdenom) );
        Nx = 1.0/Nmag;
        Ny = -y/(ydenom*Nmag);
        Nz =  z/(zdenom*Nmag);

        tmp = y * cos(alpha) - z * sin(alpha);
        z   = y * sin(alpha) + z * cos(alpha);
        y   = tmp;
        tmp = Ny * cos(alpha) - Nz * sin(alpha);
        Nz  = Ny * sin(alpha) + Nz * cos(alpha);
        Ny  = tmp;

        aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
        aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));
    }

    /* Make additional rings of points */
    npts = n;
    for( j=2; j<=nRings; j++ ) {
        beta = j*phi;
        if( j<pointGrowthLimit ) npts *= 2;
        for( i=0; i<npts; i++ ) {
            theta = i*2.0*PI/(double)npts;
            if( j>=pointGrowthLimit ) {
                theta -= (1+j-pointGrowthLimit)*PI/(double)npts;
            }
            x = r * cos( beta );
            y = r * sin( beta ) * cos( theta );
            z = r * sin( beta ) * sin( theta );
            if( y*y >= a*a ) {
                printf( "Error: y(%lf) > a(%lf)\n", y, a );
                return FALSE;
            }
            ydenom = sqrt( a*a - y*y );
            if( z*z >= b*b ) {
                printf( "Error: z(%lf) > b(%lf)\n", z, b );
                return FALSE;
            }
            zdenom = sqrt( b*b - z*z );
            x = a - b + zdenom - ydenom;

            Nmag = sqrt( 1.0 + y*y/(ydenom*ydenom) + z*z/(zdenom*zdenom) );
            Nx = 1.0/Nmag;
            Ny = -y/(ydenom*Nmag);
            Nz =  z/(zdenom*Nmag);

            tmp = y * cos(alpha) - z * sin(alpha);
            z   = y * sin(alpha) + z * cos(alpha);
            y   = tmp;
            tmp = Ny * cos(alpha) - Nz * sin(alpha);
            Nz  = Ny * sin(alpha) + Nz * cos(alpha);
            Ny  = tmp;

            aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
            aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));
        }
    }

    /* Generate faces */
    for( i=0; i<n; i++ ) {
        aonv.need(0);
        aonv += aoVs[0];
        aonv += aoVs[1+i];
        aonv += aoVs[1+((i+1)%n)];
        if( new_mesh.IsOkToAdd( aonv ) == 0 )
            printf("Error adding face\n");
        aoFs += new_mesh.AddFace(aonv);
    }

    offset = 1;
    npts = n;
    for( j=1; j<nRings; j++ ) {
        for( i=0; i<npts; i++ ) {
            if( j<pointGrowthLimit-1 ) {
                i1 = offset + i;
                i2 = offset + (i+1)%npts;
                i3 = offset + npts + 2*i;
                i4 = i3 + 1;
                i5 = offset + npts + (2*i+2)%(2*npts);

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i3];
                aonv += aoVs[i4];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i4];
                aonv += aoVs[i2];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);

                aonv.need(0);
                aonv += aoVs[i2];
                aonv += aoVs[i4];
                aonv += aoVs[i5];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);
            } else {
                i1 = offset + i;
                i2 = offset + (i+1)%npts;
                i3 = i1 + npts;
                i4 = i2 + npts;

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i3];
                aonv += aoVs[i4];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i4];
                aonv += aoVs[i2];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);
            }
        }
        offset += npts;
        if( j<pointGrowthLimit-1 ) npts *= 2;
    }

    return TRUE;
}

/***************** MakeSphericalNRing ***********************************
  This routine generates a mesh for a N-ring neighborhood around a 
  point on a sphere
  *******************************************************************/
WINbool
MakeSphericalNRing(PMeshLite &new_mesh,
                   const double r, const int n, const double phi,
                   const double dr_edge, const double dr_cent,
                   const double dphi_edge, const double dphi_cent,
                   const double dtheta, const int nRings )
{
    int i, j, npts, offset, i1, i2, i3, i4, i5, pointGrowthLimit;
    double x, y, z, PI, Nx, Ny, Nz, alpha, beta, theta, tmp;
    Array< Vertex > aoVs;
    Array< Vertex > aonv;
    Array< Face > aoFs;

    int verbose = 0;

    new_mesh.Clear();

    PI = 4.0 * atan( 1.0 );
    alpha = 0.25 * PI;
    pointGrowthLimit = 4;

    /* Print theoretical curvature */

    float k1, k2, kbar, kg;
    k1 = (float) (1.0 / r);
    k2 = k1;
    kbar = (float) (0.5 * ( k1 + k2 ));
    kg = k1 * k2;

    if( verbose > 0 ) {
        printf("\n%s: Target Curvature K1 = %f, K2 = %f, Kbar = %f, Kg = %f\n",
                "Sphere", k1, k2, kbar, kg );
    }

    /* Generate the center point (which can include error) */

    x = ( r + dr_cent ) * cos( dphi_cent );
    y = ( r + dr_cent ) * sin( dphi_cent );
    z = 0.0;
    Nx = cos( dphi_cent );
    Ny = sin( dphi_cent );
    Nz = 0.0;
    x -= r;

    tmp = y * cos(alpha) - z * sin(alpha);
    z   = y * sin(alpha) + z * cos(alpha);
    y   = tmp;
    tmp = Ny * cos(alpha) - Nz * sin(alpha);
    Nz  = Ny * sin(alpha) + Nz * cos(alpha);
    Ny  = tmp;

    aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
    aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));

    /* Generate the first 1-ring edge point (which can include error) */

    x = ( r + dr_edge ) * cos( phi + dphi_edge );
    y = ( r + dr_edge ) * sin( phi + dphi_edge ) * cos( dtheta );
    z = ( r + dr_edge ) * sin( phi + dphi_edge ) * sin( dtheta );
    Nx = cos( phi + dphi_edge );
    Ny = sin( phi + dphi_edge ) * cos( dtheta );
    Nz = sin( phi + dphi_edge ) * sin( dtheta );
    x -= r;

    tmp = y * cos(alpha) - z * sin(alpha);
    z   = y * sin(alpha) + z * cos(alpha);
    y   = tmp;
    tmp = Ny * cos(alpha) - Nz * sin(alpha);
    Nz  = Ny * sin(alpha) + Nz * cos(alpha);
    Ny  = tmp;

    aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
    aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));

    /* Make the remainder of the first ring of points */

    for( i=1; i<n; i++ ) {
        theta = i*2.0*PI/(double)n;
        x = r * cos( phi );
        y = r * sin( phi ) * cos( theta );
        z = r * sin( phi ) * sin( theta );
        Nx = cos( phi );
        Ny = sin( phi ) * cos( theta );
        Nz = sin( phi ) * sin( theta );
        x -= r;

        tmp = y * cos(alpha) - z * sin(alpha);
        z   = y * sin(alpha) + z * cos(alpha);
        y   = tmp;
        tmp = Ny * cos(alpha) - Nz * sin(alpha);
        Nz  = Ny * sin(alpha) + Nz * cos(alpha);
        Ny  = tmp;

        aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
        aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));
    }

    /* Make additional rings of points (no errors) */

    npts = n;
    for( j=2; j<=nRings; j++ ) {
        beta = j*phi;
        if( j<pointGrowthLimit ) npts *= 2;
        for( i=0; i<npts; i++ ) {
            theta = i*2.0*PI/(double)npts;
            if( j>=pointGrowthLimit ) {
                theta -= (1+j-pointGrowthLimit)*PI/(double)npts;
            }
            x = r * cos( beta );
            y = r * sin( beta ) * cos( theta );
            z = r * sin( beta ) * sin( theta );
            Nx = cos( beta );
            Ny = sin( beta ) * cos( theta );
            Nz = sin( beta ) * sin( theta );
            x -= r;

            tmp = y * cos(alpha) - z * sin(alpha);
            z   = y * sin(alpha) + z * cos(alpha);
            y   = tmp;
            tmp = Ny * cos(alpha) - Nz * sin(alpha);
            Nz  = Ny * sin(alpha) + Nz * cos(alpha);
            Ny  = tmp;

            aoVs += new_mesh.AddVertex( R3Pt(y,z,x) );
            aoVs.last()->SetNorm(R3Vec( Ny, Nz, Nx ));
        }
    }

    /* Generate faces */
    for( i=0; i<n; i++ ) {
        aonv.need(0);
        aonv += aoVs[0];
        aonv += aoVs[1+i];
        aonv += aoVs[1+((i+1)%n)];
        if( new_mesh.IsOkToAdd( aonv ) == 0 )
            printf("Error adding face\n");
        aoFs += new_mesh.AddFace(aonv);
    }

    offset = 1;
    npts = n;
    for( j=1; j<nRings; j++ ) {
        for( i=0; i<npts; i++ ) {
            if( j<pointGrowthLimit-1 ) {
                i1 = offset + i;
                i2 = offset + (i+1)%npts;
                i3 = offset + npts + 2*i;
                i4 = i3 + 1;
                i5 = offset + npts + (2*i+2)%(2*npts);

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i3];
                aonv += aoVs[i4];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i4];
                aonv += aoVs[i2];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);

                aonv.need(0);
                aonv += aoVs[i2];
                aonv += aoVs[i4];
                aonv += aoVs[i5];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);
            } else {
                i1 = offset + i;
                i2 = offset + (i+1)%npts;
                i3 = i1 + npts;
                i4 = i2 + npts;

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i3];
                aonv += aoVs[i4];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);

                aonv.need(0);
                aonv += aoVs[i1];
                aonv += aoVs[i4];
                aonv += aoVs[i2];
                if( new_mesh.IsOkToAdd( aonv ) == 0 )
                    printf("Error adding face\n");
                aoFs += new_mesh.AddFace(aonv);
            }
        }
        offset += npts;
        if( j<pointGrowthLimit-1 ) npts *= 2;
    }

    return TRUE;
}
