#include "StdAfx.H"

#include <utils/Pm_MeshLite.H>
#include <boost/dynamic_bitset/dynamic_bitset.hpp>
#include <utils/Utils_Dots.H>

void PMeshLite::GrowShortestPathTree( const Vertex                  & in_v1,
                                      const boost::dynamic_bitset<> & in_abNoCross,
                                      const Array<double>           & in_adEdgeWeights,
                                      Array<Vertex>           & out_avFrom,
                                      Array<double>           & out_adDist,
                                      boost::dynamic_bitset<> & out_abTreeEdges, 
                                      const double              in_dStopGrow ) const
{
    const int nV = NumVertices();
    const int nE = NumEdges();
    Array<Vertex> avMesh( nV );

    // do initializations
    out_adDist.need( nV );
    out_adDist.fill( in_dStopGrow );
    out_avFrom.need( nV );
    out_avFrom.memfill( NULL );
    out_abTreeEdges.resize(nE);
    out_abTreeEdges.reset();
    AllMeshVertices(avMesh);

    // initialize the source node info
    out_adDist[ in_v1->IndexId() ] = 0.0;
    out_avFrom[ in_v1->IndexId() ] = in_v1;

    // initialize the priority queue
    BHPqueue pq( nV );
    for (int i = 0; i < nV; i++) {
        pq.enterUnsorted((Univ) avMesh[i], out_adDist[i]);
    }
    pq.sort();

    // do dijkstra's
    Array<Vertex> avAdj;
    Array<Edge> aeAdj;
    while(!pq.empty()) {
        Vertex u = (Vertex)pq.removemin();
        int uid = u->IndexId();

        if ( out_adDist[uid] > in_dStopGrow )
            break;

        if ( !in_abNoCross[uid] ) {
            avAdj = VertexVertices(u);
            aeAdj = VertexEdges(u);
            for (int iV = 0; iV < avAdj.num(); iV++) {
                Vertex v = avAdj[iV];
                int vid = v->IndexId();
                int eid = aeAdj[iV]->IndexId();

                if (out_adDist[vid] > out_adDist[uid] + in_adEdgeWeights[eid] ) {
                    out_adDist[vid] = out_adDist[uid] + in_adEdgeWeights[eid];
                    pq.update((Univ)v, out_adDist[vid]);
                    if ( out_avFrom[vid] != NULL ) 
                        out_abTreeEdges.reset( EdgeFromVertices(out_avFrom[vid], v)->IndexId() );
                    out_abTreeEdges.set(eid);
                    out_avFrom[vid] = u;
                } 
            }
        }
    }
}

// same as above, only it goes on the faces instead of the vertices.
// this code gives the minimum spanning tree, but when I call it I tweak the 
// input weights so that it functions as a maximum spanning tree
void PMeshLite::DualMST( const Face                    & in_f1,
                         const boost::dynamic_bitset<> & in_abNoCross,
                         const Array<double>           & in_adEdgeWeights,
                               boost::dynamic_bitset<> & out_abTreeEdges) const
{
    const double dMax = 1e30;
    const int nF = NumFaces();
    const int nE = NumEdges();
    Array<Face> afMesh( nF );
    Array<Face> afFrom( nF );        // the predecessors of each face
    Array<double> adDist( nF );      // the distance from the source to each face

    // do initializations
    adDist.fill( dMax );
    afFrom.memfill( NULL );
    out_abTreeEdges.resize( nE );
    out_abTreeEdges.reset();
    AllMeshFaces(afMesh);
    
    // initialize the source node info
    adDist[ in_f1->IndexId() ] = 0.0;
    afFrom[ in_f1->IndexId() ] = in_f1;

    // initialize the priority queue
    BHPqueue pq( nF );
    for (int i = 0; i < nF; i++) {
        pq.enterUnsorted((Univ) afMesh[i], adDist[i]);
    }
    pq.sort();

    // do dijkstra's
    while(!pq.empty()) {
        Face f = (Face)pq.removemin();
        int fid = f->IndexId();

        Array<Edge> aeAdj = FaceEdges(f);
        for (int iE = 0; iE < aeAdj.num(); iE++) {
            Edge edgeToCross = aeAdj[iE];
            if ( !in_abNoCross[edgeToCross->IndexId()] ) {
                Face g = EdgeOtherFace(edgeToCross, f);
                int gid = g->IndexId();

                if (adDist[gid]   >   adDist[fid] + in_adEdgeWeights[ edgeToCross->IndexId() ]) {
                    adDist[gid] = adDist[fid] + in_adEdgeWeights[ edgeToCross->IndexId() ];

                    pq.update((Univ)g, adDist[gid]);
                    if (afFrom[gid] != NULL) {
                        Edge oldCrossedEdge = EdgeFromFaces(afFrom[gid], g);
                        out_abTreeEdges.reset(oldCrossedEdge->IndexId());
                    }
                    out_abTreeEdges.set(edgeToCross->IndexId());
                    afFrom[gid] = f;
                } 
            }
        }

    }
}


Array<Vertex> PMeshLite::FindPath( const Vertex                   & in_v1, 
                                   const Array<Vertex>            & in_aopBdryVs, 
                                   const boost::dynamic_bitset<>  & in_abNoCross, 
                                   const Array< double >          & in_adEdgeWeights  ) const
{
    const int nV = NumVertices();
    const int nE = NumEdges();
    Array<Vertex> avFrom( nV );   // the predecessors of each vertex
    Array<double> adDist( nV );   // the distance from the source to each vertex
    boost::dynamic_bitset<> abTreeEdges( nE );  // true if edge is in the tree

    GrowShortestPathTree(in_v1, in_abNoCross, in_adEdgeWeights, avFrom, adDist, abTreeEdges);

    // find the dest vertex with the smallest distance
    int iMin = 0;
    for ( int i = 0; i < in_aopBdryVs.num(); i++ ) {
        if ( adDist[ in_aopBdryVs[i]->IndexId() ] < adDist[ in_aopBdryVs[iMin]->IndexId() ] )
            iMin = i;
    }

    // build up the array of vertices to return, starting with the dest vertex found above
    Array<Vertex> avRet;
    avRet.need(0);
    avRet += in_aopBdryVs[iMin];
    while ( avFrom.Ok(avRet.last()->IndexId()) && avFrom[ avRet.last()->IndexId() ] != in_v1 ) {
        avRet += avFrom[ avRet.last()->IndexId() ];
    }
    avRet += in_v1;
    avRet.reverse();

    return avRet;
}


Array<Vertex> PMeshLite::FindPath( const Vertex &in_v1, 
                                   const Vertex &in_v2, 
                                   const boost::dynamic_bitset<> &in_abEdgeNoCross, 
                                   const Array< double >         &in_adEdgeWeights ) const
{
    Array<Vertex> av;
    av += in_v2;
    return FindPath( in_v1, av, in_abEdgeNoCross, in_adEdgeWeights );
}

Array<Vertex> PMeshLite::FindPath( const Vertex &in_v1, 
                                   const Vertex &in_v2, 
                                   const Array< double > &in_adEdgeWeights ) const
{
    Array<Vertex> av;
    av += in_v2;

    boost::dynamic_bitset<> abNoCross( NumVertices() );
    abNoCross.reset( );

    return FindPath( in_v1, av, abNoCross, in_adEdgeWeights );
}

Array<Vertex> PMeshLite::FindPath( const Vertex &in_v1, 
                                   const Vertex &in_v2 ) const
{
    Array<Vertex> av;
    av += in_v2;

    boost::dynamic_bitset<> abNoCross( NumVertices() );
    abNoCross.reset( );

    Array<double> adEdgeLen( NumEdges() );
    adEdgeLen.fill( 1 );

    return FindPath( in_v1, av, abNoCross, adEdgeLen );
}

Array<Vertex> PMeshLite::FindPath( const Vertex &in_v1, 
                                   const Array<Vertex> &in_av2, 
                                   const Array< double > &in_adEdgeWeights ) const
{
    boost::dynamic_bitset<> abNoCross( NumVertices() );
    abNoCross.reset( );

    return FindPath( in_v1, in_av2, abNoCross, in_adEdgeWeights );
}

Array<Vertex> PMeshLite::FindPath( const Vertex &in_v1, 
                                   const Vertex &in_v2, 
                                   const boost::dynamic_bitset<> &in_abEdgeNoCross ) const
{
    Array<double> adEdgeLen( NumEdges() );
    adEdgeLen.fill( 1 );

    Array<Vertex> av;
    av += in_v2;
    return FindPath( in_v1, av, in_abEdgeNoCross, adEdgeLen );
}

Array<Vertex> PMeshLite::FindPath( const Vertex &in_v1, 
                                   const Array<Vertex> &in_av, 
                                   const boost::dynamic_bitset<> &in_abEdgeNoCross ) const
{
    Array<double> adEdgeLen( NumEdges() );
    adEdgeLen.fill( 1 );

    return FindPath( in_v1, in_av, in_abEdgeNoCross, adEdgeLen );
}

Array<Vertex> PMeshLite::FindPath( const Vertex &in_v1, 
                                   const Array<Vertex> &in_av ) const
{
    boost::dynamic_bitset<> abNoCross( NumVertices() );
    abNoCross.reset( );

    Array<double> adEdgeLen( NumEdges() );
    adEdgeLen.fill( 1 );

    return FindPath( in_v1, in_av, abNoCross, adEdgeLen );
}

///
const Array<double> PMeshLite::EdgeLengths() const
{
    Array<double> adEdgeLengths( NumEdges() );

    ForMeshEdge( Mesh(), e ) {
        adEdgeLengths[ e->IndexId() ] = GetEdgeLength( e );
    } EndFor;

    return adEdgeLengths;
}



WINbool PMeshLite::FindPath( const Face in_oF1,
                             const Face in_oF2,
                             const boost::dynamic_bitset<> & in_abEdgeNoCross,
                             Array<Face >                  & out_aopFs ) const
{
    Array<Face> afMesh( NumFaces() );
    AllMeshFaces(afMesh);

    const int iMax = 10 * NumEdges();
    Array<Face> aopFFrom( NumFaces() );
    Array<int> aiDist( NumFaces() );
    Array< Array<Face> > aafNeigh( NumFaces() );
    for ( int i = 0; i < aiDist.num(); i++ ) {
        aiDist[i] = iMax;
        aopFFrom[i] = NULL;

        ASSERT( i == afMesh[i]->IndexId() );
        Array<Edge> aopEs = FaceEdges( afMesh[i] );
        for ( int iE = 0; iE < aopEs.num(); iE++ ) {
            if ( !in_abEdgeNoCross[ aopEs[iE]->IndexId() ] ) {
                Face f1 = EdgeFace1( aopEs[iE] );
                Face f2 = EdgeFace2( aopEs[iE] );
                if ( f1 == afMesh[i] ) {
                    aafNeigh[i] += f2;
                } else {
                    ASSERT( f2 == afMesh[i] );
                    aafNeigh[i] += f1;
                }
            }
        }
    }

    WINbool bAllFound = FALSE;

    aiDist[ in_oF1->IndexId() ] = 0;
    aopFFrom[ in_oF1->IndexId() ] = in_oF1;
    while ( bAllFound == FALSE ) {
        bAllFound = TRUE;
        for ( int i = 0; i < NumFaces(); i++ ) {
            const int iDist = aiDist[ i ];
            if ( iDist == iMax ) continue;

            const Array<Face> &afNeigh = aafNeigh[i];
            for ( int j = 0; j < afNeigh.num(); j++ ) {
                const int iFId = afNeigh[j]->IndexId();
                if ( iDist + 1 < aiDist[ iFId ] ) {
                    //TRACE("%d owner %d\n", iVId, m_aiVertexOwner[ iVId ] );
                    aopFFrom[iFId] = afMesh[i];
                    aiDist[ iFId ] = iDist + 1;
                    bAllFound = FALSE;
                }
            }
        }
    }

    Face opMe = in_oF2;
    Face opF =  aopFFrom[ in_oF2->IndexId() ];
    Face opHome = in_oF1;
    out_aopFs.need(0);

    if ( aopFFrom[ opMe->IndexId() ] == NULL ) {
        return FALSE;
    }

    //ASSERT( aopVFrom[ in_oV2->IndexId() ] != NULL );

    out_aopFs += opMe;
    while ( ! ( out_aopFs.last() == opHome ) ) {
        out_aopFs += opF;
        opMe = opF;
        opF = aopFFrom[ opF->IndexId() ];
    }

    return TRUE;
}

int PMeshLite::SwapEdges( const Array<double> in_adDist ) {
   int iSwap = 0;

    Array<Edge> aeAll;
    AllMeshEdges( aeAll );

    Array<Face> afNorm;
    for ( int i = 0; i < aeAll.num(); i++ ) {
        if ( IsOkToSwap( aeAll[i] ) == FALSE ) 
            continue;

        Vertex v1 = EdgeVertex1( aeAll[i] );
        Vertex v2 = EdgeVertex2( aeAll[i] );
        Vertex vOp1 = SideVertex1( aeAll[i] );
        Vertex vOp2 = SideVertex2( aeAll[i] );
        Face f1 = EdgeFace1( aeAll[i] );
        Face f2 = EdgeFace2( aeAll[i] );

        const Array<Vertex> avF1 = FaceVertices( f1 );
		  const R3Vec vecF1 = Cross( avF1[1]->GetLoc() - avF1[0]->GetLoc(), avF1[2]->GetLoc() - avF1[0]->GetLoc() );

		  const R3Vec vecFNew1 = Cross( vOp1->GetLoc() - v1->GetLoc(), vOp2->GetLoc() - v1->GetLoc() );
		  const R3Vec vecFNew2 = Cross( vOp2->GetLoc() - v2->GetLoc(), vOp1->GetLoc() - v2->GetLoc() );
	
        const double dDot1 = Dot( vecF1, vecFNew1 );
        const double dDot2 = Dot( vecF1, vecFNew2 );

        if ( dDot1 > 0.0 || dDot2 > 0.0 )
            continue;

        const double dLen = Length( v1->GetLoc() - v2->GetLoc() );
        const double dLenSwap = Length( vOp1->GetLoc() - vOp2->GetLoc() );

        if ( VertexVertices( v1 ).num() == 4 )
            continue;
        if ( VertexVertices( v2 ).num() == 4 )
            continue;

        const double dIdealLen = 0.5 * (in_adDist[v1->IndexId()] + in_adDist[v2->IndexId()]);

        if ( fabs( dLen - dIdealLen ) > fabs( dLenSwap - dIdealLen ) ) {
            EdgeSwap( aeAll[i] );

            v1->SetColor( UTILSColor(1,0,1) );
            v2->SetColor(  UTILSColor(1,0,1) );

            afNorm += f1;
            afNorm += f2;

            iSwap++;
        }
    }

    Array<Vertex> av;
    Compute_normals( av, afNorm );

    return iSwap;
}

int PMeshLite::Grow( const Array<Vertex> &in_aopVs, 
                     Array<Edge> &out_aopEs,
                     Array<Face> &out_aopFs ) const
{
    int iCount = 0;

    boost::dynamic_bitset<> abEs( Mesh().nextEdgeId() ), abFs( Mesh().nextFaceId() );
    abEs.reset();
    abFs.reset();
    for ( int iE = 0; iE < out_aopEs.num(); iE++ )
        abEs.set(out_aopEs[iE]->IndexId());
    for ( int iF = 0; iF < out_aopFs.num(); iF++ )
        abFs.set(out_aopFs[iF]->IndexId());

    for ( int i = 0; i < in_aopVs.num(); i++ ) {
        ForVertexEdge( Mesh(), in_aopVs[i], e ) {
            if ( !abEs[ e->IndexId() ] ) {
                out_aopEs += e;
                abEs.set( e->IndexId() );
                iCount++;
            }
        } EndFor;

        ForVertexFace( Mesh(), in_aopVs[i], f ) {
            if ( !abFs[ f->IndexId() ] ) {
                out_aopFs += f;
                abFs.set( f->IndexId() );
                iCount++;
            }
        }EndFor;
    }
    return iCount;
}

int PMeshLite::Grow( const Array<Face> &  in_aopFs, 
                           Array<Vertex> &out_aopVs,
                           Array<Edge  > &out_aopEs ) const
{
    int iCount = 0;

    boost::dynamic_bitset<> abVs( Mesh().nextVertexId() ), abEs( Mesh().nextEdgeId() );
    abVs.reset();
    abEs.reset();
    for ( int iV = 0; iV < out_aopVs.num(); iV++ )
        abVs.set(out_aopVs[iV]->IndexId());
    for ( int iE = 0; iE < out_aopEs.num(); iE++ )
        abEs.set(out_aopEs[iE]->IndexId());

    for ( int i = 0; i < in_aopFs.num(); i++ ) {
        ForFaceVertex( Mesh(), in_aopFs[i], v ) {
            if ( !abVs[ v->IndexId() ] ) {
                out_aopVs += v;
                abVs.set(v->IndexId());
                iCount++;
            }
        } EndFor;

        ForFaceEdge( Mesh(), in_aopFs[i], e ) {
            if ( !abEs[ e->IndexId() ] ) {
                out_aopEs += e;
                abEs.set( e->IndexId() );
                iCount++;
            }
        }EndFor;
    }
    return iCount;
}
    ///
int PMeshLite::Grow( const Array<Edge> &  in_aopEs, 
                           Array<Vertex> &out_aopVs,
                           Array<Face  > &out_aopFs ) const
{ 
    int iCount = 0;

    boost::dynamic_bitset<> abVs( Mesh().nextVertexId() ), abFs( Mesh().nextFaceId() );
    abVs.reset();
    abFs.reset();
    for ( int iV = 0; iV < out_aopVs.num(); iV++ )
        abVs.set(out_aopVs[iV]->IndexId());
    for ( int iF = 0; iF < out_aopFs.num(); iF++ )
        abFs.set(out_aopFs[iF]->IndexId());

    for ( int i = 0; i < in_aopEs.num(); i++ ) {
        Vertex v1 = EdgeVertex1( in_aopEs[i] );
        Vertex v2 = EdgeVertex2( in_aopEs[i] );
        Face   f1 = EdgeFace1( in_aopEs[i] );
        Face   f2 = EdgeFace2( in_aopEs[i] );

        if ( !abVs[v1->IndexId()] ) {
            out_aopVs += v1;
            abVs.set(v1->IndexId());
            iCount++;
        }

        if ( !abVs[v2->IndexId()] ) {
            out_aopVs += v2;
            abVs.set(v2->IndexId());
            iCount++;
        }

        if ( f1 && !abFs[f1->IndexId()] ) {
            out_aopFs += f1;
            abFs.set( f1->IndexId() );
            iCount++;
        }
        if ( f2 && !abFs[f2->IndexId()] ) {
            out_aopFs += f2;
            abFs.set( f1->IndexId() );
            iCount++;
        }
    }
    return iCount;
}
int PMeshLite::GrowRings( const Vertex &in_vertex, 
                          const int nRings,
                          Array<Vertex> &out_aopVs,
				          Array<Face>   &out_aopFs ) const
{
	int i, j, k, n;
	R3Pt P;
	Array<R3Pt>   apt;
    Array<Vertex> aov;
	Array<Edge>   aopEs;

	out_aopVs.clearcompletely();
	out_aopFs.clearcompletely();
	out_aopVs += in_vertex;
	apt += in_vertex->GetLoc();

	for( i=0; i<nRings; i++ ) {

		// Find adjacent faces for next ring
		n = Grow( out_aopVs, aopEs, out_aopFs );

		// Update list vertices for new ring
		for( j=0; j<out_aopFs.num(); j++ ) {
			aov = FaceVertices(out_aopFs[j]);
			for( k=0; k<aov.num(); k++ ) {
				P = aov[k]->GetLoc();
				const int iIndx = apt.index( P );
				if( iIndx == -1 ) {
					out_aopVs += aov[k];
						apt += P;
				}
			}
		}
	}

	return out_aopFs.num();
}
//might be the same as ExtractCopy below
void PMeshLite::MakeSubMesh( const boost::dynamic_bitset<> &in_abFaces, PMeshLite &newMesh ) const
{
    newMesh.Clear();
    WINbool s_bTrace = FALSE;
    Array<int> aiVNewMesh;
    for ( int fid = in_abFaces.find_first(); fid != -1; fid = in_abFaces.find_next(fid) ) {
        Array<Vertex> avFace = FaceVertices(FaceFromIndexId(fid));
        Array<Vertex> avFaceNew(avFace.num()) ;
        for ( int iV = 0; iV < avFace.num(); iV ++ ) {
            if(aiVNewMesh.index( avFace[iV]->IndexId() ) == -1){
                Vertex v = newMesh.AddVertex(avFace[iV]->GetLoc());
                aiVNewMesh += avFace[iV]->IndexId();
                avFaceNew[iV] = v;
            }
            else
                avFaceNew[iV] = newMesh.VertexFromIndexId(aiVNewMesh.index( avFace[iV]->IndexId() ));         

        }
        newMesh.AddFace(avFaceNew);
    }
    
    //newMesh.Write(WINDataHome("test/chartMesh.m"));
    if( s_bTrace && (newMesh.NumVertices() - newMesh.NumEdges() + newMesh.NumFaces() != 1))
            TRACE("Chart is not a disk\n");

    if(s_bTrace){
        UTILSDotArray dots;
        ForMeshVertex( newMesh.Mesh(), v ) {
            dots.Add( v->GetLoc(), UTILSColor::GREY );
        } EndFor;

	    dots.Write( WINDataHome("test/newMesh.txt") );
        
    }
}

int PMeshLite::ExtractCopy(const Array<Face> &in_aopFs,
					       PMeshLite &newMesh ) const
{
	int i, j;
	R3Pt P;
	Array<R3Pt> apt;
    int newId;

	newMesh.Clear();
	Array<Vertex> aov;
    Array<Vertex> aopVs;
    Array<Vertex> aopF;

	for( i=0; i<in_aopFs.num(); i++ ) {
		aov = FaceVertices(in_aopFs[i]);
		aopF.need(0);
		for( j=0; j<aov.num(); j++ ) {
			P = aov[j]->GetLoc();
			const int iIndx = apt.index( P );
			if ( iIndx == -1 ) {
                newId = apt.num();
                aopVs += newMesh.AddVertex( P );
				// Copy normal
                apt += P;
            } else {
                newId = iIndx;
            }
			aopF += aopVs[newId];
		}
        if ( newMesh.IsOkToAdd( aopF ) )
            newMesh.AddFace( aopF );
        else
            ASSERT(FALSE);
	}
	
    newMesh.Renumber();

	return newMesh.NumVertices();
}

int PMeshLite::ExtractCopy(const Array<Face> &in_aopFs,
					       PMeshLite &newMesh, Array<Vertex> &out_aopVs ) const
{
	int i, j;
	R3Pt P;
	Array<R3Pt> apt;
    int newId;

	newMesh.Clear();
	Array<Vertex> aov;
    Array<Vertex> aopVs;
    Array<Vertex> aopF;

	FILE *out = NULL;
	out = fopen(WINDataHome("extract.dat"),"a");

	out_aopVs.need(0);

	if( out != NULL ) {
		fprintf(out,"\nExtracting %d faces from $d vertices, %d faces\n",
			        in_aopFs.num(), NumVertices(), NumFaces() );
	}

	for( i=0; i<in_aopFs.num(); i++ ) {
		aov = FaceVertices(in_aopFs[i]);
		aopF.need(0);
		for( j=0; j<aov.num(); j++ ) {
			P = aov[j]->GetLoc();
			const int iIndx = apt.index( P );
			if ( iIndx == -1 ) {
				out_aopVs += aov[j];
                newId = apt.num();
                aopVs += newMesh.AddVertex( P );
				// Copy normal
                apt += P;
				if( out != NULL ) {
					fprintf(out,"j=%d, IndexId=%d, NewId=%d (%lf %lf %lf)\n",
						        j, aov[j]->IndexId(), newId, P[0], P[1], P[2] );
				}
            } else {
                newId = iIndx;
            }
			aopF += aopVs[newId];
		}
        if ( newMesh.IsOkToAdd( aopF ) )
            newMesh.AddFace( aopF );
        else
            ASSERT(FALSE);
	}
	
    newMesh.Renumber();

	if(out != NULL ) fclose(out);

	return newMesh.NumVertices();
}

int PMeshLite::GetComponents( Array< unsigned short > &out_ausLabels, Array<int> &out_aiLabels ) const
{
    TRACE("Find components\n");
    out_ausLabels.need( NumVertices() );
    out_ausLabels.memfill( 0 );

    out_aiLabels.need(1);
    out_aiLabels[0] = 0;

    int iTouched = 0;
    Array<int> aiStack;
    Array<Vertex> av;
    AllMeshVertices( av );

    while ( iTouched < NumVertices() ) {
        const int iLabel = out_aiLabels.num();
        out_aiLabels += 1;
        iTouched++;
        for ( int i = 0; i < out_ausLabels.num(); i++ ) {
            if ( out_ausLabels[i] == 0 ) {
                aiStack += i;
                out_ausLabels[i] = iLabel;
                break;
            }
        }
        while ( aiStack.num() ) {
            const int iStack = aiStack.last();
            aiStack.sub(1);
            ForVertexVertex( Mesh(), av[iStack], vv ) {
                if ( out_ausLabels[ vv->IndexId() ] == 0 ) {
                    out_ausLabels[ vv->IndexId() ] = iLabel;
                    aiStack += vv->IndexId();
                    out_aiLabels[ iLabel ]++;
                    iTouched++;
                } else {
                    ASSERT( out_ausLabels[ vv->IndexId() ] == iLabel );
                }
            } EndFor;
        }
    }

    TRACE("Num components %d: ", out_aiLabels.num() - 1 );
    int iKeep = 1;
    for ( int i = 1; i < out_aiLabels.num(); i++ ) {
        TRACE("%d,%d ", i,out_aiLabels[i]);
    }
    TRACE("\n");

    return out_aiLabels.num() - 1;
}


void PMeshLite::GetOneComponent( PMeshLite &out_mesh ) const
{
    Array<unsigned short> ausLabel( NumVertices() );
    Array<int> aiLabels;

    GetComponents( ausLabel, aiLabels );

    int iKeep = 1;
    for ( int i = 1; i < aiLabels.num(); i++ ) {
        if ( aiLabels[iKeep] < aiLabels[i] )
            iKeep = i;
    }

    out_mesh = *this;

    ForMeshVertex( out_mesh.Mesh(), v ) {
        if ( ausLabel[v->IndexId()] != iKeep ) {
            out_mesh.RemoveAdjacentFaces( v );
        } else if ( out_mesh.VertexFaces(v).num() == 0 ) {
            out_mesh.RemoveVertex(v);
        }
    } EndFor;

    ForMeshVertex( out_mesh.Mesh(), v ) {
        if ( ausLabel[v->IndexId()] != iKeep ) {
            out_mesh.RemoveVertex( v );
        } 
    } EndFor;

    out_mesh.Renumber();
}

// Called by TryCorners
void PMeshLite::CheckBoundary( const Array<Vertex> &in_av, const Array<Edge> &in_ae) const
{
    for ( int i = 0; i < in_av.num(); i++ ) {
        Vertex v1 = EdgeVertex1( in_ae[i] );
        Vertex v2 = EdgeVertex2( in_ae[i] );

        ASSERT( (in_av.wrap( i-1 ) == v1 && in_av.wrap( i ) == v2) ||
                (in_av.wrap( i-1 ) == v2 && in_av.wrap( i ) == v1)
            );

    }
}

WINbool PMeshLite::CheckVs( Face                  in_of, 
                             const Array<Vertex > &in_aopVs ) const
{
    const Array<Vertex > &aopFVs = FaceVertices( in_of );
    Array<int> aiIndex( aopFVs.num() );
    for (int i = 0; i < aopFVs.num(); i++ ) {
        aiIndex[i] = in_aopVs.index( aopFVs[i] );
    }
    int iEnd1 = -1, iEnd2 = -1, iJoin = -1;
    WINbool bFoundE1 = FALSE, bFoundE2 = FALSE;
    WINbool bFoundContiguous = FALSE, bFoundOrder = FALSE;
    for ( FORINT i = 0; i < aopFVs.num(); i++ ) {
        if ( aiIndex[i] != -1 && aiIndex.wrap(i+1) != -1 ) {
            if ( aiIndex.wrap(i+2) != -1 ) {
                bFoundContiguous = TRUE;
                const int iNextIndex = (aiIndex[i] - 1 + in_aopVs.num()) % in_aopVs.num();
                const int iNextNextIndex = (aiIndex.wrap(i+1) - 1 + in_aopVs.num()) % in_aopVs.num();
                if ( aiIndex.wrap(i+1) == iNextIndex && aiIndex.wrap(i+2) == iNextNextIndex ) {
                    bFoundOrder = TRUE;
                }
            } else {
                const int iNextIndex = (aiIndex[i] - 1 + in_aopVs.num()) % in_aopVs.num();
                if ( aiIndex.wrap(i+1) != iNextIndex ) {
                    return FALSE;
                }
            }
        } else if ( aiIndex[i] != -1 && aiIndex.wrap(i+1) == -1  ) {
            if ( bFoundE1 == TRUE )
                return FALSE;
            bFoundE1 = TRUE;
        } else if ( aiIndex[i] == -1 && aiIndex.wrap(i-1) != -1 ) {
            if ( bFoundE2 == TRUE )
                return FALSE;
            bFoundE2 = TRUE;
        }
    }

    if ( bFoundContiguous == TRUE && bFoundOrder == TRUE )
        return TRUE;

    if ( bFoundContiguous == TRUE && bFoundOrder == FALSE )
        return FALSE;

    ASSERT( bFoundE1 == TRUE && bFoundE2 == TRUE );

    return TRUE;
}

WINbool PMeshLite::EatWedge( Array<Vertex> & io_avBdry, 
                             Array<Edge>   & io_aeBdry, 
                             boost::dynamic_bitset<>       & io_abInDisk, 
                             const boost::dynamic_bitset<> & in_abNoCross ) const
{

    for ( int iV = 0; iV < io_avBdry.num(); iV++ ) {
        const Face f1 = EdgeFace1( io_aeBdry[iV] );
        const Face f2 = EdgeFace2( io_aeBdry[iV] );

        Face fWedge = NULL;
        // Only one of these should be true
        if ( f1 && !io_abInDisk[f1->IndexId()] ) {
            const Array<Vertex> avF = FaceVertices( f1 );
            const int iIndex = avF.index( io_avBdry[iV] );
            if ( avF.wrap(iIndex - 1) == io_avBdry.wrap(iV+1) && avF.wrap(iIndex + 1) == io_avBdry.wrap(iV-1) ) {
                fWedge = f1;
            }
        }
        if ( f2 && !io_abInDisk[f2->IndexId()] ) {
            const Array<Vertex> avF = FaceVertices( f2 );
            const int iIndex = avF.index( io_avBdry[iV] );
            if ( avF.wrap(iIndex - 1) == io_avBdry.wrap(iV+1) &&
                 avF.wrap(iIndex + 1) == io_avBdry.wrap(iV-1) ) {
                fWedge = f2;
            }
        }
        if ( fWedge ) {
            io_avBdry.del( iV, 1 );
            io_aeBdry.del( iV, 1 );
            io_aeBdry.wrap(iV) = EdgeFromVertices( io_avBdry.wrap(iV-1), io_avBdry.wrap(iV) );
            io_abInDisk.set( fWedge->IndexId() );
            CheckBoundary( io_avBdry, io_aeBdry );
            return TRUE;
        }
    }
    return FALSE;
}

static WINbool s_bTrace = FALSE;

WINbool PMeshLite::GrowDisk( Array<Vertex> & io_avBdry, 
                             Array<Edge>   & io_aeBdry, 
                             boost::dynamic_bitset<> & io_abInDisk, 
                             const boost::dynamic_bitset<> & in_abNoCross ) const
{
    if ( s_bTrace )
        CheckBoundary( io_avBdry, io_aeBdry );

    // Edge i has vertex i and i-1
    Array< Edge > aopVEs;
    Array< Face>  aopVFs;
    boost::dynamic_bitset<> abBdryF( NumFaces() );

    abBdryF.reset();

    // Get all faces touching a boundary vertex that are
    // not already owned.
    for ( int i = 0; i < io_avBdry.num(); i++ ) {
        VertexEdgeFace( io_avBdry[i], aopVEs, aopVFs );
        
        const int iIndx = aopVEs.index( io_aeBdry[i] );

        if ( in_abNoCross[ io_aeBdry[i]->IndexId() ] )
            continue;

        if ( io_abInDisk[ aopVFs[iIndx]->IndexId() ] ) 
            continue;

        if ( abBdryF[ aopVFs[iIndx]->IndexId() ] ) 
            continue;

        abBdryF.set( aopVFs[iIndx]->IndexId() );
    }

    if ( abBdryF.count() == 0 ) {
        if ( s_bTrace )
            TRACE("GrowCap: No grow %d\n", io_abInDisk.count() );
        return FALSE;
    }

    // Need to mark any edges which are no longer boundaries
    // Mark edges which are now boundaries
    // Replace string of old edges with new edges
    // start, stop of replaced edges/vertices on current string
    // start, stop of new edges/vertices in edges/vertices around face
    const int iStop = abBdryF.count();
    int iAdded = 0;
    Array<Vertex> aopFVs, aopNewVs;
    Array<Edge>   aopFEs, aopNewEs;
    Array<int>    aiLabelE;
    for ( FORINT i = abBdryF.find_first(); i != -1; i = abBdryF.find_next(i) ) {
        Face f = FaceFromIndexId(i);

        // Check that the intersection of the face and the boundary chain is contiguous
        if ( CheckVs( f, io_avBdry ) == FALSE ) {
            continue;
        }

        aopFVs = FaceVertices( f );
        aopFEs = FaceEdges( f );
        aiLabelE.need( aopFEs.num() );

        // Walk through the edges of the face
        //   If this edge is on the boundary, get its index
        for ( int j = 0; j < aopFVs.num(); j++ ) {
            int iEIndx = io_aeBdry.index( aopFEs[j] );
            if ( iEIndx != -1 ) {
                // This edge is on the boundary chain
                for ( int k = iEIndx + 1; k < io_aeBdry.num(); k++ ) {
                    if ( io_aeBdry[k] != aopFEs[j] ) 
                        continue;

                    // Look for latest occurance of this face in the edge list
                    Face opF1 = EdgeFace1( io_aeBdry[k] );
                    Face opF2 = EdgeFace2( io_aeBdry[k] );
                    Vertex opV1 = EdgeVertex1( io_aeBdry[k] );
                    if ( opV1 == io_avBdry.wrap( k-1 ) ) {
                        if ( opF2 == f )
                            iEIndx = k;
                    } else {
                        if ( opF1 == f )
                            iEIndx = k;
                    }
                }
            }
            aiLabelE[j] = iEIndx; // Where in old edge bdry
            //TRACE("V %d E %d Label %d\n", aopFVs[j]->IndexId(), aopFEs[j]->IndexId(), aiLabelE[j] );
        }
        int iStartOldBdry = -1, iEndOldBdry = -1;
        int iStartNewBdry = -1, iEndNewBdry = -1;
        int iDel = 0, iAdd = 0;
        for ( FORINT j = 0; j < aiLabelE.num(); j++ ) {
            if ( aiLabelE[j] == -1 && aiLabelE.wrap(j+1) != -1 ) {
                if (iStartOldBdry != -1)
                    ASSERT(FALSE);
                iStartOldBdry = (j+1) % aiLabelE.num();
                iEndNewBdry = j;
            }
            if ( aiLabelE[j] != -1 && aiLabelE.wrap(j+1) == -1 ) {
                if ( iEndOldBdry != -1 ) ASSERT(FALSE);
                iEndOldBdry = j;
                iStartNewBdry = (j+1) % aiLabelE.num();
            } 
            if ( aiLabelE[j] != -1 ) {
                iDel++;
            } else {
                iAdd++;
            }
        }

        if ( iStartOldBdry == -1 && iEndOldBdry == -1 ) {
            int iMin = aiLabelE[0];
            for ( int j = 0; j < aiLabelE.num(); j++ )
                iMin = WINmin( aiLabelE[j], iMin );
            ASSERT( io_avBdry.wrap( iMin - 1 ) == io_avBdry.wrap( iMin + 2 ) );
            io_avBdry.del( (iMin - 1 + io_avBdry.num()) % io_avBdry.num() , 3 );
            io_aeBdry.del( iMin, 3 );
        } else {

            //TRACE("%d %d  %d %d\n", iStartOldBdry, iDel, iStartNewBdry, iAdd );

            const int iKill = aiLabelE[ iEndOldBdry ];
            int iAddAtV = iKill;
            int iAddAtE = iKill;
            if ( iKill + iDel > io_avBdry.num() + 1 ) {
                ASSERT( iDel == 2 );

                io_avBdry.del( iKill, iDel - 1 );
                Array<Vertex > aopNewVs;
                for ( FORINT j = 0; j < iAdd-1; j++ )
                    aopNewVs += aopFVs.wrap( iStartNewBdry + j );
                io_avBdry.add( aopNewVs, 0 );
        
                io_aeBdry.del( iKill, 1 );
                io_aeBdry.del( 0, 1 );
                aopNewEs.need(0);
                for ( FORINT j = 0; j < iAdd; j++ )
                    aopNewEs += aopFEs.wrap( iStartNewBdry + j );
                io_aeBdry.add( aopNewEs, 0  );
            } else {
            //TRACE("Kill %d  AddV %d AddE %d\n", iKill, iAddAtV, iAddAtE );
                if ( iKill + iDel == io_avBdry.num() + 1 ) {
                    iAddAtV = 0;
                    iAddAtE = 0;
                }

                io_avBdry.del( iKill, iDel - 1 );
                aopNewVs.need(0);
                for ( FORINT j = 0; j < iAdd-1; j++ )
                    aopNewVs += aopFVs.wrap( iStartNewBdry + j );
                io_avBdry.add( aopNewVs, iAddAtV );
        
                io_aeBdry.del( iKill, iDel );
                aopNewEs.need(0);
                for ( FORINT j = 0; j < iAdd; j++ )
                    aopNewEs += aopFEs.wrap( iStartNewBdry + j );
                io_aeBdry.add( aopNewEs, iAddAtE  );
            }
        }

        //TRACE("\n");

        ASSERT( io_avBdry.num() == io_aeBdry.num() );

        if ( s_bTrace )
            CheckBoundary( io_avBdry, io_aeBdry );

        io_abInDisk.set( i );
        iAdded++;
    }

    if ( iAdded > 0 ) {
        while ( EatWedge( io_avBdry, io_aeBdry, io_abInDisk, in_abNoCross ) )
            ;
    }
    return (iAdded > 0) ? TRUE : FALSE;
}
int PMeshLite::GrowDiskByFace( const Face &in_face,
                                  boost::dynamic_bitset<> &io_abInDisk,
                                  Array<Vertex> & io_avBdry, 
                                  Array<Edge>   & io_aeBdry,
                                  Array<int>   & out_wedgeEaten) const
{
    WINbool faceAdded = AddFaceToDisk(in_face, io_abInDisk, io_avBdry, io_aeBdry);
    boost::dynamic_bitset<> abNoCross( 0 );
    abNoCross.reset();
    if(faceAdded){
        out_wedgeEaten.need(0);
        boost::dynamic_bitset<> temp_abInDisk = io_abInDisk;
        WINbool wedgeEaten = EatWedge( io_avBdry, io_aeBdry, io_abInDisk, abNoCross );
        for ( int iF = io_abInDisk.find_first(); iF != -1; iF = io_abInDisk.find_next(iF) ) {
            if(!temp_abInDisk[iF])
                out_wedgeEaten += iF;
        }
    }
    return (faceAdded? 1:0) + out_wedgeEaten.num();
}
///given current Vs and Es on the boundary, add the new face to chart
WINbool PMeshLite::AddFaceToDisk( const Face &in_face,
                                  boost::dynamic_bitset<> &io_abInDisk,
                                  Array<Vertex> & io_avBdry, 
                                  Array<Edge>   & io_aeBdry) const
{
   //actually adding the face
    ASSERT(io_avBdry.num()!=0);
    ASSERT(io_aeBdry.num()!=0);
    //CheckBoundary( io_avBdry, io_aeBdry );    

    // Need to mark any edges which are no longer boundaries
    // Mark edges which are now boundaries
    // Replace string of old edges with new edges
    // start, stop of replaced edges/vertices on current string
    // start, stop of new edges/vertices in edges/vertices around face
    Array<Vertex> avFVs, avNewVs;
    Array<Edge>   aeFEs, aeNewEs;
    Array<int>    aiLabelE;  

    avFVs = FaceVertices( in_face );
    aeFEs = FaceEdges( in_face );
    aiLabelE.need( aeFEs.num() );

    // Walk through the edges of the face
    //   If this edge is on the boundary, get its index
    for ( int iE = 0; iE < aeFEs.num(); iE++ ) {
        int iEIndx = io_aeBdry.index( aeFEs[iE] );
        aiLabelE[iE] = iEIndx; // Where in old edge bdry        
    }
    int iStartOldBdry = -1, iEndOldBdry = -1;
    int iStartNewBdry = -1, iEndNewBdry = -1;
    int iDel = 0, iAdd = 0;
    for ( FORINT iE = 0; iE < aiLabelE.num(); iE++ ) {
        if ( aiLabelE[iE] == -1 && aiLabelE.wrap(iE+1) != -1 ) {
            ASSERT(iStartOldBdry == -1);
            iStartOldBdry = (iE+1) % aiLabelE.num();
            iEndNewBdry = iE;
        }
        if ( aiLabelE[iE] != -1 && aiLabelE.wrap(iE+1) == -1 ) {
            ASSERT(iEndOldBdry == -1 );
            iEndOldBdry = iE;
            iStartNewBdry = (iE+1) % aiLabelE.num();
        } 
        if ( aiLabelE[iE] != -1 ) {
            iDel++;
        } else {
            iAdd++;
        }
    }

    if ( iStartOldBdry == -1 && iEndOldBdry == -1 ) {
        int iMin = aiLabelE[0];
        for ( int j = 0; j < aiLabelE.num(); j++ )
            iMin = WINmin( aiLabelE[j], iMin );
        ASSERT( io_avBdry.wrap( iMin - 1 ) == io_avBdry.wrap( iMin + 2 ) );
        io_avBdry.del( (iMin - 1 + io_avBdry.num()) % io_avBdry.num() , 3 );
        io_aeBdry.del( iMin, 3 );
    } else {
        const int iKill = aiLabelE[ iEndOldBdry ];
        int iAddAtV = iKill;
        int iAddAtE = iKill;
        if ( iKill + iDel > io_avBdry.num() + 1 ) {
            ASSERT( iDel == 2 );

            io_avBdry.del( iKill, iDel - 1 );
            Array<Vertex > aopNewVs;
            for ( int j = 0; j < iAdd-1; j++ )
                avNewVs += avFVs.wrap( iStartNewBdry + j );
            io_avBdry.add( aopNewVs, 0 );
    
            io_aeBdry.del( iKill, 1 );
            io_aeBdry.del( 0, 1 );
            aeNewEs.need(0);
            for ( FORINT j = 0; j < iAdd; j++ )
                aeNewEs += aeFEs.wrap( iStartNewBdry + j );
            io_aeBdry.add( aeNewEs, 0  );
        } else {
            if ( iKill + iDel == io_avBdry.num() + 1 ) {
                iAddAtV = 0;
                iAddAtE = 0;
            }

            io_avBdry.del( iKill, iDel - 1 );
            avNewVs.need(0);
            for ( int j = 0; j < iAdd-1; j++ )
                avNewVs += avFVs.wrap( iStartNewBdry + j );
            io_avBdry.add( avNewVs, iAddAtV );
    
            io_aeBdry.del( iKill, iDel );
            aeNewEs.need(0);
            for ( FORINT j = 0; j < iAdd; j++ )
                aeNewEs += aeFEs.wrap( iStartNewBdry + j );
            io_aeBdry.add( aeNewEs, iAddAtE  );
        }
    }
    ASSERT( io_avBdry.num() == io_aeBdry.num() );

    //CheckBoundary( io_avBdry, io_aeBdry );

    io_abInDisk.set(in_face->IndexId());

    return TRUE;
}

///check if the chart will remain a disk after adding this face
//might be duplicated with another method in this class
WINbool PMeshLite::OkayToAdd( const Face &in_of, const Array<Vertex > &in_aopVs ) const
{
    const Array<Vertex > &aopFVs = FaceVertices( in_of );
    Array<int> aiIndex( aopFVs.num() );
    for (int i = 0; i < aopFVs.num(); i++ ) {
        aiIndex[i] = in_aopVs.index( aopFVs[i] );
    }
    int iEnd1 = -1, iEnd2 = -1, iJoin = -1;
    WINbool bFoundE1 = FALSE, bFoundE2 = FALSE;
    WINbool bFoundContiguous = FALSE, bFoundOrder = FALSE;
    for ( FORINT i = 0; i < aopFVs.num(); i++ ) {
        if ( aiIndex[i] != -1 && aiIndex.wrap(i+1) != -1 ) {
            if ( aiIndex.wrap(i+2) != -1 ) {
                bFoundContiguous = TRUE;
                const int iNextIndex = (aiIndex[i] - 1 + in_aopVs.num()) % in_aopVs.num();
                const int iNextNextIndex = (aiIndex.wrap(i+1) - 1 + in_aopVs.num()) % in_aopVs.num();
                if ( aiIndex.wrap(i+1) == iNextIndex && aiIndex.wrap(i+2) == iNextNextIndex ) {
                    bFoundOrder = TRUE;
                }
            } else {
                const int iNextIndex = (aiIndex[i] - 1 + in_aopVs.num()) % in_aopVs.num();
                if ( aiIndex.wrap(i+1) != iNextIndex ) {
                    return FALSE;
                }
            }
        } else if ( aiIndex[i] != -1 && aiIndex.wrap(i+1) == -1  ) {
            if ( bFoundE1 == TRUE )
                return FALSE;
            bFoundE1 = TRUE;
        } else if ( aiIndex[i] == -1 && aiIndex.wrap(i-1) != -1 ) {
            if ( bFoundE2 == TRUE )
                return FALSE;
            bFoundE2 = TRUE;
        }
    }

    if ( bFoundContiguous == TRUE && bFoundOrder == TRUE )
        return TRUE;

    if ( bFoundContiguous == TRUE && bFoundOrder == FALSE )
        return FALSE;

    ASSERT( bFoundE1 == TRUE && bFoundE2 == TRUE );

    return TRUE;    
}

void PMeshLite::MakeTriangular()
{
    Array<Face> af;
    AllMeshFaces(af);

    Array<Vertex> avf, avTri(3);
    for ( int i = 0; i < af.num(); i++ ) {
        avf = FaceVertices( af[i] );
        if ( avf.num() == 3 )
            continue;
        if ( avf.num() == 4 ) {
            RemoveFace( af[i] );
            avTri[0] = avf[0];
            avTri[1] = avf[1];
            avTri[2] = avf[2];
            AddFace( avTri );

            avTri[0] = avf[0];
            avTri[1] = avf[2];
            avTri[2] = avf[3];
            AddFace( avTri );
        } else {
            ASSERT(FALSE);
        }
    }
    Renumber();
}



pair<int,int> PMeshLite::ExtremePoints( const boost::dynamic_bitset<> &in_abVs ) const
{
    int nv = in_abVs.count();
    if(nv == 0 ) return pair<int,int>(-1,-1);

    R3Pt minPt, maxPt, currentPt;
    pair<int,int> extremesX,extremesY,extremesZ;

	int i = in_abVs.find_first(); 
    minPt = VertexFromIndexId(i)->GetLoc();
    maxPt = VertexFromIndexId(i)->GetLoc();
    extremesX = pair<int,int>(i,i);
    extremesY = pair<int,int>(i,i);
    extremesZ = pair<int,int>(i,i);

    for ( i = in_abVs.find_next(i); i != -1; i = in_abVs.find_next(i) ) {
        currentPt = VertexFromIndexId(i)->GetLoc();
        if(minPt[0] > currentPt[0]){
            minPt[0] = currentPt[0];
            extremesX.first = i;
        }
        else if(maxPt[0] < currentPt[0]){
            maxPt[0] = currentPt[0];            
            extremesX.second = i;
        }
        
        if(minPt[1] > currentPt[1]){
            minPt[1] = currentPt[1];            
            extremesY.first = i;
        }
        else if(maxPt[1] < currentPt[1]){
            maxPt[1] = currentPt[1];
            extremesY.second = i;
        }
        
        if(minPt[2] > currentPt[2]){
            minPt[2] = currentPt[2];
            extremesZ.first = i;
        }
        else if(maxPt[2] < currentPt[2]){
            maxPt[2] = currentPt[2];     
            extremesZ.second = i;
        }
    }
    double distX,distY,distZ;
    distX = VertexFromIndexId(extremesX.second)->GetLoc()[0] - VertexFromIndexId(extremesX.first)->GetLoc()[0];
    distY = VertexFromIndexId(extremesY.second)->GetLoc()[1] - VertexFromIndexId(extremesY.first)->GetLoc()[1];
    distZ = VertexFromIndexId(extremesZ.second)->GetLoc()[2] - VertexFromIndexId(extremesZ.first)->GetLoc()[2];
    if(distX >= distY && distX >= distZ )
        return extremesX;
    else if(distY >= distX && distY >= distZ )
        return extremesY;
    else if(distZ >= distX && distZ >= distY )
        return extremesZ;
    else return pair<int,int>(-1,-1);
}


bool PMeshLite::GrowDiskByDijkstra(  const Face &in_face,
                          const int targetFsInDisk,
                          boost::dynamic_bitset<> &out_abFsInDisk,
                          Array<Vertex> & out_avBdry, 
                          Array<Edge>   & out_aeBdry ) const
{    
    out_abFsInDisk.resize( NumFaces() );
    out_abFsInDisk.reset();
    out_abFsInDisk.set( in_face->IndexId() );

    out_avBdry.need(0);
    out_aeBdry.need(0);
    out_avBdry = FaceVertices( in_face );
    out_aeBdry = FaceEdges( in_face );
    CheckBoundary(out_avBdry,out_aeBdry);
    
    Array<int> wedgesEaten(0);
    // initialize the priority queue
    BHPqueue facesInQueue( 0 );

    ForFaceFace( Mesh(), in_face, fNext){
        const int fNextId = fNext->IndexId();
        double dist = Length(GetFaceCentroid(fNext) - GetFaceCentroid(in_face));

        ASSERT ( !out_abFsInDisk[ fNextId ] );
        facesInQueue.enterUnsorted((Univ) fNext, dist );
    }EndFor;  
    facesInQueue.sort();
    
    // do dijkstra's
    while(!facesInQueue.empty() && (int) out_abFsInDisk.count() <  targetFsInDisk) {
        
        const double dClosestDist = facesInQueue.minpriority();
        const Face fClosest = (Face)facesInQueue.removemin();
        const int fClosestId = fClosest->IndexId();

        //skip if face already added
        if(out_abFsInDisk[fClosestId])
            continue;

        int facesAdded = 0;
        if ( OkayToAdd( fClosest, out_avBdry ) ){
            wedgesEaten.need(0);
            //do add and update data
            facesAdded = GrowDiskByFace( fClosest, out_abFsInDisk, out_avBdry, out_aeBdry, wedgesEaten);   
        }

        if(facesAdded > 0){
            ASSERT(out_abFsInDisk[fClosestId]);
            CheckBoundary( out_avBdry, out_aeBdry );        
            
            ForFaceFace( Mesh(), fClosest, fNext){                
                const int fNextId = fNext->IndexId();
            
                if ( out_abFsInDisk[ fNextId ] )
                    continue;

                double dNextDist = dClosestDist + Length(GetFaceCentroid(fNext) - GetFaceCentroid(fClosest));
                if(facesInQueue.contains((Univ)fNext)){
                    if(facesInQueue.retrieve((Univ)fNext) > dNextDist)
                        facesInQueue.update((Univ)fNext, dNextDist);
                }
                else
                    facesInQueue.enter((Univ)fNext, dNextDist);

            }EndFor;
            if(wedgesEaten.num() > 0){
                for(int iW = 0; iW < wedgesEaten.num(); iW++){
                    Face curWedge = FaceFromIndexId(wedgesEaten[iW]);
                    ForFaceFace( Mesh(), curWedge, fNext){
                        const int fNextId = fNext->IndexId();
            
                        if ( out_abFsInDisk[ fNextId ] )
                            continue;

                        double dNextDist = dClosestDist + Length(GetFaceCentroid(fNext) - GetFaceCentroid(curWedge));
                        if(facesInQueue.contains((Univ)fNext)){
                            if(facesInQueue.retrieve((Univ)fNext) > dNextDist)
                                facesInQueue.update((Univ)fNext, dNextDist);
                        }
                        else
                            facesInQueue.enter((Univ)fNext, dNextDist);

                    }EndFor;
                }
            }
            bool bTrace = FALSE;
            
            if ( bTrace && out_abFsInDisk.count() % 500 == 0) {
                UTILSDotArray dots;
                for ( int iV = 0; iV < out_avBdry.num(); iV++ ) 
                    dots.Add( out_avBdry[iV]->GetLoc(), UTILSColor::GREY );

                for ( unsigned int iF = 0; iF < out_abFsInDisk.size(); iF++ ) {
                    if(out_abFsInDisk[iF])                            
                        if(iF == in_face->IndexId())
                            dots.Add( GetFaceCentroid(in_face), UTILSColor::BLACK );
                        else if(iF == fClosest->IndexId())
                            dots.Add( GetFaceCentroid(fClosest), UTILSColor::YELLOW );                        
                        else
                            dots.Add( GetFaceCentroid(FaceFromIndexId(iF)), UTILSColor::RED );                        
                    else {
                        if(facesInQueue.contains((Univ)FaceFromIndexId(iF)))
                            dots.Add( GetFaceCentroid(FaceFromIndexId(iF)), UTILSColor::PURPLE );                            
                    }
                }
                dots.Write(WINDataHome("test/grow_disk.txt"));
            }

        }
        else 
            ASSERT(!out_abFsInDisk[fClosestId]);
    }
    return true;
}


