#include "StdAfx.H"
#include <utils/R2_Contour.H>


double  R2Contour::Area() const 
{
    double dArea = 0.0;
    for (int i = 0; i < m_apolygons.num(); i++)
        dArea += m_apolygons[i].Area();

    return dArea;
}

///
R2Pt R2Contour::Centroid() const
{
    R2Pt ptCenter(0,0);

    for (int i = 0; i < m_apolygons.num(); i++) {
        R2Pt pt = m_apolygons[i].Centroid();
        ptCenter[0] += pt[0];
        ptCenter[1] += pt[1];
    }

    if ( m_apolygons.num() ) {
        ptCenter[0] /= (double) m_apolygons.num();
        ptCenter[1] /= (double) m_apolygons.num();
    }

    return ptCenter;
}

/// Is it on or inside the boundary?
WINbool R2Contour::Inside( const R2Pt &in_pt ) const
{
    for (int i = 0; i < m_apolygons.num(); i++) {
        if ( m_apolygons[i].Inside( in_pt ) == TRUE )
            return TRUE;
    }
    return FALSE;
}

/// Is it within epsilon of the boundary?
WINbool R2Contour::IsBoundary( const R2Pt &in_pt ) const
{
    for (int i = 0; i < m_apolygons.num(); i++) {
        if ( m_apolygons[i].IsBoundary( in_pt ) == TRUE )
            return TRUE;
    }
    return FALSE;
}

///
WINbool R2Contour::IsConvex() const
{
    for (int i = 0; i < m_apolygons.num(); i++) {
        if ( m_apolygons[i].IsConvex( ) == FALSE )
            return FALSE;
    }
    return TRUE;
}

/** Use the winding test to see if the point is inside.
    Assumes the polygon is convex. Counts the number of crossings 
    on a ray from the point to infinity.  */
WINbool R2Contour::Inside_winding(const R2Pt &in_pt) const
{
    for (int i = 0; i < m_apolygons.num(); i++) {
        if ( m_apolygons[i].Inside_winding( in_pt ) == TRUE )
            return TRUE;
    }
    return FALSE;
}

/// Project onto plane 
R3Pt R2Contour::Project( const R2Pt &in_pt ) const
{
    R3Pt pt = R3Pt( in_pt[0], in_pt[1], 0.0 ) + m_vecPlaneNormal * m_dZLevel;

    return pt;
}

/// Closest point on the boundary, if not inside already
R2Pt R2Contour::Closest_pt( const R2Pt &in_pt ) const
{
    Array<double> adDist;
    Array<R2Pt>   aptClosest;

    for (int i = 0; i < m_apolygons.num(); i++) {
        aptClosest += m_apolygons[i].Closest_pt( in_pt );
        adDist += Length( aptClosest.last() - in_pt );
    }
    aptClosest.sort( adDist );

    if ( aptClosest.num() == 0 ) {
        cerr << "R2Contour::Closest_pt no contour\n";
        return in_pt;
    }

    return aptClosest[0];
}

///
WINbool R2Contour::operator==( const R2Contour &in_oContour ) const
{
    if ( m_apolygons.num() != in_oContour.m_apolygons.num() )
        return FALSE;

    for (int i = 0; i < m_apolygons.num(); i++) {
        if ( ( m_apolygons[i] == in_oContour.m_apolygons[i] ) == FALSE )
            return FALSE;
    }
    return TRUE;
    
}

///
R2Contour R2Contour::Intersect(const R2Contour &in_contour) const
{
    R2Contour out_contour;

    R3Matrix mat;
    R3Matrix::MatrixVecToVec( mat, R3Vec(0,0,1), out_contour.PlaneNormal() );

    Array<R3Vec> avec;
    for (int i = 0; i < m_apolygons.num(); i++) {
        for (int j = 0; j < in_contour.m_apolygons.num(); j++) {
            R2Polygon oPoly = m_apolygons[i].Intersect( in_contour.m_apolygons[j] );
            const R2Pt ptCenter = oPoly.Centroid();

            avec.need( oPoly.Num_pts() );
            for ( int k = 0; k < avec.num(); k++ ) {
                const R3Vec vecNorm = mat * R3Vec( oPoly[k][0] - ptCenter[0], oPoly[k][1] - ptCenter[1], 0.0 );
                avec[k] = R3Vec( UnitSafe( vecNorm ) );
            }
            if ( oPoly.Valid() )
                out_contour.Add( oPoly, avec );
        }
    }
    return out_contour;
}

/// Debugging print.
void R2Contour::Print() const 
{
    cout << "N polygons " << m_apolygons.num() << " z level " << 
        m_dZLevel << " height " << m_dWidth << "\n";
    for (int i = 0; i < m_apolygons.num(); i++)
        m_apolygons[i].Print();
}

/// Writes points to file.
void R2Contour::Write ( ofstream &out ) const 
{
    out << "Contour " << m_apolygons.num() << " " << 
        m_dZLevel << "  " << m_dWidth << "\n";
    for (int i = 0; i < m_apolygons.num(); i++)
        m_apolygons[i].Write(out);
    out << "EndContour\n";
}

/// Reads points from file.
WINbool R2Contour::Read  ( ifstream &in )
{
    char strTemp[512];
    in >> strTemp;
    ASSERT( !strcmp( strTemp, "Contour" ) );

    int iN = -1;
    in >> iN;
    m_apolygons.need( iN );

    WINbool bRes = TRUE;
    for (int i = 0; i < m_apolygons.num(); i++)
        if ( m_apolygons[i].Read(in) == FALSE )
            bRes = FALSE;

    in >> strTemp;
    ASSERT( !strcmp( strTemp, "EndContour" ) );

    return bRes;
}


WINbool R2Contour::Test()
{
    return TRUE;
}
