#include "StdAfx.H"
#include <utils/Rn_Sphere.H>


R2Pt R2Sphere::Closest( const R2Pt  & in_p1 ) const
{
    if ( ApproxEqual( in_p1, m_ptCenter ) )
        return m_ptCenter + R2Vec(0,1) * m_dRadius;

    const R2Vec vecAtOrig = in_p1 - m_ptCenter;

    return m_ptCenter + UnitSafe( vecAtOrig ) * (Radius() * (1.0 - RNEpsilon_d));
}

/// Generate evenly spaced points, number determined by io_apt.num
void R2Sphere::Points( Array<R2Pt> &io_apt, const double in_dTS, const double in_dTE ) const
{
    if ( io_apt.num() == 0 )
        io_apt.need( 10 );

    const double dDiv = (in_dTE - in_dTS) / io_apt.num();

    for (int i = 0; i < io_apt.num(); i++) {
        const R2Pt pt( cos( in_dTS + dDiv * i ), sin( in_dTS + dDiv * i ) );
        io_apt[i] = R2Pt( m_ptCenter[0] + m_dRadius * pt[0], m_ptCenter[1] + m_dRadius * pt[1] );
    }
}



/* -------------------------------------------------------------------------
 * DESCR   :	Intersect the line segment with the line
  >     x -> (2 c - 2 b m + 2 d m + 
 
                                     2
>         Sqrt[(-2 c + 2 b m - 2 d m)  - 
 
                    2    2    2            2    2               2
>           4 (1 + m ) (b  + c  - 2 b d + d  - r )]) / (2 (1 + m ))}}
 
 * ------------------------------------------------------------------------- */
int R2Sphere::Intersect(const R2Line   &in_line, 
                        R2Pt &out_p1, 
                        R2Pt &out_p2) const
{
    double b = in_line.Intercept();
    double m = in_line.Slope();
    double r = Radius();
    double c = Center()[0];
    double d = Center()[1];
    
   if (in_line.Vertical() == TRUE) {
       R2Pt e_pt;
       e_pt[0] = b;
       e_pt[1] = Center()[1];
       
       double len = Length(Center() - e_pt);
       if (len < Radius()) {
           out_p1[0] = b;
           out_p2[0] = b;
           out_p1[1] = Y_pos(out_p1[0]);
           out_p2[1] = Y_neg(out_p2[0]);
           return 2;
       } 
       if (RNIsZero(len - Radius())) {
           out_p1 = e_pt;
           return 1;
       } 
       return 0;
   }
   double radical = ((2 * (-c + m * (b - d))) * (2 * (-c + m * (b - d))) -
                     4 * (1 + m*m) * (b*b + c*c - 2 * b*d + d*d - r*r));
   
   if (radical < 0)
       return 0;
   
   if (RNIsZero(radical)) {
       out_p1[0] = (2 * (c - m * (d - b))) / (2 * (1 + m * m));
       out_p1[1] = in_line.Y(out_p1[0]);
       return 1;
   }
   radical = sqrt(radical);
   
   out_p1[0] = (2 * (c - m * (b - d)) + radical) / (2 * (1 + m * m));
   out_p2[0] = (2 * (c - m * (b - d)) - radical) / (2 * (1 + m * m));
   out_p1[1] = in_line.Y(out_p1[0]);
   out_p2[1] = in_line.Y(out_p2[0]);
   
   return 2;
}

int
R2Sphere::Intersect(const R2Line_seg &in_seg,
                    R2Pt &out_p1,
                    R2Pt &out_p2) const
{
    R2Line line(in_seg.P1(), in_seg.P2());
    
    int num = Intersect(line, out_p1, out_p2);
    
    double t, d, scl = 0.001f;
    R2Pt pt;
    
    if (num == 0) return 0;
    if (num == 1) {
        if (in_seg.FindPtOnSeg(out_p1, pt, t, d) == TRUE)
            return 1;
        else if (RNIsZero(d * scl))
            return 1;
        else
            return 0;
    }
    
    in_seg.FindPtOnSeg(out_p1, pt, t, d);
    if (RNIsZero(d * scl)) {
        in_seg.FindPtOnSeg(out_p2, pt, t, d);
        if (RNIsZero(d * scl)) {
            return 2;
        } else {
            return 1; 
        }
    } 
    in_seg.FindPtOnSeg(out_p2, pt, t, d);
    if (RNIsZero(d * scl)) {
        out_p1 = out_p2;
        return 1;
    }
    return 0;
}



R2Polygon
R2Sphere::Intersect(const R2Polygon &in_poly ) const
{
    R2Polygon poly;

    poly.Clear();
    if ( in_poly.Num_pts() < 3 )
        return poly;

    Array<R2Pt> aptInter;
    R2Pt ptOut1, ptOut2, ptCenter(0,0);
    for ( int i = 0; i < in_poly.Num_pts(); i++ ) {
        const R2Line_seg seg = in_poly.Edge( i );
        const int iN = Intersect( seg, ptOut1, ptOut2 );
        if ( iN == 1 || iN == 2 ) {
            aptInter += ptOut1;
            ptCenter[0] += ptOut1[0];
            ptCenter[1] += ptOut1[1];
        }
        if ( iN == 2 ) {
            aptInter += ptOut2;
            ptCenter[0] += ptOut2[0];
            ptCenter[1] += ptOut2[1];
        }
    }
    if ( aptInter.num() == 0 ) {
        WINbool bInside = TRUE;

        bInside = TRUE;
        for ( FORINT i = 0; i < in_poly.Num_pts(); i++ ) {
            if ( Inside( in_poly[i] ) == FALSE ) {
                bInside = FALSE;
                break;
            }
        }
        if ( bInside == TRUE )
            return in_poly;

        return poly;
    }

    for ( FORINT i = 0; i < in_poly.Num_pts(); i++ ) {
        if ( Inside( in_poly[i] ) ) {
            aptInter += in_poly[i];
            ptCenter[0] += in_poly[i][0];
            ptCenter[1] += in_poly[i][1];
        }
    }

    if ( aptInter.num() == 2 )
        ASSERT( FALSE );

    Array<double> adDot( aptInter.num() );
    ptCenter[0] /= (double) aptInter.num(); 
    ptCenter[1] /= (double) aptInter.num();

    for ( FORINT i = 0; i < aptInter.num(); i++ ) {
        adDot[i] = atan2( aptInter[i][1] - ptCenter[1], aptInter[i][0] - ptCenter[0] );
    }

    aptInter.sort( adDot );

    for ( FORINT i = 0; i < adDot.num(); i++ ) {
        if ( ApproxEqual( aptInter[i], aptInter.wrap(i+1) ) ) {
            adDot.del(i, 1);
            aptInter.del(i, 1);
            i--;
        }
    }
    R2Polygon polyRet( aptInter.num() );

    for ( FORINT i = 0; i < aptInter.num(); i++ )
        polyRet[i] = aptInter[i];

    return polyRet;
    
}


WINbool
R2Sphere::Overlaps(const R2Polygon &in_poly ) const
{
    R2Pt p1, p2;
    if ( in_poly.Num_pts() == 0 ) {
        return FALSE;
    } else if ( in_poly.Num_pts() == 1 ) {
        return Inside( in_poly[0] );
    } else if ( in_poly.Num_pts() == 2 ) {
        return Intersect( R2Line_seg( in_poly[0], in_poly[1] ), p1, p2 ) == 0 ? FALSE : TRUE;
    }

    for ( int i = 0; i < in_poly.Num_pts(); i++ ) {
        if ( Inside( in_poly[i] ) )
            return TRUE;
    }

    for ( FORINT i = 0; i < in_poly.Num_pts(); i++ ) {
        const R2Line_seg seg = in_poly.Edge( i );
        if ( Intersect( seg, p1, p2 ) )
            return TRUE;
    }

    if ( in_poly.Inside( Center() ) )
        return TRUE;

    return FALSE;
    
}

/* -------------------------------------------------------------------------
 * DESCR   :	The positive/negative x given y

 >> y given x:
                         2       2    2    2            2
           2 d + Sqrt[4 d  - 4 (c  + d  - r  - 2 c x + x )]
>    {y -> ------------------------------------------------}}
                                  2

 * ------------------------------------------------------------------------- */
double R2Sphere::X_pos(double in_dY)
const
{
   const double c = Center()[0];
   const double d = Center()[1];
   const double r = Radius();

   return (2 * c + sqrt(4 * c*c - 
                        4 * (d*d + c*c - r*r - 2*d*in_dY + in_dY*in_dY))) / 2;
}

double R2Sphere::X_neg( double in_dY)
const
{
   const double c = Center()[0];
   const double d = Center()[1];
   const double r = Radius();

   return (2 * c - sqrt(4 * c*c - 4*(d*d + c*c - r*r - 2*d*in_dY + in_dY*in_dY))) / 2;
}

double R2Sphere::Y_pos(double in_dX)
const
{
   const double c = Center()[0];
   const double d = Center()[1];
   const double r = Radius();

   return (2 * d + sqrt(4 * d*d - 4*(c*c + d*d - r*r - 2*c*in_dX + in_dX*in_dX))) / 2;
}

double R2Sphere::Y_neg(double in_dX)
const
{
   const double c = Center()[0];
   const double d = Center()[1];
   const double r = Radius();

   return (2 * d - sqrt(4 * d*d - 4*(c*c + d*d - r*r - 2*c*in_dX + in_dX*in_dX))) / 2;
}

void R2Sphere::Print() const 
{
    cout << "Sphere dim " << Dim() << " radius " << Radius() << "  ";
    cout << "Center: " << Center() << "\n";

}

void R2Sphere::Write(ofstream &out) const 
{
    out << Center() << " " << Radius() << "\n";
}

WINbool R2Sphere::Read(ifstream &in) 
{
    WINbool bRes = Center().Read( in );
    in >> Radius();
    return (bRes == TRUE && in.good()) ? TRUE : FALSE;
}


WINbool R2Sphere::Test()
{
   R2Sphere s1(1.0), s2(R2Pt(1,1), 2), s3(s1);

   VERIFY(s1 == s3);
   
   VERIFY(s1.X_pos(0.0) == 1.0);
   VERIFY(s1.Y_pos(0.2) == sqrt(1.0 - 0.2 * 0.2));
   VERIFY(s1.X_pos(0.1) == - s1.X_neg(0.1));
   VERIFY(s1.Y_pos(0.1) == - s1.Y_neg(0.1));

   VERIFY(s1.Area() == M_PI);
   VERIFY(s1.SurfaceArea() == 2.0 * M_PI);
   
   s1 = s2;
   VERIFY(s1 == s2);
   
   R2Line line(R2Pt(1,1), R2Vec(0.5, 0.5));
   R2Line line2(R2Pt(1,3), R2Vec(1,0));
   R2Line linev1(1.0), linev2(3.0);
   R2Line line3(R2Pt(1,3.5), R2Vec(1,0.001));
   R2Pt p1, p2;
   VERIFY(s1.Intersect(line, p1, p2) == 2);
   VERIFY(RNApproxEqual(p1[0], 1.0 + 2.0 * sqrt(0.5)));
   VERIFY(RNApproxEqual(p1[1], 1.0 + 2.0 * sqrt(0.5)));
   VERIFY(RNApproxEqual(p2[0], 1.0 - 2.0 * sqrt(0.5)));
   VERIFY(RNApproxEqual(p2[1], 1.0 - 2.0 * sqrt(0.5)));

   VERIFY(s1.Intersect(linev1, p1, p2) == 2);
   VERIFY(RNApproxEqual(p1[0], 1.0));
   VERIFY(RNApproxEqual(p1[1], 3.0));
   VERIFY(RNApproxEqual(p2[0], 1.0));
   VERIFY(RNApproxEqual(p2[1], -1.0));

   VERIFY(s1.Intersect(linev2, p1, p2) == 1);
   VERIFY(RNApproxEqual(p1[0], 3.0));
   VERIFY(RNApproxEqual(p1[1], 1.0));

   VERIFY(s1.Intersect(line2, p1, p2) == 1);
   VERIFY(ApproxEqual(line2.Pt(), p1));

   VERIFY(s1.Intersect(line3, p1, p2) == 0);


   R2Line_seg seg(R2Pt(-2,-2), R2Pt(4, 4));
   R2Line_seg seg2(R2Pt(0.9,3), R2Pt(1.1,3));
   R2Line_seg seg3(R2Pt(-0.5,1), R2Pt(0.5,1));
   VERIFY(s1.Intersect(seg, p1, p2) == 2);
   VERIFY(RNApproxEqual(p1[0], 1.0 + 2.0 * sqrt(0.5)));
   VERIFY(RNApproxEqual(p1[1], 1.0 + 2.0 * sqrt(0.5)));
   VERIFY(RNApproxEqual(p2[0], 1.0 - 2.0 * sqrt(0.5)));
   VERIFY(RNApproxEqual(p2[1], 1.0 - 2.0 * sqrt(0.5)));

   VERIFY(s1.Intersect(seg2, p1, p2) == 1);
   VERIFY(ApproxEqual(R2Pt(1,3), p1));

   VERIFY(s1.Intersect(seg3, p1, p2) == 0);

   s1 = R2Sphere(1.0);
   s3 = R2Sphere(3.0);
   s2 = R2Sphere(R2Pt(0.1, 0), 2.0);

   VERIFY(s1.Dim() == 2);
   VERIFY(s1.Radius() == 1.0);
   VERIFY(s1.Center() == R2Pt(0,0));

   R2Pt pt(0.1, 0.0);
   VERIFY(s1.Inside(pt) == TRUE);
   VERIFY(s2.Inside(pt) == TRUE);
   pt[0] = 1.0;
   pt[1] = 0;
   pt[2] = 0;
   VERIFY(s1.On(pt) == TRUE);
   VERIFY(s2.On(pt) == FALSE);
   
   pt[0] = 3.0;
   VERIFY(s1.Inside(pt) == FALSE);

   s3 = s2;
   VERIFY(s3 == s2);
   VERIFY(!(s3 == s1));

   s3.Print();
   ofstream out("RNtest.txt", ios::out);
   s3.Write(out);
   out.close();
   ifstream in("RNtest.txt", ios::in);
   s1.Read(in);
   in.close();
   VERIFY(s3 == s1);
   return TRUE;
}
