#include "StdAfx.H"
#include <utils/Rn_Line.H>

/* ------------------------------------------------------------------------- */

WINbool R3Line::Intersect( const R3Line &in_line, R3Pt &out_pt ) const
{
    out_pt = R3Pt(0,0,0);
    if ( IsParallel( *this, in_line ) ) {
        const double dDist = in_line.Dist_to_line( in_line.Pt() );
        if ( RNIsZero( dDist ) ) {
            out_pt = in_line.Pt();
            return TRUE;
        } else {
            out_pt = Lerp( in_line.Pt(), Pt(), 0.5 );
            return FALSE;
        }
    }

    int i1 = 0, i2 = 1;

    for ( int i = 0; i < 3; i++ ) {
        i1 = i;
        for (int j = 0; j < 3; j++ ) {
            i2 = j;
            if ( i == j )
                continue;
            if ( !RNIsZero( Vec()[i1] ) && !RNIsZero( Vec()[i2] ) )
                break;
        }
        if ( !RNIsZero( Vec()[i1] ) && !RNIsZero( Vec()[i2] ) )
            break;
    }

    if ( RNIsZero( Vec()[i1] ) && RNIsZero( Vec()[i2] ) ) {
        ASSERT(FALSE);
        return FALSE;
    }

    const double dPX1 = Pt()[i1];
    const double dPY1 = Pt()[i2];
    const double dPX2 = in_line.Pt()[i1];
    const double dPY2 = in_line.Pt()[i2];
    const double dVX1 = Vec()[i1];
    const double dVY1 = Vec()[i2];
    const double dVX2 = in_line.Vec()[i1];
    const double dVY2 = in_line.Vec()[i2];

    const double dDenom = ( dVX1 * dVY2 ) / dVY1;
    const double dS = ( -dPX1 + dPX2 + (dVX1 / dVY1) * dPY1 - (dVX1 / dVY1) * dPY2 ) / dDenom;

    const double dT = ( - dPY1 + dPY2 + dVY2 * dS ) / dVY1;

    const R3Pt p1 = (*this)(dT);
    const R3Pt p2 = in_line(dS);

    const double dLen = Length( p1 - p2 );

    const double dLenTP = Length( (*this)(dT + 0.01) - p2 );
    const double dLenTM = Length( (*this)(dT - 0.01) - p2 );
    const double dLenSP = Length( (*this)(dS + 0.01) - p1 );
    const double dLenSM = Length( (*this)(dS - 0.01) - p1 );
    
    out_pt = Lerp( p1, p2, 0.5 );
    if ( RNIsZero( dLen ) ) {
        return TRUE;
    }

    return FALSE;
}

/* -------------------------------------------------------------------------
 * DESCR   :	Return a perpendicular line
 * ------------------------------------------------------------------------- */
R3Line
R3Line::Perpendicular(const R3Pt &in_pt) const 
{
    R3Vec vec_opp(Vec());
    if (vec_opp[1] == 1.0) vec_opp[0] = 1;
    else                   vec_opp[1] = 1;
    R3Pt pt_random(Pt() + vec_opp);
    
    return R3Line(in_pt, Project(pt_random) - pt_random);
}

/* -------------------------------------------------------------------------
 * DESCR   :	Return a parallel line
 * ------------------------------------------------------------------------- */
R3Line
R3Line::Parallel(const R3Pt &in_pt) const 
{
    return R3Line(in_pt, Vec());
}

/* -------------------------------------------------------------------------
 * DESCR   :	Print this 
 * ------------------------------------------------------------------------- */
void R3Line::Print() const
{
    cout << "RNline" << Dim() << " point " << Pt() << " vec " << Vec() << "\n";
}

void R3Line::Write(ofstream &out) const
{
    out << Pt() << " " << Vec() << "\n";
}

WINbool R3Line::Read(ifstream &in)
{
    WINbool bRes1 = m_pt.Read(in);
    WINbool bRes2 = m_vec.Read(in);
    return (bRes1 && bRes2) ? TRUE : FALSE;
}

WINbool R3Line::Test()
{
   R3Pt pt1d(0,0.1,0), pt2d(2,0,0);
   R3Vec vecd(2, 0, 0);
   R3Line line1d, line3d;
   R3Line line2d(pt1d, vecd);

   line1d = line2d;
   VERIFY(line2d.Dim() == 3);
   VERIFY(line1d == line2d);
   VERIFY(Length(line1d.Vec()) == 1.0);

   pt2d = line1d(0.2);
   cerr << "Point " << pt2d << "\n";
   VERIFY(pt2d[0] == 0.2 && pt2d[1] == 0.1 && pt2d[2] == 0);
   VERIFY(IsParallel(line1d, line2d));

   double out_t_on_line = -1, out_d_to_line = -1;
   vecd = R3Vec(0, 0.1, 0.2);
   pt1d = pt2d + vecd;
   VERIFY(line1d.FindPtOnLine(pt1d, pt2d, out_t_on_line, out_d_to_line) == FALSE);
   cerr << "t " << out_t_on_line << " " << pt1d << " " << pt2d << "\n";
   cerr << line1d(0.2) << "\n";
   VERIFY(pt2d == line1d(0.2));
   ASSERT(out_t_on_line == 0.2);
   VERIFY(out_d_to_line == Length(vecd));

   line1d.Print();
   ofstream out("RNtest.txt", ios::out);
   line1d.Write(out);
   out.close();
   ifstream in("RNtest.txt", ios::in);
   line3d.Read(in);
   VERIFY(line1d == line3d);

   R3Pt pt1(0,0.1,0), pt2(1,0,0);
   R3Vec vec(0, 0.2, 0.2);
   R3Line line1, line2(pt1, pt2), line3(pt1, vec);
   R3Line line4(line1);

   line1 = line2;
   VERIFY(line1 == line2);
   line1 = line2.Perpendicular(pt2);
   VERIFY(IsPerpendicular(line1, line2));

   line1 = line2.Parallel(pt2);
   VERIFY(IsParallel(line1, line2));

   VERIFY(IsPerpendicular(line2, line3) == FALSE);
   VERIFY(IsParallel(line2, line3) == FALSE);

   line2.SetPt(line1.Pt());
   line2.SetVec(line1.Vec() * 3.0);
   VERIFY(line1 == line2);
   VERIFY(!(line1 == line3));

   return TRUE;
}

