/** \defgroup Geometry Geometrical entities and shapes
   All classes which start with RN, R1, R2, R3, R4, or S4. 

   Points, vectors, matrices, planes, lines, line segments, circles, ellipses, spheres, polynomials, polygons, bounding boxes, co-vectors, quaternions and contours.

   Libraries needed: Geometry

   Include files: include/utils/R*, S1_CircleAngleTC.H and S4_quaternion.h

   Source files: src/utils/geometry

   Note: Most of these files are templated. 
*/

/** \defgroup HelperGeometry Approximate numerical assertions, helper functions
  \ingroup Geometry

These are some helper functions for doing approximate comparisons and
  read/write. Almost all classes have some form of approximate equal as
  well as exact (==) equals.

<br> Files: include/utils/Rn_Defs.H utils/geometry/RNDefs.cpp <br>
  
File: include/utils/Rn_Defs.H
*/

/** \defgroup SimpleGeometry Points, vectors, covectors, rays, and matrices

  \ingroup Geometry

These classes are all defined in header files because they are
  templated. For example, the three dimensional point class is R3PointTC<template>; this is typedef'd in Rn_defs.H to: R3Pt (double), R3Pt_f (float), R3Pt_i (integer). 

- Methods are found in Rn_<class name>\#_i.H
- Functions that take one (or no) arguments (e.g., Length) are found in Rn_Unary\#_i.H
- Functions that take two arguments (e.g., add) are found in Rn_Binary\#_i.H
- Input/output functions are found in Rn_IO\#_i.H
- Rn_Defs.H includes all of the files and also typedefs classes, e.g., R2Pt is the two dimensional point class with a double for the template.

Files: 
- include/utils/Rn_Defs.H <br>
- include/utils/Rn_Affine\#.H <br>
- include/utils/Rn_Projective.H <br>
- include/utils/Rn_Point\#_i.H <br>
- include/utils/Rn_Vector\#_i.H <br>
- include/utils/Rn_CoVector\#_i.H <br>
- include/utils/Rn_Matrix\#_i.H <br>
- include/utils/Rn_Unary\#_i.H <br>
- include/utils/Rn_Binary\#_i.H <br>
- include/utils/Rn_Io\#_i.H <br>

- src/utils/geometry/Rn_Defs.cpp <br>
  */
//@{
/** \defgroup SimpleGeometryElems One, two, three, and four dimensional points, vectors, co-vectors, rays, and matrices

Note: operators such as addition and Length are written as global methods, so they don't appear in the class definition but as functions in the documentation
*/
/** \defgroup SimpleGeometryUnary Length, normalization, maximum

UnitSafe differs from Unit in that it checks to see if the length is zero before dividing; so it won't die from a divide by zero.
*/
/** \defgroup SimpleGeometryConversion Conversion between vec and co-vec */
/** \defgroup SimpleGeometryAlgebraic Algebraic operators

Adding, subtracting, multiplying, dot products, linear interpolation
*/
/** \defgroup SimpleGeometryCompare Comparison operators (equal, not equal)
 
  operator == and != do exact comparison; use ApproxEqual to do epsilon comparison. The default comparison values is set in RNEpsilon_d and RNEpsilon_f. Optionally, pass in the epsilon value */
/** \defgroup SimpleGeometryLinInterp Matrix linear interpolation operators
  
    These operators, together, allow you to do linear interpolation of matrices. See Marc Alexa's paper in Siggraph '03 (?)
    They essentially operate by doing the log( exp() ). Note that these are all found by optimization and there are matrices that will fail (for example, a big rotation)
    */
/** \defgroup SimpleGeometryIO Input/output operators */
//@}

/** \defgroup QuatGeometry Quaternions
  \ingroup Geometry

This is a templated quaternion class.

File directory: include/utils <br>
  - S4_quaternion.H
 */
/** \defgroup LinearGeometry Lines, line segments, bounding boxes, and planes
  \ingroup Geometry

File directory: include/utils <br>
  - Rn_Line.H, R2_Line_i.H, R3_Line_i.H<br>
  - Rn_Line_seg.H, R2_Line_seg_i.H, R3_Line_seg_i.H<br>
  - R3_Plane.H, R3_Plane_i.H
  - RN_BBox.H, RN_BBox_i.H

File directory: src/utils/geometry <br>
  - R2Line.cpp, R3Line.cpp <br>
  - R2Line_seg.cpp, R3Line_seg.cpp <br>
  - R3Plane.cpp
  - RNBBox.cpp

Content:     2 and 3 dimensional lines
  -          R2Line - 2D line <br>
  -          R3Line - 3D line <br>
  -          R2Line_seg  - 2D line segment <br>
  -          R3Line_seg  - 3D line segment <br>
  -          R3Plane - a plane
  -          RNBBox - a bounding box. Templated on point/vector/matrix type
 */

/** \defgroup PolyGeometry Polygons and contours
  * 2D and 3D polygons and 2D contours.
  \ingroup Geometry

File directory: include/utils <br>
  - Rn_Polygon.H <br>
  - R2_Polygon_i.H <br>
  - R3_Polygon_i.H <br>
  - R2_Boxed_polygon_i.H <br>
  - R2_Contour.H <br>
  - R2_Contour_i.H <br>

File directory: src/utils/geometry <br>
  - RNPolygon.cpp <br>
  - R2Polygon.cpp <br>
  - R3Polygon.cpp <br>
  - R2Contour.cpp <br>

Content:     2 and 3 dimensional polygons
  -          RNPolygon - routines common to 2 and 3 dimensional polygons
  -          R2Polygon - 2D polygon <br>
  -          R3Polygon - 3D polygon <br>
  -          R2BoxedPolygon  - 2D polygon with bounding box <br>

Content:     2  dimensional contours
  -          R2Contour - 2D contour with 1 or more polygons, a z level, and a width <br>
  */

/** \defgroup ImplicitGeometry Circles, ellipses and spheres
  \ingroup Geometry

File directory: include/utils <br>
  - S1_CircleAngle.H For representing an angle on the circle
  - Rn_Sphere.H Has both 2D and 3D class
  - R2_Sphere_i.H 2D specific methods
  - R3_Sphere_i.H 3D specific methods
  - R2_Ellipse.H, R2_Ellipse_i.H

File directory: src/utils/geometry <br>
  - S1CircleAngleTC.cpp 
  - RNSphere.cpp <br>
  - R2Sphere.cpp <br>
  - R3Sphere.cpp <br>
  - R2Ellipse.cpp <br>

Content:     
  -          S1CircleAngleTC - a circle angle. Can add, subtract, average angles, always keeping in the range 0 to 2 pi
  -          R2Sphere - Geometric circle <br>
  -          R3Sphere - Sphere <br>
  -          R2Ellipse - 2D ellipse <br>
  */

/** \defgroup Polynomials Polynomials
  \ingroup Geometry
  * Polynomial in x or x,y

File directory: include/utils <br>
  - RN_Polynomial.H <br>
  - RN_Polynomial_i.H <br>

File directory: src/utils/geometry <br>
  - RNPolynomial.cpp <br>

Content:     1 and 2 variable polynomials
  -          RNPolynomial - Polynomial in 1 or 2 dimensions <br>
  */


