// This may look like C code, but it is really -*- C++ -*-
// Copyright (c) 1992 Hugues H. Hoppe; All rights reserved.

#include <utils/Mesh_Hh.H>
#include <utils/Mesh_Map.H>                // for hhwarning()
#include <errno.h>              // for errno with __GNUG__
#include <time.h>               // gettimeofday(), timeval, time(), ctime()
#include <signal.h>             // signal()
#include <new>                // set_new_handler()
#include <stdarg.h>             // to use with hform()

#if defined(__WIN32)

#define WIN32_EXTRA_LEAN
#define WIN32_LEAN_AND_MEAN
#include <windows.h>            // FILETIME struct
#include <time.h>
#undef WIN32_EXTRA_LEAN
#undef WIN32_LEAN_AND_MEAN
#include <fcntl.h>              // for _O_TEXT
#include <io.h>
#else

#include <WINSystemDefines.H>
#include <unistd.h>

#include <sys/types.h>          // for socket struct
#include <sys/socket.h>         // for socketpair()
#include <fcntl.h>              // for fcntl()
#include <sys/time.h>           // gettimeofday(), timeval, setitimer()
#include <netinet/in.h>
#include <inttypes.h>

extern "C" {
#if defined(__DECCXX)
    int socketpair(int, int, int, int sv[2]);
    int gettimeofday(struct timeval*, struct timezone*);
#if !defined(__alpha)
    int putenv(const char* string);
#endif
#elif __GNUG__==2 && __GNUC_MINOR__==5 && defined(ultrix)
    int socketpair(int, int, int, int sv[2]);
    int putenv(const char* string);
    int setitimer(int, struct itimerval*, struct itimerval*);
    int gettimeofday(struct timeval*, struct timezone*);
#elif defined(__alpha)
    int socketpair(int, int, int, int sv[2]);
#endif
}
extern "C" char** environ;

#endif

extern "C" {
#if defined(__DECCXX)
	int socketpair(int, int, int, int sv[2]);
	int gettimeofday(struct timeval*, struct timezone*);
#ifndef __alpha
	int putenv(const char* string);
#endif
#elif __GNUG__==2 && __GNUC_MINOR__==5
	int socketpair(int, int, int, int sv[2]);
	int putenv(const char* string);
	int setitimer(int, const struct itimerval*,struct itimerval*);
	int gettimeofday(struct timeval*, struct timezone*);
#elif defined(__alpha)
	int socketpair(int, int, int, int sv[2]);
#endif
}

static void new_failed()
{
	MESHassertnever("new is out of memory");
}

static int init()
{
#ifndef WIN32
  //Leon says: I'm not really sure what set_new_handler does, but
  //it's blocking my build, so bye-bye.
//	set_new_handler(&new_failed);
#endif
	// Assumptions present in my C++ code (Hugues Hoppe)
	MESHassertx(sizeof(int)>=4);
	MESHassertx(sizeof(void*)>=sizeof(int));
	MESHassertx(sizeof(void*)>=sizeof(float));
	return 0;
}

static int dummy=init();

void* hmalloc(void* oldp, int size)
{
	MESHassertx(size>=0);
	void* r=oldp?realloc((char*)oldp,size):malloc(size);
	return MESHassertv(r);
}

void hfree(void* e)
{
	if (e) free((char*)e);
}

int Setfdnodelay(int fd, int nodelay)
{
#if defined(__sgi)
    // on sgi, setting nodelay on terminal fd may cause window closure
    if (nodelay) MESHassertx(!isatty(fd));
#endif
#if defined(hpux) || defined(__sgi)
    return fcntl(fd,F_SETFL, nodelay ? O_NONBLOCK : 0)!=-1;
#elif defined(FNDELAY)
    return fcntl(fd,F_SETFL, nodelay ? FNDELAY : 0)!=-1;
#elif defined(O_NDELAY)
    return fcntl(fd,F_SETFL, nodelay ? O_NDELAY : 0)!=-1;
#elif defined(__WIN32)
    MESHWarning("Setfdnodelay not implemented");
    // Note: probably would have to use recv() to make this work.
    // unsigned long param=nodelay;
    // return ioctlsocket(fd,FIONBIO,&param);
    return 0;
#else
    ??;
#endif
}

void SpawnOnSockets(int* fdi, int* fdo, FILE** fi, FILE** fo,
		    void (*spawn)())
{
#ifndef WIN32
	int s[2],child;
	MESHassertx(!socketpair(AF_UNIX,SOCK_STREAM,0,s));
	MESHassertx((child=fork())>=0);
	if (!child) {
		close(s[1]);
		dup2(s[0],0);
		dup2(s[0],1);
		if (s[0]<2)
			close(s[0]);
		(*spawn)();
		_exit(0);
	}
	close(s[0]);
	if (fi) *fi=fdopen(s[1],"r");
	if (fo) *fo=fdopen(s[1],"w");
	if (fdi) *fdi=s[1];
	if (fdo) *fdo=s[1];
#endif
}


static struct Warningclass {
	~Warningclass() { flush(); }
	void flush();
	MMap<const char*,int> m; // string -> number of times printed
} Warnings;

void Warningclass::flush()
{
	if (m.empty()) return;
	SHOWDF("Summary of warnings:\n");
	ForMapKeyValue(m,const char*,s,int,n) {
		SHOWDF(" %5d '%s'\n",n,s);
	} EndFor;
	m.clear();
}

void FlushWarnings() { Warnings.flush(); }

int hhassert(int warncode, const char* ps, const char* file, int line)
{
	const char* s=(ps?ps:"?");
	if (warncode>=0) {
		if (Warnings.m.contains(s)) {
			Warnings.m.replace(s,Warnings.m.get(s)+1);
			if (warncode==1) return 0;
		} else {
			Warnings.m.enter(s,1);
		}
	}
	cerr << (warncode<0?"Fatal assertion error":"assertion warning") <<
		": " << s << " in line " << line <<
		" of file " << file << "\n";
	if (GetenvValue("ASSERTABORT") ||
	    warncode>=0 && GetenvValue("ASSERTWABORT") ||
	    warncode<0 && GetenvValue("ASSERTXABORT")) abort();
	if (warncode<0) {
		if (errno) perror("possible error");
		ASSERT(FALSE);
	}
	return 1;
}

#if defined(__WIN32)
inline double to_seconds(const FILETIME& ft)
{
    const double low_to_sec=100e-9; // 100 nanoseconds
    const double high_to_sec=low_to_sec*4294967296.0;
    return ft.dwLowDateTime*low_to_sec+ft.dwHighDateTime*high_to_sec;
}
#endif


double GetPreciseTime()
{
#if defined(__WIN32)
    // GetSystemTime(&system_time); SystemTimeToFileTime(...);
    FILETIME file_time;         // 64-bits, 1unit==100ns
    GetSystemTimeAsFileTime(&file_time); // not available on Win95?
    return to_seconds(file_time);
#else
    struct timeval ti;
    MESHassertx(!gettimeofday(&ti,(struct timezone*)0));
    return double(ti.tv_sec)+double(ti.tv_usec)*1e-6;
#endif
}

int GetTime()
{
	return (int) time((time_t*)0);
}

const char* CTime()
{
	time_t ti=GetTime();
	char* s=ctime(&ti);
	s[24]=0;		// strip out trailing '\n'
	return s;
}



#if !defined(__WIN32)
static void nullfunc(int )
{
    signal(SIGALRM,HHSIG_PF(nullfunc)); // for ATT unix
}
#endif


void Delay(double sec)
{
#if defined(WIN32)
    Sleep(int(sec*1000));       // argument specifies milliseconds
#else
    struct itimerval t;
    signal(SIGALRM,HHSIG_PF(nullfunc));
    t.it_value.tv_sec=int(sec);
    t.it_value.tv_usec=int(sec*1e6)%1000000;
    t.it_interval.tv_sec=0;
    t.it_interval.tv_usec=10000; // 10 millisec, so pause() doesn't stick
    if (setitimer(ITIMER_REAL,&t,0)) perror("setitimer");
    pause();
    t.it_value.tv_sec=0;
    t.it_value.tv_usec=0;
    t.it_interval.tv_sec=0;
    t.it_interval.tv_usec=0;
    if (setitimer(ITIMER_REAL,&t,0)) perror("setitimer");
    // SIGALRM not reset to SIG_DFL just to be sure
#endif
}


const int bufsiz=512;

// Warning: I have found that routines which use va_list should not be
// compiled with -O under cxx.
// Could be because vprintf() routines are defined in terms of varargs.h
// and not stdarg.h

const char* hform(const char* format, ...) // not reentrant!!!!
{
	const int numbuf=10;
	static char buf[numbuf][bufsiz];
	static int bufn;
	va_list ap;
	va_start(ap,format);
	bufn=(bufn+1)%numbuf;
	vsprintf(buf[bufn],format,ap);
	va_end(ap);
	return buf[bufn];
}

void SHOWF(const char* format, ...)
{
	char buf[bufsiz];
	va_list ap;
	va_start(ap,format);
	vsprintf(buf,format,ap);
	va_end(ap);
	cerr << buf;
}

// old
//	isatty1	isatty2	cout	cerr
//	0	0	1	0
//	0	1	1	1
//	1	0	1	0
//	1	1	0	1
// new
//	isatty1	isatty2	cout	cerr
//	0	0	0	1
//	0	1	1	1
//	1	0	1	1
//	1	1	0	1
void SHOWDF(const char* format, ...)
{
	static int done=0, needcout, needcerr;
	if (!done++) {
		int is1=isatty(1), is2=isatty(2);
		// old
		needcout=!is1 || !is2;
		needcerr=is2;
		// new
		needcout=is1+is2==1;
		needcerr=1;
		// SHOWN(is1); SHOWN(is2); SHOWN(needcout); SHOWN(needcerr);
	}
	char buf[bufsiz];
	va_list ap;
	va_start(ap,format);
	vsprintf(buf,format,ap);
	va_end(ap);
	if (needcout) cout << "# " << buf;
	if (needcerr) cerr << "# " << buf;
}

void SHOWFF(const char* format, ...)
{
	static int done=0, wantcout;
	if (!done++) wantcout=!isatty(1);
	if (!wantcout) return;
	char buf[bufsiz];
	va_list ap;
	va_start(ap,format);
	vsprintf(buf,format,ap);
	va_end(ap);
	cout << "# " << buf;
}

char* newString(const char* s)
{
    return s ? strcpy(new char[strlen(s)+1],s) : 0;
}

int GetenvValue(const char* varname)
{
	const char* s=getenv(varname);
	if (!s || *s=='0' || *s=='n' || *s=='N' || *s=='f' || *s=='F')
		return 0;
	return WINmax(atoi(s),1);
}



static const char* _findenv(const char* name, int& offset)
{
	int len=0;
     const char * c = NULL;
	for (c=name;*c && *c!='=';c++,len++);
        for (char** p=environ;*p;p++)
                if (!strncmp(*p,name,len))
                        if (*(c=*p+len)=='=') {
                                offset=p-environ;
                                return ++c;
                        }
        return 0;
}

// Access to 'environ' is not reentrant
//Leon found that this function was already defined on his linux
//system.
#if !defined(__linux__) && !defined(powerpc)
int unsetenv(const char* name)
{
	int offset;
	while (_findenv(name,offset)) // if set multiple times
                for (char** p=&environ[offset];;p++)
                        if (!(*p=*(p+1)))
                                break;

    return 0;
}
#endif

#if !defined(powerpc)
// return: 0=success, -1=out_of_mem
int setenv(const char* name, const char* value, int overwrite)
{
	MESHassertx(overwrite);
	return putenv(newString(hform("%s=%s",name,value)));
}
#endif

const char* CreateHeader(int argc, char** argv)
{
	int len=0;
	const char* s="# ";
	for (int i=0;i<argc;i++) {
		len+=strlen(argv[i])+1;
		if (i && len>75) s=hform("%s\n#  ",s),len=strlen(argv[i])+1;
		s=hform("%s %s",s,argv[i]);
	}
	return hform("%s\n",s);
}

