// This may look like C code, but it is really -*- C++ -*-
// Copyright (c) 1993 Hugues H. Hoppe; All rights reserved.

#include <utils/Mesh_Hh.H>
#include <utils/Mesh_Mesh.H>
#include <utils/Mesh_Set.H>
#include <utils/Mesh_Array.H>
#include <utils/Mesh_Random.H>

#pragma warning( disable : 4291 )

#define ForArray(A,T,V) ForIndex(zzz,(A).num()) T const& V=(A)[zzz];

ALLOCATEPOOL(MVertex);
ALLOCATEPOOL(MFace);
ALLOCATEPOOL(MEdge);
ALLOCATEPOOL(MHEdge);

// Note: changes to mesh go through enterHEdge() and removeHEdge() !

//            v2
//       /    ^    \
//   vo1 f1 he|  f2  vo2
//       \    |    /
//            v1            if (isBoundary(he)), f2==0, vo2==0

//  Vertex v1 -> stack of HEdges (v1,?)

//*** Mesh

int Mesh::sdebug=GetenvValue("DEBUG");

Mesh::Mesh() : vertexnum(1), facenum(1), edgenum(1), nedges(0) { }

Mesh::~Mesh()
{
    clear();
}

void Mesh::clear()
{
    if (sdebug>=1) OK();
    for (;; ) {
        MapIter<int,Face> mi(id2face);
        if (!mi) break;
        destroyFace(mi.value());
    }
    for (;; ) {
        MapIter<int,Vertex> mi(id2vertex);
        if (!mi) break;
        destroyVertex(mi.value());
    }
    vertexnum=1;
    facenum=1;
    edgenum=1;
    nedges=0;
}

void Mesh::copy(const Mesh& m)
{
    MESHassertx(!numVertices());
    gmodFlag(m.gflag(ALL),1);
    ForMeshVertex(m,v) {
        Vertex vn=createVertexI(m.vertexid(v));
        modFlag(vn,m.flag(v,ALL),1);
    } EndFor;
    Array<Vertex> va;
    ForMeshFace(m,f) {
        m.vertices(f,va);
        ForIndex(i,va.num()) { va[i]=idvertex(m.vertexid(va[i])); } EndFor;
        Face fn=createFaceI(m.faceid(f),va);
        modFlag(fn,m.flag(f,ALL),1);
    } EndFor;
    ForMeshEdge(m,e) {
        int fl=m.flag(e,ALL); if (!fl) continue;
        Edge en=orderedEdge(idvertex(m.vertexid(m.vertex1(e))),
                            idvertex(m.vertexid(m.vertex2(e))));
        modFlag(en,fl,1);
    } EndFor;
}

//** Raw manipulation

Vertex Mesh::createVertex()
{
    return createVertexI(vertexnum);
}

Vertex Mesh::createVertexI(int id)
{
    MESHassertx(id>=1);
    Vertex v=new MVertex;
    // v->sthe set to empty
    v->id=id;
    v->flags=0;
    // v->point undefined
    if (sdebug>=1) MESHassertx(!id2vertex.contains(id));
    id2vertex.enter(id,v);
    vertexnum=WINmax(vertexnum,id+1);
    return v;
}

void Mesh::destroyVertex(Vertex v)
{
    MESHassertx(!herep(v));
    MESHassertx(id2vertex.remove(v->id));
    if (vertexnum-1==v->id) vertexnum--;
    delete v;
}

int Mesh::legalCreateFace(const Array<Vertex>& va) const
{
    MESHassertx(va.num()>=3);
    if (sdebug>=1) { ForArray(va,Vertex,v) { valid(v); } EndFor; }
    if (va.num()==3) {                // cheap check
        if (va[0]==va[1] || va[1]==va[2] || va[0]==va[2]) 
            return 0;
    } else {
        Set<Vertex> setv;
        ForIndex(i,va.num()) { 
            if (!setv.add(va[i])) 
                return 0; 
        } EndFor;
    }
    Vertex vo=va[va.num()-1];
    ForIndex(i,va.num()) {
        HEdge hedge = queryHEdge(vo,va[i]);
        if ( hedge ) 
            return 0;
        vo=va[i];
    } EndFor;
    return 1;
}

Face Mesh::createFace(const Array<Vertex>& va)
{
    return createFaceI(facenum,va);
}

Face Mesh::createFace(Vertex v1, Vertex v2, Vertex v3, int id)
{
    static Array<Vertex> va(3);
    va[0]=v1; va[1]=v2; va[2]=v3;
    if( id<0 )
        return createFace(va);
    else
        return createFaceI( id, va );
}

Face Mesh::createFaceI(int id, const Array<Vertex>& va)
{
    MESHassertx(id>=1);
    MESHassertx(va.num()>=3);
    if (sdebug>=1) MESHassertx(legalCreateFace(va));
    Face f=new MFace;
    // f->herep defined below
    f->id=id;
    f->flags=0;
    if (sdebug>=1) 
        MESHassertx(!id2face.contains(id));
    id2face.enter(id,f);
    HEdge hep=0;
    int nv=va.num();
    ForIndex(i,nv) {
        Vertex v2=va[i+1==nv?0:i+1];
        HEdge he=new MHEdge;
        he->prev=hep;
        // he->next is set below
        he->vert=v2;
        enterHEdge(he,va[i]);
        // he->sym and he->edge were set in enterHEdge
        he->face=f;
        hep=he;
    } EndFor;
    HEdge helast=hep;
    for (;;) {
        HEdge hepp=hep->prev;
        if (!hepp) break;
        hepp->next=hep;
        hep=hepp;
    }
    hep->prev=helast;
    helast->next=hep;
    f->herep=helast;            // such that f->herep->vert==va[0]
    facenum=WINmax(facenum,id+1);
    if (sdebug>=3) OK();
    return f;
}

void Mesh::destroyFace(Face f)
{
    NEST {
        HEdge he=MESHassertv(herep(f)), hef=he;
        Vertex v1=he->prev->vert;
        for (;;) {
            HEdge hen=he->next;
            Vertex v1n=he->vert;
            removeHEdge(he,v1);
            delete he;
            he=hen;
            v1=v1n;
            if (he==hef) break;
        }
    }
    MESHassertx(id2face.remove(f->id));
    if (facenum-1==f->id) 
        facenum--;
    delete f;
    if (sdebug>=3) OK();
}

//** Vertex

int Mesh::isNice(Vertex v) const
{
    // return numBoundaries(v)<2;  // ignores 2 interior vertices touching
    HEdge her=herep(v);
    if (!her) return 1;
    int nhe=0;
    HEdge he;
    for (he=her;;) {
        nhe++;
        he=clwHEdge(he);
        if (!he || he==her) break;
    }
    if (he!=her) for (he=her;he=ccwHEdge(he);) nhe++;
    return nhe==v->sthe.height();
}

int Mesh::degree(Vertex v) const
{
    int i=0;
    ForStack(v->sthe,HEdge,he) { i+=isBoundary(he)?2:1; } EndFor;
    return i;
}

int Mesh::numBoundaries(Vertex v) const
{
    int i=0;
    ForStack(v->sthe,HEdge,he) { if (isBoundary(he)) i++; } EndFor;
    return i;
}

int Mesh::isBoundary(Vertex v) const
{
    MESHassertx(herep(v));
    int nb=numBoundaries(v);

    return nb;
}

Edge Mesh::oppEdge(Vertex v, Face f) const
{
    MESHassertx(isTriangle(f));
    return getHEdge(v,f)->prev->edge;
}

Vertex Mesh::oppVertex(Vertex v, Edge e) const
{
    Vertex v1=vertex1(e), v2=vertex2(e);
    if (v==v1) return v2;
    if (v==v2) return v1;
    MESHassertnever("Vertex not on Edge"); return 0;
}

Vertex Mesh::mostClwVertex(Vertex v) const
{
    HEdge he=mostClwHEdge(v);
    return he ? he->next->vert : 0;
}

Vertex Mesh::mostCcwVertex(Vertex v) const
{
    HEdge he=mostCcwHEdge(v);
    return he ? he->prev->vert : 0;
}

Vertex Mesh::clwVertex(Vertex v, Vertex vext) const
{
    HEdge he=queryHEdge(vext,v); if (he) return he->next->vert;
    if (sdebug>=1) MESHassertx(queryHEdge(v,vext));
    return 0;
}

Vertex Mesh::ccwVertex(Vertex v, Vertex vext) const
{
    HEdge he=queryHEdge(v,vext); if (he) return he->prev->prev->vert;
    if (sdebug>=1) MESHassertx(queryHEdge(vext,v));
    return 0;
}

Face Mesh::mostClwFace(Vertex v) const
{
    HEdge he=mostClwHEdge(v);
    return he ? he->face : 0;
}

Face Mesh::mostCcwFace(Vertex v) const
{
    HEdge he=mostCcwHEdge(v);
    return he ? he->face : 0;
}

Face Mesh::clwFace(Vertex v, Face f) const
{
    return oppFace(f,ccwEdge(f,v));
}

Face Mesh::ccwFace(Vertex v, Face f) const
{
    return oppFace(f,clwEdge(f,v));
}

Edge Mesh::mostClwEdge(Vertex v) const
{
    HEdge he=mostClwHEdge(v);
    return he ? he->next->edge : 0;
}

Edge Mesh::mostCcwEdge(Vertex v) const
{
    HEdge he=mostCcwHEdge(v);
    return he ? he->edge : 0;
}

Edge Mesh::clwEdge(Vertex v, Edge e) const
{
    HEdge he=hefromev2(e,v);
    return he ? he->next->edge : 0;
}

Edge Mesh::ccwEdge(Vertex v, Edge e) const
{
    HEdge he=hefromev1(e,v);
    return he ? he->prev->edge : 0;
}

Face Mesh::ccwFace(Vertex v, Edge e) const
{
    if (vertex1(e)==v) return face1(e);
    if (vertex2(e)==v) return face2(e);
    MESHassertnever("Vertex not on edge"); return 0;
}

Face Mesh::clwFace(Vertex v, Edge e) const
{
    if (vertex1(e)==v) return face2(e);
    if (vertex2(e)==v) return face1(e);
    MESHassertnever("Vertex not on edge"); return 0;
}

//** Face

int Mesh::isNice(Face f) const
{
    // non-nice iff all its neighbors are the same
    Face fp=0; int i=0;
    ForFaceFace(*this,f,ff) {
        if (!i++) { fp=ff; }
        else if (ff!=fp) { fp=0; break; }
    } EndFor;
    return i<=1 || !fp;
}

int Mesh::numVertices(Face f) const
{
    int i=0;
    ForFaceHEdge(*this,f,he) { dummy_use(he); i++; } EndFor;
    return i;
}

int Mesh::isTriangle(Face f) const
{
    HEdge he=herep(f);
    return he->next->next->next==he;
}

int Mesh::isBoundary(Face f) const
{
    ForFaceVertex(*this,f,v) { if (isBoundary(v)) return 1; } EndFor;
    return 0;
}

Face Mesh::oppFace(Face f, Edge e) const
{
    Face f1=face1(e), f2=face2(e);
    if (f==f1) return f2;
    if (f==f2) return f1;
    MESHassertnever("Face not adjacent to Edge"); return 0;
}

void Mesh::vertices(Face f, Array<Vertex>& va) const
{
    va.init(0);
    ForFaceVertex(*this,f,v) { va+=v; } EndFor;
}

void Mesh::corners(Face f, Array<Corner>& ca) const
{
    ca.init(0);
    ForFaceCorner(*this,f,c) { ca+=c; } EndFor;
}

Edge Mesh::clwEdge(Face f, Edge e) const
{
    return hefromef(e,f)->prev->edge;
}

Edge Mesh::ccwEdge(Face f, Edge e) const
{
    return hefromef(e,f)->next->edge;
}

Vertex Mesh::clwVertex(Face f, Vertex v) const
{
    return getHEdge(v,f)->prev->vert;
}

Vertex Mesh::ccwVertex(Face f, Vertex v) const
{
    return getHEdge(v,f)->next->vert;
}

Edge Mesh::clwEdge(Face f, Vertex v) const
{
    return getHEdge(v,f)->edge;
}

Edge Mesh::ccwEdge(Face f, Vertex v) const
{
    return getHEdge(v,f)->next->edge;
}

//** Edge

Vertex Mesh::sideVertex1(Edge e) const
{
    return oppVertex(e,face1(e));
}

Vertex Mesh::sideVertex2(Edge e) const
{
    return face2(e) ? oppVertex(e,face2(e)) : 0;
}

Vertex Mesh::oppVertex(Edge e, Face f) const
{
    HEdge he=hefromef(e,f);
    MESHassertx(he->next->next->next==he);
    return he->next->vert;
}

Edge Mesh::oppBoundary(Edge e, Vertex v) const
{
    HEdge he=herep(e);
    MESHassertx(isBoundary(he));
    if (he->vert==v) {
        HEdge hen;
        while (hen=clwHEdge(he)) he=hen;
        return he->next->edge;
    }
    if (he->prev->vert==v) {
        HEdge hen;
        he=he->prev;
        while (hen=ccwHEdge(he)) he=hen;
        return he->edge;
    }
    MESHassertnever("Vertex not on Edge"); return 0;
}

Edge Mesh::clwBoundary(Edge e) const
{
    return oppBoundary(e,vertex2(e));
}

Edge Mesh::ccwBoundary(Edge e) const
{
    return oppBoundary(e,vertex1(e));
}

//** Corner

Corner Mesh::corner(Vertex v, Face f) const
{
    return getHEdge(v,f);
}

Corner Mesh::ccwCorner(Vertex v, Edge e) const
{
    HEdge he=hefromev1(e,v);
    return he ? he->prev : 0;
}

Corner Mesh::clwCorner(Vertex v, Edge e) const
{
    HEdge he=hefromev2(e,v);
    return he;
}

//** Other associations

Edge Mesh::queryEdge( Vertex v, Vertex w) const
{
    HEdge he;
    he=queryHEdge(v,w); if (he) return he->edge;
    he=queryHEdge(w,v); if (he) return he->edge;
    return 0;
}

Edge Mesh::queryEdgeOrdered( Vertex v, Vertex w) const
{
    HEdge he;
    he=queryHEdge(v,w); if (he) return he->edge;
    return 0;
}

Edge Mesh::edge( Vertex v, Vertex w) const
{
    return MESHassertv(queryEdge(v,w));
}

Edge Mesh::orderedEdge(Vertex v1, Vertex v2) const
{
    HEdge he=MESHassertv(queryHEdge(v1,v2));
    Edge e=he->edge;
    MESHassertx(e->herep==he);
    return e;
}

Face Mesh::face(Vertex v, Vertex vccw) const
{
    return clwFace(vccw,edge(v,vccw));
}

//** Counting routines

int Mesh::numVertices() const
{
    return id2vertex.num();
}

int Mesh::numFaces() const
{
    return id2face.num();
}
int Mesh::numEdges() const
{
    return nedges;
}

Vertex Mesh::randomVertex(MeshRandom& r) const
{
    MapIter<int,Vertex> mi(id2vertex,r);
    MESHassertx(mi); return mi.value();
}

Face Mesh::randomFace(MeshRandom& r) const
{
    MapIter<int,Face> mi(id2face,r);
    MESHassertx(mi); return mi.value();
}

Edge Mesh::randomEdge(MeshRandom& r) const
{
    Face f=randomFace(r);
    int vi=r.getint()%numVertices(f);
    HEdge hef=0;
    ForFaceHEdge(*this,f,he) {
        if (!vi--) { hef=he; break; }
    } EndFor;
    return MESHassertv(hef)->edge;
}

//** Mesh operations

int Mesh::legalEdgeCollapse(Edge e) const
{
    if (sdebug>=1) valid(e);
    Vertex v1=vertex1(e), v2=vertex2(e);
    Vertex vo1=sideVertex1(e), vo2=sideVertex2(e); // vo2 may be 0
    // Check that substituting v2 to v1 will not duplicate an edge in any
    // faces adjacent to v2 (besides f1 and f2).
    // (case of Face vertices being duplicated cannot happen here
    //  since only f1 and f2 can have both v1 and v2)
    ForVertexVertex(*this,v2,v) {
        if (v==v1 || v==vo1 || v==vo2) continue;
        if (queryEdge(v,v1)) return 0;
    } EndFor;
    return 1;
}

int Mesh::niceEdgeCollapse(Edge e) const
{
    if (sdebug>=1) valid(e);
    if (!legalEdgeCollapse(e))
        return 0;
    Vertex v1=vertex1(e), v2=vertex2(e);
    Face f1=face1(e), f2=face2(e);
    MESHassertx(isTriangle(f1));
    if (f2) MESHassertx(isTriangle(f2));
    if (sdebug>=1) MESHassertx(isNice(v1) && isNice(v2));
    // Requirements:
    //* 1 - If v1 and v2 are both boundary, (v1,v2) is a boundary edge
    if (!isBoundary(e) && isBoundary(v1) && isBoundary(v2)) return 0;
    //* 2 - For all vertices adjacent to both v1 and v2, exists a face
    Vertex vo1=sideVertex1(e), vo2=sideVertex2(e);
    Set<Vertex> set;
    ForVertexVertex(*this,v1,v) {
        if (v!=vo1 && v!=vo2) set.enter(v);
    } EndFor;
    ForVertexVertex(*this,v2,v) {
        if (v!=vo1 && v!=vo2 && !set.add(v)) return 0;
    } EndFor;
    //* 3 - two small base cases: single face and tetrahedron
    if (set.num()==2 && isBoundary(e))
        return 0;               // single face
    if (set.num()==2 && !isBoundary(v1) && !isBoundary(v2))
        return 0;               // tetrahedron
    return 1;
}

int Mesh::legalEdgeSwap(Edge e) const
{
    if (sdebug>=1) valid(e);
    if (isBoundary(e)) return 0;
    // illegal if cross edge already exists (as in tetrahedron)
    if (queryEdge(sideVertex1(e),sideVertex2(e))) return 0;
    return 1;
}

void Mesh::collapseEdgeVertex(Edge e, Vertex vs)
{
    MESHassertx(legalEdgeCollapse(e));
    HEdge he1=hefromev1(e,vs);
    HEdge he2=hefromev2(e,vs);
    Vertex vt=he1?he1->vert:he2->prev->vert;
    MESHassertx(vt==oppVertex(vs,e)); // optional
    // Create bogus hedges if boundaries
    Array<HEdge> ar_he;
    if (he1) ar_he+=he1->prev;
    if (he2) ar_he+=he2->next;
    create_bogus_hedges(ar_he);
    if (he1) {
        MESHassertx(isTriangle(he1->face));
        destroyFace(he1->face);
    }
    if (he2) {
        MESHassertx(isTriangle(he2->face));
        destroyFace(he2->face);
    }
    // Change remaining faces around vt to have vs instead
    ForVertexFace(*this,vt,f) {
        // used to be substituteFaceVertex(f,vt,vs)
        // ends up deleting and recreating MEdge structures, which is great.
        HEdge he=getHEdge(vt,f);
        removeHEdge(he,he->prev->vert); removeHEdge(he->next,he->vert);
        he->vert=vs;
        enterHEdge(he,he->prev->vert); enterHEdge(he->next,he->vert);
    } EndFor;
    // Destroy vertex vt
    destroyVertex(vt);
    remove_bogus_hedges(ar_he);
}

void Mesh::collapseEdge(Edge e)
{
    collapseEdgeVertex(e,vertex1(e));
}

Vertex Mesh::splitEdge(Edge e, int id)
{
    if (sdebug>=1) valid(e);
    Vertex v1=vertex1(e), v2=vertex2(e);
    Face f1=face1(e), f2=face2(e); // f2 could be 0
    Vertex vo1=sideVertex1(e), vo2=sideVertex2(e); // implies triangles
    // Create bogus hedges if boundaries
    Array<HEdge> ar_he; ForEdgeFace(*this,e,f) {
        ForFaceHEdge(*this,f,he) { if (he->edge!=e) ar_he+=he; } EndFor;
    } EndFor; create_bogus_hedges(ar_he);
    // Create new vertex
    Vertex vn=id ? createVertexI(id) : createVertex();
    int f1Id( f1->Id() ),
        f2Id;
    //destroy and recreate face 1 (waste no IDs)
    destroyFace(f1);
    createFace(vn,v2,vo1,f1Id); createFace(vn,vo1,v1);
    if (f2){
        f2Id = f2->Id();
        //destroy and recreate face 2 (waste no IDs)
        destroyFace(f2);
        createFace(vn,v1,vo2,f2Id); createFace(vn,vo2,v2);
    }
    remove_bogus_hedges(ar_he);
    return vn;
}

Edge Mesh::swapEdge(Edge e)
{
    MESHassertx(legalEdgeSwap(e));
    Vertex v1=vertex1(e), v2=vertex2(e);
    Face f1=face1(e), f2=face2(e);
    Vertex vo1=sideVertex1(e), vo2=sideVertex2(e); // implies triangles
    // Create bogus hedges if boundaries
    Array<HEdge> ar_he;
    
    ForEdgeFace(*this,e,f) {
        ForFaceHEdge(*this,f,he) { 
            ar_he+=he; 
        } EndFor;
    } EndFor; 
    
    create_bogus_hedges(ar_he);

    const int iFId1 = f1->id;
    const int iFId2 = f2->id;
    const int iEId2 = e->id;

    // Destroy faces
    destroyFace(f1);
    destroyFace(f2);
    // Create new faces; may die if illegal
    createFace(v1,vo2,vo1, iFId1);
    createFace(v2,vo1,vo2, iFId2);

    Edge eNew = queryEdge( vo1, vo2 );

    if ( eNew->id == edgenum-1 )
        edgenum--;

    MESHassertx(id2edge.remove(eNew->id)==eNew);
    eNew->id=iEId2;
    id2edge.enter(iEId2,eNew);

    remove_bogus_hedges(ar_he);
    return edge(vo1,vo2);
}

Vertex Mesh::splitVertex(Vertex v1, Vertex vs1, Vertex vs2, int v2i)
{
#if 0
    Vertex v2=v2i?createVertexI(v2i):createVertex();
    Set<Face> setf;             // faces clw of vs1 and ccw of vs2
    Face f=clwFace(v1,edge(v1,vs1));
    for (;f;f=clwFace(v1,f)) {
        setf.enter(f);
        if (vs2 && ccwVertex(f,v1)==vs2) break;
    }
    if (vs2 && !f) {
        f=ccwFace(v1,edge(v1,vs2));
        for (;f;f=ccwFace(v1,f)) {
            setf.enter(f);
            MESHassertx(clwVertex(f,v1)!=vs1);
        }
    }
    ForSet(setf,Face,f) {
        // ends up deleting and recreating MEdge structures, which is great.
        HEdge he=getHEdge(v1,f);
        removeHEdge(he,he->prev->vert); removeHEdge(he->next,he->vert);
        he->vert=v2;
        enterHEdge(he,he->prev->vert); enterHEdge(he->next,he->vert);
    } EndFor;
    // info on edges around v1 still valid; edges on v2 all new
    return v2;
#endif
    Vertex v2=v2i?createVertexI(v2i):createVertex();
    Stack<Corner> stackc;       // faces clw of vs1 and ccw of vs2
    Corner c;
    if (1) {
        c=clwCorner(v1,edge(v1,vs1));
        for (;c;c=clwCorner(c)) {
            stackc.push(c);
            if (vs2 && corner_vertex(ccwFaceCorner(c))==vs2) break;
        }
    }
    if (vs2 && !c) {
        c=ccwCorner(v1,edge(v1,vs2));
        for (;c;c=ccwCorner(c)) {
            stackc.push(c);
            MESHassertx(corner_vertex(clwFaceCorner(c))!=vs1);
        }
    }
    ForStack(stackc,Corner,c) {
        // ends up deleting and recreating MEdge structures, which is great.
        HEdge he=c;
        removeHEdge(he,he->prev->vert); removeHEdge(he->next,he->vert);
        he->vert=v2;
        enterHEdge(he,he->prev->vert); enterHEdge(he->next,he->vert);
    } EndFor;
    // info on edges around v1 still valid; edges on v2 all new
    return v2;
}

Vertex Mesh::center_split_face(Face f)
{
    if (sdebug>=1) valid(f);
    Array<Vertex> va; vertices(f,va);
    // Create bogus hedges if boundaries
    Array<HEdge> ar_he; ForFaceHEdge(*this,f,he) { ar_he+=he; } EndFor;
    create_bogus_hedges(ar_he);
    // Destroy face
    destroyFace(f);
    // Create new vertex and faces
    Vertex vn=createVertex();
    ForIndex(i,va.num()) {
        createFace(va[i],va[(i+1)%va.num()],vn);
    } EndFor;
    remove_bogus_hedges(ar_he);
    return vn;
}

void Mesh::center_split_face(Face f, Vertex vn)
{
    if (sdebug>=1) valid(f);
    Array<Vertex> va; vertices(f,va);
    // Create bogus hedges if boundaries
    Array<HEdge> ar_he; ForFaceHEdge(*this,f,he) { ar_he+=he; } EndFor;
    create_bogus_hedges(ar_he);

    const int iID = f->Id();
    // Destroy face
    destroyFace(f);
    // Create new vertex and faces
    Array<Vertex> avNew(3);
    avNew[2] = vn;
    ForIndex(i,va.num()) {
        avNew[0] = va[i];
        avNew[1] = va[(i+1)%va.num()];
        if ( i == 0 )
            createFaceI(iID, avNew);
        else
            createFace(avNew);
    } EndFor;
    remove_bogus_hedges(ar_he);
}

Edge Mesh::split_face(Face f, Vertex v1, Vertex v2)
{
    if (sdebug>=1) valid(v1),valid(v2);
    MESHassertx(!queryEdge(v1,v2));
    Array<Vertex> va1, va2;
    For (Vertex v=v1;;) { va1+=v; if (v==v2) break; v=ccwVertex(f,v); } EndFor;
    For (Vertex v=v2;;) { va2+=v; if (v==v1) break; v=ccwVertex(f,v); } EndFor;
    // Create bogus hedges if boundaries
    Array<HEdge> ar_he; ForFaceHEdge(*this,f,he) { ar_he+=he; } EndFor;
    create_bogus_hedges(ar_he);
    // Destroy face
    int oldId = f->Id();
    destroyFace(f);
    // Create new faces
    if (sdebug>=1) MESHassertx(legalCreateFace(va1) && legalCreateFace(va2));
    createFaceI(oldId,va1); createFace(va2);
    remove_bogus_hedges(ar_he);
    return edge(v1,v2);
}

void Mesh::gatheredgecoalescevertices(Edge e, Array<Vertex>& va) const
{
    Vertex v1=vertex1(e), v2=vertex2(e);
    Face f1=face1(e), f2=face2(e);
    va.init(0);
    if (1) {
        static Array<Vertex> va1, va2;
        vertices(f1,va1); vertices(f2,va2);
        int nv1=va1.num(), nv2=va2.num();
        int i1, i2, ic;
        // Find one vertex common to both faces (v1)
        for (i1=0;i1<nv1;i1++) if (va1[i1]==v1) break;
        for (i2=0;i2<nv2;i2++) if (va2[i2]==v1) break;
        MESHassertx(i1<nv1 && i2<nv2);
        // Find most clw vertex on face1 common to both
        for (;va1[(i1-1+nv1)%nv1]==va2[(i2+1)%nv2];)
            i1=(i1-1+nv1)%nv1,i2=(i2+1)%nv2;
        // Let ic be the number of vertices common to both faces
        for (ic=1;;ic++) if (va1[(i1+ic)%nv1]!=va2[(i2-ic+nv2)%nv2]) break;
        For (int i=ic-1;i<nv1;i++) { va+=va1[(i1+i)%nv1]; } EndFor;
        For (int i=0;i<nv2-ic+1;i++) { va+=va2[(i2+i)%nv2]; } EndFor;
    } else {
        // this version only works for a single edge between 2 faces
        For (Vertex v=v2;;) {
            va+=v; v=ccwVertex(f1,v); if (v==v1) break;
        } EndFor;
        For (Vertex v=v1;;) {
            va+=v; v=ccwVertex(f2,v); if (v==v2) break;
        } EndFor;
    }
}

int Mesh::legal_coalesce_faces(Edge e)
{
    if (MESHassertw(!isBoundary(e))) return 0;
    Array<Vertex> va; gatheredgecoalescevertices(e,va);
    NEST {                      // check for duplicate vertices
        Set<Vertex> setv;
        ForArray(va,Vertex,v) { if (!setv.add(v)) return 0; } EndFor;
    }
    NEST {                      // check that we get a "nice" face
        Face f1=face1(e), f2=face2(e);
        Face fp=0; int i=0;
        ForEdgeFace(*this,e,f) {
            ForFaceEdge(*this,f,ee) {
                if (isBoundary(ee)) { fp=0; break; }
                Face ff=oppFace(f,ee);
                if (ff==f1 || ff==f2) continue;
                if (!i++) fp=ff;
                else if (ff!=fp) { fp=0; break; }
            } EndFor;
        } EndFor;
        if (fp) { MESHWarning("coalesce -> non-nice face"); return 0; }
    }
    return 1;
}

Face Mesh::coalesce_faces(Edge e)
{
    if (sdebug>=1) MESHassertx(legal_coalesce_faces(e));
    Face f1=face1(e), f2=face2(e);
    Array<Vertex> va; gatheredgecoalescevertices(e,va);
    // See if any vertices can be deleted
    Set<Vertex> vbefore;
    ForFaceVertex(*this,f1,v) { vbefore.enter(v); } EndFor;
    ForFaceVertex(*this,f2,v) { vbefore.add(v); } EndFor;
    // Create bogus hedges if boundaries
    Array<HEdge> ar_he; ForEdgeFace(*this,e,f) {
        ForFaceHEdge(*this,f,he) { ar_he+=he; } EndFor;
    } EndFor; create_bogus_hedges(ar_he);
    // Destroy faces
    destroyFace(f1);
    destroyFace(f2);
    // Create new face
    if (sdebug>=1) MESHassertx(legalCreateFace(va));
    Face fn=createFace(va);
    remove_bogus_hedges(ar_he);
    // Delete any hanging vertices
    ForArray(va,Vertex,v) { (void)vbefore.remove(v); } EndFor;
    ForSet(vbefore,Vertex,v) { destroyVertex(v); } EndFor;
    if (sdebug>=3) OK();
    return fn;
}

Vertex Mesh::insert_vertex_on_edge(Edge e)
{
    // If >1 edge shared between f1 and f2, other shared edges will be
    //  destroyed and recreated -> will lose attributes.
    if (sdebug>=1) valid(e);
    // Create bogus hedges if boundaries
    Array<HEdge> ar_he; ForEdgeFace(*this,e,f) {
        ForFaceHEdge(*this,f,he) { ar_he+=he; } EndFor;
    } EndFor; create_bogus_hedges(ar_he);
    Vertex v1=vertex1(e), v2=vertex2(e);
    Face f1=face1(e), f2=face2(e);
    Array<Vertex> va1, va2;
    For (Vertex v=v2;;) {
        va1+=v; v=ccwVertex(f1,v); if (v==v2) break;
    } EndFor;
    if (f2) {
        For (Vertex v=v1;;) {
            va2+=v; v=ccwVertex(f2,v); if (v==v1) break;
        } EndFor;
    }
    destroyFace(f1);
    if (f2) destroyFace(f2);
    Vertex vn=createVertex();
    va1+=vn; f1=createFace(va1);
    if (f2) { va2+=vn; f2=createFace(va2); }
    remove_bogus_hedges(ar_he);
    return vn;
}

Edge Mesh::remove_vertex_between_edges(Vertex vr)
{
    Array<Face> fa; ForVertexCcwFace(*this,vr,f) { fa+=f; } EndFor;
    // Create bogus hedges if boundaries
    Array<HEdge> ar_he; ForArray(fa,Face,f) {
        ForFaceHEdge(*this,f,he) { ar_he+=he; } EndFor;
    } EndFor; create_bogus_hedges(ar_he);
    Array<Vertex> va[2];
    ForIndex(i,fa.num()) {
        For (Vertex v=vr;;) {
            v=ccwVertex(fa[i],v); if (v==vr) break;
            va[i]+=v;
        } EndFor;
    } EndFor;
    ForIndex(i,fa.num()) { destroyFace(fa[i]); } EndFor;
    destroyVertex(vr);
    ForIndex(i,fa.num()) { fa[i]=createFace(va[i]); } EndFor;
    remove_bogus_hedges(ar_he);
    return edge(va[0][0],va[0].last());
}

int Mesh::isNice() const
{
    ForMeshVertex(*this,v) { if (!isNice(v)) return 0; } EndFor;
    ForMeshFace(*this,f) { if (!isNice(f)) return 0; } EndFor;
    return 1;
}

void Mesh::renumber()
{
    int id=1;
    ForMeshOrderedVertex(*this,v) {
        MESHassertx(id2vertex.remove(v->id)==v);
        v->id=id;
        id2vertex.enter(id,v);
        id++;
    } EndFor;
    vertexnum = id2vertex.num()+1;

    id=1;
    ForMeshOrderedFace(*this,f) {
        MESHassertx(id2face.remove(f->id)==f);
        f->id=id;
        id2face.enter(id,f);
        id++;
    } EndFor;
    facenum = id2face.num()+1;

    id=1;
    ForMeshOrderedEdge(*this,e) {
        MESHassertx(id2edge.remove(e->id)==e);
        e->id=id;
        id2edge.enter(id,e);
        id++;
    } EndFor;
    edgenum = id2edge.num()+1;
}

void Mesh::OK() const
{
    // Check that sets and maps have unique entries
    id2vertex.OK();
    id2face.OK();
    id2edge.OK();
    // Check consistency of id2x (one way)
    ForMapKeyValue(id2vertex,int,id,Vertex,v) {
        valid(v);
        MESHassertx(v->id==id);
    } EndFor;
    ForMapKeyValue(id2face,int,id,Face,f) {
        valid(f);
        MESHassertx(f->id==id);
    } EndFor;
    // Look over Vertices
    Set<HEdge> sethe;
    ForMeshVertex(*this,v1) {
        MESHassertx(id2vertex.get(v1->id)==v1);
        Set<Vertex> set;
        ForStack(v1->sthe,HEdge,he) {
            // Check that HEdges are valid
            valid(he);
            MESHassertx(he->prev->vert==v1);
            Vertex v2=he->vert;
            // Check that v2's are not duplicated
            MESHassertx(set.add(v2));
            // Check that sym matches that in sthe
            MESHassertx(queryHEdge(v2,v1)==he->sym);
            // Check that HEdge sym is valid
            if (he->sym) {
                valid(he->sym);
                MESHassertx(he->sym->edge==he->edge);
                MESHassertx(he->sym->sym==he);
                MESHassertx(he->sym->vert==he->prev->vert);
                MESHassertx(he->sym->prev->vert==he->vert);
            }
            // Check that hedges are unique, and one-to-one with ste (1)
            MESHassertx(sethe.add(he));
            // Check that Faces reachable from Edges are valid
            valid(he->face);
            // Check that each HEdge appears in its face
            int found=0;
            ForFaceHEdge(*this,he->face,hee) {
                if (hee==he) { found=1; break; }
            } EndFor;
            MESHassertx(found);
            // Check that Edge is valid
            valid(he->edge);
        } EndFor;
    } EndFor;
    // Look over Faces
    ForMeshFace(*this,f) {
        MESHassertx(id2face.get(f->id)==f);
        // Check that Face has valid erep
        MESHassertx(herep(f)->face==f);
        valid(herep(f)->edge);
        Set<Vertex> set;
        ForFaceVertex(*this,f,v) {
            // Check Face contains no duplicate Vertices
            MESHassertx(set.add(v));
            // Check Vertices in face are valid
            valid(v);
        } EndFor;
        MESHassertx(set.num()>=3);
        // Check Edges in Faces are one-to-one with sthe (2)
        ForFaceHEdge(*this,f,he) {
            MESHassertx(he->face==f);
            MESHassertx(sethe.remove(he));
        } EndFor;
    } EndFor;
    // Check Edges in Faces are one-to-one with sthe (3)
    MESHassertx(sethe.empty());
    // Check nedges
    NEST {
        int i=0;
        ForMeshEdge(*this,e) { dummy_use(e); i++; } EndFor;
        MESHassertx(nedges==i);
    }
}

WINbool Mesh::valid(Vertex v) const
{
    if ( !v ) return FALSE;
    if ( !id2vertex.contains( vertexid(v) ) )
        return FALSE;

    if ( v != id2vertex.get( vertexid(v) ) )
        return FALSE;
    return TRUE;
}

WINbool Mesh::valid(Face f) const
{
    if ( !f ) return FALSE;
    if ( !id2face.contains( faceid(f) ) )
        return FALSE;
    if ( f != id2face.get( faceid(f) ) )
        return FALSE;
    return TRUE;
}

WINbool Mesh::valid(Edge e) const
{
    if ( !e ) return FALSE;
    if ( !id2edge.contains( edgeid(e) ) )
        return FALSE;
    if ( e != id2edge.get( edgeid(e) ) )
        return FALSE;
    if (valid( vertex1(e) ) == FALSE )
        return FALSE;
    if (valid(vertex2(e)) == FALSE )
        return FALSE;

    if (valid(face1(e)) == FALSE )
        return FALSE;

    MESHassertx(orderedEdge(vertex1(e),vertex2(e))==e);
    return TRUE;
}

void Mesh::valid(Corner c) const
{
    HEdge he=c;                 // they are equivalent
    MESHassertx(he && he->next && he->prev && he->face && he->vert);
    Edge e=MESHassertv(he->edge);
    MESHassertx(hefromev2(e,he->vert)==he);
}

//** Statics

int Mesh::sflag;                // initially zero
int MVertex::sflag;             // initially zero
int MFace::sflag;               // initially zero
int MEdge::sflag;               // initially zero

//** Mesh protected

int Mesh::isBoundary(HEdge he) const
{
    return !he->sym;
}

HEdge Mesh::mostClwHEdge(Vertex v) const
{
    MESHassertx(isNice(v));
    HEdge he=herep(v);
    if (!he) return 0;
    for (HEdge hef=he;;) {
        HEdge hen=clwHEdge(he);
        if (!hen || hen==hef) break;
        he=hen;
    }
    return he;
}

HEdge Mesh::mostCcwHEdge(Vertex v) const
{
    MESHassertx(isNice(v));
    HEdge he=herep(v);
    if (!he) return 0;
    for (HEdge hef=he;;) {
        HEdge hen=ccwHEdge(he);
        if (!hen || hen==hef) break;
        he=hen;
    }
    return he;
}

HEdge Mesh::clwHEdge(HEdge he) const
{
    return he->next->sym;
}

HEdge Mesh::ccwHEdge(HEdge he) const
{
    return he->sym ? he->sym->prev : 0;
}

HEdge Mesh::hefromev1(Edge e, Vertex v) const
{
    if (vertex1(e)==v) return herep(e);
    if (vertex2(e)==v) return herep(e)->sym;
    MESHassertnever("Vertex not on Edge"); return 0;
}

HEdge Mesh::hefromev2(Edge e, Vertex v) const
{
    if (vertex1(e)==v) return herep(e)->sym;
    if (vertex2(e)==v) return herep(e);
    MESHassertnever("Vertex not on Edge"); return 0;
}

HEdge Mesh::hefromef(Edge e, Face f) const
{
    if (face1(e)==f) return herep(e);
    if (face2(e)==f) return herep(e)->sym;
    MESHassertnever("Face not adjacent to edge"); return 0;
}

HEdge Mesh::getHEdge(Vertex v, Face f) const
{
    ForVertexHEdge(*this,v,he) {
        if (he->face==f) { return he; }
    } EndFor;
    MESHassertnever("Face not adjacent to Vertex");
    return 0;
}

HEdge Mesh::queryHEdge(Vertex v1, Vertex v2) const
{
    ForStack(v1->sthe,HEdge,he) {
        if (he->vert==v2) return he;
    } EndFor;
    return 0;
}

// Must have:           vert
// Defines:             sym, edge
// Must define elsewhere: next, prev, face, edge
void Mesh::enterHEdge(HEdge he, Vertex v1)
{
    Vertex v2=he->vert;
    Stack<HEdge>& stack=v1->sthe;
    if (sdebug>=1) {
        valid(v1); valid(v2);
        MESHassertx(!queryHEdge(v1,v2) && !stack.contains(he));
    }
    stack.push(he);
    HEdge hes=he->sym=queryHEdge(v2,v1);
    if (hes) {
        MESHassertx(!hes->sym);
        hes->sym=he;
        Edge e=hes->edge;
        he->edge=e;
        if (he->vert->id>hes->vert->id) e->herep=he;
    } else {
        nedges++;
        Edge e=new MEdge;
        e->id = edgenum++;
        id2edge.enter(e->id,e);
        e->herep=he;
        e->flags=0;
        he->edge=e;
    }
}

// Must have:           vert, sym, edge
// Must do later:       delete he
// Note: must be careful: could have he==herep(he->face) ! :
//   destroyFace() : deletes face -> ok.
//   substituteFaceVertex() : reintroduces edge -> ok.
void Mesh::removeHEdge(HEdge he, Vertex v1)
{
    Vertex v2=he->vert;
    Stack<HEdge>& stack=v1->sthe;
    if (sdebug>=1) {
        valid(v1); valid(v2);
        MESHassertx(he==queryHEdge(v1,v2));
        MESHassertx(he->sym==queryHEdge(v2,v1));
    }
    Edge e=he->edge;
    HEdge hes=he->sym;
    if (hes) {
        hes->sym=0;
        if (e->herep==he) e->herep=hes;
    } else {
        nedges--;
        // e->herep=0;             // optional
        MESHassertx(id2edge.remove(e->id));

        delete e;
    }
    // he->edge=0;                 // optional
    MESHassertx(stack.remove(he)); // slow, shucks
}

void Mesh::create_bogus_hedges(Array<HEdge>& ar_he)
{
    ForIndex(i,ar_he.num()) {
        HEdge& he=ar_he[i];
        if (isBoundary(he)) {
            HEdge heo=he;
            Vertex v1=heo->vert;
            he=new MHEdge;
            he->vert=heo->prev->vert;
            he->prev=0;
            he->next=0;
            he->face=(Face)v1;  // real HACK!!
            enterHEdge(he,v1);
        } else {
            he=0;
        }
    } EndFor;
}

void Mesh::remove_bogus_hedges(const Array<HEdge>& ar_he)
{
    ForArray(ar_he,HEdge,he) {
        if (!he) continue;
        Vertex v1=(Vertex)he->face; // real HACK!!
        removeHEdge(he,v1);
        delete he;
    } EndFor;
}

//*** ITERATORS
//** Mesh
//* MeshHEdgeIter

static const Stack<HEdge> StackHEdgeEmpty;

MeshHEdgeIter::MeshHEdgeIter(const Mesh& m)
: mi(m.id2vertex),
  si(*(mi?&mi.value()->sthe:&StackHEdgeEmpty)) { }

//* MeshEdgeIter

MeshEdgeIter::MeshEdgeIter(const Mesh& m)
: it(m) { }

//* MeshOrderedVertexIter

MeshOrderedVertexIter::MeshOrderedVertexIter(const Mesh& m)
{
    ForMeshVertex(m,v) {
        int vi=m.vertexid(v);
        MESHassertw1(vi<(1<<23));   // precision available in float
        pq.enterUnsorted(v,vi);
    } EndFor;
    pq.sort();
}

//* MeshOrderedFaceIter

MeshOrderedFaceIter::MeshOrderedFaceIter(const Mesh& m)
{
    ForMeshFace(m,f) {
        int fi=m.faceid(f);
        MESHassertw1(fi<(1<<23));   // precision available in float
        pq.enterUnsorted(f,fi);
    } EndFor;
    pq.sort();
}

MeshOrderedEdgeIter::MeshOrderedEdgeIter(const Mesh& m)
{
    ForMeshEdge(m,e) {
        int fi=m.edgeid(e);
        MESHassertw1(fi<(1<<23));   // precision available in float
        pq.enterUnsorted(e,fi);
    } EndFor;
    pq.sort();
}

//** Vertex
//* VertexHEdgeIter is inline

//* VertexVertexIter

VertexVertexIter::VertexVertexIter(const Mesh& m, Vertex v)
: it(m,v), extrav(0) { }

Vertex VertexVertexIter::next()
{
    if (extrav) { Vertex vr=extrav; extrav=0; return vr; }
    HEdge he=it.next();
    if (!he) return 0;
    if (!he->next->sym) extrav=he->next->vert;
    return he->prev->vert;
}

//* VertexFaceIter

VertexFaceIter::VertexFaceIter(const Mesh& m, Vertex v)
: it(m,v) { }

Face VertexFaceIter::next()
{
    HEdge he=it.next();
    return he ? he->face : 0;
}

//* VertexEdgeIter

VertexEdgeIter::VertexEdgeIter(const Mesh& m, Vertex v)
: it(m,v), extrae(0) 
{ 
}

Edge VertexEdgeIter::next()
{
    if (extrae) { Edge er=extrae; extrae=0; return er; }
    HEdge he=it.next();
    if (!he) return 0;

    if (!he->next->sym) extrae=he->next->edge;
    return he->edge;
}

//** Face
//* FaceHEdge is inline

//* FaceVertexIter is inline

//* FaceFaceIter

FaceFaceIter::FaceFaceIter(const Mesh& m, Face f)
: it(m,f) { }

Face FaceFaceIter::next()
{
    for (;;) {
        HEdge he=it.next();
        if (!he) return 0;
        if (he->sym) return he->sym->face;
    }
}

//* FaceEdgeIter

FaceEdgeIter::FaceEdgeIter(const Mesh& m, Face f)
: it(m,f) { }

Edge FaceEdgeIter::next()
{
    HEdge he=it.next();
    return he ? he->edge : 0;
}

//** Edge
//* EdgeVertexIter

EdgeVertexIter::EdgeVertexIter(const Mesh& m, Edge e)
: he(m.herep(e)), i(0) { }

Vertex EdgeVertexIter::next()
{
    return i>=2 ? 0 : !i++ ? he->vert : he->prev->vert;
}

//* EdgeFaceIter


EdgeFaceIter::EdgeFaceIter(const Mesh& m, Edge e)
: he(m.herep(e)), i(0) { }

Face EdgeFaceIter::next()
{
    return i>=2 ? 0 : !i++ ? he->face : he->sym ? he->sym->face : 0;
}

//** VertexCcw
//* VertexCcwHEdge

VertexCcwHEdgeIter::VertexCcwHEdgeIter(const Mesh& m, Vertex vp)
: mesh(m), v(vp), hec(0) { }

HEdge VertexCcwHEdgeIter::next()
{
    if (!hec) return hec=hef=mesh.mostClwHEdge(v);
    hec=mesh.ccwHEdge(hec);
    if (hec==hef) hec=0;
    return hec;
}

//* VertexCcwVertex

VertexCcwVertexIter::VertexCcwVertexIter(const Mesh& m, Vertex v)
: it(m,v), extrav(0) { }

Vertex VertexCcwVertexIter::next()
{
    if (extrav) { Vertex vr=extrav; extrav=0; return vr; }
    HEdge he=it.next();
    if (!he) return 0;
    if (!he->sym) extrav=he->prev->vert;
    return he->next->vert;
}

//* VertexCcwFace

VertexCcwFaceIter::VertexCcwFaceIter(const Mesh& m, Vertex v)
: it(m,v) { }

Face VertexCcwFaceIter::next()
{
    HEdge he=it.next();
    if (!he) return 0;
    return he->face;
}


int Mesh::substituteFaceVertex(Face f, Vertex vold, Vertex vnew)
{
    /*
	MESHassertx(vold!=vnew);
	if (sdebug>=1) valid(vold),valid(vnew);
	ForFaceVertex(*this,f,v) {
		if (v==vnew) return 0;
	} EndFor;
	Edge e = (clwEdge(m, vold,f);
	MESHassertx(e->vert==vold);	// optional
	if (lookupEdge(e->prev->vert,vnew)) return 0;
	if (lookupEdge(vnew,e->next->vert)) return 0;
	// all systems go
	removeHEdge(e);
	removeHEdge(e->next);
	e->vert=vnew;
	enterHEdge(e);
	enterHEdge(e->next);
     */
     MESHassertx(0);
	return 1;
}

