/* ------------------------------------------------------------
  A non uniform rational bspline surface
  Operations implemented:
   ------------------------------------------------------------ */

#include "StdAfx.H"
#include <utils/Srf_NurbTC.H>

/* --------------- protected Routines -------------- */
/* -------------------------- Private Routines ---------------- */


/* ------------- Public Routines ----------------------------- */

void SRFNurb3TC_CrvEval( const SRFNub3TC &in_srf, const R2Pt  &in_pt,
                         R2Vec &out_vecP1, R2Vec &out_vecP2,
                         R3Vec &out_vecNorm,
                         R3Vec &out_vecS1, R3Vec &out_vecS2,
                         double & out_dCrv1, double & out_dCrv2,
                         double & out_dArea )
{
    R3Pt pt;
    R3Vec vecDs, vecDt, vecDss, vecDtt, vecDst;
    in_srf.AllEval( in_pt, pt, vecDs, vecDt, vecDss, vecDtt, vecDst );

    const double dE = Dot( vecDs, vecDs );
    const double dF = Dot( vecDs, vecDt );
    const double dG = Dot( vecDt, vecDt );

    out_vecNorm = Cross( vecDs, vecDt );

    const double de = Dot( vecDss, out_vecNorm );
    const double df = Dot( vecDst, out_vecNorm );
    const double dg = Dot( vecDtt, out_vecNorm );

    out_dArea = sqrt( dE * dG - dF * dF );

    const double dA = (dF * dg - dG * df);
    const double dB = (dE * dg - dG * de);
    const double dC = (dE * df - dF * de);

    const double dRoot = dB * dB - 4.0 * dA * dC;
    if ( dRoot < 0 ) {
        ASSERT(FALSE);
    } 

    const double dL1 = ( -dB + sqrt(dRoot) ) / (2.0 * dA);
    const double dL2 = ( -dB - sqrt(dRoot) ) / (2.0 * dA);

    out_dCrv1 = (de + df * dL1) / (dE + dF * dL1);
    out_dCrv2 = (de + df * dL2) / (dE + dF * dL2);

    const double dAng1 = atan( dL1 );
    const double dAng2 = atan( dL2 );

    const double dC1 = cos( dAng1 );
    const double dS1 = sin( dAng1 );
    const double dC2 = cos( dAng2 );
    const double dS2 = sin( dAng2 );
    out_vecP1 = R2Vec( dC1, dS1 );
    out_vecP2 = R2Vec( dC2, dS2 );

    out_vecS1 = vecDs * dC1 + vecDt * dS1;
    out_vecS2 = vecDs * dC2 + vecDt * dS2;

    //
    const double dDot = Dot( UnitSafe( out_vecS1 ), UnitSafe( out_vecS2 ) );
    ASSERT( RNIsZero( dDot, 1e-6 ) );
}


template<>
WINbool SRFNurbTC<R2Pt, R2Pt, R2Vec >::Test()
{
   SRFNub2TC srf(R2Pt_i(1,1), R2Pt_i(4, 4) );

   for (int u = 0; u < 4; u++)
     for (int v = 0; v < 4; v++)
       srf.Pt(u,v) = R2Pt (u,v);

   ASSERT(srf.InDim() == 2);
   ASSERT(srf.OutDim() == 2);

   ASSERT(srf.Num_pts() == 16);
   R2Pt out_pt = srf( R2Pt( 0.5, 0.5) );
   ASSERT(out_pt[0] == out_pt[1]);

   out_pt = srf.Closest_pt(R2Pt (2, 2), out_pt );

   SRFNub2TC srf2(srf);
   ASSERT(srf2 == srf);
   srf2.Pt(1, 0) = R2Pt (0.3, 0.2);

   ofstream out("RNtest.txt", ios::out);
   srf.Write(out);
   out.close();
   ifstream in("RNtest.txt", ios::in);
   VERIFY(srf2.Read(in));
   in.close();
   VERIFY(srf == srf2);


   in.open("RNtest.txt", ios::in);
   SRFNub2TC srf4(in, (WINbool) FALSE);
   in.close();
   VERIFY(srf == srf4);

   srf2.Print();
   srf2.Pt(1, 2) = R2Pt (0.3, 0.2);

   WINWriteBinary("RNtest.bin", out);
   srf.Write_binary(out);
   out.close();
   WINReadBinary("RNtest.bin", in);
   VERIFY(srf2.Read_binary(in));
   in.close();
   VERIFY(srf == srf2);

   WINReadBinary("RNtest.bin", in);
   SRFNub2TC srf3(in, (WINbool) TRUE);
   in.close();
   VERIFY(srf == srf3);

   return FALSE;
}

