/* Copyright 1991, Brown Computer Graphics Group.  All Rights Reserved. */

/* -------------------------------------------------------------------------
Mapping from a point on the plane to barycoords and back
   ------------------------------------------------------------------------- */

#include "StdAfx.H"
#include <utils/Utils_Barys.H>

/* -------------------------- protected Routines --------------------------- */

/* -------------------------- Private Routines ----------------------------- */

/* -------------------------- public routines  ----------------------------- */

/* -------------------------------------------------------------------------
 * DESCR   :	Test
 * ------------------------------------------------------------------------- */

WINbool UTILSBarys::Test()
{
    /// General case
    R2Polygon p(3);
    p[0][0] = 0.0;
    p[0][1] = 0.0;
    p[1][0] = 0.3;
    p[1][1] = 0.5;
    p[2][0] = 1.0;
    p[2][1] = 0.1;
    
    const R2Pt pt(0.1, 0.1);
    
    UTILSBarys barys(p);
    R3Pt bary = barys.Inverse(pt);
    ASSERT(bary[0] > 0 && bary[1] > 0);
    ASSERT( RNIsZero( 1.0 - ( bary[0] + bary[1] + bary[2] ) ) );
    
    ASSERT( ApproxEqual( barys(bary), pt ));

    R3Pt bary2 = UTILSBarys::BaryCoords( p, pt );
    ASSERT( ApproxEqual( bary2, bary ) );

    ASSERT( ApproxEqual( pt, UTILSBarys::ReconstructBaryCoords( p, bary2 ) ) );

    bary = barys.Inverse( R2Pt(-0.2, 0.2 ) );
    bary2 = UTILSBarys::BaryCoords( p, R2Pt(-0.2, 0.2 ) );
    ASSERT( ApproxEqual( bary2, bary ) );
    ASSERT( ApproxEqual( barys(bary), R2Pt(-0.2, 0.2 ) ));

    // Co-linear case
    p[0][0] = 0.0;
    p[0][1] = 0.0;
    p[1][0] = 0.3;
    p[1][1] = 0.3;
    p[2][0] = 0.5;
    p[2][1] = 0.5;
    
    barys = p;
    bary = barys.Inverse(pt);
    ASSERT( RNIsZero( bary[1] ) );
    ASSERT( RNIsZero( 1.0 - ( bary[0] + bary[1] + bary[2] ) ) );
    
    ASSERT( ApproxEqual( barys(bary), pt ));

    bary2 = UTILSBarys::BaryCoords( p, pt );
    ASSERT( ApproxEqual( bary2, R3Pt(2.0/3.0,1.0/3.0,0.0) ) );

    ASSERT( ApproxEqual( pt, UTILSBarys::ReconstructBaryCoords( p, bary2 ) ) );

    // All same case
    p[0] = p[1] = p[2] = pt;

    barys = p;
    bary = barys.Inverse(pt);
    ASSERT( RNIsZero( 1.0 - ( bary[0] + bary[1] + bary[2] ) ) );
    
    ASSERT(barys(bary) == pt);

    bary2 = UTILSBarys::BaryCoords( p, pt );
    ASSERT( ApproxEqual( bary2, bary ) );

    ASSERT( ApproxEqual( pt, UTILSBarys::ReconstructBaryCoords( p, bary2 ) ) );

    // Denom1 zero case
    p[0][0] = 0.0;
    p[0][1] = 0.0;
    p[1][0] = 0.0;
    p[1][1] = 0.3;
    p[2][0] = 0.5;
    p[2][1] = 0.0;

    R2Polygon p2(3);
    p2[0] = p[1];
    p2[1] = p[2];
    p2[2] = p[0];

    barys = p2;
    bary = barys.Inverse(pt);
    ASSERT( RNIsZero( 1.0 - ( bary[0] + bary[1] + bary[2] ) ) );

    ASSERT( p.Inside( pt ) == TRUE );
    ASSERT( ApproxEqual( barys(bary), pt ));

    // Denom 1 zero case
    barys = p;
    bary = barys.Inverse(pt);
    ASSERT( RNIsZero( 1.0 - ( bary[0] + bary[1] + bary[2] ) ) );
    
    ASSERT( ApproxEqual( barys(bary), pt ));
    
    
    // Wedge case
    p[1] = p[0];
    p[2][0] = 0.5;
    p[2][1] = 0.5;

    
    barys = p;
    bary = barys.Inverse(pt);
    ASSERT( RNIsZero( 1.0 - ( bary[0] + bary[1] + bary[2] ) ) );
    
    ASSERT( ApproxEqual( barys(bary), pt ));
    
    bary2 = UTILSBarys::BaryCoords( p, pt );
    ASSERT( ApproxEqual( bary2, bary ) );

    ASSERT( ApproxEqual( pt, UTILSBarys::ReconstructBaryCoords( p, bary2 ) ) );


    // Test spherical bary coords
    Array<R3Vec> avec(3);
    avec[0] = R3Vec(0,0,1);
    avec[1] = R3Vec(0,1,0);
    avec[2] = R3Vec(1,0,0);

    bary = UTILSBarys::SphericalBaryCoords( avec, R3Vec(0,0,1) );
    ASSERT( RNApproxEqual( bary[0], 1.0 ) );
    ASSERT( RNIsZero( bary[1] ) );
    ASSERT( RNIsZero( bary[2] ) );

    R3Vec vecRes = UTILSBarys::ReconstructSphericalBaryCoords( avec, bary );
    ASSERT( ApproxEqual( vecRes, R3Vec(0,0,1) ) );

    R3Vec vecTry( UnitSafe( R3Vec(1.0/3.0,1.0/3.0,1.0/3.0) ) );
    bary = UTILSBarys::SphericalBaryCoords( avec, vecTry );
    ASSERT( RNApproxEqual( bary[0], 1.0/3.0 ) );
    ASSERT( RNApproxEqual( bary[1], 1.0/3.0 ) );
    ASSERT( RNApproxEqual( bary[2], 1.0/3.0 ) );

    vecRes = UTILSBarys::ReconstructSphericalBaryCoords( avec, bary );
    ASSERT( ApproxEqual( vecRes, vecTry ) );


    vecTry = UnitSafe( R3Vec(1.0/2.0,1.0/3.0,1.0/3.0) );
    bary = UTILSBarys::SphericalBaryCoords( avec, vecTry );

    vecRes = UTILSBarys::ReconstructSphericalBaryCoords( avec, bary );
    ASSERT( ApproxEqual( vecRes, vecTry ) );


    vecTry = UnitSafe( R3Vec(1.0/2.0,1.0/4.0,1.0/5.0) );
    bary = UTILSBarys::SphericalBaryCoords( avec, vecTry );

    vecRes = UTILSBarys::ReconstructSphericalBaryCoords( avec, bary );
    ASSERT( ApproxEqual( vecRes, vecTry ) );


    vecTry = UnitSafe( R3Vec(-1.0/2.0,1.0/3.0,1.0/3.0) );
    bary = UTILSBarys::SphericalBaryCoords( avec, vecTry );

    vecRes = UTILSBarys::ReconstructSphericalBaryCoords( avec, bary );
    ASSERT( ApproxEqual( vecRes, vecTry ) );

    avec[0] = UnitSafe( R3Vec(0,0,1) );
    avec[1] = UnitSafe( R3Vec(0,.8,1.0) );
    avec[2] = UnitSafe( R3Vec(0.6,0,1.0) );

    vecTry = UnitSafe( R3Vec(0.4,0.5,2.0) );
    bary = UTILSBarys::SphericalBaryCoords( avec, vecTry );

    vecRes = UTILSBarys::ReconstructSphericalBaryCoords( avec, bary );
    ASSERT( ApproxEqual( vecRes, vecTry ) );
    return TRUE;
}

/* -------------------------------------------------------------------------
 * DESCR   :	Return bary coords
p = {px, py}
p1 = {p1x, p1y}
p2 = {p2x, p2y}
p3 = {p3x, p3y}

Solve[ p == (1 - a - b) p1 + a p2 + b p3, {a,b}]

         -p1y p3x + p1x p3y + p1y px - p3y px - p1x py + p3x py
a -> -   --------------------------------------------------------
         -p1y p2x + p1x p2y + p1y p3x - p2y p3x - p1x p3y + p2x p3y


         -p1y p2x + p1x p2y + p1y px - p2y px - p1x py + p2x py
b->      -----------------------------------------------------------
         -p1y p2x + p1x p2y + p1y p3x - p2y p3x - p1x p3y + p2x p3y

 * ------------------------------------------------------------------------- */
R3Pt UTILSBarys::Inverse(const R2Pt &in_ptPoly) const
{
    R3Pt out_pt( 1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0 );

    const R2Vec vecE1 = UnitSafe( m_poly[1] - m_poly[0] );
    const R2Vec vecE2 = UnitSafe( m_poly[2] - m_poly[0] );
    
    // Zero Area
    if ( ApproxEqual( m_poly[0], m_poly[1] ) && ApproxEqual( m_poly[0], m_poly[2] ) ) {
        return out_pt;
    }

    WINbool bFlat = FALSE;
    for ( int i = 0; i < 3; i++ ) {
        if ( RNIsZero( Length(m_poly[i] - m_poly.PtWrap(i+1)) ) ) {
            bFlat = TRUE;
        }
    }
    // Co-linear
    if ( bFlat == TRUE || RNApproxEqual( fabs( Dot( vecE1, vecE2 ) ), 1.0 ) ) {
        Array<double> adLen(3);
        for (int i = 0; i < 3; i++)
            adLen[i] = LengthSq( m_poly[i] - m_poly.PtWrap(i+1) );

        int iWhich = adLen.max_i( adLen );
        R2Line_seg seg( m_poly[iWhich], m_poly.PtWrap(iWhich+1) );
                
        /// t between 0 and 1
        double t = seg.Projected_dist_on_seg( in_ptPoly );
                
        out_pt[iWhich] = 1.0 - t;
        out_pt[(iWhich+1)%3] = t;
        out_pt[(iWhich+2)%3] = 0;

        return out_pt;
    }

    const double p1x = m_poly[0][0];
    const double p1y = m_poly[0][1];
    const double p2x = m_poly[1][0];
    const double p2y = m_poly[1][1];
    const double p3x = m_poly[2][0];
    const double p3y = m_poly[2][1];

    const double px = in_ptPoly[0];
    const double py = in_ptPoly[1];

    double dNumerA = -p1y * p3x + p1x * p3y + p1y * px - p3y * px - p1x * py + p3x * py;
    double dNumerB = -p1y * p2x + p1x * p2y + p1y * px - p2y * px - p1x * py + p2x * py;

    double dDenom = -p1y * p2x + p1x * p2y + p1y * p3x - p2y * p3x - p1x * p3y + p2x * p3y;

    ASSERT( ! RNIsZero( dDenom ) );

    out_pt[1] = - dNumerA / dDenom;
    out_pt[2] = dNumerB / dDenom;
    out_pt[0] = 1.0 - out_pt[1] - out_pt[2];

    return out_pt;
}


R3Pt UTILSBarys::BaryCoords( const R2Polygon &in_poly, const R2Pt &in_pt )
{
    R3Pt ptBary( -1, -1, -1 );

    const R2Vec v01 = in_poly[1] - in_poly[0];
    const R2Vec v12 = in_poly[2] - in_poly[1];
    const R2Vec v20 = in_poly[0] - in_poly[2];

    const R2Vec vp0 = in_pt - in_poly[0];
    const R2Vec vp1 = in_pt - in_poly[1];
    const R2Vec vp2 = in_pt - in_poly[2];

    const double dA0 = 0.5 * ( v01[1] * vp0[0] - v01[0] * vp0[1] );
    const double dA1 = 0.5 * ( v12[1] * vp1[0] - v12[0] * vp1[1] );
    const double dA2 = 0.5 * ( v20[1] * vp2[0] - v20[0] * vp2[1] );

    const double dA = 0.5 * ( v01[1] * v12[0] - v01[0] * v12[1] );

    if ( RNIsZero( dA, 1e-12 ) ) {
        ptBary[0] = dA1;
        ptBary[1] = dA2;
        ptBary[2] = dA0;
    } else {
        ptBary[0] = dA1 / dA;
        ptBary[1] = dA2 / dA;
        ptBary[2] = dA0 / dA;
    }
    double dSum = ptBary[0] + ptBary[1] + ptBary[2];
    if(!RNApproxEqual( dSum, 1.0, 1e-2 )){
        if(!RNIsZero( dSum, 1e-12 )){
            for(int i = 0; i<3; i++){
                if ( RNApproxEqual( ptBary[i], 0.0, 1e-4 ) )
                    ptBary[i] = 1e-12;
                else
                    ptBary[i] = ptBary[i]/dSum;
            }
        }
        else{
            for(int i = 0; i<3; i++){
                ptBary[i] = 0.1/0.3;
            }
        }
    }
    dSum = ptBary[0] + ptBary[1] + ptBary[2];
    ASSERT(RNApproxEqual( dSum, 1.0, 1e-2 ));
    return ptBary;
}

Array<double> UTILSBarys::BaryCoordsFull( const R2Polygon &in_poly, const R2Pt &in_pt ){    Array<double> adBary( in_poly.Num_pts() );    const double dAvg = 1.0 / (double) adBary.num();    for ( int i = 0; i < adBary.num(); i++ )        adBary[i] = dAvg;    if ( in_poly.Num_pts() == 3 ) {        const R3Pt ptBary = BaryCoords( in_poly, in_pt );        for ( int i = 0; i < 3; i++ )            adBary[i] = ptBary[i];    } else {        const R2Pt ptCentroid = in_poly.Centroid();        R2Polygon tri(3);        tri[2] = ptCentroid;        for ( int i = 0; i < in_poly.Num_pts(); i++ ) {            tri[0] = in_poly[i];            tri[1] = in_poly.PtWrap(i+1);            const R3Pt ptBary = BaryCoords( tri, in_pt );            if ( ptBary[0] >= 0.0 && ptBary[0] <= 1.0 &&                  ptBary[1] >= 0.0 && ptBary[1] <= 1.0 &&                  ptBary[2] >= 0.0 && ptBary[2] <= 1.0 ) {                for ( int j = 0; j < adBary.num(); j++ ) {                    adBary[j] = dAvg * ptBary[2];                }                adBary[i] += ptBary[0];                adBary.wrap(i+1) += ptBary[1];                break;            }        }    }    return adBary;}

///
R2Pt UTILSBarys::ReconstructBaryCoords( const R2Polygon &in_poly, const R3Pt &in_ptBary )
{
    R2Pt ptRet(0.0, 0.0);

    ASSERT( in_poly.Num_pts() == 3 );

    for ( int i = 0; i < 3; i++ ) {
        ptRet[0] += in_poly[i][0] * in_ptBary[i]; 
        ptRet[1] += in_poly[i][1] * in_ptBary[i]; 
    }
    return ptRet;
}

R2Pt UTILSBarys::ReconstructBaryCoords( const R2Polygon &in_poly, const Array<double> &in_adBary )
{
    R2Pt ptRet(0.0, 0.0);

    ASSERT( in_poly.Num_pts() == in_adBary.num() );

    for ( int i = 0; i < in_adBary.num(); i++ ) {
        ptRet[0] += in_poly[i][0] * in_adBary[i]; 
        ptRet[1] += in_poly[i][1] * in_adBary[i]; 
    }
    return ptRet;
}

///Spherical bary coords
R3Pt UTILSBarys::SphericalBaryCoords( const Array<R3Vec> &in_avecTri, const R3Vec &in_vec )
{
    R3Vec avecVs[3], avecVt[3], avecE1[3], avecE2[3], vecVs, vecVt;
    R3Pt  ptBary(1.0/3.0, 1.0/3.0, 1.0/3.0);

    ASSERT( in_avecTri.num() == 3 );
    for ( int i = 0; i < 3; i++ ) {
        ASSERT( RNApproxEqual( Length( in_avecTri[i] ), 1.0 ) );

        // Two vectors from this point to the other two vertices
        avecE1[i] = UnitSafe( in_avecTri.wrap(i+1) - in_avecTri[i] );
        avecE2[i] = UnitSafe( in_avecTri.wrap(i-1) - in_avecTri[i] );

        if ( RNIsZero( Length( avecE1[i] ) ) ) {
            return ptBary;
        }

        // perpendicular to normal 
        avecVt[i] = UnitSafe( Cross( in_avecTri[i], avecE1[i] ) );
        // perpendicular to normal and vecVt
        avecVs[i] = UnitSafe( Cross( avecVt[i], in_avecTri[i] ) );
    }
    const R3Vec vecDir = Cross( avecE1[0], avecE2[0] );

    ASSERT( RNApproxEqual( Length( in_vec ), 1.0 ) );
    vecVs = UnitSafe( in_vec.Perpendicular() );
    vecVt = Cross( vecVs, in_vec );

    // Area of entire triangle
    double dArea = -M_PI;
    for ( FORINT i = 0; i < 3; i++ ) {
        const R2Vec vec1 = UnitSafe( R2Vec( Dot( avecVs[i], avecE1[i] ), Dot( avecVt[i], avecE1[i] ) ) );
        const R2Vec vec2 = UnitSafe( R2Vec( Dot( avecVs[i], avecE2[i] ), Dot( avecVt[i], avecE2[i] ) ) );
        const double dDot = Dot(vec1, vec2);
        const double dAngle = acos( dDot );
        dArea += dAngle;
    }
    if ( RNIsZero( dArea ) )
        return ptBary;

    // Area of vec with triangle
    Array<R3Vec> avec(3), avecVsArea(3), avecVtArea(3);
    for ( FORINT i = 0; i < 3; i++ ) {
        avec[0] = in_vec;
        avec[1] = in_avecTri.wrap(i+1);
        avec[2] = in_avecTri.wrap(i-1);
        avecVsArea[0] = vecVs;
        avecVsArea[1] = avecVs[ (i+1)%3 ];
        avecVsArea[2] = avecVs[ (i+2)%3 ];
        avecVtArea[0] = vecVt;
        avecVtArea[1] = avecVt[ (i+1)%3 ];
        avecVtArea[2] = avecVt[ (i+2)%3 ];

        double dAreaMe = -M_PI;
        double dDir = 1.0;
        for ( int j = 0; j < 3; j++ ) {
            const R3Vec vecE1 = UnitSafe( avec.wrap(j+1) - avec[j] );
            const R3Vec vecE2 = UnitSafe( avec.wrap(j-1) - avec[j] );
            const R3Vec vecMyDir = Cross( vecE1, vecE2 );
            if ( Dot( vecMyDir, vecDir ) < -RNEpsilon_d ) {
                dDir = -1.0;
            } else if ( Dot( vecMyDir, vecDir ) > RNEpsilon_d ) {
                dDir = 1.0;
            }

            // Project e1 and e2 onto tangent plane
            const R2Vec vec1 = UnitSafe( R2Vec( Dot( avecVsArea[j], vecE1 ), Dot( avecVtArea[j], vecE1 ) ) );
            const R2Vec vec2 = UnitSafe( R2Vec( Dot( avecVsArea[j], vecE2 ), Dot( avecVtArea[j], vecE2 ) ) );
            const double dDot = Dot(vec1, vec2);
            const double dAngle = acos( dDot );
            dAreaMe += dAngle;
        }
        ptBary[i] = dDir * dAreaMe / dArea;
    }

    return ptBary;
}

///
R3Vec UTILSBarys::ReconstructSphericalBaryCoords( const Array<R3Vec> &in_avecTri, const R3Pt &in_ptBary )
{
    R3Vec vec(0,0,0);
    for ( int i = 0; i < 3; i++ ) {
        for ( int j = 0; j < 3; j++ )
            vec[j] += in_avecTri[i][j] * in_ptBary[i];
    }
    return UnitSafe( vec );
}

