/* Copyright 1991, Brown Computer Graphics Group.  All Rights Reserved. */

/* -------------------------------------------------------------------------
Mapping from quadrilaterals to unit squares
   ------------------------------------------------------------------------- */
#include "StdAfx.H"

#include <utils/Utils_Polygon_mapping.H>

/* -------------------------- protected Routines --------------------------- */

/* -------------------------- Private Routines ----------------------------- */
R3Matrix &UTILSPolygon::MakeMatrix(const R2Polygon &in_poly)
{
    static R3Matrix Q, Qinv;
    static R3Pt Q4(0,0,1);
    static R3Pt lambda(0,0,1);
    
    ASSERT(in_poly.Num_pts() == 4);
    
    // Matrix taking in_poly to proj. points
    // Computing lambda, mu and nu (Q * lambda = Q4)
    int r;
    for ( r = 0; r < 3; r++)
        Q(2, r) = 1;
    
    for (r = 0; r < 3; r++)
        for (int c = 0; c < 2; c++) {
            Q(c,r) = in_poly[r][c];
        }

    WINbool bSuc = FALSE;
    Qinv = Q.Inverse( bSuc );
    
    Q4[0] = in_poly[3][0];
    Q4[1] = in_poly[3][1];
    Q4[2] = 1;
    
    lambda = Qinv * Q4;
    
    // Computing matrix to take 100 010 001 111 to in_poly
    static R3Matrix T2;
    for (r = 0; r < 3; r++) {
        for (int c = 0; c < 2; c++)
            T2(c, r) = lambda[r] * in_poly[r][c];
        T2(2, r) = lambda[r];
    }

    return T2;
}


R3Matrix
UTILSPolygon::Unit_to_polygon(const R2Polygon &in_poly)
{
    static R3Matrix T1, T1inv;
    static WINbool T1_initd = FALSE;
    static R2Polygon unit_square(4);
    static R3Polygon poly(4);
    static double dDeterminant = 0;
    
    if (T1_initd == FALSE) {
        T1_initd = TRUE;
        unit_square[1][0] = 1;
        unit_square[2][0] = unit_square[2][1] = 1;
        unit_square[3][1] = 1;
        poly[0] = R3Pt(1,0,0);
        poly[1] = R3Pt(0,1,0);
        poly[2] = R3Pt(0,0,1);
        poly[3] = R3Pt(1,1,1);
        
        T1 = MakeMatrix(unit_square); // takes poly to unit square
        WINbool bSuc = FALSE;
        T1inv = T1.Inverse( bSuc );

        R3Matrix matIdent = T1 * T1inv;
        R3Matrix matIdent2 = T1inv * T1;

        ASSERT( bSuc == TRUE );
    }
    static R3Matrix T2inv, T2;
    T2 = MakeMatrix(in_poly);  // Takes poly to in_poly
    WINbool bSuc = FALSE;
    T2inv = T2.Inverse( bSuc );
    //ASSERT( bSuc == TRUE );
    
    R3Matrix matIdent = T2 * T2inv;

    // Matrix to take a point on the unit square to in_poly
    R3Matrix T = T2 * T1inv;
    
    return T;
}


R3Matrix
UTILSPolygon::Polygon_to_unit(const R2Polygon &in_poly)
{
   static R3Matrix T, Tinv;
   T = Unit_to_polygon(in_poly);

   WINbool bSuc = FALSE;
   Tinv = T.Inverse( bSuc );
   //ASSERT( bSuc == TRUE );

   return Tinv;
}



R3Matrix
UTILSPolygon::Polygon_to_polygon(const R2Polygon &in_polyFrom,
                                 const R2Polygon &in_polyTo
   )
{
    R3Matrix m1 = Polygon_to_unit(in_polyFrom);
    R3Matrix m2 = Unit_to_polygon(in_polyTo);
    
    return m2 * m1;
}

/* -------------------------------------------------------------------------
 * DESCR   :	Map the 2D vector through the projective transform
 * ------------------------------------------------------------------------- */

R2Vec 
UTILSPolygon::Map(const R3Matrix &m, const R2Vec &v)
{
   R3Vec temp(0,0,1);
   R2Vec result(0,0);
   
   temp[0] = v[0];
   temp[1] = v[1];
   temp[2] = 1;

   temp = m * temp;

   ASSERT(!RNIsZero(temp[2]));

   result[0] = temp[0] / temp[2];
   result[1] = temp[1] / temp[2];

   return result;
}


R2Pt
UTILSPolygon::Map(const R3Matrix &m, const R2Pt &v)
{
   R3Pt temp(0,0,1);
   R2Pt res(0,0);
   
   for (int j = 0; j < 2; j++)
   temp[j] = v[j];
   temp[1] = v[1];
   temp[2] = 1;

   temp = m * temp;

   ASSERT(!RNIsZero(temp[2]));
   res[0] = temp[0] / temp[2];
   res[1] = temp[1] / temp[2];

   return res;
}

WINbool UTILSPolygon::Test()
{
   R2Polygon poly(4), poly1(4);
   poly[0][0] = 0.1;
   poly[0][1] = 0.2;
   poly[1][0] = 0.6;
   poly[1][1] = 0.25;
   poly[2][0] = 0.55;
   poly[2][1] = 0.67;
   poly[3][0] = 0.15;
   poly[3][1] = 0.67;

   poly1[0][0] = -0.5;
   poly1[0][1] = -0.5;
   poly1[1][0] = 0.5;
   poly1[1][1] = -0.5;
   poly1[2][0] = 0.5;
   poly1[2][1] = 0.5;
   poly1[3][0] = -0.5;
   poly1[3][1] = 0.5;

   R3Matrix m1 = Polygon_to_unit(poly);
   R3Matrix m3 = Polygon_to_unit(poly1);
   R3Matrix m2 = Polygon_to_polygon(poly, poly);
   R3Matrix m4 = Polygon_to_polygon(poly1, poly);
   
   R2Pt pt = poly1[0];
   R2Vec vec = R2Vec(poly1[1][0], poly1[1][1]);
   R2Pt ptres = Map(m4, pt);
   R2Vec vecres = Map(m4, vec);
   
   ASSERT(ApproxEqual(ptres, poly[0]) && RNApproxEqual(vecres[0], poly[1][0]));

   ptres = Map(m2, pt);
   RNEpsilon_d = 1e-6;
   ASSERT(ApproxEqual(ptres, pt));
   vecres = Map(m2, vec);
   ASSERT(ApproxEqual(vecres, vec));
   return TRUE;
}
